# UPC parsing Python module

# Import statements
import tokenize
import helper
import rules
import state




# Setup the parser state.
# --------------------------------------------------------------------------- #
def new_state(**args):
    """Create new starting state."""
    return state.UPCParserState(**args)

# Required/framing operations
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #

# Divide the line into 3 sets of tokens.
# --------------------------------------------------------------------------- #
def get_tokens(pstate, line, code_tokens, comment_tokens, directive_tokens):
    """Calls all tokenizing operations on a line of the source code."""
    tokens = tokenize.tokenize(line)
    code_tokens.extend(tokens)
    pstate = tokenize.comments(pstate, code_tokens, comment_tokens)
    pstate = tokenize.directives(pstate, code_tokens, directive_tokens)
    return pstate

# Apply all the rules in upc_rules module to the tokens list.
# --------------------------------------------------------------------------- #
def apply_rules(pstate, code_tokens, comment_tokens, directive_tokens):
    """Applies all the parsing rules to the tokens lists."""
    pstate = rules.linenumber(pstate)
    # Apply rules to the code tokens.
    pstate = rules.scopecheck(pstate, code_tokens)
    pstate = rules.typedefcheck(pstate, code_tokens)
    pstate = rules.staticvarcheck(pstate, code_tokens)
    pstate = rules.upc_sharedvarcheck(pstate, code_tokens)
    pstate = rules.upc_sharedptrcheck(pstate, code_tokens)
    pstate = rules.globalvarcheck(pstate, code_tokens)
    pstate = rules.maincheck(pstate, code_tokens)
    pstate = rules.upc_statementcheck(pstate, code_tokens)
    pstate = rules.upc_forallcheck(pstate, code_tokens)
    pstate = rules.upc_identifiercheck(pstate, code_tokens)
    pstate = rules.upc_expressioncheck(pstate, code_tokens)
    # Apply rules to the directive tokens.
    pstate = rules.upc_headerscheck(pstate, directive_tokens)
    return pstate

# Print a line to the output file
# --------------------------------------------------------------------------- #
def writeline(pstate, line, code_tokens, comment_tokens, directive_tokens, outfile):
    """Prints a line of code to the output file."""
    line_modified = pstate.modified()    # Determine whether output should be generated from input line or tokens
    line_scope    = pstate.scope()       # Determine indentation of output from scope count (maybe?)
    outstring = ''
    verbose_text = "modified line generated"
    if line_modified:
        newline_count = 0
        outstring = '    ' * line_scope
        if len(code_tokens) > 0:
            for token in code_tokens:
                outstring = outstring + token + ' '
                if token == '\n': newline_count = newline_count + 1
        if len(directive_tokens) > 0:
            for token in directive_tokens: outstring = outstring + token + ' '
        if len(comment_tokens) > 0:
            for token in comment_tokens: outstring = outstring + token + ' '
        outstring = outstring[0:len(outstring)-1] + '\n'   # Replace last ' ' with a newline
        # Print the old line before the new if the '-r' flag is passed
        if pstate.replicate():
            outstring = '//' + line + outstring
            verbose_text = verbose_text + " from input line " + str(pstate.linenumber())
            pstate.linenumber_out_inc()   # 2 lines of output
        pstate.modified_unset()
        pstate.modified_count_inc()
        pstate.linenumber_out(pstate.linenumber_out() + newline_count)
        pstate.output_append(verbose_text + '.', 1)
    else:
        outstring = line
    outfile.write(outstring)
    helper.print_output(pstate)
    return pstate

