# upc_parse_state.py

# An object class that maintains the state of the parser across lines/rules.
# =========================================================================== #
class UPCParserState:
    # Internal state variables.
    _linenumber = 0
    _linenumber_out = 0
    _scope_counter = 0
    _line_modified = False
    _line_modified_count = 0
    _inside_comment = False
    _inside_directive = False
    # Command line argument state variables.
    _verbosity_level = 0
    _replicate_mod_lines = False
    # List of per-line output strings.
    _verbose_output = []
    
# Constructor methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def __init__(self, verbosity=0, replicate=False, **args):
        """Basic UPC parser state constructor."""
        if isinstance(verbosity, int):  self._verbosity_level = verbosity
        if isinstance(replicate, bool): self._replicate_mod_lines = replicate
    
# Line number methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def linenumber(self, ln = None):
        """Set and return the input file's current line number."""
        if isinstance(ln, int):
            self._linenumber = ln
        return self._linenumber
    
    def linenumber_inc(self):
        """Increment and return the input file's the current line number."""
        self._linenumber = self._linenumber+1
        return self._linenumber
    
    def linenumber_out(self, ln = None):
        """Set and return the output file's current line number."""
        if isinstance(ln, int):
            self._linenumber_out = ln
        return self._linenumber_out
    
    def linenumber_out_inc(self):
        """Increment and return the output file's current line number."""
        self._linenumber_out = self._linenumber_out+1
        return self._linenumber_out
    
# Scope counter methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def scope(self, s = None):
        """Set and return current scope."""
        if isinstance(s, int):
            self._scope_counter = s
        return self._scope_counter
    
    def scope_inc(self):
        """Increment scope level."""
        self._scope_counter = self._scope_counter+1
        return self._scope_counter
    
    def scope_dec(self):
        """Decrement the scope level."""
        self._scope_counter = self._scope_counter-1
        return self._scope_counter
    
# Verbosity level methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def verbosity(self, v = None):
        """Set and return the verbosity level."""
        if isinstance(v, int):
            self._verbosity_level = v
        return self._verbosity_level

# Replicate modified lines methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def replicate(self, r = None):
        """Set and return the flag determining whether modified lines are replicated."""
        if isinstance(r, bool):
            self._replicate_mod_lines = r
        return self._replicate_mod_lines

# Line modified flag methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def modified(self, m = None):
        """Set and return line modified flag."""
        if isinstance(m, bool):
            self._line_modified = m
        return self._line_modified
    
    def modified_set(self):
        """Set modified flag to true."""
        self._line_modified = True
        return self._line_modified
    
    def modified_unset(self):
        """Set modified flag to false."""
        self._line_modified = False
        return self._line_modified
    
    def modified_count(self, mc = None):
        """Set and return the count of the number of lines that have been modified."""
        if isinstance(mc, int):
            self._line_modified_count = mc
        return self._line_modified_count
    
    def modified_count_inc(self):
        """Increment the count of the number of lines that have been modified."""
        self._line_modified_count = self._line_modified_count+1
        return self._line_modified_count
    
# Inside comment flag methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def comment(self, c = None):
        """Set and return inside comment flag."""
        if isinstance(c, bool):
            self._inside_comment = c
        return self._inside_comment
    
    def comment_set(self):
        """Set inside comment flag to true."""
        self._inside_comment = True
        return self._inside_comment
    
    def comment_unset(self):
        """Set inside comment flag to false."""
        self._inside_comment = False
        return self._inside_comment
    
# Inside directive flag methods.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def directive(self, d = None):
        """Set and return inside directive flag."""
        if isinstance(d, bool):
            self._inside_directive = d
        return self._inside_directive
    
    def directive_set(self):
        """Set inside directive flag to true."""
        self._inside_directive = True
        return self._inside_directive
    
    def directive_unset(self):
        """Set inside directive flag to false."""
        self._inside_directive = False
        return self._inside_directive

# Output string list.
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~ #
    def output_append(self, outstring, verbosity = 0):
        """Adds a string to the end of the output list."""
        if self._verbosity_level >= verbosity:
            self._verbose_output.append(outstring)
            
    def output_get(self):
        """Returns the list of output created for the current line."""
        return self._verbose_output

    def output_reset(self):
        """Resets the output list in preparation for a new line."""
        self._verbose_output = []
