/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_SKELETONS_MINI_GTC_GTC_MODULES_H_INCLUDED
#define SSTMAC_SOFTWARE_SKELETONS_MINI_GTC_GTC_MODULES_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/software/api/mpi/sstmac_mpi.h>
#include <sstmac/common/rng.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/sim_parameters.h>

namespace gtc
{

#define SMALL 1e-20
#define BIG 1e20;

  /**
   * Base type for modules that encapsulate variables (a la fortran)
   */
  class gtc_module : public sstmac::ptr_type
  {
  protected:
    gtc_module()
    {
      if (!dbg_.is_real())
        {
          dbg_ = sstmac::debug("<debug> gtc | gtc_module",
              sstmac::logger::blue);
          sstmac::logger::register_logger(&dbg_);
        }
    }

    static sstmac::debug dbg_;

  public:
    typedef boost::intrusive_ptr<gtc_module> ptr;

    virtual std::string
    to_string() const
    {
      return "gtc_module";
    }
  };

  /**
   * Holds all the parameters that come from the config file
   */
  class config_parameters : public gtc_module
  {

  protected:
    config_parameters(sstmac::sim_parameters::ptr params);

  public:
    typedef boost::intrusive_ptr<config_parameters> ptr;

    static ptr
    construct(sstmac::sim_parameters::ptr params)
    {
      return ptr(new config_parameters(params));
    }

    virtual
    ~config_parameters()
    {
    }

    virtual std::string
    to_string() const
    {
      return "gtc::config_parameters";
    }

    // ---- control params ----
    int irun_; // 0 for initial run, any non-zero value for restart
    int mstep_; // # of time steps
    int msnap_; // # of snapshots
    int ndiag_; // do diag when mod(istep,ndiag)=0
    double nonlinear_; // 1.0: nonlinear run, 0.0: linear run
    double paranl_; // 0: adiabatic electron, >1: kinetic electron
    int mode00_; // 1: keep parallel nonlinearity

    // --- run size (both mtheta and mzetamax should be multiples of # of PEs)
    double tstep_; // 1 include (0,0) mode, 0 exclude (0,0) mode
    int micell_; // time step (unit=L_T/v_th), tstep*\omega_transit<0.1
    int mecell_; // # of ions per grid cell
    int mpsi_; // # of electrons per grid cell
    int mthetamax_; // total # of radial grid points
    int mzetamax_; // poloidal grid, even and factors of 2,3,5 for FFT
    int npartdom_; // number of particle domain partitions per tor dom.
    int ncycle_; // subcycle electron

    // ----  run geometry
    double a_; // minor radius, unit=R_0
    double a0_; // inner boundary, unit=a
    double a1_; // outer boundary, unit=a
    double q0_; // q_profile, q=q0 + q1*r/a + q2 (r/a)^2
    double q1_; //
    double q2_; //
    double rc_; // kappa=exp{-[(r-rc)/rw]**6}
    double rw_; // rc in unit of (a1+a0) and rw in unit of (a1-a0)

    // ----  species information
    double aion_; // species isotope #
    double qion_; // charge state
    double aelectron_; //
    double qelectron_; //

    // ----  equilibrium unit: R_0=1, Omega_c=1, B_0=1, m=1, e=1
    double kappati_; // grad_T/T
    double kappate_; //
    double kappan_; // inverse of eta_i, grad_n/grad_T
    bool fixed_Tprofile; // Maintain Temperature profile (0=no, >0 =yes)
    double tite_; // T_i/T_e
    double flow0_; // d phi/dpsi=gyroradius*[flow0+flow1*r/a+
    double flow1_; //                      flow2*(r/a)**2]
    double flow2_; //

    // ----  physical unit
    double r0_; // major radius (unit=cm)
    double b0_; // on-axis vacuum field (unit=gauss)
    double temperature_; // electron temperature (unit=ev)
    double edensity0_; // electron number density (1/cm^3)

    // -----  standard output: use 0 or 6 to terminal and 11 to file 'stdout.out'
    int nbound_; // 0 for periodic, >0 for zero boundary
    double umax_; // unit=v_th, maximum velocity in each direction
    int iload_; // 0: uniform, 1: non-uniform
    double tauii_; // -1.0: no collisions, 1.0: collisions
    bool track_particles_; // true: keep track of some particles
    int nptrack_; // track nptrack particles every time step
    long rng_control_; // controls seed and algorithm for random num. gen.

    // ---- mode diagnostic: 8 modes.
    std::vector<int> nmode_; // n: toroidal mode number
    std::vector<int> mmode_; // m: poloidal mode number

    int nhybrid_;
    int ntracer_;
    int ntracer1_;
    double gyroradius_;
    int numberpe_;
    int mype_;

    bool do_collision_;
    int mtdiag_;


    int idiag1_;
    int idiag2_;


    int limit_vpara_;

    int mflux_;
    int num_mode_;
    int m_poloidal_;



    int nparam_;


    //these fields change over the course of the run!
    int idiag_;
    int irk_;
    int ihybrid_;
  };

  /**
   * The particle_decomp module from setup.F90
   */
  class particle_decomp : public gtc_module
  {
  protected:
    particle_decomp();

  public:

    int ntoroidal_;

    MPI_Comm partd_comm_;
    int nproc_partd_;
    int myrank_partd_;
    MPI_Comm toroidal_comm_;
    int nproc_toroidal_;
    int myrank_toroidal_;
    int left_pe_;
    int right_pe_;
    int toroidal_domain_location_;
    int particle_domain_location_;
    int mzeta_;
    double zetamin_;
    double zetamax_;
    double deltaz_;

  public:
    typedef boost::intrusive_ptr<particle_decomp> ptr;

    static ptr
    construct()
    {
      return ptr(new particle_decomp());
    }

    virtual
    ~particle_decomp()
    {
    }

    virtual std::string
    to_string() const
    {
      return "gtc::particle_decomp";
    }

    void
    setup(config_parameters::ptr params);
  };

  /**
   * The field_array module from module.F90
   */
  class field_array : public gtc_module
  {
  protected:
    field_array(int mmpsi = 192);

  public:
    typedef boost::intrusive_ptr<field_array> ptr;

    static ptr
    construct(int mpsi)
    {
      return ptr(new field_array(mpsi));
    }

    virtual
    ~field_array()
    {
    }

    virtual std::string
    to_string() const
    {
      return "gtc::field_array";
    }

    void
    setup(config_parameters::ptr params, double gyroradius,
        particle_decomp::ptr partdecomp);

    // --- derived things
    int mimax_;
    int memax_;
    int mgrid_;

    double mi_local_;
    double me_local_;

    int mi_;
    int me_;

    int me1_;

    double deltar_;
    std::vector<int> mtheta_;
  };







} /* namespace sstmac */
#endif /* GTC_MODULES_H_ */
