/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <HPCCG/hpccg.h>

#include <sstmac/software/api/hpx/components/hpx_distributing_factory_block.h>
#include <sstmac/software/api/hpx/lcos/promise.h>
#include <sstmac/software/api/hpx/lcos/future_wait.h>
#include <iostream>
#include <fstream>

#include "HPCCG/hpccg_matrix.h"
#include "HPCCG/hpccg_matrix_stubs.h"

#include <sstmac/sstmacro.h>

namespace hpx_hpccg
{
  SSTRegisterApp("hpx_hpccg", hpccg, "parallax version of HPCCG");

  using namespace sstmac;
  using namespace sstmac::sw;
  using namespace hpx;
  using namespace std;



  double
  hpccg::mytimer()
  {
    return now().sec();
  }

  void
  hpccg::consume_params(const sim_parameters::ptr &params)
  {
    int nx = params->get_int_param("hpccg_nx");
    int ny = params->get_int_param("hpccg_ny");
    int nz = params->get_int_param("hpccg_nz");

    int domains = params->get_int_param("hpccg_domains");

    par_ = config_params::construct();
    par_->nx_ = nx;
    par_->ny_ = ny;
    par_->nz_ = nz;
    par_->domains_ = domains;
  }

  void
  hpccg::register_actions()
  {

  }

  void
  hpccg::register_components()
  {
    hpx::globalid dumm;
    threadman_->register_component<matrix> (matrix::component_name(),
        matrix::construct(dumm));

    threadman_->register_component<hpx::components::hpx_distributing_factory> (
        hpx::components::hpx_distributing_factory::component_name(),
        hpx::components::hpx_distributing_factory::construct(dumm));
  }

  void
  hpccg::hpx_main()
  {

    timestamp start = now();

    // -------- problem setup from original ---------- //

    double norm, d;
    int ierr = 0;
    int i, j;
    int ione = 1;
    double t6 = 0.0;

    // Create a distributing factory. This object can be used to create
    // a distributed array of components.
    hpx::components::hpx_distributing_factory_block::ptr factory_points =
        hpx::components::hpx_distributing_factory_block::construct(
            agas_->new_gid(hpx::lva::COMPONENT));

    std::vector<matrix::stub_ptr> mat_stubs =
        factory_points->create_components<matrix::matrix_stubs> (
            matrix::component_name(), par_->domains_);

    SSTMAC_DEBUG << "hpccg: done distributing points \n";

    std::vector<hpx::lcos::promise<int>::ptr> initial_phase;

    for (std::size_t i = 0; i < mat_stubs.size(); i++)
      {
        initial_phase.push_back(mat_stubs[i]->init_async(i, par_, mat_stubs));
      }
    hpx::lcos::wait(initial_phase);

    SSTMAC_DEBUG << "hpccg: done initializing points \n";

    // Transform matrix indices from global to local values.
    // Define number of columns for the local matrix.

    double t1 = mytimer(); // Initialize it (if needed)

    int max_iter = 150;
    double tolerance = 0.0; // Set tolerance to zero to make all runs do max_iter iterations
    std::vector<hpx::lcos::promise<int>::ptr> main_phase;

    for (std::size_t i = 0; i < mat_stubs.size(); i++)
      {
        main_phase.push_back(
            mat_stubs[i]->HPCCG_Solve(max_iter, tolerance));
      }

    hpx::lcos::wait(main_phase);

    // if (ierr)
    //  cerr << "Error in call to CG: " << ierr << ".\n" << endl;

    // ---------------------------------------------------- //


  }

}
