/*
 * hpccg_matrix.h
 *
 *  Created on: Nov 7, 2012
 *      Author: ghendry
 */

#ifndef HPCCG_MATRIX_H_
#define HPCCG_MATRIX_H_

#include "HPCCG/hpccg.h"

#include <sstmac/software/api/hpx/components/hpx_component.h>
#include "HPCCG/buffer_wrapper.h"

#include <sstmac/software/api/hpx/lcos/sync.h>

namespace hpx_hpccg
{

  class hpccg::matrix : public hpx::components::hpx_component
  {

  public:
    class matrix_stubs;

    typedef boost::intrusive_ptr<matrix> ptr;
    typedef boost::intrusive_ptr<matrix_stubs> stub_ptr;

  protected:
    matrix(hpx::globalid gid);

    std::size_t idx_;
    std::vector<stub_ptr> mstubs_;
    config_params::ptr par_;
    HPC_Sparse_Matrix *A_;
    double *b_;
    double *x_;
    double* p_;
    double *xexact_;

    static long bytes_sent_;

    int* global_index_offsets_;
    int* tmp_neighbors_;
    int* tmp_buffer_;
    hpx::lcos::sync::ptr tmp_neighbors_lock_;


    int sends_recvd_;
    int lengths_recvd_;
    int num_recv_neighbors_;
    int* send_list_;
    hpx::lcos::sync::ptr sends_lock_;
    hpx::lcos::sync::ptr lengths_lock_;
    hpx::lcos::sync::ptr send_list_lock_;

    int element_idx_;

    double* times_;

    //double* x_external_;

    hpx::lcos::sync::ptr neighbors_lock_;

    double ddot_local_result_;
    hpx::lcos::sync::ptr ddot_local_lock_;
    double ddot_global_result_;

    int ddots_accessed_;

    // -- helper functions for initialization --
    // --  -- look in hpccg_matrix_init.cc for the implementations
    void
    generate_matrix(int nx, int ny, int nz, HPC_Sparse_Matrix **A, double **x,
        double **b, double **xexact);

    void
    read_HPC_row(const char *data_file, HPC_Sparse_Matrix **A, double **x,
        double **b, double **xexact);

    void
    make_local_matrix(HPC_Sparse_Matrix * A);
    // -------------------------------------------

    void
    exchange_externals();

    int
    ddot(const int n, const double * const x,
        const double * const y, double * const result, double & time_allreduce);

    int
    HPC_sparsemv(const double * const x, double * const y);

    // ------- some callbacks ----------
    bool
    reduce_start_row(std::size_t i, buffer_wrapper<int> const& buf);

    bool
    reduce_tmp_neighbors(std::size_t i, buffer_wrapper<int> const& buf);

    bool
    reduce_ddot_local(std::size_t i, double const& val);

  public:

    virtual
    ~matrix();

    static ptr
    construct(hpx::globalid gid)
    {
      return ptr(new matrix(gid));
    }

    virtual void
    register_actions(boost::intrusive_ptr<hpx::hpx_thread_manager> threadman);

    virtual hpx::components::hpx_component::ptr
    clone(hpx::globalid gid)
    {
      return construct(gid);
    }

    double
    mytimer()
    {
      return curr_loc_->now().sec();
    }

    static std::string
    component_name()
    {
      return "hpccg::matrix";
    }

    enum ACTION
    {
      INIT,
      GET_START_ROW,
      GET_TMP_NEIGHBORS,
      ADD_SEND_NEIGHBOR,
      ADD_SEND_LENGTH,
      ADD_SEND_ELEMENT,
      HPCCG_SOLVE,
      ADD_ELEMENTS,
      GET_DDOT_LOCAL

    };

    static std::string
    get_action_name(ACTION a);

    // -----------------------  action functions ----------------------------- //
    // ---- these are the functions that actually do something,
    // ------- they are protected so that only actions can access them
  protected:

    int
    init(std::size_t objectid, config_params::ptr par,
        std::vector<stub_ptr> points);

    // -------- accessor functions ------ //
    buffer_wrapper<int>
    get_start_row();

    buffer_wrapper<int>
    get_tmp_neighbors();

    int
    add_send_neighbor(int val);

    int
    add_send_length(int val);

    //this one is used during initialization
    int
    add_send_element(buffer_wrapper<int> buf);

    //this one is used during the external exchange
    int
    add_elements(buffer_wrapper<double> el);

    int
    HPCCG_Solve(int max_iter, double tolerance);

    double
    get_ddot_local();

    //------------------------ typedefs for actions ------------------------ //
  public:

    typedef hpx::actions::component_result_action3<int, hpccg::matrix,
        std::size_t, config_params::ptr, std::vector<stub_ptr>, // argument
        &hpccg::matrix::init // function
    > init_action;
    typedef hpx::hpx_action_params3<size_t, config_params::ptr,
        std::vector<stub_ptr> > init_paramtype;

    typedef hpx::actions::component_result_action2<int, hpccg::matrix, int,
        double, // argument
        &hpccg::matrix::HPCCG_Solve // function
    > hpccg_solve_action;
    typedef hpx::hpx_action_params2<int, double> hpccg_solve_paramtype;

    typedef hpx::actions::component_result_action1<int, hpccg::matrix,
        buffer_wrapper<double> , // argument
        &hpccg::matrix::add_elements // function
    > add_elements_action;
    typedef hpx::hpx_action_params1<buffer_wrapper<double> >
        add_elements_paramtype;

    // --------- accessor actions ------------ //
    typedef hpx::actions::component_result_action0<buffer_wrapper<int> ,
        hpccg::matrix, &hpccg::matrix::get_start_row // function
    > get_start_row_action;
    typedef hpx::hpx_action_params get_start_row_paramtype;

    typedef hpx::actions::component_result_action0<buffer_wrapper<int> ,
        hpccg::matrix, &hpccg::matrix::get_tmp_neighbors // function
    > get_tmp_neighbors_action;
    typedef hpx::hpx_action_params get_tmp_neighbors_paramtype;

    typedef hpx::actions::component_result_action1<int, hpccg::matrix, int, // argument
        &hpccg::matrix::add_send_neighbor // function
    > add_send_neighbor_action;
    typedef hpx::hpx_action_params1<int> add_send_neighbor_paramtype;

    typedef hpx::actions::component_result_action1<int, hpccg::matrix, int, // argument
        &hpccg::matrix::add_send_length // function
    > add_send_length_action;
    typedef hpx::hpx_action_params1<int> add_send_length_paramtype;

    typedef hpx::actions::component_result_action1<int, hpccg::matrix,
        buffer_wrapper<int> , // argument
        &hpccg::matrix::add_send_element // function
    > add_send_element_action;
    typedef hpx::hpx_action_params1<buffer_wrapper<int> >
        add_send_element_paramtype;

    typedef hpx::actions::component_result_action0<double, hpccg::matrix,
        &hpccg::matrix::get_ddot_local // function
    > get_ddot_local_action;
    typedef hpx::hpx_action_params get_ddot_local_paramtype;

  };

}

#endif /* HPCCG_MATRIX_H_ */
