#include "HPCCG/hpccg_matrix.h"

#include <sstmac/software/libraries/compute/sstmac_compute.h>
#include "HPCCG/buffer_wrapper.h"
#include <sstmac/software/api/hpx/lcos/future_wait.h>
#include "HPCCG/hpccg_matrix_stubs.h"

namespace hpx_hpccg
{

  using namespace std;

  bool
  hpccg::matrix::reduce_start_row(std::size_t i, buffer_wrapper<int> const& buf)
  {
    global_index_offsets_[i] = *buf.get_buf();
    return HPX_SUCCESS;
  }

  bool
  hpccg::matrix::reduce_tmp_neighbors(std::size_t i,
      buffer_wrapper<int> const& buf)
  {
    SSTMAC_DEBUG2 << "rank " << idx_ << ": reduce_tmp_neighbors from " << i
          << "\n";

    int* raw = buf.get_buf();
    for (int j = 0; j < buf.num(); j++)
    {
      tmp_buffer_[j] += raw[j];
    }
    return HPX_SUCCESS;
  }

  buffer_wrapper<int>
  hpccg::matrix::get_start_row()
  {
    return buffer_wrapper<int> (&A_->start_row, 1);
  }

  buffer_wrapper<int>
  hpccg::matrix::get_tmp_neighbors()
  {
    tmp_neighbors_lock_->lock(); // make sure it's been initialized first
    return buffer_wrapper<int> (tmp_neighbors_, par_->domains_);

  }

  int
  hpccg::matrix::add_send_neighbor(int val)
  {
    SSTMAC_DEBUG2 << "make_local_matrix(" << idx_
          << "): recvd send neighbor \n";
    send_list_lock_->lock(); //make sure it was created
    send_list_[sends_recvd_] = val;
    sends_recvd_++;
    sends_lock_->unlock();

    return HPX_SUCCESS;
  }

  int
  hpccg::matrix::add_send_length(int val)
  {
    SSTMAC_DEBUG2 << "make_local_matrix(" << idx_ << "): recvd send length \n";
    A_->send_length[lengths_recvd_] = val;
    lengths_recvd_++;

    lengths_lock_->unlock();

    return HPX_SUCCESS;
  }

  int
  hpccg::matrix::add_send_element(buffer_wrapper<int> buf)
  {
    SSTMAC_DEBUG2 << "make_local_matrix(" << idx_
          << "): recvd send elements \n";
    int* raw = buf.get_buf();
    for (int i = 0; i < buf.num(); i++)
    {
      A_->elements_to_send[element_idx_ + i] = raw[i];
    }
    element_idx_ += buf.num();
    sends_lock_->unlock();


    return HPX_SUCCESS;
  }

  void
  hpccg::matrix::generate_matrix(int nx, int ny, int nz, HPC_Sparse_Matrix **A,
      double **x, double **b, double **xexact)

  {
    long size = par_->domains_;
    int rank = idx_;

    int local_nrow = nx * ny * nz; // This is the size of our subblock
    assert(local_nrow>0); // Must have something to work with
    int local_nnz = 27 * local_nrow; // Approximately 27 nonzeros per row (except for boundary nodes)

    int total_nrow = local_nrow * size; // Total number of grid points in mesh
    long long total_nnz = 27 * (long long) total_nrow; // Approximately 27 nonzeros per row (except for boundary nodes)

    int start_row = local_nrow * rank; // Each processor gets a section of a chimney stack domain
    int stop_row = start_row + local_nrow - 1;

    // Allocate arrays that are of length local_nrow
    int *nnz_in_row = new int[local_nrow];
    double **ptr_to_vals_in_row = new double*[local_nrow];
    int **ptr_to_inds_in_row = new int *[local_nrow];
    double **ptr_to_diags = new double*[local_nrow];

    *x = new double[local_nrow];
    *b = new double[local_nrow];
    *xexact = new double[local_nrow];

    // Allocate arrays that are of length local_nnz
    double *list_of_vals = new double[local_nnz];
    int *list_of_inds = new int[local_nnz];

    double * curvalptr = list_of_vals;
    int * curindptr = list_of_inds;

    long long nnzglobal = 0;
    for (int iz = 0; iz < nz; iz++)
      for (int iy = 0; iy < ny; iy++)
        for (int ix = 0; ix < nx; ix++)
        {
          int curlocalrow = iz * nx * ny + iy * nx + ix;
          int currow = start_row + iz * nx * ny + iy * nx + ix;
          int nnzrow = 0;
          ptr_to_vals_in_row[curlocalrow] = curvalptr;
          ptr_to_inds_in_row[curlocalrow] = curindptr;
          for (int sz = -1; sz <= 1; sz++)
            for (int sy = -1; sy <= 1; sy++)
              for (int sx = -1; sx <= 1; sx++)
              {
                int curcol = currow + sz * nx * ny + sy * nx + sx;

                if (curcol >= 0 && curcol < total_nrow)
                {
                  if (curcol == currow)
                  {
                    ptr_to_diags[curlocalrow] = curvalptr;
                    *curvalptr++ = 27.0;
                  }
                  else
                    *curvalptr++ = -1.0;
                  *curindptr++ = curcol;

                  nnzrow++;

                }
              }
          nnz_in_row[curlocalrow] = nnzrow;
          nnzglobal += nnzrow;
          (*x)[curlocalrow] = 0.0;
          (*b)[curlocalrow] = 27.0 - ((double) (nnzrow - 1));
          (*xexact)[curlocalrow] = 1.0;
        } // end ix loop

    hpx_app::compute_loop3(0, nx, 0, ny, 0, nz, 5 + 7 * 27);

    // dump("b", *b, local_nrow, rank);


    SSTMAC_DEBUG << "Process " << idx_ << " of " << size << " has "
          << local_nrow;

    SSTMAC_DEBUG << " rows. Global rows " << start_row << " through "
          << stop_row << "\n";

    SSTMAC_DEBUG << "Process " << idx_ << " of " << size << " has "
          << local_nnz << " nonzeros." << "\n";

    *A = new HPC_Sparse_Matrix; // Allocate matrix struct and fill it
    (*A)->title = 0;
    (*A)->start_row = start_row;
    (*A)->stop_row = stop_row;
    (*A)->total_nrow = total_nrow;
    (*A)->total_nnz = total_nnz;
    (*A)->local_nrow = local_nrow;
    (*A)->local_ncol = local_nrow;
    (*A)->local_nnz = local_nnz;
    (*A)->nnz_in_row = nnz_in_row;
    (*A)->ptr_to_vals_in_row = ptr_to_vals_in_row;
    (*A)->ptr_to_inds_in_row = ptr_to_inds_in_row;
    (*A)-> ptr_to_diags = ptr_to_diags;

    return;
  }

  void
  hpccg::matrix::read_HPC_row(const char *data_file, HPC_Sparse_Matrix **A,
      double **x, double **b, double **xexact)

  {
    FILE *in_file;

    int i, j, max_nnz_in_row;
    int total_nrow;
    long long total_nnz;
    int l;
    int * lp = &l;
    double v;
    double * vp = &v;

    printf("Reading matrix info from %s...\n", data_file);

    in_file = fopen(data_file, "r");
    if (in_file == NULL)
    {
      printf("Error: Cannot open file: %s\n", data_file);
      exit(1);
    }

    fscanf(in_file, "%d", &total_nrow);
    fscanf(in_file, "%d", &total_nnz);

    long size = par_->domains_;
    int rank = idx_;

    int n = total_nrow;
    int chunksize = n / size;
    int remainder = n % size;

    int mp = chunksize;
    if (rank < remainder)
      mp++;
    int local_nrow = mp;

    int off = rank * (chunksize + 1);
    if (rank > remainder)
      off -= (rank - remainder);
    int start_row = off;
    int stop_row = off + mp - 1;

    // Allocate arrays that are of length local_nrow
    int *nnz_in_row = new int[local_nrow];
    double **ptr_to_vals_in_row = new double*[local_nrow];
    int **ptr_to_inds_in_row = new int *[local_nrow];
    double **ptr_to_diags = new double*[local_nrow];

    *x = new double[local_nrow];
    *b = new double[local_nrow];
    *xexact = new double[local_nrow];

    // Find nnz for this processor
    int local_nnz = 0;
    int cur_local_row = 0;

    for (i = 0; i < total_nrow; i++)
    {
      fscanf(in_file, "%d", lp); /* row #, nnz in row */
      if (start_row <= i && i <= stop_row) // See if nnz for row should be added
      {
        local_nnz += l;
        nnz_in_row[cur_local_row] = l;
        cur_local_row++;
      }
    }
    assert(cur_local_row==local_nrow); // cur_local_row should equal local_nrow


    // Allocate arrays that are of length local_nnz
    double *list_of_vals = new double[local_nnz];
    int *list_of_inds = new int[local_nnz];

    // Define pointers into list_of_vals/inds
    ptr_to_vals_in_row[0] = list_of_vals;
    ptr_to_inds_in_row[0] = list_of_inds;
    for (i = 1; i < local_nrow; i++)
    {
      ptr_to_vals_in_row[i] = ptr_to_vals_in_row[i - 1] + nnz_in_row[i - 1];
      ptr_to_inds_in_row[i] = ptr_to_inds_in_row[i - 1] + nnz_in_row[i - 1];
    }

    cur_local_row = 0;
    for (i = 0; i < total_nrow; i++)
    {
      int cur_nnz;
      fscanf(in_file, "%d", &cur_nnz);
      if (start_row <= i && i <= stop_row) // See if nnz for row should be added
      {
        SSTMAC_DEBUG << "Process " << rank << " of " << size << " getting row "
              << i << "\n";
        for (j = 0; j < cur_nnz; j++)
        {
          fscanf(in_file, "%lf %d", vp, lp);
          ptr_to_vals_in_row[cur_local_row][j] = v;
          ptr_to_inds_in_row[cur_local_row][j] = l;
        }
        cur_local_row++;
      }
      else
        for (j = 0; j < cur_nnz; j++)
          fscanf(in_file, "%lf %d", vp, lp);
    }

    cur_local_row = 0;
    double xt, bt, xxt;
    for (i = 0; i < total_nrow; i++)
    {
      if (start_row <= i && i <= stop_row) // See if entry should be added
      {
        SSTMAC_DEBUG << "Process " << rank << " of " << size << " getting RHS "
              << i << "\n";
        fscanf(in_file, "%lf %lf %lf", &xt, &bt, &xxt);
        (*x)[cur_local_row] = xt;
        (*b)[cur_local_row] = bt;
        (*xexact)[cur_local_row] = xxt;
        cur_local_row++;
      }
      else
        fscanf(in_file, "%lf %lf %lf", vp, vp, vp); // or thrown away
    }

    fclose(in_file);
    SSTMAC_DEBUG << "Process " << rank << " of " << size << " has "
          << local_nrow;

    SSTMAC_DEBUG << " rows. Global rows " << start_row << " through "
          << stop_row << "\n";

    SSTMAC_DEBUG << "Process " << rank << " of " << size << " has "
          << local_nnz << " nonzeros." << "\n";

    *A = new HPC_Sparse_Matrix; // Allocate matrix struct and fill it
    (*A)->title = 0;
    (*A)->start_row = start_row;
    (*A)->stop_row = stop_row;
    (*A)->total_nrow = total_nrow;
    (*A)->total_nnz = total_nnz;
    (*A)->local_nrow = local_nrow;
    (*A)->local_ncol = local_nrow;
    (*A)->local_nnz = local_nnz;
    (*A)->nnz_in_row = nnz_in_row;
    (*A)->ptr_to_vals_in_row = ptr_to_vals_in_row;
    (*A)->ptr_to_inds_in_row = ptr_to_inds_in_row;
    (*A)-> ptr_to_diags = ptr_to_diags;

    return;
  }

  void
  hpccg::matrix::make_local_matrix(HPC_Sparse_Matrix * A)
  {
    std::map<int, int> externals;
    int i, j, k;
    int num_external = 0;
    double t0;

    int * neighbors = new int[max_num_neighbors];
    int * recv_length = new int[max_num_neighbors];

    A->neighbors = neighbors;
    A->recv_length = recv_length;
    A->send_length = new int[max_num_neighbors];

    // Get MPI process info

    int size = par_->domains_;
    int rank = idx_;

    // Extract Matrix pieces

    int start_row = A->start_row;
    int stop_row = A->stop_row;
    int total_nrow = A->total_nrow;
    long long total_nnz = A->total_nnz;
    int local_nrow = A->local_nrow;
    int local_nnz = A->local_nnz;
    int * nnz_in_row = A->nnz_in_row;
    double ** ptr_to_vals_in_row = A->ptr_to_vals_in_row;
    int ** ptr_to_inds_in_row = A->ptr_to_inds_in_row;

    // We need to convert the index values for the rows on this processor
    // to a local index space. We need to:
    // - Determine if each index reaches to a local value or external value
    // - If local, subtract start_row from index value to get local index
    // - If external, find out if it is already accounted for.
    //   - If so, then do nothing,
    //   - otherwise
    //     - add it to the list of external indices,
    //     - find out which processor owns the value.
    //     - Set up communication for sparse MV operation.


    ///////////////////////////////////////////
    // Scan the indices and transform to local
    ///////////////////////////////////////////


    t0 = mytimer();

    int *external_index = new int[max_external];
    int *external_local_index = new int[max_external];
    A->external_index = external_index;
    A->external_local_index = external_local_index;

    int countit = 0;

    for (i = 0; i < local_nrow; i++)
    {
      for (j = 0; j < nnz_in_row[i]; j++)
      {
        countit++;

        int cur_ind = ptr_to_inds_in_row[i][j];
        SSTMAC_DEBUG4 << "Process " << rank << " of " << size
              << " getting index " << cur_ind << " in local row " << i << "\n";

        if (start_row <= cur_ind && cur_ind <= stop_row)
        {
          ptr_to_inds_in_row[i][j] -= start_row;
        }
        else // Must find out if we have already set up this point
        {
          if (externals.find(cur_ind) == externals.end())
          {
            externals[cur_ind] = num_external++;
            if (num_external <= max_external)
            {
              external_index[num_external - 1] = cur_ind;
              // Mark index as external by negating it
              ptr_to_inds_in_row[i][j] = -(ptr_to_inds_in_row[i][j] + 1);
            }
            else
            {
              cerr << "Must increase max_external in HPC_Sparse_Matrix.hpp"
                  << "\n";
              abort();
            }
          }
          else
          {
            // Mark index as external by adding 1 and negating it
            ptr_to_inds_in_row[i][j] = -(ptr_to_inds_in_row[i][j] + 1);
          }
        }
      }
    }

    // a little compute modeling here
    hpx_app::compute_loop(0, countit, 4);

    t0 = mytimer() - t0;

    SSTMAC_DEBUG << "            Time in transform to local phase = " << t0
          << "\n";
    SSTMAC_DEBUG << "Processor " << rank << " of " << size
          << ": Number of external equations = " << num_external << "\n";

    ////////////////////////////////////////////////////////////////////////////
    // Go through list of externals to find out which processors must be accessed.
    ////////////////////////////////////////////////////////////////////////////

    t0 = mytimer();

    A->num_external = num_external;
    tmp_buffer_ = new int[size]; // Temp buffer space needed below

    // Build list of global index offset

    global_index_offsets_ = new int[size];
    for (i = 0; i < size; i++)
      tmp_buffer_[i] = 0; // First zero out

    //tmp_buffer[rank] = start_row; // This is my start row

    // This call sends the start_row of each ith processor to the ith
    // entry of global_index_offset on all processors.
    // Thus, each processor know the range of indices owned by all
    // other processors.
    // Note:  There might be a better algorithm for doing this, but this
    //        will work...


    {//MPI_Allreduce(tmp_buffer, global_index_offsets, size, MPI_INT, MPI_SUM,
      //    MPI_COMM_WORLD);
      typedef std::vector<hpx::lcos::promise<buffer_wrapper<int> >::ptr>
          lazy_results_type;
      lazy_results_type lazy_results;
      for (size_t i = 0; i < mstubs_.size(); i++)
      {
        lazy_results.push_back(mstubs_[i]->get_start_row());
      }
      ptr self(this);
      hpx::lcos::wait<buffer_wrapper<int> , hpccg::matrix>(lazy_results,
          &hpccg::matrix::reduce_start_row, self);
    }

    // Go through list of externals and find the processor that owns each
    int * external_processor = new int[num_external];
    int * new_external_processor = new int[num_external];
    for (i = 0; i < num_external; i++)
    {
      int cur_ind = external_index[i];
      for (int j = size - 1; j >= 0; j--)
        if (global_index_offsets_[j] <= cur_ind)
        {
          external_processor[i] = j;
          break;
        }
    }

    t0 = mytimer() - t0;

    SSTMAC_DEBUG << "          Time in finding processors phase = " << t0
          << "\n";

    ////////////////////////////////////////////////////////////////////////////
    // Sift through the external elements. For each newly encountered external
    // point assign it the next index in the sequence. Then look for other
    // external elements who are update by the same node and assign them the next
    // set of index numbers in the sequence (ie. elements updated by the same node
    // have consecutive indices).
    ////////////////////////////////////////////////////////////////////////////

    t0 = mytimer();

    int count = local_nrow;
    for (i = 0; i < num_external; i++)
      external_local_index[i] = -1;

    countit = 0;

    for (i = 0; i < num_external; i++)
    {
      if (external_local_index[i] == -1)
      {
        external_local_index[i] = count++;

        for (j = i + 1; j < num_external; j++)
        {
          countit++;
          if (external_processor[j] == external_processor[i])
            external_local_index[j] = count++;
        }
      }
    }

    hpx_app::compute_loop(0, countit, 2);

    t0 = mytimer() - t0;
    SSTMAC_DEBUG << "           Time in scanning external indices phase = "
          << t0 << "\n";

    t0 = mytimer();

    countit = 0;

    for (i = 0; i < local_nrow; i++)
    {
      for (j = 0; j < nnz_in_row[i]; j++)
      {
        countit++;
        if (ptr_to_inds_in_row[i][j] < 0) // Change index values of externals
        {
          int cur_ind = -ptr_to_inds_in_row[i][j] - 1;
          ptr_to_inds_in_row[i][j] = external_local_index[externals[cur_ind]];
        }
      }
    }

    hpx_app::compute_loop(0, countit, 2);

    for (i = 0; i < num_external; i++)
      new_external_processor[i] = 0;

    for (i = 0; i < num_external; i++)
      new_external_processor[external_local_index[i] - local_nrow]
          = external_processor[i];

    hpx_app::compute_loop(0, num_external, 1);

    t0 = mytimer() - t0;
    SSTMAC_DEBUG << "           Time in assigning external indices phase = "
          << t0 << "\n";

    for (i = 0; i < num_external; i++)
    {
      SSTMAC_DEBUG3 << "Processor " << rank << " of " << size
            << ": external processor[" << i << "] = " << external_processor[i]
            << "\n";
      SSTMAC_DEBUG3 << "Processor " << rank << " of " << size
            << ": new external processor[" << i << "] = "
            << new_external_processor[i] << "\n";
    }

    ////////////////////////////////////////////////////////////////////////////
    ///
    // Count the number of neighbors from which we receive information to update
    // our external elements. Additionally, fill the array tmp_neighbors in the
    // following way:
    //      tmp_neighbors[i] = 0   ==>  No external elements are updated by
    //                              processor i.
    //      tmp_neighbors[i] = x   ==>  (x-1)/size elements are updated from
    //                              processor i.
    ///
    ////////////////////////////////////////////////////////////////////////////

    t0 = mytimer();
    tmp_neighbors_ = new int[size];
    for (i = 0; i < size; i++)
      tmp_neighbors_[i] = 0;

    num_recv_neighbors_ = 0;
    int length = 1;

    for (i = 0; i < num_external; i++)
    {
      if (tmp_neighbors_[new_external_processor[i]] == 0)
      {
        num_recv_neighbors_++;
        tmp_neighbors_[new_external_processor[i]] = 1;
      }
      tmp_neighbors_[new_external_processor[i]] += size;
    }

    for (int i = 0; i < size; i++)
    {
      SSTMAC_DEBUG3 << "contents of tmp_neighbors_[" << i << "]: "
            << tmp_neighbors_[i] << "\n";
    }

    //it's been initialized, unlock it
    tmp_neighbors_lock_->unlock(mstubs_.size());

    hpx_app::compute_loop(0, num_external, 3);

    /// sum over all processors all the tmp_neighbors arrays ///


    {// MPI_Allreduce(tmp_neighbors, tmp_buffer, size, MPI_INT, MPI_SUM,
      //MPI_COMM_WORLD);
      typedef std::vector<hpx::lcos::promise<buffer_wrapper<int> >::ptr>
          lazy_results_type;
      lazy_results_type lazy_results;
      for (size_t i = 0; i < mstubs_.size(); i++)
      {
        lazy_results.push_back(mstubs_[i]->get_tmp_neighbors());
      }
      ptr self(this);
      hpx::lcos::wait<buffer_wrapper<int> , hpccg::matrix>(lazy_results,
          &hpccg::matrix::reduce_tmp_neighbors, self);
    }

    /// decode the combined 'tmp_neighbors' (stored in tmp_buffer)
    //  array from all the processors

    for (int i = 0; i < size; i++)
    {
      SSTMAC_DEBUG3 << "contents of tmp_buffer_[" << i << "]: "
            << tmp_buffer_[i] << "\n";
    }

    int num_send_neighbors = tmp_buffer_[rank] % size;

    /// decode 'tmp_buffer[rank] to deduce total number of elements
    //  we must send

    int total_to_be_sent = (tmp_buffer_[rank] - num_send_neighbors) / size;

    //
    // Check to see if we have enough workspace allocated.  This could be
    // dynamically modified, but let's keep it simple for now...
    //

    if (num_send_neighbors > max_num_messages)
    {
      cerr << "Must increase max_num_messages in HPC_Sparse_Matrix.hpp" << "\n";
      cerr << "Must be at least " << num_send_neighbors << "\n";
      abort();
    }

    if (total_to_be_sent > max_external)
    {
      cerr << "Must increase max_external in HPC_Sparse_Matrix.hpp" << "\n";
      cerr << "Must be at least " << total_to_be_sent << "\n";
      abort();
    }

    t0 = mytimer() - t0;
    SSTMAC_DEBUG << "           Time in finding neighbors phase = " << t0
          << "\n";

    SSTMAC_DEBUG << "Processor " << rank << " of " << size
          << ": Number of send neighbors = " << num_send_neighbors << "\n";

    SSTMAC_DEBUG << "Processor " << rank << " of " << size
          << ": Number of receive neighbors = " << num_recv_neighbors_ << "\n";

    SSTMAC_DEBUG << "Processor " << rank << " of " << size
          << ": Total number of elements to send = " << total_to_be_sent
          << "\n";

    A_->total_to_be_sent = total_to_be_sent;
    A_->elements_to_send = new int[total_to_be_sent];

    for (i = 0; i < total_to_be_sent; i++)
      A_->elements_to_send[i] = 0;

    /////////////////////////////////////////////////////////////////////////
    ///
    // Make a list of the neighbors that will send information to update our
    // external elements (in the order that we will receive this information).
    ///
    /////////////////////////////////////////////////////////////////////////

    int * recv_list = new int[max_external];
    j = 0;
    recv_list[j++] = new_external_processor[0];
    for (i = 1; i < num_external; i++)
    {
      if (new_external_processor[i - 1] != new_external_processor[i])
      {
        recv_list[j++] = new_external_processor[i];
      }
    }

    hpx_app::compute_loop(1, num_external, 2);

    //
    // Send a 0 length message to each of our recv neighbors
    //

    send_list_ = new int[num_send_neighbors];
    for (i = 0; i < num_send_neighbors; i++)
      send_list_[i] = 0;

    send_list_lock_->unlock(num_send_neighbors);

    for (i = 0; i < num_recv_neighbors_; i++)
    {

      // MPI_Send(tmp_buffer + i, 1, MPI_INT, recv_list[i], MPI_MY_TAG,
      //     MPI_COMM_WORLD);

      mstubs_[recv_list[i]]->add_send_neighbor(rank);

    }

    sends_lock_->lock(num_recv_neighbors_); //this one should block until all are recvd



    /////////////////////////////////////////////////////////////////////////
    ///
    //  Compare the two lists. In most cases they should be the same.
    //  However, if they are not then add new entries to the recv list
    //  that are in the send list (but not already in the recv list).
    ///
    /////////////////////////////////////////////////////////////////////////

    for (j = 0; j < num_send_neighbors; j++)
    {
      int found = 0;
      for (i = 0; i < num_recv_neighbors_; i++)
      {
        if (recv_list[i] == send_list_[j])
          found = 1;
      }

      if (found == 0)
      {
        SSTMAC_DEBUG << "Processor " << rank << " of " << size
              << ": recv_list[" << num_recv_neighbors_ << "] = "
              << send_list_[j] << "\n";
        recv_list[num_recv_neighbors_] = send_list_[j];
        (num_recv_neighbors_)++;

      }
    }

    hpx_app::compute_loop2(0, num_send_neighbors, 0, num_recv_neighbors_, 1);

    delete[] send_list_;
    num_send_neighbors = num_recv_neighbors_;

    if (num_send_neighbors > max_num_messages)
    {
      cerr << "Must increase max_external in HPC_Sparse_Matrix.hpp" << "\n";
      abort();
    }

    /////////////////////////////////////////////////////////////////////////
    /// Start filling HPC_Sparse_Matrix struct
    /////////////////////////////////////////////////////////////////////////


    //
    // Create 'new_external' which explicitly put the external elements in the
    // order given by 'external_local_index'
    //

    int * new_external = new int[num_external];
    for (i = 0; i < num_external; i++)
    {
      new_external[external_local_index[i] - local_nrow] = external_index[i];
      //  SSTMAC_DEBUG2 << "rank(" << idx_ << "): new_external(" << i << ") is " << new_external[external_local_index[i] - local_nrow] << "\n";
    }

    hpx_app::compute_loop(0, num_external, 1);

    /////////////////////////////////////////////////////////////////////////
    //
    // Send each processor the global index list of the external elements in the
    // order that I will want to receive them when updating my external elements
    //
    /////////////////////////////////////////////////////////////////////////


    j = 0;
    for (i = 0; i < num_recv_neighbors_; i++)
    {
      int start = j;
      int newlength = 0;

      // go through list of external elements until updating
      // processor changes

      while ((j < num_external) && (new_external_processor[j] == recv_list[i]))
      {
        newlength++;
        j++;
        if (j == num_external)
          break;
      }

      recv_length[i] = newlength;
      neighbors[i] = recv_list[i];

      length = j - start;

      // MPI_Send(&length, 1, MPI_INT, recv_list[i], MPI_MY_TAG, MPI_COMM_WORLD);
      mstubs_[recv_list[i]]->add_send_length(length);
    }

    lengths_lock_->lock(num_recv_neighbors_); //this one should block

    ///////////////////////////////////////////////////////////////////
    // Build "elements_to_send" list.  These are the x elements I own
    // that need to be sent to other processors.
    ///////////////////////////////////////////////////////////////////


    j = 0;
    for (i = 0; i < num_recv_neighbors_; i++)
    {
      int start = j;
      int newlength = 0;

      // Go through list of external elements
      // until updating processor changes.  This is redundant, but
      // saves us from recording this information.

      while ((j < num_external) && (new_external_processor[j] == recv_list[i]))
      {

        newlength++;
        j++;
        if (j == num_external)
          break;
      }
      // MPI_Send(new_external + start, j - start, MPI_INT, recv_list[i],
      //    MPI_MY_TAG, MPI_COMM_WORLD);
      buffer_wrapper<int> rap(new_external + start, j - start);
      SSTMAC_DEBUG << "make_local_matrix(" << idx_ << "): sending " << (j
            - start) << " elements to " << recv_list[i] << " which is "
            << rap.get_byte_length() << " bytes \n";
      mstubs_[recv_list[i]]->add_send_element(
          buffer_wrapper<int> (new_external + start, j - start));
    }

    sends_lock_->lock(num_recv_neighbors_);


    /// replace global indices by local indices ///
    SSTMAC_DEBUG << "make_local_matrix(" << idx_ << "): correcting indices \n";

    for (i = 0; i < total_to_be_sent; i++)
      A_->elements_to_send[i] -= start_row;

    SSTMAC_DEBUG << "make_local_matrix(" << idx_
          << "): I have the following send lengths: \n";
    for (i = 0; i < num_send_neighbors; i++)
    {
      SSTMAC_DEBUG << A->send_length[i] << "\n";
    }

    ////////////////
    // Finish up !!
    ////////////////

    A->num_send_neighbors = num_send_neighbors;
    A->local_ncol = A->local_nrow + num_external;

    //Used in exchange_externals
    double *send_buffer = new double[total_to_be_sent];
    A->send_buffer = send_buffer;

    delete[] tmp_buffer_;
    delete[] global_index_offsets_;
    delete[] recv_list;
    delete[] external_processor;
    delete[] new_external_processor;
    delete[] tmp_neighbors_;
    return;
  }

}
