/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "gtc_params.h"

#include <sstmac/common/sstmac_env.h>

#define _USE_MATH_DEFINES
#include <math.h>
#include <functional>
#include <numeric>

namespace hpx_gtc
{
  using namespace std;
  using namespace sstmac;
  using namespace sstmac::sw;

  sstmac::debug gtc_module::dbg_;

  config_parameters::config_parameters(sstmac::sim_parameters::ptr params)
  {
    bool err = false;
    irun_ = params->get_int_param("gtc_main_irun");
    mstep_ = params->get_int_param("gtc_main_mstep");
    msnap_ = params->get_int_param("gtc_main_msnap");
    ndiag_ = params->get_int_param("gtc_main_ndiag");
    nonlinear_ = params->get_double_param("gtc_main_nonlinear");
    paranl_ = params->get_double_param("gtc_main_paranl");
    mode00_ = params->get_int_param("gtc_main_mode00");
    tstep_ = params->get_double_param("gtc_main_tstep");
    micell_ = params->get_int_param("gtc_main_micell");
    mecell_ = params->get_int_param("gtc_main_mecell");
    mpsi_ = params->get_int_param("gtc_main_mpsi");
    mthetamax_ = params->get_int_param("gtc_main_mthetamax");
    mzetamax_ = params->get_int_param("gtc_main_mzetamax");
    npartdom_ = params->get_int_param("gtc_main_npartdom");
    ncycle_ = params->get_int_param("gtc_main_ncycle");
    a_ = params->get_double_param("gtc_main_a");
    a0_ = params->get_double_param("gtc_main_a0");
    a1_ = params->get_double_param("gtc_main_a1");
    q0_ = params->get_double_param("gtc_main_q0");
    q1_ = params->get_double_param("gtc_main_q1");
    q2_ = params->get_double_param("gtc_main_q2");
    rc_ = params->get_double_param("gtc_main_rc");
    rw_ = params->get_double_param("gtc_main_rw");
    aion_ = params->get_double_param("gtc_main_aion");
    qion_ = params->get_double_param("gtc_main_qion");
    aelectron_ = params->get_double_param("gtc_main_aelectron");
    qelectron_ = params->get_double_param("gtc_main_qelectron");
    kappati_ = params->get_double_param("gtc_main_kappati");
    kappate_ = params->get_double_param("gtc_main_kappate");
    kappan_ = params->get_double_param("gtc_main_kappan");
    fixed_Tprofile_ = params->get_optional_bool_param(
        "gtc_main_fixed_Tprofile", true);
    tite_ = params->get_double_param("gtc_main_tite");
    flow0_ = params->get_double_param("gtc_main_flow0");
    flow1_ = params->get_double_param("gtc_main_flow1");
    flow2_ = params->get_double_param("gtc_main_flow2");
    r0_ = params->get_double_param("gtc_main_r0");
    b0_ = params->get_double_param("gtc_main_b0");
    temperature_ = params->get_double_param("gtc_main_temperature");
    edensity0_ = params->get_double_param("gtc_main_edensity0");
    nbound_ = params->get_int_param("gtc_main_nbound");
    umax_ = params->get_double_param("gtc_main_umax");
    iload_ = params->get_int_param("gtc_main_iload");
    tauii_ = params->get_double_param("gtc_main_tauii");
    track_particles_ = params->get_bool_param("gtc_main_track_particles");
    nptrack_ = params->get_int_param("gtc_main_nptrack");
    rng_control_ = params->get_long_param("gtc_main_rng_control");
    params->get_vector_param("gtc_main_nmode", nmode_);
    params->get_vector_param("gtc_main_mmode", mmode_);

    nhybrid_ = params->get_optional_int_param("gtc_main_nhybrid", 0);

    num_mode_ = nmode_.size();
    mflux_ = 5;
    m_poloidal_ = 9;

  }

  void
  config_parameters::setup(size_t numberpe)
  {
    numberpe_ = numberpe;

    // ----- First we verify the consistency of ntoroidal and npartdom -------
    // The number of toroidal domains (ntoroidal) times the number of particle
    // "domains" (npartdom) needs to be equal to the number of processor "numberpe".
    // numberpe cannot be changed since it is given on the command line.

    // numberpe must be a multiple of npartdom so change npartdom accordingly
    while (numberpe_ % npartdom_ != 0)
    {
      npartdom_ = npartdom_ - 1;
      if (npartdom_ == 1)
        break;
    }
    ntoroidal_ = sstmac_env::params->get_optional_int_param("gtc_ntoroidal",
        numberpe_ / npartdom_);
  }

  void
  config_parameters::print_thyself()
  {
    SSTMAC_DEBUG << "------------------- GTC parameters -------------- \n";
    SSTMAC_DEBUG << "irun:\t\t\t" << irun_ << "\n";
    SSTMAC_DEBUG << "mstep:\t\t\t" << mstep_ << "\n";
    SSTMAC_DEBUG << "msnap:\t\t\t" << msnap_ << "\n";
    SSTMAC_DEBUG << "ndiag:\t\t\t" << ndiag_ << "\n";
    SSTMAC_DEBUG << "nonlinear:\t\t" << nonlinear_ << "\n";
    SSTMAC_DEBUG << "paranl:\t\t\t" << paranl_ << "\n";
    SSTMAC_DEBUG << "mode00:\t\t\t" << mode00_ << "\n";
    SSTMAC_DEBUG << "tstep:\t\t\t" << tstep_ << "\n";
    SSTMAC_DEBUG << "numberpe:\t\t" << numberpe_ << "\n";
    SSTMAC_DEBUG << "ntoroidal:\t\t" << ntoroidal_ << "\n";

  }

  particle_decomp::particle_decomp(config_parameters::ptr params)
  {

  }

  void
  particle_decomp::setup(config_parameters::ptr params, size_t mype)
  {
    SSTMAC_DEBUG << "gtc(" << mype << ")::particle_decomp - starting..\n";

    ntoroidal_ = params->ntoroidal_;
    mype_ = mype;

    // make sure that mzetamax is a multiple of ntoroidal
    params->mzetamax_ = ntoroidal_ * max(1,
        int(double(params->mzetamax_) / double(ntoroidal_) + 0.5));

    // Make sure that "mpsi", the total number of flux surfaces, is an even
    // number since this quantity will be used in Fast Fourier Transforms
    params->mpsi_ = 2 * (params->mpsi_ / 2);

    // We now give each PE (task) a unique domain identified by 2 numbers: the
    // particle and toroidal domain numbers.
    //    particle_domain_location = rank of the particle domain holding mype
    //    toroidal_domain_location = rank of the toroidal domain holding mype
    //
    // On the IBM SP, the MPI tasks are distributed in an orderly fashion to each
    // node unless the LoadLeveler instruction "#@ blocking = unlimited" is used.
    // On Seaborg for example, the first 16 tasks (mype=0-15) will be assigned to
    // the first node that has been allocated to the job, then the next 16
    // (mype=16-31) will be assigned to the second node, etc. When not using the
    // OpenMP, we want the particle domains to sit on the same node because
    // communication is more intensive. To achieve this, successive PE numbers are
    // assigned to the particle domains first.
    // It is easy to achieve this ordering by simply using mype/npartdom for
    // the toroidal domain and mod(mype,npartdom) for the particle domain.
    //
    //  pe_number=0
    //  do j=0,ntoroidal-1
    //     do i=0,npartdom-1
    //        pe_grid(i,j)=pe_number
    //        particle_domain_location(pe_number)=i
    //        toroidal_domain_location(pe_number)=j
    //        pe_number=pe_number+1
    //     enddo
    //  enddo

    particle_domain_location_ = (mype % params->npartdom_);
    toroidal_domain_location_ = mype / params->npartdom_;

    myrank_toroidal_ = mype_;

    //  write(0,*)'mype=',mype,"  particle_domain_location =",&
    //            particle_domain_location,' toroidal_domain_location =',&
    //            toroidal_domain_location,' pi=',pi

    // Domain decomposition in toroidal direction.
    mzeta_ = params->mzetamax_ / ntoroidal_;
    zetamin_ = 2.0 * M_PI * double(toroidal_domain_location_) / double(
        ntoroidal_);
    zetamax_ = 2.0 * M_PI * double(toroidal_domain_location_ + 1) / double(
        ntoroidal_);

    //  write(0,*)mype,' in set_particle_decomp: mzeta=',mzeta,'  zetamin=',&
    //            zetamin,'  zetamax=',zetamax

    // grid spacing in the toroidal direction
    deltaz_ = (zetamax_ - zetamin_) / double(mzeta_);

    // We now find the toroidal neighbors of the current toroidal domain and
    // store that information in 2 easily accessible variables. This information
    // is needed several times inside the code, such as when particles cross
    // the domain boundaries. We will use the toroidal communicator to do these
    // transfers so we don't need to worry about the value of myrank_partd.
    // We have periodic boundary conditions in the toroidal direction so the
    // neighbor to the left of myrank_toroidal=0 is (ntoroidal-1).

    left_pe_ = ((myrank_toroidal_ - 1 + ntoroidal_) % ntoroidal_);
    right_pe_ = ((myrank_toroidal_ + 1) % ntoroidal_);

    SSTMAC_DEBUG << "******** decomposition ************** \n";
    SSTMAC_DEBUG << "npartdom: " << params->npartdom_ << "\n";
    SSTMAC_DEBUG << "myrank_toroidal: " << myrank_toroidal_ << "\n";
    SSTMAC_DEBUG << "ntoroidal: " << ntoroidal_ << "\n";
    SSTMAC_DEBUG << "left_pe: " << left_pe_ << "\n";
    SSTMAC_DEBUG << "right_pe: " << right_pe_ << "\n";
    SSTMAC_DEBUG << "************************************* \n";

    SSTMAC_DEBUG << "gtc(" << mype << ")::particle_decomp - done \n";
  }

  void
  field_array::setup(config_parameters::ptr params, double gyroradius,
      particle_decomp::ptr partdecomp, size_t mype)
  {

    // --- Define poloidal grid ---
    // grid spacing
    deltar_ = (params->a1_ - params->a0_) / double(params->mpsi_);

    std::vector<int> mtheta_(params->mpsi_ + 1);
    // grid shift associated with fieldline following coordinates
    double tdum = 2.0 * M_PI * params->a1_ / double(params->mthetamax_);
    double r, q;
    for (int i = 0; i <= params->mpsi_; i++)
    {
      r = params->a0_ + deltar_ * double(i);
      mtheta_[i] = max(2,
          min(params->mthetamax_, 2 * int(M_PI * r / tdum + 0.5))); // even # poloidal grid

    }
    // un-comment the next two lines to use magnetic coordinate
    //  qtinv=0.0
    //  itran=0

    // When doing mode diagnostics, we need to switch from the field-line following
    // coordinates alpha-zeta to a normal geometric grid in theta-zeta. This
    // translates to a greater number of grid points in the zeta direction, which
    // is mtdiag. Precisely, mtdiag should be mtheta/q but since mtheta changes
    // from one flux surface to another, we use a formula that gives us enough
    // grid points for all the flux surfaces considered.
    params->mtdiag_ = (params->mthetamax_ / params->mzetamax_)
        * params->mzetamax_;
    // In the case of a non-linear run, we need only half the points.
    if (params->nonlinear_ > 0.5)
      params->mtdiag_ = params->mtdiag_ / 2;

    // number of grids on a poloidal plane
    mgrid_ = std::accumulate(mtheta_.begin(), mtheta_.end(), 1);
    mi_local_ = params->micell_ * (mgrid_ - params->mpsi_) * partdecomp->mzeta_; //# of ions in toroidal domain
    mi_ = params->micell_ * (mgrid_ - params->mpsi_) * partdecomp->mzeta_
        / params->npartdom_; //# of ions per processor
    if (mi_ < ((int) mi_local_ % params->npartdom_))
      mi_ = mi_ + 1;
    me_local_ = params->mecell_ * (mgrid_ - params->mpsi_) * partdecomp->mzeta_; //# of electrons in toroidal domain
    me_ = params->mecell_ * (mgrid_ - params->mpsi_) * partdecomp->mzeta_
        / params->npartdom_; //# of electrons per processor
    if (me_ < ((int) me_local_ % params->npartdom_))
      me_ = me_ + 1;
    mimax_ = mi_ + 100 * ceil(sqrt(double(mi_))); // ions array upper bound
    memax_ = me_ + 100 * ceil(sqrt(double(me_))); // electrons array upper bound

    if (mype == 0)
    {
      SSTMAC_DEBUG << "#ions per proc= " << mi_ << " #electrons per proc= "
            << me_ << "\n";
      SSTMAC_DEBUG << "#ions per grid cell= " << params->micell_
            << "   #electrons per grid cell= " << params->mecell_ << "\n";
      SSTMAC_DEBUG << "#ions per toroidal domain=" << mi_local_
            << "   #electrons per toroidal domain= " << me_local_ << "\n";
      SSTMAC_DEBUG << "mzeta=" << partdecomp->mzeta_ << "\n";
      SSTMAC_DEBUG << "mgrid=" << mgrid_ << "\n";
      SSTMAC_DEBUG << "mpsi=" << params->mpsi_ << "\n";

    }

  }

}
