/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <example_shared_memory.h>

namespace exshmem {

    SSTRegisterApp("example_shared_memory", example_shared_memory);

    // here's a C++ shared variable which can be accessed by all the threads
    int example_shared_memory::worker::shared_var_ = 0;

    // A c-style static variable.
    int WORKERS_ACTIVE = 0;

    /**
     * consume_params : this function should grab any parameters it
     * needs from the parameters, in place of passing in argv and argc
     * through main
     * @param params - the simulation parameters passed in through the
     * simulation configuration file and overrides on the command line
     */
    void
    example_shared_memory::consume_params(const sim_parameters::ptr &params)
    {
      // Here's how we grab parameters.
      // The simulation will automatically complain if you forgot
      //  to provide this parameter, but you can explicitly
      //  check if it was provided with params->has_param("paramname")
      //  which is useful for optional parameters
      num_workers_ = params->get_int_param("sharedmemory_workers");

    }

    /**
     * skeleton_main replaces main(...) as the entry point of
     * the application.  The simulator will make sure the application
     * starts up in the right places and calls this.
     */
    void
    example_shared_memory::skeleton_main()
    {
      // if multiple threads were started up as a part of the same
      //  application, which doesn't really make sense for real
      //  applications but is still possible in the simulator,
      //  you can get the "number" of this instance
      //  with the id_ member variable.
      int rank = id_.task_.id;

      //however, chances are you are the single "main" thread managing
      // worker threads, so lets spawn some dudes.
      int amount = 10;
      for (int i = 0; i < num_workers_; i++)
        {
          softwareid sid(id_.app_, taskid(rank++));
          worker::ptr wrk = worker::construct(sid, amount);

          //here's how we tell to start him up.
          // We can also use os_->add_application if we want
          // to pass parameters using a sim_parameters::ptr,
          // which might be useful for many parameters from
          // a file, but for a small number of paramters
          // just pass them through the constructor.
          os_->start_thread(wrk);
        }

      // An admittedly poor way of managing threads - by
      //  polling every so often.
      while(WORKERS_ACTIVE > 0){
          sleep(timestamp(1e-3));
      }

      SSTMAC_DEBUG << "Application: done \n";

    }

    /**
     * Here's the run() function for our worker thread
     */
    void
    example_shared_memory::worker::skeleton_main()
    {
      WORKERS_ACTIVE++;

      SSTMAC_DEBUG << "Starting worker " << id_.task_.id  << "\n";

      for(int i = 0; i < amount_of_work_; i++){
          shared_var_++;

          //This is how we debug.  It only prints out if we turn it on,
          // and the right-hand-side is only std::stringified if it's turned on,
          // so you shouldn't worry too much about the performance implications
          // of putting these in.
          SSTMAC_DEBUG << "Worker " << id_.task_.id << ": computing for some time at iteration " << i << "\n";

          //Here, I'm saying that I want to compute for
          // 1 microsecond, which is one way of describing
          // computation
          timestamp compute_time(1e-6);
          compute(compute_time);
      }

      while(shared_var_ > 0){


          SSTMAC_DEBUG << "Worker " << id_.task_.id << ": doing some matrix things with shared_var = " << shared_var_ << "\n";

          // Here we use a libary which has already been defined which models
          //  some matrix operations.  See sstmac/software/libraries/compute/lib_compute_matrix.h/cc
          //  for what it does, which you can do right in here if you wanted to.
          //  However, defining libraries for compute-type things which
          //  may be repeated are the preferred way of organizing apps, because
          //  it allows you to separate the app's logic from the description
          //  of it's performance to the simulator.
          lib_mat_->double_mxm(100, 10, 100, 1);

          //Here we're going show that this variable can be shared
          shared_var_ = shared_var_ - 2;
      }

      SSTMAC_DEBUG << "Finishing worker " << id_.task_.id  << "\n";

      WORKERS_ACTIVE--;
    }

} //end namespace 

