#include <sstmac/backends/native/driver.h>
#include <sstmac/backends/native/nodeid.h>
//#include <sstmac/backends/native/threadparallel/logicalprocess.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/software/services/launch/indexing/indexstrategy.h>
#include <sstmac/common/logger.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/processor/processor.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/hardware/memory/memorymodel.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/services/launch/instantlaunch.h>
#include <sstmac/software/process/key.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/backends/native/eventheap.h>

namespace sstmac
{
  namespace native
  {

    void
    clear_statics()
    {
      nodeid::clear_statics();
      sw::allocationstrategy::clear_statics();
      sw::indexstrategy::clear_statics();
      logger::clear_statics();
      hw::interconnect::clear_statics();
      hw::processor::clear_statics();
      hw::networkinterface::clear_statics();
      hw::memorymodel::clear_statics();
      sw::OS_API::clear_statics();
      sw::mpiqueue::clear_statics();
      sw::mpiapi::clear_statics();
      sw::app_factory::clear_apps();
      sw::instantlaunch::clear_statics();
      sw::key::clear_statics();
    }




    eventmanager::ptr
    driver_run(const sim_parameters::ptr &params, timestamp until, bool print)
    {

      //read environmental variables as potential overrides
      char* param = getenv("MPICH_GNI_MAX_VSHORT_MSG_SIZE");
      if (param)
        params->add_param_override("max_vshort_msg_size", param);

      param = getenv("MPICH_GNI_MAX_EAGER_MSG_SIZE");
      if (param)
        params->add_param_override("max_eager_msg_size", param);

      param = getenv("MPICH_SMP_SINGLE_COPY_SIZE");
      if (param)
        params->add_param_override("smp_single_copy_size", param);

      /** set to an absurdly high value - no single copies */
      param = getenv("MPICH_SMP_SINGLE_COPY_OFF");
      if (param)
        params->add_param_override("smp_single_copy_size", "99999999999");

      debug dbg_("<debug> sstmac | sim | simulation");

      manager::ptr mgr;

      try
      {
        std::string mgr_type = params->get_optional_param("event_manager",
            "eventmap");
        eventmanager::ptr ev_mgr;
        if (mgr_type == "eventmap")
        {
          ev_mgr = eventmap::construct();
        }else if(mgr_type == "eventheap"){
          ev_mgr = eventheap::construct();
        }

        else if (mgr_type == "eventlist")
        {
          long one_million = 1000 * 1000;
          /** for ps ticks, this covers a ms */
          long nticks_window = params->get_optional_long_param(
              "eventlist_nticks_window", one_million);
          /** for ps ticks, this covers a us */
          long nwindows = params->get_optional_long_param("eventlist_nwindows",
              one_million);
          ev_mgr = eventlist::construct(nticks_window, nwindows);
        }
        else
        {
          std::string errormsg = "driver::run: invalid event manager "
              + mgr_type;
          throw inputerror(errormsg);
        }

        if (!params->has_param("manager") || params->get_param("manager")
            == "simple")
        {
          mgr = manager::construct(ev_mgr);
        }
        else if (params->get_param("manager") == "degraded")
        {
          mgr = manager_degraded::construct(ev_mgr);
        }

        manager::global = mgr;

        mgr->build_interconnect(params);

        SSTMAC_DEBUG << "-- nodes added -- \n";

        timestamp runtime = mgr->run(until);

        if (print)
        {
          dbg_ << sstprintf("Estimated total runtime of %20.8f seconds\n",
              runtime.sec());
        }

        mgr->finish();
        sstmac_env::params = 0;
        return mgr->get_eventmanager();

      }
      catch (const std::exception& e)
      {
        if (mgr)
          mgr->done();
        throw;
      }
    }

  }
}
