#include <sstmac/backends/native/eventlist.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sstmac-config.h>

#ifndef SSTMAC_USE_GMPXX
#define time_to_int(x) x.ticks()
#else
#define time_to_int(x) ((int64_t) x.ticks().get_d())
#endif

namespace sstmac
{
  namespace native
  {

    eventunit::eventunit() :
      head_(0), tail_(0)
    {
    }

    void
    eventunit::append(const event::ptr &ev)
    {
      ev->next = 0;
      if (head_)
        {
          tail_->next = ev;
          tail_ = ev;
        }
      else
        {
          head_ = ev;
          tail_ = ev;
        }
    }

    void
    eventunit::clear()
    {
      head_ = 0;
      tail_ = 0;
    }

    bool
    eventunit::nonempty() const
    {
      return head_;
    }

    bool
    eventunit::empty() const
    {
      return head_ == 0;
    }

    event::ptr
    eventunit::pop()
    {
      event::ptr next = head_;
      head_ = head_->next;
      return next;
    }

    void
    eventlist::cancel_all_messages(const eventmodule::ptr &handler)
    {
      throw ssterror(
          "eventmanager::cancel all messages is not supported with event list yet");
    }

    eventlist::eventlist(long nticks_window, long nwindows) :
        future_windows_(0),
        event_units_(0),
        distant_future_horizon_(1e9), //set to something stupid large for now
        current_window_tick_start_(0),
        nticks_time_window_(nticks_window),
          current_window_(0),
          nwindows_(nwindows),
          window_limit_(nwindows),
          max_scheduled_window_(0), running_(false)
    {
      event_units_ = new eventunit[nticks_window];
      future_windows_ = new eventunit[nwindows];
    }

    eventlist::~eventlist() throw ()
    {
    }

    void
    eventlist::schedule(const timestamp& start_time, const event::ptr& ev)
    {
      int64_t ticks = time_to_int(start_time);
      schedule(ticks, ev);
    }

    void
    eventlist::schedule(int64_t ticks, const event::ptr& ev)
    {
      int64_t bin_index = ticks / nticks_time_window_;
      if (bin_index == current_window_)
        {
          int64_t event_index = ticks - current_window_tick_start_;

          SSTMAC_DEBUG << "eventlist: adding event ticks=" << ticks
                << " into current-window=" << current_window_
                << " window-limit=" << window_limit_ << " bin-index="
                << bin_index << " event-index=" << event_index << "\n";

          event_units_[event_index].append(ev);
        }
      else if (bin_index >= window_limit_)
        {
          if (distant_future_head_)
            {
              distant_future_tail_->next = ev;
              distant_future_tail_ = ev;
            }
          else
            {
              distant_future_head_ = ev;
              distant_future_tail_ = ev;
            }
          if (bin_index < distant_future_horizon_)
            distant_future_horizon_ = bin_index;

          SSTMAC_DEBUG << "eventlist: adding at future event horizion="
                << distant_future_horizon_ << "\n";
        }
      else
        {
          if (bin_index > max_scheduled_window_)
            max_scheduled_window_ = bin_index;

          bin_index = bin_index % nwindows_;
          SSTMAC_DEBUG << "eventlist: adding event ticks=" << ticks
                << " into future bin " << bin_index << " with max window "
                << max_scheduled_window_ << "\n";
          future_windows_[bin_index].append(ev);
        }
    }

    void
    eventlist::do_one_event()
    {

    }

    void
    eventlist::schedule(const event::ptr &ev)
    {
      schedule(time_to_int(ev->time()), ev);
    }

    void
    eventlist::reschedule_future_events()
    {
      SSTMAC_DEBUG << "eventlist: rescheduling events at future horizon="
            << distant_future_horizon_ << "\n";
      /** add some ridiculous increment for now */
      distant_future_horizon_ += 1e9;
      while (distant_future_head_)
        {
          schedule(distant_future_head_);
          distant_future_head_ = distant_future_head_->next;
        }
    }

    void
    eventlist::find_next_window()
    {
      bool empty = true;
      while (current_window_ < max_scheduled_window_ && empty)
        {
          /** keep going to try to find more events */
          ++current_window_;
          ++window_limit_;
          current_window_tick_start_ += nticks_time_window_;

          if (current_window_ == distant_future_horizon_)
            reschedule_future_events();

          empty = future_windows_[current_window_ % nwindows_].empty();
          SSTMAC_DEBUG << "eventlist: moving on to window=" << current_window_
                << " max-scheduled-window=" << max_scheduled_window_
                << " window-limit=" << window_limit_ << " future-horizon="
                << distant_future_horizon_ << " empty=" << empty << "\n";
        }
    }

    void
    eventlist::run()
    {
      if (running_)
        {
          throw illformederror("eventlist::run: already running");
        }
      running_ = true;
      stopped_ = false;
      current_window_ = 0;
      while (1)
        {
          for (int64_t unit = 0; unit < nticks_time_window_; ++unit)
            {
              eventunit& ev_unit = event_units_[unit];
              while (ev_unit.nonempty())
                {
                  event::ptr ev = ev_unit.pop();
                  SSTMAC_DEBUG << "eventlist: running event in window="
                        << current_window_ << " ticks="
                        << time_to_int(ev->time()) << "\n";
                  set_now(ev->time());
                  ev->execute();
                }
            }
          /** having finished the current window... */
          find_next_window();

          eventunit& next_unit = future_windows_[current_window_ % nwindows_];
          if (next_unit.empty())
            {
              /** There are no more events */
              if (distant_future_head_ == 0)
                {
                  break;
                }

              /** The only events left are in the DISTANT FUTURE !!! */
              int64_t delta_window = distant_future_horizon_ - current_window_;
              current_window_ = distant_future_horizon_;
              window_limit_ = current_window_ + nwindows_;
              current_window_tick_start_ += delta_window * nticks_time_window_;
              reschedule_future_events();
            }
          else
            {
              /** schedule the next events */
              while (next_unit.nonempty())
                {
                  event::ptr ev = next_unit.pop();
                  schedule(ev);
                }
              next_unit.clear();
            }
        }

      //at this point, we have no more events

      running_ = false;
      complete_ = true;
      finish();
    }

    void
    eventlist::step()
    {
      throw illformederror("Cannot step through eventlist");
    }

    eventlist::ptr
    eventlist::construct(long nticks_window, long nwindows)
    {
      return new eventlist(nticks_window, nwindows);
    }

    void
    eventlist::clear(const timestamp& zero_time)
    {
      throw illformederror("eventlist::clear: eventlist cannot be cleared");
    }

    void
    eventlist::finish()
    {
      std::list<ptr_type::ptr>::iterator it, end = finishers_.end();
      for (it = finishers_.begin(); it != end; it++)
        {
          eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(*it);
          mod->finish();
        }

      finish_stats();
    }

    void
    eventlist::register_finish(ptr_type::ptr p)
    {
      finishers_.push_back(p);
    }

  }
}

