/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

//#include <backends/sstmacro-native/manager-env.h>
#include <sstmac/backends/native/manager.h>
#include <sstmac/backends/native/nodelist.h>

#include <sstmac/common/driver_util.h>
#include <sstmac/backends/native/eventmap.h>

#include <sstmac/common/counter.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/cast.h>
#include <sstmac/software/ami/environment.h>

#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>

#include <iostream>
#include <iterator>
#include <cstdlib>

namespace sstmac
{
  namespace native
  {
    using namespace sstmac::sw;
    using namespace sstmac::hw;

    manager::ptr manager::global = 0;

    //
    // Default constructor.
    //
    manager::manager(const eventmanager::ptr& ev_mgr) :
      eventmanager_(ev_mgr), running_(false),
          next_ppid_(0), dbg_("<debug> manager")
    //stats(statscollector::construct("stats.log"))
    {
      SSTMAC_DEBUG << "manager::manager()\n";
      eventmanager::global = eventmanager_;
      logger::timer_ = eventmanager_;
    }

    //
    // Construct a manager.
    //
    manager::ptr
    manager::construct(const eventmanager::ptr& ev_mgr)
    {
      manager::ptr retval(new manager(ev_mgr));
      return retval;
    }

    void
    manager::finish()
    {
        if (interconnect_){
            interconnect_->finish();
        }
        interconnect_ = 0;
    }

    //
    // Goodbye.
    //
    manager::~manager() throw ()
    {
      SSTMAC_DEBUG << "manager::~manager()\n";
      if (this->running_)
      {
          std::cerr
              << "FATAL:  manager going out of scope while still running.\n";
          abort();
      }
      SSTMAC_DEBUG << "manager::~manager() leaving\n";
    }

    //
    // Define a network.
    //
    void
    manager::build_interconnect(const sim_parameters::ptr& params)
    {
      if (running_){
          sst_throw(illformederror, "manager::build_interconnect: simulation is running");
      }
      if (interconnect_){
          sst_throw(illformederror, "manager::build_interconnect: interconnect already built");
      }

      /** sstkeyword {
            docstring = Specify the general type of network congestion model that will be used
                        for the interconnect;
            gui = train;
      } */
      // If distributed, need to build a distributed interconnect.
      interconnect_ = interconnect_factory::get_param("network_name", params);

      eventmanager_->register_finish(interconnect_);

      SSTMAC_DEBUG << "manager::build_interconnect(" << interconnect_->to_string() << ")\n";
    }

    //
    // Access the interconnect.
    //
    interconnect::ptr
    manager::get_interconnect() const
    {
      return interconnect_;
    }

    //
    // Set the interconnect (for parallel simulations).
    //
    void
    manager::set_interconnect(const interconnect::ptr& fab)
    {
        interconnect_ = fab;
        eventmanager_->register_finish(fab);
    }

    //
    // Setup the simulation.
    //
    void
    manager::setup()
    {
      SSTMAC_DEBUG << "manager::setup: entered\n";
      if (interconnect_ == 0)
        {
          throw nullerror("manager::setup:  Null interconnect pointer.");
        }
      if (running_)
        {
          throw illformederror("manager::setup:  Simulation already running.");
        }
      // Initial setup.
      if (eventmanager_ == 0)
        {
          eventmanager_ = eventmap::construct();

          eventmanager::global = eventmanager_;
        }
      eventmanager_->params_ = sstmac_env::params;
      // Make sure the network simulator knows about our event manager.
      interconnect_->set_eventmanager(eventmanager_);

      vis::vis_display::ptr visd =
          boost::dynamic_pointer_cast<vis::vis_display>(interconnect_);
      if (visd && sstmac_env::params->has_param("vis_engine"))
        {
          visd->vis_start(true);
          std::cout
              << "--- VTK visualization is on, press any key to continue \n";
          std::cin.ignore();
        }

      // Schedule job startups.
      // for (startup_t::iterator it = startup_.begin(); it != startup_.end(); ++it)
      //  {
      //    eventmanager_->schedule(it->first, it->second, sst_message::ptr());
      //  }
      startup_.clear();
      running_ = true;
    }

    //
    // Start the simulation.
    //
    timestamp
    manager::run(timestamp until)
    {

      // setup the simulation before starting
      setup();

      SSTMAC_DEBUG << "manager::run: calling eventmanager::run\n";
      if (until.sec() > 0)
        {
          eventmanager_->schedule(until, stophandler::construct(),
              sst_message::ptr());
        }
      eventmanager_->run();
      // Now call done routine to end simulation and print Stats.
      done();

      return eventmanager_->now();

    }

    void
    manager::done()
    {
      SSTMAC_DEBUG << "manager::done: entered\n";

      running_ = false;
      /*if (allocated_.size() > 0)
       {
       std::cout << "manager no longer running but " << allocated_.size()
       << " nodes are still allocated--deallocating all nodes."
       << std::endl;
       std::copy(allocated_.begin(), allocated_.end(),
       std::inserter(available_, available_.begin()));
       allocated_.clear();
       }*/

      vis::vis_display::ptr visd =
          boost::dynamic_pointer_cast<vis::vis_display>(interconnect_);
      if (visd && sstmac_env::params->has_param("vis_engine"))
        {
          std::cout << "--- Simulation has finished, press any key to kill \n";
          std::cin.ignore();
          visd->vis_complete();
        }

      if (interconnect_){
        interconnect_->finish();
      }
      interconnect_ = 0;

      eventmanager::global = 0;
      // stats->printStats(eventmanager_->now());
    }

  }
} // end of namespace sstmac
