/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/backends/native/mpiparallel/mpi_serializer.h>
#include <sstmac/common/messages/serializable.h>
#include <sstmac/common/sstmac_env.h>
namespace sstmac
{
  namespace native
  {

    template<class T>
      void
      mpi_ser_packer::do_pack(T t)
      {
        char* off = buffer_ + at_;
        T* cast = (T*) off;
        *cast = t;
        at_ += sizeof(T);

        SSTMAC_DEBUG << "serializer: packed a " << typeid(T).name() << " " << t
              << ", now at " << at_ << "\n";
      }

    template<class T>
      void
      mpi_ser_unpacker::do_unpack(T &t)
      {
        char* off = buffer_ + at_;
        T* cast = (T*) off;
        t = *cast;
        at_ += sizeof(T);
        SSTMAC_DEBUG << "serializer: unpacked a " << typeid(T).name() << " "
              << t << ", now at " << at_ << "\n";
      }

#define ImplementSerType(ty) \
    void \
    mpi_ser_packer::pack(ty t){ \
        do_pack<ty>(t); \
    } \
    void \
    mpi_ser_unpacker::unpack(ty& t){ \
        do_unpack<ty>(t); \
    } \
    void \
    mpi_ser_sizer::size(ty t){ \
      size_ += sizeof(ty); \
    }

    ImplementSerType(short)
    ImplementSerType(float)
    ImplementSerType(double)
    ImplementSerType(int)
    ImplementSerType(unsigned int)
    ImplementSerType(long int)
    ImplementSerType(long long int)
    ImplementSerType(long long unsigned int)
    ImplementSerType(char)
    ImplementSerType(long unsigned int)
    ImplementSerType(bool)

    void
    mpi_ser_packer::pack(std::string i)
    {
      char* off = buffer_ + at_;
      short* cast = (short*) off;
      strcpy(off, i.c_str());

      at_ += strlen(i.c_str()) + 1;

      SSTMAC_DEBUG << "serializer: packed a string " << i << ", now at " << at_
            << "\n";

    }
    void
    mpi_ser_unpacker::unpack(std::string &i)
    {
      char* off = buffer_ + at_;
      short* cast = (short*) off;
      char buf[100];
      strcpy(buf, off);
      i = std::string(buf);

      at_ += strlen(i.c_str()) + 1;
      SSTMAC_DEBUG << "serializer: unpacked a string " << i << ", now at "
            << at_ << "\n";
    }

    void
    mpi_ser_sizer::size(std::string i)
    {
      size_ += i.length() + 1;
    }

    void
    mpi_ser_packer::pack(timestamp i)
    {
      pack((long int) i.ticks_int64());
    }

    void
    mpi_ser_unpacker::unpack(timestamp &i)
    {
      long int t;
      unpack(t);
      i = timestamp::exact_ticks(t);
    }

    void
    mpi_ser_sizer::size(timestamp t)
    {
      size_ += sizeof(long int);
    }

    void
    mpi_ser_packer::pack_array(void* arr, size_t size)
    {
      SSTMAC_DEBUG2 << "serializer, packing an array that contains...";
      if (dbg_.is_active(2))
      {
        unsigned char* cast = (unsigned char*)arr;
        for (int i = 0; i < size; i++)
        {
          SSTMAC_DEBUG2 << i << " " << cast[i] << "\n";
        }
      }
      char* off = buffer_ + at_;
      memcpy(off, arr, size);
      at_ += size;

      SSTMAC_DEBUG << "serializer: packed an array of size " << size
            << " at position " << at_ << "\n";
    }

    void
    mpi_ser_unpacker::unpack_array(void* arr, size_t size)
    {
      char* off = buffer_ + at_;
      memcpy(arr, off, size);
      at_ += size;

      SSTMAC_DEBUG << "serializer: unpacked an array of size " << size
            << " at position " << at_ << "\n";

      SSTMAC_DEBUG2 << "serializer, unpacked an array that contains...";
      if (dbg_.is_active(2))
      {
        unsigned char* cast = (unsigned char*) arr;
        for (int i = 0; i < size; i++)
        {
          SSTMAC_DEBUG2 << i << " " << cast[i] << "\n";
        }
      }
    }

  }
}

