/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#include <sstmac/backends/native/mpiparallel/par_eventmap.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

#include <sstmac/hardware/network/switches/dist_dummyswitch.h>

#include <sstmac/backends/native/mpiparallel/null_event.h>
#include <sstmac/backends/native/ipc_event.h>

void
break_deadlock_fn(int *invec, int *inoutvec, int *len, MPI_Datatype *dtype)
{
  long int* castin = (long int*) invec;
  long int* castout = (long int*) inoutvec;

  long int final;

  if (*castin < 0)
    return;
  else if (*castout < 0)
    final = *castin;
  else
  {
    final = std::min(*castin, *castout);
  }

  *castout = final;

}

namespace sstmac
{
  namespace native
  {

    par_eventmap::par_eventmap(const timestamp &tv, int lpid, int numlps,
        partition::ptr part, const timestamp &iplat) :
      lpid_(lpid), numlps_(numlps), part_(part), iplat_(iplat)
    {

      set_now(tv);

      //waiting_on_update_ = 0;
      // last_update_sent_ = timestamp(0);

      for (int i = 0; i < numlps; i++)
      {
        //comm_buffers_[i] = (int*)malloc(sizeof(int) * (LAST_COMM_STATE));
        state_[i] = INIT;

        last_update_time_[i] = iplat;

        if (i != lpid_)
        {
          queues_[i] = par_q(i, lpid_);
        }
      }

      MPI_Op_create((MPI_User_function*) break_deadlock_fn, 1, &dlop_);

    }

    void
    par_eventmap::run()
    {

      if (state_[lpid_] == INIT)
      { // we never got anything to start off
        change_state(NOLOC);
      }

      while (state_[lpid_] != FINISHED)
      {
        SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: top of run, state: "
              << state_[lpid_] << "\n";

        check_for_updates();

        if (state_[lpid_] == LOC_NOK || state_[lpid_] == NOLOC)
        {
          wait_for_updates();
        }
        else if (state_[lpid_] == LOC_OK)
        {
          do_one_event();
        }

      }

      finish();
    }

    /**
     * I'm changing my state locally, so tell everyone else about it
     * @param s
     */
    void
    par_eventmap::change_state(STATE s)
    {

      if (state_[lpid_] != s)
      {
        SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: changing from state "
              << state_[lpid_] << " to " << s << "\n";

        queue_t::iterator it = localevents_.begin();
        timestamp min_t = timestamp(-1);
        if (localevents_.size() > 0)
          min_t = it->first;

        for (int i = 0; i < numlps_; i++)
        {
          if (i != lpid_)
          {
            last_sent_time_[i] = now_ + iplat_;
            par_msg_state* parm = new par_msg_state();
            parm->state_ = s;
            parm->now_ = now_;
            parm->min_ = min_t;
            queues_[i].send_message(parm);
          }
        }
      }

      state_[lpid_] = s;
      //check_deadlock();
    }

    /**
     * I've received a state update from someone else, so note that
     * @param s
     * @param stat
     */
    void
    par_eventmap::update_state(STATE s, int lp)
    {
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: updating state of " << lp
            << " to " << s << "\n";
      if (s == FINISHED)
      {
        state_[lpid_] = FINISHED; //i'm done too then
      }
      state_[lp] = s;

      /* if (s == LOC_NOK && state_[lpid_] != NOLOC)
       {
       queue_t::iterator it = localevents_.begin();
       timestamp min_t = timestamp(-1);
       if (localevents_.size() > 0)
       min_t = it->first;

       //send him an update
       null_event::ptr nu = null_event::construct(min_t);
       par_msg_message* pm = new par_msg_message();
       pm->event_ = nu;
       queues_[lp].send_message(pm);
       }*/

      //check_deadlock(); //must do it here too
    }

    void
    par_eventmap::check_deadlock()
    {
      int cnt = 0;
      for (int i = 0; i < numlps_; i++)
      {
        if (state_[i] == LOC_OK || state_[i] == INIT || state_[i] == FINISHED)
        {
          return;
        }

        if (last_sent_time_[i] > last_update_time_[i])
          return;

        if (state_[i] == LOC_NOK)
          cnt++;
      }

      // if(cnt < 2)
      //   return;

      //  if(last_sent)

      /*   par_msg_poss_deadlock* dl = new par_msg_poss_deadlock();
       dl->started_ = lpid_;

       queues_[(lpid_ + 1) % numlps_].send_message(dl);*/

      queue_t::iterator it = localevents_.begin();
      timestamp min_t = timestamp(-1);
      if (localevents_.size() > 0)
        min_t = it->first;

      long int t = min_t.ticks();
      long int gmin_t;

      SSTMAC_DEBUG << "par_eventmap[" << lpid_
            << "]: breaking deadlock, sending " << min_t << "\n";

      MPI_Allreduce(&t, &gmin_t, 1, MPI_LONG, dlop_, MPI_COMM_WORLD);

      timestamp gt = timestamp::exact_ticks(gmin_t);
      SSTMAC_DEBUG << "par_eventmap[" << lpid_
            << "]: after deadlock, got global min: " << gt << "\n";

      if (gmin_t < 0)
      {
        sst_throw(ssterror, "got negative global time back from deadlock, meaning noone had any events");
      }
      else
      {

        for (int i = 0; i < numlps_; i++)
        {
          if (i != lpid_)
          {
            last_update_time_[i] = gt;
          }
        }
        if (state_[lpid_] == LOC_NOK)
        {

          if (check_valid(min_t))
          {
            change_state(LOC_OK);
          }
        }
      }

    }

    /**
     * Check for various things that could affect my state
     */
    void
    par_eventmap::check_for_updates()
    {

      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: check_for_updates \n";
      for (int i = 0; i < numlps_; i++)
      {
        if (i != lpid_)
        {
          bool good = true;
          par_message* p = NULL;
          do
          {
            p = queues_[i].check_recv();
            good = (p != NULL);
            parse_message(p);
          }
          while (good);
        }
      }

      //check if I need to update my state at all
      queue_t::iterator it = localevents_.begin();
      timestamp min_t = timestamp(-1);
      if (localevents_.size() > 0)
        min_t = it->first;

      if (state_[lpid_] == LOC_OK)
      {
        check_valid(min_t);

      }
      else if (state_[lpid_] == LOC_NOK)
      {
        if (check_valid(min_t))
        {
          change_state(LOC_OK);
        }
      }
      else if (state_[lpid_] == NOLOC)
      {
        if (localevents_.size() > 0)
        {
          if (check_valid(min_t))
          {
            change_state(LOC_OK);
          }
          else
          {
            change_state(LOC_NOK);
          }
        }
      }

      // check_deadlock();
      //check for deadlock
      /* if (state_[lpid_] != FINISHED && waiting_on_update_ == 0)
       {
       bool fine = false; //is anyone running?
       for (int i = 0; i < numlps_; i++)
       {
       if (state_[i] == LOC_OK || state_[i] == INIT)
       {
       fine = true;
       break;
       }
       }
       if (!fine)
       {
       dump_info();
       sst_throw(ssterror, "par_eventmap::check_for_updates - it seems we have deadlocked");

       }
       }*/

    }

    bool
    par_eventmap::check_valid(timestamp min_t)
    {
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: check_valid \n";
      bool ok = true;
      for (int i = 0; i < numlps_; i++)
      {
        if (i != lpid_)
        {
          if (last_update_time_[i] < min_t && (state_[i] == LOC_OK || state_[i]
              == LOC_NOK))
          {
            // if (state_[lpid_] == LOC_OK)
            //  {
            //SSTMAC_DEBUG << "par_eventmap[" << lpid_
            //      << "]: sending update request to " << i << "\n";
            change_state(LOC_NOK);

            SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: waiting on " << i
                  << " since their update time is " << last_update_time_[i]
                  << " and my min is " << min_t << " and last sent is "
                  << last_sent_time_[i] << "\n";
            //    par_msg_update2* pu = new par_msg_update2();
            //    pu->min_ = min_t;
            //    queues_[i].send_message(pu);

            // }

            ok = false;
            break;

          }
        }
      }
      return ok;
    }

    void
    par_eventmap::dump_info()
    {

      std::cout << "------------------------------- \n";
      std::cout << "par_eventmap[" << lpid_ << "]: \n";

      if (localevents_.size() > 0)
      {
        queue_t::iterator it = localevents_.begin();
        timestamp min_t = it->first;
        std::cout << "min_t: " << min_t << "\n";
      }
      else
      {
        std::cout << "queue empty \n";
      }

      std::cout << "states: \n";
      for (int i = 0; i < numlps_; i++)
      {
        std::cout << "state[" << i << "]: " << state_[i] << "\n";
        std::cout << "last update[" << i << "]: " << last_update_time_[i]
            << "\n";
      }
      std::cout << "------------------------------- \n";
    }

    /**
     * I can't progress any further, so wait for anything to happen
     */
    void
    par_eventmap::wait_for_updates()
    {
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: wait_for_updates \n";
      int numreqs = numlps_;
      int which = 1;
      MPI_Status stat;
      MPI_Request reqs[numreqs];

      for (int i = 0; i < numlps_; i++)
      {
        if (i == lpid_)
          reqs[i] = MPI_REQUEST_NULL;
        else
          reqs[i] = queues_[i].get_req();
      }

      MPI_Waitany(numreqs, reqs, &which, &stat);

      par_message* p = queues_[which].parse_request(reqs[which], stat);
      parse_message(p);

      queues_[which].repost_recv();

    }

    /**
     * Side Effect!!!! this function deletes p
     * @param p
     */
    void
    par_eventmap::parse_message(par_message *p)
    {

      if (!p)
        return;

      switch (p->type_)
      {

      case par_message::MESSAGE:
      {
        par_msg_message* pm = (par_msg_message*) p;

        null_event::ptr nu =
            boost::dynamic_pointer_cast<null_event>(pm->event_);
        if (nu)
        {
          SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: from " << p->from_
                << " : got null message with time " << nu->time()
                << " at recent " << now_ << "\n";

          if (state_[pm->from_] == LOC_NOK)
          { //they're waiting (possibly on me)
            if (last_sent_time_[pm->from_] < nu->time())
            { //they are waiting on me
              queue_t::iterator it = localevents_.begin();
              timestamp min_t = timestamp(-1);
              if (localevents_.size() > 0)
                min_t = it->first;
              if (state_[lpid_] == LOC_NOK && last_update_time_[pm->from_]
                  < min_t)
              { //i'm also waiting on them
                timestamp timetosend = (min_t < nu->time()) ? min_t : nu->time(); // we should both take the min
            //    timetosend = timetosend + iplat_;

                SSTMAC_DEBUG << "par_eventmap[" << lpid_
                      << "]: trying to break deadlock by sending "
                      << timetosend << "\n";
                last_sent_time_[pm->from_] = timetosend;
                // last_update_sent_ = timetosend;
                null_event::ptr nu = null_event::construct(timetosend);
                par_msg_message* npm = new par_msg_message();
                npm->event_ = nu;
                queues_[pm->from_].send_message(npm);

              }
            }
          }

          if (last_update_time_[pm->from_] < nu->time())
            last_update_time_[pm->from_] = nu->time();

          queue_t::iterator it = localevents_.begin();
          timestamp min_t = timestamp(-1);
          if (localevents_.size() > 0)
            min_t = it->first;

          /*    if (state_[lpid_] == LOC_NOK && min_t > last_update_time_[pm->from_])
           { //that wasn't good enough, request a bigger update
           par_msg_update2* pu2 = new par_msg_update2();
           pu2->min_ = min_t;
           queues_[pm->from_].send_message(pu2);
           }*/

          // if(state_[lpid_] == LOC_NOK && nu->time() > now_ && nu->time() < min_t){
          //    set_now(nu->time());
          //  }


        }
        else
        {
          ipc_event::ptr iev = boost::dynamic_pointer_cast<ipc_event>(
              pm->event_);
          SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: got msg going to "
                << iev->get_dest_switch()->to_string() << " \n";
          // << ev->get_message()->to_string() << "\n";

          iev->set_event_handler(
              network_->get_topology()->find_switch(network_->get_switches(),
                  iev->get_dest_switch()));
          SSTMAC_DEBUG << "set event handler to "
                << iev->get_handler()->to_string() << "\n";
          SSTMAC_DEBUG << "has message " << iev->get_message()->to_string()
                << "\n";
          localevents_.insert(
              std::make_pair<timestamp, ipc_event::ptr>(iev->time(), iev));

          if (iev->time() < now_)
          {
            sst_throw(
                ssterror,
                "par_eventmap::deserialize() - got an event with time %ld that is before now %ld",
                (long) iev->time().ticks(), (long) now_.ticks());
          }

          if (last_update_time_[pm->from_] < iev->time())
            last_update_time_[pm->from_] = iev->time();

        }

        break;
      }

      case par_message::STATE:
      {
        par_msg_state* ps = (par_msg_state*) p;
        update_state((STATE) ps->state_, ps->from_);

        if (ps->now_ + iplat_ > last_update_time_[ps->from_])
          last_update_time_[ps->from_] = ps->now_ + iplat_;

        if (ps->state_ == LOC_NOK)
        { //they're waiting (possibly on me)
          if (last_sent_time_[ps->from_] < ps->min_)
          { //they are waiting on me
            queue_t::iterator it = localevents_.begin();
            timestamp min_t = timestamp(-1);
            if (localevents_.size() > 0)
              min_t = it->first;
            if (state_[lpid_] == LOC_NOK && last_update_time_[ps->from_]
                < min_t)
            { //i'm also waiting on them
              timestamp timetosend = (min_t < ps->min_) ? min_t : ps->min_; // we should both take the min
           //   timetosend = timetosend + iplat_;
              SSTMAC_DEBUG << "par_eventmap[" << lpid_
                    << "]: trying to break deadlock by sending " << timetosend
                    << "\n";
              last_sent_time_[ps->from_] = timetosend;
              // last_update_sent_ = timetosend;
              null_event::ptr nu = null_event::construct(timetosend);
              par_msg_message* pm = new par_msg_message();
              pm->event_ = nu;
              queues_[ps->from_].send_message(pm);

            }
          }
        }
        break;
      }

      case par_message::UPDATE:
      {
        par_msg_update* pu = (par_msg_update*) p;

        send_update2(pu->from_, pu->min_);
        break;
      }
      case par_message::UPDATE2:
      {
        par_msg_update2* pu = (par_msg_update2*) p;

        send_update2(pu->from_, pu->min_);
        break;
      }

      case par_message::POSSDEADLOCK:
      {
        /* par_msg_poss_deadlock* pdl = (par_msg_poss_deadlock*) p;

         if (state_[lpid_] != LOC_OK && state_[lpid_] != INIT)
         { // could be locked according to us
         if (pdl->started_ == lpid_)
         { //we got it back

         }
         else
         { //pass it along
         par_msg_poss_deadlock* dl = new par_msg_poss_deadlock();
         dl->started_ = pdl->started_;
         queues_[(lpid_ + 1) % numlps_].send_message(dl);
         }
         } //else just forget about it
         break;*/sst_throw(ssterror, "possdeadlock not used right now");
      }

      }

      delete p;
    }

    /**
     * Other partition is stuck waiting on me, and has requested
     * an update of my time, so send a null event.
     * @param stat
     */
    void
    par_eventmap::send_update(int lp, timestamp theirs)
    {
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: sending update to " << lp
            << " \n";
      timestamp timetosend = now_ + iplat_; //can definitely gaurantee this

      last_sent_time_[lp] = timetosend;
      // last_update_sent_ = timetosend;
      null_event::ptr nu = null_event::construct(timetosend);
      par_msg_message* pm = new par_msg_message();
      pm->event_ = nu;
      queues_[lp].send_message(pm);

    }

    /**
     * Other partition is stuck waiting on me, and has requested
     * an update of my time, so send a null event.
     * @param stat
     */
    void
    par_eventmap::send_update2(int lp, timestamp theirs)
    {
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]: sending update2 to "
            << lp << "\n";

      timestamp timetosend = now_ + iplat_; //can definitely gaurantee this
      if (theirs > timetosend) //it won't do them any good even if we send an update

      {
        // std::cout << "they have: " << theirs << "\n";
        //  std::cout << "I could send them: " << (now_ + iplat_) << "\n";
        //  std::cout << "iplat: " << iplat_ << "\n";
        //  sst_throw(ssterror, "par_eventmap::send_update - won't do no good");
        if (localevents_.size() > 0)
        {
          queue_t::iterator it = localevents_.begin();
          timestamp min_t = it->first;
          if (min_t > timetosend)
          {
            timetosend = min_t + iplat_;

          }
        }
        else
        {
          // last_update_time_[stat.MPI_SOURCE] = theirs;
          //  timetosend = theirs;
        }
      }

      last_sent_time_[lp] = timetosend;
      // last_update_sent_ = timetosend;
      null_event::ptr nu = null_event::construct(timetosend);
      par_msg_message* pm = new par_msg_message();
      pm->event_ = nu;
      queues_[lp].send_message(pm);

    }

    void
    par_eventmap::do_one_event()
    {
      queue_t::iterator it = localevents_.begin();
      timestamp min_t = it->first;
      event::ptr min_ev = it->second;

      set_now(min_t);
      SSTMAC_DEBUG << "par_eventmap[" << lpid_
            << "]::do_one_event:  Processing event " << min_ev << " of type "
            << min_ev->to_string() << " at time " << now() << "\n";
      min_ev->execute();
      SSTMAC_DEBUG << "par_eventmap[" << lpid_ << "]::do_one_event:  Event "
            << min_ev << " complete, " << localevents_.size()
            << " events pending" << "\n";

      localevents_.erase(it);

      if (localevents_.size() == 0)
      {
        change_state(NOLOC);
      }/*else{
       it = localevents_.begin();
       if(it->first > now_ + iplat_){
       null_event::ptr nu = null_event::construct(now_ + iplat_);
       schedule(now_ + iplat_, nu);
       }
       }*/
    }

    void
    par_eventmap::end_gracefully()
    {
      change_state(FINISHED);
    }

    void
    par_eventmap::set_switches(const hw::switchinterconnect::ptr& fab)
    {
      network_ = fab;
    }

    /**
     * The time set on ev is now + latency
     */
    void
    par_eventmap::ipc_schedule(nodeaddress::ptr goingto, event::ptr &ev)
    {
      ipc_event::ptr ie = boost::dynamic_pointer_cast<ipc_event>(ev);
      if (!ie)
        sst_throw(ssterror, "par_eventmap::ipc_schedule - need ipc_event, got something else");
      int lp = part_->get_parent(goingto->unique_id());

      SSTMAC_DEBUG << "par_eventmap[" << lpid_
            << "]:ipc_schedule, sending event to" << "partition " << lp
            << "...";

      last_sent_time_[lp] = ev->time();

      par_msg_message* parm = new par_msg_message();
      parm->event_ = ev;
      queues_[lp].send_message(parm);

    }

    void
    par_eventmap::schedule(const timestamp& start_time, const event::ptr& ev)
    {

      if (state_[lpid_] == INIT)
      {
        change_state(LOC_OK);
      }

      timestamp thetime = start_time;

      simple_event::ptr sev = boost::dynamic_pointer_cast<simple_event>(ev);
      if (sev)
      {
        hw::dist_dummyswitch::ptr sw = boost::dynamic_pointer_cast<
            hw::dist_dummyswitch>(sev->get_handler());

        if (sw)
        { //this is a correction for latency, since the dummy switches have to schedule
          // their messages [latency] in advance
          thetime = start_time - iplat_;

        }
      }

      ev->set_time(thetime);

      localevents_.insert(std::make_pair(thetime, ev));
      SSTMAC_DEBUG << "par_eventmap[" << lpid_
            << "]::schedule:  Scheduling event " << ev << " which is "
            << ev->to_string() << thetime << ", " << localevents_.size()
            << " events pending.\n";
      ++nevents_;
      // }
    }

    void
    par_eventmap::step()
    {
      sst_throw(ssterror,
          "par_eventmap::step - this function is not compatible with parallel simulation");
    }

    void
    par_eventmap::finish()
    {
      SSTMAC_DEBUG << "reporting for " << finishers_.size() << " modules \n";
      std::list<ptr_type::ptr>::iterator it, end = finishers_.end();
      for (it = finishers_.begin(); it != end; it++)
      {
        eventmodule::ptr mod = boost::dynamic_pointer_cast<eventmodule>(*it);
        mod->finish();
      }
      finishers_.clear();
    }

    void
    par_eventmap::register_finish(ptr_type::ptr p)
    {
      finishers_.push_back(p);
    }
  }
}
