/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "par_q.h"
#include <sstmac/backends/native/mpiparallel/mpi_serializer.h>

namespace sstmac
{
  namespace native
  {

    debug par_message::dbg_;
    debug par_q::dbg_;
#define INITIAL_Q_RECV_BUFFER_SIZE 1000000

    par_q::par_q(int lp, int host) :
      lp_(lp), hostlp_(host)
    {

      current_recv_size_ = INITIAL_Q_RECV_BUFFER_SIZE;
      send_buffer_ = (char*) malloc(sizeof(char) * current_recv_size_);
      recv_buffer_ = (char*) malloc(sizeof(char) * current_recv_size_);

      //post it again
      MPI_Irecv(recv_buffer_, current_recv_size_, MPI_CHAR, lp_, lp_,
          MPI_COMM_WORLD, &inreq_);

      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> parq");
        logger::register_logger(&dbg_);
      }
    }

    par_message*
    par_q::parse_request(MPI_Request req, const MPI_Status &stat)
    {
      par_message* ret = NULL;
      int* cast = (int*) recv_buffer_;
      if (cast[0] == par_message::STATE)
      {
        SSTMAC_DEBUG << "parse_request::state...";
        par_msg_state* p = new par_msg_state();
        p->state_ = cast[1];
        long int* lcast = (long int*) (&cast[2]);
        p->now_ = timestamp::exact_ticks(lcast[0]);
        p->min_ = timestamp::exact_ticks(lcast[1]);
        SSTMAC_DEBUG << p->state_ << "\n";
        ret = p;
      }
      else if (cast[0] == par_message::UPDATE)
      {
        SSTMAC_DEBUG << "parse_request::update...";
        long int* lcast = (long int*) (&cast[1]);

        par_msg_update* p = new par_msg_update();
        SSTMAC_DEBUG << lcast[0] << "\n";
        p->min_ = timestamp::exact_ticks(lcast[0]);

        ret = p;
      }
      else if (cast[0] == par_message::MESSAGE)
      {
        SSTMAC_DEBUG << "parse_request::message...\n";
        event::ptr e = deserialize(req, stat, (char*) &cast[1]);
        par_msg_message* p = new par_msg_message();
        p->event_ = e;
        ret = p;
      }
      else if (cast[0] == par_message::UPDATE2)
      {
        SSTMAC_DEBUG << "parse_request::update2...";
        long int* lcast = (long int*) (&cast[1]);

        par_msg_update2* p = new par_msg_update2();
        SSTMAC_DEBUG << lcast[0] << "\n";
        p->min_ = timestamp::exact_ticks(lcast[0]);
        ret = p;
      }
      else
      {
        sst_throw(ssterror, "par_q::check_recv() - recvd unknown message id %d", cast[0]);
      }

      ret->from_ = lp_;

      return ret;
    }

    par_message*
    par_q::check_recv()
    {
      par_message* ret = NULL;
      int flag = 1;
      MPI_Status stat;

      MPI_Test(&inreq_, &flag, &stat);

      if (flag)
      {
        ret = parse_request(inreq_, stat);
        repost_recv();
      }
      return ret;

    }

    void
    par_q::repost_recv()
    {
      MPI_Irecv(recv_buffer_, current_recv_size_, MPI_CHAR, lp_, lp_,
          MPI_COMM_WORLD, &inreq_);
    }

    par_message*
    par_q::wait_message()
    {

      par_message* ret;
      int flag = 1;
      MPI_Status stat;

      MPI_Wait(&inreq_, &stat);

      ret = parse_request(inreq_, stat);
      repost_recv();

      return ret;
    }

    int
    par_msg_state::pack_stuff(char* buf, int max)
    {
      int* cast = (int*) buf;
      cast[0] = state_;
      long int* lcast = (long int*)(&cast[1]);
      lcast[0] = now_.ticks();
      lcast[1] = min_.ticks();
      return sizeof(int) + 2 * sizeof(long int);
    }

    int
    par_message::pack_event(event::ptr ev, char* buf, int max)
    {
      if (!ev)
      {
        sst_throw(ssterror, "par_message::pack_event - null event");
      }
      //first, size that thing up
      ser_sizer::ptr s = mpi_ser_sizer::construct();
      sst_serializer* sersize = new sst_serializer(s);
      sersize->do_obj(ev);
      int bufsize = s->get_size();
      SSTMAC_DEBUG << "it sized at " << bufsize << "...";

      if (bufsize > max)
      {
        /* size_t newsize = bufsize;
         MPI_Send(&newsize, 1, MPI_LONG, path_.second, mpisize_ + path_.first,
         MPI_COMM_WORLD);
         int fake;
         MPI_Status stat;
         MPI_Recv(&fake, 1, MPI_BYTE, path_.second, mpisize_ + path_.first,
         MPI_COMM_WORLD, &stat);
         current_recv_size_ = newsize;
         sendbuf_ = (char*) realloc(sendbuf_, newsize);*/sst_throw(ssterror, "par_eventmap::ipc_schedule - buffer resize is not implemented right now");
      }

      //pack it up, pack it in
      mpi_ser_packer::ptr p = mpi_ser_packer::construct(buf);
      sst_serializer* serpack = new sst_serializer(p);
      serpack->do_obj(ev);

      SSTMAC_DEBUG << "serialized " << p->get_bytes_serialized() << " bytes \n";

      if (bufsize != p->get_bytes_serialized())
      {
        sst_throw(ssterror, "mpipar_ipq_out::insert_event - event sized at %d but packed %d", bufsize, p->get_bytes_serialized());
      }

      return bufsize;
    }

    int
    par_msg_update::pack_stuff(char* buf, int max)
    {
      long int* cast = (long int*) buf;
      cast[0] = (long int) min_.ticks();
      return sizeof(long int);
    }

    int
    par_msg_update2::pack_stuff(char* buf, int max)
    {
      long int* cast = (long int*) buf;
      cast[0] = (long int) min_.ticks();
      return sizeof(long int);
    }

    int
    par_msg_message::pack_stuff(char* buf, int max)
    {
      return pack_event(event_, buf, max);
    }

    void
    par_q::send_message(par_message *m)
    {
      if (!m)
      {
        sst_throw(ssterror, "par_q::send_message - trying to send null");
      }
      //first pack header
      int* cast = (int*) send_buffer_;
      cast[0] = m->type_;

      int bufsize = m->pack_stuff((char*) &cast[1],
          current_recv_size_ - sizeof(int));

      MPI_Send(send_buffer_, bufsize + sizeof(int), MPI_CHAR, lp_, hostlp_,
          MPI_COMM_WORLD);

      delete m;
    }

    event::ptr
    par_q::deserialize(MPI_Request req, const MPI_Status &stat, char * buf)
    {

      //SSTMAC_DEBUG << "par_eventmap[" << lp_ << "]: from " << stat.MPI_SOURCE
      //      << " : check_recv - deserializing..";
      int count;
      MPI_Status stemp = stat;
      MPI_Get_count(&stemp, MPI_CHAR, &count);
      //SSTMAC_DEBUG << "got " << count << " bytes...";

      event::ptr ev;
      //unpack
      ser_unpacker::ptr u = mpi_ser_unpacker::construct(buf);
      sst_serializer* serunpack = new sst_serializer(u);
      serunpack->do_obj(ev);

      return ev;

    }

  }
}
