/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_PAR_Q_H_
#define SSTMAC_BACKENDS_NATIVE_PAR_Q_H_

#include <sstmac/common/ptr_type.h>
#include <mpi.h>
#include <sstmac/common/sstevent.h>
#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace native
  {

    class par_message
    {
    public:
      enum MESSAGE_TYPE
      {
        STATE, //here's my state
        UPDATE, //need a time update
        UPDATE2, //that wasn't good enough
        MESSAGE, //either null event or real message
        POSSDEADLOCK,
        BREAKDEADLOCK,
        NONE
      //not a valid message
      };

    protected:
      par_message(MESSAGE_TYPE t) :
        type_(t)
      {
        if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> parq");
          logger::register_logger(&dbg_);
        }

      }

      int
      pack_event(event::ptr e, char* buf, int max);

    public:

      par_message() :
        type_(NONE)
      {
      }

      /**
       * returns number of bytes packed
       * @param buf
       * @return
       */
      virtual int
      pack_stuff(char* buf, int max) = 0;

      MESSAGE_TYPE type_;
      int from_; //only used for recv functions

      static debug dbg_;
    };

    class par_msg_state : public par_message
    {
    public:
      par_msg_state() :
        par_message(STATE)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      int state_;
      timestamp now_;
      timestamp min_;
    };

    class par_msg_update : public par_message
    {
    public:
      par_msg_update() :
        par_message(UPDATE)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      timestamp min_;
    };

    class par_msg_poss_deadlock : public par_message
    {
    public:
      par_msg_poss_deadlock() :
        par_message(POSSDEADLOCK)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      int started_;
    };

    class par_msg_break_deadlock : public par_message
    {
    public:
      par_msg_break_deadlock() :
        par_message(BREAKDEADLOCK)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      timestamp min_;
    };

    class par_msg_update2 : public par_message
    {
    public:
      par_msg_update2() :
        par_message(UPDATE2)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      timestamp min_;
    };

    class par_msg_message : public par_message
    {
    public:
      par_msg_message() :
        par_message(MESSAGE)
      {

      }

      virtual int
      pack_stuff(char* buf, int max);

      event::ptr event_;
    };

    class par_q
    {

    protected:
      int lp_;
      int hostlp_;
      char* send_buffer_;
      char* recv_buffer_;
      int current_recv_size_;

      MPI_Request inreq_;

      static debug dbg_;

    public:
      par_q(int lp, int host);

      par_q()
      {
        lp_ = -1;
        hostlp_ = -1;
      }

      virtual
      ~par_q()
      {

      }

      MPI_Request
      get_req()
      {
        return inreq_;
      }

      par_message*
      check_recv();

      par_message*
      wait_message();

      /**
       * Side Effect!!!! This function deletes the message
       * @param m
       */
      void
      send_message(par_message* m);

      par_message*
      parse_request(MPI_Request req, const MPI_Status &stat);

      void
      repost_recv();

    protected:
      event::ptr
      deserialize(MPI_Request req, const MPI_Status &stat, char * buf);

    };

  }
}
#endif /* PAR_Q_H_ */
