/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_EVENTMANAGER_H_INCLUDED
#define SSTMAC_COMMON_EVENTMANAGER_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/sstevent.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/stats/stat_logger.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/backends/native/sim_partition.h>

namespace sstmac
{

  /**
   * Base type for implementations of an engine that
   * is able to schedule events and advance simulation time
   * in the right order.
   */

  class eventmodule;
  typedef boost::intrusive_ptr<eventmodule> eventschedulerPtr;

  class eventmanager : public ptr_type
  {

  protected:
    timestamp now_;

    static std::list<stat_logger::ptr> stats_;

    void
    set_now(const timestamp &ts);

    eventmanager();

    static debug dbg_;
    static debug updates_;

    time_t sim_start_;
    time_t last_update_;
    time_t update_interval_;

    long long nevents_;

    bool complete_;

  public:
    bool finish_on_stop_;

    virtual std::string
    to_string() const
    {
      return "eventmanager";
    }

    bool
    is_complete()
    {
      return complete_;
    }

    typedef boost::intrusive_ptr<eventmanager> ptr;

    static eventmanager::ptr global;

    /// Goodbye.
    virtual
    ~eventmanager();

    /// Clear all events and set time back to a zero of your choice.
    /// This call shall not be permitted while the event manager is running.
    virtual void
    clear(const timestamp &zero_time = timestamp(0)) = 0;

    /// Run the eventmanager.
    /// The eventmanager shall return control when no more messages remain.
    virtual void
    run() = 0;

    virtual void
    step() = 0;

    /// Get the current eventmanager time.
    /// This call must work regardless of whether the eventmanager is currently
    /// running or not.  After a run, the call should return the time at which
    /// the last event was completed.
    const timestamp&
    now() const
    {
      return now_;
    }

    long long
    nevents() const
    {
      return nevents_;
    }

    static void
    register_stat(stat_logger::ptr stat)
    {
      stats_.push_back(stat);
    }

    sim_parameters::ptr params_;

    /// Set off the given eventhandler at the given time.
    virtual void
    schedule(const timestamp &start_time, const eventhandler::ptr &handler,
        const sst_message::ptr &msg);

    virtual void
    schedule(const timestamp& start_time, const event::ptr& event) = 0;

    virtual void
    register_finish(ptr_type::ptr p) = 0;

    virtual void
    cancel_all_messages(const eventschedulerPtr& mod) = 0;

    virtual void
    end_gracefully()
    {
    }

    virtual void
    finish_stats();

    static bool stopped_;


    // ---- These are interface functions for PDES, they should
    // ----   only get called when running in parallel mode
    virtual void
    ipc_schedule(nodeaddress::ptr goingto, event::ptr &ev)
    {
      sst_throw(ssterror, "eventmanager::ipc_schedule - not implemented yo, not good");
    }

    virtual int
    get_lpid() const
    {
      return -1;
    }

    virtual native::partition::ptr
    get_partition() const
    {
      return NULL;
    }
    // ---------------------------------

  };

} // end of namespace sstmac
#endif
