/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/eventmodule.h>
#include <sstmac/common/config.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/eventmanager.h>
#include <sstmac/common/messages/sst_message.h>



namespace sstmac {

  void eventmodule::schedule(const timestamp &t,
      const eventhandler::ptr &handler,
      const sst_message::ptr &msg)
  {
#if SST_SANITY_CHECK
    if (eventman_ == 0)
      {
        sst_throw(nullerror, "eventscheduler::schedule: null event manager type %s!",
                  typeid(*this).name());
      }
    double delta_t = (t - now()).sec();
    if (delta_t < -1e-9){
        sst_throw(illformederror, "time has gone backwards %8.4e seconds", delta_t);
    }
    if (!handler){
        sst_throw(nullerror, "eventscheduler::schedule: null handler passed in");
    }
#endif
    event::ptr ev = simple_event::construct(msg, handler);
    //ev->set_module(eventscheduler::ptr(this));
    eventman_->schedule(t, ev);
  }

  void eventmodule::schedule(const timestamp &t, const event::ptr &ev)
  {
#if SST_SANITY_CHECK
      double delta_t = (t - now()).sec();
      if (delta_t < -1e-9){
          sst_throw(illformederror, "time has gone backwards %8.4e seconds", delta_t);
      }
#endif
      //ev->set_module(eventscheduler::ptr(this));
      eventman_->schedule(t, ev);
  }

  void
  eventmodule::set_eventmanager(const eventmanager::ptr &m)
  {
    eventman_ = m;

    std::list<stat_logger::ptr>::iterator it, end = need_reg_.end();
    for(it = need_reg_.begin(); it != end; it++){

        eventman_->register_stat(*it);
    }
    need_reg_.clear();
  }

  timestamp
  eventmodule::now() const
  {
#if SST_SANITY_CHECK
    if (eventman_ == 0)
    {
        sst_throw(nullerror, "eventscheduler::now: null event manager for eventscheduler %s!",
                  typeid(*this).name());
    }
#endif
    return eventman_->now();
  }

  void
  eventmodule::schedule_delay(const timestamp &delay, const event::ptr &ev)
  {
      schedule(now() + delay, ev);
  }

  void
  eventmodule::register_stat(stat_logger::ptr stat)
  {
      if (!eventman_){
          need_reg_.push_back(stat);
      }
      else {
          eventman_->register_stat(stat);
      }
  }

  void
  eventmodule::cancel_all_messages()
  {
        eventman_->cancel_all_messages(ptr(this));
  }

  void
  eventmodule::schedule_delay(
    const timestamp &delay,
    const eventhandler::ptr& handler,
    const sst_message::ptr& msg)
  {
      schedule(now() + delay, handler, msg);
  }

  void
  eventmodule::schedule_now(const event::ptr &ev)
  {
      schedule(now(), ev);
  }

  eventmanager::ptr
  eventmodule::get_eventmanager() const
  {
      return eventman_;
  }

} // end of namespace sstmac
