/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMACRO_COMMON_FACTORIES_FACTORY_H_INCLUDED
#define SSTMACRO_COMMON_FACTORIES_FACTORY_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/sim_parameters.h>

namespace sstmac
{
  class factory_type : virtual public ptr_type
  {
  protected:
    std::string factory_name_;

  public:
    typedef boost::intrusive_ptr<factory_type> ptr;

    void
    set_factory_name(const std::string& name)
    {
      factory_name_ = name;
    }

    std::string
    get_factory_name() const
    {
      return factory_name_;
    }

    virtual void
    init_factory_params(const sim_parameters::ptr& params)
    {
      //by default, nothing
    }

  };

  class SSTDesc_base
  {
  public:

    virtual factory_type::ptr
    build(const sim_parameters::ptr& params) = 0;

    virtual void
    clear();

  };

  class SSTFactory_base
  {
  public:
    static void
    add_to_map(const std::string& namestr, SSTDesc_base* desc,
        std::map<std::string, SSTDesc_base*> *m);

  };

  template<class T>
    class SSTFactoryTemplateBase : public SSTFactory_base
    {
    protected:
      typedef std::map<std::string, SSTDesc_base*> descr_map;
      static descr_map* object_map_;
      static const char* name_;

    public:
      static void
      register_name(const std::string& name, SSTDesc_base* descr)
      {
        if (!object_map_)
          object_map_ = new descr_map;
        add_to_map(name, descr, object_map_);
      }

      static void
      validate_map(const std::string& valname){
        if (!object_map_){
            sst_throw(illformederror,
                "could not find name %s for factory %s. no classes are registered",
                valname.c_str(),
                name_);
        }
      }
    };

  template<class T>
    class SSTFactory : public SSTFactoryTemplateBase<T>
    {

    public:
      typedef SSTFactoryTemplateBase<T> parent;
      typedef boost::intrusive_ptr<T> ptr;
      using SSTFactoryTemplateBase<T>::object_map_;
      using SSTFactoryTemplateBase<T>::name_;

      static ptr
      get_value(const std::string& valname, const sim_parameters::ptr& params)
      {
        parent::validate_map(valname);

        std::map<std::string, SSTDesc_base*>::const_iterator it =
            object_map_->find(valname), end = object_map_->end();

        if (it == end)
        {
          std::cerr << "Valid factories are:" << std::endl;
          for (it = object_map_->begin(); it != end; ++it)
          {
            std::cerr << it->first << std::endl;
          }
          sst_throw(valueerror, "could not find name %s for factory %s",
              valname.c_str(), name_);
        }

        SSTDesc_base* descr = it->second;
        if (!descr)
        {
          sst_throw(valueerror, "initialized name %s with null descr for factory %s",
              valname.c_str(), name_);
        }

        factory_type::ptr p = descr->build(params);
        p->set_factory_name(valname);
        ptr cast = boost::dynamic_pointer_cast<T>(p);
        cast->init_factory_params(params);
        return cast;
      }

      static ptr
      get_extra_param(const std::string& param_name,
            const sim_parameters::ptr& params)
      {
        if (params->has_param(param_name)){
            return get_param(param_name,params);
        }
        else {
            return ptr();
        }
      }

      static ptr
      get_param(const std::string& param_name,
          const sim_parameters::ptr& params)
      {
        std::string valname = params->get_param(param_name);
        return get_value(valname, params);
      }

      static ptr
      get_optional_param(const std::string& param_name,
          const std::string& defval, const sim_parameters::ptr& params)
      {
        std::string valname = params->get_optional_param(param_name, defval);
        return get_value(valname, params);
      }

    };

  template<class T, class U>
    class SSTFactory1InitParam : public SSTFactory<T>
    {
    private:
      typedef SSTFactory<T> parent;

    public:
      typedef boost::intrusive_ptr<T> ptr;

      static ptr
      get_value(const std::string& valname, const sim_parameters::ptr& params,
          const U& param1)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1);
        return p;
      }

      static ptr
      get_param(const std::string& param_name,
          const sim_parameters::ptr& params, const U& param1)
      {
        ptr p = parent::get_param(param_name, params);
        p->init(param1);
        return p;
      }

      static ptr
      get_optional_param(const std::string& param_name,
          const std::string& defval, const sim_parameters::ptr& params,
          const U& param1)
      {
        ptr p = parent::get_optional_param(param_name, defval, params);
        p->init(param1);
        return p;
      }
    };

  template<class T, class U, class V>
    class SSTFactory2InitParams : public SSTFactory<T>
    {
    private:
      typedef SSTFactory<T> parent;

    public:
      typedef boost::intrusive_ptr<T> ptr;

      static ptr
      get_value(const std::string& valname, const sim_parameters::ptr& params,
          const U& param1, const V& param2)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1, param2);
        return p;
      }

      static ptr
      get_param(const std::string& param_name,
          const sim_parameters::ptr& params, const U& param1, const V& param2)
      {
        ptr p = parent::get_param(param_name, params);
        p->init(param1, param2);
        return p;
      }

      static ptr
      get_optional_param(const std::string& param_name,
          const std::string& defval, const sim_parameters::ptr& params,
          const U& param1, const V& param2)
      {
        ptr p = parent::get_optional_param(param_name, defval, params);
        p->init(param1, param2);
        return p;
      }
    };

  template<class T, class U, class V, class W>
    class SSTFactory3InitParams : public SSTFactory<T>
    {
    private:
      typedef SSTFactory<T> parent;

    public:
      typedef boost::intrusive_ptr<T> ptr;

      static ptr
      get_value(const std::string& valname, const sim_parameters::ptr& params,
          const U& param1, const V& param2, const W& param3)
      {
        ptr p = parent::get_value(valname, params);
        p->init(param1, param2, param3);
        return p;
      }

      static ptr
      get_param(const std::string& param_name,
          const sim_parameters::ptr& params, const U& param1, const V& param2,
          const W& param3)
      {
        ptr p = parent::get_param(param_name, params);
        p->init(param1, param2, param3);
        return p;
      }

      static ptr
      get_optional_param(const std::string& param_name,
          const std::string& defval, const sim_parameters::ptr& params,
          const U& param1, const V& param2, const W& param3)
      {
        ptr p = parent::get_optional_param(param_name, defval, params);
        p->init(param1, param2, param3);
        return p;
      }
    };

  template<class Parent, class Child>
    class SSTFactory_desc : public SSTDesc_base
    {

    public:
      SSTFactory_desc(const char* name)
      {
        std::string n(name);
        /** This should only ever be declared within
         the static initialization routine. There is therefore
         nothing unsafe about passing "this" out.  This object
         will be persisent throughout the entire simulation */
        SSTFactory<Parent>::register_name(n, this);
      }

      factory_type::ptr
      build(const sim_parameters::ptr& params)
      {
        return Child::construct(params);
      }

    };

}

#define DeclareFactory(type_name) \
    typedef SSTFactory<type_name> type_name##_factory;

#define DeclareFactory1Param(type_name, param1_name) \
    typedef SSTFactory1Param<type_name, param1_name> type_name##_factory;

#define DeclareFactory1InitParam(type_name, param1_name) \
    typedef SSTFactory1InitParam<type_name, param1_name> type_name##_factory;

#define DeclareFactory2InitParams(type_name, param1_name, param2_name) \
    typedef SSTFactory2InitParams<type_name, param1_name, param2_name> type_name##_factory;

#define DeclareFactory3InitParams(type_name, param1_name, param2_name, param3_name) \
    typedef SSTFactory3InitParams<type_name, param1_name, param2_name, param3_name> type_name##_factory;

#define ImplementFactory(type_name) \
    namespace sstmac { \
    template<> const char* ::sstmac::SSTFactoryTemplateBase<type_name>::name_ = #type_name; \
    template<> std::map<std::string, SSTDesc_base*>* SSTFactoryTemplateBase<type_name>::object_map_ = 0; \
    } 

#define SSTRegister(cls_str, parent_cls, child_cls, ...) \
    static sstmac::SSTFactory_desc<parent_cls,child_cls> child_cls##_cd(cls_str)

#endif
