/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_PROCESSOR_EVENTDATA_H_INCLUDED
#define SSTMAC_HARDWARE_PROCESSOR_EVENTDATA_H_INCLUDED

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/ptr_type.h>
#include <stdint.h>

namespace sstmac {

  /**
   * Input for processor models that use
   * performance counter data. Is basically just a map
   * that maps std::string keys to integer values.
   * Keys are defined in the libraries that use them.
   */
  class eventdata :
    public library_message
  {

  protected:
    eventdata();

    eventdata(const std::string& libn);

  public:

    typedef boost::intrusive_ptr<eventdata> ptr;
    typedef boost::intrusive_ptr<const eventdata> const_ptr;

    static const char INST_FLOP[];
    static const char INST_MEMOP[];
    static const char INST_MEMOP_LOCAL[];
    static const char INST_MEMOP_RAND[];

    /**
      * Construct a new eventdata object.
      * @param libn The library name which is notified when this event is done
      * @param t The default multi-use time parameter
      * @return A smart pointer to an eventdata object.
      */
    static eventdata::ptr construct(const std::string& libn);

    virtual ~eventdata() {}

    /**
     *
     * @param key the key
     * @param val value to associate with the given key
     */
    void
    set_event_value(const std::string& key, unsigned long long int val);


    /**
      * Get event data
      * @param key the key
      * @return value associated with given key
      */
    unsigned long long int get_event_value(const std::string& key) const;

    unsigned long long int get_optional_event_value(const std::string& key) const;

    /**
      * Test for event key presence
      * @param key to test
      * @return true if this eventdata has the given key
      */
    bool has_event_value(const std::string& key) const;

    void set_event_time(const timestamp& t);

    const timestamp& get_event_time() const;

    void set_core(int core);

    int get_core() const;

  private:
    int core_;

    typedef std::map<std::string, unsigned long long int> data_map;

    data_map event_data_;

    timestamp event_time_;



  };

}  // end of namespace sstmac

#endif
