/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/messages/sst_metadata.h>
#include <sstmac/software/process/operatingsystem.h>

DeclareSerializable(sstmac::sst_message);

namespace sstmac
{

  sst_message::sst_message(MESSAGE_TYPES t) :
    msgtype_(t), is_scheduled_(false)
  {
  }

  sst_message::~sst_message()
  {
  }

  //
  // Allocate an object of this type.
  //

  sst_message::ptr
  sst_message::construct(MESSAGE_TYPES ty)
  {
    sst_message::ptr retval(new sst_message(ty));

    return retval;
  }

  void
  sst_message::serialize_order(sst_serializer* ser)
  {

    ser->do_enum(msgtype_);
  }


  const sw::taskid&
  sst_message::dest_task() const
  {
    throw illformederror("generic sst message does not have dest task");
  }

  const sw::taskid&
  sst_message::source_task() const
  {
    throw illformederror("generic sst message does not have src task");
  }

  long
  sst_message::source_thread(const sw::operatingsystem::ptr& os) const
  {
    throw illformederror("generic sst message does not have source thread");
  }

  long
  sst_message::dest_thread(const sw::operatingsystem::ptr& os) const
  {
    throw illformederror("generic sst message does not have dest thread");
  }

  void
  sst_message::add_delay(const timestamp &delay)
  {
  }

  bool
  sst_message::has_callback() const
  {
    return false;
  }

  void
  sst_message::add_callback(const eventhandler::ptr& callback)
  {
    sst_throw(unimplementederror, "cannot call add_callback on message %s",
        to_string().c_str());
  }

  eventhandler::ptr
  sst_message::pop_callback()
  {
    sst_throw(unimplementederror, "cannot call pop_callback on message %s",
        to_string().c_str());
  }

  sst_message::ptr
  sst_message::clone(MESSAGE_TYPES t) const
  {
    return construct(t);
  }

  sst_message::ptr
  sst_message::clone() const
  {
    return construct(msgtype_);
  }

  void
  sst_message::mutate(mutation_t mut)
  {
    throw unimplementederror("message cannot be mutated");
  }

  long
  sst_message::get_byte_length() const
  {
    return 0;
  }

  const timestamp&
  sst_message::get_event_time() const
  {
    throw unimplementederror(
        "sst_message does not implement get_event_time interface");
  }

  void
  sst_message::set_event_time(const timestamp &t)
  {
    throw unimplementederror(
        "sst_message does not implement set_event_time interface");
  }

  int
  sst_message::get_core() const
  {
    throw unimplementederror(
        "sst_message does not implement get_core interface");
  }

  void
  sst_message::set_core(int core)
  {
    throw unimplementederror(
        "sst_message does not implement set_core interface");
  }

  bool
  sst_message::is_network_message() const
  {
    return false;
  }

  bool
  sst_message::get_needs_ack() const
  {
    sst_throw(unimplementederror, "cannot call get needs ack on %s",
        to_string().c_str());
  }

  nodeaddress::ptr
  sst_message::fromaddr() const
  {
    sst_throw(unimplementederror, "cannot call fromaddr on message %s",
        to_string().c_str());
  }

  nodeaddress::ptr
  sst_message::toaddr() const
  {
    sst_throw(unimplementederror, "cannot call toaddr on message %s",
        to_string().c_str());
  }

  void
  sst_message::set_fromaddr(const nodeaddress::ptr &addr)
  {
    sst_throw(unimplementederror, "cannot call set_toaddr on message %s",
              to_string().c_str());
  }

  void
  sst_message::set_toaddr(const nodeaddress::ptr &addr)
  {
    sst_throw(unimplementederror, "cannot call set_toaddr on message %s",
              to_string().c_str());
  }

  sw::AMI::AMI_COMP_FUNC
  sst_message::get_function() const
  {
    throw unimplementederror(
        "sst_message does not implement get_function interface");
  }

  bool
  sst_message::has_event_value(const std::string& key) const
  {
    throw unimplementederror(
        "sst_message does not implement has_event_value interface");
  }

  bool
  sst_message::is_tail() const
  {
    return true;
  }

  std::string
  sst_message::to_string() const
  {
    return sstprintf("sst_message of type %s", tostr(msgtype_));
  }

  sst_message::ptr
  sst_message::get_parent_msg() const
  {
    return sst_message::ptr(const_cast<sst_message*> (this));
  }

  unsigned long long
  sst_message::get_event_value(const std::string& key) const
  {
    throw unimplementederror(
        "sst_message does not implement get_event_value interface");
  }

  unsigned long long
  sst_message::get_optional_event_value(const std::string& key) const
  {
    throw unimplementederror(
        "sst_message does not implement get_optional_event_value interface");
  }

  void
  sst_message::set_event_value(const std::string& key, unsigned long long val)
  {
    throw unimplementederror(
        "sst_message does not implement set_event_value interface");
  }

#define casestr(x) case x: return #x
  const char*
  sst_message::tostr(MESSAGE_TYPES t)
  {
      switch(t)
      {
          casestr(SST);
          casestr(MPI);
          casestr(ACTOR);
          casestr(SEND_ACK);
          casestr(COMPUTE);
          casestr(LAUNCH);
          casestr(DISK);
          casestr(FILEMETA);
          casestr(FILEDATA);
          casestr(DATA);
          casestr(ACK);
          casestr(HPX);
          casestr(FAILURE);
          casestr(SDMA);
          casestr(CREME);
          casestr(EIGER);
          casestr(NETWORK);
          casestr(RDMA_GET_REQ);
          casestr(RDMA_PUT_REQ);
          casestr(RDMA_GET_RSP);
          casestr(RDMA_PUT_RSP);
      }
      sst_throw(valueerror, "sst_message: invalid message type tostr %d", t);
  }

  const char*
  sst_message::tostr(mutation_t mut)
  {
    switch (mut)
    {
    casestr(RDMA_GET_REQ_TO_RSP)
      ;
    casestr(RDMA_PUT_REQ_TO_RSP)
      ;
    casestr(GNI_HEADER_TO_RDMA_GET_REQ)
      ;
    casestr(GNI_HEADER_TO_RDMA_PUT_REQ)
      ;
    }
    sst_throw(valueerror, "sst_message: invalid mutation type tostr %d", mut);
  }

  std::string
  sst_message::get_lib_name() const
  {
    sst_throw(illformederror, "invalid use of get_lib_name non-library message %s", to_string().c_str());
  }

  library_message::library_message(const std::string& libn) :
    lib_name_(libn)
  {
  }

  library_message::library_message()
  {
  }

  std::string
  library_message::get_lib_name() const
  {
    return lib_name_;
  }

  void
  library_message::set_lib_name(const std::string& libn)
  {
    lib_name_ = libn;
  }

  timed_message::timed_message()
  {
  }

  timed_message::timed_message(const timestamp& t) :
    time_(t)
  {
  }

  timed_message::ptr
  timed_message::construct(const timestamp& t)
  {
    return timed_message::ptr(new timed_message(t));
  }

  timestamp
  timed_message::get_time() const
  {
    return time_;
  }

  void
  timed_message::set_time(const timestamp& time)
  {
    time_ = time;
  }

  sleep_message::sleep_message(const std::string& libname, const timestamp& t) :
    library_message(libname), timed_message(t)
  {
  }

  sleep_message::ptr
  sleep_message::construct(const std::string& libname, const timestamp& t)
  {
    return new sleep_message(libname, t);
  }

  display_message::display_message(const std::string& libname,
      MESSAGE_TYPES ty, int k) :
    sst_message(ty), library_message(libname), kind_message(k)
  {
  }

  display_message::ptr
  display_message::construct(const std::string& libname, MESSAGE_TYPES ty,
      int k)
  {
    return new display_message(libname, ty, k);
  }

  kind_message::kind_message(int k) :
    kind_(k)
  {
  }

  kind_message::ptr
  kind_message::construct(int k)
  {
    return new kind_message(k);
  }

  int
  kind_message::get_kind() const
  {
    return kind_;
  }

  void
  kind_message::set_kind(int k)
  {
    kind_ = k;
  }

  callback_message::callback_message()
  {
  }

  bool
  callback_message::has_callback() const
  {
    return !callbacks_.empty();
  }

  void
  callback_message::add_callback(const eventhandler::ptr& callback)
  {
    callbacks_.push_front(callback);
  }

  eventhandler::ptr
  callback_message::pop_callback()
  {
    if (callbacks_.empty()){
        sst_throw(illformederror,
            "sst_message::pop_callback: callback list is empty for %s",
            to_string().c_str());
    }
    eventhandler::ptr tmp = callbacks_.front();
    callbacks_.pop_front();
    return tmp;
  }

} // end of namespace sstmac
