/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_SST_MESSAGE_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_SST_MESSAGE_H_INCLUDED

#include <sstmac/common/nodeaddress.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/ami/ami.h>

#include <sstmac/common/messages/sst_serializer.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/malloc.h>

#include <sstmac/common/factories/factory.h>
#include <sstmac/common/messages/serializable.h>

namespace sstmac
{
  
  class eventcallback;
  class eventhandler;

  namespace sw { class operatingsystem; }

  /**
   * A class describing an event.
   */
  class sst_message :
    public serializable,
    public serializable_type<sst_message>
  {
  ImplementSerializable(sst_message)

  public:

    typedef boost::intrusive_ptr<sst_message> ptr;
    
    /**
     * Child classes should register their type here, for convenience
     */
    enum MESSAGE_TYPES
    {
      SST,
      MPI,
      ACTOR,
      SEND_ACK,
      COMPUTE,
      LAUNCH,
      DISK,
      DATA,
      ACK,
      FILEMETA,
      FILEDATA,
      HPX,
      FAILURE,
      SDMA,
      NETWORK,
      RDMA_GET_REQ,
      RDMA_PUT_REQ,
      RDMA_GET_RSP,
      RDMA_PUT_RSP,
      CREME,
      EIGER
    };

    typedef enum {
        RDMA_GET_REQ_TO_RSP,
        RDMA_PUT_REQ_TO_RSP,
        GNI_HEADER_TO_RDMA_GET_REQ,
        GNI_HEADER_TO_RDMA_PUT_REQ
    } mutation_t;

  protected:

    /**
     * Protected constructor
     * @param t The type (child class, if any) of this message
     */
    sst_message(MESSAGE_TYPES ty = SST);

  public:

    virtual ~sst_message();

    /**
     * Smart pointer constructor
     * @param libn The name of the library this message should be delivered to
     * @param t  A multi-use time parameter
     * @param ty Message type (child class)
     * @param k An int describing the kind, multi-use, non-critical
     * @return A smart pointer to an sst_message
     */
    static ptr
    construct(MESSAGE_TYPES ty = SST);

    virtual void
    add_callback(const boost::intrusive_ptr<eventhandler>& callback);

    virtual bool has_callback() const;

    virtual boost::intrusive_ptr<eventhandler>
    pop_callback();

    virtual bool
    cancellable() const {
      return true;
    }

    /**
     * Stringifier
     * @return String description
     */
    virtual std::string
    to_string() const;

    /**
     * Null pointer convenience function
     * @return A null smart pointer
     */
    static ptr
    null()
    {
      ptr ret;
      return ret;
    }
    // --------------------------------------//

  public:

    static const char*
    tostr(MESSAGE_TYPES ty);

    static const char*
    tostr(mutation_t mut);


    /**
     * Copy smart constructor
     * @return A smart pointer to an identical but different sst_message
     */
    virtual ptr clone() const;

    /**
     * Convenient smart constructor which changes the message type
     * @param t new message type
     * @return A smart pointer to a new sst_message with the given message type
     */
    virtual ptr clone(MESSAGE_TYPES t) const;

    virtual void mutate(mutation_t mut);

    /**
     * Virtual function to return size. Child classes should impement this
     * if they want any size tracked / modeled.
     * @return Zero size, meant to be implemented by children.
     */
    virtual long get_byte_length() const;

    virtual void add_delay(const timestamp& delay);


    /**
     * Serialize this message during parallel simulation.
     * @param ser The serializer to use
     */
    virtual void
    serialize_order(sst_serializer* ser);


    /**
     * Message type getter
     * @return message type (child class)
     */
    MESSAGE_TYPES
    get_type() const
    {
      return msgtype_;
    }

    void
    set_type(MESSAGE_TYPES t)
    {
      msgtype_ = t;
    }


    /**
     * Library name getter
     * @return Name of deliver-to library
     */
    virtual std::string
    get_lib_name() const;


    bool
    is_scheduled() const
    {
      return is_scheduled_;
    }

    void
    set_scheduled(bool sched)
    {
      is_scheduled_ = sched;
    }

    virtual nodeaddress::ptr
    toaddr() const;

    virtual nodeaddress::ptr
    fromaddr() const;

    virtual void
    set_toaddr(const nodeaddress::ptr& addr);

    virtual void
    set_fromaddr(const nodeaddress::ptr& addr);

    virtual bool get_needs_ack() const;

    virtual bool is_tail() const;

    virtual bool is_network_message() const;

    virtual sst_message::ptr get_parent_msg() const;

    virtual const sw::taskid&
    source_task() const;

    virtual const sw::taskid&
    dest_task() const;

    virtual long
    source_thread(const boost::intrusive_ptr<sw::operatingsystem>& os) const;

    virtual long
    dest_thread(const boost::intrusive_ptr<sw::operatingsystem>& os) const;

    virtual int
    get_core() const;

    virtual void
    set_core(int core);

    virtual sw::AMI::AMI_COMP_FUNC
    get_function() const;

    /**
     *
     * @param key the key
     * @param val value to associate with the given key
     */
    virtual void
    set_event_value(const std::string& key, unsigned long long int val);

    /**
      * Get event data
      * @param key the key
      * @return value associated with given key
      */
    virtual unsigned long long int get_event_value(const std::string& key) const;

    virtual unsigned long long int get_optional_event_value(const std::string& key) const;

    /**
      * Test for event key presence
      * @param key to test
      * @return true if this eventdata has the given key
      */
    virtual bool has_event_value(const std::string& key) const;

    virtual void set_event_time(const timestamp& t);

    virtual const timestamp& get_event_time() const;

  protected:

    //used by libraries to distinguish categories of messages
    MESSAGE_TYPES msgtype_;

    bool is_scheduled_;


  };

  class callback_message :
    virtual public sst_message
  {
    protected:
        typedef std::list<boost::intrusive_ptr<eventhandler> > callback_queue;
        callback_queue callbacks_;

        callback_message();

    public:
        typedef boost::intrusive_ptr<callback_message> ptr;

        virtual void
        add_callback(const boost::intrusive_ptr<eventhandler>& callback);

        virtual bool has_callback() const;

        virtual boost::intrusive_ptr<eventhandler>
        pop_callback();
  };

  class library_message :
    virtual public sst_message
  {

    protected:
        std::string lib_name_;

        library_message(const std::string& libname);

        library_message();

    public:
        typedef boost::intrusive_ptr<library_message> ptr;

        /**
         * Library name getter
         * @return Name of deliver-to library
         */
        std::string
        get_lib_name() const;

        /**
         * Library name setter
         * @param s Name of deliver-to library
         */
        void
        set_lib_name(const std::string& s);
        
  };

  class timed_message :
    virtual public sst_message
  {
    protected:
        timed_message();

        timed_message(const timestamp& t);

        timestamp time_;

    public:

        typedef boost::intrusive_ptr<timed_message> ptr;

        static ptr
        construct(const timestamp& t);

        /**
         * Time getter
         * @return time field
         */
        timestamp
        get_time() const;

        /**
         * Time setter
         * @param t time value
         */
        void
        set_time(const timestamp& t);

  };

  class sleep_message :
    public library_message,
    public timed_message
  {
    protected:
        sleep_message(const std::string& libname, const timestamp& t);

    public:

        typedef boost::intrusive_ptr<sleep_message> ptr;

        static ptr
        construct(const std::string& libname, const timestamp& t);
        
  };

  class kind_message :
    virtual public sst_message
 {
    protected:
        int kind_;
        
        kind_message(int k);

    public:

        typedef boost::intrusive_ptr<kind_message> ptr;

        static ptr
        construct(int k);

        /**
         * Message kind getter
         * @return message kind
         */
        int get_kind() const;

        /**
         * Message kind setter
         * @param k message kind value
         */
        void set_kind(int k);
 };

  class display_message :
    public library_message,
    public kind_message
  {
    protected:
        display_message(const std::string& libname, MESSAGE_TYPES ty, int k);

    public:

        typedef boost::intrusive_ptr<display_message> ptr;

        static ptr
        construct(const std::string& libname, MESSAGE_TYPES ty, int k);
        
  };




} // end of namespace sstmac
#endif
