/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_MESSAGES_SST_SERIALIZER_H_INCLUDED
#define SSTMAC_COMMON_MESSAGES_SST_SERIALIZER_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/common/richpair.h>
#include <sstmac/common/sstmac-config.h>
//#include <sstmac/common/messages/serializable.h>  //forward delcare
#include <typeinfo>
#include <sstmac/common/debug.h>

namespace sstmac
{

  class serializable;



#define SER_PACKER_PURE_VIRTUAL_PACK(thetype) virtual void pack(thetype t) = 0;

  class ser_packer :
    public ptr_type
  {

  protected:
    ser_packer()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> packer | parallel");
        logger::register_logger(&dbg_);
      }
    }

    static debug dbg_;

  public:
    typedef boost::intrusive_ptr<ser_packer> ptr;

    SER_PACKER_PURE_VIRTUAL_PACK(int)
    SER_PACKER_PURE_VIRTUAL_PACK(float)
    SER_PACKER_PURE_VIRTUAL_PACK(short)
    SER_PACKER_PURE_VIRTUAL_PACK(double)
    SER_PACKER_PURE_VIRTUAL_PACK(long int)
    SER_PACKER_PURE_VIRTUAL_PACK(long long int)
    SER_PACKER_PURE_VIRTUAL_PACK(unsigned int)
    SER_PACKER_PURE_VIRTUAL_PACK(long unsigned int)
    SER_PACKER_PURE_VIRTUAL_PACK(long long unsigned int)
    SER_PACKER_PURE_VIRTUAL_PACK(char)
    SER_PACKER_PURE_VIRTUAL_PACK(bool)
    SER_PACKER_PURE_VIRTUAL_PACK(std::string)
    SER_PACKER_PURE_VIRTUAL_PACK(timestamp)

    virtual void pack_array(void* arr, size_t size) = 0;

    virtual
    ~ser_packer()
    {
    }


    virtual std::string
    to_string() const
    {
      return "ser_packer";
    }
  };
#undef SER_PACKER_PURE_VIRTUAL_PACK

#define SER_UNPACKER_PURE_VIRTUAL_PACK(thetype) virtual void unpack(thetype &t) = 0;

  class ser_unpacker : public ptr_type {

  protected:
    ser_unpacker()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> unpacker | parallel");
        logger::register_logger(&dbg_);
      }
    }
    static debug dbg_;

  public:

    typedef boost::intrusive_ptr<ser_unpacker> ptr;
    SER_UNPACKER_PURE_VIRTUAL_PACK(int)
    SER_UNPACKER_PURE_VIRTUAL_PACK(float)
    SER_UNPACKER_PURE_VIRTUAL_PACK(short)
    SER_UNPACKER_PURE_VIRTUAL_PACK(double)
    SER_UNPACKER_PURE_VIRTUAL_PACK(long)
    SER_UNPACKER_PURE_VIRTUAL_PACK(long long)
    SER_UNPACKER_PURE_VIRTUAL_PACK(unsigned int)
    SER_UNPACKER_PURE_VIRTUAL_PACK(unsigned long)
    SER_UNPACKER_PURE_VIRTUAL_PACK(unsigned long long)
    SER_UNPACKER_PURE_VIRTUAL_PACK(char)
    SER_UNPACKER_PURE_VIRTUAL_PACK(bool)
    SER_UNPACKER_PURE_VIRTUAL_PACK(std::string)
    SER_UNPACKER_PURE_VIRTUAL_PACK(timestamp)

    virtual void unpack_array(void* arr, size_t size) = 0;

    virtual
    ~ser_unpacker()
    {
    }



    virtual std::string
    to_string() const
    {
      return "ser_unpacker";
    }
  };
#undef SER_UNPACKER_PURE_VIRTUAL_PACK


#define SER_SIZER_PURE_VIRTUAL_PACK(thetype) virtual void size(thetype t) = 0;

  class ser_sizer : public ptr_type {

  protected:
    ser_sizer() :
      size_(0)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> sizer | parallel");
        logger::register_logger(&dbg_);
      }
    }

    static debug dbg_;

    size_t size_;
  public:
    typedef boost::intrusive_ptr<ser_sizer> ptr;

    SER_SIZER_PURE_VIRTUAL_PACK(int)
    SER_SIZER_PURE_VIRTUAL_PACK(float)
    SER_SIZER_PURE_VIRTUAL_PACK(short)
    SER_SIZER_PURE_VIRTUAL_PACK(double)
    SER_SIZER_PURE_VIRTUAL_PACK(long int)
    SER_SIZER_PURE_VIRTUAL_PACK(long long int)
    SER_SIZER_PURE_VIRTUAL_PACK(unsigned int)
    SER_SIZER_PURE_VIRTUAL_PACK(long unsigned int)
    SER_SIZER_PURE_VIRTUAL_PACK(long long unsigned int)
    SER_SIZER_PURE_VIRTUAL_PACK(char)
    SER_SIZER_PURE_VIRTUAL_PACK(bool)
    SER_SIZER_PURE_VIRTUAL_PACK(std::string)
    SER_SIZER_PURE_VIRTUAL_PACK(timestamp)


    void
    add_obj_overhead() {
      size_ += sizeof(long);
    }

    void
    add(size_t s){
      size_ += s;
    }

    size_t
    get_size() {
      return size_;
    }

    virtual ~ser_sizer() { }

    virtual std::string
    to_string() const {
      return "ser_sizer";
    }
  };
#undef SER_SIZER_PURE_VIRTUAL_PACK

  class sst_serializer;

  template<typename T>
  class obj_deserializer {
  public:
    static boost::intrusive_ptr<T>
    deserialize_object(long cls_id, sst_serializer* ser);

    static bool
    sanity_check(serializable* ser, long cls_id);
  };

  /**
    * This class is basically a wrapper for objects to declare the order in
    * which their members should be ser/des
    */
   class sst_serializer
   {
   public:
     enum SERIALIZE_MODE
     {
       SIZER, PACK, UNPACK
     };
     SERIALIZE_MODE mode_;

   protected:
     typedef boost::intrusive_ptr<serializable> serializablePtr;
     typedef boost::intrusive_ptr<const serializable> serializableConstPtr;

     //only one of these is going to be valid for this sst_serializer
     //  not very good class design, but a little more convenient
     ser_packer::ptr packer_;
     ser_unpacker::ptr unpacker_;
     ser_sizer::ptr sizer_;

     static const long null_ptr_id = -1;

    static debug dbg_;

  public:

    sst_serializer(ser_packer::ptr p) :
      packer_(p)
    {
      mode_ = PACK;
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> parallel | serializer");
        logger::register_logger(&dbg_);
      }
    }

    sst_serializer(ser_unpacker::ptr p) :
      unpacker_(p)
    {
      mode_ = UNPACK;
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> parallel | serializer");
        logger::register_logger(&dbg_);
      }
    }

    sst_serializer(ser_sizer::ptr p) :
      sizer_(p)
    {
      mode_ = SIZER;
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> parallel | serializer");
        logger::register_logger(&dbg_);
      }
    }

    virtual
    ~sst_serializer()
    {
    }

    SERIALIZE_MODE
    get_mode() const {
      return mode_;
    }

    template<typename T>
      void
      do_primitive(T &t)
      {
        if (mode_ == SIZER)
        {
          sizer_->size(t);
        }
        else if (mode_ == PACK)
        {
          packer_->pack(t);
        }
        else if (mode_ == UNPACK)
        {
          unpacker_->unpack(t);
        }
      }


    template<typename T>
      void
      do_enum(T &t)
      {

        if (mode_ == SIZER)
        {
          short s = (short)t;
          sizer_->size(s);
        }
        else if (mode_ == PACK)
        {
          short s = (short)t;
          packer_->pack(s);
        }
        else if (mode_ == UNPACK)
        {
          short s;
          unpacker_->unpack(s);
          t = (T)s;
        }
      }


    /**
     * size is in bytes
     */
    template<typename T>
      void
      do_array(T* t, size_t size)
      {
        if (mode_ == SIZER)
        {
          sizer_->add(size);
        }
        else if (mode_ == PACK)
        {
          packer_->pack_array(t, size);
        }
        else if (mode_ == UNPACK)
        {
          unpacker_->unpack_array(t, size);
        }
      }

    template<typename T>
      void
      do_stl_vector(std::vector<T> &v)
      {
        if (mode_ == SIZER)
        {
          sizer_->size(v.size());

          for(int i = 0; i < v.size(); i++){
            sizer_->size(v[i]);
          }
        }
        else if (mode_ == PACK)
        {
          packer_->pack(v.size());

          for (int i = 0; i < v.size(); i++)
          {
            packer_->pack(v[i]);
          }
        }
        else if (mode_ == UNPACK)
        {
          size_t s;
          unpacker_->unpack(s);
          v.resize(s);
          for(int i = 0; i < s; i++){
            unpacker_->unpack(v[s]);
          }

        }
      }

    template<typename T>
      void
      do_stl_vector_obj(std::vector<boost::intrusive_ptr<T> > &v)
      {
        if (mode_ == SIZER)
        {
          sizer_->size(v.size());

          for (int i = 0; i < v.size(); i++)
          {
            do_obj(v[i]);
          }
        }
        else if (mode_ == PACK)
        {
          packer_->pack(v.size());

          for (int i = 0; i < v.size(); i++)
          {
            do_obj(v[i]);
          }
        }
        else if (mode_ == UNPACK)
        {
          size_t s;
          unpacker_->unpack(s);
          v.resize(s);
          for (int i = 0; i < s; i++)
          {
            do_obj(v[s]);
          }

        }
      }

    template<typename T>
      void
      do_stl_vector_const_obj(std::vector<boost::intrusive_ptr<const T> > &v)
      {
        if (mode_ == SIZER)
        {
          sizer_->size(v.size());

          for (int i = 0; i < v.size(); i++)
          {
            do_const_obj(v[i]);
          }
        }
        else if (mode_ == PACK)
        {
          packer_->pack(v.size());

          for (int i = 0; i < v.size(); i++)
          {
            do_const_obj(v[i]);
          }
        }
        else if (mode_ == UNPACK)
        {
          size_t s;
          unpacker_->unpack(s);
          v.resize(s);
          for (int i = 0; i < s; i++)
          {
            do_const_obj(v[i]);
          }

        }
      }



    template<typename T>
      void
      do_obj(boost::intrusive_ptr<T> &o)
      {
        SSTMAC_DEBUG << "serializer (do_obj)...";
        if (mode_ == SIZER)
        {
          SSTMAC_DEBUG << "sizer\n";
          sizer_->add_obj_overhead();
          if (o)
            o->serialize_order((this));
        }
        else if (mode_ == PACK)
        {
          SSTMAC_DEBUG << "packing...";
          if (o)
          {
#if SST_SANITY_CHECK
            if (!obj_deserializer<T>::sanity_check(o.get(), o->cls_id()))
            {
              sst_throw(ssterror, "trying to pack an object %s that might not have inherited from serializable_type<T>", o->to_string().c_str());
            }
#endif
            SSTMAC_DEBUG << "object with class id " << o->cls_id() << ": "
                  << o->to_string() << "\n";
            packer_->pack(o->cls_id());
            o->serialize_order((this));
          }
          else
          {
            SSTMAC_DEBUG << "null object \n";
            packer_->pack(null_ptr_id);
          }
        }
        else if (mode_ == UNPACK)
        {
          SSTMAC_DEBUG << "unpacking...";
          long cls_id;
          unpacker_->unpack(cls_id);
          if (cls_id == null_ptr_id)
          {
            SSTMAC_DEBUG << "null pointer object \n";
            o = boost::intrusive_ptr<T>();
          }
          else
          {
            SSTMAC_DEBUG << "object with class id "
                  << cls_id << "\n";
            o = obj_deserializer<T>::deserialize_object(cls_id, this);
            SSTMAC_DEBUG << "unpacked object " << o->to_string()
                  << "\n";

          }
        }
      }

   /** template<class T>
      class Deconstifier
      {
      public:
        typedef T unconst_type;
      };

    template<class T>
      class Deconstifier<const T>
      {
      public:
        typedef T unconst_type;
      };

    template <class constPTR>
      void
      do_const_obj(constPTR &o)
      {
        typedef typename constPTR::element_type constT;
        typedef typename Deconstifier<constT>::unconst_type T;
        typedef typename boost::intrusive_ptr<T> PTR;
        PTR ptr = boost::dynamic_pointer_cast<>o->serialization_clone();
        do_obj(ptr);
        o = ptr;

      } */

    template<typename T>
      void
      do_const_obj(boost::intrusive_ptr<const T> &o)
      {
        boost::intrusive_ptr<T> p;
        if (o)
        {
          SSTMAC_DEBUG << "sst_serializer::do_const_obj - cloning " << o->to_string() << "...";
          p = boost::dynamic_pointer_cast<T>(o->serialization_clone());
          SSTMAC_DEBUG << " got " << p->to_string() << "\n";
        }
        do_obj(p);
        o = p;
      }

    //const reference wrapper
    template<typename T>
      void
      do_obj(const boost::intrusive_ptr<T> &o)
      {
        if(mode_ == UNPACK){
          sst_throw(ssterror, "sst_serializer::do_obj(const &) - can't unpack with a const ref");
        }
        boost::intrusive_ptr<T> thing = o;
        do_obj(thing);
      }

   };

} // end of namespace sstmac
#endif
