/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_NODEADDRESS_H_INCLUDED
#define SSTMAC_COMMON_NODEADDRESS_H_INCLUDED

#include <sstmac/common/messages/sst_serializer.h>
#include <sstmac/common/messages/serializable.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/logger.h>

#include <sstmac/common/errors.h>

namespace sstmac
{

  /**
   * The abstract base class for an address of a node. This allows different
   * backends to structure themselves however they want and have convenient
   * addressing.
   */
  class nodeaddress :
    public serializable
  {
  public:
      typedef boost::intrusive_ptr<nodeaddress> ptr;

      static nodeaddress::ptr any_addr;

  protected:
    nodeaddress()
    {
    }

    /**
     * Protected internal class to convert a long unique id
     * back into a smart nodeaddress pointer.  An implementation
     * of this should be set to nodeaddress::converter_
     */
    class address_converter : public ptr_type
    {
    public:

      virtual nodeaddress::ptr
      convert(long uid) = 0;

      virtual
      ~address_converter()
      {
      }

      typedef boost::intrusive_ptr<address_converter> ptr;

      virtual std::string
      to_string() const
      {
        return "address_converter";
      }
    };

    static address_converter::ptr converter_;

  public:
    virtual std::string
    to_string() const
    {
      return "nodeaddress";
    }

    virtual
    ~nodeaddress()
    {

    }

    /**
     * Comparer of nodeaddresses for sorted collections.
     */
    struct nodeaddresscomp
    {
      bool
      operator()(const nodeaddress::ptr& lhs, const nodeaddress::ptr& rhs) const
      {
        return lhs->unique_id() < rhs->unique_id();
      }
    };

    typedef boost::unordered_map<long, nodeaddress::ptr> id_to_node_map_t;
    static id_to_node_map_t id_to_node_map_;

    static nodeaddress::ptr
    get_addr(long id);

    virtual bool
    equals(const nodeaddress::ptr &p) const = 0;

    /// This needs to happen for working with topologies and network models in SSTMM
    virtual long
    unique_id() const = 0;

    static nodeaddress::ptr
    convert_from_uniqueid(long uid)
    {
      return converter_->convert(uid);
    }
    
  };

  inline std::ostream&
  operator<<(std::ostream &os, const nodeaddress &n)
  {
    return (os << n.to_string());
  }
  inline logger&
  operator<<(logger &os, const nodeaddress &n)
  {
    return (os << n.to_string());
  }

  bool
  operator==(const nodeaddress::ptr &a, const nodeaddress::ptr &b);

  bool
  operator!=(const nodeaddress::ptr &a, const nodeaddress::ptr &b);

  class any_node_address :
      public nodeaddress,
      public serializable_type<any_node_address>
  {
    ImplementSerializable(any_node_address)

    public:
        any_node_address(){}

        static const long id = -111;

        long unique_id() const;

        bool equals(const nodeaddress::ptr& p) const;

        void serialize_order(sst_serializer *ser);




  };

} // end of namespace sstmac


namespace boost
{
  size_t
  hash_value(const sstmac::nodeaddress::ptr &t);
}
#endif
