
/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/driver_util.h>

namespace sstmac {
  debug sim_parameters::dbg_;


  void
  sim_parameters::get_vector_param(const std::string& key, std::vector<std::string>& vals) const
  {
      std::deque<std::string> tok;
      std::string space = " ";
      std::string param_value_str = get_param(key);
      pst::BasicStringTokenizer::tokenize(param_value_str, tok, space);
      std::deque<std::string>::const_iterator it, end = tok.end();
      for (it = tok.begin(); it != end; ++it)
      {
          std::string core = *it;
          if (core.size() > 0)
            vals.push_back(core);
      }
  }

  void
  sim_parameters::register_optional_param(const std::string &key, const std::string &val) const
  {
      // every time you const cast pointer this in a const function
      // god murders a baby dolphin
      const_cast<sim_parameters*>(this)->optional_params_read_[key].insert(val);
  }

  std::string
  sim_parameters::get_optional_param(const std::string &key, const std::string &def) const
  {
    if (has_param(key))
      return get_param(key);
    register_optional_param(key, def);
    return def;
  }

  sim_parameters::ptr
  sim_parameters::get_param_scope(const std::string& ns)
  {
    return ptr(this);
  }

  long
  sim_parameters::get_long_param(const std::string &key) const
  {
      std::string v = get_param(key);
      const char* begin = v.c_str();
      char* end = const_cast<char*>(begin);
      long ret = ::strtol(begin, &end, 0);
      if (begin == end){
          sst_abort("sim_parameters::get_long_param: param %s with value %s is not formatted as an integer",
                    key.c_str(), v.c_str());
      }
      return ret;
  }

  long
  sim_parameters::deprecated_long_param(const std::string &key) const
  {
    return get_long_param(key);
  }

  long
  sim_parameters::deprecated_optional_long_param(const std::string &key, long def) const
  {
    return get_optional_long_param(key, def);
  }

  long
  sim_parameters::get_optional_long_param(const std::string &key, long def) const
  {
    if (has_param(key))
      return get_long_param(key);

    register_optional_param(key, sstprintf("%ld", def));
    return def;
  }


  timestamp
  sim_parameters::get_time_param(const std::string& key) const
  {
    bool error = false;
    return get_timestamp(get_param(key).c_str(), error);

  }

  timestamp
  sim_parameters::deprecated_time_param(const std::string &key) const
  {
    return get_time_param(key);
  }

  timestamp
  sim_parameters::get_optional_time_param(const std::string &key, timestamp def) const
  {
    if (has_param(key))
      return get_time_param(key);

    register_optional_param(key, sstprintf("%fs", def.sec()));
    return def;
  }

  timestamp
  sim_parameters::deprecated_optional_time_param(const std::string &key, timestamp def) const
  {
    return get_optional_time_param(key, def);
  }

  double
  sim_parameters::reread_double_param(const std::string &key) const
  {
    return get_double_param(key);
  }

  double
  sim_parameters::reread_optional_double_param(const std::string &key, double def) const
  {
    return get_optional_double_param(key, def);
  }

  double
  sim_parameters::get_double_param(const std::string& key) const
  {
      std::string v = get_param(key);
      const char* begin = v.c_str();
      char* end = const_cast<char*>(begin);
      double ret = ::strtod(begin, &end);
      if (begin == end){
          sst_abort("sim_parameters::get_double_param: param %s with value %s is not formatted as a double",
                    key.c_str(), v.c_str());
      }
      return ret;
  }

  double
  sim_parameters::get_optional_double_param(const std::string &key, double def) const
  {
    if (has_param(key))
      return get_double_param(key);

    register_optional_param(key, sstprintf("%16.12e", def));
    return def;
  }

  double
  sim_parameters::deprecated_double_param(const std::string &key) const
  {
      return get_double_param(key);
  }

  double
  sim_parameters::deprecated_optional_double_param(const std::string &key, double def) const
  {
    return get_optional_double_param(key, def);
  }

  sim_parameters::ptr
  sim_parameters::parse_param(const std::string& p, std::pair<std::string, std::string>& keypair)
  {
    std::string lhs = p.substr(0, p.find("="));
    std::string rhs = p.substr(p.find("=") + 1);

    sim_parameters::ptr cur_scope = this;

    size_t ns_pos = lhs.find(".");
    while (ns_pos != std::string::npos){
        std::string ns = lhs.substr(0, ns_pos);
        lhs  = lhs.substr(ns_pos+1);
        if (ns == "**"){
            break;
        }
        cur_scope = cur_scope->get_param_scope(ns);
        ns_pos = lhs.find(".");
    }

    keypair.first = sstmac::trim_str(lhs);
    keypair.second = sstmac::trim_str(rhs);
    return cur_scope;
  }

  int
  sim_parameters::deprecated_optional_int_param(const std::string &key, int def) const
  {
    return get_optional_int_param(key, def);
  }

  int
  sim_parameters::get_optional_int_param(const std::string &key, int def) const
  {
    if (has_param(key))
      return get_int_param(key);
    register_optional_param(key, sstprintf("%d", def));
    return def;
  }

  int
  sim_parameters::get_int_param(const std::string& key) const
  {
    std::string v = get_param(key);
    const char* begin = v.c_str();
    char* end = const_cast<char*>(begin);
    long ret = ::strtol(begin, &end, 0);
    if (begin == end){
        sst_abort("sim_parameters::get_int_param: param %s with value %s is not formatted as an integer",
                  key.c_str(), v.c_str());
    }
    return ret;
  }

  int
  sim_parameters::reread_optional_int_param(const std::string &key, int def) const
  {
      return get_optional_int_param(key,def);
  }

  int
  sim_parameters::reread_int_param(const std::string &key) const
  {
      return get_int_param(key);
  }

  int
  sim_parameters::deprecated_int_param(const std::string &key) const
  {
      return get_int_param(key);
  }

  bool
  sim_parameters::deprecated_optional_bool_param(const std::string &key, bool def) const
  {
    return get_optional_bool_param(key, def);
  }

  bool
  sim_parameters::get_optional_bool_param(const std::string &key, int def) const
  {
    if (has_param(key))
      return get_bool_param(key);
    register_optional_param(key, def ? "true" : "false");
    return def;
  }

  bool
  sim_parameters::reread_optional_bool_param(const std::string &key, bool def) const
  {
      return get_optional_bool_param(key,def);
  }

  bool
  sim_parameters::reread_bool_param(const std::string &key) const
  {
      return get_bool_param(key);
  }

  bool
  sim_parameters::deprecated_bool_param(const std::string &key) const
  {
      return get_bool_param(key);
  }

  bool
  sim_parameters::get_bool_param(const std::string &key) const
  {
    std::string v = get_param(key);
    if (v == "true" || v == "1")
      {
        return true;
      }
    else if (v != "false" && v != "0")
      {
        sst_abort("sim_parameters::get_bool_param: param %s with value %s is not formatted as a proper boolean",
                  key.c_str(), v.c_str());
      }
    return false;
  }

  void
  sim_parameters::get_vector_param(const std::string& key, std::vector<int>& vals) const
  {
      bool errorflag = false;
      std::string param_value_str = get_param(key);
      get_intvec(param_value_str.c_str(), errorflag, vals);
      if (errorflag){
          sst_abort("improperly formatted integer vector (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
  }

  double
  sim_parameters::get_freq_param(const std::string &key) const
  {
      bool errorflag = false;
      std::string param_value_str = get_param(key);
      double val = get_frequency(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted frequency (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  double
  sim_parameters::deprecated_freq_param(const std::string &key) const
  {
    return get_freq_param(key);
  }

  double
  sim_parameters::deprecated_optional_freq_param(const std::string &key, double def) const
  {
    return get_optional_freq_param(key, def);
  }

  double
  sim_parameters::get_optional_freq_param(const std::string &key, double def) const
  {
      bool errorflag = false;
      std::string freq_str = sstprintf("%eHz", def);
      std::string param_value_str = get_optional_param(key, freq_str);
      double val = get_frequency(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted frequency (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  long
  sim_parameters::get_byte_length_param(const std::string &key) const
  {
      bool errorflag = false;
      std::string param_value_str = get_param(key);
      double val = get_byte_length(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted byte_length (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  long
  sim_parameters::deprecated_byte_length_param(const std::string &key) const
  {
      return get_byte_length_param(key);
  }

  long
  sim_parameters::deprecated_optional_byte_length_param(const std::string &key, long def) const
  {
    return get_optional_byte_length_param(key, def);
  }

  long
  sim_parameters::get_optional_byte_length_param(const std::string& key, long length) const
  {
      bool errorflag = false;
      std::string length_str = sstprintf("%ldB", length);
      std::string param_value_str = get_optional_param(key, length_str);
      double val = get_byte_length(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted byte length (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  double
  sim_parameters::get_bandwidth_param(const std::string &key) const
  {
      bool errorflag = false;
      std::string param_value_str = get_param(key);
      double val = get_bandwidth(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted bandwidth (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  double
  sim_parameters::reread_bandwidth_param(const std::string &key) const
  {
    return get_bandwidth_param(key);
  }

  double
  sim_parameters::reread_optional_bandwidth_param(const std::string &key, double def) const
  {
    return get_optional_bandwidth_param(key, def);
  }

  double
  sim_parameters::deprecated_bandwidth_param(const std::string &key) const
  {
    return get_bandwidth_param(key);
  }

  double
  sim_parameters::deprecated_optional_bandwidth_param(const std::string &key, double def) const
  {
    return get_optional_bandwidth_param(key, def);
  }

  double
  sim_parameters::get_optional_bandwidth_param(const std::string &key, const std::string& def) const
  {
      bool errorflag = false;
      std::string param_value_str = get_optional_param(key, def);
      double val = get_bandwidth(param_value_str.c_str(), errorflag);
      if (errorflag){
          sst_abort("improperly formatted bandwidth (%s) for parameter %s",
                    param_value_str.c_str(), key.c_str());
      }
      return val;
  }

  double
  sim_parameters::get_optional_bandwidth_param(const std::string &key, double def) const
  {
       std::string bwstr = sstprintf("%ebytes/sec", def);
       return get_optional_bandwidth_param(key, bwstr);
  }

  void
  sim_parameters::print_params(const maptype &pmap, std::ostream &os) const
  {
      maptype::const_iterator it = pmap.begin(), end = pmap.end();
      os << "SST/macro parameters: \n";
      for ( ; it != end; ++it){
          os << sstprintf("%25s = %s\n", it->first.c_str(), it->second.c_str());
      }
      os << "Optional parameters read: \n";
      params_read_map::const_iterator itkey = optional_params_read_.begin(), endkey = optional_params_read_.end();
      for ( ; itkey != endkey; ++itkey){
          const value_set& valset = itkey->second;
          const std::string& key = itkey->first;
          value_set::const_iterator itval = valset.begin(), endval= valset.end();
          os << sstprintf("%25s = %s", key.c_str(), (*itval).c_str());
          ++itval;
          for ( ; itval != endval; ++itval){
              os << sstprintf(", %s", (*itval).c_str());
          }
          os << "\n";
      }
  }

  void
  empty_parameters::print_params(std::ostream &os) const
  {
    sim_parameters::print_params(params_, os);
  }


}
