/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_COMMON_VIS_VIS_H_INCLUDED
#define SSTMAC_COMMON_VIS_VIS_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/common/nodeaddress.h>
#include <boost/unordered_map.hpp>
#include <sstmac/common/sstmac_env.h>
#include <sstmac/common/factories/factory.h>

namespace sstmac
{
  namespace vis
  {

    /**
     * This is a parent class for an vis object that is passed back to
     * the simulation models to be able to manipulate
     */
    class vis_obj :
        public ptr_type
    {
    public:
      typedef boost::intrusive_ptr<vis_obj> ptr;

      enum VIS_COLOR
      {
        RED, BLUE, BLACK, GREEN, YELLOW, ORANGE, WHITE, PURPLE, CLEAR
      };

      virtual void
      set_activity(double val) = 0;

      virtual void
      set_color(VIS_COLOR col) = 0;

      static double clear_r;
      static double clear_g;
      static double clear_b;
      static bool clear_set;

    protected:

      vis_obj()
      {
        clear_set = false;
      }

      virtual
      ~vis_obj()
      {
      }

      void
      get_color(VIS_COLOR col, double &r, double &g, double &b);

      virtual std::string
      to_string() const
      {
        return "vis_obj";
      }

    };

    /**
     * this is a parent class for a way of displaying a set of actions or components.
     */
    class vis_engine :
        public factory_type
    {
    public:
      typedef boost::intrusive_ptr<vis_engine> ptr;

      enum VIS_DIR
      {
        VIS_X = 0, VIS_Y, VIS_Z, VIS_NONE
      };

      virtual void
      init() = 0;

      virtual void
      draw() = 0;

      virtual void
      update(timestamp now) = 0;

      virtual vis_obj::ptr
      create_cube(double x, double y, double z, double len) = 0;

      virtual vis_obj::ptr
      create_cylinder(double x, double y, double z, double rad, double len,
          int dir) = 0;

      virtual void
      write_to_file(const std::list<vis_obj::ptr> &objs, std::string fname) = 0;

      virtual void
      display_label(std::string text) = 0;

      virtual void
      update_label(timestamp now) = 0;

    protected:

      vis_engine()
      {

      }

      virtual
      ~vis_engine()
      {
      }

      virtual std::string
      to_string() const
      {
        return "vis_engine";
      }
    };
    DeclareFactory(vis_engine);

    /**
     * This is a parent class for a simulation model that is associated with a vis_obj
     */
    class vis_comp : public virtual ptr_type
    {
    public:
      typedef boost::intrusive_ptr<vis_comp> ptr;

      void
      set_vis_obj(vis_obj::ptr o)
      {
        vobj_ = o;
      }

      virtual void
      set_link_vis(vis::vis_obj::ptr vo, int dim, int dir)
      {

      }

      vis_obj::ptr
      get_vis_obj()
      {
        return vobj_;
      }

      void
      set_vis_engine(vis_engine::ptr e)
      {
        veng_ = e;
      }

      void
      vis_set_activity(double val, timestamp now);

      void
      vis_set_color(vis_obj::VIS_COLOR col, timestamp now);

    protected:

      vis_obj::ptr vobj_;
      vis_engine::ptr veng_;

      vis_comp()
      {

      }

      virtual
      ~vis_comp()
      {
      }

      virtual std::string
      to_string() const
      {
        return "vis_comp";
      }
    };

    class vis_topology : public virtual ptr_type
    {
    public:
      typedef boost::intrusive_ptr<vis_topology> ptr;

      typedef boost::unordered_map<nodeaddress::ptr, vis::vis_comp::ptr> vismaptype;

      virtual void
      display_nodes(const vismaptype &switches, const vis::vis_engine::ptr& eng,
          std::list<vis::vis_obj::ptr> &objs) = 0;

    protected:

      vis_topology()
      {
      }

      virtual
      ~vis_topology()
      {
      }

      virtual std::string
      to_string() const
      {
        return "vis_topology";
      }
    };

    /**
     * vis_display is a window that will be displayed.  simulation components
     * can become children of this class to display themselves.  they use the
     * provided vis_engine to request actions.
     */
    class vis_display :
        virtual public ptr_type
    {
    public:
      typedef boost::intrusive_ptr<vis_display> ptr;

      virtual vis_topology::vismaptype
      vis_get_nodes()
      {
        return vismap_;
      }

      virtual vis_topology::ptr
      vis_get_topology()
      {
        return vistop_;
      }

      virtual void
      vis_start(bool tofile);

      virtual void
      vis_update(timestamp now);

      virtual void
      vis_complete()
      {

      }

      static ptr
      construct(const std::string& name, const vis_engine::ptr& e,
          vis_topology::vismaptype m, const vis_topology::ptr& t)
      {
        return ptr(new vis_display(name, e, m, t));
      }

    protected:

      vis_engine::ptr eng_;
      vis_topology::vismaptype vismap_;
      vis_topology::ptr vistop_;
      std::string name_;

      std::list<vis_obj::ptr> objs_;

      timestamp file_interval_;
      timestamp last_file_;
      int filenum_;

      /**
       * Use this constructor when making a vis_display by itself
       */
      vis_display(const std::string& name, vis_engine::ptr e,
          vis_topology::vismaptype m, vis_topology::ptr t) :
        eng_(e), vismap_(m), vistop_(t), name_(name)
      {
        last_file_ = timestamp(0);
        filenum_ = 0;
      }

      /**
       * Use this constructor when inheriting from vis_display.  Make sure to
       * override vis_get_nodes() and vis_get_topology()
       */
      vis_display(const std::string& name) :
        name_(name)
      {
      }

      void
      init_factory_params(const sim_parameters::ptr &params);

      virtual
      ~vis_display()
      {
      }

      virtual std::string
      to_string() const
      {
        return "vis_display";
      }
    };

  }
}

#endif
