/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstream>

#include <sstmac/common/vis/vtk_engine.h>

#include <vtkCubeSource.h>
#include <vtkCylinderSource.h>

#include <vtkPolyDataMapper.h>
#include <vtkProperty.h>
#include <vtkPolyDataWriter.h>
#include <vtkXMLPolyDataWriter.h>
#include <vtkAppendPolyData.h>

#include <vtkMatrixToLinearTransform.h>
#include <vtkTransform.h>
#include <vtkCellData.h>
#include <vtkPointData.h>
#include <vtkSmartPointer.h>

namespace sstmac
{
  namespace vis
  {

    SSTRegister("vtk", vis_engine, vtk_engine,
          "Implements a network visualization tool for linking with VTK");

    vtk_obj::vtk_obj(vtkPolyDataAlgorithm* al, vtkActor* ac, vtkMapper* ma) :
      alg_(al), act_(ac), map_(ma)
    {

    }

    void
    vtk_obj::set_activity(double val)
    {

      double red, green, blue;

      get_heatmap_color(val, red, green, blue);

      act_->GetProperty()->SetColor(red, green, blue);
      act_->GetProperty()->SetOpacity(0.5);
    }

    void
    vtk_obj::set_color(VIS_COLOR col)
    {
      double red, green, blue;

      get_color(col, red, green, blue);

      act_->GetProperty()->SetColor(red, green, blue);
      act_->GetProperty()->SetOpacity(0.5);
    }

    void
    vtk_obj::get_heatmap_color(double val, double &r, double &g, double &b)
    {
      if (val < 0.2)
        {
          r = 0.1;
          g = 0.1;
          b = std::max(0.1, (val) / 0.2);
        }
      else if (val < 0.4)
        {
          r = 0.1;
          g = std::max(0.1, (val - 0.2) / 0.2);
          b = 1.0;
        }
      else if (val < 0.6)
        {
          r = 0.1;
          g = 1.0;
          b = std::max(0.1, 1.0 - (val - 0.4) / 0.2);
        }
      else if (val < 0.8)
        {
          r = std::max(0.1, (val - 0.6) / 0.2);
          g = 1.0;
          b = 0.1;
        }
      else
        {
          r = 1.0;
          g = std::max(0.1, 1.0 - (val - 0.8) / 0.2);
          b = 0.1;
        }
    }

    vtk_engine::vtk_engine() :
      hasTurned_(false), lastupdate_(timestamp(0))
    {
    }

    void
    vtk_engine::init_factory_params(const sim_parameters::ptr &params)
    {
       update_ = params->get_optional_time_param("vis_update",
          timestamp(0.001));
      start_ = params->get_optional_time_param("vis_start",
          timestamp(0.000));
    }

    void
    vtk_engine::init()
    {
      //
      // Finally we create the render window which will show up on the screen.
      // We put our renderer into the render window using AddRenderer. We also
      // set the size to be 300 pixels by 300.
      //
      rend_ = vtkRenderer::New();
      rend_->SetBackground(0.1, 0.2, 0.4);

      renWin_ = vtkRenderWindow::New();
      renWin_->AddRenderer(rend_);
      renWin_->SetSize(900, 900);
    }

    void
    vtk_engine::draw()
    {
      renWin_->Render();

      if (!hasTurned_)
        {
          rend_->GetActiveCamera()->Azimuth(30);
          // rend_->GetActiveCamera()->Elevation(30);
          // rend_->GetActiveCamera()->
          hasTurned_ = true;
          renWin_->Render();
        }

    }

    vis_obj::ptr
    vtk_engine::create_cube(double x, double y, double z, double len)
    {
      vtkCubeSource* cube = vtkCubeSource::New();
      cube->SetBounds(x, x + len, y, y + len, z, z + len);

      //create mapper
      vtkPolyDataMapper *cubeMapper = vtkPolyDataMapper::New();
      cubeMapper->SetInputConnection(cube->GetOutputPort());

      //create actor
      vtkActor *cubeActor = vtkActor::New();
      cubeActor->SetMapper(cubeMapper);
      cubeActor->GetProperty()->SetOpacity(0.5);

      rend_->AddActor(cubeActor);

      vtk_obj::ptr o = vtk_obj::construct(cube, cubeActor, cubeMapper);
      //objs_.push_back(o);

      if (!vis_obj::clear_set)
        {

          cubeActor->GetProperty()->GetColor(vis_obj::clear_r,
              vis_obj::clear_g, vis_obj::clear_b);
          vis_obj::clear_set = true;
        }

      return o;
    }

    vis_obj::ptr
    vtk_engine::create_cylinder(double x, double y, double z, double rad,
        double len, int dir)
    {
      vtkCylinderSource* cube = vtkCylinderSource::New();
      cube->SetCenter(x, y, z);
      cube->SetRadius(rad);
      cube->SetHeight(len);
      cube->CappingOn();

      //create mapper
      vtkPolyDataMapper *cubeMapper = vtkPolyDataMapper::New();
      cubeMapper->SetInputConnection(cube->GetOutputPort());

      //create actor
      vtkActor *cubeActor = vtkActor::New();
      cubeActor->SetMapper(cubeMapper);
      cubeActor->GetProperty()->SetOpacity(0.5);
      cubeActor->SetOrigin(cube->GetCenter());
      if (dir == VIS_X)
        {

          cubeActor->RotateWXYZ(90, 1, 0, 0);
        }
      else if (dir == VIS_Y)
        {

          cubeActor->RotateWXYZ(90, 0, 1, 0);
        }
      else if (dir = VIS_Z)
        {

          cubeActor->RotateWXYZ(90, 0, 0, 1);
        }

      rend_->AddActor(cubeActor);

      vtk_obj::ptr o = vtk_obj::construct(cube, cubeActor, cubeMapper);
      //objs_.push_back(o);

      if (!vis_obj::clear_set)
        {

          cubeActor->GetProperty()->GetColor(vis_obj::clear_r,
              vis_obj::clear_g, vis_obj::clear_b);
          vis_obj::clear_set = true;
        }

      return o;
    }

    void
    vtk_engine::write_to_file(const std::list<vis_obj::ptr> &objs,
        std::string fname)
    {

      vtkXMLPolyDataWriter* writer = vtkXMLPolyDataWriter::New();

      writer->SetFileName(fname.c_str());

      vtkAppendPolyData* app = vtkAppendPolyData::New();
      vtkMatrixToLinearTransform* mtl = vtkMatrixToLinearTransform::New();
      std::list<vis_obj::ptr>::const_iterator it, end = objs.end();
      long i = 0;
      for (it = objs.begin(); it != end; it++)
        {
          vtk_obj::ptr vob = boost::dynamic_pointer_cast<vtk_obj>(*it);
          vtkPolyDataAlgorithm* a = vob->get_alg();
          a->Update();
          vtkActor* act = vob->get_act();
          vtkPolyData* pdata = a->GetOutput();
          vtkPoints* p = pdata->GetPoints();
          vtkMatrix4x4* mat = act->GetMatrix();
          mtl->SetInput(mat);
          vtkPoints* pout = vtkPoints::New(); ///delete?

          mtl->TransformPoints(p, pout);
          vtkPolyData* pd = vtkPolyData::New();
          pd->DeepCopy(pdata);

          pd->SetPoints(pout);

          vtkSmartPointer<vtkUnsignedCharArray> colors = vtkSmartPointer<
              vtkUnsignedCharArray>::New();
          colors->SetNumberOfComponents(4);
          colors->SetName("Colors");

          double r, g, b;
          act->GetProperty()->GetColor(r, g, b);

          //the color vector
          unsigned char col[4] =
            { r * 255, g*255, b*255, 255 };


          // Add 6 of them for the 6 sizes of the polygon
          colors->InsertNextTupleValue(col);
          colors->InsertNextTupleValue(col);
          colors->InsertNextTupleValue(col);
          colors->InsertNextTupleValue(col);
          colors->InsertNextTupleValue(col);
          colors->InsertNextTupleValue(col);

          pd->GetCellData()->SetScalars(colors);


          app->AddInputData(pd);

          i++;
        }

      app->Update();

      writer->SetInputConnection(app->GetOutputPort());

      writer->Write();

      writer->Delete();
      app->Delete();
      mtl->Delete();
    }

    void
    vtk_engine::display_label(std::string text)
    {
      labeltext_ = text;
      label_ = vtkTextActor::New();
      label_->SetDisplayPosition(1, 1);
      label_->SetHeight(1);
      update_label(timestamp(-1));
      rend_->AddActor(label_);
    }

    void
    vtk_engine::update_label(timestamp now)
    {
      std::stringstream ss;
      ss << labeltext_ << " -- (t=" << now.sec() << ")\n";
      label_->SetInput(ss.str().c_str());
    }

    void
    vtk_engine::update(timestamp now)
    {

      update_label(now);
      if (now >= start_ && now >= lastupdate_ + update_)
        {

          draw();

          lastupdate_ = now;
        }
    }
  }
}
