/*
 * disk.cc
 *
 * Represents a simulated disk, loaded from an XML-format file.
 *
 * Jonathan Pearson
 * November 12, 2005 -- Created
 *
 */

#include <sstmac/hardware/disk/vesper/disk.h>
#include <sstmac/hardware/disk/vesper/mlparser.h>

//from Nachos
#include <sstmac/hardware/disk/vesper/utility.h>
#include <math.h>


namespace sstmac
{
  namespace hw
  {

    //metric prefixes
#define KILO 1024
#define MEGA (KILO * KILO)
#define GIGA (KILO * MEGA)

    // We put this at the front of the file representing the
    // disk, to make it less likely we will accidentally treat a useful file
    // as a disk (which would probably trash the file's contents).
#define MagicNumber 	0x456789ab
#define MagicSize 		sizeof(int)

    using std::ifstream;

     int Disk::logicalBlocks;
     int Disk::blockSize;
     bool Disk::disableCache;
     int Disk::maxRead;
     int Disk::outerTrackSize;
     double Disk::speed;

    debug Disk::dbg_;

    // Remember, the whole point of XML is that if you see something that you don't recognize, IGNORE IT (or at least, don't crash)
    // Things that you don't recognize can be from a newer version, which probably still contains all the things you DO want

    Disk::Disk(const char *profile, const char *name, eventhandler::ptr callWhenDone)
    {

      if(!dbg_.is_real()){
          dbg_ = debug("<debug> disk");
          logger::register_logger(&dbg_);
      }

      //load profile
      int i = OpenForReadWrite(profile, true); //crash if it's not there
      MLParser *ml = new MLParser(i);
      Close(i);

      //now load from ml
      loadData(ml);

      SSTMAC_DEBUG << "Disk: data loaded \n";

      delete ml; //no more need for it

      //initialize disk
      headSector = 0; //start the disk off at 0
      diskOperationTicks = 0;

      cacheEnabled = true; //default to enabled

      opStart = opEnd = timestamp(0);

      callback = callWhenDone;


      /*diskFile = OpenForReadWrite(name, false);
      if (diskFile >= 0)
        { //file exists, check magic number
          int temp;
          Read(diskFile, (char *) &temp, MagicSize);
          ASSERT(temp == MagicNumber);
        }
      else
        {
          //doesn't exist, create it
          int temp = MagicNumber;
          diskFile = OpenForWrite(name); //this will create and open for read/write (OpenForWrite isn't very accurate...)
          WriteFile(diskFile, (char *) &temp, MagicSize);

          //make sure the file extends all the way
          temp = 0;
          Lseek(diskFile, getSectorCount() * getSectorSize(), 0);
          WriteFile(diskFile, (char *) &temp, sizeof(int));
        }*/
    }

    Disk::~Disk()
    {

      int i;

      free(vendor);
      free(id);

      for (i = 0; i < readCount; i++)
        {
          delete readRate[i];
        }

      free(readRate);

      for (i = 0; i < seekCount; i++)
        {
          delete seekRate[i];
        }

      free(seekRate);

    }

    bool
    Disk::ReadRequest(int sectorNumber, int sectors, char *data)
    {

      timestamp timing;
      int cycles;

      if (sectorNumber < 0 || sectorNumber + sectors > logicalBlocks)
        return false; //can't do anything, past the end of the disk on one end or the other
      if (sectors > maxRead)
        return false; //can't operate on more sectors than 'maxRead'

      //are we in the middle of an operation?
      SSTMAC_ASSERT(opEnd <= now());
      //if the end is in the future (op has not completed), then crash

      //before doing anything, update the head position on the disk (spin the disk around a few times)
      updateHeadPos();

      timestamp totalSeekTime = seekToTime(sectorNumber);
      timestamp totalReadTime = readTime(sectorNumber, sectors);
      timing = totalSeekTime + totalReadTime;


      SSTMAC_DEBUG << "Disk: read is going to take " << timing << "\n";

      //update our own info so we know what's going on when the callback is called
      reading = true;
      startSector = sectorNumber;
      sectorCount = sectors;
      dataBuffer = data;

      opStart = now();
      opEnd = opStart + timing;

      //stats->numDiskReads++;

      //set up a scheduler point to let us know when the "read" will complete
      if (cycles <= 0)
        cycles = 1; //prevent interrupt from crashing
      schedule(opEnd, ptr(this), sst_message::construct());

      return true;

    }

    bool
    Disk::WriteRequest(int sectorNumber, int sectors, char *data)
    {

      timestamp timing;
      int cycles;

      if (sectorNumber < 0 || sectorNumber + sectors > logicalBlocks)
        return false; //can't do anything, past the end of the disk on one end or the other
      if (sectors > maxRead)
        return false; //can't operate on more sectors than 'maxRead'

      //are we in the middle of an operation?
      SSTMAC_ASSERT(opEnd <= now());
      //if the end is in the future (op has not completed) then crash

      //before doing anything, update the head position on the disk
      updateHeadPos();

      timestamp totalSeekTime = seekToTime(sectorNumber);
      timestamp totalWriteTime = writeTime(sectorNumber, sectors);
      timing = totalSeekTime + totalWriteTime;



      SSTMAC_DEBUG << "Disk: write is going to take " << timing << "\n";

      //update our own data so we know what to do when the callback gets called
      reading = false;
      startSector = sectorNumber;
      sectorCount = sectors;
      dataBuffer = data;

      opStart = now();
      opEnd = opStart + timing;

      //stats->numDiskWrites++;

      //set up a scheduler point to let us know when the "write" will complete
      if (cycles <= 0)
        cycles = 1; //prevent interrupt from crashing

      schedule(opEnd, ptr(this), sst_message::construct());

      return true;

    }

    void
    Disk::handle(const sst_message::ptr &msg)
    {

      SSTMAC_ASSERT(opEnd != timestamp(0));
      //got called when there was no op in progress, something weird going on...

      //diskOperationTicks += stats->GetTargetCycle(opEnd - opStart); //update by number of ticks disk was working

      opStart = opEnd = timestamp(0); //in case there were multiple for some reason

      //correct time, move the head here and read/write the data
      headSector = startSector + sectorCount; //scheduler point was set for the end of the operation (i.e. when the head reached the end of the range)

      if (reading)
        {
          diskRead();
        }
      else
        {
          diskWrite();
        }

      SSTMAC_DEBUG << "Disk: Calling callback procedure from disk\n";

      callback->handle(sst_message::ptr());

    }

    timestamp
    Disk::seekToTime(int sec)
    {
      //NOTE: this assumes that the head position has been updated recently (headTime == stats->GetTimeMillis())

      timestamp _seekTime(0.0);
      timestamp _rotateTime(0.0);

      //get sectors-per-millisecond passing by
      double spms = getRotationSpeed() * getOuterTrackSize() / (double) 60; //getSpeed() returns KRPM = tracks per second / 1000

      //firstly, are we already on the right track?
      int targTrack = sec / getOuterTrackSize();
      int curTrack = headSector / getOuterTrackSize();

      if (targTrack == curTrack)
        {
          //make sure we've spun around enough (if the cache is enabled), or calculate how much more time is necessary to spin back around
          _seekTime = timestamp(0.0);

          if (cacheEnabled)
            {
              SSTMAC_DEBUG << "Cache enabled, on the same track as the request.\n";

              //check whether the head has passed 'sec' since arriving
              timestamp tdiff = now() - cacheInitTime; //how long since we got to this track?
              int sectorsPassed = (int) (tdiff.msec() * spms); //how many sectors have we passed in that time?
              if (cacheInit + sectorsPassed >= sec)
                { //we passed 'sec'
                  _rotateTime = timestamp(0); //already got it
                }
              else
                {
                  //how much further before we hit it?
                  int relativeSector =
                      sec
                      - ((int) (sec / getOuterTrackSize())
                          * getOuterTrackSize()); //how far around this track is 'sec'?
                  int sectorsLeft = (relativeSector - sectorsPassed)
                          % getOuterTrackSize(); //how many more before we hit 'sec'?
                  while (sectorsLeft < 0)
                    sectorsLeft += getOuterTrackSize(); //proper mod (make sure it's not negative)

                  _rotateTime = timestamp(((double)sectorsLeft / spms)/1000.0); //time until we hit it
                }
            }
          else
            { //no cache
              //same code as for a cache but not having hit 'sec' yet
              SSTMAC_DEBUG << "Cache disabled, on the same track as the request.\n";
              timestamp tdiff = now() - cacheInitTime; //in this case, 'cacheInitTime' = when we arrived at the track so we know how far around we are
              int sectorsPassed = (int) (tdiff.msec() * spms);

              //how much further before we hit it?
              int relativeSector = sec
                  - ((sec / getOuterTrackSize()) * getOuterTrackSize());
              int sectorsLeft = (relativeSector - sectorsPassed)
                      % getOuterTrackSize();
              while (sectorsLeft < 0)
                sectorsLeft += getOuterTrackSize();

              _rotateTime = timestamp(((double)sectorsLeft / spms)/1000.0); //time until we hit it
            }
        }
      else
        { //wrong track
          //find how long it will take to move the head to the track, then how long it will take for the disk to rotate around to sec

          int index = findSeek((int) abs(headSector - sec)); //will find the index for a seek going the same number of tracks

          //this is how long it will take to move the head from where it is currently to the target track
          _seekTime = timestamp(seekRate[index]->time * seekRate[index]->units); //time * units = seconds, * 1000 = milliseconds

          //now figure out how far around the disk has gone in that time so we know how much longer to wait for the correct sector
          //so the numer of sectors that flash past is time * spms
          int sectors = (int) (_seekTime.msec() * spms);

          //now how far around will we be when we get there?
          int arrival = (headSector + sectors) % getOuterTrackSize();

          cacheInit = arrival;
          cacheInitTime = now() + _seekTime;

          //and how much further do we need to go once we're there?
          int difference = (sec - arrival) % getOuterTrackSize();
          while (difference < 0)
            difference += getOuterTrackSize(); //make sure it's a positive number

          difference = getOuterTrackSize() / 2;

          //how long will that take? add it onto 'time' (now that we've recorded where we'll arrive)
          _rotateTime = timestamp(((double)difference / spms)/1000.0); //time until we hit it

          //lastly, set the beginning and ending sectors for the track we're moving to
          cacheBegin = (sec / getOuterTrackSize()) * getOuterTrackSize();
          cacheEnd = cacheBegin + getOuterTrackSize() - 1;
        }

      SSTMAC_DEBUG << "Disk: Time to seek to the correct track: " << _seekTime << "\n";
      SSTMAC_DEBUG << "Disk: Time to seek to the correct sector: " << _rotateTime << "\n";
      // stats->seekCnt++;
      //stats->seekTicks += stats->GetTargetCycle(_seekTime);
      // stats->rotateCnt++;
      // stats->rotateTicks += stats->GetTargetCycle(_rotateTime);
      return _seekTime + _rotateTime;
    }

    timestamp
    Disk::readTime(int start, int count)
    {
      //NOTE: This assumes that the head is on the correct track

      //find the transfer rate starting at sector closest to 'start' (but lower)
      //use that to figure out how long it'd take to transfer 'count' sectors of data

      timestamp _readTime = timestamp(0);

      int index = findRead(start);
      double spms = getRotationSpeed() * getOuterTrackSize() / (double) 60; //sectors per millisecond

      //is the cache enabled, and if so is the data in the cache?
      if (cacheEnabled)
        {
          //how many sectors has the head passed since arriving at the track?
          timestamp timeDiff = now() - cacheInitTime;
          int sectorDiff = (int) (timeDiff.msec() * spms);
          if (cacheInit + sectorDiff > start + count)
            { //TODO: did it definitely read the correct 'count' sectors? or some other 'count' sectors?
              SSTMAC_DEBUG << "Cache enabled, target data already read.\n";
              _readTime = timestamp(0); //in the cache, no time to read
            }
          else
            {
              SSTMAC_DEBUG << "Cache enabled, target data not already read.\n";
              //hasn't come far enough, how much further is needed?
              int more = (start + count) - (cacheInit + sectorDiff);
              if (more > count)
                more = count; //seek takes care of getting to 'start', use read rate to read sectors

              int bytes = more * getSectorSize();
              double rate = readRate[index]->rate * readRate[index]->units; //bytes per second

              _readTime = timestamp((double)bytes / (double)rate);
            }
        }
      else
        {
          SSTMAC_DEBUG << "Cache disabled.";

          //cache not enabled, seek will tell us how far around to go
          //just figure out how long to transfer 'count' sectors
          int bytes = count * getSectorSize();
          double rate = readRate[index]->rate * readRate[index]->units;

          _readTime = timestamp((double)bytes / (double)rate);
        }

      SSTMAC_DEBUG << "Time to read: " << _readTime << "\n";
      // stats->transferCnt++;
      // stats->transferTicks += stats->GetTargetCycle(_readTime);
      return _readTime;
    }

    timestamp
    Disk::writeTime(int start, int count)
    {
      //no caching here, just raw time to write count bytes
      int index = findRead(start); //still using read rates, though

      int bytes = count * getSectorSize();
      double rate = readRate[index]->rate * readRate[index]->units;
      timestamp _writeTime = timestamp((double)bytes / (double)rate);

      SSTMAC_DEBUG << "Time to write: " << _writeTime << "\n";
      //stats->transferCnt++;
      //stats->transferTicks += stats->GetTargetCycle(_writeTime);
      return _writeTime;
    }

    int
    Disk::findRead(int sec)
    {
      //binary search for nearest (but lower) readRate by starting sector
      int index = readCount >> 1;
      int step = readCount >> 2;

      //quick check in case there's only one
      if (readCount == 1)
        return 0;

      while (step != 0)
        {
          if (readRate[index]->start == sec)
            return index;
          else if (readRate[index]->start < sec)
            {
              index += step;
              step >>= 1;
            }
          else
            {
              index -= step;
              step >>= 1;
            }
        }

      //make sure we're one below (this shouldn't take more than about 2 * log log n steps)
      while (index < readCount - 1 && readRate[index]->start < sec)
        index++;
      while (index > 0 && readRate[index]->start > sec)
        index--;

      return index;
    }

    int
    Disk::findSeek(int sec)
    {
      //binary search for nearest (but lower) seekRate by starting sector
      int index = seekCount >> 1;
      int step = seekCount >> 2;

      //quick check in case there's only one
      if (seekCount == 1)
        return 0;

      while (step != 0)
        {
          if (seekRate[index]->distance == sec)
            return index;
          else if (seekRate[index]->distance < sec)
            {
              index += step;
              step >>= 1;
            }
          else
            {
              index -= step;
              step >>= 1;
            }
        }

      //make sure we're one below (this shouldn't take more than about 2 * log log n steps)
      while (index < seekCount - 1 && seekRate[index]->distance < sec)
        index++;
      while (index > 0 && seekRate[index]->distance > sec)
        index--;

      return index;
    }

    void
    Disk::loadData(MLParser *ml)
    {
      //loop through tags
      for (tagvec_iter_t i = ml->getTags(); i != ml->getTagsEnd(); i++)
        {
          if ((*i)->getName() == "disk")
            {
              loadDisk(*i);
            }
        }
    }

    void
    Disk::loadDisk(MLTag *tag)
    {
      //load data from tag
      std::string temp;

      temp = tag->stringAttribute("vendor");
      vendor = (char *) malloc(temp.length() * sizeof(char));
      strcpy(vendor, temp.c_str());

      temp = tag->stringAttribute("id");
      id = (char *) malloc(temp.length() * sizeof(char));
      strcpy(id, temp.c_str());

      revision = tag->intAttribute("revision");

      //loop through elements
      for (evec_iter_t i = tag->getContents(); i != tag->getContentsEnd(); i++)
        {
          element *elem = (*i);

          if (!elem->str)
            { //tag element
              MLTag *ctag = elem->st.tag; //child tag

              //pass along to proper loader
              if (ctag->getName() == "info")
                {
                  loadInfo(ctag);
                }
              else if (ctag->getName() == "profile")
                {
                  loadProfile(ctag);
                }
            } //don't expect to see std::string elements, so ignore them
        }
    }

    void
    Disk::loadInfo(MLTag *tag)
    {
      //load data from tag
      logicalBlocks = tag->longAttribute("logicalblocks");
      blockSize = tag->intAttribute("blocksize");
      disableCache = tag->booleanAttribute("disablecache");
      maxRead = tag->intAttribute("maxread");
      outerTrackSize = tag->intAttribute("outertracksize");
      speed = tag->floatAttribute("speed");

      //should not have any children, and if it does we're not interested in them anyway
    }

    void
    Disk::loadProfile(MLTag *tag)
    {
      //doesn't have any attributes (or if it does, we're not interested in them)
      //loop through elements
      for (evec_iter_t i = tag->getContents(); i != tag->getContentsEnd(); i++)
        {
          element *elem = (*i);

          if (!elem->str)
            { //tag element
              MLTag *ctag = elem->st.tag; //child tag

              //pass along to proper loader
              if (ctag->getName() == "read")
                {
                  loadReadData(ctag);
                }
              else if (ctag->getName() == "seek")
                {
                  loadSeekData(ctag);
                }
            } //don't expect to see std::string elements, so ignore them
        }
    }

    void
    Disk::loadReadData(MLTag *tag)
    {
      //don't expect attributes
      //loop through elements (should all be "rate" tags)
      int pos = 0;

      readCount = tag->countContents();
      readRate = (readRate_t **) malloc(readCount * sizeof(readRate_t *));

      for (evec_iter_t i = tag->getContents(); i != tag->getContentsEnd(); i++)
        {
          element *elem = (*i);

          if (!elem->str)
            { //tag element
              MLTag *ctag = elem->st.tag; //child tag

              //don't bother passing along, just read data
              if (ctag->getName() == "rate")
                {
                  std::string unit;

                  readRate[pos] = new readRate_t;

                  readRate[pos]->start = ctag->intAttribute("startblock");
                  readRate[pos]->rate = ctag->floatContents();

                  unit = ctag->stringAttribute("unit");
                  if (unit == "GB/sec")
                    readRate[pos]->units = (double) GIGA;
                  else if (unit == "MB/sec")
                    readRate[pos]->units = (double) MEGA;
                  else if (unit == "KB/sec")
                    readRate[pos]->units = (double) KILO;
                  else if (unit == "B/sec")
                    readRate[pos]->units = (double) 1.0;
                  else
                    {
                      fprintf(stderr,
                          "Unknown unit specification, assuming MB/sec: %s\n",
                          unit.c_str());
                      readRate[pos]->units = (double) MEGA; //most common rate, so just assume if unknown
                    }

                  pos++;
                }
              else
                { //don't expect anything else...
                  readCount--; //take one off so we don't make a mistake searching later on
                }
            }
          else
            { //don't expect contents...
              readCount--; //take one off so we don't make a mistake searching later on
            }
        }
    }

    void
    Disk::loadSeekData(MLTag *tag)
    {
      //don't expect attributes
      //loop through elements (should all be "time" tags)
      int pos = 0;

      seekCount = tag->countContents();
      seekRate = (seekRate_t **) malloc(seekCount * sizeof(seekRate_t *));

      for (evec_iter_t i = tag->getContents(); i != tag->getContentsEnd(); i++)
        {
          element *elem = (*i);

          if (!elem->str)
            { //tag element
              MLTag *ctag = elem->st.tag; //child tag

              //don't bother passing along, just read data
              if (ctag->getName() == "time")
                {
                  std::string unit;

                  seekRate[pos] = new seekRate_t;

                  seekRate[pos]->distance = ctag->intAttribute("distance");
                  seekRate[pos]->time = ctag->floatContents();

                  unit = ctag->stringAttribute("unit");
                  if (unit == "ns")
                    seekRate[pos]->units = (double) 1.0 / 1000000000.0;
                  else if (unit == "us")
                    seekRate[pos]->units = (double) 1.0 / 1000000.0;
                  else if (unit == "ms")
                    seekRate[pos]->units = (double) 1.0 / 1000.0;
                  else if (unit == "s")
                    seekRate[pos]->units = (double) 1.0;
                  else
                    {
                      fprintf(
                          stderr,
                          "Unknown unit specification, assuming milliseconds: %s\n",
                          unit.c_str());
                      seekRate[pos]->units = (double) 1.0 / 1000.0; //most common rate, so just assume if unknown
                    }

                  pos++;
                }
              else
                { //don't expect anything else...
                  seekCount--; //take one off so we don't make a mistake searching later on

                  fprintf(stderr, "Unexpected tag in disk profile: %s\n",
                      ctag->getName().c_str());
                }
            }
          else
            { //don't expect contents...
              seekCount--; //take one off so we don't make a mistake searching later on
            }
        }
    }

    void
    Disk::diskRead()
    {
      //read data from disk file into 'dataBuffer'
     // Lseek(diskFile, startSector * getSectorSize() + MagicSize, 0);
     // Read(diskFile, dataBuffer, sectorCount * getSectorSize());
    }

    void
    Disk::diskWrite()
    {
      //write data from 'dataBuffer' into disk file
     // Lseek(diskFile, startSector * getSectorSize() + MagicSize, 0);
     // WriteFile(diskFile, dataBuffer, sectorCount * getSectorSize());
    }

    void
    Disk::updateHeadPos()
    {
      //how long since the last update?
      timestamp timeDiff = now() - headTime; //milliseconds

      //how many sectors flash past in a millisecond?
      double spms = getRotationSpeed() * getOuterTrackSize() / (double) 60;

      //number of sectors that flashed past since the last update
      int sectorDiff = (int) (timeDiff.msec() * spms);
      sectorDiff %= getOuterTrackSize();

      //relative position on the track (how far from the beginning of the track?)
      int relativePos = headSector
          - ((headSector / getOuterTrackSize()) * getOuterTrackSize());
      relativePos += sectorDiff;
      relativePos %= getOuterTrackSize();

      //sector number of the beginning of the track
      int trackStart = (headSector / getOuterTrackSize()) * getOuterTrackSize();
      headSector = trackStart + relativePos;

      //last update just occured now, reset it
      headTime = now();
    }

  }
}
