/*
 * disk.h
 *
 * Represents a simulated disk.
 *
 * Jonathan Pearson
 * November 12, 2005 -- Created
 *
 */

#ifndef SIMDISK_H
#define SIMDISK_H

#define VESPER

#include <sstmac/hardware/disk/vesper/mlparser.h>

//from Nachos
#include <sstmac/hardware/disk/vesper/utility.h>

#include <sstmac/common/eventmodule.h>

namespace sstmac
{
  namespace hw
  {

    typedef struct
    {
      int start;
      double rate;
      double units; //rate * units = bytes per second, so for MB/sec, should be 2^20
    } readRate_t;

    typedef struct
    {
      int distance;
      double time;
      double units; //time * units = seconds, so for milliseconds, should be 1/1000
    } seekRate_t;

    class Disk : public eventmodule
    {
    protected:
      //name is ignored
      Disk(const char *profile, const char *name, eventhandler::ptr callWhenDone); //bit of a double meaning -- file containing the profile ('file' is a substring of 'profile')

      static debug dbg_;

    public:
      typedef boost::intrusive_ptr<Disk> ptr;

      static ptr
      construct(const char *profile, const char *name, eventhandler::ptr callWhenDone){
        return ptr(new Disk(profile, name, callWhenDone));
      }

      virtual void handle(const sst_message::ptr &msg);


      int diskOperationTicks;

      virtual
      ~Disk();

      //These will tell the 'disk' to read/write a 'sectors' sectors, starting at 'sectorNumber'
      //When finished, 'callWhenDone' from the SimDisk constructor will be called with 'arg' as an argument (also from the constructor)
      //DO NOT CHANGE 'data' after calling one of these and before 'callWhenDone' is called, as it will produce unpredictable results (at least, for a real disk)
      //You should not call either of these while waiting for 'callWhenDone', but if you do, an approximation of the expected disk behavior will result
      virtual bool
      ReadRequest(int sectorNumber, int sectors, char *data);
      virtual bool
      WriteRequest(int sectorNumber, int sectors, char *data);


      virtual const char *
      getVendor() const
      {
        return vendor;
      }
      virtual const char *
      getID() const
      {
        return id;
      }
      virtual int
      getRevision() const
      {
        return revision;
      }
      virtual long long
      getSectorCount() const
      {
        return logicalBlocks;
      }
      virtual int
      getSectorSize() const
      {
        return blockSize;
      }
      virtual int
      getMaxRead() const
      {
        return maxRead;
      }
      virtual int
      getOuterTrackSize() const
      {
        return outerTrackSize;
      }
      virtual double
      getRotationSpeed() const
      {
        return speed;
      }

      virtual bool
      getCacheEnabled() const
      {
        return cacheEnabled;
      }
      virtual void
      setCacheEnabled(bool b)
      {
        cacheEnabled = b;
      }

      virtual int
      getTrackCount() const
      {
        return getSectorCount() / getOuterTrackSize();
      } //this is an estimate; it assumes all tracks are the same size

    private:

      //starting and ending time (in millisecond) for the disk operation
      timestamp opStart;
      timestamp opEnd; //if opComplete is called during a cycle not equal to this, it is ignored

      //how much time will it take to do various operations? all return milliseconds
      virtual timestamp
      seekToTime(int sec); //return how long it would take to seek the head to the new sector (and set up the cache variables)
      virtual timestamp
      readTime(int start, int count); //returns how long it would take to read 'count' sectors from the disk
      virtual timestamp
      writeTime(int start, int count); //returns how long it would take to write 'count' sectors to the disk

      //find the index of the proper value in the profile data
      virtual int
      findRead(int sec);
      virtual int
      findSeek(int sec);

      //these load data from a parsed XML file
      virtual void
      loadData(MLParser *ml);

      virtual void
      loadDisk(MLTag *tag);
      virtual void
      loadInfo(MLTag *tag);
      virtual void
      loadProfile(MLTag *tag);

      virtual void
      loadReadData(MLTag *tag);
      virtual void
      loadSeekData(MLTag *tag);

      //updates the head position by spinning the disk around based on the last time it was updated
      virtual void
      updateHeadPos();

      char *vendor;
      char *id;
      int revision;

    public:
      static int logicalBlocks;
      static int blockSize;
      static bool disableCache;
      static int maxRead;
      static int outerTrackSize;
      static double speed;

    private:
      bool cacheEnabled;
      int cacheInit; //sector where the head reached the current track (so we know when it has finished reading the entire track)
      timestamp cacheInitTime; //time that the head reached cacheInit (so we know how far the disk as spun since)
      int cacheBegin; //beginning sector of this track
      int cacheEnd; //ending sector of this track

      int headSector; //sector the head is currently pointing at
      timestamp headTime; //last time the head position was updated (so we know how far the disk as spun since)

      readRate_t **readRate;
      seekRate_t **seekRate;

      int readCount;
      int seekCount;

      //these are the true I/O functions that actually move data to/from the "disk"
      //the rest of the functions are for accurate timing of interrupts and whatnot
      virtual void
      diskRead();
      virtual void
      diskWrite();

      eventhandler::ptr callback;

      int diskFile;

      //these let us know what should happen when opComplete is called
      bool reading;
      long long startSector;
      int sectorCount;
      char *dataBuffer;
    };

  }
}
#endif
