/*
 * mlparser.h
 *
 * Reads and parses an XML file into tags with attributes and contents (which may include more tags)
 * Supports multiple top-level tags
 *
 * Jonathan Pearson
 * July 20, 2004 -- Original version written in Java
 * February 18, 2005 -- Translated from Java to C#
 * November 2, 2005 -- Translated from C# to C++
 *
 */
 
#include <sstmac/hardware/disk/vesper/mlparser.h>
#include <sstmac/hardware/disk/vesper/utility.h>

#include <cstdio>
#include <utility>
#include <stdlib.h>


namespace sstmac { namespace hw {

//exceptions
static const int openCloseMismatch = 1;
static const int expectedWhitespace = 2;
static const int expectedAlphaNum = 3;
static const int unexpectedChar = 4; //set expected (below) when throwing this

static char expected = 0;

static std::string toLower(const std::string &s) {
	int i, len;
    std::string ans = "";
	
	len = s.length();
	for (i = 0; i < len; i++) {
		if ('A' <= s[i] && s[i] <= 'Z') {
			ans += s[i] + ('a' - 'A');
		} else {
			ans += s[i];
		}
	}
	
	return ans;
}

static bool isLetter(char c) {
	return ('a' <= c && c <= 'z') || ('A' <= c && c <= 'Z');
}

static bool isDigit(char c) {
	return '0' <= c && c <= '9';
}

static bool isLetterOrDigit(char c) {
	return isLetter(c) || isDigit(c);
}

static bool isWhitespace(char c) {
	return c == ' ' || c == '\r' || c == '\n' || c == '\t' || c == 0xB || c == 0xC;
}

MLTag::MLTag() {
	name = "";
}

MLTag::~MLTag() {
	//delete contents
	for (evec_iter_t ei = contents.begin(); ei != contents.end(); ei++) {
		const element *elem = (*ei);
		if (elem->str) delete elem->st.str; //could be a std::string *
		else delete elem->st.tag; //or an MLTag * (should recursively delete downward)
		
		delete elem;
	}
	
	//delete attributes
	for (attmap_iter_t ai = attributes.begin(); ai != attributes.end(); ai++) {
		delete (*ai).first;
	}
}

std::string MLTag::getName() const {
	return name;
}

void MLTag::setName(const std::string &nm) {
	name = nm;
}

void MLTag::setAttribute(std::string attname, std::string attval) {
	char *nm = (char *)malloc(attname.length() * sizeof(char));
	
	strcpy(nm, attname.c_str());
	
	attributes.insert(make_pair(nm, attval));
}

void MLTag::addContent(std::string *elem) {
	element *e = new element();
	e->str = true;
	e->st.str = elem;
	
	contents.push_back(e);
}

void MLTag::addContent(MLTag *elem) {
	element *e = new element();
	e->str = false;
	e->st.tag = elem;
	
	contents.push_back(e);
}

evec_iter_t MLTag::getContents() const {
	return contents.begin();
}

evec_iter_t MLTag::getContentsEnd() const {
	return contents.end();
}

int MLTag::countContents() const {
	return contents.size();
}

std::string MLTag::stringAttribute(const std::string &nm) {
    std::string str = toLower(nm);
    std::string ans = attributes[str.c_str()]; //convert from char * to std::string
	
	return ans;
}

int MLTag::intAttribute(const std::string &nm) {
    return atoi(stringAttribute(nm).c_str());
}

long long MLTag::longAttribute(const std::string &nm) {
    return atol(stringAttribute(nm).c_str());
}

float MLTag::floatAttribute(const std::string &nm) {
    return atof(stringAttribute(nm).c_str());
}

bool MLTag::booleanAttribute(const std::string &nm) {
    std::string orig = stringAttribute(nm);
    std::string temp = toLower(orig);
	
	if (temp == "true" || temp == "t" || temp == "yes" || temp == "y" || temp == "1") return true;
	else if (temp == "false" || temp == "f" || temp == "no" || temp == "n" || temp == "0") return false;
	else return false;
}

std::string MLTag::textContents() const {
    std::string ans = "";
	
	for (evec_iter_t i = contents.begin(); i != contents.end(); i++) {
		if ((*i)->str) ans += *((*i)->st.str);
	}
	
	return ans;
}

std::string MLTag::allTextContents() const {
    std::string ans = "";
	
	for (evec_iter_t i = contents.begin(); i != contents.end(); i++) {
		if ((*i)->str) ans += *((*i)->st.str);
		else ans += (*i)->st.tag->allTextContents();
	}
	
	return ans;
}

int MLTag::intContents() const {
	return atoi(textContents().c_str());
}

int MLTag::intAllContents() const {
	return atoi(allTextContents().c_str());
}

long long MLTag::longContents() const {
	return atol(textContents().c_str());
}

long long MLTag::longAllContents() const {
	return atol(allTextContents().c_str());
}

float MLTag::floatContents() const {
	return atof(textContents().c_str());
}

float MLTag::floatAllContents() const {
	return atof(allTextContents().c_str());
}

bool MLTag::booleanContents() const {
    std::string orig = textContents();
    std::string temp = toLower(orig);
	
	if (temp == "true" || temp == "t" || temp == "yes" || temp == "y" || temp == "1") return true;
	else if (temp == "false" || temp == "f" || temp == "no" || temp == "n" || temp == "0") return false;
	else return false;
}

bool MLTag::booleanAllContents() const {
    std::string orig = allTextContents();
    std::string temp = toLower(orig);
	
	if (temp == "true" || temp == "t" || temp == "yes" || temp == "y" || temp == "1") return true;
	else if (temp == "false" || temp == "f" || temp == "no" || temp == "n" || temp == "0") return false;
	else return false;
}

const MLTag *MLTag::findChild(const std::string &nm) const {
    std::string temp = toLower(nm);
	
	for (evec_iter_t i = contents.begin(); i != contents.end(); i++) {
		if (!((*i)->str)) {
			if ((*i)->st.tag->getName() == nm) return (*i)->st.tag;
		}
	}
	
	return NULL;
}

const MLTag *MLTag::findChild(const std::string &nm, const MLTag *def) const {
	const MLTag *child = findChild(nm);
	if (child == NULL) return def;
	else return child;
}


ParserData::ParserData(int input) {
	inp = input;
	
	ch = 0;
	peek = 0;
	curTag = NULL;
}

ParserData::~ParserData() { }

int ParserData::Read() {
	char ch;
	int i = ReadPartial(inp, &ch, 1);
	if (i < 1) return -1;
	return (int)ch;
}


MLParser::MLParser(int input) {
	parse(input);
}

MLParser::~MLParser() { }

tagvec_iter_t MLParser::getTags() const {
	return tags.begin();
}

tagvec_iter_t MLParser::getTagsEnd() const {
	return tags.end();
}

int MLParser::tagCount() const {
	return tags.size();
}

const MLTag *MLParser::getTag(int index) const {
	return tags[index];
}

const MLTag *MLParser::findChild(const std::string &name) const {
    std::string nm = toLower(name);
	
	for (tagvec_iter_t i = getTags(); i != getTagsEnd(); i++) {
		//names are stored in lower case, so we don't need to convert (*i)->getName() before checking
		if ((*i)->getName() == nm) return (*i);
	}
	
	return NULL;
}

void MLParser::parse(int input) {
	ParserData *pd = new ParserData(input);
	
	try {
		if (consume(pd)) { //read into peek
			consume(pd); //copy to ch and read a new peek
			start(pd);
		}
	} catch (int ex) {
		switch (ex) {
			case openCloseMismatch:
				fprintf(stderr, "Close tag does not match corresponding open tag: line %d, column %d.\n", pd->line, pd->column);
				break;
			case expectedWhitespace:
				fprintf(stderr, "Expected whitespace, found %c, line %d, column %d.\n", pd->c, pd->line, pd->column);
				break;
			case expectedAlphaNum:
				fprintf(stderr, "Expected alphanumeric character, found %c, line %d, column %d.\n", pd->c, pd->line, pd->column);
				break;
			case unexpectedChar:
				fprintf(stderr, "Expected %c, found %c, line %d, column %d.\n", expected, pd->c, pd->line, pd->column);
		}
	}
	
	delete pd;
}

void MLParser::start(ParserData *pd) {
	while (pd->ch != -1) {
		text(pd);
	}
}

void MLParser::text(ParserData *pd) {
    if (pd->c == '<') { //opening tag
		match(pd, '<');

		//check for a comment
		if (pd->c == '!') {
			match(pd, '!');
			chars(pd);
			match(pd, '>');
		} else {
			if (pd->curTag && pd->curTag->getName().length() > 0) { //we're still nested in another tag
				//create a new nested tag
				MLTag *ntag = new MLTag();
				//add it to its parent's contents
				//pd->curTag->addContent(ntag);
				//push the parent
				pd->openTags.push(pd->curTag);
				//work on the nested tag
				pd->curTag = ntag;
			} else {
				//make sure we're working with a fresh tag, discard anything that wasn't in a tag
				pd->curTag = new MLTag();
			}

			//name the new tag
            pd->curTag->setName(toLower(stringx(pd)));
			
			//record its attributes
			attributes(pd);

			if (pd->c == '/') { //no contents or close tag
				match(pd, '/');
				match(pd, '>');
				if (pd->openTags.empty()) {
					tags.push_back(pd->curTag);
					pd->curTag = new MLTag();
				} else {
					pd->openTags.top()->addContent(pd->curTag); //add current to parent
					pd->curTag = pd->openTags.top();
					pd->openTags.pop();
				}
			} else {
				match(pd, '>');

				//watch for the close tag, but until then let everything through
				while (!(pd->c == '<' && pd->peek == '/') && pd->ch != -1)
					text(pd);

				if (pd->ch == -1) { //end of stream without closing tag, close it ourselves
					if (!pd->openTags.empty()) {
						pd->curTag = pd->openTags.top();
						pd->openTags.pop();
					}
				} else {
					//closing tag
					match(pd, '<');
					closeTag(pd);
					match(pd, '>');
				}
			}
		}
	} else { //just text
		//ignore preceeding spaces
		spacesOpt(pd);
		pd->token = chars(pd);
		if (pd->token.length() > 0) pd->curTag->addContent(new std::string(pd->token));
	}
}

void MLParser::closeTag(ParserData *pd) {
	match(pd, '/');

    std::string name = toLower(stringx(pd));
	if (name == pd->curTag->getName()) {
		if (pd->openTags.empty()) {
			tags.push_back(pd->curTag);
			pd->curTag = new MLTag();
		} else {
			pd->openTags.top()->addContent(pd->curTag); //add current to parent
			pd->curTag = pd->openTags.top();
			pd->openTags.pop();
		}
	} else {
		throw openCloseMismatch;
	}
}

void MLParser::attributes(ParserData *pd) {
	//attributes end when we see a >
	if (pd->c == '>') return; //no more attributes
	if (pd->c == '/') return; //can be a slash without spaces in between

	//there is always a space before the attribute name or a immediate /
	spaces(pd);

	//check for end of tag after spaces
	if (pd->c == '>') return;
	if (pd->c == '/') return;

	//read the attribute name
    pd->attName = toLower(stringx(pd));
	
	//there may be spaces before the =
	spacesOpt(pd);
	
	match(pd, '=');
	
	//there may be spaces after the =
	spacesOpt(pd);

	//read the contents of the attribute, it may be enclosed in ', ", or nothing
	if (pd->c == '\'') {
		quote1(pd);
	} else if (pd->c == '\"') {
		quote2(pd);
	} else {
		//not enclosed in anything, just read a String and add it to the current tag's attributes
        pd->token = stringx(pd);
		pd->curTag->setAttribute(pd->attName, pd->token);
	}

	//there might be more attributes
	attributes(pd);
}

void MLParser::quote1(ParserData *pd) {
	match(pd, '\'');

	//allow any characters other than ' in the value
	pd->token = charsWithoutQuote1(pd);
	match(pd, '\'');
	pd->curTag->setAttribute(pd->attName, pd->token);
}

void MLParser::quote2(ParserData *pd) {
	match(pd, '\"');

	//allow any characters other than " in the value
	pd->token = charsWithoutQuote2(pd);
	match(pd, '\"');
	pd->curTag->setAttribute(pd->attName, pd->token);
}

std::string MLParser::chars(ParserData *pd) {
	//allow for any characters other than < and >
    std::string ans = "";
	while (pd->c != '<' && pd->c != '>' && pd->ch != -1) {
		ans += pd->c;
		match(pd, pd->c);
	}
	
	return ans;
}

std::string MLParser::spaces(ParserData *pd) {
	//reads whitespace, at least one character of it
    std::string ans = "";
	if (isWhitespace(pd->c)) {
		ans += pd->c;
		match(pd, pd->c);
		ans += spacesOpt(pd);
	} else {
		throw expectedWhitespace;
	}

	return ans;
}

std::string MLParser::spacesOpt(ParserData *pd) {
	//reads any amount of whitespace
    std::string ans = "";
	while (isWhitespace(pd->c) && pd->ch != -1) {
		ans += pd->c;
		match(pd, pd->c);
	}

	return ans;
}

std::string MLParser::stringx(ParserData *pd) {
	//read at least one alphanumeric character
    std::string ans = "";
	if (isLetterOrDigit(pd->c)) {
		ans += pd->c;
		match(pd, pd->c);
        ans += stringOpt(pd);
	} else {
		throw expectedAlphaNum;
	}

	return ans;
}

std::string MLParser::stringOpt(ParserData *pd) {
	//reads any number of alphanumeric characters
    std::string ans = "";
	while (isLetterOrDigit(pd->c) && pd->ch != -1) {
		ans += pd->c;
		match(pd, pd->c);
	}

	return ans;
}

std::string MLParser::charsWithoutQuote1(ParserData *pd) {
	//reads any number of characters other than '
    std::string ans = "";
	while (pd->c != '\'' && pd->ch != -1)
	{
		ans += pd->c;
		match(pd, pd->c);
	}

	return ans;
}

std::string MLParser::charsWithoutQuote2(ParserData *pd) {
	//reads any number of characters other than "
    std::string ans = "";
	while (pd->c != '\"' && pd->ch != -1)
	{
		ans += pd->c;
		match(pd, pd->c);
	}

	return ans;
}

void MLParser::match(ParserData *pd, char c) {
	if (pd->c != c) {
		expected = c;
		throw unexpectedChar;
	} else {
		consume(pd);
	}
}

bool MLParser::consume(ParserData *pd) {
	pd->column++;
	pd->ch = pd->peek;
	while ((pd->peek = pd->Read()) == '\r') ;

	if (pd->ch == '\n')
	{
		pd->line++;
		pd->column = 0; //it'll be incremented when we call consume again

		consume(pd);
	}
	if (pd->ch == -1) return false;

	pd->c = (char)pd->ch;
	return true;
}

} }
