// utility.cc 
//	Debugging routines.  Allows users to control whether to 
//	print DEBUG statements, based on a command line argument.
//
// Copyright (c) 1992-1993 The Regents of the University of California.
// All rights reserved.  See copyright.h for copyright notice and limitation 
// of liability and disclaimer of warranty provisions.

#include <sstmac/hardware/disk/vesper/copyright.h>
#include <sstmac/hardware/disk/vesper/utility.h>
#include <stddef.h>
#include <stdio.h>
#include <sys/file.h>
#include <sys/types.h>
#include <unistd.h>
#include <sstmac/common/errors.h>
// this seems to be dependent on how the compiler is configured.
// if you have problems with va_start, try both of these alternatives
//#if (defined(HOST_SNAKE) || defined(HOST_SPARC) || (defined(HOST_i386) && defined(Linux)))
#include <stdarg.h>


namespace sstmac { namespace hw {


static char *enableFlags = NULL; // controls which DEBUG messages are printed 

//----------------------------------------------------------------------
// DebugInit
//      Initialize so that only DEBUG messages with a flag in flagList 
//	will be printed.
//
//	If the flag is "+", we enable all DEBUG messages.
//
// 	"flagList" is a std::string of characters for whose DEBUG messages are 
//		to be enabled.
//----------------------------------------------------------------------

void
DebugInit(char *flagList)
{
    enableFlags = flagList;
}

//----------------------------------------------------------------------
// DebugIsEnabled
//      Return true if DEBUG messages with "flag" are to be printed.
//----------------------------------------------------------------------

bool
DebugIsEnabled(char flag)
{
    if (enableFlags != NULL)
       return (strchr(enableFlags, flag) != 0) 
		|| (strchr(enableFlags, '+') != 0);
    else
      return false;
}

//----------------------------------------------------------------------
// DEBUG
//      Print a debug message, if flag is enabled.  Like printf,
//	only with an extra argument on the front.
//----------------------------------------------------------------------

void 
VESPER_DEBUG(char flag, char *format, ...)
{
    if (DebugIsEnabled(flag)) {
	va_list ap;
	// You will get an unused variable message here -- ignore it.
	va_start(ap, format);
	vfprintf(stdout, format, ap);
	va_end(ap);
	fflush(stdout);
    }
}

//----------------------------------------------------------------------
// OpenForWrite
//      Open a file for writing.  Create it if it doesn't exist; truncate it
//      if it does already exist.  Return the file descriptor.
//
//      "name" -- file name
//----------------------------------------------------------------------

int
OpenForWrite(const char *name)
{
    int fd = open(name, O_RDWR|O_CREAT|O_TRUNC, 0666);

    SSTMAC_ASSERT(fd >= 0);
    return fd;
}

//----------------------------------------------------------------------
// OpenForReadWrite
//      Open a file for reading or writing.
//      Return the file descriptor, or error if it doesn't exist.
//
//      "name" -- file name
//----------------------------------------------------------------------

int
OpenForReadWrite(const char *name, bool crashOnError)
{
    int fd = open(name, O_RDWR, 0);

    SSTMAC_ASSERT(!crashOnError || fd >= 0);
    return fd;
}

//----------------------------------------------------------------------
// Read
//      Read characters from an open file.  Abort if read fails.
//----------------------------------------------------------------------

void
Read(int fd, char *buffer, int nBytes)
{
    int retVal = read(fd, buffer, nBytes);
    SSTMAC_ASSERT(retVal == nBytes);
}

//----------------------------------------------------------------------
// ReadPartial
//      Read characters from an open file, returning as many as are
//      available.
//----------------------------------------------------------------------

int
ReadPartial(int fd, char *buffer, int nBytes)
{
    return read(fd, buffer, nBytes);
}


//----------------------------------------------------------------------
// WriteFile
//      Write characters to an open file.  Abort if write fails.
//----------------------------------------------------------------------

void
WriteFile(int fd, char *buffer, int nBytes)
{
    int retVal = write(fd, buffer, nBytes);
    SSTMAC_ASSERT(retVal == nBytes);
}

//----------------------------------------------------------------------
// Lseek
//      Change the location within an open file.  Abort on error.
//----------------------------------------------------------------------

void
Lseek(int fd, int offset, int whence)
{
    int retVal = lseek(fd, offset, whence);
    SSTMAC_ASSERT(retVal >= 0);
}


//----------------------------------------------------------------------
// Close
//      Close a file.  Abort on error.
//----------------------------------------------------------------------

void
Close(int fd)
{
    int retVal = close(fd);
    SSTMAC_ASSERT(retVal >= 0);
}


} }
