/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/failure/failure_config.h>
#include <sstmac/hardware/failure/generic_failure_component.h>
#include <sstmac/hardware/failure/failure_rate_model.h>
#include <sstmac/hardware/failure/failure_trace_model.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>

namespace sstmac
{
  namespace hw
  {

    failure_config::failure_config(sim_parameters::ptr params)
    {
      std::string modelname = params->get_param("failure_model");
      if (modelname == "rate")
        {
          type_ = RATE;
        }
      else if (modelname == "trace")
        {
          type_ = TRACE;
          trace_name_ = params->get_param("failure_trace_file");
        }
      else
        {
          throw ssterror(
              "failure_config: unknown failure model type: " + modelname);
        }
      nodes_per_ps_ = params->get_int_param("failure_nodes_per_ps");
      params_ = params;
    }

    void
    failure_config::config_nodes(const std::vector<node::ptr>& list,
        eventmanager::ptr ev_man)
    {
      generic_failure_component::ptr ps;
      std::list<failable::ptr> comps;
      std::map<failable::ptr, failable::ptr> psmap;

      for (long i = 0; i < list.size(); i++)
        {
          if ((i % nodes_per_ps_) == 0)
            {
              ps = generic_failure_component::construct("ps", i);
              ps->set_eventmanager(ev_man);
              comps.push_back(ps);
            }
          node::ptr n = list[i];
          std::list<failable::ptr> thesecomps = n->get_failure_components();
          std::list<failable::ptr>::iterator it, end = thesecomps.end();
          for (it = thesecomps.begin(); it != end; it++)
            {
              psmap[*it] = ps;
            }
          comps.merge(thesecomps);
          //n->init_failure_model(lambda_proc_, lambda_mem_, lambda_disk_,
          //  lambda_nic_, ps, p_ps_node_);
        }

      //now initialize them all
      std::map<failable::ptr, failure_model::ptr> psmodelmap;
      std::list<failable::ptr>::iterator it, end = comps.end();
      for (it = comps.begin(); it != end; it++)
        {
          failure_model::ptr model;
          if (type_ == RATE)
            {
              std::string paramname = "failure_rate_" + (*it)->to_string();
              double rate = params_->get_double_param(paramname);
              model = failure_rate_model::construct(rate);

              if ((*it)->to_string() == "failure_component_ps")
                {
                  psmodelmap[*it] = model;
                }
              else
                {

                }
            }
          else if (type_ == TRACE)
            {
              model = failure_trace_model::construct(trace_name_, (*it)->to_string(), (*it)->failure_id());
            }
          eventmodule::ptr casted = boost::dynamic_pointer_cast<eventmodule>(
              *it);
          if (!casted)
            {
              throw ssterror(
                  "failure_config::config_nodes: failed to cast component to eventmodule");
            }
          (*it)->init_failable(model, casted);
        }

      if (type_ == RATE)
        {
          std::map<failable::ptr, failable::ptr>::iterator it, end =
              psmap.end();
          for (it = psmap.begin(); it != end; it++)
            {
              failure_model::ptr mod = psmodelmap[it->second];

              failure_rate_model::ptr casted = boost::dynamic_pointer_cast<
                  failure_rate_model>(mod);
              double prob = params_->get_double_param("failure_ps_prop");
              casted->add_failable_child(it->first, prob);
            }
        }

    }

    void
    failure_config::config_network(const interconnect::ptr& net)
    {
      long num_switches = net->get_topology()->num_switches();
      for (long i = 0; i < num_switches; i++)
        {
          failable::ptr fail =
              generic_failure_component::construct("switch", i);

          failure_model::ptr model;
          if (type_ == RATE)
            {
              std::string paramname = "failure_rate_switch";
              double rate = params_->get_double_param(paramname);
              model = failure_rate_model::construct(rate);
            }
          else if (type_ == TRACE)
            {
              model = failure_trace_model::construct(trace_name_, "switch", i);
            }

          fail->init_failable(model, net);

          //fails[vertex_index(i)] = fail;
        }
      //net->init_failure_model(fails);
    }
  }

}
