/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_HARDWARE_FAILURE_FAILURE_TRACE_MODEL_H_INCLUDED
#define SSTMAC_HARDWARE_FAILURE_FAILURE_TRACE_MODEL_H_INCLUDED

#include <sstream>

#include <sstmac/hardware/failure/failure_model.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/units.h>
#include <sstmac/common/fileio.h>
namespace sstmac
{
  namespace hw
  {

    class failure_trace_model : public failure_model
    {
    protected:
      std::string name_;
      long uid_;
      bool found_;

      std::ifstream in_;
      failure_trace_model(std::string fname, std::string name, long uid) :
        name_(name), uid_(uid)
      {
        SSTFileIO::open_file(in_, fname);

        if (!in_.is_open())
          {
            std::stringstream ss;
            ss << "failure_trace_model: component " << name << "-" << uid
                << " failed to open input file \"" << fname << "\""
                << std::endl;

            throw valueerror(ss.str());
          }
        else
          {
            SSTMAC_DEBUG << "component " << name << "-" << uid
                  << " successfully opened " << fname
                  << " for failure trace \n";

            found_ = false;
            std::string line;
            while (in_.good())
              {
                std::getline(in_, line);

                //SSTMAC_DEBUG << "failure line: " << line << "\n";
                if (line.size() < 3)
                  continue;
                std::deque<std::string> tok;
                std::string space = " ";
                pst::BasicStringTokenizer::tokenize(line, tok, space);

                if (tok[1] == name_)
                  {
                    long id = atol(tok[2].c_str());

                    if (id == uid_)
                      {
                        found_ = true;
                      }
                  }
              }
            if (!found_)
              {
                in_.close();
              }
            else
              {
                in_.close();
                SSTFileIO::open_file(in_, fname);
              }
          }
      }

    public:
      typedef boost::intrusive_ptr<failure_trace_model> ptr;

      static ptr
      construct(std::string fname, std::string name, long uid)
      {
        return ptr(new failure_trace_model(fname, name, uid));
      }

      virtual
      ~failure_trace_model()
      {
      }

      virtual std::string
      to_string() const
      {
        return "failure_trace_model";
      }

      virtual void
      next_failure(boost::intrusive_ptr<eventmodule> parentmodel)
      {
        if (found_)
          {
            std::string line;

            while (in_.good())
              {
                std::getline(in_, line);

                //SSTMAC_DEBUG << "failure line: " << line << "\n";
                if (line.size() < 3)
                  continue;
                std::deque<std::string> tok;
                std::string space = " ";
                pst::BasicStringTokenizer::tokenize(line, tok, space);

                if (tok[1] == name_)
                  {
                    long id = atol(tok[2].c_str());

                    if (id == uid_)
                      {
                        bool err = false;
                        double val = atof(tok[0].c_str());
                        timestamp time = timestamp(val);

                        if (err)
                          {
                            throw ssterror(
                                "failure_trace_model: couldn't parse time stamp: "
                                    + tok[0]);
                          }

                        SSTMAC_DEBUG
                              << "failure_trace_model: scheduling a failure for "
                              << name_ << "-" << uid_ << " at time " << time
                              << "\n";
                        failure_event::ptr evnt = failure_event::construct();
                        parentmodel->schedule(time, parentmodel, evnt);
                        break;
                      }
                  }
              }
          }
      }

      /**
       * propagate - this function is called when a parent must signal that its
       * children should possibly fail directly because of it (like a power supply)
       */
      virtual void
      propagate()
      {

      }
    };

  }
}

#endif
