#include <sstmac/hardware/flow/flow.h>

#define ZERO_BYTE_CUTOFF 0.7

namespace sstmac {
    namespace hw {

debug flow::dbg_;

flow::flow(
    const sst_message::ptr& msg,
    const eventmodule::ptr& parent,
    const eventhandler::ptr& completion,
    int max_nsegments,
    long min_bytes_per_segment,
    const timestamp& idle_time,
    const event_step_callback::ptr& flow_update_cb
) :
    idle_time_per_segment_(0),
    bytes_per_segment_(0),
    nsegments_(0),
    segment_number_(0),
    msg_(msg),
    completion_(completion),
    parent_(parent),
    flow_update_cb_(flow_update_cb),
    last_bw_(0)
{
    long nbytes = msg->get_byte_length();
    if (nbytes < min_bytes_per_segment){
        nsegments_ = 1;
        bytes_per_segment_ = nbytes;
    }
    else {
        nsegments_ = nbytes / min_bytes_per_segment;
        if (nsegments_ > max_nsegments){
            nsegments_ = max_nsegments;
        }
        bytes_per_segment_ = nbytes / nsegments_;
    }

    if (nbytes % bytes_per_segment_ != 0){
        //make sure at least enough bytes are getting sent
        ++bytes_per_segment_;
    }

    if (idle_time.ticks_int64() != 0)
        idle_time_per_segment_ = idle_time.sec() / nsegments_;

    if (!dbg_.is_real())
    {
        dbg_ = debug("<debug> flow");
        logger::register_logger(&dbg_);
    }

}

flow::ptr
flow::construct(
    const sst_message::ptr &msg,
    const eventmodule::ptr &parent,
    const eventhandler::ptr &completion,
    int max_nsegments,
    long min_nbytes_per_segment,
    const timestamp& idle_time,
    const event_step_callback::ptr& flow_update_cb
)
{
    return new flow(msg, parent, completion,
                    max_nsegments, min_nbytes_per_segment,
                    idle_time,
                    flow_update_cb);
}

std::string
flow::to_string() const
{
    return "flow";
}

void
flow::handle(const sst_message::ptr &msg)
{
    if (flow_update_cb_ && segment_number_ > 0){
        long byte_offset = bytes_per_segment_ * (segment_number_ - 1);
        flow_update_cb_->callback(bytes_per_segment_, byte_offset);
    };

    if (segment_number_ == nsegments_){
        stop();
    }
    else if (segment_number_ == 0){
        register_flow();
        push();
    }
    else {
        push();
    }
}

void
flow::start()
{
    double latency = 0;
    const_link_iterator it = links_.begin(), end = links_.end();
    for ( ; it !=end; ++it){
        link::ptr next_link = *it;
        latency += next_link->latency();
    }

    timestamp next = parent_->now() + timestamp(latency);

    SSTMAC_DEBUG << "Starting flow " << (long) this
                 << " latency=" << latency
                 << " bytes=" << msg_->get_byte_length()
                 << " nsegments=" << nsegments_
                 << " bytes_per_segment=" << bytes_per_segment_
                 << "\n";

    parent_->schedule(next, eventhandler::ptr(this));
}

void
flow::register_flow()
{
    const_link_iterator it = links_.begin(), end = links_.end();
    for ( ; it != end; ++it){
        link::ptr next_link = *it;
        next_link->register_flow();
    }
}

void
flow::push()
{
    SSTMAC_DEBUG << "Scheduling flow update on " << (long) this
                 << " for segment " << segment_number_ << "\n";

    //walk all the links and figure out our bandwidth
    const_link_iterator it = links_.begin(), end = links_.end();
    double link_bw = 0, min_bw = 1e100; //absurd number for now
    for ( ; it != end; ++it){
        link::ptr next_link = *it;
        next_link->deallocate_bandwidth(last_bw_);
        link_bw = next_link->suggest_bandwidth();
        if (link_bw < min_bw)
            min_bw = link_bw;

        if (link_bw < 0){
            sst_throw(illformederror, "flow::push: got negative link bandwidth.\n"
                      "link_bw=%e\n"
                      "max_link_bw=%e\n"
                      "alloc_bw=%e\n"
                      "link_nflows=%d\n"
                      "last_bw=%e\n",
                      link_bw, next_link->max_bw(),
                      next_link->allocated_bandwidth(),
                      next_link->nflows(), last_bw_);
        }

        SSTMAC_DEBUG << "    link: bw=" << min_bw
                     << " max_bw=" << next_link->max_bw()
                     << " alloc_bw=" << next_link->allocated_bandwidth()
                     << " nflows=" << next_link->nflows()
                     << "\n";
    }
    if (min_bw < 1){ //just try again in 10 ns
        timestamp next  = parent_->now() + timestamp(10e-9);
        parent_->schedule(next, eventhandler::ptr(this));
    }

    double delta_t = bytes_per_segment_ / min_bw;

    if (idle_time_per_segment_ != 0){
        //intermittent in the flow, we have some sort of operation that means we're
        //not completely utilizing the available bandwidth
        double percent_utilized = delta_t / (delta_t + idle_time_per_segment_);
        delta_t += idle_time_per_segment_;
        min_bw *= percent_utilized;
    }


    SSTMAC_DEBUG << "flow-delta_t=" << delta_t
                 << " idle-time=" << idle_time_per_segment_
                 << " min-bw=" << min_bw
                 << " mss-bytes=" << bytes_per_segment_ << " bytes\n";


    it = links_.begin();
    for ( ; it != end; ++it){
        link::ptr next_link = *it;
        next_link->allocate_bandwidth(min_bw);
    }

    timestamp next  = timestamp(parent_->now().sec() + delta_t);
    parent_->schedule(next, eventhandler::ptr(this));
    last_bw_ = min_bw;
    ++segment_number_;

}

void
flow::stop()
{
    SSTMAC_DEBUG << "Stopping flow " << (long) this
                 << "\n";

    const_link_iterator it = links_.begin(), end = links_.end();
    for ( ; it != end; ++it){
        link::ptr next_link = *it;
        next_link->deallocate_bandwidth(last_bw_);
        next_link->unregister_flow();
    }
    parent_->schedule(parent_->now(), completion_, msg_);
}

void
flow::add_link(const link::ptr &lnk)
{
    links_.push_back(lnk);
}

flow::link::link(double max_bw, double max_single_bw, const timestamp& latency)
    : max_single_bw_(max_single_bw),
      max_bw_(max_bw),
      latency_(latency.sec()),
      bw_allocated_(0),
      nflows_(0)
{
}

flow::link::link(double max_bw)
    : max_single_bw_(max_bw), max_bw_(max_bw),
      latency_(0), bw_allocated_(0), nflows_(0)
{
}

flow::link::ptr
flow::link::construct(double max_bw, double max_single_bw, const timestamp &latency)
{
    return new link(max_bw, max_single_bw, latency);
}

flow::link::ptr
flow::link::construct(double max_bw)
{
    return new link(max_bw);
}

int
flow::link::nflows() const
{
    return nflows_;
}

double
flow::link::max_bw() const
{
    return max_bw_;
}

void
flow::link::allocate_bandwidth(double bw)
{
    bw_allocated_ += bw;
}

void
flow::link::deallocate_bandwidth(double bw)
{
    bw_allocated_ -= bw;
}

void
flow::link::register_flow()
{
    ++nflows_;
}

void
flow::link::unregister_flow()
{
    --nflows_;
}

double
flow::link::allocated_bandwidth() const
{
    return bw_allocated_;
}

double
flow::link::suggest_bandwidth()
{
    //turn off congestion
    if (nflows_ == 1)
        return max_single_bw_;
    else if (nflows_ == 0)
        sst_throw(illformederror, "flow::link suggesting bandwidth but there are no flows");

    /** rebalance the bws for all the other flows */
    double rebalanced_bw = max_bw_ / nflows_;
    double remaining_bw = max_bw_ - bw_allocated_;
    if (remaining_bw > rebalanced_bw)
        rebalanced_bw = remaining_bw;

    if (rebalanced_bw < max_single_bw_)
        return rebalanced_bw;
    else
        return max_single_bw_;
}

double
flow::link::latency() const
{
    return latency_;
}

    }
}
