#include <sstmac/hardware/memory/numa_memorymodel.h>
#include <sstmac/common/messages/sst_metadata.h>
#include <sstmac/software/libraries/compute/mem_access.h>

namespace sstmac
{
  namespace hw
  {

    SSTRegister("numa",memorymodel,numa_memorymodel,
                "Implements a memory model compatible with a NUMA architecture. "
                "Uses a flow congestion model");

    numa_memorymodel::~numa_memorymodel()
    {
    }

    void
    numa_memorymodel::handle(const sst_message::ptr &msg)
    {
      if(msg->get_type() == sst_message::FAILURE){
          failure_occurred(msg, ptr(this));
          return;
      }
   }

    int
    numa_memorymodel::get_socket(long threadnum) const
    {
        /** For now, just assume the NIC is hooked up
            to socket 0 */
        if (threadnum == sw::thread::nic_thread)
            return 0;
        /** Also assume the RDMA engine is hooked up
            to socket 0 */
        else if (threadnum == sw::thread::rdma_thread)
            return 0;

        int socket = sched_->get_socket(threadnum);
        if (socket == sw::thread::no_socket_affinity)
            return 0;
        return socket;
    }

    void
    numa_memorymodel::finish()
    {
        if (finished_)
            return;

        memorymodel::finish();
        done_ = 0;
    }

    flow::ptr
    numa_memorymodel::construct_flow(const sst_message::ptr &sstmsg)
    {
        sw::mem_access::ptr msg = boost::dynamic_pointer_cast<sw::mem_access>(sstmsg);
        if (!msg){
            sst_throw(valueerror, "numa_memorymodel::do_access: received msg that is not a mem_access");
        }

        event_step_callback::ptr page_event_cb = msg->pop_page_event_callback();
        hw::flow::ptr new_flow = flow::construct(msg, eventmodule::ptr(this), done_,
                                                 flow_nsegments_, flow_nbytes_per_segment_,
                                                 msg->get_event_time(),
                                                 page_event_cb);
        return new_flow;
    }

    void
    numa_memorymodel::access(const sst_message::ptr& msg)
    {
      use_me();

      if (is_skippable(msg)){
          skip_access(msg);
          return;
      }

      flow::ptr new_flow = construct_flow(msg);

      mem_bytes_->record(msg->get_byte_length());

      int dst_socket = get_socket(msg->dest_thread(get_os()));
      int src_socket = get_socket(msg->source_thread(get_os()));


      if (src_socket != dst_socket){
          flow::link::ptr ht_link = socket_links_[src_socket][dst_socket];
          if (!ht_link){
              throw illformederror("numa memory model cannot yet handle anything but a fully connected set of sockets");
          }
          new_flow->add_link(ht_link);
      }
      flow::link::ptr bus_link = bus_links_[dst_socket];
      new_flow->add_link(bus_link);

      new_flow->start();
    }

    numa_memorymodel::ptr
    numa_memorymodel::construct(const sim_parameters::ptr &params)
    {
        timestamp memory_latency = params->get_time_param("node_mem_latency");
        double memory_bw = params->get_bandwidth_param("node_mem_bandwidth");
        double bus_bw = params->get_bandwidth_param("numa_bus_bandwidth");
        int flow_nsegments = params->get_optional_int_param("numa_max_num_pages", flow::default_max_nsegments);
        long flow_nbytes_per_segment = params->get_optional_long_param("numa_page_size", flow::default_min_bytes_per_segment);
        double nic_bw = params->get_bandwidth_param("nic_mem_link_bandwidth");
        timestamp nic_lat = params->get_time_param("nic_mem_link_latency");
        int nsockets = params->reread_optional_int_param("node_sockets", 1);

        numa_memorymodel::ptr model = new numa_memorymodel(
                                                       memory_latency, bus_bw, memory_bw,
                                                       nic_lat, nic_bw,
                                                       nsockets,
                                                       flow_nsegments,
                                                       flow_nbytes_per_segment);


        char param_name[100];
        //for now - I don't know a better way to do this - loop all possible connections
        for (int i=0; i < nsockets; ++i){
            for (int j=0; j < nsockets; ++j){
                if (i==j)
                    continue;

                sprintf(param_name, "numa_hypertransport_bandwidth_%d_%d", i, j);
                if (params->has_param(param_name)){
                    double link_bw = params->get_bandwidth_param(param_name);
                    model->add_hypertransport_link(i, j, link_bw);
                }
            }
        }
        return model;
    }

    void
    numa_memorymodel::init(const node::ptr &parent_node, const computescheduler::ptr &sched)
    {
        memorymodel::init(parent_node, sched);
        nsocket_ = parent_node->nsockets();
    }

    void
    numa_memorymodel::skip_access(const sst_message::ptr& msg)
    {

      //skip all memory accesses
      sw::mem_access::ptr mem = safe_cast(msg, sw::mem_access);
      event_step_callback::ptr flow_update_cb = mem->pop_page_event_callback();
      if (flow_update_cb){
        flow_update_cb->callback(mem->get_byte_length(), 0);
      }
      schedule(now() + timestamp(1e-9), done_, msg);
    }

    bool
    numa_memorymodel::is_skippable(const sst_message::ptr& msg) const
    {
        return false;
        //return msg->get_byte_length() < 512;
    }

    void
    numa_memorymodel::set_eventmanager(const eventmanager::ptr &m)
    {
        memorymodel::set_eventmanager(m);
    }

    sw::operatingsystem::ptr
    numa_memorymodel::get_os() const
    {
        return parent_node_->get_os();
    }

    void
    numa_memorymodel::add_hypertransport_link(int src_socket, int dst_socket, double bw)
    {
        //assume symmetric bidirectional for now
        flow::link::ptr out_link = flow::link::construct(bw);
        socket_links_[src_socket][dst_socket] = out_link;
        flow::link::ptr in_link = flow::link::construct(bw);
        socket_links_[dst_socket][src_socket] = in_link;
    }

    numa_memorymodel::numa_memorymodel(
        const timestamp &activate_latency,
        double bus_bw,
        double mem_bw,
        const timestamp& nic_link_lat,
        double nic_link_bw,
        int nsocket,
        int flow_nsegments,
        long flow_nbytes_per_segment)
        :
          socket_links_(nsocket),
          bus_links_(nsocket),
          nsocket_(nsocket),
          flow_nsegments_(flow_nsegments),
          flow_nbytes_per_segment_(flow_nbytes_per_segment)
      {
        for (int i=0; i < nsocket; ++i){
            bus_links_[i] = flow::link::construct(bus_bw, mem_bw, activate_latency);
        }
        nic_link_ = flow::link::construct(nic_link_bw, nic_link_bw, nic_link_lat);
      }

  }
} /* namespace sstmac */
