#ifndef TRAIN_MEMORYMODEL_H
#define TRAIN_MEMORYMODEL_H

#include <sstmac/hardware/memory/memorymodel.h>
#include <sstmac/hardware/processor/computescheduler.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/hardware/packettrain/trainstation.h>
#include <sstmac/hardware/packettrain/trainlink.h>
#include <sstmac/hardware/node/node.h>

namespace sstmac {
    namespace hw {


  /**
    @class memoryrouter
    Endpoint for memory trainlink.  After memory operation, it routes
    messages to node handler/network as appropriate.
  */
  class memoryrouter :
    public trainstation
  {
      private:
        trainpath::ptr injector_;

        nodeaddress::ptr my_addr_;

        void handle_network(const packettrain::ptr& train);

        static debug dbg_;

      protected:
        memoryrouter(const nodeaddress::ptr& addr, const eventhandler::ptr& handler);

      public:
        typedef boost::intrusive_ptr<memoryrouter> ptr;

        static ptr
        construct(const nodeaddress::ptr& addr, const eventhandler::ptr& handler);

        /**
          Finish reading a message from memory.  If this message is a local
          memory read or an arriving network message, add in total number
          of bytes received.  If a remote send, inject into the network.
          @param msg The message being read in memory
        */
        void handle(const sst_message::ptr &msg);

        /**
          Set the path that will be used for injecting messages into the network.
          @param link The path a message will travel to go from memory to network injection
        */
        void set_injector(const trainpath::ptr& link);

        virtual void finish();

  };

  /**
    @class train_memorymodel
    A memory model compatible with the train message framework
  */
  class train_memorymodel :
    public memorymodel
  {
  protected:
    double freq_hz_;
    double bytes_per_cycle_;
    timestamp cas_latency_;
    bool multipath_;
    bool congestion_;
    output_bw_arbitrator::ptr arb_;

    train_memorymodel(double freq_hz,
                      uint64_t latency_num_cycles,
                      double bytes_per_cycle,
                      bool enable_multipath,
                      bool ignore_congestion,
                      const output_bw_arbitrator::ptr& arb);

    /** The ''router'' for messages being read from memory. Routes
        messages either to the node handler or to the network */
    memoryrouter::ptr router_;

    /** The link to the memory router.  This link's bandwidth
        matches the memory bandwidth */
    trainlink::ptr to_router_;

  public:
    typedef boost::intrusive_ptr<train_memorymodel> ptr;

    static ptr
    construct(const sim_parameters::ptr& params);

    virtual ~train_memorymodel();

    void set_eventmanager(const eventmanager::ptr &m);

    virtual void init(const nodePtr& parent_node, const computescheduler::ptr& sched);

    /**
      @param sw The switch messages are injected to by the NIC
    */
    void set_injector(const trainpath::ptr& sw);

    virtual void finish();

    virtual void handle(const sst_message::ptr &msg);

    virtual void do_access(const sst_message::ptr& msg);

    virtual void access(const sst_message::ptr& msg);

  };

} } /* namespace sstmac */

#endif // TRAIN_MEMORYMODEL_H
