/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/nic/networkinterface.h>
//#include <sstmac/backends/native/dist_eventmap.h>

ImplementFactory(sstmac::hw::interconnect)
;

namespace sstmac
{
  namespace hw
  {
    debug interconnect::dbg_;

    bool interconnect::registered_stats_ = false;
    stat_sum::ptr interconnect::pwrstat_;
    stat_time_window_sum_duration::ptr interconnect::energystat_;
    stat_time_window_sum::ptr interconnect::bandwidthstat_;
    stat_time_window_sum::ptr interconnect::msg_count_;

    interconnect::interconnect()
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> interconnect | network");
        logger::register_logger(&dbg_);
      }

      if (!pwrstat_)
      {
        const std::string fname = "stats_interconnect_pwr.csv";
        pwrstat_ = stat_sum::construct("<stats> interconnect | network", fname);
        register_stat(pwrstat_);
      }
      if (!energystat_)
      {
        const std::string fname = "stats_interconnect_energy.csv";
        energystat_ = stat_time_window_sum_duration::construct(
            "<stats> interconnect | network", fname);
        register_stat(energystat_);
      }
      if (!bandwidthstat_)
      {
        const std::string fname = "stats_interconnect_bandwidth.csv";
        bandwidthstat_ = stat_time_window_sum::construct(
            "<stats> interconnect | network", fname);
        register_stat(bandwidthstat_);
      }
      if (!msg_count_)
      {
        const std::string fname = "stats_interconnect_msg_count.csv";
        msg_count_ = stat_time_window_sum::construct(
            "<stats> interconnect | network", fname);
        register_stat(msg_count_);
      }
    }

    void
    interconnect::init_factory_params(const sim_parameters::ptr& params)
    {
      factory_type::init_factory_params(params);
       /** sstkeyword {
           gui=torus;
           docstring=Specify the basic topology type used for the
           network interconnect;
       }
       */
      topology_ = topology_factory::get_param("topology_name", params);
    }

    //
    // Add one or more copies of the given node type to the manager.
    //
    void
    interconnect::add_nodes(const sim_parameters::ptr& params)
    {
      SSTMAC_DEBUG << "interconnect::add_nodes\n";

      /**
       sstkeyword {
       gui=simple;
       docstring=Specify the model used for the compute node.
       This determines the processor model used in
       estimating compute time;
       }
       */
      sstmac::hw::node::ptr node_tmpl = node_factory::get_param("node_name",
          params);
      /**
       sstkeyword {
       gui=train;
       docstring=Specify the type of network interface controller.
       This determines how injection is modeled and also what
       level of hardware support for things like RDMA is
       available;

       }
       */
      sstmac::hw::networkinterface::ptr nictype =
          networkinterface_factory::get_param("nic_name", params, node_tmpl,
              ptr(this));

      /** I would like to avoid doing this, but there are
       so many calls to parameter node_num it's not worth
       going through the code to fix them all. Just
       take the interconnect object and override
       the parameter. JJW 01/09/12
       */
      std::vector<long>* nodes_to_add = topology_->nodes_to_add();
      long count = nodes_to_add->size();
      // params->add_param_override("node_num", sstprintf("%d", num_nodes()));
      SSTMAC_DEBUG << "interconnect::add_nodes(" << node_tmpl << ", " << count
            << ")\n";

      nodes_.resize(count);
      int i = 0;
      std::vector<long>::iterator it, end = nodes_to_add->end();
      for (it = nodes_to_add->begin(); it != end; it++)
      {
        nodeaddress::ptr id = nodeaddress::convert_from_uniqueid(*it);
        node::ptr newnode = node_tmpl->clone(id);

        sstmac::hw::networkinterface::ptr newnic = nictype->clone(newnode,
            ptr(this), params);
        newnode->set_nic(newnic);

        nodes_[i] = newnode;
        i++;
        SSTMAC_DEBUG << "interconnect: adding nodeid " << id->to_string() << "\n";
      }

      //finish off the template node
      node_tmpl->finish();
      nictype->finish();
    }

    interconnect::~interconnect()
    {
    }

    std::string
    interconnect::label(long node_id) const
    {
      sst_throw(unimplementederror, "interconnect::label for %s",
          to_string().c_str());
    }

    void
    interconnect::set_eventmanager(const eventmanager::ptr &m)
    {
      SSTMAC_DEBUG << "interconnect::set_eventmanager \n";
      eventmodule::set_eventmanager(m);

      if (sstmac_env::run_mode_ != sstmac_env::SHARED_MEMORY)
      {

        dbg_.set_partition_id(m->get_lpid());
      }

      // Initialize the nodes.
      for (int i = 0; i < nodes_.size(); ++i)
      {
        node::ptr the_node = nodes_[i];
        the_node->initialize(m, topology_);
        the_node->get_nic()->set_eventmanager(m);
      }

      if (fail_)
      {
        fail_->config_nodes(nodes_, m);
        fail_->config_network(ptr(this));
      }
    }

    void
    interconnect::init_failure_model(const sim_parameters::ptr &params)
    {
      /**
       sstkeyword {
       docstring = Specify the failure model [if any] for the nodes;
       gui = none;
       none = No failures occur;
       }
       */
      std::string failmodel = params->get_optional_param("failure_model",
          "none");
      if (!fail_ && failmodel != "none")
      {
        SSTMAC_DEBUG << "manager::init_failure model\n";
        fail_ = failure_config::construct(params);
      }
    }

    void
    interconnect::finish()
    {
      eventmodule::finish();

      long num_nodes = nodes_.size();
      for (long i = 0; i < num_nodes; ++i)
      {
        node::ptr node = nodes_[i];
        if (node)
        {
          node->finish();
        }
      }
      nodes_.clear();
    }

    void
    interconnect::clear_statics()
    {
      pwrstat_ = 0;
      energystat_ = 0;
      bandwidthstat_ = 0;
      msg_count_ = 0;
    }

  }
}
