/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_CONGESTION_INTERCONNECT_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_CONGESTION_INTERCONNECT_H_INCLUDED

#include <sstmac/common/timestamp.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/stats/power_config.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/common/component.h>
#include <sstmac/common/stats/stats_common.h>
#include <sstmac/common/stats/stat_time_window.h>
#include <sstmac/hardware/failure/failable.h>
#include <sstmac/hardware/failure/generic_failure_component.h>
#include <sstmac/hardware/failure/failure_config.h>

namespace sstmac
{
  namespace hw
  {

    /**
     * Base class for network congestion models.
     */
    class interconnect :
            virtual public nocopy,
            public factory_type,
            public eventmodule
    {
    public:
      typedef boost::intrusive_ptr<interconnect> ptr;
      typedef boost::intrusive_ptr<const interconnect> const_ptr;

      virtual std::string
      to_string() const
      {
        return "interconnect";
      }

      virtual ~interconnect();

      // Optional finish function for children
      virtual void finish();

      static void clear_statics();

      /// The maximum number of nodes on this network.
      /// Return a negative (really <= 0) value to indicate an infinite capacity.
      virtual long
      num_nodes() const {
        return topology_->num_nodes();
      }

      virtual std::string label(long node_id) const;

      virtual topology::ptr
      get_topology() const
      {
        return topology_;
      }

/* removing graph/vertex breaks this
      virtual void
      init_failure_model(const std::map<vertex_index, failable::ptr>& c)
      {
        failuremodel_ = true;
        fail_switches_ = c;
      }
*/

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      virtual void
      init_factory_params(const sim_parameters::ptr &params);

    protected:

      interconnect();

      /// The node objects.
      std::vector<node::ptr> nodes_;

      failure_config::ptr fail_;

      void add_nodes(const sim_parameters::ptr& params);

      void init_failure_model(const sim_parameters::ptr &params);

      static debug dbg_;

      // stats
      //netstat::ptr nstat;
      // powerstat::ptr pwrstat;

      power_config::ptr pwr_config_;

      static bool registered_stats_;
      static stat_sum::ptr pwrstat_;
      static stat_time_window_sum_duration::ptr energystat_;
      static stat_time_window_sum::ptr bandwidthstat_;
      static stat_time_window_sum::ptr msg_count_;

      //std::map<vertex_index, failable::ptr> fail_switches_;
      //bool failuremodel_;

      /// The network topology.  This object should eventually take over the
      /// ownership of per-link latencies and bandwidths.
      topology::ptr topology_;

    };

    DeclareFactory(interconnect);

  }
} // end of namespace sstmac

#endif
