/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstream>

#include <sstmac/hardware/network/switches/cycleaccurate/cycleaccurateswitch.h>
#include <sstmac/hardware/network/switches/cycleaccurate/ca_routing_info.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("cycleaccurate | accurate", networkswitch, cycleaccurateswitch);

    stat_time_window_sum::ptr cycleaccurateswitch::power_stat_;
    double cycleaccurateswitch::static_power_ = 0;

    stat_sum::ptr cycleaccurateswitch::stat_bytes_ = 0;

    long long cycleaccurateswitch::packetinjector::MSG_COUNT = 0;

    cycleaccurateswitch::cycleaccurateswitch() :
      using_orion_(false)
    {
      nodebufcnt_ = 0;
      clockmsg_ = sst_message::construct();

      if (!stat_bytes_)
      {
        const std::string fname = "stat_cycleaccurate_bytes.csv";
        stat_bytes_ = stat_sum::construct(
            "<stats> network | interconnect | cycleaccurate", fname);
        register_stat(stat_bytes_);
      }

    }

    networkswitch::ptr
    cycleaccurateswitch::clone(router::ptr route, nodeaddress::ptr addr) const
    {
      ptr ret = new cycleaccurateswitch;
      clone_into(ret, addr, route);
      ret->init_members(const_ptr(this));
      return ret;
    }

    injector::ptr
    cycleaccurateswitch::get_new_injector(sim_parameters::ptr params)
    {
      return packetinjector::construct(params, ptr(this));
    }

    ejector::ptr
    cycleaccurateswitch::get_new_ejector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetejector::construct(params, self, my_addr_);
    }

    void
    cycleaccurateswitch::init_members(const cycleaccurateswitch::const_ptr &sw)
    {
      discrete_network_switch::init_members(sw);

      numVC_ = sw->numVC_;
      inBuffSize_ = sw->inBuffSize_;
      outBuffSize_ = sw->outBuffSize_;

      arbLatency_ = sw->arbLatency_;
      switchLatency_ = sw->switchLatency_;
      switchBandwidth_ = sw->switchBandwidth_;

      mtu = sw->mtu;

      clockFreq_ = sw->clockFreq_;
      clockper_ = sw->clockper_;

      link_lat_ = sw->link_lat_;
      inj_bw_ = sw->inj_bw_;

      int nodes = top_->nps(my_addr_);
      for (int i = 0; i < nodes; i++)
      {
        long port = router_->convert_to_outport(topology::EJECT, i);
        portbw_[port] = inj_bw_;
      }

      packet_timeout_ = sw->packet_timeout_;

      if (sw->using_orion_)
      {
        init_orion();
      }

    }

    void
    cycleaccurateswitch::init_factory_params(const sim_parameters::ptr& params)
    {
      srand(time(NULL));

      discrete_network_switch::init_factory_params(params);

      numVC_ = params->get_int_param("cycleaccurateswitch_vc");
      inBuffSize_ = params->get_int_param("cycleaccurateswitch_input_buffer");
      outBuffSize_ = params->get_int_param("cycleaccurateswitch_output_buffer");

      arbLatency_ = params->get_time_param(
          "cycleaccurateswitch_latency_arbitration");
      switchLatency_ = params->get_time_param(
          "cycleaccurateswitch_latency_xbar");
      switchBandwidth_ = params->get_bandwidth_param(
          "cycleaccurateswitch_bandwidth_xbar");

      mtu = params->get_int_param("cycleaccurateswitch_MTU");

      inj_bw_
          = params->get_bandwidth_param("cycleaccurateswitch_bandwidth_n2r");

      clockFreq_ = params->get_freq_param("cycleaccurateswitch_freq");
      clockper_ = timestamp(1.0 / clockFreq_);

      link_lat_ = params->get_time_param("cycleaccurateswitch_latency_link");
      packet_timeout_ = params->get_optional_time_param(
          "cycleaccurateswitch_packet_timeout", timestamp(10e-6));

      if (!power_stat_)
      {
        stat_time_window::TIME_INTERVAL = timestamp(0.0001);
        std::stringstream ss;
        ss << "network_power.csv";
        stat_name_ = ss.str();
        power_stat_ = stat_time_window_sum::construct("<stats> router",
            stat_name_);
        //this is okay - logger doesn't handle smart ptrs
        logger::register_logger(power_stat_.get());
      }

      if (params->get_optional_bool_param("cycleaccurateswitch_orion", false))
      {
        init_orion();
      }

    }

    void
    cycleaccurateswitch::initialize()
    {
      init_buffers();

      arb_inport_ = inbuffcnt_.begin();
    }

    void
    cycleaccurateswitch::set_inport(const cycleaccurateswitch::ptr &sw,
        int dim, int dir)
    {
      int port = router_->get_inport(sw->get_addr(), get_addr(), dim, dir);
      inbuffcnt_[port] = 0;

      if (port >= 0)
      {
        for (int j = 0; j < numVC_; j++)
        {
          inbuffs_[port * numVC_ + j] = std::queue<cycleaccuratepayload::ptr>();

          credits_[port * numVC_ + j] = inBuffSize_;

          totalinbufspace_ += inBuffSize_;
        }
      }
      inports_free_[port] = timestamp(0);
    }

    void
    cycleaccurateswitch::init_orion()
    {
      //--------------------------------------------------------------------------
      //------- ORION stuff --------
      //--------------------------------------------------------------------------


      using_orion_ = true;

      int technology = 32;
      int trans_type = NVT;
      double voltage = 0.8;
      double frequency = clockFreq_;

      ORION_Tech_Config *conf = new ORION_Tech_Config(technology, trans_type,
          voltage, frequency);

      // ---------- input buffer ---------------
      int buf_type = SRAM;
      int is_fifo = true;
      int n_read_port = MAX_GRANTS;
      int n_write_port = 1;
      line_width_ = 128;
      int n_entry = inBuffSize_ / line_width_;

      int outdrv = false;

      ORION_Array_Info* info = new ORION_Array_Info();
      info->init(buf_type, is_fifo, n_read_port, n_write_port, n_entry,
          line_width_, outdrv, conf);

      power_buffer_in_ = new ORION_Array();
      power_buffer_in_->init(info);

      //std::cout << "bufferin: " << power_buffer_in_->stat_static_power() << "\n";
      static_power_ += numVC_ * outports_.size()
          * power_buffer_in_->stat_static_power();

      // ------------ output buffer --------------
      int out_buf_type = SRAM;
      int out_is_fifo = true;
      int out_n_read_port = 1;
      int out_n_write_port = 1;

      int out_n_entry = outBuffSize_ / line_width_;

      int out_outdrv = true;

      ORION_Array_Info* out_info = new ORION_Array_Info();
      out_info->init(out_buf_type, out_is_fifo, out_n_read_port,
          out_n_write_port, out_n_entry, line_width_, out_outdrv, conf);

      power_buffer_out_ = new ORION_Array();
      power_buffer_out_->init(out_info);

      static_power_ += outports_.size()
          * power_buffer_out_->stat_static_power();

      // --------------- arbiter ------------------
      int model = RR_ARBITER;
      int ff_model = NEG_DFF;
      int req_width = (outports_.size() - 1) * numVC_;
      double length = outports_.size() * n_entry
          * conf->PARM("CrsbarCellWidth");

      power_arbiter_ = new ORION_Arbiter();

      power_arbiter_->init(model, ff_model, req_width, length, conf);
      static_power_ += power_arbiter_->report_static_power();

      // ----------------- crossbar -----------------
      int xmodel = MATRIX_CROSSBAR;
      int ins = outports_.size();
      int outs = outports_.size();
      int in_seg = 0; //used by segmented crossbar
      int out_seg = 0; //used by segmented crossbar
      int degree = 4; //used by multree crossbar
      int connect_type = TRISTATE_GATE;
      double in_len = 0; //use 0 if unknown, ORION will figure it out
      double out_len = 0;

      power_xbar_ = new ORION_Crossbar();
      power_xbar_->init(xmodel, ins, outs, in_seg, out_seg, line_width_,
          degree, connect_type, trans_type, in_len, out_len, conf);
      static_power_ += power_xbar_->report_static_power();

      // --------------- links -------------
      double link_len = 1000; // in um
      link_e_per_bm_ = ORION_Link::LinkDynamicEnergyPerBitPerMeter(link_len,
          conf);
      static_power_ += ORION_Link::LinkLeakagePowerPerMeter(link_len, conf);
      link_length_ = link_len / 1000000; // in meters


    }

    void
    cycleaccurateswitch::init_buffers()
    {
      totalinbufspace_ = 0;
      totalinbufocc_ = 0;

      boost::unordered_map<long, eventhandler::ptr>::iterator it, end =
          outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        long port = it->first;
        SSTMAC_DEBUG << "cycleaccurateswitch: initializing buffer for port "
              << port << " with bw " << portbw_[port] << "\n";
        outbuffs_[port] = outbuffer::construct(port, ptr(this), link_lat_,
            portbw_[port]);
        outports_free_[port] = timestamp(0);

        arb_vc_[port] = 0;
      }
      outports_free_[-1] = timestamp(0);

      //input buffers coming from nodes
      for (int i = 1; i <= router_->num_nodes(); i++)
      {
        inports_free_[-i] = timestamp(0);
        arb_vc_[-i] = 0;

        for (int j = 0; j < numVC_; j++)
        {
          inbuffs_[-i * numVC_ + j] = std::queue<cycleaccuratepayload::ptr>();
          totalinbufspace_ += inBuffSize_;
        }

        inbuffcnt_[-i] = 0;

      }
    }

    void
    cycleaccurateswitch::finish()
    {
      if (finished_)
        return;

      discrete_network_switch::finish();

      SSTMAC_DEBUG << "cycleaccurateswitch: ------- finish -------- \n";
      boost::unordered_map<long, eventhandler::ptr>::iterator it, end =
          outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        long port = it->first;
        for (int j = 0; j < numVC_; j++)
        {

          int index = port * numVC_ + j;

          if (port >= 0 && credits_[index] != inBuffSize_)
          {
            std::cout << "cycleaccurateswitch[" << my_addr_->to_string()
                << "]: credits for VC " << j << " on inport " << port << " ("
                << credits_[index] << ") do not equal buff size("
                << inBuffSize_ << ")" << std::endl;
          }

          if (!inbuffs_[index].empty())
          {
            std::cout << "cycleaccurateswitch[" << my_addr_->to_string()
                << "]: " << inbuffs_[index].size()
                << " messages still in inbuff " << port << ", vc = " << j
                << std::endl;

            cycleaccuratepayload::ptr msg = inbuffs_[index].front();
            //int dim;
            //int dir;
            //ca_routing_info::ptr ri = ca_routing_info::construct(
            //    msg->orig()->toaddr());
            //router_->route(ri, dim, dir);
            int outport = router_->route(msg);

            std::cout << "cycleaccurateswitch[" << my_addr_->to_string()
                << "]: front message size " << msg->get_byte_length()
                << ", outport: " << outport << ", outport free: "
                << outports_free_[outport];

            std::cout << ", outport buff size: "
                << outbuffs_[outport]->get_occupancy() << ", credits: "
                << credits_[outport * numVC_ + j] << std::endl;

            std::cout << "message is: " << msg->to_string() << "\n";
          }
        }

        if (outbuffs_[port]->get_occupancy() > 0)
        {
          std::cout << "cycleaccurateswitch[" << my_addr_->to_string() << "]: "
              << outbuffs_[port]->get_occupancy()
              << " messages still in outbuff " << port << std::endl;
        }
      }

      if (using_orion_)
      {
        delete power_arbiter_;
        delete power_xbar_;
        delete power_buffer_in_;
        delete power_buffer_out_;
      }

      stat_bytes_->simulation_finished(now());
    }

    void
    cycleaccurateswitch::outbuffer::clear_outbuff()
    {
      std::list<timestamp> clear;
      std::map<timestamp, long>::iterator it, end = outbuff_occ_clear_.end();
      for (it = outbuff_occ_clear_.begin(); it != end; it++)
      {
        if (it->first > parent_->now())
        {
          break;
        }
        else
        {
          clear.push_back(it->first);
          outbuff_occ_ -= it->second;
        }
      }

      std::list<timestamp>::iterator it2, end2 = clear.end();
      for (it2 = clear.begin(); it2 != end2; it2++)
      {
        outbuff_occ_clear_.erase(*it2);
      }
    }

    long
    cycleaccurateswitch::outbuffer::get_occupancy()
    {
      clear_outbuff();

      return outbuff_occ_;
    }

    void
    cycleaccurateswitch::outbuffer::handle(const sst_message::ptr &msg)
    {
      if (first_time_)
      {
        parent_->schedule(parent_->now() + timestamp(10e-6),
            boost::intrusive_ptr<outbuffer>(this), vis_update_);
        first_time_ = false;
      }
      if (msg == vis_update_)
      {
        if (parent_->linkvis_[pid_])
        {
          double bw = (double) total_bytes_ / (double) 10e-6;
          double act = bw / link_bw_;
          parent_->linkvis_[pid_]->set_activity(act);
          total_bytes_ = 0;
          parent_->schedule(parent_->now() + timestamp(10e-6),
              boost::intrusive_ptr<outbuffer>(this), vis_update_);
        }
        return;
      }

      //messages arrive from the crossbar, so schedule them to leave
      SSTMAC_DEBUG << "outbuffer(" << parent_->get_addr()->to_string()
            << "- port " << pid_ << ")::handling message \n";

      cycleaccuratepayload::ptr n = safe_cast(msg, cycleaccuratepayload);

      if (parent_->using_orion_)
      {
        //power for arriving to outbuff
        parent_->power_stat_->record(
            parent_->now(),
            parent_->power_buffer_out_->stat_energy(0,
                n->get_byte_length() / parent_->line_width_, 0));

        //power for coming out of the crossbar
        double initial = parent_->power_xbar_->report();
        //parent_->power_xbar_->record(0, n->inport_, pid_, n->get_byte_length());
        parent_->power_stat_->record(parent_->now(),
            parent_->power_xbar_->report() - initial);

        //power for leaving outbuff
        parent_->power_stat_->record(
            parent_->now(),
            parent_->power_buffer_out_->stat_energy(
                n->get_byte_length() / parent_->line_width_, 0, 0));

        parent_->power_stat_->record(
            parent_->now(),
            parent_->link_e_per_bm_ * parent_->link_length_
                * parent_->line_width_ * n->get_byte_length() * 8);
      }

      outbuff_occ_ += n->get_byte_length();

      vc_bytes_[n->vc_] += n->get_byte_length();

      //assign new vc


      timestamp txFinishTime = finish_;
      if (txFinishTime < parent_->now())
        txFinishTime = parent_->now();
      timestamp delay = link_lat_ + timestamp(n->get_byte_length() / link_bw_);

      if (pid_ < 0)
      {
        //n->inport_ = -1;
      }
      else
      {
        if (!parent_->outports_[pid_])
        {
          sst_throw(ssterror,
              "cycleaccurateswitch::outbuffer::handle: can't find outport %d",
              pid_);
        }
        discrete_network_switch::ptr net = boost::dynamic_pointer_cast<
            discrete_network_switch>(parent_->outports_[pid_]);
        n->inport_ = pid_;

        total_bytes_ += n->get_byte_length();
      }

      SSTMAC_DEBUG << "outbuffer(" << parent_->get_addr()->to_string()
            << "- port " << pid_ << ") message going to "
            << parent_->outports_[pid_]->to_string()
        //<< " at inport " << n->inport_
            << " with delay " << delay << "\n";

      parent_->schedule(txFinishTime + delay, parent_->outports_[pid_], n);

      finish_ = txFinishTime + delay - link_lat_;

      if (outbuff_occ_clear_.find(finish_) != outbuff_occ_clear_.end())
      {
        throw ssterror(
            "cycleaccurateswitch::outbuffer:: finish time already exists in out buff clear map");
      }
      outbuff_occ_clear_.insert(std::make_pair(finish_, n->get_byte_length()));

      if (!parent_->clockmsg_->is_scheduled())
      {
        parent_->schedule(parent_->now() + parent_->clockper_, parent_,
            parent_->clockmsg_);
      }

    }

    void
    cycleaccurateswitch::handle(const sst_message::ptr &cmsg)
    {
      SSTMAC_DEBUG << "cycleaccurateswitch::handle \n";
      if (!cmsg || cmsg == clockmsg_)
      {
        arbitrate();
      }
      else
      { //it's a data message arriving


        if (cmsg->get_type() == sst_message::ACK)
        {
          cycleaccurateack::ptr n = safe_cast(cmsg, cycleaccurateack);
          int inport = n->inport_;
          int inbuffindex = inport * numVC_ + n->vc_;
          SSTMAC_DEBUG << "switch[" << my_addr_->to_string() << "]: message "
                << n->to_string() << " which is an ACK arrived on inport " << inport
                << " on VC " << n->vc_ << " with index " << inbuffindex
                << " ...";

          int creds = n-> get_byte_length();

          credits_[inbuffindex] += creds;

          SSTMAC_DEBUG << "credits up by " << creds << " to "
                << credits_[inbuffindex] << "\n";

          if (!clockmsg_->is_scheduled())
          {
            schedule(now() + clockper_, ptr(this), clockmsg_);
          }

          return;
        }
        else
        { //it's data

          cycleaccuratepayload::ptr n = safe_cast(cmsg, cycleaccuratepayload);
          int inport = n->inport_;
          int inbuffindex = inport * numVC_ + n->vc_;
          SSTMAC_DEBUG << "switch[" << my_addr_->to_string() << "]: message "
                << n->to_string() << " arrived on inport " << inport
                << " on VC " << n->vc_ << " with index " << inbuffindex
                << " ...";

          SSTMAC_DEBUG << "data \n";

          n->bufferarrive_ = now();

          totalinbufocc_ += n->get_byte_length();
          double act = (double) totalinbufocc_ / (double) totalinbufspace_;
          SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
                << "]:total input buffer occupancy: " << act << "\n";
          vis_set_activity(act, now());

          inbuffs_[inbuffindex].push(n);
          inbuffcnt_[inport]++;

          if (using_orion_)
          {
            //power for entering inport
            power_stat_->record(
                now(),
                power_buffer_in_->stat_energy(0,
                    n->get_byte_length() / line_width_, 0));
          }

          if (!clockmsg_->is_scheduled())
          {
            schedule(now() + clockper_, ptr(this), clockmsg_);
          }

          stat_bytes_->record(n->get_byte_length());

        }
      }

    }

    void
    cycleaccurateswitch::arbitrate()
    {
      SSTMAC_DEBUG << "cycleaccurateswitch::arbitrate \n";
      //iterate over input ports, starting with the last one that went
      int maxGrants = MAX_GRANTS;
      int numGrants = 0;
      bool clock = false;
      int began = arb_inport_->first;
      bool firstinport = true;
      for (; firstinport || (arb_inport_->first != began && numGrants
          < maxGrants);)
      {
        firstinport = false;
        numGrants += check_inport(arb_inport_->first, clock);
        arb_inport_++;
        if (arb_inport_ == inbuffcnt_.end())
        {
          arb_inport_ = inbuffcnt_.begin();
        }
      }

      if (clock)
      {

        if (using_orion_)
        {
          //power for using the arbiter
          double initial = power_arbiter_->report();
          power_arbiter_->record(numGrants, numGrants);
          power_stat_->record(now(), power_arbiter_->report() - initial);
        }
        schedule(now() + clockper_, ptr(this), clockmsg_);
      }
    }

    int
    cycleaccurateswitch::get_queue_length(int port) const
    {
      std::map<int, outbuffer::ptr>::const_iterator it = outbuffs_.find(port);
      if (it == outbuffs_.end())
      {
        sst_throw(valueerror,
            "cycleaccurateswitch::get_queue_length: invalid port %d",
            port);
      }
      const outbuffer::ptr& outbuf = it->second;
      return outbuf->get_occupancy() / networkswitch::packet_length_bytes();
    }

    int
    cycleaccurateswitch::check_inport(int inport, bool &clock)
    {
      int ret = 0;

      /* SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
       << "]: examining inport " << inport << ", num_in_buff = "
       << inbuffcnt_[inport] << ", time = " << now() << ", inport_free = "
       << inports_free_[inport].sec() << "\n";*/

      if (inbuffcnt_[inport] > 0)
      {
        if (inports_free_[inport] <= now())
        {

          int beginVC = (arb_vc_[inport] + 1) % numVC_;
          bool firstVC = true;
          for (int vc = beginVC; firstVC || vc != beginVC; vc = (vc + 1)
              % numVC_)
          {

            firstVC = false;
            int index = inport * numVC_ + vc;

            /*SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
             << "]: examining vc " << vc << ", with index " << index
             << ", queue size = " << inbuffs_[index].size() << "\n";*/

            if (!inbuffs_[index].empty())
            {
              cycleaccuratepayload::ptr msg = inbuffs_[index].front();
              //int dim;
              //int dir;
              //ca_routing_info::ptr ri = ca_routing_info::construct(addr);
              //router_->route(ri, dim, dir);
              int outport = router_->route(msg);
              //int outport = router_->convert_to_outport(dim, dir);

              if (msg->vc_ != vc)
              {
                throw ssterror(
                    "cycleaccurate switch: vcs don't match for some reason");
              }

              SSTMAC_DEBUG << "packet going to outport " << outport
                    << ", outports_free = " << outports_free_[outport].sec();

              SSTMAC_DEBUG << ", outbuff occupancy = "
                    << outbuffs_[outport]->get_occupancy() << "\n";

              if (outport >= 0)
              {
                SSTMAC_DEBUG << ", credits = " << credits_[outport * numVC_
                      + vc] << "\n";
              }

              bool creditcheck = true;

              if (outport >= 0)
              {
                if (inport < 0 || msg->prevport_ != outport)
                {
                  creditcheck = credits_[outport * numVC_ + vc]
                      >= msg->get_byte_length() + 2 * mtu;
                }
                else
                {
                  creditcheck = credits_[outport * numVC_ + vc]
                      >= msg->get_byte_length();
                }
              }

              if (outports_free_[outport] <= now() && (outBuffSize_
                  - outbuffs_[outport]->get_occupancy()
                  >= msg->get_byte_length() && creditcheck))
              {
                SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
                      << "]: arbitration good from " << inport << " to "
                      << outport << " \n";

                inbuffs_[inport * numVC_ + vc].pop();
                inbuffcnt_[inport]--;
                //now we're ok to send through the crossbar
                //  sendDelayed(msg, arbLatency_, outports[outport]);
                timestamp outbuffarrival = now() + switchLatency_ + timestamp(
                    msg->get_byte_length() / switchBandwidth_);
                int prevport = msg->prevport_;
                discrete_network_switch::ptr prevswitch = msg->prevswitch_;
                msg->prevswitch_ = ptr(this);
                msg->prevport_ = outport;

                schedule(outbuffarrival, outbuffs_[outport], msg);

                totalinbufocc_ -= msg->get_byte_length();
                double act = (double) totalinbufocc_
                    / (double) totalinbufspace_;
                SSTMAC_DEBUG << "switch[" << my_addr_->to_string()
                      << "]:total input buffer occupancy: " << act << "\n";
                vis_set_activity(act, now());

                if (using_orion_)
                {
                  //power for leaving inport
                  power_stat_->record(
                      now(),
                      power_buffer_in_->stat_energy(
                          msg->get_byte_length() / line_width_, 0, 0));

                  //power for entering crossbar
                  double initial = power_xbar_->report();
                  power_xbar_->record(1, inport, outport,
                      msg->get_byte_length());
                  power_stat_->record(now(), power_xbar_->report() - initial);
                }

                inports_free_[inport] = outbuffarrival - switchLatency_;
                outports_free_[outport] = outbuffarrival;
                arb_vc_[inport] = vc;
                ret++;
                clock = true;

                if (outport >= 0)
                {
                  int inbuffindex = outport * numVC_ + vc;
                  credits_[inbuffindex] -= msg->get_byte_length();
                }

                if (inport >= 0)
                {
                  //send credit ack back
                  cycleaccurateack::ptr ack = cycleaccurateack::construct(
                      msg->get_byte_length(), vc);

                  ack->inport_ = prevport;

                  SSTMAC_DEBUG << "ca-switch(" << my_addr_->to_string()
                        << "): sending ack back to " << prevswitch->to_string()
                        << "\n";

                  schedule(now() + link_lat_, prevswitch, ack);
                }

                break;

              }
              else if (msg->bufferarrive_ + packet_timeout_ >= now())
              {
                //cycle it
                inbuffs_[inport * numVC_ + vc].pop();
                inbuffs_[inport * numVC_ + vc].push(msg);
              }
            }
          }
        }
      }

      return ret;
    }

    void
    cycleaccurateswitch::packetinjector::send(long nbytes, long byte_offset,
        const sst_message::ptr &msg)
    {
      timestamp delay(0);
      //int vc = rand() % parent_->numVC_;
      int vc = next_vc_;
      next_vc_ = (next_vc_ + 1) % parent_->numVC_;
      int inport;

      if (parent_->nodebuffers_.find(msg->fromaddr())
          == parent_->nodebuffers_.end())
      {
        SSTMAC_DEBUG << "cycleaccurateswitch("
              << parent_->get_addr()->to_string()
              << ")::packetinjector - adding node inport "
              << (parent_->nodebufcnt_ - 1) << " for address "
              << msg->fromaddr()->to_string() << "\n";
        std::map<nodeaddress::ptr, int>::iterator it, end =
            parent_->nodebuffers_.end();
        for (it = parent_->nodebuffers_.begin(); it != end; it++)
        {
          SSTMAC_DEBUG << "in map: " << it->first->to_string() << "\n";
          SSTMAC_DEBUG << "equals: " << it->first->equals(msg->fromaddr())
                << "\n";
          SSTMAC_DEBUG << "==: " << (it->first == msg->fromaddr()) << "\n";
        }

        parent_->nodebuffers_[msg->fromaddr()] = --parent_->nodebufcnt_;
      }
      inport = parent_->nodebuffers_[msg->fromaddr()];

      long bytes_sent = byte_offset;
      long bytes_total = msg->get_byte_length();
      bool tail_packet = false;
      if (MTU_ > 0)
      {
        SSTMAC_DEBUG << "packetinjector: injecting packets: "
              << msg->to_string() << ", bytes-to-send=" << nbytes
              << ", total-msg-size=" << msg->get_byte_length() << ", MTU="
              << MTU_ << "\n";
        long bytes_left = std::min(nbytes, bytes_total - byte_offset);
        long long msgid =  MSG_COUNT++;
        while (bytes_left > 0)
        {
          long bytes_to_send = std::min(MTU_, bytes_left);
          bytes_sent += bytes_to_send;
          bytes_left -= bytes_to_send;
          tail_packet = bytes_sent >= bytes_total;
          cycleaccuratepayload::ptr pack = cycleaccuratepayload::construct(msg,
              bytes_to_send, tail_packet, vc);
          pack->msg_id_ = msgid;
          pack->inport_ = inport;

          parent_->schedule(parent_->now() + delay, parent_, pack);
          delay = delay + timestamp((bytes_to_send / n2r_));

        }
      }
      else
      {
        delay = timestamp((nbytes / n2r_));
        SSTMAC_DEBUG << "packetinjector: injecting whole message: "
              << msg->to_string() << " with delay " << delay
              << ", bytes-to-send=" << ", total-msg-size="
              << msg->get_byte_length() << ", bw=" << n2r_ << "\n";

        bytes_sent = byte_offset + nbytes;
        tail_packet = bytes_sent >= bytes_total;
        cycleaccuratepayload::ptr pack = cycleaccuratepayload::construct(msg,
            nbytes, tail_packet, vc);
        pack->msg_id_ = MSG_COUNT++;
        pack->inport_ = inport;
        parent_->schedule(parent_->now() + delay, parent_, pack);
      }

      if (msg->get_needs_ack() && tail_packet)
      {
        sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
        parent_->schedule(parent_->now() + delay, sendhandle_[msg->fromaddr()],
            sendack);
      }

    }

    void
    cycleaccurateswitch::packetejector::handle(const sst_message::ptr &msg)
    {
      cycleaccuratepayload::ptr pack = safe_cast(msg, cycleaccuratepayload);

      SSTMAC_DEBUG << "packetejector: got packet: " << pack->to_string()
            << "\n";

      if (!parent_->clockmsg_->is_scheduled())
      {
        parent_->schedule(parent_->now() + parent_->clockper_, parent_,
            parent_->clockmsg_);
      }

      if (pack->is_tail())
      {
        tails_[pack->msg_id_] = pack;
        SSTMAC_DEBUG << "ejector got tail for msg id " << pack->msg_id_ << "\n";
      }
      recv_bytes_[pack->msg_id_] += pack->get_byte_length();
      if (recv_bytes_[pack->msg_id_] >= pack->orig()->get_byte_length())
      {

        if (tails_.find(pack->msg_id_) == tails_.end())
        {
          sst_throw(ssterror, "caswitch::ejector::handle - tail not here yet, problem");
        }
        if (recvhandles_.find(pack->orig()->toaddr()) == recvhandles_.end())
        {
          std::cerr << "my addr: " << addr_->to_string() << "\n";
          std::cerr << "message headed to: "
              << pack->orig()->toaddr()->to_string() << "\n";
          std::cerr << "message came from: "
              << pack->orig()->fromaddr()->to_string() << "\n";
          std::cerr << "message: " << msg->to_string() << "\n";
          throw ssterror(
              "cycleaccurateswitch::packetejector::handle() -- can't find recv handler, that's a problem");
        }
        SSTMAC_DEBUG << "packetejector: ejecting message to "
              << pack->orig()->toaddr()->to_string() << "\n";
        parent_->schedule(parent_->now(), recvhandles_[pack->orig()->toaddr()],
            tails_[pack->msg_id_]);

        recv_bytes_.erase(pack->msg_id_);
        tails_.erase(pack->msg_id_);
      }
      else
      {
        SSTMAC_DEBUG << "packetejector: recv bytes up to "
              << recv_bytes_[pack->msg_id_] << " but waiting for "
              << pack->orig()->get_byte_length() << "\n";
      }

    }

  }
}
