/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_NETWORKSWITCH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_NETWORKSWITCH_H_INCLUDED

#include <sstmac/hardware/common/component.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/hardware/network/topology/routing/router.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/switches/entryexit.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/vis/vis.h>

namespace sstmac
{
  namespace hw
  {

    class networkswitch : public component,
        public factory_type,
        public vis::vis_comp
    {
    protected:

      nodeaddress::ptr my_addr_;
      router::ptr router_;
      topology::ptr top_;

      static int packet_length_bytes_;

    public:

      typedef boost::intrusive_ptr<networkswitch> ptr;

      std::string
      to_string() const
      {
        return "network switch";
      }

      virtual
      ~networkswitch()
      {
      }

      static int
      packet_length_bytes() {
        return packet_length_bytes_;
      }

      topology::ptr
      get_topology() const
      {
        return top_;
      }

      nodeaddress::ptr
      get_addr() const
      {
        return my_addr_;
      }

      router::ptr
      get_router() const
      {
        return router_;
      }

      virtual void
      init_factory_params(const sim_parameters::ptr& params);

      virtual std::string
      get_interprocess_link_name() const = 0;

      virtual ptr
      clone(router::ptr route, nodeaddress::ptr addr) const = 0;

      void
      clone_into(const networkswitch::ptr& sw, const nodeaddress::ptr& me,
          const router::ptr& route) const
      {
        sw->init(top_);
        sw->init_addr(me, route);
      }

      coordinates
      get_coords() const;

      virtual void
      initialize()
      {
        //nothing to do by default
      }

      virtual std::vector<nodeaddress::ptr>
      get_connected_switches() const
      {
        sst_throw(ssterror, "networkswitch::get_connected_switches() in base class is not implemented (child classes must implement)");
      }

      virtual void
      init(const topology::ptr& top)
      {
        top_ = top;
      }

      virtual void
      init_addr(const nodeaddress::ptr& me, const router::ptr& route)
      {
        my_addr_ = me;
        router_ = route;
        router_->set_switch(ptr(this));
      }

      virtual int
      get_queue_length(int port) const {
        return 0;
      }
    };

    class discrete_network_switch : public networkswitch
    {
    public:
      typedef boost::intrusive_ptr<discrete_network_switch> ptr;
      typedef boost::intrusive_ptr<const discrete_network_switch> const_ptr;

    protected:
      double link_bw_;

      int nps_;

      discrete_network_switch()
      {
        if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> networkswitch | network");
          logger::register_logger(&dbg_);
        }
      }

      void
      init_members(const discrete_network_switch::const_ptr& sw)
      {
        this->link_bw_ = sw->link_bw_;
      }

      /**
       * have to make sure this is called after init_addr()
       */
      void
      init_nps()
      {
        if (!top_)
        {
          sst_throw(valueerror, "discrete_network_switch::init_nps - topology is null, call init first");
        }
        nps_ = top_->nps(my_addr_);
      }

      void
      clone_into(const ptr& sw, const nodeaddress::ptr &me,
          const router::ptr &route) const
      {
        networkswitch::clone_into(sw, me, route);
        sw->init_nps();
      }

      static debug dbg_;

      //      dim            dir    switch
      //typedef boost::unordered_map<int, eventhandler::ptr> outport_dir_map;
      //typedef boost::unordered_map<int, outport_dir_map> outport_dim_map;
      //outport_dim_map outports_;

      typedef boost::unordered_map<long, eventhandler::ptr> outport_map;
      outport_map outports_;

      //this map tracks which connections have multiple
      //  link going to the same place
      //nested_type<int, int, int>::unordered_map redund_;
      boost::unordered_map<long, int> redund_;

      //      dim            dir  bandwidth
      boost::unordered_map<long, double> portbw_;

      //      dim            dir    vis obj
      boost::unordered_map<long, vis::vis_obj::ptr> linkvis_;

      virtual injector::ptr
      get_new_injector(sim_parameters::ptr params) = 0;

      virtual ejector::ptr
      get_new_ejector(sim_parameters::ptr params) = 0;

    public:

      eventhandler::ptr
      get_outport_handler(long port) const
      {
        outport_map::const_iterator it = outports_.find(port);
        if (it == outports_.end())
          return eventhandler::ptr();

        return it->second;
      }

      virtual std::vector<nodeaddress::ptr>
      get_connected_switches() const;

      //returns same order as get_connected_switches()
      virtual std::vector<int>
      get_connected_redund() const;

      virtual void
      init_factory_params(const sim_parameters::ptr& params);

      virtual void
      connect(const nodeaddress::ptr&, int port, connection_type_t ty,
          const connectable::ptr& comp);

      virtual void
      connect(int dim, int port, connection_type_t ty,
          const connectable::ptr& comp, double weight, int red);

      virtual void
      connect(int dim, int dir, eventhandler::ptr sw, double weight,
          int redund = 1);

      virtual void
      finish()
      {
        outports_.clear();
      }

      virtual std::string
      to_string() const
      {
        return "discrete network switch";
      }

      virtual void
      set_link_vis(vis::vis_obj::ptr vo, int dim, int dir)
      {
        long port = top_->convert_to_outport(dim, dir);
        linkvis_[port] = vo;
      }

      virtual
      ~discrete_network_switch()
      {
      }

    };
    DeclareFactory1InitParam(networkswitch, topology::ptr)
    ;
  }
}

#endif
