/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/packet/packetswitch.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>
#include <sstmac/hardware/network/topology/routing/router.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("packet", networkswitch, packetswitch);

    stat_time_window_avg::ptr packetswitch::stat_Q_wait_ = NULL;
    stat_time_window_sum::ptr packetswitch::stat_bytes_ = NULL;
    stat_time_window_sum_duration::ptr packetswitch::stat_pwr_ = NULL;

    packetswitch::packetswitch()
    {
      redundRNG_ = RNG::SimpleCombo::construct(time(NULL));
      maxtime_ = timestamp(0);
    }

    networkswitch::ptr
    packetswitch::clone(router::ptr route, nodeaddress::ptr addr) const
    {
      if (!route)
      {
        sst_throw(valueerror, "packetswitch::clone - null router?");
      }
      if (!addr)
      {
        sst_throw(valueerror, "packetswitch::clone - null address?");
      }
      ptr ret = new packetswitch;
      clone_into(ret, addr, route);
      ret->init_members(const_ptr(this));
      return ret;
    }

    injector::ptr
    packetswitch::get_new_injector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetinjector::construct(params, self);
    }

    ejector::ptr
    packetswitch::get_new_ejector(sim_parameters::ptr params)
    {
      ptr self(this);
      return packetejector::construct(params, self, my_addr_);
    }

    packetswitch::~packetswitch()
    {
    }

    void
    packetswitch::init_factory_params(const sim_parameters::ptr& params)
    {
      discrete_network_switch::init_factory_params(params);

      if (!stat_Q_wait_)
        stat_Q_wait_ = stat_time_window_avg::construct("<stats> network",
            "stats_packet_Qwait.csv");

      if (!stat_bytes_)
        stat_bytes_ = stat_time_window_sum::construct("<stats> network",
            "stats_packet_bytes.csv");

      if (!stat_pwr_)
        stat_pwr_ = stat_time_window_sum_duration::construct("<stats> network",
            "stats_power_network.csv");

      lat_r2r_ = params->get_time_param("packetswitch_latency_r2r");
      bw_n2r_ = params->get_bandwidth_param("packetswitch_bandwidth_n2r");

    }

    void
    packetswitch::init_members(const packetswitch::const_ptr &sw)
    {
      discrete_network_switch::init_members(sw);
      lat_r2r_ = sw->lat_r2r_;
      bw_n2r_ = sw->bw_n2r_;
      init_eject();
    }

    void
    packetswitch::init_eject()
    {
      SSTMAC_DEBUG << "packetswitch(" << my_addr_->to_string()
            << ") - initializing " << nps_ << " ejectors \n";
      int nodes = top_->nps(my_addr_);
      for (int i = 0; i < nodes; i++)
      {
        int dir = i;
        long port = router_->convert_to_outport(topology::EJECT, dir);
        portbw_[port] = bw_n2r_;
        //make singly redundant
        finishtimes_[port] = packet_finish(1);
      }
    }

    void
    packetswitch::finish()
    {
      discrete_network_switch::finish();
      stat_Q_wait_ = 0;
      stat_pwr_ = 0;
      stat_bytes_ = 0;
    }

    void
    packetswitch::update_congestion_info(routing_info::ptr info)
    {

    #if 0
      congestion_routing_info::ptr cast = boost::dynamic_pointer_cast<
          congestion_routing_info>(info);
      if (!cast)
      {
        sst_throw(ssterror, "trainswitch::update_router_cong: routing_info didn't cast to a congestion_routing_info");
      }
      timestamp rightnow(now());
      outdim_to_finish_map::iterator it, end = finishtimes_.end();
      for (it = finishtimes_.begin(); it != end; it++)
      {
        int dim = it->first;
        outdir_to_finish_map::const_iterator it2, end2 = it->second.end();
        for (it2 = it->second.begin(); it2 != end2; it2++)
        {
          int dir = it2->first;
          int port = router_->convert_to_outport(dim, dir);
          double avg = 0;
          int num = 0;
          const packet_finish& fin = it2->second;
          int nredund = fin.times.size();
          for (int r = 0; r < nredund; ++r)
          {
            const timestamp& time = fin.times[r];
            if (time.sec() - rightnow.sec() >= 0)
            {
              avg += time.sec() - rightnow.sec();
              num++;
            }
          }
          if (num > 0)
          {
            avg /= (double) num;

            cast->set_congestion_info(port, avg);

          }
          else
          {
            cast->set_congestion_info(port, 0);
          }

        }

      }
      #endif
    }

    void
    packetswitch::handle(const sst_message::ptr &msg)
    {

      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
      {
        throw ssterror("packetswitch::handle() - can't cast to packetmessage");
      }

      int dim;
      int dir;

      //if (router_->needs_congestion_info())
      //{
      //  update_congestion_info(pack->rinfo());
      //}
      //router_->route(pack->rinfo(), dim, dir);
      long port = router_->route(pack);
      if (redund_.find(port) == redund_.end())
        redund_[port] = 1;
      else if (redund_[port] < 1)
        redund_[port] = 1;

      //if(portbw_[dim].find(dir) == portbw_[dim].end()){
      //  sst_throw(valueerror, "packetswitch::handle: can't find dir %d in dim %d sending from switch %ld to node %ld",
      //            dir, dim, get_addr()->unique_id(), pack->toaddr()->unique_id());
      //}

      //SSTMAC_DEBUG << "port: " << dim << ", dir: " << dir << ", redund: "
      //      << redund_[dim][dir] << ", portbw: " << portbw_[dim][dir] << "\n";

      //int r = rand() % redund_[dim][dir];
      packet_finish& finish = finishtimes_[port];
      int r = finish.round_robin_index;
      //update the round robin index for the next packet
      finish.increment_index();
      timestamp txFinishTime = finish.times[r];

      if (txFinishTime < now())
        txFinishTime = now();
      timestamp delay = lat_r2r_ + timestamp(
          pack->get_byte_length() / portbw_[port]);

      SSTMAC_DEBUG << "packetswitch[" << my_addr_->to_string() << "]: message going to " << pack->toaddr()->unique_id() << " which is dim " << dim << " and dir " << dir
            << " which is switch " << outports_[port]->to_string()
            << " with delay " << delay << "\n";
      schedule(txFinishTime + delay, outports_[port], msg);

      finish.times[r] = txFinishTime + delay;

      if (finish.times[r] - now() > maxtime_)
      {
        maxtime_ = finish.times[r] - now();
      }

      double wait = (txFinishTime.sec() - now().sec());
      stat_Q_wait_->record(txFinishTime, wait);

      stat_bytes_->record(txFinishTime, pack->get_byte_length());

    }

    void
    packetswitch::packetinjector::finish()
    {
      parent_ = 0;
      injector::finish();
    }

    void
    packetswitch::packetinjector::send(long nbytes, long byte_offset,
        const sst_message::ptr& msg)
    {
      timestamp delay(0);
      long bytes_sent = byte_offset;
      long bytes_total = msg->get_byte_length();
      bool tail_packet = false; //assume false for now
      if (MTU_ > 0)
      {
        SSTMAC_DEBUG << "packetinjector: injecting packets: "
              << msg->to_string() << ", bytes-to-send=" << nbytes
              << ", bytes-offset=" << byte_offset << ", total-msg-size="
              << msg->get_byte_length() << ", MTU=" << MTU_ << "\n";

        long bytes_left = std::min(nbytes, bytes_total - byte_offset);
        while (bytes_left > 0)
        {
          long bytes_to_send = std::min(MTU_, bytes_left);
          bytes_sent += bytes_to_send;
          bytes_left -= bytes_to_send;
          tail_packet = bytes_sent >= bytes_total;
          packetmessage::ptr pack = packetmessage::construct(msg,
              bytes_to_send, tail_packet);
          parent_->schedule(parent_->now() + delay, parent_, pack);
          delay = delay + timestamp(bytes_to_send / n2r_);
        }
      }
      else
      {
        delay = timestamp(nbytes / n2r_);
        SSTMAC_DEBUG << "packetinjector: injecting whole message: "
              << msg->to_string() << " with delay " << delay
              << ", bytes-to-send=" << ", bytes-offset=" << byte_offset
              << ", total-msg-size=" << msg->get_byte_length() << ", bw="
              << n2r_ << "\n";

        bytes_sent = byte_offset + nbytes;
        tail_packet = bytes_sent >= bytes_total;
        parent_->schedule(parent_->now() + delay, parent_,
            packetmessage::construct(msg, nbytes, tail_packet));
      }

      if (msg->get_needs_ack() && tail_packet)
      {
        sst_message::ptr sendack = msg->clone(sst_message::SEND_ACK);
        parent_->schedule(parent_->now() + delay,
            get_sendhandler(msg->fromaddr()), sendack);
      }
    }

    int
    packetswitch::get_queue_length(int port) const
    {
        boost::unordered_map<long,packet_finish>::const_iterator
            it = finishtimes_.find(port);
        if (it == finishtimes_.end()){
            sst_throw(valueerror,
                "packetswith::get_queue_length: invalid port %d",
                port);
        }
        const packet_finish& fin = it->second;
        int min_length = 100000000;
        const timestamp& now_ = now();
        for (int r=0; r < fin.times.size(); ++r)
        {
            int queue_length = 0;
            if (fin.times[r] > now_){
                timestamp delta_t = fin.times[r] - now_;
                long bytes_left = delta_t.sec()*link_bw_;
                queue_length = bytes_left / networkswitch::packet_length_bytes();
            }
            min_length = std::min(queue_length, min_length);
        }
        return min_length;
    }

    void
    packetswitch::packetejector::finish()
    {
      parent_ = 0;
      ejector::finish();
    }

    void
    packetswitch::connect(int dim, int dir, eventhandler::ptr sw,
        double bw, int redund)
    {
      SSTMAC_DEBUG << "packetswitch::connect dim:" << dim << " dir:" << dir
            << "\n";
      discrete_network_switch::connect(dim, dir, sw, bw, redund);
      long port = router_->convert_to_outport(dim, dir);
      finishtimes_[port] = packet_finish(redund);
    }

    void
    packetswitch::packetejector::handle(const sst_message::ptr &msg)
    {
      packetmessage::ptr pack = boost::dynamic_pointer_cast<packetmessage>(msg);

      if (!pack)
      {
        throw ssterror(
            "packetswitch::packetejector::handle() - can't cast to packetmessage");
      }

      /** send all packets to the nic now */
      bool send_to_nic = true; //pack->is_tail()

      if (send_to_nic)
      {
        if (recvhandles_.find(pack->orig()->toaddr()) == recvhandles_.end())
        {
          std::cerr << "my addr: " << addr_->to_string() << "\n";
          std::cerr << "message headed to: "
              << pack->orig()->toaddr()->to_string() << "\n";
          std::cerr << "message came from: "
              << pack->orig()->fromaddr()->to_string() << "\n";
          throw ssterror(
              "packetswitch::packetejector::handle() -- can't find recv handler, that's a problem");
        }
        SSTMAC_DEBUG << "packetejector: ejecting message to "
              << pack->orig()->toaddr()->to_string() << ": "
              << msg->to_string() << "\n";
        parent_->schedule(parent_->now(), recvhandles_[pack->orig()->toaddr()],
            pack);
      }
    }

  }
}
