/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_SWITCHES_PACKETSWITCH_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_SWITCHES_PACKETSWITCH_H_INCLUDED

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/network/switches/packet/packetmessage.h>
#include <sstmac/common/stats/stat_time_window.h>
#include <sstmac/common/units.h>
#include <sstmac/common/rng.h>

namespace sstmac
{
  namespace hw
  {

    class packetswitch : public discrete_network_switch
    {
    public:
        typedef boost::intrusive_ptr<packetswitch> ptr;
        typedef boost::intrusive_ptr<const packetswitch> const_ptr;

    protected:
      //outdim_to_finish_map finishtimes_;
      boost::unordered_map<long, packet_finish> finishtimes_;

      timestamp maxtime_;

      timestamp lat_r2r_;

      double bw_n2r_;

      static sstmac::stat_time_window_avg::ptr stat_Q_wait_;
      static sstmac::stat_time_window_sum_duration::ptr stat_pwr_;
      static sstmac::stat_time_window_sum::ptr stat_bytes_;

      RNG::UniformInteger::ptr redundRNG_;

      //friend class packetinjector;
      // friend class packetejector;

      packetswitch();

      virtual void finish();

      void init_members(const packetswitch::const_ptr& sw);

      void init_eject();

      void
      update_congestion_info(routing_info::ptr info);

      virtual injector::ptr
      get_new_injector(sim_parameters::ptr params);

      virtual ejector::ptr
      get_new_ejector(sim_parameters::ptr params);

      class packetinjector : public injector
      {

      protected:
        packetinjector(sim_parameters::ptr params, packetswitch::ptr par)
        {
          MTU_ = params->get_long_param("packetswitch_MTU");

          n2r_ = params->get_bandwidth_param("packetswitch_bandwidth_n2r");

          parent_ = par;
          dbg_ = parent_->dbg_;
        }

        long MTU_;
        double n2r_;
        debug dbg_;

        packetswitch::ptr parent_;

      public:
        typedef boost::intrusive_ptr<packetinjector> ptr;

        virtual std::string
        to_string() const
        {
          return "packetinjector";
        }

        virtual
        ~packetinjector()
        {
        }

        static ptr
        construct(sim_parameters::ptr params, packetswitch::ptr par)
        {
          return ptr(new packetinjector(params, par));
        }

        void send(long nbytes, long bytes_offset, const sst_message::ptr &msg);

        virtual void finish();

      };

      class packetejector : public ejector
      {

      protected:
        packetejector(sim_parameters::ptr params, packetswitch::ptr par,
            nodeaddress::ptr addr)
        {
          parent_ = par;
          dbg_ = parent_->dbg_;
          addr_ = addr;
        }

        packetswitch::ptr parent_;
        debug dbg_;
        nodeaddress::ptr addr_;

      public:
        typedef boost::intrusive_ptr<packetejector> ptr;

        virtual std::string
        to_string() const
        {
          return "packetejector";
        }

        virtual
        ~packetejector()
        {
        }

        static ptr
        construct(sim_parameters::ptr params, packetswitch::ptr par,
            nodeaddress::ptr addr)
        {
          return ptr(new packetejector(params, par, addr));
        }

        virtual void
        handle(const sst_message::ptr &msg);

        virtual void finish();

      };

    public:

      static ptr
      construct(const sim_parameters::ptr& params)
      {
          return new packetswitch;
      }

      virtual std::string
      to_string() const
      {
        return "packetswitch(" + my_addr_->to_string() + ")";
      }

      virtual ~packetswitch();

      virtual void
      connect(int dim, int dir, eventhandler::ptr sw, double bw, int redund);

      virtual void
      init_factory_params(const sim_parameters::ptr& params);

      virtual networkswitch::ptr
      clone(router::ptr route, nodeaddress::ptr addr) const;

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      get_interprocess_link_name() const
      {
        return "packetswitch_latency_r2r";
      }

      virtual int
      get_queue_length(int port) const;
  
    };

  }
}
#endif
