/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// butterfly.cc: Implementation of butterfly networks.
//
// Based on butterfly.c by Jim Schutt
// Adapted by Curtis Janssen <cljanss@ca.sandia.gov>

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>
#include <sstmac/hardware/network/topology/butterfly.h>
#include <sstmac/common/driver_util.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("butterfly | bfly", topology, butterfly);

    const int butterfly::up_dimension = 0;

    abstract_butterfly::abstract_butterfly(int kary, int nfly, int nps) :
        regular_topology(nps),
        kary_(kary), nfly_(nfly)
    {
        //a 4-ary 3-fly has three levels of router
        //assume for simplicity nps = kary
        //we have 4^3 = 64 nodes
        //we need 4^(3-1) = 16 switches per level to support 64 nodes
        nswitches_per_col_ = pow(kary, nfly - 1);
    }

    void
    abstract_butterfly::compute_switch_coords(long uid, coordinates& coords) const
    {
        long group_size = nswitches_per_col_;
        long next_group_size = group_size / kary_;
        for (int l=0; l < (nfly_ - 1); ++l){
            //figure out the group offset
            long group_relative_row = uid % group_size;
            long group_number = group_relative_row / next_group_size;
            coords[l] = group_number;

            group_size = next_group_size;
            next_group_size /= kary_;
        }
    }

    long
    abstract_butterfly::get_switch_id(const coordinates& coords) const
    {
        long index_multiplier = nswitches_per_col_ / kary_;
        long nid = 0;
        for (int l=0; l < (nfly_ - 1); ++l){
            nid += coords[l] * index_multiplier;
            index_multiplier /= kary_;
        }
        return nid;
    }

    void
    butterfly::compute_switch_coords(long uid, coordinates& coords) const
    {
        abstract_butterfly::compute_switch_coords(uid, coords);

        //figure out which column
        int col = uid / nswitches_per_col_;
        coords[nfly_-1] = col;
    }

    butterfly::butterfly(int kary, int nfly, int nps) :
        abstract_butterfly(kary, nfly, nps)
    {
        //a 4-ary 3-fly has three levels of router
        //assume for simplicity nps = kary
        //we have 4^3 = 64 nodes
        //we need 4^(3-1) = 16 switches per level to support 64 nodes
        last_col_index_start_ = nswitches_per_col_ * (nfly - 1);
    }

    butterfly::ptr
    butterfly::construct(sim_parameters::ptr params)
    {
      /**
       sstkeyword {
       gui=4 3;
       docstring=Specify the geometry of the butterfly network.ENDL
       Should be a vector of size 2 for K-ary N-fly.ENDL
       First index, K, is the radix of the butterfly.ENDL
       Second index, N, is the number of stages.ENDL
       Network will have K^(N-1) leaf switches and K^N network endpoints.;
       }
       */
      std::vector<int> args; params->get_vector_param("topology_geometry", args);
      int kary = args[0];
      int nfly = args[1];
      /**
       sstkeyword {
       gui=4;
       docstring=Specify the number of compute nodes per leaf switch.ENDL
       In most cases, for K-ary N-fly, this should just match K.;
       }
       */
      int nps = params->get_optional_int_param("network_nodes_per_switch", kary);
      return new butterfly(kary, nfly, nps);
    }

    int
    butterfly::get_productive_dir(
        int& dim,
        const coordinates &src,
        const coordinates &dst) const
    {
    #if SST_SANITY_CHECK
        for (int i=0; i < src.size(); ++i){
            if (src[i] != dst[i]){
                if (i != dim){
                    sst_throw(illformederror,
                        "butterfly::get_productive_dir: requested dim %d, but dim %d differs."
                        " butterfly must proceed in dimension-order.");
                }
                return dst[i];
            }
        }
    #else
        return dst[dim];
    #endif
    }

    int
    butterfly::convert_to_outport(int dim, int dir)
    {
      if (dim == topology::EJECT)
        {
          return -(1+dir);
        }
      return dir;
    }



    void
    butterfly::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = topology::EJECT;
          dir = (-outport) - 1;
        }
      else
        {
          dim = butterfly::up_dimension;
          dir = outport;
        }
    }

    void
    butterfly::connect_switches(topology::maptype &switches)
    {
      double link_weight = 1.0;
      int red = 1;

      /**
        In 4-ary 3-fly, we have 16 switches per col
        with 3 columns or stages.  Thus we can label
        each switch by its row,col index
        Lets label switch 0 in each col A0, B0, C0...
        B0 = 16, B1 = 17 when using absolute indexing
        C0 = 32, C1 = 33 when using absolute indexing
        The first two cols connect in strides of 4
        A0 -> B0, B4, B8, B12 = 16, 20, 24, 28
        A1 -> B1, B5, B9, B13 = 17, 21, 25, 29
        ...
        A4 -> B0, B4, B8, B12
        etc
        The second and third cols connect in groups of 4
        with stride 1
        B0 -> C0, C1, C2, C3 = 32, 33, 34, 35
        B1 -> C0, C1, C2, C3
        B2 -> C0, C1, C2, C3
        B3 -> C0, C1, C2, C3
        B4 -> C4, C5, C6, C7 = 36, 37, 38, 39
        etc

        We distinguish between blocks and groups.
        A block is set of switches whose links "cross" and
        cannot be separated.  A group is a set of switches
        who connect to the same partners.
        On dimension 0, connecting A to B,
        we have one block of size 16 and 4 groups of size 4.
        On dimension1, connecting B to C,
        we now have four blocks of size 4 and still
        4 groups of size 4.
      */

      long connection_stride = nswitches_per_col_ / kary_;
      long block_size = nswitches_per_col_;
      long group_size = kary_;
      for (int l=0; l < (nfly_-1); ++l)
      {
        long col_start_index = l * nswitches_per_col_;
        for (long i=0; i < nswitches_per_col_; ++i)
        {
            long my_switch_index = col_start_index + i;
            nodeaddress::ptr my_addr = nodeaddress::convert_from_uniqueid(my_switch_index);
            networkswitch::ptr my_sw = switches[my_addr];

            long my_block = i / block_size;
            long my_intra_block_index = i % block_size;
            long my_block_index_start = my_block * block_size;

            long my_group_offset = my_intra_block_index % connection_stride;
            long my_group_index_start = my_block_index_start + my_group_offset;

            //make sure to include column offset
            long up_group_partner = my_group_index_start + col_start_index + nswitches_per_col_;
            long num_connections = kary_;

            for (long c=0; c < num_connections; ++c, up_group_partner += connection_stride)
            {
                //printf("Connecting %ld:%ld->%ld\n", my_switch_index, c, up_group_partner);
                nodeaddress::ptr up_group_partner_addr = nodeaddress::convert_from_uniqueid(up_group_partner);
                networkswitch::ptr partner_sw = switches[up_group_partner_addr];
                my_sw->connect(up_dimension, c,
                               connectable::network_link,
                               partner_sw,
                               link_weight, red);
            }
        }

        //the next set of connections is more compact - lower the stride
        connection_stride /= kary_;
        block_size /= kary_;
      }

    }

    void
    butterfly::minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const
    {
        //we have to route our current level
        int current_dim = current_coords[nfly_ - 1];
        dim = up_dimension;
        dir = dest_coords[current_dim];
    }

    int
    butterfly::minimal_distance(const coordinates &current_coords, const coordinates &dest_coords) const
    {
        int eject_dim = nfly_ - 1;
        int lastidx = eject_dim;
        int current_dim = current_coords[lastidx];
        int dest_dim = dest_coords[lastidx];
        if (dest_dim != eject_dim){
            sst_throw(valueerror,
                "invalid butterfly destination coordinates");
        }
        //this many hops remaining
        return (eject_dim - current_dim);
    }

    nodeaddress::ptr
    butterfly::node_to_ejector_addr(const nodeaddress::ptr& addr, int& switch_port) const
    {
          long node_idx = addr->unique_id();
          //we inject on the first row - eject on the last row
          long ej_idx = node_idx / nps_ + last_col_index_start_;
          switch_port = node_idx % nps_;
          return nodeaddress::convert_from_uniqueid(ej_idx);
    }

    long
    butterfly::get_switch_id(const coordinates &coords) const
    {
        //get the row number
        long row_id = abstract_butterfly::get_switch_id(coords);
        long col_id = coords[nfly_-1];
        long nid = col_id * nswitches_per_col_ + row_id;
        return nid;
    }

  }
} //end of namespace sstmac
