/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
#include <sstmac/hardware/network/topology/dragonfly.h>
#include <math.h>
#include <sstream>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    using namespace std;
    SSTRegister("dragonfly | dfly", topology, dragonfly);

    const int dragonfly::x_dimension = 0;
    const int dragonfly::y_dimension = 1;
    const int dragonfly::g_dimension = 2;

    dragonfly::dragonfly(long x, long y, long g,
                         int gc, int nps,
                         double opt_link_weight) :
                         regular_topology(nps), x_(x), y_(y), g_(g), group_con_(gc),
      opt_link_weight_(opt_link_weight),
      outputgraph_(false)
    {

    }

    dragonfly::ptr
    dragonfly::construct(sim_parameters::ptr params)
    {
      std::vector<int> args; params->get_vector_param("topology_geometry", args);
      double opt_link_weight = 1.0;
      if (params->has_param("optical_link_weight")){
        opt_link_weight = params->get_double_param("optical_link_weight");
      }
      else if (params->has_param("optical_link_bandwidth")){
        double opt_linkbw = params->get_bandwidth_param("optical_link_bandwidth");
        double net_linkbw = params->reread_bandwidth_param("network_bandwidth_link");
        opt_link_weight = opt_linkbw / net_linkbw;
      }
      else if (params->has_param("network_train_optical_freq")){
           double opt_link_freq = params->reread_double_param("network_train_optical_freq");
           double reg_link_freq = params->reread_double_param("network_train_link_freq");
           opt_link_weight = opt_link_freq / reg_link_freq;
      }
      else {
        sst_throw(inputerror,
            "need either optical_link_bandwidth "
            "or optical_link_weight for dragonfly topology");
      }
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);
      int group_con = params->get_int_param("topology_group_connections");
      int nx = args[0];
      int ny = args[1];
      int ng = args[2];
      return new dragonfly(nx, ny, ng,
                           group_con, nps,
                           opt_link_weight);
    }

    int
    dragonfly::get_productive_dir(
        int& dim,
        const coordinates &src,
        const coordinates &dst) const
    {
        if (dim == x_dimension){
            return minimal_route_to_X(dst[x_dimension]);
        }
        else if (dim == y_dimension){
            return minimal_route_to_Y(dst[y_dimension]);
        }
        else if (dim == g_dimension){
            int myX = src[x_dimension];
            int myY = src[y_dimension];
            int hisg = dst[g_dimension];
            int dir;
            minimal_route_to_group(myX, myY, hisg, dim, dir);
            return dir;
        }
        else {
            sst_throw(valueerror,
                "invalid dragonfly dimension %d given",
                dim);
        }
        return -1;
    }

    void
    dragonfly::get_coords(long uid, int &x, int &y, int &g) const
    {
      x = uid % x_;
      y = (uid / x_) % y_;
      g = uid / (x_ * y_);
    }

    void
    dragonfly::compute_switch_coords(long uid, coordinates& coords) const
    {
        get_coords(uid, coords[0], coords[1], coords[2]);
    }

    int
    dragonfly::convert_to_outport(int dim, int dir)
    {
      if (dim == topology::EJECT)
      {
        return -(1 + dir);
      }
      if (dim == 0)
      {
        return dir;
      }
      else if (dim == 1)
      {
        return x_ + dir;
      }
      else
        return x_ + y_ + dir;
    }

    void
    dragonfly::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
      {
        dim = topology::EJECT;
        dir = (-outport) - 1;
      }
      else
      {
        if (outport < x_)
        {
          dim = 0;
          dir = outport;
        }
        else if (outport >= x_ && outport < x_ + y_)
        {
          dim = 1;
          dir = outport - x_;
        }
        else
        {
          dim = 2;
          dir = outport - (x_ + y_);
        }

      }
    }

    std::string
    dragonfly::coord_string(int x, int y, int g) const
    {
      std::stringstream ss;
      ss << "(" << x << ", " << y << ", " << g << ")";

      return ss.str();
    }

    long
    dragonfly::get_uid(int x, int y, int g) const
    {
      return g * (x_ * y_) + y * (x_) + x;
    }

    long
    dragonfly::get_switch_id(const coordinates &coords) const
    {
        return get_uid(coords[0], coords[1], coords[2]);
    }

    nodeaddress::ptr
    dragonfly::get_random_intermediate_switch(const nodeaddress::ptr &current_sw, const sst_message::ptr &msg)
    {
        long curid = current_sw->unique_id();
        long nid = curid;
        while (curid == nid)
        {
            int srcX, srcY, srcG;
            get_coords(current_sw->unique_id(), srcX, srcY, srcG);

            //randomly select a group
            int hisG = rng_->value_in_range(g_);

            //now figure out which x,y,g fills the path
            int hisX, hisY;
            find_path_to_group(srcX, srcY, hisG, hisX, hisY);
            nid = get_uid(hisX, hisY, hisG);
        }
        return nodeaddress::convert_from_uniqueid(nid);
    }

    bool
    dragonfly::xy_connected_to_group(int dstx, int dsty, int dstg) const
    {
      int gspace = std::max(1, g_ / group_con_);
      int myid = dstx + dsty * x_;
      int gset = myid % g_;

      for (int g = 0; g < group_con_; g++)
      {
        int gid = (gset + gspace * g) % g_;
        if (gid == dstg)
            return true;
      }
      return false;
    }

    int
    dragonfly::find_y_path_to_group(int myX, int dstg) const
    {
      int ystart = rng_->value_in_range(y_);
      for (int yy = 0; yy < y_; ++yy)
      {
        int dsty = (ystart + yy) % y_;
        if (xy_connected_to_group(myX, dsty, dstg))
            return dsty;
      }
      return -1;
    }

    int
    dragonfly::find_x_path_to_group(int myY, int dstg) const
    {
        int xstart = rng_->value_in_range(x_);
        for (int xx = 0; xx < x_; ++xx)
        {
          int dstx = (xstart + xx) % x_;
          if (xy_connected_to_group(dstx, myY, dstg))
            return dstx;
        }
        return -1;
    }

    void
    dragonfly::find_path_to_group(
        int myX, int myY, int dstg,
        int& dstx, int& dsty
    ) const
    {
        //see if we can go directly to the group
        if (xy_connected_to_group(myX, myY, dstg)){
            dstx = myX;
            dsty = myY;
            return;
        }

        int x_or_y = rng_->value_in_range(2);
        //try to find x,y path first
        int x_first = 0;
        if (x_or_y == x_first){
            dstx = find_x_path_to_group(myY, dstg);
            if (dstx >= 0){
                dsty = myY;
                return;
            }
        }
        else {
            dsty = find_y_path_to_group(myX, dstg);
            if (dsty >= 0){
                dstx = myX;
                return;
            }
        }

        //both x and y need to change
        int xstart = rng_->value_in_range(x_);
        for (int xx = 0; xx < x_; ++xx)
        {
          dstx = (xstart + xx) % x_;
          dsty = find_y_path_to_group(dstx, dstg);
          if (dsty >= 0)
            return; //path found
        }
    }

    void
    dragonfly::minimal_route_to_group(
        int myX, int myY, int hisg,
        int& dim, int& dir
    ) const
    {
      int dsty, dstx;
      find_path_to_group(myX, myY, hisg, dstx, dsty);

      if (dstx != myX){
        dim = dragonfly::x_dimension;
        dir = dstx;
      }
      else if (dsty != myY){
        dim = dragonfly::y_dimension;
        dir = dsty;
      }
      else {
        dim = dragonfly::g_dimension;
        dir = hisg;
      }
    }

    int
    dragonfly::minimal_route_to_X(int hisx) const
    {
        return hisx;
    }

    int
    dragonfly::minimal_route_to_Y(int hisy) const
    {
        return hisy;
    }

    void
    dragonfly::minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const
    {
      int myX = current_coords[0];
      int myY = current_coords[1];
      int myG = current_coords[2];
      int hisx = dest_coords[0];
      int hisy = dest_coords[1];
      int hisg = dest_coords[2];

      if (hisg != myG)
      { //route to the right group first
        minimal_route_to_group(myX, myY, hisg, dim, dir);
      }
      else if (hisx != myX)
      {
        dim = dragonfly::x_dimension;
        dir = minimal_route_to_X(hisx);
        SSTMAC_DEBUG << "in group, headed to x pos " << hisx << "\n";
      }
      else
      {
        dim = dragonfly::y_dimension;
        dir = minimal_route_to_Y(hisy);
        SSTMAC_DEBUG << "in group, headed to y pos " << hisy << "\n";
      }
    }

    int
    dragonfly::minimal_distance(const coordinates &current_coords, const coordinates &dest_coords) const
    {
        int dstx, dsty;
        if (current_coords[2] == dest_coords[2]){
            int path_length = 0;
            if (dest_coords[0] != current_coords[0]) ++path_length;
            if (dest_coords[1] != current_coords[1]) ++path_length;
            //same group, all that matters is x,y difference
            return path_length;
        }

        find_path_to_group(
            current_coords[0],
            current_coords[1],
            dest_coords[2],
            dstx, dsty);

        int path_length = 1; //group jump
        //we might need x,y jump before group jump
        if (dstx != current_coords[0]) ++path_length;
        if (dsty != current_coords[1]) ++path_length;
        //we might need x,y jump after group jump
        if (dstx != dest_coords[0]) ++path_length;
        if (dsty != dest_coords[1]) ++path_length;
        return path_length;
    }

    /// coordinates go x,y - group
    void
    dragonfly::connect_switches(topology::maptype &switches)
    {
      double intra_group_weight = 1.0;
      for (long i = 0; i < switches.size(); i++)
      {
        std::vector<int> connected;
        std::vector<int> connected_red;

        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);
        int myx;
        int myy;
        int myg;
        get_coords(i, myx, myy, myg);

        topology::maptype::iterator itme = switches.find(me);

        if (itme == switches.end())
        {
          sst_throw(
              ssterror,
              "dragon::connect_switches(): can't find switch when connecting:   uid=%d  addr=%s",
              get_uid(myx, myy, myg), me->to_string().c_str());
        }

        bool errorflag = false;



        for (int x = 0; x < x_; x++)
        {
          if (x != myx)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(x, myy, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(x, myy, myg) << "\n";

            topology::maptype::iterator it1 = switches.find(them);

            if (it1 == switches.end())
            {
              sst_throw(
                  ssterror,
                  "dragon::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
                  get_uid(x, myy, myg), them->to_string().c_str());
            }

            switches[me]->connect(x_dimension, x,
                                  connectable::network_link,
                                  switches[them],
                                  intra_group_weight, red_[0]);

            if (outputgraph_)
            {
              connected.push_back(them->unique_id());
              connected_red.push_back(red_[0]);
            }
          }
        }

        for (int y = 0; y < y_; y++)
        {
          if (y != myy)
          {

            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(
                get_uid(myx, y, myg));
            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, y, myg) << "\n";

            switches[me]->connect(y_dimension, y,
                                  connectable::network_link,
                                  switches[them],
                                  intra_group_weight, red_[1]);

            if (outputgraph_)
            {
              connected.push_back(them->unique_id());
              connected_red.push_back(red_[1]);
            }
          }
        }

        //each switch connects to a max of 5 different groups

        int gspace = max(1, g_ / group_con_);
        int myid = myx + myy * x_;
        int gset = myid % g_;
        for (int g = 0; g < group_con_; g++)
        {

          int gid = (gset + gspace * g) % g_;
          if (gid != myg)
          {
            long uid = get_uid(myx, myy, gid);
            nodeaddress::ptr them = nodeaddress::convert_from_uniqueid(uid);

            SSTMAC_DEBUG << "node " << me->to_string() << " " << coord_string(
                  myx, myy, myg) << " connecting to node " << them->to_string()
                  << " " << coord_string(myx, myy, gid) << "\n";
            switches[me]->connect(g_dimension, gid,
                                  connectable::network_link,
                                  switches[them],
                                  opt_link_weight_, red_[2]);

            if (outputgraph_)
            {
              connected.push_back(them->unique_id());
              connected_red.push_back(red_[2]);
            }
          }

        }

        if (outputgraph_)
        {
          for (int k = 0; k < nps_; k++)
          {
            std::cout << (i * nps_ + k);
            for (int l = 0; l < connected.size(); l++)
            {
              for (int m = 0; m < nps_; m++)
              {
                std::cout << " " << (connected[l] * nps_ + m) << ":"
                    << connected_red[l];

              }
            }

            for (int m = 0; m < nps_; m++)
            {
              if (m != k)
              {
                std::cout << " " << (i * nps_ + m) << ":" << 100;
              }
            }
            std::cout << "\n";
          }

        }

      }

    }

    static std::string
    get_link_name(const std::vector<int>& src, const std::vector<int>& dst)
    {
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) src.size(); ++i)
      {
        sstr << sstprintf(" %2d", src[i]);
      }
      sstr << " ]->[";
      for (int i = 0; i < (int) dst.size(); ++i)
      {
        sstr << sstprintf(" %2d", dst[i]);
      }
      sstr << "]";
      return sstr.str();
    }

    std::string
    dragonfly::label(long id) const
    {
      int x, y, g;

      get_coords(id, x, y, g);
      std::stringstream sstr;
      sstr << "[";
      sstr << sstprintf(" %2d", x);
      sstr << sstprintf(" %2d", y);
      sstr << sstprintf(" %2d", g);

      sstr << " ]";
      return sstr.str();
    }

    void
    dragonfly::get_nearest_neighbor_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        //+X,-X,+Y,-Z and group connections
        partners.clear();
        partners.resize(4 + group_con_);
        int idx = 0; long nid;
        coordinates tmp_coords(src_sw_coords);

        //get the plus,minus 1 in the x-direction
        tmp_coords[x_dimension] = (src_sw_coords[x_dimension] + 1) % x_;
        nid = get_node_id(tmp_coords, port);
        partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

        tmp_coords[x_dimension] = (src_sw_coords[x_dimension] -1 + x_) % x_;
        nid = get_node_id(tmp_coords, port);
        partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

        //reset coords
        tmp_coords[x_dimension] = src_sw_coords[x_dimension];

        tmp_coords[y_dimension] = (src_sw_coords[y_dimension] + 1) % y_;
        nid = get_node_id(tmp_coords, port);
        partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

        tmp_coords[y_dimension] = (src_sw_coords[y_dimension] - 1 + y_) % y_;
        nid = get_node_id(tmp_coords, port);
        partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

        //reset coords
        tmp_coords[y_dimension] = src_sw_coords[y_dimension];

        //now we have to figure out our group connections
        int gspace = std::max(1, g_ / group_con_);
        int myX = src_sw_coords[x_dimension];
        int myY = src_sw_coords[y_dimension];
        int myG = src_sw_coords[g_dimension];
        int myid = myX + myY * x_;
        int gset = myid % g_;

        for (int g = 0; g < group_con_; g++)
        {
            int gid = (gset + gspace * g) % g_;
            if (gid == myG)
                continue;

            tmp_coords[g_dimension] = gid;
            long nid = get_node_id(tmp_coords, port);
            partners[idx++] = nodeaddress::convert_from_uniqueid(nid);
        }
        partners.resize(idx);
    }

    void
    dragonfly::get_tornado_recv_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        tmp_coords[x_dimension] = (src_sw_coords[x_dimension] - (x_ - 1)/2 + x_) % x_;
        tmp_coords[y_dimension] = (src_sw_coords[y_dimension] - (y_ - 1)/2 + y_) % y_;
        tmp_coords[g_dimension] = (src_sw_coords[g_dimension] - (g_ - 1)/2 + g_) % g_;

        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }

    void
    dragonfly::get_tornado_send_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        tmp_coords[x_dimension] = (src_sw_coords[x_dimension] + (x_ - 1)/2) % x_;
        tmp_coords[y_dimension] = (src_sw_coords[y_dimension] + (y_ - 1)/2) % y_;
        tmp_coords[g_dimension] = (src_sw_coords[g_dimension] + (g_ - 1)/2) % g_;

        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }

    void
    dragonfly::get_bit_complement_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        tmp_coords[x_dimension] = x_ - src_sw_coords[x_dimension] - 1;
        tmp_coords[y_dimension] = y_ - src_sw_coords[y_dimension] - 1;
        tmp_coords[g_dimension] = g_ - src_sw_coords[g_dimension] - 1;
        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }


  }
} //end of namespace sstmac


