/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_DRAGONFLY_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_DRAGONFLY_H_INCLUDED

#include <sstmac/hardware/network/topology/regular_topology.h>

namespace sstmac
{
  namespace hw
  {

    class dragonfly
        : public regular_topology
    {

    protected:
      dragonfly(long x, long y, long g,
                int gc, int nps,
                double opt_link_weight);

      int x_;
      int y_;
      int g_;
      int group_con_;

      bool outputgraph_;
      double opt_link_weight_;

      void
      minimal_route_to_group(int myX, int myY, int hisg,
                             int& dim, int& dir) const;

      void
      find_path_to_group(
        int myX, int myY, int dstg,
        int& dsty, int& dstx) const;

      int
      minimal_route_to_X(int hisx) const;

      int
      minimal_route_to_Y(int hisy) const;

      virtual void
      compute_switch_coords(long uid, coordinates& coords) const;

      int find_y_path_to_group(int myX, int dstg) const;

      int find_x_path_to_group(int myY, int dstg) const;

      bool xy_connected_to_group(int dstx, int dsty, int dstg) const;

    public:
      static const int x_dimension;
      static const int y_dimension;
      static const int g_dimension;

      typedef boost::intrusive_ptr<dragonfly> ptr;

      static ptr
      construct(sim_parameters::ptr params);

      virtual std::string
      to_string() const
      {
        return "dragonfly";
      }

      int ndimensions() const {
        return 3;
      }

      int
      numX() const
      {
        return x_;
      }

      int
      numY() const
      {
        return y_;
      }

      int
      numG() const
      {
        return g_;
      }

      int
      group_con() const
      {
        return group_con_;
      }

      void
      get_coords(long uid, int &x, int &y, int &g) const;

      long
      get_uid(int x, int y, int g) const;

      long
      get_switch_id(const coordinates &coords) const;

      virtual long
      num_switches() const
      {
        return x_ * y_ * g_;
      }

      long num_leaf_switches() const
      {
        return x_ * y_ * g_;
      }

      virtual std::string
      label(long id) const;

      virtual
      ~dragonfly()
      {
      }

      virtual int
      radix() const
      {
        return x_ + y_ + group_con_;
      }

      std::string
      coord_string(int x, int y, int g) const;

      virtual void
      connect_switches(topology::maptype &switches);

      virtual int64_t
      diameter() const
      {
        return -1;
      }

      virtual nodeaddress::ptr
      get_random_intermediate_switch(const nodeaddress::ptr& current_sw, const sst_message::ptr& msg);

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      int
      get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const;

      void
      minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const;

      int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const;

      void
      get_nearest_neighbor_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_bit_complement_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_tornado_send_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_tornado_recv_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

    };

  }
} //end of namespace sstmac

#endif
