/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */
// fattree.cc: Implementation of fat tree networks.
//
// Author: Curtis Janssen <cljanss@ca.sandia.gov>
#include <sstream>
#include <sstmac/hardware/network/topology/fattree.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("fattree | ftree", topology, fattree);

    const int fattree::up_dimension = 0;
    const int fattree::down_dimension = 1;

    fattree::ptr
    fattree::construct(sim_parameters::ptr params)
    {
      std::vector<int> args;
      params->get_vector_param("topology_geometry", args);
      if (args.size() < 2 || args.size() > 4){
        sst_throw(inputerror,
            "fattree::construct: topology_geometry: need 2, 3, or 4 parameters. got %d",
            args.size());
      }
      int num_levels = args[0];
      int num_branches = args[1];

      if (num_levels <= 2){
        sst_throw(inputerror, "fattree::construct: fat tree only valid with 3 or more levels");
      }
      if (num_branches % 2){
        sst_throw(inputerror, "fattree::construct: fat tree must have an even number of branches per level");
      }

      int nps = 0;
      if (args.size() >= 3){
        nps = args[2];
      }
      else {
        nps = params->get_optional_int_param("network_nodes_per_switch", num_branches);
      }

      int width = args.size() == 4 ? args[3] : 1;

      return new fattree(num_levels, num_branches, nps, width);
    }

    int
    fattree::get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const
    {
        sst_throw(
            illformederror,
            "fattree::get_productive_dir should never be called."
            "productive outports are determined differently from other topologies");
    }

    std::string
    fattree::label(long id) const
    {
        return sstprintf("fat tree %ld", id);
    }

    fattree::fattree(int l, int k, int nndpersw, int widtharg)
        : regular_topology(nndpersw)
    {
      init_fattree(l, k, nndpersw, widtharg);
    }

    void
    fattree::init_fattree(int l, int k, int nndpersw, int widtharg)
    {
      width = widtharg;
      l_ = l;
      k_ = k;
      toplevel_ = l_ - 1;
      numleafswitches_ = pow(k_, l_ - 1);
      toplevel_id_start_ = numleafswitches_ * (l_ - 1);
    }

    std::string
    fattree::name() const
    {
      std::ostringstream ostr;
      ostr << "FatTree(" << l_ << "," << k_ << ")";
      return ostr.str();
    }

    void
    fattree::connect_group(
        maptype& switches,
        int group_stride,
        int down_group_size,
        long down_group_offset,
        int up_group_size,
        long up_group_offset)
    {
        double link_weight = 1.0;
        int red = 1;
        long down_partner = down_group_offset;
        for (int d=0; d < down_group_size; ++d, down_partner += group_stride){
            nodeaddress::ptr down_addr = nodeaddress::convert_from_uniqueid(down_partner);
            networkswitch::ptr down_switch = switches[down_addr];

            long up_partner = up_group_offset;
            for (int u=0; u < up_group_size; ++u, up_partner += group_stride){
                // the top level is a little bit weird
                // we only have half the number of switches on the top level
                // the 1/2 number of switches sort of simulates a full level
                // because it only connects downward and has no up connections
                long actual_up_partner = up_partner;
                int port_offset = 0;
                if (up_partner >= toplevel_id_start_){
                    //in an l=3,k=2 tree for example
                    //switch 8 connects as if it were both switch 8 and 9
                    //switch 9 connects as if it were both switch 9 and 10
                    long level_id = up_partner - toplevel_id_start_;
                    actual_up_partner = (level_id / 2) + toplevel_id_start_;
                    port_offset = k_ * (level_id % 2);
                }
                  nodeaddress::ptr up_addr = nodeaddress::convert_from_uniqueid(actual_up_partner);
                  networkswitch::ptr up_switch = switches[up_addr];

                  SSTMAC_DEBUG << sstprintf("fattree:: connection %ld:%d <-> %ld:%d\n",
                                            down_partner, u,
                                            actual_up_partner, d + port_offset);

                  down_switch->connect(up_dimension, u,
                                       connectable::network_link,
                                       up_switch,
                                       link_weight, red); // up
                  up_switch->connect(down_dimension, d + port_offset,
                                     connectable::network_link,
                                     down_switch,
                                     link_weight, red); //down
            }
        }
    }

    void
    fattree::connect_section(
        maptype& switches,
        int group_stride,
        int num_groups_per_section,
        int down_group_size,
        long down_section_offset,
        int up_group_size,
        long up_section_offset
    )
    {
        for (int grp=0; grp < num_groups_per_section; ++grp)
        {
            long up_group_offset = up_section_offset + grp;
            long down_group_offset = down_section_offset + grp;
            connect_group(switches, group_stride,
                          down_group_size, down_group_offset,
                          up_group_size, up_group_offset);
        }
    }

    void
    fattree::connect_switches(topology::maptype &switches)
    {
      long level_size = numleafswitches_;
      int group_size = k_;
      int group_stride = 1;
      long down_level_start = 0;
      long up_level_start = numleafswitches_;

      /**
      Consider an l=4, k=2 tree
      On L=0, we have
      0 <-> 8 <-> 1
      0 <-> 9 <-> 1
      2 <-> 10 <-> 3
      2 <-> 11 <-> 3
      Etc...
      6 <-> 15 <-> 7
      This level is divided into 4 sections/4 groups
      {0,1,8,9} form one connection group/section
      {2,3,10,11} form another connection group/section, etc

      On L=1, we have
      8 <-> 16 <-> 10
      8 <-> 18 <-> 10
      9 <-> 17 <-> 11
      9 <-> 19 <-> 11
      ---
      12 <-> 20 <-> 14
      12 <-> 22 <-> 14
      13 <-> 21 <-> 15
      13 <-> 23 <-> 15
      This level is divided into 2 sections/4 groups. Even switches
      8,9 do connect to the same switches, their links "cross"
      putting them into the same tree "cross" section
      {8,9,19,11,16,17,18,19} form a connection section
            {8,10,16,18} form a connection group
            {9,11,17,19} form another connection group
      {12,13,14,15,20,21,22,23} for a connection section
            {12,14,20,22} form a connection group
            {13,15,21,23} form another connection group

      For L=3, the top level, we have
      16 <-> 24* <-> 20
      16 <-> 28* <-> 20
      17 <-> 25* <-> 21
      17 <-> 29* <-> 21
      18 <-> 26* <-> 22
      18 <-> 30* <-> 22
      19 <-> 27* <-> 23
      19 <-> 31* <-> 23
      The level is now all one section (all switches have cross links)
      We still have 4 groups.
      Switches 24-31 have an asterisk indicating it's actually just
      a 'virtual' label.  The top level only has 4 switches, not 8.
      We only need 4 switches because top level switches can use all
      2*k ports for down connection.  Thus, each top level switch
      performs the role of two 'virtual' switches.
      Switch 24 actually performs the virtual role 24,25
      Switch 25 actually performs the virtual role 26,27
      Thus, the actual set of connections is
      16 <-> 24 <-> 20
      16 <-> 26 <-> 20
      17 <-> 24 <-> 21
      17 <-> 26 <-> 21
      18 <-> 25 <-> 22
      18 <-> 27 <-> 22
      19 <-> 25 <-> 23
      19 <-> 27 <-> 23
      */
      for (int l = 0; l < toplevel_; l++)
      {
        int num_tree_sections = level_size / group_stride / k_;
        int section_size = level_size / num_tree_sections;
        int num_groups = level_size / group_size;
        int num_groups_per_section = num_groups / num_tree_sections;
        for (int secnum=0; secnum < num_tree_sections; ++secnum){
            long up_section_offset = up_level_start + section_size * secnum;
            long down_section_offset = down_level_start + section_size * secnum;
            connect_section(switches, group_stride,
                            num_groups_per_section,
                            group_size, down_section_offset,
                            group_size, up_section_offset);
        }
        down_level_start += numleafswitches_;
        up_level_start += numleafswitches_;
        group_stride *= k_;
    }
   }

   long
   fattree::get_switch_id(const coordinates &coords) const
   {
    long levels_down = coords[0];
    if (levels_down == 0){
        long my_col = coords[1];
        //we have l - 1 rows of numleafswitches
        //before getting to my row
        return (l_ - 1) * numleafswitches_ + my_col;
    }

    long branch_offset = 0;
    long router_degeneracy = numleafswitches_ / k_;
    for (int i=1; i <= levels_down; ++i){
        long offset_increment = coords[i] * router_degeneracy;
        branch_offset += offset_increment;
        router_degeneracy /= k_;
    }
    long replicate_num = coords[levels_down+1];
    long levels_up = l_ - 1 - levels_down;
    return branch_offset + replicate_num + levels_up * numleafswitches_;
   }

   void
   fattree::compute_switch_coords(long uid, coordinates& coords) const
   {
    // first we indicate which level of the tree
    int levels_up = uid / numleafswitches_;
    int levels_down = l_ - 1 - levels_up;
    long my_col = uid % numleafswitches_;
    if (levels_down == 0){
        coords[0] = 0;
        coords[1] = my_col;
        coords.resize(2);
        return;
    }

    // first coordinate indicates the level counting top-down
    // second set of coordinates indicates each branch taken on the way down
    // final coordinate indicates which "replicate" this switch is
    // a "virtual" fat-tree switch is implemented via multiple commodity switches
    int num_coords = 1 + levels_down + 1;
    coords.resize(num_coords);
    coords[0] = levels_down;
    long router_degeneracy = numleafswitches_ / k_;
    long replicate_num = 0;
    long branch_offset = 0;
    for (int lvl=1; lvl <= levels_down; ++lvl){
        //figure out which branch to take
        long idx = my_col - branch_offset;
        long branch = idx / router_degeneracy;
        coords[lvl] = branch;
        replicate_num = idx % router_degeneracy;
        branch_offset += branch*router_degeneracy;
        router_degeneracy /= k_;
    }
    coords[levels_down+1] = replicate_num;
   }

   void
   fattree::minimal_route_to_coords(
    const coordinates &current_coords,
    const coordinates &dest_coords,
    int &dim, int &dir) const
   {
        sst_throw(unimplementederror, "fattree::minimal_route_to_coords");
   }

   void
   fattree::minimal_route_to_switch(
    const nodeaddress::ptr &current_sw_addr,
    const nodeaddress::ptr &dest_sw_addr,
    int &dim, int &dir) const
   {
        sst_throw(unimplementederror, "fattree::minimal_route_to_switch");
   }

    int
    fattree::convert_to_outport(int dim, int dir)
    {
      if (dim == topology::EJECT)
      {
        return -(1 + dir);
      }
      return dim * k_ + dir;
    }

    int
    fattree::minimal_distance(const coordinates &current_coords, const coordinates &dest_coords) const
    {
        //can we go straight down? or do we need to go up
        //first coordinate is the number of levels down
        int smallest_depth = std::min(current_coords[0], dest_coords[0]);
        //find the largest index that differs
        int num_branches_different = 0;
        for (int i=smallest_depth; i >= 1; --i){
            if (current_coords[i] != dest_coords[i])
                num_branches_different = smallest_depth - i + 1;
        }

        int csize = current_coords.size();
        int dsize = dest_coords.size();
        if (num_branches_different == 0 && csize == dsize){
            //we might be on the same branch, but
            //we might be different replicate numbers
            int lastidx = csize - 1;
            //if different, two hops are needed
            //one up, then one back down to correct replicate number
            return current_coords[lastidx] == dest_coords[lastidx] ? 0 : 2;
        }
        else {
            // we need to go up and back down max_idx steps
            // then fill out the difference
            return 2*num_branches_different + abs(csize - dsize);
        }
    }

    void
    fattree::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
      {
        dim = topology::EJECT;
        dir = -(outport + 1);
      }
      else
      {
          dim = outport / k_;
          dir = outport % k_;
      }
    }



  }
} //end of namespace sstmac
