/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// hdtorus.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov> 
// 

#include <sstmac/hardware/network/topology/hdtorus.h>
#include <stdio.h> 
#include <sstream>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("torus | hdtorus", topology, hdtorus,
        "hdtorus implements a high-dimension torus with an arbitrary number of dimensions");

    hdtorus::ptr
    hdtorus::construct(const sim_parameters::ptr& params)
    {
      return new hdtorus;

    }

    static bool
    equals(const std::vector<int>& coords, int x, int y, int z)
    {
      if (coords.size() != 3)
      {
        return false;
      }
      return coords[0] == x && coords[1] == y && coords[2] == z;
    }

    /*  Generates a high dimensional torus network
     Inputs:
     dimensionarray: specifies the number of processors in each dimension
     and whether  the links are wrapped around  (torus) or not (mesh).
     dimensionarray[i] <0 refers to absence of  wrap around edges.
     dimensionarray[i] >0 refers to presence of wrap around edges.
     nndpersw is the number of nodes per switch.
     widtharrg specifies the width between the switches and the  nodes.
     */
    hdtorus::hdtorus() :
      regular_topology(1) //initialize to 1 for now
    {
    }

    void
    hdtorus::init_factory_params(const sim_parameters::ptr &params)
    {
      std::vector<int> args;
      /**
       sstkeyword {
       gui=10 5 10;
       docstring=The number of switches in each dimension of the torus.ENDL
       An arbitrary number of dimensions can be used.ENDL
       The total number of switches is:ENDL
       N_Switches=N_X*N_Y*...*N_Z.ENDL
       The total number of compute nodes is:ENDL
       N_Comp_Nodes=N_Switches*N_Nodes_Per_switch;
       }
       */
      params->get_vector_param("topology_geometry", args);

      if (args.size() == 0)
      {
        sst_throw(valueerror, "empty topology vector for hdtorus");
      }
      int nps = params->get_optional_int_param("network_nodes_per_switch", 1);

      init(args, nps);
      regular_topology::init_factory_params(params);
    }

    void
    hdtorus::init(const std::vector<int>& dimensionarray, int nndspersw)
    {
      int P, i;

      // Pre-process  the input;
      int64_t dsize = dimensionarray.size();
      nested_type<int>::vector vertices;

      dimensions_.resize(dsize);
      wrapped_.resize(dsize);

      for (i = 0; i < dsize; i++)
      {
        if (dimensionarray[i] == 0)
        {
          throw illformederror(
              "hdtorus::hdtorus: dimension of size 0 not allowed");
        }
        else if (dimensionarray[i] < 0)
        {
          wrapped_[i] = 0;
          dimensions_[i] = -dimensionarray[i];
        }
        else if (dimensionarray[i] > 2)
        {
          dimensions_[i] = dimensionarray[i];
          wrapped_[i] = 1;
        }
        else if (dimensionarray[i] == 2)
        {
          // This handles the case of dimension = 1 or 2, where
          // wrapping is turned off.
          dimensions_[i] = dimensionarray[i];
          wrapped_[i] = 0;
        }
        else
        {
          sst_throw(illformederror, "hdtorus::hdtorus: dimension of size 1 is not allowed");
        }
      }

      num_switches_ = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        num_switches_ *= dimensions_[i];
      }
      nps_ = nndspersw;
    }

    /// Returns a descriptive name of the graph.
    std::string
    hdtorus::name() const
    {
      int64_t dsize = dimensions_.size();
      std::ostringstream ostr;
      ostr << "hdtorus(";
      for (int i = 0; i < dsize; i++)
        ostr << dimensions_[i] << ",";
      ostr << ")";
      return ostr.str();
    }

    std::string
    hdtorus::label(long id) const
    {
      std::vector<int> coords = get_switch_coords(id);
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) coords.size(); ++i)
      {
        sstr << sstprintf(" %2d", coords[i]);
      }
      sstr << " ]";
      return sstr.str();
    }

    void
    hdtorus::compute_switch_coords(long uid, coordinates& coords) const
    {
      long div = 1;
      int ndim = dimensions_.size();
      for (int i = 0; i < ndim; i++)
      {
        coords[i] = (uid / div) % dimensions_[i];
        div *= dimensions_[i];
      }
    }

    long
    hdtorus::get_switch_id(const coordinates& coords) const
    {
      long ret = 0;
      long mult = 1;
      for (int i = 0; i < (int) dimensions_.size(); i++)
      {
        if (coords[i] >= dimensions_[i])
        {
          sst_throw(illformederror, "hdtorus::get_switch_id: dimension %d coordinate %d exceeds maximum %d\n"
              "check that topology_geometry matches map file",
              i, coords[i], dimensions_[i] - 1);
        }
        ret += coords[i] * mult;
        mult *= dimensions_[i];
      }
      return ret;
    }

    std::string
    hdtorus::coord_string(const std::vector<int>& coords)
    {
      if (coords.size() == 0){
        return "[empty]";
      }

      std::stringstream ss;
      ss << "[" << coords[0];
      for (int i = 1; i < (int) dimensions_.size(); i++)
      {
        ss << "," << coords[i];
      }
      ss << "]";
      return ss.str();
    }

    static std::string
    get_link_name(const std::vector<int>& src, const std::vector<int>& dst)
    {
      std::stringstream sstr;
      sstr << "[";
      for (int i = 0; i < (int) src.size(); ++i)
      {
        sstr << sstprintf(" %2d", src[i]);
      }
      sstr << " ]->[";
      for (int i = 0; i < (int) dst.size(); ++i)
      {
        sstr << sstprintf(" %2d", dst[i]);
      }
      sstr << "]";
      return sstr.str();
    }

    int
    hdtorus::get_distance(int dim, int dir, int src, int dst) const
    {
        if (dir == topology::POS){
            return ((dst - src) + dimensions_[dim]) % dimensions_[dim];
        }
        else {
            return ((src - dst) + dimensions_[dim]) % dimensions_[dim];
        }
    }

    int
    hdtorus::get_shortest_distance(int dim, int src, int dst) const
    {
        int up_distance, down_distance;
        if (dst > src){
            up_distance = dst - src;
            down_distance = src + (dimensions_[dim] - dst);
        }
        else {
            up_distance = dst + (dimensions_[dim] - src);
            down_distance = src - dst;
        }

        if (up_distance > down_distance){
            //shorter to go down
            return ((src - dst) + dimensions_[dim]) % dimensions_[dim];
        }
        else {
            //shorter to go up
            return ((dst - src) + dimensions_[dim]) % dimensions_[dim];
        }
    }

    int
    hdtorus::get_shortest_direction(int dim, int src, int dst) const
    {
        int up_distance, down_distance;
        if (dst > src){
            up_distance = dst - src;
            down_distance = src + (dimensions_[dim] - dst);
        }
        else {
            up_distance = dst + (dimensions_[dim] - src);
            down_distance = src - dst;
        }

        if (up_distance > down_distance){
            //shorter to go down
            return topology::NEG;
        }
        else {
            //shorter to go up
            return topology::POS;
        }
    }

    void
    hdtorus::minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const
    {
        for (int i=0; i < current_coords.size(); ++i){
            if (current_coords[i] != dest_coords[i]){
                dim = i;
                dir = get_shortest_direction(i, current_coords[i], dest_coords[i]);
                return;
            }
        }
    }

    int
    hdtorus::minimal_distance(
        const coordinates &current_coords,
        const coordinates &dest_coords) const
    {
        int dist = 0;
        for (int i=0; i < current_coords.size(); ++i){
            dist += get_shortest_distance(i, current_coords[i], dest_coords[i]);
        }
        return dist;
    }

    void
    hdtorus::connect_switches(topology::maptype &switches)
    {
      double link_weight = 1.0;
      SSTMAC_DEBUG << "hdtorus: connecting " << switches.size() << " switches" << "\n";
      topology::maptype::iterator it;
      for (it = switches.begin(); it != switches.end(); ++it)
      {
        nodeaddress::ptr me = it->first;
        coordinates coords = get_switch_coords(me->unique_id());

        SSTMAC_DEBUG << "switch " << me->unique_id() << " " << coord_string(coords) << "... \n";

        std::vector<int> connected;
        std::vector<int> connected_red;

        for (int z = 0; z < (int) dimensions_.size(); z++)
        {
          coordinates connect1 = coords;
          coordinates connect2 = coords;
          if (coords[z] == 0)
          {
            connect1[z]++;
            connect2[z] = dimensions_[z] - 1;
          }
          else if (coords[z] == (dimensions_[z] - 1))
          {
            connect1[z] = 0;
            connect2[z]--;
          }
          else
          {
            connect1[z]++;
            connect2[z]--;
          }
          long uid1 = get_switch_id(connect1);
          long uid2 = get_switch_id(connect2);

          nodeaddress::ptr addr1 = nodeaddress::convert_from_uniqueid(uid1);
          nodeaddress::ptr addr2 = nodeaddress::convert_from_uniqueid(uid2);

          networkswitch::ptr dest1 = NULL, dest2 = NULL;

          dest1 = find_switch(switches, addr1);
          dest2 = find_switch(switches, addr2);

          if (outputgraph_)
          {
            connected.push_back(addr1->unique_id());
            connected.push_back(addr2->unique_id());
            connected_red.push_back(red_[z]);
            connected_red.push_back(red_[z]);

          }

          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the pos dir to " << coord_string(connect1)
                << " which is addr " << addr1->to_string() << " and object " << dest1->to_string() << "\n";
          SSTMAC_DEBUG << "switch " << me->to_string() << " with coords "
                << coord_string(coords) << " for dim " << z
                << " connecting in the neg dir to " << coord_string(connect2)
                << " which is addr " << addr2->to_string() << " and object " << dest2->to_string() << "\n";

          switches[me]->connect(z, topology::POS, connectable::network_link,
              dest1, link_weight, red_[z]);

          switches[me]->connect(z, topology::NEG, connectable::network_link,
              dest2, link_weight, red_[z]);

        }

        if (outputgraph_)
        {
          for (int k = 0; k < nps_; k++)
          {
            std::cout << (me->unique_id() * nps_ + k);
            for (int l = 0; l < (int) connected.size(); l++)
            {
              for (int m = 0; m < nps_; m++)
              {
                std::cout << " " << (connected[l] * nps_ + m) << ":"
                    << connected_red[l];

              }
            }
            for (int m = 0; m < nps_; m++)
            {
              if (m != k)
              {
                std::cout << " " << (me->unique_id() * nps_ + m) << ":"
                    << 100000;
              }
            }
            std::cout << "\n";
          }

        }
      }

    }

    void
    hdtorus::display_nodes(const vismaptype &switches,
        const vis::vis_engine::ptr& eng, std::list<vis::vis_obj::ptr> &objs)
    {
      for (long i = 0; i < (int) switches.size(); i++)
      {
        nodeaddress::ptr me = nodeaddress::convert_from_uniqueid(i);

        std::vector<int> coords = get_switch_coords(i);

        double x = 0, y = 0, z = 0;
        x = coords[0] * 2;
        if (coords.size() > 1)
          y = coords[1] * 2;

        if (coords.size() > 2)
          z = coords[2] * 2;

        vis::vis_obj::ptr vo = eng->create_cube(x, y, z, 1);
        const vismaptype::mapped_type& myswitch = mapget(switches, me,
            "network switch for hdtorus");
        myswitch->set_vis_obj(vo);
        myswitch->set_vis_engine(eng);
        objs.push_back(vo);

        for (int z = 0; z < (int) dimensions_.size(); z++)
        {
          double x1 = coords[0] * 2;
          double y1 = coords[1] * 2;
          double z1 = coords[2] * 2;

          double x2 = coords[0] * 2;
          double y2 = coords[1] * 2;
          double z2 = coords[2] * 2;
          int dir;

          bool onegood = (coords[z] != dimensions_[z] - 1);
          bool twogood = (coords[z] != 0);

          if (z == 0)
          {

            x1 += 1.5;
            z1 += 0.5;
            y1 += 0.25;

            x2 -= 0.5;
            z2 += 0.5;
            y2 += 0.75;

            dir = vis::vis_engine::VIS_Z;
          }
          else if (z == 1)
          {
            y1 += 1.5;
            z1 += 0.25;
            x1 += 0.5;

            y2 -= 0.5;
            z2 += 0.75;
            x2 += 0.5;
            dir = vis::vis_engine::VIS_Y;
          }
          else
          {
            z1 += 1.5;
            y1 += 0.25;
            x1 += 0.5;

            z2 -= 0.5;
            y2 += 0.75;
            x2 += 0.5;
            dir = vis::vis_engine::VIS_X;
          }

          const vismaptype::mapped_type& myswitch = mapget(switches, me,
              "vis switch for hdtorus");
          if (onegood)
          {
            vis::vis_obj::ptr c1 = eng->create_cylinder(x1, y1, z1, 0.125, 1,
                dir);
            objs.push_back(c1);
            SSTMAC_DEBUG << "switch " << i << " setting vis link for dim " << z
                  << " dir " << topology::POS << "\n";
            myswitch->set_link_vis(c1, z, topology::POS);
          }

          if (twogood)
          {
            vis::vis_obj::ptr c2 = eng->create_cylinder(x2, y2, z2, 0.125, 1,
                dir);
            SSTMAC_DEBUG << "switch " << i << " setting vis link for dim " << z
                  << " dir " << topology::NEG << "\n";
            myswitch->set_link_vis(c2, z, topology::NEG);
            objs.push_back(c2);
          }
        }
      }
    }

    int
    hdtorus::get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const
    {
        return get_shortest_direction(dim, src[dim], dst[dim]);
    }

    int
    hdtorus::convert_to_outport(int dim, int dir)
    {
      if (dim == topology::EJECT)
        {
          return -(1 + dir);
        }
      return dim * 2 + (dir == topology::POS ? 0 : 1);
    }

    void
    hdtorus::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0)
        {
          dim = topology::EJECT;
          dir = (-outport) - 1;
        }
      else
        {
          dim = outport / 2;
          dir = (outport % 2 == 0 ? topology::POS : topology::NEG);
        }
    }

    void
    hdtorus::get_nearest_neighbor_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        int ndim = dimensions_.size();
        partners.resize(ndim * 2);
        coordinates tmp_coords(src_sw_coords);
        int idx = 0;
        for (int i=0; i < ndim; ++i){
            int dim_size = dimensions_[i];

            //plus partner
            tmp_coords[i] = (src_sw_coords[i] + 1) % dim_size;
            long nid = get_node_id(tmp_coords, port);
            partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

            //minus partner
            tmp_coords[i] = (src_sw_coords[i] + dim_size - 1) % dim_size;
            nid = get_node_id(tmp_coords, port);
            partners[idx++] = nodeaddress::convert_from_uniqueid(nid);

            //reset coords
            tmp_coords[i] = src_sw_coords[i];
        }
    }

    void
    hdtorus::get_tornado_recv_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        int ndim = dimensions_.size();
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        for (int i=0; i < ndim; ++i){
            int dim_size = dimensions_[i];
            tmp_coords[i] = (src_sw_coords[i] - (dim_size - 1)/2 + dim_size) % dim_size;
        }
        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }

    void
    hdtorus::get_tornado_send_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        int ndim = dimensions_.size();
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        for (int i=0; i < ndim; ++i){
            int dim_size = dimensions_[i];
            tmp_coords[i] = (src_sw_coords[i] + (dim_size - 1)/2) % dim_size;
        }
        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }

    void
    hdtorus::get_bit_complement_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const
    {
        int ndim = dimensions_.size();
        partners.resize(1);
        coordinates tmp_coords(src_sw_coords);
        for (int i=0; i < ndim; ++i){
            int dim_size = dimensions_[i];
            tmp_coords[i] = dim_size - src_sw_coords[i] - 1;
        }
        long nid = get_node_id(tmp_coords, port);
        partners[0] = nodeaddress::convert_from_uniqueid(nid);
    }

  }
} //end of namespace sstmac


