/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

// hdtorus.h: Interface for torus networks.
//
// Author: Ali Pinar <apinar@sandia.gov>

#ifndef SSTMAC_HARDWARE_NETWORK_TOPOLOGY_HDTORUS_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_TOPOLOGY_HDTORUS_H_INCLUDED
#include <sstmac/hardware/network/topology/regular_topology.h>

namespace sstmac
{
  namespace hw
  {

    /** Implements a high Dimensional torus network.

     <b>Configuring the 2HDTorus network</b>

     The inputs for this network can specify the following attributes:

     <table border="1">
     <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>

     <b>Configuring the 3HDTorus network</b>

     The inputs for this network
     can specify the following attributes:

     <table border="1">
     <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>

     <b>Configuring the 4HDTorus network</b>

     The inputs for this network
     can specify the following attributes:

     <table border="1">      <tr><td>Keyword<td>Type<td>Default<td>Description
     <tr><td><tt>n1</tt><td>int<td>4<td>The number of nodes in the first
     dimension.
     <tr><td><tt>n2</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>n3</tt><td>int<td>0<td>The number of nodes in the third
     dimension.
     <tr><td><tt>nx</tt><td>int<td>4<td>The number of nodes in the second
     dimension.
     <tr><td><tt>nndpersw</tt><td>int<td>0<td>The number of nodes per switch.
     <tr><td>width<td>int<td>0<td>The width of edges connecting two switches.
     </table>
     */

    class hdtorus :
        public regular_topology
    {

    protected:
      long num_switches_;
      std::vector<int> dimensions_;
      std::vector<bool> wrapped_;

    protected:
      hdtorus();

      void
      init(const std::vector<int>& dimensionarray, int nndpersw);

      virtual void
      compute_switch_coords(long uid, coordinates& coords) const;

      void
      get_nearest_neighbor_partners(
        const coordinates& src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_bit_complement_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_tornado_send_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

      void
      get_tornado_recv_partners(
        const coordinates &src_sw_coords,
        int port,
        std::vector<nodeaddress::ptr>& partners) const;

    public:
      typedef boost::intrusive_ptr<hdtorus> ptr;

      static ptr
      construct(const sim_parameters::ptr& params);

      void
      init_factory_params(const sim_parameters::ptr &params);

      int get_shortest_direction(int dim, int src, int dst) const;

      int get_shortest_distance(int dim, int src, int dst) const;

      int get_distance(int dim, int dir, int src, int dst) const;

      int
      get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const;

      long
      get_switch_id(const coordinates& v) const;

      virtual std::string label(long id) const;

      virtual int
      radix() const
      {
        return 2 * dimensions_.size();
      }

      std::string
      coord_string(const std::vector<int>& coords);

      virtual
      ~hdtorus()
      {
      }

      std::string
      name() const;

      /// Returns the vector giving each dimension of the torus.
      const std::vector<int> &
      dimensions() const
      {
        return dimensions_;
      }

      virtual long
      num_switches() const {
        return num_switches_;
      }

      virtual long
      num_leaf_switches() const {
        return num_switches();
      }

      virtual void
      display_nodes(const vismaptype &switches,
          const vis::vis_engine::ptr& eng,
          std::list<vis::vis_obj::ptr> &objs);

      virtual int
      ndimensions() const {
          return dimensions_.size();
      }

      virtual void
      minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const;

      virtual void
      connect_switches(topology::maptype &switches);

      virtual int
      convert_to_outport(int dim, int dir);

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir);

      virtual std::string
      to_string() const
      {
        return "hdtorustopology";
      }

      int
      minimal_distance(
        const coordinates& current_coords,
        const coordinates& dest_coords
      ) const;

      void
      get_partners(
        traffic_pattern::type_t ty,
        const coordinates &src_sw_coords,
        std::list<nodeaddress::ptr>& partners
      ) const;

    };

  }
} //end of namespace sstmac

#endif
