/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

// hypercube.cc: Implementation of a high dimensional torus network.
//
// Author: Ali Pinar <apinar@sandia.gov>
//

#include <sstmac/hardware/network/topology/hypercube.h>
#include <stdio.h>
#include <sstream>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac
{
  namespace hw
  {
    SSTRegister("hypercube", topology, hypercube,
        "hypercube implements a high-dimension torus with an arbitrary number of dimensions");

    hypercube::ptr
    hypercube::construct(const sim_parameters::ptr& params)
    {
      return new hypercube;
    }

    hypercube::hypercube() :
        hdtorus()
    {
    }

    void
    hypercube::init_factory_params(const sim_parameters::ptr &params)
    {
        hdtorus::init_factory_params(params);
        ndim_ = dimensions_.size();
        dim_to_outport_.resize(ndim_);
        int offset = 0;
        for (int i=0; i < ndim_; ++i){
            dim_to_outport_[i] = offset;
            offset += dimensions_[i];
        }
        radix_ = offset;
    }

    /// Returns a descriptive name of the graph.
    std::string
    hypercube::name() const
    {
      int64_t dsize = dimensions_.size();
      std::ostringstream ostr;
      ostr << "hypercube(";
      for (int i = 0; i < dsize; i++)
        ostr << dimensions_[i] << ",";
      ostr << ")";
      return ostr.str();
    }

    void
    hypercube::minimal_route_to_coords(
        const coordinates &current_coords,
        const coordinates &dest_coords,
        int &dim, int &dir) const
    {
        for (int i=0; i < current_coords.size(); ++i){
            if (current_coords[i] != dest_coords[i]){
                dim = i;
                dir = dest_coords[i];
                return;
            }
        }
    }

    int
    hypercube::minimal_distance(
        const coordinates &current_coords,
        const coordinates &dest_coords) const
    {
        int dist = 0;
        for (int i=0; i < current_coords.size(); ++i){
            if (current_coords[i], dest_coords[i]){
                ++dist;
            }
        }
        return dist;
    }

    void
    hypercube::connect_switches(topology::maptype &switches)
    {
      double link_weight = 1.0;
      SSTMAC_DEBUG << "hypercube: connecting " << switches.size() << " switches" << "\n";
      topology::maptype::iterator it;
      for (it = switches.begin(); it != switches.end(); ++it)
      {
        nodeaddress::ptr me = it->first;
        coordinates coords = get_switch_coords(me->unique_id());

        //loop every dimension and connect to all "neighbors"
        for (int dim=0; dim < dimensions_.size(); ++dim)
        {
            coordinates neighbor_coords = coords;
            long my_id = get_switch_id(coords);
            int dimsize = dimensions_[dim];
            int my_idx = coords[dim];
            int red = red_[dim];
            for (int dir=0; dir < dimsize; ++dir)
            {
                if (dir != my_idx){
                    neighbor_coords[dim] = dir;
                    networkswitch::ptr neighbor_sw = get_switch(neighbor_coords, switches);
                    SSTMAC_DEBUG << sstprintf("Connecting %s(%ld) to %s(%ld) at dim=%d,dir=%d\n",
                                            coord_string(coords).c_str(), my_id,
                                            coord_string(neighbor_coords).c_str(), neighbor_sw->get_addr()->unique_id(),
                                            dim, dir);
                    switches[me]->connect(dim, dir, connectable::network_link,
                      neighbor_sw, link_weight, red);
                }
            }
          }
        }
    }

    void
    hypercube::display_nodes(const vismaptype &switches,
        const vis::vis_engine::ptr& eng, std::list<vis::vis_obj::ptr> &objs)
    {
        sst_throw(unimplementederror, "hypercube::display_nodes");
    }

    int
    hypercube::get_productive_dir(
        int &dim,
        const coordinates &src,
        const coordinates &dst) const
    {
        return dst[dim];
    }

    int
    hypercube::convert_to_outport(int dim, int dir)
    {
      if (dim == topology::EJECT){
          return -(1 + dir);
      }
      return dim_to_outport_[dim] + dir;
    }

    void
    hypercube::convert_to_dimdir(int outport, int &dim, int&dir)
    {
      if (outport < 0){
          dim = topology::EJECT;
          dir = (-outport) - 1;
      }
      else {
          for (int i=0; i < ndim_; ++i){
            int offset = dim_to_outport_[outport];
            if (offset >= dim_to_outport_[dim]){
                dim = i;
                dir = outport - offset;
                return;
            }
          }
      }
    }


  }
} //end of namespace sstmac



