/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/routing/fattreerouter.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

namespace sstmac
{
  namespace hw
  {

    SSTRegister("fattree | ftree", router, fattreerouter);
    RNG::Combo::ptr fattreerouter::rng_;

     fattreerouter::fattreerouter()
      {
        if (!rng_)
        {
          long seed = time(NULL);
          if (sstmac::sstmac_env::params->has_param("router_seed"))
            seed = sstmac::sstmac_env::params->get_long_param("router_seed");

          std::vector<RNG::rngint_t> seeds;
          seeds.push_back(seed);
          rng_ = RNG::Combo::construct(seeds);

        }

      }

    int
    fattreerouter::convert_to_outport(int dim, int dir)
    {
        if (is_top_level()){
            //only one dimesion - down
            return dir;
        }
        else {
            return sw_->get_topology()->convert_to_outport(dim, dir);
        }
    }

    void
    fattreerouter::convert_to_dimdir(int outport, int &dim, int &dir)
    {
        if (is_top_level()){
            dim = fattree::down_dimension;
            dir = outport;
        }
        else {
            sw_->get_topology()->convert_to_dimdir(outport, dim, dir);
        }
    }

    void
    fattreerouter::get_productive_ports_to_switch(const nodeaddress::ptr &dst, outports &ports)
    {
        int dim, dir;
        minimal_route_to_switch(dst, dim, dir);
        if (is_top_level()){
            ports.resize(2);
            int path_group = dir % k_;
            //we have two possible down connections
            ports[0] = path_group;
            ports[1] = path_group + k_;
        }
        else if (dim == fattree::down_dimension){
            //we have no choice - only one path down is correct
            ports.resize(1);
            ports[0] = convert_to_outport(dim, dir);
        }
        else {
            // we have k productive paths up
            ports.resize(k_);
            for (int i=0; i < k_; ++i){
                ports[i] = convert_to_outport(dim, i);
            }
        }
    }

    void
    fattreerouter::minimal_route_to_switch(
        const nodeaddress::ptr& ej_addr,
        int& dim, int& dir)
    {
     long ej_id = ej_addr->unique_id();
      if (myL_ == 0)
      { // if we're on the first level and we're not there yet, go up
        dim = 0;
        dir = choose_up_path();
        SSTMAC_DEBUG << "routing up from first level \n";
      }
      else if (is_top_level())
      {
        dim = fattree::down_dimension;
        //in an l=3,k=2 train, e.g.
        // switch 8 performs the "role" of switches 8,9
        // switch 9 performs the "role" of switches 10,11
        // choose one of the roles to perform
        int my_virtual_role = choose_down_path();
        long relative_ej_id = ej_id - min_reachable_leaf_id_;
        // the leaf ids are divided into groups of size num_leaf_switches_per_path
        // based on which leaf nodes are reachable following a given path
        long path_group = relative_ej_id / num_leaf_switches_per_path_;
        // each node can play "two" roles giving two degenerate paths down
        dir = path_group + my_virtual_role * k_;
        // Because
        SSTMAC_DEBUG << sstprintf("routing down with dir %d: eject-id=%ld rel-eject-id=%ld\n",
                                  dir, ej_id, relative_ej_id);
      }
      else if (ej_id >= min_reachable_leaf_id_ && ej_id < max_reachable_leaf_id_)
      {
        dim = fattree::down_dimension;
        long relative_ej_id = ej_id - min_reachable_leaf_id_;
        dir = relative_ej_id / num_leaf_switches_per_path_;
        SSTMAC_DEBUG << sstprintf("routing down with dir %d: eject-id=%ld rel-eject-id=%ld\n",
                                  dir, ej_id, relative_ej_id);
      }
      else
      { //route up
        dim = fattree::up_dimension;
        dir = choose_up_path();
        SSTMAC_DEBUG << sstprintf("routing up with dir %d\n");
      }
    }

    int
    fattreerouter::choose_up_path()
    {
      int ret = numpicked_;
      numpicked_ = (numpicked_ + 1) % k_;
      return ret;
    }

      void
      fattreerouter::init(const nodeaddress::ptr& me, const topology::ptr& top)
      {
        fattree::ptr ft_top = safe_cast(top, fattree);
        k_ = ft_top->k();
        l_ = ft_top->l();

        regular_router::init(me, top);

        int switchesperlevel = pow(k_, l_ - 1);
        myL_ = me->unique_id() / switchesperlevel;
        logicalid_ = (me->unique_id() - (myL_ * switchesperlevel)) / pow(k_,
            myL_);

        num_leaf_switches_reachable_ = pow(k_, myL_);
        num_leaf_switches_per_path_ = num_leaf_switches_reachable_ / k_;
        level_relative_id_ = me->unique_id() - myL_ * switchesperlevel;
        if (is_top_level())
            level_relative_id_ *= 2;

        int my_leaf_group = level_relative_id_ / num_leaf_switches_reachable_;
        min_reachable_leaf_id_ = my_leaf_group * num_leaf_switches_reachable_;
        max_reachable_leaf_id_ = min_reachable_leaf_id_ + num_leaf_switches_reachable_;

        pickstart_ = rng_->value() % k_;
        numpicked_ = 0;

        numpicktop_ = 0;
      }

    int
    fattreerouter::choose_down_path()
    {
        int ret = numpicktop_;
        int two_down_connections_per_top_switch = 2;
        numpicktop_ = (numpicktop_ + 1) % two_down_connections_per_top_switch;
        return ret;
    }

    int
    fattreerouter::get_number_paths(const sst_message::ptr& msg) const
    {
      nodeaddress::ptr ej_addr = sw_->get_topology()->node_to_ejector_addr(msg->toaddr());
      long ej_id = ej_addr->unique_id();
      if (ej_addr->equals(me_))
      {
        return 1;
      }
      else if (myL_ == 0)
      {
        return k_;
      }
      else if (is_top_level())
      {
        return 2;
      }
      else if (ej_id >= min_reachable_leaf_id_ && ej_id < max_reachable_leaf_id_)
      {
        return 1;
      }
      else
      {
        //route up
        return k_;
      }
    }

    void
    fattreerouter::set_inport(nodeaddress::ptr from, int port)
    {
      inports_[from->unique_id()] = port;
    }

    int
    fattreerouter::get_inport(nodeaddress::ptr from, int dim, int dir)
    {
      return inports_[from->unique_id()];
    }



  }
}
