/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_NETWORK_ROUTING_ROUTER_H_INCLUDED
#define SSTMAC_HARDWARE_NETWORK_ROUTING_ROUTER_H_INCLUDED

#include <sstmac/common/debug.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/network/interconnect/interconnect_message.h>
#include <sstmac/common/factories/factory.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/topology/routing/algorithm.h>

namespace sstmac
{
  namespace hw
  {


    /**
      @class router
      Class that computes the next step a messag should taken in traversing
      the network.  This performs routing operations only and is not actually
      a 'component' in the network - those are switches.  Switch and router
      are not synonymous in SST/macro.  All switches have routers.
    */
    class router :
        public factory_type
    {
    protected:

      virtual int
      do_route(const interconnect_message::ptr &msg);

      void
      init_routing(const interconnect_message::ptr& msg) const;

      routing_info::routing_algorithm_t default_algo_;

      routing_info::routing_algorithm_t
      str_to_algo(const std::string& str);

      typedef std::map<routing_info::routing_algorithm_t, routing_algorithm::ptr> algo_map;
      algo_map algos_;

    public:
      typedef std::vector<int> outports;

      typedef boost::intrusive_ptr<router> ptr;

      virtual std::string
      to_string() const {
        return "router";
      }

      virtual ~router();

      static bool
      is_eject(long port){
        return port < 0;
      }

      virtual void
      init_factory_params(const sim_parameters::ptr &params);

      virtual void
      init(const nodeaddress::ptr& me, const topology::ptr& top);

      /**
        @return The outport to be routed to
      */
      int
      route(const interconnect_message::ptr& msg);

      virtual int
      basic_route(const interconnect_message::ptr& msg) = 0;

      virtual int
      convert_to_outport(int dim, int dir) = 0;

      virtual int
      convert_to_inport(int dim, int dir){
        return convert_to_outport(dim, dir);
      }

      virtual void
      convert_to_dimdir(int outport, int &dim, int&dir) = 0;

      virtual int num_nodes() const = 0;

      /**
       * Override this to indicate that packets can take multiple paths
       */
      virtual int
      get_number_paths(const interconnect_message::ptr& msg) const;

      virtual void
      set_switch(const networkswitchPtr& sw);

      /**
        @return The inport
      */
      virtual int
      get_inport(const nodeaddress::ptr& output_buf_sw,
                 const nodeaddress::ptr& input_but_bw,
                 int dim, int dir) = 0;

      /**
      */
      virtual bool
      is_fixed_routing() const
      {
        return false;
      }

      /**
        @return Whether the router needs congestion info to properly make decisions.
                This is usally used to notify the switch that it should update
                congestion info - i.e. the router makes the routing decisions
                but the switch is what actually knows about congestion.
      */
      virtual bool
      needs_congestion_info() const {
        return false;
      }

      /**
        @param The node the port is connected to
        @param The port number
      */
      virtual void
      set_inport(nodeaddress::ptr from, int port)
      {
      }


      /**
       @return Whether we are ejecting
       @param dst The destination node
       @param ports The ports that make progress towards the final destination.
                    If ejecting, this contains only the ejection port.
      */
      virtual bool
      get_productive_ports_to_node(
        const nodeaddress::ptr& dst,
        outports& ports);

      virtual void
      get_productive_ports_to_switch(
        const nodeaddress::ptr& dst,
        outports& ports) = 0;

    protected:
      nodeaddress::ptr me_;

      networkswitchPtr sw_;

      static sstmac::debug dbg_;

      static bool template_init_;

      router();

      virtual void clear_routing_info(){}


    };

    DeclareFactory2InitParams(router,nodeaddress::ptr,topology::ptr);




  }
}
#endif
