/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/topology/routing/tff_router_djikstra.h>
#include <sstmac/hardware/network/topology/routing/basic_routing.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/common/factories/factory.h>
namespace sstmac
{
  namespace hw
  {
    using namespace sstmac;
    using namespace sstmac::hw;
    SSTRegister("top_from_file_djikstra | djikstra", router, tff_router_djikstra);

    int
    tff_router_djikstra::basic_route(const interconnect_message::ptr &msg)
    {
        int dim, dir;
        route(msg->toaddr(), dim, dir);
        return convert_to_outport(dim, dir);
    }

    void
    tff_router_djikstra::route(const nodeaddress::ptr& toaddr, int &dim, int &dir)
    {
      long dest = toaddr->unique_id();
      long myid = me_->unique_id();

      if (!built_rtable_)
      {
        build_rtable();
      }
      SSTMAC_DEBUG << "router " << myid << " going to " << dest
            << " at router " << reverse_table_[dest] << " ...";

      if (reverse_table_[dest] == myid)
      {
        SSTMAC_DEBUG << "ejecting \n";
        dim = topology::EJECT;
        dir = dest;
      }
      else
      {
        dim = 0;
        if (rconnections_[myid].size() == 0)
        {
          sst_throw(
              ssterror,
              "top_from_file_djikstra::route() - router %d says: strange, I don't have any connections when going to %d",
              myid, dest);
        }

        long ind = reverse_table_[dest];
        dir = rtable_[ind];

        SSTMAC_DEBUG << " taking port " << dir << "\n";
      }

    }

    void
    tff_router_djikstra::get_productive_ports_to_switch(const nodeaddress::ptr &dst, outports &ports)
    {
        sst_throw(unimplementederror,
            "tff_router_djikstra::get_productive_ports_to_switch");
    }

    void
    tff_router_djikstra::build_rtable()
    {
      SSTMAC_DEBUG << "top_from_file_djikstra:: building rtable for "
            << nconnections_.size() << " switches ...\n";

      long myid = me_->unique_id();

      long num_routers = nconnections_.size();
      boost::unordered_map<long, double> dist;
      boost::unordered_map<long, long> previous;
      boost::unordered_map<long, long> Q;
      const long MAXLONG = 1000000000;

      for (long i = 0; i < num_routers; i++)
      {
        dist[i] = MAXLONG;
        //previous[i] = -1;
      }

      dist[myid] = 0; // source to itself
      Q[myid] = 0;

      for (long r = 0; r < num_routers; r++)
      {
        //find minimum in Q
        long minval = MAXLONG;
        long u = -1;
        boost::unordered_map<long, long>::iterator it, end = Q.end();
        for (it = Q.begin(); it != end; it++)
        {
          if (dist[it->first] < minval)
          {
            minval = dist[it->first];
            u = it->first;
          }
        }

        if (u < 0)
        { //can't reach any others
          break;
        }

        if (u != myid)
        {
          int go = u;
          SSTMAC_DEBUG2 << "[" << myid << "]: considering router " << u << "\n";
          //trace it back to a neighbor of me
          while (previous.find(go) != previous.end())
          {
            if (previous[go] == myid)
              break;
            go = previous[go];
          }

          SSTMAC_DEBUG2 << "[" << myid << "]: way to get there from here is "
                << go << "\n";

          //got the neighbor, find it in my connections to be sure
          int p = -1;
          for (int i = 0; i < rconnections_[myid].size(); i++)
          {
            SSTMAC_DEBUG2 << "[" << myid << "]: connected to me is "
                  << rconnections_[myid][i].id_ << "\n";
            if (rconnections_[myid][i].id_ == go)
            {
              p = i;
              break;
            }
          }

          if (p < 0)
          {
            sst_throw(
                ssterror,
                "build_rtable[%d]: didn't find port when looking for neighbor %d",
                myid, go);
          }

          rtable_[u] = go; // set it in the table
        }

        Q.erase(u);

        SSTMAC_DEBUG2 << "[" << myid << "]: distance to " << u << " is "
              << dist[u] << "\n";

        if (dist[u] >= MAXLONG)
        { //can't reach any other routers
          break;
        }

        for (int i = 0; i < rconnections_[u].size(); i++)
        {
          connection c = rconnections_[u][i];
          double alt = dist[u] + 1.0 / c.red_; //one way of taking link redundancy into account
          if (alt < dist[c.id_])
          {
            dist[c.id_] = alt;
            previous[c.id_] = u;
            Q[c.id_] = 0;
          }
        }

      }

      //build reverse table
      for (long i = 0; i < num_routers; i++)
      {
        std::vector<long>::iterator it, end = nconnections_[i].end();
        for (it = nconnections_[i].begin(); it != end; it++)
        {
          reverse_table_[*it] = i;

          SSTMAC_DEBUG2 << "destination " << (*it) << " is off of router " << i
                << "\n";
        }
      }

      built_rtable_ = true;
    }

    tff_router_djikstra::tff_router_djikstra()
    {
        built_rtable_ = false;
    }

    void
    tff_router_djikstra::init_factory_params(const sim_parameters::ptr &params)
    {
        tff_router::init_factory_params(params);
        default_algo_ = routing_info::basic;
        algos_[routing_info::basic] = basic_routing::construct(params);
    }

  }
}
