#include <sstmac/hardware/network/topology/routing/ugal_routing.h>

namespace sstmac {
    namespace hw {

SSTRegister("ugal", routing_algorithm, ugal_routing,
            "router implementing ugal congestion-aware routing");

ugal_routing::ugal_routing(int thr)
  : val_threshold_(thr)
{
}

void
ugal_routing::init_routing(const interconnect_message::ptr &msg) const
{
    routing_info::ptr rinfo
        = ugal_routing_info::construct();
    msg->init_routing(rinfo);
}


int
ugal_routing::initial_step(
    const valiant_routing_info::ptr& rinfo,
    const interconnect_message::ptr& msg)
{
    int dim, dir;
    nodeaddress::ptr ej_addr = regtop_->node_to_ejector_addr(msg->toaddr(), dir);
    if (ej_addr->equals(sw_->get_addr())){
        return router_->convert_to_outport(topology::EJECT, dir);
    }

    coordinates src = regtop_->get_switch_coords(sw_->get_addr());
    coordinates dst = regtop_->get_switch_coords(ej_addr);
    int min_dst = regtop_->minimal_distance(src, dst);
    if (min_dst <= val_threshold_){
        //too close - ignore valiant
        router_->minimal_route_to_switch(ej_addr, dim, dir);
        return router_->convert_to_outport(dim, dir);
    }

    regtop_->minimal_route_to_coords(src, dst, dim, dir);
    int min_port = router_->convert_to_outport(dim, dir);
    int min_queue_length = sw_->get_queue_length(min_port);

    //select an intermediate node
    nodeaddress::ptr inter_addr = regtop_->get_random_intermediate_switch(sw_->get_addr(), msg);
    coordinates inter = regtop_->get_switch_coords(inter_addr);


    int valiant_dst = regtop_->minimal_distance(src, inter) + regtop_->minimal_distance(inter, dst);


    regtop_->minimal_route_to_coords(src, inter, dim, dir);
    int valiant_port = router_->convert_to_outport(dim, dir);
    int valiant_queue_length = sw_->get_queue_length(valiant_port);

    SSTMAC_DEBUG << sstprintf("UGAL routing: min=%d x %d to sw %ld, valiant=%d x %d to sw %ld\n",
                              min_dst, min_queue_length, ej_addr->unique_id(),
                              valiant_dst, valiant_queue_length, inter_addr->unique_id());

    int minimal_weight = min_queue_length * min_dst;
    int valiant_weight = valiant_queue_length * valiant_dst;

    if (minimal_weight <= valiant_weight){
        //keep routing minimally
        SSTMAC_DEBUG << sstprintf("UGAL minimal routing to port %d\n", min_port);
        return min_port;
    }
    else {
        //switch to valiant routing
        rinfo->set_dest_switch(inter_addr);
        rinfo->set_stage(valiant_routing_info::intermediate);
        SSTMAC_DEBUG << sstprintf("UGAL valiant routing to switch %ld, port %d\n",
                                  inter_addr->unique_id(), min_port);
        return valiant_port;
    }
}

    }
}
