#include <sstmac/hardware/network/topology/routing/valiant_routing.h>
#include <sstmac/hardware/network/switches/networkswitch.h>

DeclareSerializable(sstmac::hw::valiant_routing_info)
;

namespace sstmac
{
  namespace hw
  {
    SSTRegister("valiant", routing_algorithm, valiant_routing,
        "router implementing valiant routing");

    void
    valiant_routing_info::serialize_order(sst_serializer *ser)
    {
      ser->do_obj(dest_switch_);
      ser->do_enum(stage_);
    }

    valiant_routing::valiant_routing()
    {
    }

    void
    valiant_routing::init_routing(const interconnect_message::ptr &msg) const
    {
      routing_info::ptr rinfo = valiant_routing_info::construct();
      msg->init_routing(rinfo);
    }

    int
    valiant_routing::initial_step(const valiant_routing_info::ptr& rinfo,
        const interconnect_message::ptr& msg)
    {
      SSTMAC_DEBUG << "valiant: initial routing \n";
      nodeaddress::ptr middle_switch =
          sw_->get_topology()->get_random_intermediate_switch(sw_->get_addr(),
              msg);
      rinfo->set_dest_switch(middle_switch);
      rinfo->set_stage(valiant_routing_info::intermediate);
      return intermediate_step(rinfo, msg);
    }

    int
    valiant_routing::intermediate_step(const valiant_routing_info::ptr& rinfo,
        const interconnect_message::ptr& msg)
    {
      SSTMAC_DEBUG << "valiant: intermediate routing \n";
      if (!rinfo->dest_switch()->equals(sw_->get_addr()))
      {
        int dim, dir;
        router_->minimal_route_to_switch(rinfo->dest_switch(), dim, dir);
        return router_->convert_to_outport(dim, dir);
      }
      else
      {
        rinfo->set_stage(valiant_routing_info::final);
        //proceed to final destination
        return final_step(rinfo, msg);
      }
    }

    int
    valiant_routing::final_step(const valiant_routing_info::ptr& rinfo,
        const interconnect_message::ptr& msg)
    {
      SSTMAC_DEBUG << "valiant: final routing \n";
      int dim, dir;
      router_->minimal_route_to_node(msg->toaddr(), dim, dir);
      return router_->convert_to_outport(dim, dir);
    }

    int
    valiant_routing::route(const interconnect_message::ptr& msg)
    {
      valiant_routing_info::ptr rinfo =
          safe_cast(msg->rinfo(), valiant_routing_info);

      switch (rinfo->stage())
      {
      case valiant_routing_info::initial:
      {
        return initial_step(rinfo, msg);
      }
      case valiant_routing_info::intermediate:
      {
        return intermediate_step(rinfo, msg);
      }
      case valiant_routing_info::final:
      {
        return final_step(rinfo, msg);
      }
      default:
        sst_throw(valueerror, "unknown valiant routing stage %d", rinfo->stage());
      }
    }

    int
    valiant_routing::get_number_paths(const interconnect_message::ptr& msg) const
    {
      valiant_routing_info::ptr rinfo =
          safe_cast(msg->rinfo(), valiant_routing_info);

      switch (rinfo->stage())
      {
      case valiant_routing_info::initial:
      {
        int ret = sw_->get_topology()->num_switches() - 1;
        SSTMAC_DEBUG << "valiant::get_number_paths() - returning" << ret
              << "\n";
        return ret;
      }

      default:
      {
        SSTMAC_DEBUG << "valiant::get_number_paths() - returning 1 \n";
        return 1;
      }
      }
    }

  }
}

