/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
// top_from_file.cc: implements an arbitrary topology read from file
//
// Author: Gilbert Hendry <ghendry@sandia.gov>
// 

#include <sstmac/hardware/network/topology/top_from_file.h>
#include <sstmac/hardware/network/topology/routing/tff_router.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/fileio.h>
#include <deque>
//#include "tff_router.h"

namespace sstmac
{
  namespace hw
  {

    using namespace sstmac;
    using namespace sstmac::hw;
    using namespace std;
    SSTRegister("top_from_file", topology, top_from_file,
        "top_from_file implements an arbitrary topology read from file");

    // --------------------------------------------------------------------------------------------------------
    /*  Generates a high dimensional torus network
     Inputs:
     dimensionarray: specifies the number of processors in each dimension
     and whether  the links are wrapped around  (torus) or not (mesh).
     dimensionarray[i] <0 refers to absence of  wrap around edges.
     dimensionarray[i] >0 refers to presence of wrap around edges.

     nndpersw is the number of nodes per switch.
     widtharrg specifies the width between the switches and the  nodes.
     */
    top_from_file::top_from_file(const std::string fname) :
      topology()
    {
      fname_ = fname;
      num_switches_ = -1; //hasn't been initialized yet.
      parsed_node_conn_ = false;
      parsed_switch_conn_ = false;
    }

    int
    top_from_file::nps(nodeaddress::ptr addr) const
    {
      if (nconnections_.find(addr->unique_id()) == nconnections_.end())
      {
        sst_throw(valueerror, "top_from_file::nps - can't find addr %d in nconnections", addr->unique_id());
      }
      return nconnections_[addr->unique_id()].size();
    }

    int
    top_from_file::node_to_switch_port(long nid) const
    {
      /*if (node_connections_.find(nid) == node_connections_.end())
      {
        sst_throw(valueerror, "top_from_file::node_to_switch_port - can't find nid %d in node_connections", nid);
      }
      long sw = node_connections_[nid];

      for (int i = 0; i < nconnections_[sw].size(); i++)
      {
        if (nconnections_[sw][i] == nid)
        {
          return i;
        }
      }*/
      return nid;

      sst_throw(valueerror, "top_from_file::node_to_switch_port - can't find nid %d in nconnections");
    }

    topology_descr::ptr
    top_from_file::get_descr(const nodeaddress::ptr &addr) const
    {
        sst_throw(unimplementederror,
            "top_from_file::get_descr");
    }

    void
    top_from_file::minimal_route_to_switch(
        const nodeaddress::ptr &current_sw_addr,
        const nodeaddress::ptr &dest_sw_addr,
        int &dim, int &dir) const
    {
        sst_throw(unimplementederror, "top_from_file::minimal_route_to_switch");
    }

      int
      top_from_file::convert_to_outport(int dim, int dir)
      {
        if (dim == sstmac::hw::topology::EJECT)
        {
          return -(1 + dir);
        }
        return dir;

      }

      void
      top_from_file::convert_to_dimdir(int outport, int &dim, int&dir)
      {
        if (outport < 0)
        {
          dim = sstmac::hw::topology::EJECT;
          dir = (-outport) - 1;
        }
        else
        {
          dim = 0;
          dir = outport;
        }
      }

    void
    top_from_file::get_send_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr &src,
        std::vector<nodeaddress::ptr>& partners) const
    {
        sst_throw(unimplementederror,
            "top_from_file::get_recv_partners:"
            " don't know how to do regular traffic patterns"
            " on an irregular topology");
    }

    void
    top_from_file::get_recv_partners(
        traffic_pattern::type_t ty,
        const nodeaddress::ptr &src,
        std::vector<nodeaddress::ptr>& partners) const
    {
        sst_throw(unimplementederror,
            "top_from_file::get_send_partners:"
            " don't know how to do regular traffic patterns"
            " on an irregular topology");
    }

    void
    top_from_file::find_num_nodes() const
    {
      if (!parsed_node_conn_)
      {
        std::ifstream in;
        SSTFileIO::open_file(in, fname_);

        if (in.is_open())
        {
          deque<std::string> tok;

          std::string space = " ";

          std::string line;

          boost::unordered_map<long, int> connections;

          while (in.good())
          {

            //first, take out the num of ranks
            std::getline(in, line);
            line = trim_str(line);

            if (line != "")
            {
              tok.clear();
              pst::BasicStringTokenizer::tokenize(line, tok, space);

              SSTMAC_DEBUG2 << "line: " << tok[0] << ", " << tok[1] << "\n";

              if (tok[0].find("router") != string::npos
                  && tok[1].find("router") != string::npos)
              {

              }
              else
              {

                long id0, id1;
                bool isr0 = false, isr1 = false;

                if (tok[0].find("router") != string::npos)
                {
                  id0 = atol(tok[0].substr(6).c_str());
                  isr0 = true;
                }
                else
                {
                  id0 = atol(tok[0].substr(8).c_str());
                }

                if (tok[1].find("router") != string::npos)
                {
                  id1 = atol(tok[1].substr(6).c_str());
                  isr1 = true;
                }
                else
                {
                  id1 = atol(tok[1].substr(8).c_str());
                }

                if (!isr0 && !isr1)
                {
                  sst_throw(ssterror,
                      "top_from_file::find_num_switches() - can't connect a node to a node");
                }

                if (isr0)
                {
                  node_connections_[id1] = id0;
                  SSTMAC_DEBUG << "computer " << id1 << " connected to switch "
                        << id0 << "\n";
                  nconnections_[id0].push_back(id1);
                }
                else
                {
                  node_connections_[id0] = id1;
                  SSTMAC_DEBUG << "computer " << id0 << " connected to switch "
                        << id1 << "\n";
                  nconnections_[id1].push_back(id0);
                }

              }

            }
          }

        }
        else
        {
          sst_throw(ssterror,
              "top_from_file::find_num_switches() - can't find file %s",
              fname_.c_str());
        }

        parsed_node_conn_ = true;
      }
    }

    void
    top_from_file::find_num_switches() const
    {
      if (!parsed_switch_conn_)
      {

        std::ifstream in;
        SSTFileIO::open_file(in, fname_);

        if (in.is_open())
        {
          deque<std::string> tok;

          std::string space = " ";

          std::string line;

          //boost::unordered_map<long, int> connections;

          while (in.good())
          {

            //first, take out the num of ranks
            std::getline(in, line);
            line = trim_str(line);

            if (line != "")
            {
              tok.clear();
              pst::BasicStringTokenizer::tokenize(line, tok, space);

              SSTMAC_DEBUG2 << "line: " << tok[0] << ", " << tok[1] << "\n";

              if (tok[0].find("router") != string::npos
                  && tok[1].find("router") != string::npos)
              {
                long r0 = atol(tok[0].substr(6).c_str());
                long r1 = atol(tok[1].substr(6).c_str());
                SSTMAC_DEBUG << "connecting router " << r0 << " to router "
                      << r1;

                int found = -1;
                for (int i = 0; i < rconnections_[r0].size(); i++)
                {
                  if (rconnections_[r0][i].id_ == r1)
                  {
                    found = i;
                    break;
                  }

                }

                if (found >= 0)
                {
                  rconnections_[r0][found].red_++;
                  SSTMAC_DEBUG << " with redundancy of "
                        << rconnections_[r0][found].red_;
                }
                else
                {
                  rconnections_[r0].push_back(connection(r1));
                  SSTMAC_DEBUG << " with redundancy of 1";
                }

                found = -1;
                for (int i = 0; i < rconnections_[r1].size(); i++)
                {
                  if (rconnections_[r1][i].id_ == r0)
                  {
                    found = i;
                    break;
                  }

                }

                if (found >= 0)
                {
                  rconnections_[r1][found].red_++;
                  SSTMAC_DEBUG << " and " << rconnections_[r1][found].red_
                        << "\n";
                }
                else
                {
                  rconnections_[r1].push_back(connection(r0));
                  SSTMAC_DEBUG << " and 1 \n";
                }

              }

            }
          }

        }
        else
        {
          sst_throw(ssterror,
              "top_from_file::connect_train_switches() - can't find file %s",
              fname_.c_str());
        }

        parsed_switch_conn_ = true;
      }
    }

    nodeaddress::ptr
    top_from_file::node_to_ejector_addr(const nodeaddress::ptr &nodeaddr,
        int &switch_port) const
    {
      long nid = get_switch_ejecting_to_node(nodeaddr);
      switch_port = nodeaddr->unique_id();
      return nodeaddress::convert_from_uniqueid(nid);
    }

    nodeaddress::ptr
    top_from_file::node_to_injector_addr(const nodeaddress::ptr &nodeaddr,
        int &switch_port) const
    {
      long nid = get_switch_injecting_from_node(nodeaddr);
      switch_port = nodeaddr->unique_id();
      return nodeaddress::convert_from_uniqueid(nid);
    }

    long
    top_from_file::get_switch_injecting_from_node(nodeaddress::ptr addr) const
    {
      find_num_nodes();

      node_connection_map::const_iterator it = node_connections_.find(
          addr->unique_id());
      if (it == node_connections_.end())
      {

        sst_throw(ssterror, "top_from_file::get_switch_injecting_from_node() can't find node %d", addr->unique_id());

      }
      return it->second;
    }

    long
    top_from_file::get_switch_ejecting_to_node(nodeaddress::ptr addr) const
    {
      find_num_nodes();
      nconnection_map::const_iterator it, end = nconnections_.end();
      for (it = nconnections_.begin(); it != end; it++)
      {
        std::vector<long>::const_iterator it2, end2 = it->second.end();
        for (it2 = it->second.begin(); it2 != end2; it2++)
        {
          if (addr->unique_id() == *it2)
          {
            return it->first;
          }
        }
      }

      sst_throw(ssterror, "top_from_file::get_switch_ejecting_to_node() can't find node %d", addr->unique_id());

    }

    static std::string
    get_link_name(long from, long to)
    {
      std::stringstream sstr;
      sstr << "[";

      sstr << sstprintf(" %2d", from);

      sstr << " ]->[";

      sstr << sstprintf(" %2d", to);

      sstr << "]";
      return sstr.str();
    }

    void
    top_from_file::connect_switches(maptype& switches)
    {

      if (red_.size() != 0)
      {
        sst_throw(ssterror,
            "top_from_file::connect_train_tracks: "
            " topology_redundant parameter doesn't make sense with this topology."
            " just so you know to take it out.");
      }

      double link_weight = 1.0;

      find_num_nodes();
      find_num_switches();

      boost::unordered_map<long, std::vector<connection> >::iterator it, end =
          rconnections_.end();
      for (it = rconnections_.begin(); it != end; it++)
      {
        nodeaddress::ptr addr0 = nodeaddress::convert_from_uniqueid(it->first);

        maptype::iterator it1 = switches.find(addr0);

        if (it1 == switches.end())
        {
          sst_throw(
              ssterror,
              "top_from_file::connect_train_tracks(): can't find switch 1 when connecting:   uid=%d  addr=%s",
              it->first, addr0->to_string().c_str());
        }

        for (int j = 0; j < it->second.size(); j++)
        {
          nodeaddress::ptr addr1 = nodeaddress::convert_from_uniqueid(
              it->second[j].id_);
          maptype::iterator it2 = switches.find(addr1);

          if (it2 == switches.end())
          {
            sst_throw(
                ssterror,
                "top_from_file::connect_train_tracks(): can't find switch 2 when connecting uid=%d  addr=%s",
                it->second[j].id_, addr1->to_string().c_str());
          }

          switches[addr0]->connect(0, it->second[j].id_,
              connectable::network_link, it2->second, link_weight,
              it->second[j].red_);
        }

      }

      maptype::iterator it3, end3 = switches.end();
      for (it3 = switches.begin(); it3 != end3; it3++)
      {
        tff_router::ptr tr = boost::dynamic_pointer_cast<tff_router>(
            it3->second->get_router());
        tr->init_maps(node_connections_, rconnections_, nconnections_);
      }

    }

  }

}
