/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_circuitnic_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_circuitnic_H_INCLUDED

#include <sstmac/software/services/service.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/hardware/network/switches/circuit/circuitmessage.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <boost/unordered_map.hpp>

namespace sstmac
{
  namespace hw
  {

    class circuitnic : public networkinterface
    {

    public:

      typedef boost::intrusive_ptr<circuitnic> ptr;

    protected:
      /// Hi.
      circuitnic(const sim_parameters::ptr &params);

      timestamp propdelay_;
      double bw_;

      struct job
      {
        nodeaddress::ptr recver;
        sst_message::ptr payload;
        timestamp arrived;
      };

      std::queue<job> jobs_;
      bool busy_;
      bool firstsendset_;

      timestamp setup_timeout_;
      circuitmessage::ptr current_;

      typedef boost::unordered_map<nodeaddress::ptr, circuitnic::ptr> map_type;

      static map_type node_map_;

      void
      check_jobs();

      static long got_datas_;
      static long got_path_acks_;
      static long got_setups_;
      static long got_teardowns_;
      static long numsends_;
      static bool printed_;
      static long timeouts_;

    private:
      void
      send_to_network_link(const sst_message::ptr& msg,
          const nodeaddress::ptr& recver, const timestamp &arrived);

      void
      recv(const sst_message::ptr& netmsg);

    public:
      virtual void
      init(const node::ptr &parent,
          const sstmac::hw::interconnect::ptr &interconn);

      /// Hello.
      static circuitnic::ptr
      construct(const sim_parameters::ptr& params)
      {
        return circuitnic::ptr(new circuitnic(params));
      }

      virtual void
      set_eventmanager(const eventmanager::ptr &m)
      {
        networkinterface::set_eventmanager(m);
        m->register_finish(ptr(this));
      }

      /// Goodbye.
      virtual
      ~circuitnic() throw ()
      {
      }

      virtual void
      finish();

      /// Schedule a send operation to happen at the given time.
      /// \param start_time is the time when the send is started.
      /// \param ppidarg is the parallel process id of the source and target.
      /// \param recver is the global index of the receiver node.
      /// \param payload is the data sent over the network
      /// \param completion is the eventhandler signaled when the send is complete
      virtual void
      send(const sst_message::ptr &payload);

      virtual networkinterface::ptr
      clone(const node::ptr &parent,
          const sstmac::hw::interconnect::ptr &interconn,
          const sim_parameters::ptr& params) const;


      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      to_string() const
      {
        return "circuitnic(" + parent_->get_id()->to_string() + ")";
      }

      void
      timeout(sst_message::ptr msg);

      void
      send_out_resetup(const sst_message::ptr &msg);

      bool
      msg_supported(const sst_message::ptr& msg) const;

      void
      newmsg(const sst_message::ptr &m)
      {
        busy_ = false;
        check_jobs();
      }
    };

  }
} // end of namespace sstmac.

#endif
