/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_NETWORKINTERFACE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NIC_NETWORKINTERFACE_H_INCLUDED

#include <sstmac/hardware/node/node.h>
#include <sstmac/common/timestamp.h>

#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/failure/failable.h>
#include <sstmac/common/stats/stats_common.h>
#include <sstmac/common/factories/factory.h>

#include <sstmac/hardware/node/node.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/hardware/common/component.h>

namespace sstmac
{
  namespace hw
  {

    //class interconnect;
    //typedef boost::intrusive_ptr<interconnect> interconnectPtr;

    /**
     * A networkinterface is a delegate between a node and a server module.
     * This object helps ornament network operations with information about
     * the process (ppid) involved.
     */
    class networkinterface :
            public component,
            public factory_type,
            public failable
    {

    protected:
      networkinterface();

      static debug dbg_;

      eventhandler::ptr injector_;

      interconnect::ptr interconn_;

      node::ptr parent_;

      sw::lib_compute_memmove::ptr lib_mem_;

      void finish_recv_ack(const sst_message::ptr& msg);

      void finish_recv_req(const sst_message::ptr& msg);


    public:
      typedef boost::intrusive_ptr<networkinterface> ptr;

      virtual void
      init(const node::ptr &parent,
          const interconnect::ptr &interconn);

      virtual void
      validate_msg_supported(const sst_message::ptr& msg) const;

      nodeaddress::ptr
      get_parent_id() const
      {
        return parent_->get_id();
      }

      node::ptr
      get_parent_node() const
      {
        return parent_;
      }

      interconnect::ptr
      get_interconnect() const {
        return interconn_;
      }

      void connect(connection_type_t ty, const connectable::ptr &mod);

      static void clear_statics();

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      virtual long
      failure_id() const;

      /// Goodbye.
      virtual ~networkinterface();

      /// Schedule a send operation to happen at the given time.
      /// \param start_time is the time when the send is started.
      /// \param ppidarg is the parallel process id of the source and target.
      /// \param recver is the global index of the receiver node.
      /// \param payload is the data sent over the network
      /// \param completion is the eventhandler signaled when the send is complete
      virtual void
      send(const sst_message::ptr& payload) = 0;

      virtual void
      recv(const sst_message::ptr& payload) = 0;

      virtual void
      handle(const sst_message::ptr& netmsg);

      /// Get a new nic of the same type as the current one.
      virtual networkinterface::ptr
      clone(const node::ptr& parent,
            const interconnect::ptr& interconn,
            const sim_parameters::ptr& params) const = 0;

      virtual std::string
      to_string() const
      {
        return "nic";
      }

      sw::lib_compute_memmove::ptr
      get_lib_mem() const
      {
        return lib_mem_;
      }

      virtual void finish();

    protected:

      static stat_spyplot::ptr spy_;
      static stat_spyplot::ptr spy_bytes_;
      static stat_spyplot_png::ptr spy_bytes_alloc_png_;
      static stat_spyplot_png::ptr spy_bytes_machine_png_;
      static stat_histogram<long>::ptr hist_msg_size_;
      static stat_sum::ptr total_lat_;
    };

    DeclareFactory2InitParams(networkinterface, node::ptr, interconnect::ptr);

  }
} // end of namespace sstmac.

#endif
