/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/nic/simplenic.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/process/waiter.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/hardware/network/switches/entryexit.h>

namespace sstmac
{
  namespace hw
  {

    SSTRegister("simplenic | simple", networkinterface, simplenic,
                "Basic NIC. Performs basic memory reads and does simple injection of packets.");
    // Hello.
    //
    simplenic::ptr
    simplenic::construct(const sim_parameters::ptr& params)
    {
      return simplenic::ptr(new simplenic());
    }

    //
    // Goodbye.
    //
    simplenic::~simplenic() throw ()
    {
    }

    void
    simplenic::send(const sst_message::ptr& payload)
    {
      use_me();

      validate_msg_supported(payload);

      if (spy_){
        spy_->add_one(payload->fromaddr()->unique_id(),
              payload->toaddr()->unique_id());
        spy_bytes_->add(payload->fromaddr()->unique_id(),
              payload->toaddr()->unique_id(), payload->get_byte_length());
        spy_bytes_alloc_png_->add(payload->fromaddr()->unique_id(),
                                payload->toaddr()->unique_id(), payload->get_byte_length());
        spy_bytes_machine_png_->add(payload->fromaddr()->unique_id(),
                                  payload->toaddr()->unique_id(), payload->get_byte_length());
      }
      if (hist_msg_size_){
        hist_msg_size_->add_one(payload->get_byte_length());
      }

      injector::ptr inj = safe_cast(injector_, injector);
      event_step_callback::ptr cb = new_event_step_callback(inj, &injector::send, payload);

      lib_mem_->read(payload->get_byte_length(), payload->source_task(), cb);
    }

    networkinterface::ptr
    simplenic::clone(const node::ptr &parent,
        const sstmac::hw::interconnect::ptr &interconn,
        const sim_parameters::ptr& params) const
    {
      simplenic::ptr ret = simplenic::construct(params);
      ret->init(parent, interconn);
      return ret;
    }

    void
    simplenic::recv(const sst_message::ptr& netmsg)
    {
      sst_message::ptr payload = netmsg->get_parent_msg();
      eventcallback::ptr cb = netmsg->is_tail()
          ? new_callback(ptr(this), &simplenic::send_to_os, payload)
          : donothingcallback::construct();
      lib_mem_->write(netmsg->get_byte_length(), payload->dest_task(), cb);
    }

    void
    simplenic::send_to_os(const sst_message::ptr &msg)
    {
      parent_->handle(msg);
    }

  }
} // end of namespace sstmac.
