/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED
#define SSTMAC_BACKENDS_NATIVE_COMPONENTS_NODE_NODE_H_INCLUDED

#include <sstmac/common/timestamp.h>

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/eventmodule.h>
#include <sstmac/software/ami/ami.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/hardware/failure/failable.h>


namespace sstmac
{
  namespace sw
  {
    class operatingsystem;
    class launcher;
    class environment;
    class service;
  }

  namespace sw {
    class operatingsystem;
    typedef boost::intrusive_ptr<operatingsystem> operatingsystemPtr;
    class environment;
    typedef boost::intrusive_ptr<environment> environmentPtr;
    class launcher;
    typedef boost::intrusive_ptr<launcher> launcherPtr;
    class service;
    typedef boost::intrusive_ptr<service> servicePtr;
  }

  namespace hw
  {

    // Forward declarations of relevant types.
    class networkinterface;
    typedef boost::intrusive_ptr<networkinterface> networkinterfacePtr;
    class memorymodel;
    typedef boost::intrusive_ptr<memorymodel> memorymodelPtr;

    /**
     * An interface for "physical" nodes -- may eventually be decorated
     * with additional functionality (OS/cpu modeling or whatever).
     */
    class node :
            public eventmodule,
            public factory_type,
            public sstmac::sw::AMI,
            public failable
    {

      friend class message_handler;

    public:
      typedef boost::intrusive_ptr<node> ptr;

    protected:
      class message_handler : public eventhandler
      {
      protected:
        node::ptr parent_;

        message_handler(const node::ptr &p) :
          parent_(p)
        {
        }

      public:
        typedef boost::intrusive_ptr<message_handler> ptr;

        virtual void
        handle(const sst_message::ptr &msg);

        static ptr
        construct(const node::ptr &p)
        {
          return new message_handler(p);
        }

        virtual std::string
        to_string() const
        {
          return "node::message_handler";
        }
      };

      virtual bool
      has_failed();
      virtual void
      reset_failed();

      message_handler::ptr msg_handler_;

      sw::environmentPtr env_;

      sw::operatingsystemPtr os_;

      int nsockets_;

      int ncores_;

      /// The id for this node.
      nodeaddress::ptr nodeid_;

      sw::launcherPtr launcher_;

      static debug dbg_;

      sim_parameters::ptr params_;

      node(const sim_parameters::ptr &params);

      node(const sim_parameters::ptr &params, const nodeaddress::ptr& my_id);

      virtual ~node();

      static sstmac::sw::allocationstrategy::ptr allocation_;

      networkinterfacePtr nic_;

      memorymodelPtr mem_model_;

    public:
      /// Get a new node of the same type as the current one.
      virtual node::ptr
      clone(nodeaddress::ptr my_id) const = 0;

      /// Assign a node index and event call interface for this node.
      /// \throw nullerror if ev is null.
      virtual void
      initialize(eventmanager::ptr man, const topology::ptr& top);

      void init_factory_params(const sim_parameters::ptr &params);

      void
      set_nic(const networkinterfacePtr &n);

      int nsockets() const {
          return nsockets_;
      }

      eventhandler::ptr
      get_msg_handler() const;

      networkinterfacePtr
      get_nic() const;

      memorymodelPtr
      get_mem() const;

      sw::operatingsystemPtr
      get_os() const;

      virtual void
      handle(const sst_message::ptr &msg);

      virtual std::string
      to_string() const;

      virtual void finish();

      virtual long
      failure_id() const
      {
        return nodeid_->unique_id();
      }

      nodeaddress::ptr
      get_id() const
      {
        return nodeid_;
      }

      virtual std::list<failable::ptr>
      get_failure_components() = 0;

      void
      add_service(const sw::servicePtr serv);

      //virtual void
      //init_failure_model(double l_proc, double l_mem, double l_disk, double l_nic, failable::ptr ps, double p_ps_node) = 0;

    };

    DeclareFactory(node);

  }
} // end of namespace sstmac

#endif
