/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/simpleprocessor-config.h>
#include <sstmac/hardware/processor/instructionprocessor-config.h>
#include <sstmac/hardware/memory/simple_memorymodel.h>
#include <sstmac/hardware/memory/numa_memorymodel.h>
#include <sstmac/hardware/memory/train_memorymodel.h>
#include <sstmac/hardware/node/nullnode.h>
#include <sstmac/hardware/processor/nullscheduler.h>
#include <sstmac/hardware/nic/networkinterface.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <iostream>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac
{

  namespace hw
  {
    SSTRegister("null | nullnode", node, nullnode,
        "Node which implements interface, but does no compute modeling");

    using namespace sstmac::sw;

    nullnode::nullnode(const sim_parameters::ptr &params) :
      node(params)
    {

    }

    nullnode::nullnode(const sim_parameters::ptr &params,
        const nodeaddress::ptr& my_id) :
      node(params, my_id)
    {
      os_->setAMI(sstmac::sw::AMI::AMI_COMM, this);
      os_->setAMI(sstmac::sw::AMI::AMI_COMPUTE, this);
      os_->setAMI(sstmac::sw::AMI::AMI_QUERY, this);
      os_->setAMI(sstmac::sw::AMI::AMI_HW, this);
    }

    void
    nullnode::init_factory_params(const sim_parameters::ptr &params)
    {
      //fake template node
      node::init_factory_params(params);
      if (!nodeid_)
        return;

      /**
       sstkeyword {
       gui=null;
       }
       */
      mem_model_ = memorymodel_factory::get_optional_param("node_memory_model",
          "null", params_, ptr(this), computescheduler::ptr());
      compute_sched_ = nullscheduler::construct(
          nodeid_->unique_id(), ncores_, nsockets_, timestamp(0), os_);

    }

    void
    nullnode::initialize(eventmanager::ptr man, const topology::ptr& top)
    {
      node::initialize(man, top);

      mem_model_->set_eventmanager(man);

      os_->start_apps();
      launcher_->start();
    }

    node::ptr
    nullnode::clone(nodeaddress::ptr my_id) const
    {
      node::ptr new_node = new nullnode(params_, my_id);
      new_node->init_factory_params(params_);
      return new_node;
    }

    void
    nullnode::finish()
    {
      if (finished_)
        return;

      node::finish();

      if (mem_model_)
      {
        mem_model_->finish();
        mem_model_ = 0;
      }
    }

    void
    nullnode::execute_kernel(AMI::AMI_COMM_FUNC func,
        const sst_message::ptr& data)
    {
      bool failed = false;

      SSTMAC_DEBUG << sstprintf("nullnode::execute_kernel: %s for message %s\n",
                              AMI::tostr(func),
                              data->to_string().c_str());

      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMM_SEND:
      {
        data->set_fromaddr(nodeid_);

        SSTMAC_DEBUG << "Node(" << nodeid_->to_string() << "): sending to "
              << data->toaddr()->to_string() << "\n";

        if (data->toaddr()->equals(nodeid_))
        {
          timestamp slight_delay(1e-9);
          timestamp now_plus_delay = now() + slight_delay;
          if (data->get_needs_ack())
          {
            sst_message::ptr ack = data->clone(sst_message::SEND_ACK);
            schedule(now_plus_delay, msg_handler_, ack);
          }

          //delay a nanonsecond and reschedule
          //nanosecond insures the msg_handler comes AFTER
          //any blocking operation that might occur on execute_kernel finishing
          schedule(now_plus_delay, msg_handler_, data);
        }
        else {
          nic_->send(data);
        }

        failed = nic_->has_failed() || mem_model_->has_failed();
        break;
      }
      default:
        sst_throw(ssterror, "nullnode: cannot process kernel %s", AMI::tostr(func));
        break;
      }
    }

    void
    nullnode::execute_kernel(AMI::AMI_COMP_FUNC func,
        const sst_message::ptr& data)
    {
      SSTMAC_DEBUG << sstprintf("nullnode::execute_kernel: %s for message %s\n",
                              AMI::tostr(func),
                              data->to_string().c_str());

      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMP_SLEEP:
      {
        timestamp done = now() + safe_cast(data,timed_message)->get_time();
        SSTMAC_DEBUG << "node (" << nodeid_->to_string()
              << "): sleeping until " << done << "\n";
        schedule(done, ptr(this), data);
        break;
      }
      case sstmac::sw::AMI::AMI_COMP_TIME:
      case sstmac::sw::AMI::AMI_COMP_MEMREAD:
      case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
      case sstmac::sw::AMI::AMI_COMP_INSTR:
      {
        //schedule done with a minimal delay
        timestamp done = now() + timestamp(1e-9);
        schedule(done, ptr(this), data);
        break;
      }
      case sstmac::sw::AMI::AMI_COMP_INSTR_TIME:
      {
        data->set_event_time(timestamp(0));
        break;
      }
      case sstmac::sw::AMI::AMI_COMP_SCHEDULE:
      {
        eventhandler::ptr handle = data->pop_callback();
        if (!handle){
            sst_throw(illformederror, "compute schedule reserve only works with callbacks");
        }
        handle->handle(data);
        break;
      }
      case sstmac::sw::AMI::AMI_COMP_RELEASE: //no work to do
        break;
      default:
        sst_throw(ssterror, "nullnode: cannot process kernel %s", AMI::tostr(func));
      }
    }

    bool
    nullnode::has_failed()
    {
      return nic_->has_failed() || mem_model_->has_failed() || failable::has_failed();
    }
    void
    nullnode::reset_failed()
    {
      nic_->reset_failed();
      mem_model_->reset_failed();
      failable::reset_failed();
    }

    ptr_type::ptr
    nullnode::execute_kernel(AMI::AMI_HW_FUNC func)
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_CATASTROPHIC_INTERRUPT:
      {
        cancel_all_messages();
        mem_model_->cancel_all_messages();
        nic_->cancel_all_messages();
        return ptr_type::ptr();
        break;
      }
      default:
        sst_throw(valueerror, "nullnode: cannot process kernel %s", tostr(func));
      }
    }

    ptr_type::ptr
    nullnode::execute_kernel(AMI::AMI_QUERY_FUNC func)
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
      {
        return nodeid_;
        break;
      }
      default:
      {
      sst_throw(valueerror, "nullnode: cannot process kernel %s", tostr(func));
      }
      }
      return ptr_type::ptr();
    }

    timestamp
    nullnode::now() const
    {
      return eventmodule::now();
    }

    bool
    nullnode::kernel_supported(AMI::AMI_COMP_FUNC func) const
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMP_TIME:
      case sstmac::sw::AMI::AMI_COMP_SLEEP:
      case sstmac::sw::AMI::AMI_COMP_INSTR:
      case sstmac::sw::AMI::AMI_COMP_MEMREAD:
      case sstmac::sw::AMI::AMI_COMP_SCHEDULE:
      case sstmac::sw::AMI::AMI_COMP_MEMWRITE:
        return true;
      default:
        return false;
      }
      return false;
    }

    bool
    nullnode::kernel_supported(AMI::AMI_COMM_FUNC func) const
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_COMM_SEND:
      case sstmac::sw::AMI::AMI_COMM_NIC_SEND_ACK:
        return true;
      default:
        return false;
      }
      return false;
    }

    bool
    nullnode::kernel_supported(AMI::AMI_HW_FUNC func) const
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_CATASTROPHIC_INTERRUPT:
        return true;
      }
      return false;
    }

    bool
    nullnode::kernel_supported(AMI::AMI_QUERY_FUNC func) const
    {
      switch (func)
      {
      case sstmac::sw::AMI::AMI_QUERY_ADDRESS:
        return true;
      default:
        return false;
      }
      return false;
    }

    std::list<failable::ptr>
    nullnode::get_failure_components()
    {
      std::list<failable::ptr> ret;
      if (mem_model_)
        ret.push_back(mem_model_);
      if (nic_)
        ret.push_back(nic_);

      ptr s(this);
      ret.push_back(s);
      return ret;
    }

  }
} // end of namespace sstmac.

