#ifndef PACKETFLOW_H
#define PACKETFLOW_H

#ifndef PACKET_TRAIN_H
#define PACKET_TRAIN_H

#include <sstmac/hardware/network/interconnect/interconnect_message.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sim_parameters.h>


namespace sstmac
{
  namespace hw
  {



    class packetflow_message :
        virtual public sst_message
    {

        public:
            typedef enum {
                credit,
                payload
            } type_t;

            typedef boost::intrusive_ptr<packetflow_message> ptr;

            virtual type_t type() const = 0;

    };

    /**
     @class packetflow
     Encapsulates a group of machine packets traveling together on the
     same path between endpoints.  This is usually one fraction of
     a larger message.
     */
    class packetflow_payload :
        public interconnect_message,
        public packetflow_message
    {
    public:
      typedef boost::intrusive_ptr<packetflow_payload> ptr;

    protected:
      static const double uninitialized_bw;

      long msg_id_;

      int outport_;

      int inport_;

      double bw_;

      long bytes_arrived_;

      long bytes_credited_;

      long bytes_sent_;

      long byte_offset_;

      double send_done_;

      static long id_count_;

      static long max_num_packets_;

      static long min_num_bytes_per_packet_;

      static long negligible_size_;

      packetflow_payload(
          const sst_message::ptr& parent,
          int msg_id,
          long num_bytes,
          long offset);

    public:
      static ptr
      construct(
          const sst_message::ptr& parent,
          int msg_id,
          long num_bytes,
          long offset){
        return new packetflow_payload(parent, msg_id, num_bytes, offset);
      }

      virtual
      ~packetflow_payload(){}

      type_t type() const {
        return payload;
      }

      /**
       @return The total number of bytes in the complete message
       summed over all trains
       */
      long
      num_bytes_total() const
      {
        return orig_->get_byte_length();
      }

      /**
       @return The minimum number of bytes needed in a train. An 8K
       messages with min_bytes=4K will be split into two
       messages of size 4K.
       */
      static long
      min_num_bytes_per_packet()
      {
        return min_num_bytes_per_packet_;
      }

      /**
       @return The size at which a message is considered
       'negligible' for congestion modeling. This
       message is basically just routed with
       no congestion arbitration.
       */
      bool
      negligible_size() const
      {
        return num_bytes_total() <= negligible_size_;
      }

      /**
       @return Whether the parent message is being sent as a single
       packetflow, i.e. the parent message is smaller
       than the min_bytes per train.
       */
      bool
      is_single_packet() const
      {
        return num_bytes_total() <= min_num_bytes_per_packet_;
      }

      /**
       @param num_bytes The number of bytes sent in a message
       @return Whether num_bytes would be sent as a single packet train
       */
      static bool
      is_single_packet(uint64_t num_bytes)
      {
        return num_bytes <= min_num_bytes_per_packet_;
      }

      static long
      allocate_msg_id()
      {
        long ret = id_count_;
        ++id_count_;
        return ret;
      }

      bool is_tail() const {
        return (byte_offset_ + num_bytes_) == num_bytes_total();
      }

      long
      msg_id() const
      {
        return msg_id_;
      }

      bool
      get_needs_ack() const;

      /**
       @return The number of bytes in this packetflow, NOT
       the total number of bytes in the parent message.
       See #num_bytes_total
       */
      long
      num_bytes() const
      {
        return num_bytes_;
      }

      /**
       @return The bandwidth in number of bytes per second
       */
      double
      bw() const {
        return bw_;
      }

      /**
       @param The bandwidth in number of bytes per second
       */
      void
      set_bw(double bw) {
        bw_ = bw;
      }

      double
      ser_delay() const {
        return num_bytes_ / bw_;
      }

      long
      credits_needed() const {
        return num_bytes_ - bytes_credited_;
      }

      long
      bytes_unsent() const {
        return num_bytes_ - bytes_sent_;
      }

      void
      increment_credits(long num_credits){
        bytes_credited_ += num_credits;
      }

      void
      increment_sent(long num_bytes){
        bytes_sent_ += num_bytes;
      }

      void
      set_outport(int port){
        outport_ = port;
      }

      int
      outport() const {
        return outport_;
      }

      void
      set_inport(int port){
        inport_ = port;
      }

      int
      inport() const {
        return inport_;
      }

      void
      set_send_done(double time){
        send_done_ = std::max(time, send_done_);
      }

      double
      send_done() const {
        return send_done_;
      }

      void
      clear_intermediates();

      std::string
      to_string() const;

      static void
      init_statics(int min_bytes, int max_trains, int negligible_size);

    };

    struct creditflow {
        timestamp t_start;
        double bw;
        long num_bytes;
    };

    class packetflow_credit :
        public packetflow_message
    {
        protected:
            double delta_t_;
            long num_bytes_now_;
            long num_bytes_flow_;
            int port_;

            packetflow_credit(
                int port,
                long num_bytes)
                : port_(port),
                delta_t_(-1),
                num_bytes_now_(num_bytes),
                num_bytes_flow_(0)
            {
            }

            packetflow_credit(
                int port,
                long num_bytes_now,
                long num_bytes_flow,
                double delta_t)
                : port_(port),
                delta_t_(delta_t),
                num_bytes_now_(num_bytes_now),
                num_bytes_flow_(num_bytes_flow)
            {
            }

        public:
            typedef boost::intrusive_ptr<packetflow_credit> ptr;

            static ptr
            construct(int port, double delta_t,
                      long num_bytes_now, long num_bytes_flow){
                return new packetflow_credit(port, num_bytes_now, num_bytes_flow, delta_t);
            }

            static ptr
            construct(int port, long num_bytes){
                return new packetflow_credit(port, num_bytes);
            }

            type_t type() const {
                return credit;
            }

            int port() const {
                return port_;
            }

            long num_bytes_now() const {
                return num_bytes_now_;
            }

            long num_bytes_flow() const {
                return num_bytes_flow_;
            }

            std::string
            to_string() const;

    };


  }
}

#endif // PACKET_TRAIN_H


#endif // PACKETFLOW_H
