#include <sstmac/hardware/packetflow/packetflow_sender.h>

namespace sstmac {
    namespace hw {

debug packetflow_sender::dbg_;

packetflow_sender::packetflow_sender(
    timestamp lat,
    double max_bw,
    const packetflow_handler::ptr& parent)
  : send_lat_(lat),
  max_output_bw_(max_bw),
  parent_(parent),
  total_bytes_available_(0),
  bytes_available_now_(0),
  head_(0),
  port_(0)
{
     if (!dbg_.is_real()){
        dbg_ = debug("<debug> packetflow");
     }
}

void
packetflow_sender::set_eventmanager(const eventmanager::ptr &m)
{
    eventmodule::set_eventmanager(m);

    head_ = new packetflow_epoch;
    packetflow_epoch* next = head_;
    double interval = 2e4;
    double start = now().sec();
    long bytes = interval * max_output_bw_;
    //split one million seconds into ten intervals
    for (int i=1; i < 10; ++i){
        next->start = start;
        next->stop = start + interval;
        next->num_bytes = bytes;
        start += interval;
        next->next = new packetflow_epoch;
        next = next->next;
    }
    next->next = 0;
}

void
packetflow_sender::allocate(const packetflow_payload::ptr& msg)
{
    long credits_allocated = std::min(bytes_available_now_, msg->credits_needed());
    bytes_available_now_ -= credits_allocated;
    total_bytes_available_ -= credits_allocated;
    msg->increment_credits(credits_allocated);

    long bytes_unsent = msg->bytes_unsent();
    //can never send more bytes than we have credits allocated
    long bytes_to_send = std::min(bytes_unsent, credits_allocated);

#define debug_pflow_stats(nbytes) \
    SSTMAC_DEBUG << sstprintf("On sender %s port %d, allocate %ld sent on E(%ld:%8.4e--%8.4e) to {%s}\n", \
                                parent_->to_string().c_str(), \
                                port_, \
                                nbytes, \
                                next->num_bytes, next->start, next->stop, \
                                msg->to_string().c_str())

    packetflow_epoch* next = head_;
    while (next)
    {
        if (next->num_bytes > bytes_to_send)
        {
            debug_pflow_stats(bytes_to_send);
            msg->increment_sent(bytes_to_send);
            double epoch_length = next->stop - next->start;
            double time_to_send = (epoch_length * bytes_to_send) / next->num_bytes;
            next->start += time_to_send;
            next->num_bytes -= bytes_to_send;
            msg->set_send_done(next->start);
            return;
        }
        else if (next->num_bytes == bytes_to_send)
        {
            debug_pflow_stats(bytes_to_send);
            msg->increment_sent(next->num_bytes);
            head_ = next->next;
            msg->set_send_done(next->stop);
            return;
        }
        else
        {
            debug_pflow_stats(bytes_to_send);
            bytes_to_send -= next->num_bytes;
            msg->increment_sent(next->num_bytes);
            //remove the epoch
            head_ = next->next;
            next = head_;
        }
    }
}

void
packetflow_sender::init(long num_credits)
{
    bytes_available_now_ = num_credits;
    total_bytes_available_ = num_credits;
}

void
packetflow_sender::add_credit(const packetflow_credit::ptr &credit)
{
    SSTMAC_DEBUG << sstprintf("On sender %s port %d, received {%s}\n",
                              parent_->to_string().c_str(),
                              port_,
                              credit->to_string().c_str());
    bytes_available_now_ += credit->num_bytes_now();
    total_bytes_available_ += credit->num_bytes_now() + credit->num_bytes_flow();
    if (credit->num_bytes_flow() > 0){
       //add in the credit flow to our various epochs
    }
    if (!queue_.empty())
        do_send();
}

bool
packetflow_sender::try_send()
{
    if (queue_.empty())
        return false;

    packetflow_payload::ptr msg = queue_.front();
    SSTMAC_DEBUG << sstprintf("On sender %s port %d, trying to send {%s} with %ld credits\n",
                               parent_->to_string().c_str(),
                               port_,
                               msg->to_string().c_str(),
                               bytes_available_now_);

    if (msg->credits_needed() > total_bytes_available_){
        allocate(msg);
        return false;
    }
    else {
        queue_.pop_front();
        allocate(msg);
        double now_ = now().sec();
        double msg_bw = msg->num_bytes() / (msg->send_done() - now_);
        msg->set_bw(msg_bw);
        parent_->send_along(msg);
        return true;
    }
}

void
packetflow_sender::update_epochs()
{
    double now_ = now().sec();
    packetflow_epoch* next = head_;
    while(next)
    {
        if (next->stop <= now_)
        {
            head_ = next->next;
            delete next;
            next = head_;
        }
        else if (next->start < now_)
        {
            double epoch_length = next->stop - next->start;
            double remaining = next->stop - now_;
            next->num_bytes *= remaining/epoch_length;
            next->start = now_;
            return; //we are done
        }
        else {
            //the epochs are already in the future
            return;
        }
    }
}

void
packetflow_sender::do_send()
{
    if (queue_.empty()){
        return;
    }

    update_epochs();

    SSTMAC_DEBUG << sstprintf("On sender %s port %d, doing send with %d pending messages and %ld credits\n",
                              parent_->to_string().c_str(),
                              port_,
                              queue_.size(),
                              bytes_available_now_);

    while(try_send());
}

void
packetflow_sender::send(const packetflow_payload::ptr &msg)
{
    msg->clear_intermediates();
    SSTMAC_DEBUG << sstprintf("On sender %s port %d, starting send {%s}\n",
        parent_->to_string().c_str(),
        port_,
        msg->to_string().c_str());
    queue_.push_back(msg);
    if (msg->num_bytes() <= total_bytes_available_){
        do_send();
    }
}


    }
}
