#ifndef packetflow_switch_CC
#define packetflow_switch_CC

/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packetflow/packetflow_switch.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>

namespace sstmac
{

  namespace hw
  {

    SSTRegister("packetflow", networkswitch, packetflow_switch);

    debug packetflow_switch::dbg_;

    template <class T>
    void
    set_ev_mgr(T& themap, const eventmanager::ptr& m){
        typename T::iterator it, end = themap.end();
        for (it=themap.begin(); it != end; ++it){
            it->second->set_eventmanager(m);
        }
    }

    packetflow_switch::packetflow_switch(const packetflow_params::ptr &params)
        : params_(params)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> packetflow | network");
        logger::register_logger(&dbg_);
      }
    }

    packetflow_switch::ptr
    packetflow_switch::construct(const sim_parameters::ptr& params)
    {
       packetflow_params::ptr flow_params = packetflow_params::construct();
       flow_params->link_bw = params->get_bandwidth_param("packetflow_network_link_bandwidth");
       flow_params->link_lat = params->get_time_param("packetflow_network_link_latency");
       flow_params->xbar_output_buffer_num_bytes
           = params->get_byte_length_param("packetflow_switch_output_buffer_size");
       flow_params->crossbar_bw = params->get_bandwidth_param("packetflow_switch_crossbar_bandwidth");
       flow_params->crossbar_lat = params->get_time_param("packetflow_switch_crossbar_latency");
       flow_params->xbar_input_buffer_num_bytes
           = params->get_byte_length_param("packetflow_switch_input_buffer_size");

      /**
       sstkeyword {
       docstring=The minimum number of bytes a single packet train can contain.ENDL
       Raising this value increases the coarseness and lowers the accuracy.;
       };
       */
      int min_bytes =
          params->get_optional_int_param("packetflow_min_num_bytes", 4096);
      /**
       sstkeyword {
       docstring=The maximum number of packet trains a message can be split into.ENDL
       For example, if 1MB is sent and max trains is 100, each packet trainENDL
       will contain 10KB.  If max trains is raised to 1000, then each trainENDL
       can be made smaller to 1KB.;
       }
       */
      int max_trains = params->get_optional_int_param("packetflow_max_num_messages",
          10);
      /**
       sstkeyword {
       docstring=The size below which a train is considered negligible and does notENDL
       do congestion modeling. This increases efficiency, but lowers accuracy.;
       }
       */
      int negligible_size = params->get_optional_int_param(
          "packetflow__negligible_size", 512);

      packetflow_payload::init_statics(min_bytes, max_trains, negligible_size);

      return new packetflow_switch(flow_params);
    }

    void
    packetflow_switch::clone_into(
        const ptr &sw,
        const nodeaddress::ptr &me,
        const router::ptr &route) const
    {
        networkswitch::clone_into(sw, me, route);
    }

    packetflow_switch::~packetflow_switch()
    {
    }

    void
    packetflow_switch::finish()
    {
      if (finished_)
        return;
    }

#if 0
    void
    packetflow_switch::connect_crossbar(const packetflow_crossbar::ptr& buf)
    {
        boost::unordered_map<int, packetflow_simple_buffer::ptr>
            ::iterator it, end = out_buffers_.end();
        for (it = out_buffers_.begin(); it != end; ++it){
            buf->set_output(it->first, it->second);
        }
    }
#endif

    void
    packetflow_switch::initialize()
    {
        descr_ = top_->get_descr(my_addr_);

        xbar_ = packetflow_crossbar::construct(
                    params_->crossbar_lat,
                    params_->crossbar_bw);
        xbar_->set_router(ptr(this));
        xbar_->set_top_descr(descr_);

        //connect all the input buffers to the crossbar
        { boost::unordered_map<int, packetflow_crossbar_buffer::ptr>
            ::iterator it, end = in_buffers_.end();
        for (it = in_buffers_.begin(); it != end; ++it){
            packetflow_crossbar_buffer::ptr buf = it->second;
            int port = xbar_->allocate_inport();
            xbar_->set_input(port, buf);
            buf->set_output(port, xbar_);
            buf->set_top_descr(descr_);
        } }

        { boost::unordered_map<int, packetflow_simple_buffer::ptr>
            ::iterator it, end = out_buffers_.end();
        for (it = out_buffers_.begin(); it != end; ++it){
            packetflow_simple_buffer::ptr buf = it->second;
            int outport = it->first;
            xbar_->set_output(outport, buf);
            buf->set_input(outport, xbar_);
            buf->set_top_descr(descr_);
        } }
    }

    void
    packetflow_switch::connect(int dim, int port_offset, connection_type_t ty,
        const connectable::ptr &mod, double weight, int red)
    {

      int outport = router_->convert_to_outport(dim, port_offset);
      //create an output buffer for the port
      double total_link_bw = params_->link_bw * weight * red;
      packetflow_simple_buffer::ptr out_buffer = packetflow_simple_buffer
        ::construct(total_link_bw,
               params_->link_lat,
               params_->xbar_output_buffer_num_bytes);
      packetflow_switch::ptr neighbor_sw = safe_cast(mod, packetflow_switch);
      int inport = neighbor_sw->get_router()->convert_to_inport(dim, port_offset);
      packetflow_crossbar_buffer::ptr in_buffer = neighbor_sw->get_input_buffer(inport);

      out_buffers_[outport] = out_buffer;

      out_buffer->set_output(0, in_buffer);
      in_buffer->set_input(0, out_buffer);
    }

    void
    packetflow_switch::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);
      set_ev_mgr(out_buffers_, m);
      set_ev_mgr(in_buffers_, m);
      xbar_->set_eventmanager(m);
    }

    int
    packetflow_switch::get_next_port(const packetflow_payload::ptr &msg)
    {
        return router_->route(msg);
    }

    void
    packetflow_switch::eject(const packetflow_payload::ptr &msg)
    {
        ejectors_[msg->toaddr()]->handle(msg);
    }

    void
    packetflow_switch::connect(const nodeaddress::ptr& addr, int port,
        connection_type_t ty, const connectable::ptr& mod)
    {
      connectable::ptr nic = mod;
      switch (ty)
      {
      case injection:
        nic->connect(injection, ptr(this));
        break;
      case ejection:
        nic->connect(ejection, ptr(this));
        break;
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    packetflow_switch::add_ejector(const nodeaddress::ptr& addr,
        const eventmodule::ptr& dest)
    {
      SSTMAC_DEBUG2 << sstprintf("Adding ejector at node %s to switch at %s\n",
            addr->to_string().c_str(), my_addr_->to_string().c_str());
      ejectors_[addr] = dest;
    }


    int
    packetflow_switch::get_queue_length(int port) const
    {
        return 0;
    }

    void
    packetflow_switch::handle(const sst_message::ptr &msg)
    {
        sst_throw(illformederror,
            "packetflow_switch::handle: should not be handling");
    }

    packetflow_crossbar_buffer::ptr
    packetflow_switch::get_input_buffer(int port)
    {
        packetflow_crossbar_buffer::ptr& buf = in_buffers_[port];
        if (!buf){
            buf = packetflow_crossbar_buffer::construct(
                    params_->crossbar_bw,
                    params_->crossbar_lat,
                    params_->xbar_input_buffer_num_bytes);
        }
        return buf;
    }

    std::string
    packetflow_switch::to_string() const
    {
      return "packetflow switch";
    }

  }
}

#endif // packetflow_switch_CC

