#ifndef PACKETFLOW_SWITCH_H
#define PACKETFLOW_SWITCH_H

#ifndef packetflow_switch_H
#define packetflow_switch_H

#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/hardware/network/topology/routing/router.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packetflow/packetflow_buffer.h>
#include <sstmac/hardware/packetflow/packetflow_multiplexer.h>
#include <sstmac/hardware/packetflow/packetflow_crossbar.h>

namespace sstmac
{
  namespace hw
  {

    class packetflow_params :
        public ptr_type
    {
        public:
            typedef boost::intrusive_ptr<packetflow_params> ptr;

            double link_bw;

            timestamp link_lat;

            long xbar_output_buffer_num_bytes;

            double crossbar_bw;

            timestamp crossbar_lat;

            long xbar_input_buffer_num_bytes;

            static ptr
            construct(){
                return new packetflow_params;
            }

            std::string
            to_string() const {
                return "packetflow parameters";
            }

    };

    /**
     @class packetflow_switch
     A switch in the network that arbitrates/routes packettrains
     to the next link in the network
     */
    class packetflow_switch :
        public networkswitch,
        public packetflow_router
    {

    protected:
      topology_descr::ptr descr_;

      packetflow_params::ptr params_;

      packetflow_switch(const packetflow_params::ptr& params);

      boost::unordered_map<int,packetflow_crossbar_buffer::ptr> in_buffers_;

      boost::unordered_map<int,packetflow_simple_buffer::ptr> out_buffers_;

      boost::unordered_map<nodeaddress::ptr, eventmodule::ptr> ejectors_;

      packetflow_crossbar::ptr xbar_;

      static debug dbg_;

    protected:
        void
        connect_crossbar_buffer(const packetflow_crossbar_buffer::ptr& buf);

    public:
      static const int injection_port = -1;

      typedef boost::intrusive_ptr<packetflow_switch> ptr;

      static ptr
      construct(const sim_parameters::ptr& params);

      /**
       Create an exact copy of this switch, but with different router and network node address
       @param The router for the cloned switch
       @param The network node address for the cloned switch
       @return The new packetflow_switch
       */
      virtual networkswitch::ptr
      clone(router::ptr route, nodeaddress::ptr addr) const
      {
        ptr ret = new packetflow_switch(params_);
        clone_into(ret, addr, route);
        return ret;
      }

      virtual void
      clone_into(
        const ptr& sw,
        const nodeaddress::ptr &me,
        const router::ptr &route) const;

      virtual void
      initialize();

      packetflow_crossbar_buffer::ptr
      get_input_buffer(int port);

      int
      get_queue_length(int port) const;

      virtual int
      get_next_port(const packetflow_payload::ptr &msg);

      virtual void
      eject(const packetflow_payload::ptr &msg);

      virtual void
      connect(int dim, int port, connection_type_t ty,
          const connectable::ptr& mod, double weight, int red);

      virtual void
      connect(const nodeaddress::ptr& addr, int port, connection_type_t ty,
          const connectable::ptr& mod);

      /**
       Cast message and pass to #send
       @param msg Incoming message (should cast to packettrain)
       */
      void
      handle(const sst_message::ptr &msg);

      /**
       Set the link to use when ejecting packets at their endpoint.  A packetflow_switch
       can have any number of ejectors, corresponding to the number of nodes
       per switch.
       @param addr The compute node address of the endpoint to eject to
       @param link The link to the compute node for ejection
       */
      void
      add_ejector(const nodeaddress::ptr& addr, const eventmodule::ptr& link);

      virtual void
      set_eventmanager(const eventmanager::ptr &m);

      virtual void
      finish();

      virtual std::string
      to_string() const;

      virtual
      ~packetflow_switch();

      virtual std::string
      get_interprocess_link_name() const {
        return "packetflow switch";
      }

      topology_descr::ptr
      get_descr() const {
        return descr_;
      }

    };

  }
}

#endif // packetflow_switch_H


#endif // PACKETFLOW_SWITCH_H
