#include <sstmac/hardware/packettrain/congestion.h>
#include <sstmac/hardware/packettrain/trainswitch.h>

namespace sstmac {
    namespace hw {

    debug traincongestionpoint::dbg_;
    std::map<std::string, stats_congestion::ptr> traincongestionpoint::loggers_;
    const double traincongestionpoint::send_delayed = -1;


stats_congestion::stats_congestion(const std::string& stat_str, const std::string& fname)
    : stat_logger(stat_str, fname)
{
}

stats_congestion::ptr
stats_congestion::construct(const std::string& devtype)
{
    std::string fname = sstprintf("%s_congestion.dat", devtype.c_str());
    std::string stat_str = sstprintf("<stats> congestion | %s_congestion", devtype.c_str());
    return new stats_congestion(stat_str, fname);
}

void
stats_congestion::add_congestion(const traincongestionpoint::ptr &ev_mod, double bytes_sent, double delay)
{
    link_stats& entry = link_stats_[ev_mod];
    entry.kbytes_sent += bytes_sent / 1e3;
    entry.ms_delayed += delay * 1e3;
}

stats_congestion::link_stats::link_stats()
    : kbytes_sent(0), ms_delayed(0)
{
}

    void
    stats_congestion::simulation_finished(timestamp t_end)
    {
      if (is_active() && check_open())
      {

        boost::unordered_map<traincongestionpoint::ptr, link_stats>::iterator
            it, end = link_stats_.end();
        myfile
            << sstprintf("%50s %12s %12s %12s\n",
                "           Device          ", "Sent(KB)", "Delay(ms)",
                "Av(us/KB)");
        for (it = link_stats_.begin(); it != end; ++it)
        {
          link_stats& entry = it->second;
          double delay_per_kbyte_us = (entry.ms_delayed / entry.kbytes_sent)
              * 1e3;
          myfile << sstprintf("%50s %12.6f %12.6f %12.6f\n",
              it->first->name().c_str(), entry.kbytes_sent, entry.ms_delayed,
              delay_per_kbyte_us);

        }
        myfile.close();

        link_stats_.clear();
      }
    }

void
traincongestionpoint::set_eventmanager(const eventmanager::ptr &m)
{
    eventmodule::set_eventmanager(m);
    stats_congestion::ptr& cong = loggers_[devtype_];
    if (!cong){
        cong = stats_congestion::construct(devtype_);
        logger::register_logger(cong.get());
        m->register_stat(cong);
    }
    logger_ = cong;
}

traincongestionpoint::traincongestionpoint(double num_bytes_per_send,
                                           double freq_hz,
                                           double payload_fraction,
                                           const std::string& devtype,
                                           const std::string& id)
    : arbiter_(0), enable_multipath_(false), ignore_congestion_(false),
      freq_hz_(freq_hz),
      bw_bytes_(num_bytes_per_send * freq_hz * payload_fraction),
      payload_fraction_(payload_fraction),
      num_bytes_per_send_(num_bytes_per_send),
      devtype_(devtype),
      id_(id)
{
    //std::cout << "Building " << devtype_ << " " << id_ << std::endl;
    if (!dbg_.is_real())
    {
      dbg_ = debug("<debug> train");
    }
}

void
traincongestionpoint::uncongested_input(const packettrain::ptr& train)
{
    if (!train->negligible_size()){
        SSTMAC_DEBUG << sstprintf("On %s, uncongested input %s\n",
                                  to_string().c_str(), train->to_string().c_str());
    }

    arbiter_->uncongested_input(train);
}

std::string
traincongestionpoint::name() const
{
    return devtype_ + " : " + id_;
}

void
traincongestionpoint::finish()
{
    if (finished_)
        return;

    trainpath::finish();

    arbiter_->finish();
    vc_alloc_->finish();

    arbiter_ = 0;
    vc_alloc_ = 0;
}

void
traincongestionpoint::allocate_vc_step(const packettrain::ptr& train)
{
    SSTMAC_DEBUG << sstprintf("On %s, allocate vc for %s\n",
                              to_string().c_str(), train->to_string().c_str());
    if (train->negligible_size()){
        //just skip straight to the bw step
        allocate_bw_step(train);
        return;
    }

    double t_now = now().sec();
    double free_time = vc_alloc_->allocate_vc(t_now, train);
    if (free_time == input_vc_allocator::free_vc){
        //we are good to continue
        allocate_output_buffer_step(train);
        //serialize things
        if (train->has_next()){
            timestamp delay(1e-9);
            packettrain::ptr next = train->pop_next();
            event::ptr send_ev = new_event(ptr(this), &traincongestionpoint::send, next);
            schedule_delay(delay, send_ev);
        }
    }
    else if (free_time == input_vc_allocator::queued_vc){
        SSTMAC_DEBUG << sstprintf("On %s, delaying %s until prev train clears\n",
                                  to_string().c_str(), train->to_string().c_str());
    }
    else {
        SSTMAC_DEBUG << sstprintf("On %s, delaying %s until %12.8e\n",
                                  to_string().c_str(), train->to_string().c_str(),
                                  free_time);
        //there are packets backed up ahead of me
        event::ptr send_ev = new_event(ptr(this), &traincongestionpoint::allocate_vc_step, train);
        timestamp t_finish(free_time);
        schedule(t_finish, send_ev);
    }
}

void
traincongestionpoint::allocate_output_buffer_step(const packettrain::ptr& train)
{
    SSTMAC_DEBUG << sstprintf("On %s, allocate output buffer for %s\n",
                               to_string().c_str(), train->to_string().c_str());
    double t_now = now().sec();
    train->set_start_time(t_now);
    double time_until_free = arbiter_->allocate_output_buffer(train, t_now);
    if (time_until_free == output_bw_arbitrator::free_buffer){
        allocate_bw_step(train);
    }
    else {
        event::ptr send_ev = new_event(ptr(this), &traincongestionpoint::allocate_bw_step, train);
        timestamp delay(time_until_free);
        schedule_delay(delay, send_ev);
    }
}

void
traincongestionpoint::allocate_bw_step(const packettrain::ptr& train)
{
    SSTMAC_DEBUG << sstprintf("On %s, allocate bw for %s\n",
                               to_string().c_str(), train->to_string().c_str());
    double t_now = now().sec();
    double t_start_ser, t_stop_ser;
    if (train->negligible_size()){
        uncongested_input(train);
        t_start_ser = t_now;
        t_stop_ser = t_now + train->ser_delay();
    }
    else if (ignore_congestion_){
        uncongested_input(train);
        t_start_ser = t_now;
        t_stop_ser = t_now + train->ser_delay();
    }
    else {
        new_input(train, t_start_ser, t_stop_ser);
    }

    timestamp t_start_send(t_start_ser);
    do_send(t_start_send, train);

    timestamp t_finish_send(t_stop_ser);
    ack(t_finish_send, train);

    if (!train->negligible_size()){
        event::ptr ev = new_event(ptr(this), &traincongestionpoint::deallocate_vc_step, train);
        schedule(t_finish_send, ev);
    }
}

void
traincongestionpoint::deallocate_vc_step(const packettrain::ptr& train)
{
    SSTMAC_DEBUG << sstprintf("On %s, deallocate vc for %s\n",
                               to_string().c_str(), train->to_string().c_str());

    packettrain::ptr next = vc_alloc_->deallocate_vc(now().sec(), train);
    if (next){
        event::ptr send_ev = new_event(ptr(this), &traincongestionpoint::allocate_output_buffer_step, next);
        SSTMAC_DEBUG << sstprintf("On %s, scheduling send %s at %12.8e\n",
                                    to_string().c_str(), next->to_string().c_str(),
                                    now().sec());

        schedule_now(send_ev);
    }
}

void
traincongestionpoint::send(const packettrain::ptr& train)
{
    //if (devtype_ == "network_link")
    //    std::cout << "Sending on " << devtype_ << " " << id_ << std::endl;

    SSTMAC_DEBUG << sstprintf("On %s, sending %s\n",
        to_string().c_str(),
            train->to_string().c_str());

    //this starts a recursion through all the send steps
    allocate_vc_step(train);
}

void
traincongestionpoint::new_input(const packettrain::ptr& train, double& t_start_ser, double& t_stop_ser)
{
    double t_now = now().sec();
    train->set_initial_bw(bw_bytes_);

    arbiter_->arbitrate(train, t_now, t_start_ser, t_stop_ser);

    SSTMAC_DEBUG << sstprintf("On %s, arbitrate %s to t_start=%12.8e t_stop=%12.8e\n",
        to_string().c_str(),
        train->to_string().c_str(),
        t_start_ser, t_stop_ser);

    if (logger_->is_active()){
        double uncongested_latency = arbiter_->min_delay(train);
        double actual_latency = t_stop_ser - train->get_arrival_time();
        double delay = actual_latency - uncongested_latency;
        //std::cout << sstprintf("delay=%12.8e actual=%11.8e uncong=%11.8e t_stop=%11.8e t_start=%11.8e t_now=%11.8e t_arrive=%11.8e bw=%8.4e bytes=%lu\n",
        //                       delay, actual_latency, uncongested_latency,
        //                       t_stop_ser, t_start_ser,
        //                       t_now, train->get_arrival_time(),
        //                       bw_bytes_, train->num_bytes());
        logger_->add_congestion(ptr(this), train->num_bytes(), delay);
        double delay_us = delay * 1e6;
        train->add_congestion_delay_us(delay_us * 1e6);
    }

    if (t_stop_ser != t_start_ser){
        double total_time_to_send = t_stop_ser - t_start_ser;
        double exit_bw = train->num_bytes() / total_time_to_send;
        train->set_input_bw(exit_bw);
    }

}

    }
}
