#ifndef PACKET_TRAIN_H
#define PACKET_TRAIN_H

#include <sstmac/hardware/network/interconnect/interconnect_message.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/sim_parameters.h>

#define divide_and_round_up(num, denom) ((((num) - 1)/(denom)) + 1)

/**
 sstvariable {
 ignore_congestion=Whether to model congestion.  Setting ignore
 to true speeds up the simulation and assumes
 an idealized network where packets do not
 compete for bandwidth.;
 enable_multipath=Whether to allow packets to traverse multiple
 paths to their network endpoint.  This also
 means an MPI message might experience
 congestion from running into itself.;
 payload_fraction=The fraction of each packet that actually carries payload.
 The observable MPI bandwidth is given by
 MPI_BW=Link_BW*Payload_Fraction;
 bytes_per_cycle=The total number of bytes moved in each cycle.
 The bandwidth is given by Num_Bytes*Freq;
 train_latency=On a quiet network, the number of cycles a packet
 needs to traverse the component.  The latency
 is given by Num_Cycles/Freq;
 train_freq=Set the frequency of the network component. The bandwidth
 is given by Bytes_Per_Cycle*Freq;
 train_arbitrator=The arbitration algorithm for modeling the congestion in a
 network component.  Each algorithm uses different approximations
 of varying accuracy/efficiency.;
 }
 */

namespace sstmac
{

  namespace hw
  {

#define USE_DEBUG_TRAJECTORY 0

#if USE_DEBUG_TRAJECTORY
    typedef enum
      { up=0, down=1, left=2, right=3}switch_direction;
#endif


    /**
     @class packettrain
     Encapsulates a group of machine packets traveling together on the
     same path between endpoints.  This is usually one fraction of
     a larger message.
     */
    class packettrain :
        public interconnect_message
    {
    public:
      typedef boost::intrusive_ptr<packettrain> ptr;

      static const double uninitialized_bw;

    protected:
      double input_bw_;

      uint64_t byte_offset_;

      packettrain::ptr next_pending_;

      double t_arrive_;

      bool vc_allocated_;

      uint64_t msg_id_;

      double congestion_delay_us_;

      static uint64_t id_count_;

      static uint64_t max_num_trains_;

      static uint64_t min_num_bytes_per_train_;

      static uint64_t negligible_size_;

      static debug dbg_;

      packettrain(const sst_message::ptr& parent,
          int msg_id,
          uint64_t num_bytes,
          uint64_t offset,
          double bw);

      packettrain(const sst_message::ptr& parent);

    public:
      static ptr
      construct(const sst_message::ptr& parent);

      virtual
      ~packettrain();

      /**
       @return The total number of bytes in the complete message
       summed over all trains
       */
      uint64_t
      num_bytes_total() const
      {
        return orig_->get_byte_length();
      }

      /**
       @return The minimum number of bytes needed in a train. An 8K
       messages with min_bytes=4K will be split into two
       messages of size 4K.
       */
      static uint64_t
      min_num_bytes_per_train()
      {
        return min_num_bytes_per_train_;
      }

      /**
       @return The size at which a message is considered
       'negligible' for congestion modeling. This
       message is basically just routed with
       no congestion arbitration.
       */
      bool
      negligible_size() const
      {
        return num_bytes_total() <= negligible_size_;
      }

      /**
       @return Whether or not this is the head train for a message.
       NOTE: there is no in-order arrival assumption at
       the endpoint
       */
      bool
      is_head() const
      {
        return byte_offset_ == 0;
      }

      /**
       @return Whether or not this is the tail train for a message
       NOTE: there is no in-order arrival assumption at
       the endpoint
       */
      bool
      is_tail() const
      {
        return (byte_offset_ + num_bytes_) == num_bytes_total();
      }

      /**
       @return Whether the parent message is being sent as a single
       packettrain, i.e. the parent message is smaller
       than the min_bytes per train.
       */
      bool
      is_single_packet() const
      {
        return num_bytes_total() <= min_num_bytes_per_train_;
      }

      /**
       @return The time a packettrain arrived at its current destination.
       This is used mainly for monitoring congestion
       */
      double
      get_arrival_time() const
      {
        return t_arrive_;
      }

      /**
       @return The time a packettrain arrived at its current destination.
       This is used mainly for monitoring congestion
       */
      void
      set_start_time(double time)
      {
        t_arrive_ = time;
      }

      /**
       @param num_bytes The number of bytes sent in a message
       @return Whether num_bytes would be sent as a single packet train
       */
      static bool
      is_single_packet(uint64_t num_bytes)
      {
        return num_bytes <= min_num_bytes_per_train_;
      }

#if USE_DEBUG_TRAJECTORY
      void add_step(const eventmodule::ptr& router, switch_direction dir)
        {
          mdata_->trajectory[router] = dir;
        }

      switch_direction route(const eventmodule::ptr& source)
        {
          return mdata_->trajectory[source];
        }

      void set_trajectory(const std::map<eventmodule::ptr, switch_direction>& trajectory)
        {
          mdata_->trajectory = trajectory;
        }
#endif

      /**
       @return Whether the packettrain has been assigned a
       virtual channel at its current switch/link
       */
      bool
      vc_allocated() const
      {
        return vc_allocated_;
      }

      /**
       Set the current virtual channel status of the message
       at its current switch/link
       @param flag
       */
      void
      set_vc_allocated(bool flag)
      {
        vc_allocated_ = flag;
      }

      uint64_t
      msg_id() const
      {
        return msg_id_;
      }

      bool
      get_needs_ack() const;

      /**
       */
      double
      ser_delay() const
      {
        return num_bytes_ / input_bw_;
      }

      /**
       @return The number of bytes in this packettrain, NOT
       the total number of bytes in the parent message.
       See #num_bytes_total
       */
      uint64_t
      num_bytes() const
      {
        return num_bytes_;
      }

      /**
       @return The bandwidth in number of bytes per second
       */
      double
      get_input_bw() const
      {
        return input_bw_;
      }

      /**
       @param The bandwidth in number of bytes per second
       */
      void
      set_input_bw(double bw);

      /**
       Set the bandwidth of the train to a given max bw.
       Bandwidth may already be lower than max bw, in
       which case this function has no effect.
       @param The bandwidth in number of bytes per second
       */
      void
      throttle_bw(double bw);

      /**
       Trains from the same message may back up on each other,
       forming a linked list of messages to send later. This
       pops the front of this linked list and returns it,
       allowing the next message to be sent.  This operation
       modifies the linked list.
       @return The next pending message.
       */
      packettrain::ptr
      pop_next();

      /**
       Trains from the same message may back up on each other,
       forming a linked list of messages to send later.
       @return If there a pending next message
       */
      bool
      has_next() const
      {
        return next_pending_;
      }

      /**
       Trains from the same message may back up on each other,
       forming a linked list of messages to send later. This
       adds a message to the linked list, moving it to
       the front if there are multiple messages.
       @param The next pending message that will be sent
       */
      void
      set_next(const packettrain::ptr& train)
      {
        next_pending_ = train;
      }

      /**
       Trains from the same message may back up on each other,
       forming a linked list of messages to send later. This
       adds a message to the linked list, moving it to
       the end if there are multiple messages.
       @param The next pending message that will be sent
       */
      void
      add_next(const packettrain::ptr& train);

      /**
       Trains from the same message may back up on each other,
       forming a linked list of messages to send later.
       @return The next message in the linked list
       */
      packettrain::ptr
      get_next() const
      {
        return next_pending_;
      }

      /**
       If the bandwidth for a packettrain is unassigned,
       set its initial bandwidth. Otherwise, no effect
       @param The initial bw in byes per second
       */
      void
      set_initial_bw(double bw)
      {
        input_bw_ = input_bw_ < 0 ? bw : input_bw_;
      }

      /**
       Add to a running total of congestion delays
       to this packettrain across its entire path
       @param delay_us The delay increment in microseconds
       */
      void
      add_congestion_delay_us(double delay_us)
      {
        congestion_delay_us_ += delay_us;
      }

      /**
       @return The running total (in microseconds) of congestion delays
       to this packettrain across its entire path
       */
      double
      get_congestion_delay_us() const
      {
        return congestion_delay_us_;
      }

      /**
       Commit the total congestion delay to the parent message
       */
      void
      finalize_congestion_delay();

      std::string
      to_string() const;

      static void
      init_statics(int min_bytes, int max_trains, int negligible_size);

    };

  }
}

#endif // PACKET_TRAIN_H
