#ifndef TRAINSTATION_H
#define TRAINSTATION_H

#include <sstmac/hardware/packettrain/packettrain.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/eventmodule.h>

namespace sstmac
{
  namespace hw
  {

    /**
      @class trainstation
      The final arrival point of a message represented by packet trains.
      This tracks how many bytes of a given message have been received
      and when the entire message has arrive, informs a recv event handler
      that the message is ready to go.  There is no assumption of in-order
      arrival.
    */
    class trainstation :
        public eventmodule
    {
    private:
      std::map<uint64_t, uint64_t> trains_;

      int num_;

      static debug dbg_;

      static int num_stations_;

      /**
        When all trains for a message have finished arriving
        (i.e. been unloaded) finalize the train's arrival
        and pass on to a recv event handler
      */
      void
      finalize_train(const packettrain::ptr& train);

      /**
        When a train begins arriving, schedule the time
        that it will finish arriving.
        @param train The newly arriving train
      */
      void
      schedule_unload(const packettrain::ptr& train);

    protected:
      eventhandler::ptr handler_;

      trainstation(const eventhandler::ptr& handler);

    public:
      typedef boost::intrusive_ptr<trainstation> ptr;

      /**
        Construct a new endpoint for recving messages
        @param handler The event handler that will be called
                       when all trains for a message finish
                       arriving
      */
      static ptr
      construct(const eventhandler::ptr& handler);

      /**
        Cast message and pass to #schedule_unload
        @param msg Incoming message (should cast to packettrain)
      */
      virtual void
      handle(const sst_message::ptr& msg);

      /**
        Event scheduled by #schedule_unload.  This will
        increment the total number of bytes received for
        train's parent message. If all bytes have been receved,
        train is sent to #finalize_train.
        @param train The newly arrivated train at the endpoint
      */
      void
      unload_train(const packettrain::ptr& train);

      virtual std::string
      to_string() const;

      virtual void finish();

      /**
        @return The unique trainstation identifier
      */
      int
      num() const
      {
        return num_;
      }
    };



  }
}

#endif // TRAINSTATION_H
