#ifndef TRAINSWITCH_CC
#define TRAINSWITCH_CC

/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

//#include <sstmac/backends/native/dist_eventmap.h>
//#include <sstmac/backends/native/logicalprocess.h>
#include <sstmac/hardware/network/switches/networkswitch.h>
#include <sstmac/hardware/packettrain/trainswitch.h>
#include <sstmac/hardware/packettrain/train_nic.h>
#include <sstmac/hardware/network/switches/packet/packet_routing_info.h>
#include <sstmac/software/api/mpi/mpimessage.h>

namespace sstmac
{

  namespace hw
  {

    SSTRegister("train | trainswitch", networkswitch, trainswitch);

    debug trainswitch::dbg_;

    trainconfig::ptr
    trainconfig::construct()
    {
      return new trainconfig;
    }

    trainswitch::trainswitch(const trainconfig::ptr &config) :
          latency_(config->switch_latency_cycles / config->switch_freq_hz),
          tile_bw_(
              config->switch_freq_hz * config->switch_bytes_per_cycle
                  * config->payload_fraction), vis_(config->vis),
          config_(config)
    {
      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> trainswitch | network");
        logger::register_logger(&dbg_);
      }
    }

    trainswitch::ptr
    trainswitch::construct(const sim_parameters::ptr& params)
    {
      trainconfig::ptr config = trainconfig::construct();

      /**
       sstkeyword {
       gui=800mhz;
       docstring=$train_freq;
       }
       */
      config->switch_freq_hz = params->get_freq_param(
          "network_train_switch_freq");
      /**
       sstkeyword {
       gui=30;
       docstring=$train_latency;
       }
       */
      config->switch_latency_cycles = params->get_long_param(
          "network_train_switch_latency_cycles");
      /**
       sstkeyword { gui=2; docstring=$bytes_per_cycle; }
       */
      config->switch_bytes_per_cycle = params->get_double_param(
          "network_train_switch_bytes_per_cycle");

      config->payload_fraction = params->get_optional_double_param(
          "network_train_payload_fraction", 1.0);

      /**
       sstkeyword { docstring=$ignore_congestion; }
       */
      config->ignore_switch_congestion = params->get_optional_bool_param(
          "network_train_switch_ignore_congestion", false);
      /**
       sstkeyword { docstring=$enable_multipath; }
       */
      config->enable_switch_multipath = params->get_optional_bool_param(
          "network_train_switch_enable_multipath", false);

      /**
       sstkeyword { gui=1.5; docstring=$bytes_per_cycle; }
       */
      config->link_bytes_per_cycle = params->get_double_param(
          "network_train_link_bytes_per_cycle");
      /** sstkeyword { gui=3ghz; docstring=$train_freq; } */
      config->link_freq_hz = params->get_freq_param("network_train_link_freq");
      /**
        sstkeyword {
            gui=30ns;
            docstring=The latency required to traverse the link between switches;
        }
      */
      config->link_latency = params->get_time_param(
          "network_train_link_latency");

      /**
       sstkeyword { docstring=$ignore_congestion; }
       */
      config->ignore_link_congestion = params->get_optional_bool_param(
          "network_train_link_ignore_congestion", false);
      /**
       sstkeyword { docstring=$enable_multipath; }
       */
      config->enable_link_multipath = params->get_optional_bool_param(
          "network_train_link_enable_multipath", false);

      /**
       sstkeyword { docstring=$train_arbitrator; }
       */
      config->link_arb_template
          = output_bw_arbitrator_factory::get_optional_param(
              "network_train_link_arbitrator", "infinite_buffer", params);

      //initialize the static variables
      /**
       sstkeyword {
       docstring=The minimum number of bytes a single packet train can contain.ENDL
       Raising this value increases the coarseness and lowers the accuracy.;
       };
       */
      int min_bytes =
          params->get_optional_int_param("train_min_num_bytes", 256);
      /**
       sstkeyword {
       docstring=The maximum number of packet trains a message can be split into.ENDL
       For example, if 1MB is sent and max trains is 100, each packet trainENDL
       will contain 10KB.  If max trains is raised to 1000, then each trainENDL
       can be made smaller to 1KB.;
       }
       */
      int max_trains = params->get_optional_int_param("train_max_num_messages",
          10);
      /**
       sstkeyword {
       docstring=The size below which a train is consideredENDL
       negligible and does not do congestion modeling.ENDL
       This increases efficiency, but lowers accuracy.;
       }
       */
      int negligible_size = params->get_optional_int_param(
          "train_negligible_size", 512);

      packettrain::init_statics(min_bytes, max_trains, negligible_size);

      config->vis = params->has_param("vis_engine");

      return new trainswitch(config);
    }

    void
    trainswitch::ack(const timestamp &t, const packettrain::ptr &train)
    {
      //do nothing
    }

    trainswitch::~trainswitch()
    {
    }

    void
    trainswitch::finish()
    {
      if (finished_)
        return;

      {
        boost::unordered_map<long, trainlink::ptr>::iterator it, end = outports_.end();
        for (it = outports_.begin(); it != end; it++)
        {
            trainlink::ptr link = it->second;
            if (link)
            {
              link->finish();
            }
            else
            {
              sst_throw(ssterror, "at trainswitch::finish(): null link at port %d", it->first);
            }
        }
      }
      outports_.clear();

      {
        boost::unordered_map<nodeaddress::ptr, eventmodule::ptr>::iterator it,
            end = ejectors_.end();
        for (it = ejectors_.begin(); it != end; ++it)
        {
          eventmodule::ptr& mod = it->second;
          if (mod)
            mod->finish();
        }
      }
      ejectors_.clear();

      router_ = 0;

    }

    void
    trainswitch::connect(int dim, int port_offset, connection_type_t ty,
        const connectable::ptr &mod, double weight, int red)
    {
      std::stringstream ss;

      ss << "(switch id: " << my_addr_->unique_id() << ") dim: " << dim
          << " port_offset: " << port_offset;

      long port = router_->convert_to_outport(dim, port_offset);
      outports_[port] = trainlink::construct(safe_cast(mod, component),
          config_->link_latency, config_->link_bytes_per_cycle * weight * red,
          config_->link_freq_hz, config_->payload_fraction,
          config_->enable_link_multipath, config_->ignore_link_congestion,
          config_->link_arb_template->clone(), "network_link", ss.str());

    }

    void
    trainswitch::set_eventmanager(const eventmanager::ptr &m)
    {
      eventmodule::set_eventmanager(m);
      boost::unordered_map<long, trainlink::ptr>::iterator it, end = outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
          trainlink::ptr link = it->second;
          if (link)
          {
            link->set_eventmanager(m);
          }
      }
    }

    void
    trainswitch::connect(const nodeaddress::ptr& addr, int port,
        connection_type_t ty, const connectable::ptr& mod)
    {
      connectable::ptr nic = mod;
      switch (ty)
      {
      case injection:
        nic->connect(injection, ptr(this));
        break;
      case ejection:
        nic->connect(ejection, ptr(this));
        break;
      default:
        connectable_type_invalid(ty);
      }
    }

    void
    trainswitch::add_ejector(const nodeaddress::ptr& addr,
        const trainpath::ptr& dest)
    {
      SSTMAC_DEBUG2 << sstprintf("Adding ejector at node %s to switch at %s\n",
            addr->to_string().c_str(), my_addr_->to_string().c_str());
      ejectors_[addr] = dest;
    }

#if 0
    void
    trainswitch::update_router_cong(routing_info::ptr info)
    {
      congestion_routing_info::ptr cast = boost::dynamic_pointer_cast<
          congestion_routing_info>(info);
      if (!cast)
      {
        sst_throw(ssterror, "trainswitch::update_router_cong: routing_info didn't cast to a congestion_routing_info");
      }
      boost::unordered_map<int, boost::unordered_map<int, trainlink::ptr> >::iterator
          it, end = outports_.end();
      for (it = outports_.begin(); it != end; it++)
      {
        boost::unordered_map<int, trainlink::ptr>::iterator it2, end2 =
            it->second.end();
        for (it2 = it->second.begin(); it2 != end2; it2++)
        {
          cast->set_congestion_info(
              router_->convert_to_outport(it->first, it2->first),
              it2->second->next_free());
        }
      }
    }
#endif

    void
    trainswitch::do_send(const timestamp &t_start,
        const packettrain::ptr& train)
    {
      train->throttle_bw(tile_bw_);
      long port = router_->route(train);
      //router_->route(train->rinfo(), dim, dir);
      timestamp arrive = now() + latency_;

      if (router::is_eject(port))
      {
        schedule(arrive, ejectors_[train->toaddr()], train);
      }
      else
      {
        if (vis_)
        {
          if (linkvis_[port])
          {

            double act = outports_[port]->get_util_bw();
            linkvis_[port]->set_activity(act);
            vis_parent_->vis_update(now());
          }

        }
        if(outports_.find(port) == outports_.end()){
            sst_throw(valueerror, "trainswitch::handle: can't find port %d sending from switch %ld to node %ld",
                  port, get_addr()->unique_id(), train->toaddr()->unique_id());
        }
        schedule(arrive, outports_[port], train);
      }
    }

    int
    trainswitch::get_queue_length(int port) const
    {
        boost::unordered_map<long, trainlink::ptr>::const_iterator
            it = outports_.find(port);
        if (it == outports_.end()){
            std::cerr << "Valid ports are:\n";
            for (it=outports_.begin(); it != outports_.end(); ++it){
                std::cerr << it->first << std::endl;
            }
            sst_throw(valueerror, "trainswitch::get_queue_length: "
                      "invalid port %d", port);
        }
        const trainlink::ptr& tr = it->second;
        return tr->get_queue_length();
    }

    void
    trainswitch::handle(const sst_message::ptr &msg)
    {
      packettrain::ptr train = safe_cast(msg, packettrain);
      do_send(now(), train);
    }

    std::string
    trainswitch::to_string() const
    {
      return "train switch";
    }

  }
}

#endif // TRAINSWITCH_CC
