/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/computescheduler.h>
#include <sstmac/software/process/thread.h>

namespace sstmac
{
  namespace hw
  {

    debug computescheduler::dbg_;

    computescheduler::computescheduler(long nodeid, int ncores, int nsocket, timestamp preempt, const sw::operatingsystem::ptr& os) :
        preemption_(preempt), ncores_(ncores), nsocket_(nsocket), nodeid_(nodeid), os_(os), next_affinity_(0)
    {
      if (!dbg_.is_real())
      {
          dbg_ = debug("<debug> computescheduler");
          logger::register_logger(&dbg_);
      }

      os->init_scheduler(computescheduler::ptr(this));
    }

    int
    computescheduler::get_thread_core_affinity(long threadid) const
    {
        std::map<long, int>::const_iterator it = thread_affinities_.find(threadid);
        if (it == thread_affinities_.end())
            return sw::thread::no_core_affinity;
        else
            return it->second;
    }

    int
    computescheduler::get_socket(long thread_id)
    {
        int core = get_thread_core_affinity(thread_id);
        if (core == sw::thread::no_core_affinity){
            return -1;
        }
        else {
            int socket = core * nsocket_ / ncores_;
            return socket;
        }
    }

    void
    computescheduler::set_thread_core_affinity(long threadid, int aff)
    {
        thread_affinities_[threadid] = aff;
    }

    void
    computescheduler::schedule_affinity(long threadid, int strict_affinity)
    {

        if (strict_affinity != sw::thread::no_core_affinity)
        {
            /** we have a strict affinity from start up */
            set_thread_core_affinity(threadid, strict_affinity);
        }
        else if (threadid == sw::thread::main_thread){
            /** this is the operating system or device running. */
            set_thread_core_affinity(sw::thread::main_thread, sw::thread::no_core_affinity);
        }
        else {
            int aff = get_thread_core_affinity(threadid);
            if (aff == sw::thread::no_core_affinity) //assign an affinity
            {
                aff = next_affinity_;
                set_thread_core_affinity(threadid, aff);
                next_affinity_ = (next_affinity_ + 1) % ncores_;
                SSTMAC_DEBUG << "operatingsystem[ncores=" << ncores_ << "]: "
                             <<  "assigned core affinity " << aff
                             << " to thread " << threadid << "\n";
            }
            else
            {
                //keep the old one - this can be updated later
                //with a smarter scheduler
            }
        }
    }


  }
} //end of namespace sstmac
