/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_PROCESSOR_COMPUTESCHEDULER_H_INCLUDED
#define SSTMAC_HARDWARE_PROCESSOR_COMPUTESCHEDULER_H_INCLUDED

#include <sstmac/common/eventhandler.h>
#include <sstmac/common/messages/sst_message.h>
#include <sstmac/common/eventmodule.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/failure/failable.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/common/sim_parameters.h>


namespace sstmac
{
  namespace hw
  {

    class computescheduler : public eventmodule, public failable
    {
    protected:
      computescheduler(long nodeid, int ncores, int nsocket, timestamp preempt, const sw::operatingsystem::ptr& os);

      timestamp preemption_;

      std::map<long, int> thread_affinities_;

      int ncores_;
      int nsocket_;
      long nodeid_;
      sw::operatingsystem::ptr os_;
      static debug dbg_;
      int next_affinity_;

      void set_thread_core_affinity(long threadid, int aff);

      virtual long
      failure_id() const
      {
        return nodeid_;
      }

    public:
       typedef enum { SCHEDULE, RESERVE, RELEASE } scheduler_action_t;

      virtual std::string
      to_string() const
      {
        return "computescheduler";
      }
      typedef boost::intrusive_ptr<computescheduler> ptr;

      virtual
      ~computescheduler()
      {
      }



      virtual int
      get_socket(long thread_id);

      virtual void
      schedule_computation(const timestamp& comp,
          const sst_message::ptr& to_deliver,
          const eventhandler::ptr& handler,
          long thread_id) = 0;

      virtual void
      reserve(const sst_message::ptr& msg, long thread_id) = 0;

      virtual void
      release(const sst_message::ptr& msg, long thread_id) = 0;

      virtual void
      schedule_affinity(long thread_id, int strict_affinity);

      int get_thread_core_affinity(long threadid) const;

      class compute_job : public ptr_type
      {
      protected:
        compute_job()
        {
        }

        compute_job(const sst_message::ptr& m, timestamp t, timestamp l,
            eventhandler::ptr h, long threadid, int c = -1) :
          msg_(m), time_(t), timeleft_(l), handle_(h), core_(c),
            threadid_(threadid), action_(SCHEDULE)
        {
        }

        compute_job(const sst_message::ptr& m, long threadid) :
          msg_(m), time_(0), timeleft_(0), handle_(0), core_(-1),
            threadid_(threadid), action_(RESERVE)
        {
        }

      public:
        typedef boost::intrusive_ptr<compute_job> ptr;

        static ptr
        construct(const sst_message::ptr& m, timestamp t, timestamp l,
            eventhandler::ptr h, long threadid, int c = -1)
        {
          return ptr(new compute_job(m, t, l, h, threadid, c));
        }

        static ptr
        construct(const sst_message::ptr& m, long threadid)
        {
          return ptr(new compute_job(m, threadid));
        }

        virtual std::string
        to_string() const
        {
          return "compute_job";
        }

        sst_message::ptr msg_;
        timestamp time_;
        timestamp timeleft_;
        eventhandler::ptr handle_;
        int core_;
        long threadid_;
        scheduler_action_t action_;

      };

    };

  }
} //end of namespace sstmac
#endif
