/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_HARDWARE_PROCESSOR_INSTRUCTIONPROCESSOR_CONFIG_H_INCLUDED
#define SSTMAC_HARDWARE_PROCESSOR_INSTRUCTIONPROCESSOR_CONFIG_H_INCLUDED

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/timestamp.h>
#include <sstmac/hardware/processor/processor_config.h>

namespace sstmac { namespace hw 
{

  /**
   * Store configuration info for simple processor models.
   */
  class instructionprocessor_config : public processor_config
  {
  private:
    /// Configuration file
    std::string filename_;

    // Parameters

    /// Label for flop event data.
    std::string flop_;
    /// Average service time for flops
    double tflop_;

    /// Label for memory operation event data.
    std::string memop_;
    /// Average service time for memops (ns)
    double tmemop_;

    /// Label for local memory operation event data.
    std::string memop_local_;
    /// Average service time for local memops (ns)
    double tmemop_local_;

    /// Label for random memory operation event data.
    std::string memop_rand_;
    /// Average service time for random memops (ns)
    double tmemop_rand_;

    double pwr_static;
    double pwr_flop, pwr_memop, pwr_memop_local, pwr_memop_rand;

  protected:
    /// Protected constructor. Initializes to default configuration.
    instructionprocessor_config();

    /// Protected constructor using std::strings.
    /** We should not use timestamps here = JJW 05/14/2012.  The time increment
        per flop is so small that picosecond resolution only gives you 3 sig figs.
        If you push to 100 ps resolution, you get 1 sig fig. 1!!! */
    instructionprocessor_config(int cores, const std::string &flop_label,
        double tflop, const std::string &memop_label,
        double tmemop, const std::string &memop_local_label,
        double tmemop_local, const std::string &memop_rand_label,
        double tmemop_rand,
        double pwr_st, double p_flop, double p_memop,
        double p_memop_local, double p_memop_rand);



  public:
    typedef boost::intrusive_ptr<instructionprocessor_config> ptr;

    /// Public default constructor.
    static instructionprocessor_config::ptr
    construct();

    /// Public constructor. 
    static instructionprocessor_config::ptr
    construct(int cores, const std::string& flop_label,
        double tflop,
        const std::string& memop_label = "",
        double tmemop = 0,
        const std::string& memop_local_label = "",
        double tmemop_local = 0,
        const std::string& memop_rand_label = "",
        double tmemop_rand = 0,
        double pwr_st = 30, double p_flop = 1e-12, double p_memop = 1e-9,
        double p_memop_local = 1e-11, double p_memop_rand = 1e-10);


    /// Farewell.
    virtual ~instructionprocessor_config() throw ()
    {
    }

    /// Set the configuration filename.
    void
    set_filename(const std::string &filename)
    {
      filename_ = filename;
    }

    /// Get the configuration filename.
    const std::string&
    filename() const
    {
      return filename_;
    }

    /// Set the label for flop event data.
    void
    set_flop_label(const std::string &label)
    {
      flop_ = label;
    }

    /// Get the label for flop event data.
    const std::string&
    flop_label() const
    {
      return flop_;
    }

    /// Set the average service time for flops (ns).
    void
    set_tflop(double val)
    {
      tflop_ = val;
    }

    /// Get the average serice time for flops (ns).
    double
    tflop() const
    {
      return tflop_;
    }

    /// Set the label for memop event data.
    void
    set_memop_label(const std::string &label)
    {
      memop_ = label;
    }

    /// Get the label for memop event data.
    const std::string&
    memop_label() const
    {
      return memop_;
    }

    /// Set the average service time for memops (ns).
    void
    set_tmemop(double val)
    {
      tmemop_ = val;
    }

    /// Get the average serice time for memops (ns).
    double
    tmemop() const
    {
      return tmemop_;
    }

    /// Set the label for local memop event data.
    void
    set_memop_local_label(const std::string &label)
    {
      memop_local_ = label;
    }

    /// Get the label for local memop event data.
    const std::string&
    memop_local_label() const
    {
      return memop_local_;
    }

    /// Set the average service time for local memops (ns).
    void
    set_tmemop_local(double val)
    {
      tmemop_local_ = val;
    }

    /// Get the average serice time for local memops (ns).
    double
    tmemop_local() const
    {
      return tmemop_local_;
    }

    /// Set the label for random memop event data.
    void
    set_memop_rand_label(const std::string &label)
    {
      memop_rand_ = label;
    }

    /// Get the label for random memop event data.
    const std::string&
    memop_rand_label() const
    {
      return memop_rand_;
    }

    /// Set the average service time for random memops (ns).
    void
    set_tmemop_rand(double val)
    {
      tmemop_rand_ = val;
    }

    /// Get the average serice time for random memops (ns).
    double
    tmemop_rand() const
    {
      return tmemop_rand_;
    }

    double
    get_pwr_static()
    {
      return pwr_static;
    }

    double
    get_pwr_flop()
    {
      return pwr_flop;
    }
    double
    get_pwr_memop()
    {
      return pwr_memop;
    }
    double
    get_pwr_memop_local()
    {
      return pwr_memop_local;
    }
    double
    get_pwr_memop_rand()
    {
      return pwr_memop_rand;
    }

  };

} } // end of namespace sstmac

#endif
