/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/hardware/processor/migratingcomputescheduler.h>

namespace sstmac
{
  namespace hw
  {

    migratingcomputescheduler::migratingcomputescheduler(long nodeid, int ncores, int nsocket, timestamp preempt, const sw::operatingsystem::ptr& os) :
        computescheduler(nodeid, ncores, nsocket, preempt, os)
    {


      for (int i = 0; i < ncores_; i++)
        {
          compute_slots_[i] = false;
        }
    }

    migratingcomputescheduler::~migratingcomputescheduler()
    {
    }

    int
    migratingcomputescheduler::find_open_core(long threadid)
    {
      int affinity = get_thread_core_affinity(threadid);

      /** if we have no affinity or the compute slot isn't open,
          try to reschedule this on an open core */
      if (affinity == sw::thread::no_core_affinity || compute_slots_[affinity])
      {
          for (int i = 0; i < ncores_; i++)
          {
              SSTMAC_DEBUG << "migratingcomputescheduler::request_compute checking for core "
                    << i << "\n";
              if (compute_slots_.find(i) == compute_slots_.end())
              {
                  sst_throw(ssterror,
                      "migratingcomputescheduler::request_compute: can't find core %d in compute_slots for num_cores=%d",
                              i, ncores_);
              }
              if (!compute_slots_[i])
              {

                  return i;
              }
          }
      }
      else //use the assigned affinity
      {
          SSTMAC_DEBUG << "migratingcomputescheduler::using assigned affinity "
                       << affinity << " for thread " << threadid << "\n";
          return affinity;
      }

      return no_open_cores;
    }

    void
    migratingcomputescheduler::schedule_computation(
        const timestamp& comp,
        const sst_message::ptr& to_deliver,
        const eventhandler::ptr& handler,
        long thread_id)
    {

      use_me();

      int core = find_open_core(thread_id);

      if (core == no_open_cores) //put in suspended queue
        {
          SSTMAC_DEBUG << "[t=" << now()
                << "] -- migratingcomputescheduler suspending compute request" << "\n";
          suspendedQ.push(
              compute_job::construct(to_deliver, comp, comp, handler, thread_id));

        }
      else
        {
          SSTMAC_DEBUG << "[t=" << now()
                << "] -- migratingcomputescheduler: computing for time " << comp << "\n";

          timestamp actual = comp;
          if (preemption_.sec() > 0)
            {
              actual = timestamp(std::min(comp.sec(), preemption_.sec()));
            }
          compute_slots_[core] = true;
          slots_compute_[to_deliver] = compute_job::construct(to_deliver,
              actual, comp, handler, thread_id, core);
          timestamp time = now() + actual;
          eventmodule::schedule(time, ptr(this), to_deliver);

          // opsys->block(data->get_key());
        }
    }

    void
    migratingcomputescheduler::release(const sst_message::ptr &msg, long thread_id)
    {
        int core = msg->get_core();
        if (core == sw::thread::no_core_affinity)
            throw illformederror("releasing compute on message without core assignment");
        msg->set_core(sw::thread::no_core_affinity);
        compute_slots_[core] = false;
        start_next_suspended();
    }

    void
    migratingcomputescheduler::reserve(const sst_message::ptr &msg, long thread_id)
    {
        int core = msg->get_core();
        if (core != sw::thread::no_core_affinity)
            throw illformederror("message already scheduled for compute");

        core = find_open_core(thread_id);
        if (core == no_open_cores){
            suspendedQ.push(compute_job::construct(msg, thread_id));
        }
        else {
            compute_slots_[core] = true;
            msg->set_core(core);
            eventhandler::ptr handle = msg->pop_callback();
            if (!handle){
                sst_throw(illformederror, "compute schedule reserve only works with callbacks");
            }
            handle->handle(msg);
        }
    }

    void
    migratingcomputescheduler::schedule_computation(compute_job::ptr jorb)
    {
      use_me();
      int core = find_open_core(jorb->threadid_);

      if (core < 0) //put in suspended queue
        {
          SSTMAC_DEBUG << "migratingcomputescheduler suspending compute request" << "\n";
          suspendedQ.push(jorb);

        }
      else
        {
          SSTMAC_DEBUG << "migratingcomputescheduler: computing for time "
                << jorb->timeleft_ << "\n";

          timestamp actual = jorb->timeleft_;
          if (preemption_.sec() > 0)
            {
              actual = timestamp(std::min(jorb->timeleft_.sec(), preemption_.sec()));
            }
          jorb->time_ = actual;

          compute_slots_[core] = true;
          slots_compute_[jorb->msg_] = jorb;
          timestamp time = now() + actual;
          eventmodule::schedule(time, ptr(this), jorb->msg_);

          // opsys->block(data->get_key());
        }
    }

    void
    migratingcomputescheduler::start_next_suspended()
    {
        if (!suspendedQ.empty())
          {
            SSTMAC_DEBUG << "migratingcomputescheduler("
                  << "): executing suspended kernel " << "\n";

            compute_job::ptr next_jorb = suspendedQ.front();
            suspendedQ.pop();
            if (next_jorb->action_ == SCHEDULE){
                schedule_computation(next_jorb->time_,
                    next_jorb->msg_,
                    next_jorb->handle_,
                    next_jorb->threadid_);
            }
            else if (next_jorb->action_ == RESERVE){
                reserve(next_jorb->msg_, next_jorb->threadid_);
            }
          }
    }

    void
    migratingcomputescheduler::handle(const sst_message::ptr &msg)
    {
      SSTMAC_DEBUG << "migratingcomputescheduler::handle(" << ")" << "\n";

      if (msg->get_type() == sst_message::FAILURE)
        {
          failure_occurred(msg, ptr(this));
          return;
        }

      compute_job::ptr jorb = slots_compute_[msg];
      slots_compute_.erase(msg);
      compute_slots_[jorb->core_] = false;

      start_next_suspended();

      if (jorb->time_ >= jorb->timeleft_)
        {

          SSTMAC_DEBUG << "migratingcomputescheduler("
                << "): calling handle on a finished job " << "\n";
          jorb->handle_->handle(jorb->msg_);
        }
      else
        {

          SSTMAC_DEBUG
                <<
                "migratingcomputescheduler - computation was preempted, scheduling again \n";

          jorb->timeleft_ = timestamp(
              jorb->timeleft_.sec() - jorb->time_.sec());
          schedule_computation(jorb);
        }
    }

  }
} //end of namespace sstmac
