/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/main/multisim.h>
#include <sstmac/common/parameters/file_parameters.h>
#include <sstmac/backends/native/driver.h>
#include <sstmac/main/sstmac.h>
#include <sstmac/common/fileio.h>

#include <sys/types.h>
#include <sys/shm.h>
#include <sys/time.h>
#include <sys/wait.h>

namespace sstmac
{
  namespace multisim
  {

    int _multisim_nprocs_ = 0;
    //type<int, std::ostream>::map _multisim_logs_;
    std::ofstream multilog_;

    void
    make_run(std::map<int, std::deque<std::string> > &vars,
        std::map<int, std::vector<std::string> > &runs,
        std::vector<int> &indexes)
    {
      // std::cout << "making run...";
      int next = runs.size();

      std::vector<std::string> vec(vars.size());
      for (int i = 0; i < (int) indexes.size(); i++)
      {
        std::deque<std::string>::const_iterator it, end = vars[i].end();
        int cnt = 0;
        for (it = vars[i].begin(); it != end && cnt < indexes[i]; it++)
        {
          cnt++;
        }

        vec[i] = *it;
      }

      runs[next] = vec;
      //std::cout << "done \n";
    }

    void
    rec_run(std::map<int, std::deque<std::string> > &vars,
        std::map<int, std::vector<std::string> > &runs,
        std::vector<int> &indexes, int var)
    {

      if (var == ((int) vars.size() - 1)) // at the end, make the run
      {
        for (int i = 0; i < (int) vars[var].size(); i++)
        {

          make_run(vars, runs, indexes);
          indexes[var]++;
        }
        indexes[var] = 0;
      }
      else
      {
        for (int i = 0; i < (int) vars[var].size(); i++)
        {

          rec_run(vars, runs, indexes, var + 1);
          indexes[var]++;
        }
        indexes[var] = 0;

      }

    }

    void
    print_multi_params(sstmac::sim_parameters::ptr params)
    {
      printf("%-40s%-100s\n", "Name", "Value");
      std::cout << "------------------------------------------------ \n";
      sstmac::sim_parameters::multisim_maptype p = params->get_multi_params();
      sstmac::sim_parameters::multisim_maptype::iterator it, end = p.end();

      for (it = p.begin(); it != end; it++)
      {
        printf("%-40s%-100s\n", it->first.c_str(), it->second.c_str());
      }
    }

    void
    add_to_combos(std::vector<std::vector<std::string> > &comb,
        std::deque<std::string> more)
    {

      if (comb.size() == 0)
      {
        // std::cout << "adding to comb first \n";
        std::deque<std::string>::const_iterator it = more.begin(), end =
            more.end();
        for (; it != end; it++)
        {
          std::vector<std::string> newitem;
          newitem.push_back(*it);
          comb.push_back(newitem);
        }
      }
      else
      {
        // std::cout << "adding to comb already of size " << comb.size() << "\n";
        std::vector<std::vector<std::string> > newvector;
        for (int i = 0; i < comb.size(); i++)
        {

          std::deque<std::string>::const_iterator it = more.begin(), end =
              more.end();

          for (; it != end; it++)
          {
            std::vector<std::string> newitem(comb[i]);
            newitem.push_back(*it);
            newvector.push_back(newitem);
          }
        }
        comb = newvector;
      }

      //std::cout << "add_to_combos done \n";
    }

    void
    add_to_runs(std::vector<parameter_run> &runs,
        sstmac::sim_parameters::ptr params, bool print, bool fail_on_existing)
    {
      sstmac::sim_parameters::multisim_maptype p = params->get_multi_params();
      sstmac::sim_parameters::multisim_maptype::iterator it, end = p.end();
      std::map<int, std::string> keys;

      int count = 0;
      std::vector<std::vector<std::string> > param_combos;
      for (it = p.begin(); it != end; it++)
      {
        //std::cout << it->first << "\t\t" << it->second << "\n";
        std::string val = trim_str(it->second, " \r\n{}");
        std::deque<std::string> tok;

        std::string space = "|";
        pst::BasicStringTokenizer::tokenize(val, tok, space);

        add_to_combos(param_combos, tok);

        keys[count] = it->first;
        count++;
      }

      //  std::vector<int> ind(count, 0);
      //  rec_run(parsed, runs, ind, 0);

      if (print)
      {
        std::cout << "------------------------------------ \n\n";
        std::cout << "Parameter combinations will be as follows: \n";
        std::cout << "  (values are in order as they appear above) \n";
        printf("%-10s%-100s\n", "Run#", "Values");
        std::cout << "------------------------------------------------ \n";
      }

      for (int i = 0; i < param_combos.size(); i++)
      {
        sstmac::sim_parameters::ptr newp = params->clone();
        if (print)
          printf("%-10i", (int) runs.size());
        for (int j = 0; j < param_combos[i].size(); j++)
        {
          if (fail_on_existing)
          {
            newp ->add_param(keys[j], param_combos[i][j]);
          }
          else
          {
            newp ->add_param_override(keys[j], param_combos[i][j]);
          }

          if (print)
          {
            std::cout << param_combos[i][j] << " ";
          }
        }
        if (print)
        {
          std::cout << "\n";
        }
        parameter_run r;
        r.num_ = runs.size();
        r.params_ = newp;
        runs.push_back(r);
      }

    }

    void
    multisim_init(sstmac::sim_parameters::ptr params,
        std::vector<parameter_run> &runs, bool print)
    {
      if (print)
      {
        std::cout << "-------------------------------------- \n";
        std::cout << "----- SST/macro Multi-Sim Manager ---- \n";
        std::cout << "-------------------------------------- \n \n";

      }

      // flatten all multi-includes
      sstmac::file_parameters::ptr fp = boost::dynamic_pointer_cast<
          sstmac::file_parameters>(params);

      if (fp)
      {
        std::vector<std::string> inc = fp->get_multisim_includes();

        if (inc.size() > 0)
        {

          std::vector<std::vector<std::string> > include_combos;
          for (int i = 0; i < inc.size(); i++)
          {
            std::string val = trim_str(inc[i], " \r\n{}");
            std::deque<std::string> tok;

            std::string space = "|";
            pst::BasicStringTokenizer::tokenize(val, tok, space);

            add_to_combos(include_combos, tok);
          }

          for (int i = 0; i < include_combos.size(); i++)
          {

            file_parameters::ptr newf = boost::dynamic_pointer_cast<
                file_parameters>(params->clone_multi());
            if (print)
            {
              std::cout << "**************************************************";
              std::cout << "******* For file combination: ";
            }
            for (int j = 0; j < include_combos[i].size(); j++)
            {
              if (print)
                std::cout << include_combos[i][j] << " ";

              try
              {
                newf->try_to_parse(include_combos[i][j], true);
              }
              catch (sstmac::valueerror& e)
              {
                sst_throw(
                    sstmac::inputerror,
                    "when using parameter sweeps across files, they must be mutually exclusive (for safety), but we dedected the following collisions: %s",
                    e.message.c_str());
              }

            }

            if (print)
            {
              std::cout << "\n";
              std::cout << "We found the following multi-parameters: \n";
              print_multi_params(newf);
            }

            try
            {
              add_to_runs(runs, newf, print, false);
            }
            catch (sstmac::valueerror& e)
            {
              sst_throw(
                  sstmac::inputerror,
                  "when using parameter sweeps across files, they must be mutually exclusive (for safety), but we dedected the following collisions: %s",
                  e.message.c_str());
            }

          }
        }
        else
        {
          if (print)
          {
            std::cout << "We found the following multi-parameters: \n";
            print_multi_params(params);
          }

          add_to_runs(runs, params, print, false);
        }

      }
      else
      {
        //no chance of includes
        if (print)
        {
          std::cout << "We found the following multi-parameters: \n";
          print_multi_params(params);
        }
        add_to_runs(runs, params, print, false);
      }

    }

    void
    child_done(int snum)
    {
      int pid;
      int status;

      pid = wait(&status);
      std::cout << "---------- Child simulation has finished ------------ \n";
      _multisim_nprocs_++;
    }

    sstmac::sim_parameters::ptr
    do_multisim_fork(std::vector<parameter_run> &runs)
    {
      _multisim_nprocs_ = runs[0].params_->get_optional_int_param(
          "multisim_nproc", 1);
      std::cout << "------------------------------------- \n";
      std::cout << "-> About to run " << _multisim_nprocs_
          << " simulations at a time, press enter to continue \n";
      if (!runs[0].params_->has_param("multisim_nproc"))
      {
        std::cout
            << "   (Warning: you have not set the multisim_nproc parameter, 1 is default)\n";
      }

      std::cin.ignore();

      //get busy
      signal(SIGCHLD, child_done);

      for (int r = 0; r < runs.size(); r++)
      {
        int pid;

        _multisim_nprocs_--;

        switch (pid = fork())
        {
        case 0: /* a fork returns 0 to the child */
        {

          std::stringstream ss;
          ss << "sim_output-" << runs[r].num_ << ".out";

          //std::ofstream myfile;

          multilog_.open(ss.str().c_str(), std::ios::out);

          if (!multilog_.is_open())
          {
            sst_throw(ioerror, "statlogger: cannot open file %s for writing",
                ss.str().c_str());
          }

          //_multisim_logs_[it2->first] = myfile;

          //debug::debug_stream_ = &_multisim_logs_[it2->first];
          debug::debug_stream_ = &multilog_;
          stat_logger::set_multisim(r);

          return runs[r].params_;
        }

        default: /* a fork returns a pid to the parent */

          break;

        case -1: /* something went wrong */
          perror("fork");
          exit(1);
        }

        while (_multisim_nprocs_ <= 0)
        {
          sleep(60);
        }
      }
      exit(0);
    }

  }
}
