/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/main/sstmac-mpisweep.h>
#include <sstmac/main/multisim.h>
#include <mpi.h>

using namespace sstmac;

std::ofstream multilog_;

//                   pid   run#
boost::unordered_map<int, int> pidruns;

//                   pid  proc#
boost::unordered_map<int, int> pidprocs;


void
child_done(int snum)
{
  int pid;
  int status;

  pid = wait(&status);

  std::cout << "---------- Run " << pidruns[pid]
      << " has finished ------------ \n";

  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  if (rank != 0)
  {
    int b = pidruns[pid];
    MPI_Send(&b, 1, MPI_INT, 0, pidprocs[pid], MPI_COMM_WORLD);
  }
  pidprocs.erase(pid);
  pidruns.erase(pid);

}

sstmac::sim_parameters::ptr
fork_a_run(int runid, int procid, sstmac::sim_parameters::ptr params,
    multisim::parameter_run &run, bool &goforit)
{
  //set signal handler
  // signal(SIGCHLD, child_done);

  int pid;
  switch (pid = fork())
  {
  case 0: /* a fork returns 0 to the child */
  {

    std::stringstream ss;
    ss << "sim_output-" << run.num_ << ".out";

    //std::ofstream myfile;

    multilog_.open(ss.str().c_str(), std::ios::out);

    if (!multilog_.is_open())
    {
      sst_throw(ioerror, "statlogger: cannot open file %s for writing", ss.str().c_str());
    }

    //_multisim_logs_[it2->first] = myfile;

    //debug::debug_stream_ = &_multisim_logs_[it2->first];
    debug::debug_stream_ = &multilog_;
    goforit = true;
    return run.params_;
  }

  default: /* a fork returns a pid to the parent */
    pidruns[pid] = run.num_;
    pidprocs[pid] = procid;
    break;

  case -1: /* something went wrong */
    perror("fork");
    exit(1);
  }

  goforit = false;
  return sstmac::sim_parameters::ptr();
}

sstmac::sim_parameters::ptr
do_multisim_mpi(sstmac::sim_parameters::ptr params,
    std::vector<multisim::parameter_run> &runs, int mpirank)
{
  // int num_free_procs = params->get_optional_int_param("multisim_nproc", 1);
  int num_free_procs = 1;

  if (mpirank == 0)
  {
    std::cout << "------------------------------------- \n";
    std::cout << "-> About to run " << num_free_procs
        << " simulations at a time on each rank, press enter to continue \n";
  }
  if (!params->has_param("multisim_nproc"))
  {
    if (mpirank == 0)
    {
      std::cout
          << "   (Warning: you have not set the multisim_nproc parameter, 1 is default)\n";
    }
  }

  if (mpirank == 0)
  {
    std::cin.ignore();
  }

  MPI_Barrier(MPI_COMM_WORLD);

  int mpisize;
  MPI_Comm_size(MPI_COMM_WORLD, &mpisize);

  int masterruncount = 0;
  int rundummy;

  if (mpirank == 0)
  {

    //first, fork my own and get them going
    for (int z = 0; z < num_free_procs && masterruncount < runs.size(); z++)
    {
      bool go;
      sstmac::sim_parameters::ptr p = fork_a_run(masterruncount, z, params,
          runs[masterruncount], go);

      if (go)
      { //this is the forked child
        return p;
      }
      masterruncount++;
    }

    //                   rank                      proc    req
    MPI_Request reqs[(mpisize - 1) * num_free_procs];

    //initialize correctly
    for (int z = 0; z < (mpisize - 1) * num_free_procs; z++)
    {
      reqs[z] = MPI_REQUEST_NULL;
    }

    int numwaiting = 0;

    std::cout << "master: spawning out, comm size is  " << mpisize << "\n";

    for (int i = 1; i < mpisize && masterruncount < runs.size(); i++)
    {
      std::cout << "master: considering rank " << i << "\n";

      for (int j = 0; j < num_free_procs && masterruncount < runs.size(); j++)
      {
        std::cout << "master: giving run " << masterruncount << " to rank "
            << i << " on proc " << j << "\n";
        MPI_Request req;
        MPI_Isend(&masterruncount, 1, MPI_INT, i, j, MPI_COMM_WORLD, &req);
        masterruncount++;

        MPI_Irecv(&rundummy, 1, MPI_INT, i, j, MPI_COMM_WORLD, &req);

        reqs[(i - 1) * num_free_procs + j] = req;
        numwaiting++;
      }
    }

    //keep spawning more when necessary
    while (masterruncount < runs.size() && numwaiting > 0)
    {
      int ind;
      MPI_Status stat;
      MPI_Waitany((mpisize - 1) * num_free_procs, reqs, &ind, &stat);

      int donerank = (ind / num_free_procs) + 1;
      int doneproc = ind % num_free_procs;

      std::cout << "master: got back index " << ind << ", which is donerank "
          << donerank << " and doneproc " << doneproc << "\n";

      MPI_Request req;
      MPI_Isend(&masterruncount, 1, MPI_INT, donerank, doneproc,
          MPI_COMM_WORLD, &req);
      masterruncount++;

      MPI_Irecv(&rundummy, 1, MPI_INT, donerank, doneproc, MPI_COMM_WORLD, &req);

      reqs[ind] = req;
    }

    //finish by sending all the ranks an exit code
    std::cout << "master: all done, cleaning up \n";
    int exit = -1;
    for (int i = 1; i < mpisize; i++)
    {
      MPI_Request req;

      MPI_Isend(&exit, 1, MPI_INT, i, 0, MPI_COMM_WORLD, &req);
    }

  }
  else
  {
    int* runtodo = (int*) malloc(sizeof(int) * num_free_procs);
    MPI_Request reqs[num_free_procs];

    for (int j = 0; j < num_free_procs; j++)
    {
      MPI_Request req;
      MPI_Irecv(&runtodo[j], 1, MPI_INT, 0, j, MPI_COMM_WORLD, &req);
      reqs[j] = req;
    }

    int activerun;
    do
    {
      MPI_Status stat;
      int index;
      MPI_Waitany(num_free_procs, reqs, &index, &stat);
      activerun = runtodo[index];
      if (activerun >= 0)
      {
        std::cout << "--- rank " << mpirank << " is running #" << activerun
            << " \n";

        bool go;
        sstmac::sim_parameters::ptr p = fork_a_run(activerun, index, params,
            runs[activerun], go);

        if (go)
        { //this is the forked child
        //  std::cout << "$$$$$$ here ya'll \n";
          return p;
        }

        child_done(0);

        //i'm the parent, wait for another
        MPI_Request req;
        MPI_Irecv(&runtodo[index], 1, MPI_INT, 0, index, MPI_COMM_WORLD, &req);
        reqs[index] = req;
      }
    }
    while (activerun >= 0);

    std::cout << "--- rank " << mpirank << " got the exit code, exiting \n";

  }

  MPI_Finalize();
  exit(0);
}

sstmac::sim_parameters::ptr
parsweepsim_manager::check_multisim(const opts &oo,
    sstmac::sim_parameters::ptr params)
{
  int rank;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);

  sim_parameters::ptr ret = params;

  if (params->is_multi_sim())
  {

    std::vector<multisim::parameter_run> runs;
    multisim::multisim_init(params, runs, (rank == 0));
    params = do_multisim_mpi(params, runs, rank);
    doingmultisim_ = true;
  }
  else if (rank != 0)
  {
    //std::cout
    //    << "WARNING: mpisweep is being used, but multiple parameter combinations were not found. Rank "
    //    << mpirank << " is exiting without doing anything \n";
    //return 0;
    sst_throw(ssterror, "using mpisweep, but no multisim configuration found");
  }

}

void
parsweepsim_manager::run(const opts &oo, sim_parameters::ptr params)
{
  native::driver_run(params, oo.stoptime, (oo.mode != "notime"));
}

int
main(int argc, char **argv)
{

  MPI_Init(&argc, &argv);

  int rc;

  try
  {
    boost::intrusive_ptr<parsweepsim_manager> simmgr =
        parsweepsim_manager::construct();
    rc = try_sstmacmain(argc, argv, simmgr);
    if (simmgr->doingmultisim_){
    //  std::cout << "$$$ finishin up ya'll \n";
      exit(0);
    }
  }
  catch (const std::exception &e)
  {
    std::cout << argv[0] << ": caught exception while setting up simulation:\n"
        << e.what() << "\n";
    return 1;
  }
  catch (...)
  {
    std::cout << argv[0]
        << ": caught unknown exception while setting up simulation\n";
    return 1;
  }

  MPI_Finalize();

  return rc;
}

