/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_ARMCI_ARMCI_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_ARMCI_ARMCI_H_INCLUDED

#include <sstmac/sstmacro.h>

typedef struct
{
  void **src_ptr_ar; //- Source starting addresses of each data segment.
  void **dst_ptr_ar; //- Destination starting addresses of each data segment.
  int bytes; //- The length of each segment in bytes.
  int ptr_ar_len; //- Number of data segment.
} armci_giov_t;

typedef long armci_hdl_t;
typedef size_t armci_size_t;
enum armci_domain_t {ARMCI_DOMAIN_SMP = 0};

#define ARMCI_SUCCESS 0


enum __ARMCI_types_
{
  ARMCI_INT = 0, ARMCI_LONG, ARMCI_FLOAT, ARMCI_DOUBLE
};

#ifdef __cplusplus
extern "C"
{
#endif

  // Initialization and Termination
  int
  ARMCI_Init();

  int
  ARMCI_Finalize();

  void
  ARMCI_Cleanup();

  void
  ARMCI_Error(char *message, int code);

  // Copy operations using the generalized I/O vector format

  int
  ARMCI_PutV(armci_giov_t *dsrc_arr, int arr_len, int proc);

  int
  ARMCI_NbPutV(armci_giov_t *dsrc_arr, int arr_len, int proc,
      armci_hdl_t* handle);

  int
  ARMCI_GetV(armci_giov_t *dsrc_arr, int arr_len, int proc);

  int
  ARMCI_NbGetV(armci_giov_t *dsrc_arr, int arr_len, int proc,
      armci_hdl_t* handle);

  // Copy operations using the strided format

  int
  ARMCI_PutS(void* src_ptr, int src_stride_ar[], void* dst_ptr,
      int dst_stride_ar[], int count[], int stride_levels, int proc);

  int
  ARMCI_NbPutS(void* src_ptr, int src_stride_ar[], void* dst_ptr,
      int dst_stride_ar[], int count[], int stride_levels, int proc,
      armci_hdl_t* handle);

  int
  ARMCI_GetS(void *src_ptr, int src_stride_ar[], void* dst_ptr,
      int dst_stride_ar[], int count[], int stride_levels, int proc);

  int
  ARMCI_NbGetS(void *src_ptr, int src_stride_ar[], void* dst_ptr,
      int dst_stride_ar[], int count[], int stride_levels, int proc,
      armci_hdl_t* handle);

  // Copy operations for contiguous data

  int
  ARMCI_Put(void* src, void* dst, int bytes, int proc);

  int
  ARMCI_NbPut(void* src, void* dst, int bytes, int proc, armci_hdl_t* handle);

  int
  ARMCI_Get(void* src, void* dst, int bytes, int proc);

  int
  ARMCI_NbGet(void* src, void* dst, int bytes, int proc, armci_hdl_t* handle);

  // Accumulate operation using the generalized I/O vector format

  int
  ARMCI_AccV(int datatype, void *scale, armci_giov_t *dsrc_arr, int arr_len,
      int proc, armci_hdl_t* handle);

  int
  ARMCI_NbAccV(int datatype, void *scale, armci_giov_t *dsrc_arr, int arr_len,
      int proc, armci_hdl_t* handle);

  // Accumulate operation using the strided format

  int
  ARMCI_AccS(int datatype, void *scale, void* src_ptr, int src_stride_ar[],
      void* dst_ptr, int dst_stride_ar[], int count[], int stride_levels,
      int proc);

  int
  ARMCI_NbAccS(int datatype, void *scale, void* src_ptr, int src_stride_ar[],
      void* dst_ptr, int dst_stride_ar[], int count[], int stride_levels,
      int proc, armci_hdl_t* handle);

  // Accumulate operation using contiguous format

  int
  ARMCI_Acc(int datatype, void *scale, void* src, void* dst, int bytes,
      int proc);

  int
  ARMCI_NbAcc(int datatype, void *scale, void* src, void* dst, int bytes,
      int proc, armci_hdl_t* handle);

  // Register Originated Data Transfer Operations

  int
  ARMCI_PutValueLong(long src, void* dst, int proc);

  int
  ARMCI_NbPutValueLong(long src, void* dst, int proc, armci_hdl_t* handle);

  long
  ARMCI_GetValueLong(void *src, int proc);

  int
  ARMCI_PutValueInt(int src, void* dst, int proc);

  int
  ARMCI_NbPutValueInt(int src, void* dst, int proc, armci_hdl_t* handle);

  int
  ARMCI_GetValueInt(void *src, int proc);

  int
  ARMCI_PutValueDouble(double src, void* dst, int proc);

  int
  ARMCI_NbPutValueDouble(double src, void* dst, int proc, armci_hdl_t* handle);

  double
  ARMCI_GetValueDouble(void *src, int proc);

  int
  ARMCI_PutValueFloat(float src, void* dst, int proc);

  int
  ARMCI_NbPutValueFloat(float src, void* dst, int proc, armci_hdl_t* handle);

  float
  ARMCI_GetValueFloat(void *src, int proc);

  // Completion of Outstanding Operations

  void
  ARMCI_Fence(int proc);

  void
  ARMCI_AllFence();

  int
  ARMCI_Wait(armci_hdl_t* handle);

  int
  ARMCI_Test(armci_hdl_t* handle);

  int
  ARMCI_WaitProc(int proc);

  int
  ARMCI_WaitAll();

  int
  ARMCI_Barrier();

  // Aggregated Data Transfer Operations

  void
  ARMCI_SET_AGGREGATE_HANDLE(armci_hdl_t* handle);

  void
  ARMCI_UNSET_AGGREGATE_HANDLE(armci_hdl_t* handle);

  // Atomic and Synchronization Operations

  int
  ARMCI_Rmw(int op, void *ploc, void *prem, int value, int proc);

  int
  ARMCI_Create_mutexes(int count);

  int
  ARMCI_Destroy_mutexes(void);

  void
  ARMCI_Lock(int mutex, int proc);

  void
  ARMCI_Unlock(int mutex, int proc);

  // Memory Management

  int
  ARMCI_Malloc(void* ptr[], armci_size_t bytes);

  void*
  ARMCI_Malloc_local(armci_size_t bytes);

  int
  ARMCI_Free(void *address);

  int
  ARMCI_Free_local(void *address);

  // Collective Operations
  void
  armci_msg_brdcst(void* buffer, int len, int root);

  void
  armci_msg_lgop(void *x, int n, int type, char* op);

  void
  armci_msg_igop(void *x, int n, int type, char* op);

  void
  armci_msg_dgop(void *x, int n, int type, char* op);

  void
  armci_msg_fgop(void *x, int n, int type, char* op);

  void
  armci_msg_barrier(void);

  void
  armci_msg_reduce(void *x, int n, char* op, int type, int root);

  // System Configuration

  int armci_msg_me();
  int armci_msg_nproc();

  int
  armci_domain_nprocs(armci_domain_t domain, int id);

  int
  armci_domain_count(armci_domain_t domain);
  int
  armci_domain_id(armci_domain_t domain, int glob_proc_id);

  int
  armci_domain_glob_proc_id(armci_domain_t domain, int id, int loc_proc_id);

#ifdef __cplusplus
}
#endif

#endif

