/* -*- c-basic-offset: 8; indent-tabs-mode: nil -*- */
/*
        Contains the user interface to the NIC. Kernel and User level.

        Copyright (C) 2007 Cray Inc. All Rights Reserved.
        Written by Igor Gorodetsky <igorodet@cray.com>

        This program is free software; you can redistribute it and/or modify it 
        under the terms of the GNU General Public License as published by the 
        Free Software Foundation; either version 2 of the License, 
        or (at your option) any later version.

        This program is distributed the hope that it will be useful,
        but WITHANY WARRANTY; witheven the implied warranty of
        MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
        See the GNU General Public License for more details.

        You should have received a copy of the GNU General Public License 
        along with this program; if not, write to the Free Software Foundation, Inc., 
        51 FranklSt, Fifth Floor, Boston, MA 02110-1301 USA
*/

#ifndef SSTMACRO_SOFTWARE_LIBARIES_GNI_PUB_H
#define SSTMACRO_SOFTWARE_LIBARIES_GNI_PUB_H

#include <sys/types.h>
#include <stdint.h>

#ifdef __cplusplus
extern "C"
{
#endif

/* Registered memory handle */
typedef uint64_t gni_mem_handle_t;

/* Opaque handles */
typedef uint64_t gni_nic_handle_t;
typedef uint64_t gni_cdm_handle_t;
typedef uint64_t gni_ep_handle_t;
typedef uint64_t gni_cq_handle_t;
typedef uint64_t gni_err_handle_t;
typedef uint64_t gni_msgq_handle_t;

/* Short messaging types */
typedef enum gni_smsg_type {
        GNI_SMSG_TYPE_INVALID = 0,
        GNI_SMSG_TYPE_MBOX,
        GNI_SMSG_TYPE_MBOX_AUTO_RETRANSMIT
} gni_smsg_type_t;

#define GNI_SMSG_ANY_TAG 0xFF

/* Short messaging attributes */
typedef struct gni_smsg_attr {
        gni_smsg_type_t         msg_type;
        void                    *msg_buffer;
        uint32_t                buff_size;
        gni_mem_handle_t        mem_hndl;
        uint32_t                mbox_offset;
        uint16_t                mbox_maxcredit;
        uint32_t                msg_maxsize;
} gni_smsg_attr_t;

/* Maximum SMSG retransmit count default values */

#define FMA_SMSG_MAX_RETRANS_DEFAULT	10

/* Return codes */
typedef enum gni_return {
        GNI_RC_SUCCESS = 0,
        GNI_RC_NOT_DONE,
        GNI_RC_INVALID_PARAM,
        GNI_RC_ERROR_RESOURCE,
        GNI_RC_TIMEOUT,
        GNI_RC_PERMISSION_ERROR,
        GNI_RC_DESCRIPTOR_ERROR,
        GNI_RC_ALIGNMENT_ERROR,
        GNI_RC_INVALID_STATE,
        GNI_RC_NO_MATCH,
        GNI_RC_SIZE_ERROR,
        GNI_RC_TRANSACTION_ERROR,
        GNI_RC_ILLEGAL_OP,
        GNI_RC_ERROR_NOMEM
} gni_return_t;

/* Enable transmitting RDMA posts through one BTE channel, instead of
 * defaulting to using all three channels. This may be preferred for
 * some applications.
 */
#define GNI_CDM_MODE_BTE_SINGLE_CHANNEL 0x00000040

/* By default, users may post transactions with either local or global completion 
 * notification, not both.  If receipt of both local and global events is requested
 * users must set DUAL_EVENTS.  Performing a post operation with local and global
 * events enabled withDUAL_EVENTS set will yield an error GNI_RC_INVALID_PARAM.
 * 
 * addition, during an EpBind default mode, transfer requests are allocated
 * equal size to the number of events the associated source CQ.  When
 * DUAL_EVENTS is set transfer requests are allocated 1 per 2 CQ event slots.
 * Therefore, a user is limited to posting half as many transactions as CQ events
 * when DUAL_EVENTS is set.  Exceeding this limit will yield an error
 * GNI_RC_ERROR_RESOURCE.
 */
#define GNI_CDM_MODE_DUAL_EVENTS	0x00001000

#define GNI_CDM_MODE_CQ_GNI_LOCAL_PLACEMENT 0x00080000

/* Endpoint machine state */
typedef enum gni_post_state{
        GNI_POST_PENDING,
        GNI_POST_COMPLETED,
        GNI_POST_ERROR,
        GNI_POST_TIMEOUT,
        GNI_POST_TERMINATED,
        GNI_POST_REMOTE_DATA
} gni_post_state_t;

/* Instruct NIC to enforce strict PI (Processor Interface, e.g. HT) ordering for the 
   region by disabling "Non-Posted Pass Posted Writes" rule */
#define GNI_MEM_STRICT_PI_ORDERING      0x00000020
/* Instruct NIC to issue PI (Processor Interface, e.g. HT) FLUSH command prior 
   to sending network responses for the region */
#define GNI_MEM_PI_FLUSH                0x00000040
/* Instruct NIC to allow relaxed ordering for Non-Posted and Posted write requests into the processor
   by enabling both "Non-Posted Pass Posted Writes" and "Posted Pass Posted Writes" rules.
   (only "Non-Posted Pass Posted Writes" rule is enabled by default)
   This attribute should be used with caution as it may effect ordering of the transactions.
   It gets overridden by GNI_MEM_STRICT_PI_ORDERING attribute */
#define GNI_MEM_RELAXED_PI_ORDERING     0x00000100

/* CQ modes of operation */
typedef enum gni_cq_mode {
        GNI_CQ_NOBLOCK = 0,
        GNI_CQ_BLOCKING
} gni_cq_mode_t;

/* Macros and enum for processing data component of CQEs associated with
   PostRDMA, PostFma, Short message transactions */

/* Completion queue entry (size of type field is 2 bits) */
#define GNI_CQ_EVENT_TYPE_POST  0x0ULL
#define GNI_CQ_EVENT_TYPE_SMSG  0x1ULL
#define GNI_CQ_EVENT_TYPE_DMAPP 0x2ULL
#define GNI_CQ_EVENT_TYPE_MSGQ  0x3ULL
typedef uint64_t gni_cq_entry_t;

uint64_t gni_cq_get_data(gni_cq_entry_t);
uint64_t gni_cq_get_source(gni_cq_entry_t);
uint64_t gni_cq_get_status(gni_cq_entry_t);
uint64_t gni_cq_get_info(gni_cq_entry_t);
uint64_t gni_cq_overrun(gni_cq_entry_t);
uint64_t gni_cq_get_inst_id(gni_cq_entry_t);
uint64_t gni_cq_get_tid(gni_cq_entry_t);
uint64_t gni_cq_get_msg_id(gni_cq_entry_t);
uint64_t gni_cq_get_type(gni_cq_entry_t);
uint64_t gni_cq_get_block_id(gni_cq_entry_t);
uint64_t gni_cq_get_unsuccessful_cnt(gni_cq_entry_t);
uint64_t gni_cq_get_marker_id(gni_cq_entry_t);
void gni_cq_set_inst_id(gni_cq_entry_t *,uint64_t);
void gni_cq_set_tid(gni_cq_entry_t *,uint64_t);
void gni_cq_set_msg_id(gni_cq_entry_t *,uint64_t);
void gni_cq_set_type(gni_cq_entry_t *,uint64_t);
void gni_cq_clr_status(gni_cq_entry_t *);

#define GNI_CQ_GET_DATA    gni_cq_get_data
#define GNI_CQ_GET_SOURCE  gni_cq_get_source
#define GNI_CQ_GET_STATUS  gni_cq_get_status
#define GNI_CQ_GET_INFO    gni_cq_get_info
#define GNI_CQ_GET_INST_ID gni_cq_get_inst_id
#define GNI_CQ_GET_TID     gni_cq_get_tid
#define GNI_CQ_GET_MSG_ID  gni_cq_get_msg_id
#define GNI_CQ_GET_TYPE    gni_cq_get_type
#define GNI_CQ_OVERRUN     gni_cq_overrun
#define GNI_CQ_GET_BLOCK_ID           gni_cq_get_block_id
#define GNI_CQ_GET_UNSUCCESSFUL_CNT   gni_cq_get_unsuccessful_cnt
#define GNI_CQ_GET_MARKER_ID          gni_cq_get_marker_id
#define GNI_CQ_SET_INST_ID(entry,val) gni_cq_set_inst_id(&(entry),val)
#define GNI_CQ_SET_TID(entry,val)     gni_cq_set_tid(&(entry),val)
#define GNI_CQ_SET_MSG_ID(entry,val)  gni_cq_set_msg_id(&(entry),val)
#define GNI_CQ_SET_TYPE(entry,val)    gni_cq_set_type(&(entry),val)
#define GNI_CQ_CLR_STATUS(entry)      gni_cq_clr_status(&(entry))
#define GNI_CQ_STATUS_OK(entry)      (gni_cq_get_status(entry) == 0)
#define GNI_CQ_STATUS_DLA_OVERFLOW(entry)   (gni_cq_status_dla_overflow(entry))

/* Transaction types (for type field of post descriptor) */
typedef enum gni_post_type {
        GNI_POST_RDMA_PUT = 1,
        GNI_POST_RDMA_GET,
        GNI_POST_FMA_PUT,
        GNI_POST_FMA_PUT_W_SYNCFLAG,
        GNI_POST_FMA_GET,
        GNI_POST_AMO,
        GNI_POST_CQWRITE
} gni_post_type_t;

/* FMA command types (for amo_cmd field of post descriptor) */
typedef enum gni_fma_cmd_type {
        /************ AMOs with GET semantics **************/
        GNI_FMA_ATOMIC_FADD    = 0x008,    /* atomic FETCH and ADD */
        GNI_FMA_ATOMIC_FADD_C  = 0x018,    /* cached atomic FETCH and ADD */
        GNI_FMA_ATOMIC_FAND    = 0x009,    /* atomic FETCH and AND */
        GNI_FMA_ATOMIC_FAND_C  = 0x019,    /* cached atomic FETCH and AND */
        GNI_FMA_ATOMIC_FOR     = 0x00A,    /* atomic FETCH and OR */
        GNI_FMA_ATOMIC_FOR_C   = 0x01A,    /* cached atomic FETCH and OR */
        GNI_FMA_ATOMIC_FXOR    = 0x00B,    /* atomic FETCH and XOR */
        GNI_FMA_ATOMIC_FXOR_C  = 0x01B,    /* cached atomic FETCH and XOR */
        GNI_FMA_ATOMIC_FAX     = 0x00C,    /* atomic FETCH AND exclusive OR */
        GNI_FMA_ATOMIC_FAX_C   = 0x01C,    /* cached atomic FETCH AND exclusive OR */
        GNI_FMA_ATOMIC_CSWAP   = 0x00D,    /* atomic COMPARE and SWAP */
        GNI_FMA_ATOMIC_CSWAP_C = 0x01D,    /* cached atomic COMPARE and SWAP */
        /* Second generation commands ( GET sematics ) */
        GNI_FMA_ATOMIC2_FAND     = 0x041,    /* atomic FETCH and AND */
        GNI_FMA_ATOMIC2_FAND_C   = 0x061,    /* cached atomic FETCH and AND */
        GNI_FMA_ATOMIC2_FOR      = 0x043,    /* atomic FETCH and OR */
        GNI_FMA_ATOMIC2_FOR_C    = 0x063,    /* cached atomic FETCH and OR */
        GNI_FMA_ATOMIC2_FXOR     = 0x045,    /* atomic FETCH exclusive OR */
        GNI_FMA_ATOMIC2_FXOR_C   = 0x065,    /* cached atomic FETCH exclusive OR */
        GNI_FMA_ATOMIC2_FSWAP    = 0x047,    /* atomic FETCH and SWAP */
        GNI_FMA_ATOMIC2_FSWAP_C  = 0x067,    /* cached atomic FETCH and SWAP */
        GNI_FMA_ATOMIC2_FAX      = 0x049,    /* atomic FETCH AND exclusive OR */
        GNI_FMA_ATOMIC2_FAX_C    = 0x069,    /* cached atomic FETCH AND exclusive OR */
        GNI_FMA_ATOMIC2_FCSWAP   = 0x04B,    /* atomic Fetching COMPARE and SWAP */
        GNI_FMA_ATOMIC2_FCSWAP_C = 0x06B,    /* cached atomic Fetching COMPARE and SWAP */
        GNI_FMA_ATOMIC2_FIM   = 0x051,    /* atomic Fetching integer signed two's complement Minimum */
        GNI_FMA_ATOMIC2_FIMIN_C  = 0x071,    /* cached atomic Fetching integer signed two's complement Minimum */
        GNI_FMA_ATOMIC2_FIMAX    = 0x053,    /* atomic Fetching integer signed two's complement Maximum */
        GNI_FMA_ATOMIC2_FIMAX_C  = 0x073,    /* cached atomic Fetching integer signed two's complement Maximum */
        GNI_FMA_ATOMIC2_FIADD    = 0x055,    /* atomic Fetching integer two's complement Addition */
        GNI_FMA_ATOMIC2_FIADD_C  = 0x075,    /* cached atomic Fetching integer two's complement Addition */
        GNI_FMA_ATOMIC2_FFPM  = 0x059,    /* atomic Fetching floating point Minimum (double precision) */
        GNI_FMA_ATOMIC2_FFPMIN_C = 0x079,    /* cached atomic Fetching floating point Minimum (double precision) */
        GNI_FMA_ATOMIC2_FFPMAX   = 0x05B,    /* atomic Fetching floating point Maximum (double precision) */
        GNI_FMA_ATOMIC2_FFPMAX_C = 0x07B,    /* cached atomic Fetching floating point Maximum (double precision) */
        GNI_FMA_ATOMIC2_FFPADD   = 0x05D,    /* atomic Fetching floating point Addition (double precision) */
        GNI_FMA_ATOMIC2_FFPADD_C = 0x07D,    /* cached atomic Fetching floating point Addition (double precision) */
        /************ AMOs with PUT semantics ***************/
        GNI_FMA_ATOMIC_ADD     = 0x108,    /* atomic AND */
        GNI_FMA_ATOMIC_ADD_C   = 0x118,    /* cached atomic ADD */
        GNI_FMA_ATOMIC_AND     = 0x109,    /* atomic AND */
        GNI_FMA_ATOMIC_AND_C   = 0x119,    /* cached atomic AND */
        GNI_FMA_ATOMIC_OR      = 0x10A,    /* atomic OR */
        GNI_FMA_ATOMIC_OR_C    = 0x11A,    /* cached atomic OR */
        GNI_FMA_ATOMIC_XOR     = 0x10B,    /* atomic exclusive OR */
        GNI_FMA_ATOMIC_XOR_C   = 0x11B,    /* cached atomic exclusive OR */
        GNI_FMA_ATOMIC_AX      = 0x10C,    /* atomic AND exclusive OR */
        GNI_FMA_ATOMIC_AX_C    = 0x11C,    /* cached atomic AND exclusive OR */
        /* Second generation commands ( PUT sematics ) */
        GNI_FMA_ATOMIC2_AND     = 0x141,    /* atomic AND */
        GNI_FMA_ATOMIC2_AND_C   = 0x161,    /* cached atomic AND */
        GNI_FMA_ATOMIC2_OR      = 0x143,    /* atomic  OR */
        GNI_FMA_ATOMIC2_OR_C    = 0x163,    /* cached atomic  OR */
        GNI_FMA_ATOMIC2_XOR     = 0x145,    /* atomic exclusive OR */
        GNI_FMA_ATOMIC2_XOR_C   = 0x165,    /* cached atomic exclusive OR */
        GNI_FMA_ATOMIC2_SWAP    = 0x147,    /* atomic SWAP */
        GNI_FMA_ATOMIC2_SWAP_C  = 0x167,    /* cached atomic SWAP */
        GNI_FMA_ATOMIC2_AX      = 0x149,    /* atomic AND exclusive OR */
        GNI_FMA_ATOMIC2_AX_C    = 0x169,    /* cached atomic AND exclusive OR */
        GNI_FMA_ATOMIC2_CSWAP   = 0x14B,    /* atomic COMPARE and SWAP */
        GNI_FMA_ATOMIC2_CSWAP_C = 0x16B,    /* cached atomic COMPARE and SWAP */
        GNI_FMA_ATOMIC2_IM   = 0x151,    /* atomic integer signed two's complement Minimum */
        GNI_FMA_ATOMIC2_IMIN_C  = 0x171,    /* cached atomic integer signed two's complement Minimum */
        GNI_FMA_ATOMIC2_IMAX    = 0x153,    /* atomic integer signed two's complement Maximum */
        GNI_FMA_ATOMIC2_IMAX_C  = 0x173,    /* cached atomic integer signed two's complement Maximum */
        GNI_FMA_ATOMIC2_IADD    = 0x155,    /* atomic integer two's complement Addition */
        GNI_FMA_ATOMIC2_IADD_C  = 0x175,    /* cached atomic integer two's complement Addition */
        GNI_FMA_ATOMIC2_FPM  = 0x159,    /* atomic floating point Minimum (double precision) */
        GNI_FMA_ATOMIC2_FPMIN_C = 0x179,    /* cached atomic floating point Minimum (double precision) */
        GNI_FMA_ATOMIC2_FPMAX   = 0x15B,    /* atomic floating point Maximum (double precision) */
        GNI_FMA_ATOMIC2_FPMAX_C = 0x17B,    /* cached atomic floating point Maximum (double precision) */
        GNI_FMA_ATOMIC2_FPADD   = 0x15D,    /* atomic floating point Addition (double precision) */
        GNI_FMA_ATOMIC2_FPADD_C = 0x17D    /* cached atomic floating point Addition (double precision) */
} gni_fma_cmd_type_t;

/* CQ event types */
#define GNI_CQMODE_SILENT       0x0000
#define GNI_CQMODE_LOCAL_EVENT  0x0001
#define GNI_CQMODE_GLOBAL_EVENT 0x0002
#define GNI_CQMODE_REMOTE_EVENT 0x0004
#define GNI_CQMODE_DUAL_EVENTS	( GNI_CQMODE_LOCAL_EVENT | GNI_CQMODE_GLOBAL_EVENT )

/* Delivery modes */
#define GNI_DLVMODE_PERFORMANCE 0x0000
#define GNI_DLVMODE_NO_ADAPT    0x0001
#define GNI_DLVMODE_NO_HASH     0x0002
#define GNI_DLVMODE_NO_RADAPT   0x0004
#define GNI_DLVMODE_IN_ORDER    ( GNI_DLVMODE_NO_ADAPT | GNI_DLVMODE_NO_HASH )

/* Error Event Categories */
/* WARNING: DO NOT CHANGE THESE UNLESS YOU CHANGE ghal_err_cat.h */
#define GNI_ERRMASK_CORRECTABLE_MEMORY   (1 << 0)
#define GNI_ERRMASK_CRITICAL             (1 << 1)
#define GNI_ERRMASK_TRANSACTION          (1 << 2)
#define GNI_ERRMASK_ADDRESS_TRANSLATION  (1 << 3)
#define GNI_ERRMASK_TRANSIENT            (1 << 4)
#define GNI_ERRMASK_INFORMATIONAL        (1 << 5)
#define GNI_ERRMASK_DIAG_ONLY            (1 << 6)

/* RDMA mode */
/* local_addr is a physical address (kernel only) */
#define GNI_RDMAMODE_PHYS_ADDR  0x0001
/* instruction to Gemini to wait for all responses from this post and all
 * previous posts before processing the next RDMA descriptor */
#define GNI_RDMAMODE_FENCE      0x0002

/* Maximum length bytes of a datagram transaction */
#define GNI_DATAGRAM_MAXSIZE	128

/* Transaction descriptor */
typedef struct gni_post_descriptor {
        /********************** Control **********************/
        /* points to the next descriptor the link list */
        gni_post_descriptor* next_descr;
        /* points to the previous descriptor the link list */
        gni_post_descriptor* prev_descr;
        /* holds an ID of the transaction assigned by the user */
        uint64_t post_id;    
        /* error status of the transaction */
        uint64_t status;        
        /* completion flag of the transaction */
        uint16_t cq_mode_complete;
        /********************** Common ***********************/
        /* type of the transaction */
        gni_post_type_t type;
        /* instruction to generate CQ events of the following types 
           (see GNI_CQMODE_xxx)*/
        uint16_t cq_mode;       
        /* delivery mode (see GNI_DLVMODE_xxx) */
        uint16_t dlvr_mode;     
        /* address of region on the local node: source for Put, target for Get */
        uint64_t local_addr;    
        /* local memory handle */ 
        gni_mem_handle_t local_mem_hndl;
        /* address of the remote region: target for Put, source for Get */
        uint64_t remote_addr;   
        /* remote memory handle */
        gni_mem_handle_t remote_mem_hndl;
        /* number of bytes to move during the transaction */
        uint64_t length;        
        /****************** RDMA specific ********************/
        /* see GNI_RDMAMODE_xxx */
        uint16_t rdma_mode;     
        /* defines src. CQ for the transaction */
        gni_cq_handle_t src_cq_hndl;
        /************ FMA and AMO specific *******************/
        /* synchronization value */
        uint64_t sync_flag_value;      
        /* location to deliver sync. value */
        uint64_t sync_flag_addr;     
        /****************** AMO specific *********************/
        /* AMO command for the transaction */
        gni_fma_cmd_type_t amo_cmd;
        /* first operand required by the AMO command */
        uint64_t first_operand; 
        /* second operand required by the AMO command */
        uint64_t second_operand;
} gni_post_descriptor_t;

/**
 * GNI_CdmCreate - Create Communication Domain
 *
 * Parameters:
 * IN
 * inst_id  Instance of the cdm the job (user level).
 *          Unique address of the instance withthe upper layer
 *          protocol doma(kernel level).
 * ptag     Protection Tag.
 * cookie   Unique identifier generated by ALPS. Along with ptag
 *          helps to identify the Communication Domain.
 * modes    bit mask (see GNI_CDM_MODE_xxxxxx definitions)
 *
 * OUT
 * cdm_hndl     Handle returned. The handle is used with the other functions
 *      to specify a particular instance of the Communication Domain.
 *
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - One of the input parameters was invalid.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 * 
 * Description:
 * This function creates an instance of the Communication Domain.
 **/
gni_return_t
GNI_CdmCreate(
    uint32_t inst_id,
    uint8_t ptag,
    uint32_t cookie,
    uint32_t modes,
    gni_cdm_handle_t *cdm_hndl
);

/**
 * GNI_CdmDestroy - Destroys the instance of a Communication Domain
 *
 * Parameters:
 * IN
 * cdm_hndl   Communication DomaHandle
 *
 * Returns:
 * GNI_RC_SUCCESS - The operation completioned succesffully
 * GNI_RC_INVALID_PARAM - Caller specified an invalid Communication DomaHandle
 *
 * Description:
 * Destroys the instance of a Communication Domain.  Removes associations
 * between the calling process and the NIC devices that were established via
 * the corresponding Attach function.
 **/
gni_return_t
GNI_CdmDestroy(gni_cdm_handle_t cdm_hndl);

/**
 * GNI_CdmGetNicAddress - Get the PE address of a NIC device.
 *
 * Parameters:
 * IN
 * device_id    The ID of the NIC device to query.
 *
 * OUT
 * address      The PE address of the NIC device queried.
 * cpu_id       The ID of the first CPU directly connected to the NIC device.
 *
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_NO_MATCH - Specified device_id does not exists.
 *
 * Description:
 *
 * Returns the PE address of the NIC device with ID device_id and the ID of
 * it's most closely connected CPU.
 **/
gni_return_t
GNI_CdmGetNicAddress(
    uint32_t device_id,
    uint32_t *address,
    uint32_t *cpu_id
);

/**
 * GNI_CdmAttach - Attach Communication Domato a NIC device
 * 
 * Parameters:
 * IN
 * cdm_hndl     The Communication DomaHandle.
 * device_id    The device identifier , e.g. /dev/knic1 has
 *              device_id = DEVICE_MINOR_NUMBER - GEMINI_BASE_MINOR_NUMBER = 1
 *              Setting device_id to (-1) will result attaching to the nearest
 *              Gemini NIC.
 * 
 * OUT
 * local_addr   PE address of the Gemini NIC attached
 * gni_hndl     Handle returned. The handle is used with the other functions to specify
 *              a particular instance of a Gemini NIC.
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - Caller specified an invalid CDM handle.
 * GNI_RC_NO_MATCH - Specified device_id does not exists
 * GNI_RC_ERROR_RESOURCE - The operation failed due to insufficient resources.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 * GNI_RC_PERMISSION_ERROR - Insufficient permissions to perform operation.
 * GNI_RC_INVALID_STATE - Caller attempts to attach the same CDM instance to
 *                        the same Gemini NIC device more than once.
 *                        If returned while device_id= -1, means that there
 *                        are no more devices left for this CDM to attach to.
 *
 * Description:
 * Associates the Communication Domawith a Gemini NIC and provides a
 * NIC handle to the upper layer protocol. A process is not allowed
 * to attach the same CDM instance to the same Gemini NIC more than once,
 * but it is allowed to attach multiple CDMs to the same Gemini NIC.
 **/
gni_return_t
GNI_CdmAttach(
    gni_cdm_handle_t cdm_hndl,
    uint32_t device_id,
    uint32_t *local_addr,
    gni_nic_handle_t *gni_hndl
);

/**
 * GNI_EpCreate - Create logical Endpoint
 * 
 * Parameters:
 * IN
 * gni_hndl     Handle of the associated Gemini NIC.
 * src_cq_hndl  Handle of the CQ that will be used by default to deliver events
 *              related to the transactions initiated by the local node.
 * 
 * OUT
 * ep_hndl      The handle of the newly created Endpoint instance.
 * 
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - One of the input parameters was invalid.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 *
 * Description:
 * This function creates an instance of a Logical Endpoint. 
 * A new instance is always created a non-bound state.
 * A non-bound Endpoint is able to exchange posted data with 
 * any bound remote Endpoint withthe same Communication Domain.
 * An Endpoint cannot be used to post RDMA, FMA transactions or 
 * send short messages while it is non-bound state.
 **/
gni_return_t
GNI_EpCreate(
    gni_nic_handle_t gni_hndl,
    gni_cq_handle_t src_cq_hndl,
    gni_ep_handle_t *ep_hndl
);

/**
 * GNI_EpBind - Bind logical Endpoint to a peer
 *
 * Parameters:
 * IN
 * ep_hndl      The handle of the Endpoint instance to be bound.
 * remote_addr  Physical address of the Gemini NIC at the remote peer or NTT index,
 *              when NTT is enabled for the given Communication Domain.
 * remote_id    User specified ID of the remote instance the job or unique identifier of
 *              the remote instance withthe upper layer protocol domain.
 *
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - One of the input parameters was invalid.
 * GNI_RC_ERROR_RESOURCE - The operation failed due to insufficient resources.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 *
 * Description:
 * This function binds a Logical Endpoint to the specific remote address
 * and remote instance the Communication Domain.
 * Once bound the Endpoint can be used to post RDMA and FMA transactions.
 **/
gni_return_t
GNI_EpBind(
    gni_ep_handle_t ep_hndl,
    uint32_t remote_addr,
    uint32_t remote_id
);

/**
 * GNI_EpUnbind - Unbind logical Endpoint
 *
 * Parameters:
 * IN
 * ep_hndl      The handle of the Endpoint instance to be bound.
 *
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - One of the input parameters was invalid.
 * GNI_RC_NOT_DONE - Operation is not permited
 *
 * Description:
 * This function unbinds a Logical Endpoint from the specific remote address
 * and remote instance and releases any internal short message resource.
 * A non-bound Endpoint is able to exchange posted data with
 * any bound remote Endpoint withthe same Communication Domain.
 * An Endpoint cannot be used to post RDMA, FMA transactions or
 * send short messages while it is non-bound state.
 **/
gni_return_t
GNI_EpUnbind(gni_ep_handle_t ep_hndl);

/**
 * GNI_EpDestroy - Destroy logical Endpoint
 * 
 * Parameters:
 * IN
 * ep_hndl      The handle of the Endpoint instance to be destroyed.
 * 
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - An invalid EP handle was specified.
 *
 * Description:
 * This function tears down an Endpoint.
 **/
gni_return_t
GNI_EpDestroy(gni_ep_handle_t ep_hndl);

/**
 * GNI_MemRegister - Register memory with the NIC
 *
 * Parameters:
 * IN
 * gni_hndl     Handle of a currently open NIC.
 * address      Starting address of the memory region to be registered.
 * length       Length of the memory region to be registered, bytes.
 * dst_cq_hndl  If not NULL, it will be used to notify the local process
 *              that a remote peer has delivered data from RDMA or FMA PUT
 *              into this memory region.
 * flags        Memory attributes associated with the region
 *              (see GNI_MEM_xxx gni_puh.h)
 * vmdh_index   Specifies the index withthe pre-allocated MDD block that
 *              must be used for the registration, e.g. when set to 0 will
 *              use the first entry of the MDD block. If set to (-1) relies
 *              on NIC library to allocate the next available entry from
 *              the MDD block.
 *
 * INOUT
 * mem_hndl     The new memory handle for the region.
 *
 * Returns:
 * GNI_RC_SUCCESS - The memory region was successfully registered.
 * GNI_RC_INVALID_PARAM - One on the parameters was invalid.
 * GNI_RC_ERROR_RESOURCE - The registration operation failed due to
 *                         insufficient resources.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 * GNI_RC_PERMISSION_ERROR - The user's buffer R/W permissions conflict with
 *                           the flags argument.
 *
 * Description:
 * This function allows a process to register a region of memory with
 * the Gemini NIC. The user may specify an arbitrary size region of memory,
 * with arbitrary alignment, but the actual area of memory registered will
 * be registered on MRT block granularity (or physical page granularity if
 * MRT is not enabled for this process).
 * A memory region must consist of a single segment.
 * Using a single segment to register a memory region allows an application
 * to use a virtual address the future transactions and of the
 * registered region. A single segment memory registration should be a common
 * way which an application registers its memory.
 * A new memory handle is generated for each region of memory that
 * is registered by a process.
 * The contents of the memory region being registered are not altered.
 * The memory region must be previously allocated by an application.
 **/
gni_return_t
GNI_MemRegister(
    gni_nic_handle_t gni_hndl,
    uint64_t address,
    uint64_t length,
    gni_cq_handle_t dst_cq_hndl,
    uint32_t flags,
    uint32_t vmdh_index,
    gni_mem_handle_t *mem_hndl
);

/**
 * GNI_MemDeregister - De-register memory
 *
 * Parameters:
 * IN
 * gni_hndl  The handle for the NIC that owns the memory region
 *           being de-registered.
 * mem_hndl  Memory handle for the region.
 *
 * Returns:
 * GNI_RC_SUCCESS - The memory region was successfully de-registered.
 * GNI_RC_INVALID_PARAM - One or more of the parameters was invalid.
 *
 * Description:
 * This function de-registers memory that was previously registered and unlocks
 * the associated pages from physical memory. The contents and attributes of the
 * region of memory being de-registered are not altered any way.
 **/
gni_return_t
GNI_MemDeregister(
    gni_nic_handle_t gni_hndl,
    gni_mem_handle_t *mem_hndl
);

/**
 * GNI_CqCreate - Create Completion Queue
 *
 * Parameters:
 * IN
 * gni_hndl     The handle of the associated NIC.
 * entry_count  The minimum number of completion entries that this CQ will hold.
 * delay_count  The number of events the Gemini will allow to occur before
 *              generating an interrupt.
 *              Setting this to zero results interrupt delivery with every event.
 *              For the user level this parameter is meaningful only when mode is
 *              set to GNI_CQ_BLOCKING
 * mode         The mode of operation of the new CQ: GNI_CQ_BLOCKING, GNI_CQ_NOBLOCK
 * handler      User supplied callback function to be run for each CQ entry received
 *              the CQ.
 * context      User supplied pointer to be passed to the handler callback function.
 *
 * OUT
 * cq_hndl      The handle of the newly created Completion Queue.
 *
 * Returns:
 * GNI_RC_SUCCESS - A new Completion Queue was successfully created.
 * GNI_RC_INVALID_PARAM - One or more of the parameters was invalid.
 * GNI_RC_ERROR_RESOURCE - The Completion Queue could not be created due
 *                         to insufficient resources.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 *
 * Description:
 * This function creates a new Completion Queue. The caller must specify
 * the minimum number of completion entries that the queue must contain.
 * To avoid dropped completion notifications, applications should make sure
 * that the number of operations posted on Endpoints attached to
 * a src_cq_hndl does not exceed the completion queue capacity at any time.
 *
 * Notes:
 * The handler, if specified, runs for each CQ entry that is received into
 * the CQ.  The handler is supplied with two arguments, a pointer to the
 * CQ entry, and a pointer to the context provided at CQ creation.
 * The handler is invoked at some time between when the CQ entry is deposited
 * into the CQ and the return of GNI_CqGetEvent or GNI_CqWaitEvent with
 * a status of either GNI_RC_SUCCESS or GNI_RC_TRANSACTION_ERROR.
 * Use of callback functions does not relieve the  user of the need to call
 * GNI_CqGetEvent or GNI_CqWaitEvent for each event deposited into the CQ.
 *
 * Completion Queues may be used for receipt of locally generated events
 * such as those arising from GNI_Post style transactions, etc. or
 * may be used for receipt of remote events, but not both.
 **/
gni_return_t
GNI_CqCreate(
    gni_nic_handle_t gni_hndl,
    uint32_t entry_count,
    uint32_t delay_count,
    gni_cq_mode_t mode,
    void (*handler)(gni_cq_entry_t*,void*),
    void *context,
    gni_cq_handle_t *cq_hndl
);

/**
 * GNI_CqDestroy - Destroy Completion queue
 *
 * Parameters:
 * IN
 * cq_hndl    The handle for the Completion Queue to be destroyed.
 *
 * Returns:
 * GNI_RC_SUCCESS        - The CQ was successfully destroyed.
 * GNI_RC_INVALID_PARAM  - One or more of the parameters was invalid.
 * GNI_RC_ERROR_RESOURCE - The CQ could not be destroyed because one or
 *                         more Endpoint instances are still associated with it.
 *
 * Description:
 * This function destroys a specified Completion Queue.
 * If any Endpoints are associated with the CQ, the CQ is not destroyed and
 * an error is returned.
 **/
gni_return_t
GNI_CqDestroy(gni_cq_handle_t cq_hndl);

/**
 * GNI_PostRdma - Post RDMA transaction
 * 
 * Parameters:
 * IN
 * ep_hndl      Instance of a local Endpoint.
 * post_descr   Pointer to a descriptor to be posted.
 *
 * Returns:
 * GNI_RC_SUCCESS - The descriptor was successfully posted.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid.
 * GNI_RC_ALIGNMENT_ERROR - Posted source or destination data pointers or
 *                          data length are not properly aligned.
 * GNI_RC_ERROR_RESOURCE - The transaction request could not be posted due
 *                         to insufficient resources.
 * GNI_RC_ERROR_NOMEM - Insufficient memory to complete the operation.
 * GNI_RC_PERMISSION_ERROR - The user's buffer R/W permissions conflict with
 *                           the access type.
 *
 * Description:
 * This function adds a descriptor to the tail of the RDMA queue and 
 * returns immediately.
 * 
 **/
gni_return_t
GNI_PostRdma(
    gni_ep_handle_t ep_hndl,
    gni_post_descriptor_t *post_descr
);

gni_return_t
GNI_PostRdmaGet(
    gni_ep_handle_t ep_hndl,
    gni_mem_handle_t local_mem_hndl,
    gni_mem_handle_t remote_mem_hndl
);

gni_return_t
GNI_PostRdmaGetBuffer(
    gni_ep_handle_t ep_hndl,
    void* buffer,
    uint64_t length,
    gni_mem_handle_t remote_mem_hndl
);

/**
 * GNI_PostFma - Post FMA transaction
 * 
 * Parameters:
 * IN
 * ep_hndl      Instance of a local Endpoint.
 * post_descr   Pointer to a descriptor to be posted.
 *
 * Returns:
 * GNI_RC_SUCCESS - The descriptor was successfully posted.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid.
 * GNI_RC_ALIGNMENT_ERROR - Posted source or destination data pointers or
 *                          data length are not properly aligned.
 * GNI_RC_ERROR_RESOURCE - The transaction request could not be posted due
 *                         to insufficient resources.
 *
 * Description:
 * This function executes a data transaction (Put, Get or AMO) by 
 * storing into the directly mapped FMA Window to initiate a series of 
 * FMA requests. 
 * It returns before the transaction is confirmed by the remote NIC.
 * Zero-length FMA Put operations are supported. Zero-length FMA Get and
 * zero-length FMA AMO operations are not supported.
 * 
 **/

gni_return_t
GNI_PostFma(
    gni_ep_handle_t ep_hndl,
    gni_post_descriptor_t *post_descr
);

gni_return_t
GNI_PostFmaGet(
    gni_ep_handle_t ep_hndl,
    gni_mem_handle_t local_mem_hndl,
    gni_mem_handle_t remote_mem_hndl
);

gni_return_t
GNI_PostFmaGetBuffer(
    gni_ep_handle_t ep_hndl,
    void* buffer,
    uint64_t length,
    gni_mem_handle_t remote_mem_hndl
);


/**
 * GNI_GetCompleted - Get next completed descriptor
 * 
 * Parameters:
 * IN
 * cq_hndl      The handle for the Completion Queue.
 * event_data   The event returned by CqGetEvent function.
 * 
 * OUT
 * post_desc    Address of the descriptor that has completed.
 *
 * Returns:
 * GNI_RC_SUCCESS - A completed descriptor was returned with a successful
 *                  completion status.
 * GNI_RC_DESCRIPTOR_ERROR - If the corresponding post queue (FMA, RDMA or AMO)
 *                           is empty, the descriptor pointer is set to NULL, 
 *                           otherwise, a completed descriptor is returned with 
 *                           an error completion status.
 * GNI_RC_INVALID_PARAM - The CQ handle was invalid.
 * GNI_RC_TRANSACTION_ERROR - A completed descriptor was returned with a
 *                            network error status.
 *
 * Description:
 * This function gets the descriptor from the corresponding post queue. 
 * The descriptor is removed from the head of the queue and the address 
 * of the descriptor is returned.
 * 
 **/
gni_return_t
GNI_GetCompleted(
    gni_cq_handle_t cq_hndl,
    gni_cq_entry_t event_data,
    gni_post_descriptor_t **post_descr
);

/**
 * GNI_CqGetEvent - Get next event
 * 
 * Parameters:
 * IN
 * cq_hndl      The handle for the Completion Queue.
 * 
 * OUT
 * event_data   A new event entry data, if the return status indicates success. 
 *              Undefined otherwise.
 *
 * Returns:
 * GNI_RC_SUCCESS - A completion entry was found on the Completion Queue.
 * GNI_RC_NOT_DONE - No new completion entries are on the Completion Queue.
 * GNI_RC_INVALID_PARAM - The Completion Queue handle was invalid.
 * GNI_RC_ERROR_RESOURCE - CQ is an overrun state and CQ events may
 *                         have been lost.
 * GNI_RC_TRANSACTION_ERROR - A network error was encountered processing a transaction.
 *
 * Description:
 * This function polls the specified Completion Queue for a completion entry. 
 * If a completion entry is found, it returns the event data stored the entry.
 * CqGetEvent is a non-blocking call. It is up to the calling process to 
 * subsequently invoke the appropriate function to de-queue the completed descriptor. 
 * CqGetEvent only de-queues the completion entry from the Completion Queue.
 * 
 **/
gni_return_t
GNI_CqGetEvent(
    gni_cq_handle_t cq_hndl,
    gni_cq_entry_t *event_data
);

/**
 * GNI_CqWaitEvent - Wait for the next event
 * 
 * Parameters:
 * IN
 * cq_hndl      The handle for the Completion Queue.
 * time     The number of milliseconds to block before returning
 *              to the caller, (-1) if no time-is desired.
 * 
 * OUT
 * event_data   A new event entry data, if the return status indicates success. 
 *              Undefined otherwise.
 *
 * Returns:
 * GNI_RC_SUCCESS - A completion entry was found on the Completion Queue.
 * GNI_RC_TIME- The request timed and no completion entry was found.
 * GNI_RC_INVALID_PARAM - The Completion Queue handle was invalid.
 * GNI_RC_ERROR_RESOURCE - The Completion Queue was not created in
 *                         the GNI_CQ_BLOCKING mode.
 * GNI_RC_TRANSACTION_ERROR - A network error was encountered processing a transaction.
 *
 * Description:
 * This function polls the specified Completion Queue for a completion entry. 
 * If a completion entry was found, it immediately returns event data.
 * If no completion entry is found, the caller is blocked until a completion 
 * entry is generated, or until the timevalue expires.
 * The Completion Queue must be created with the GNI_CQ_BLOCKING mode set
 * order to be able to block on it.
 * 
 **/
gni_return_t
GNI_CqWaitEvent(
    gni_cq_handle_t cq_hndl,
    uint64_t timeout,
    gni_cq_entry_t *event_data
);

/**
 * GNI_CqTestEvent - Check if there is an event on a Completion Queue
 * 
 * Parameters:
 * IN
 * cq_hndl      The handle for the Completion Queue.
 *
 * 
 * Returns:     
 * GNI_RC_SUCCESS - A completion entry was found on the Completion Queue.
 * GNI_RC_NOT_DONE - No new completion entries are on the Completion Queue.
 * GNI_RC_INVALID_PARAM - The Completion Queue handle was invalid.
 * GNI_RC_ERROR_RESOURCE - CQ is an overrun state and CQ events may have been lost.
 * 
 * Description:
 * This function polls the specified Completion Queue for a completion entry.
 * If a completion entry is found, it return GNI_RC_SUCCESS, unless the
 * CQ is overrun, which case GNI_RC_ERROR_RESOURCE.  If no completion entry
 * is found GNI_RC_NOT_DONE is returned.
 *
 * No processing of new entries is performed by this function.
 *
 **/
gni_return_t
GNI_CqTestEvent(gni_cq_handle_t cq_hndl);


/**
 * GNI_SmsgInit - Initialize short messaging resources
 * IN
 * ep_hndl              The handle of the Endpoint.
 * local_smsg_attr      Local parameters for short messaging
 * remote_smsg_attr     Remote parameters for short messaging provided by peer
 *  
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - One of the input parameters was invalid.
 * GNI_RC_INVALID_STATE - Endpoind is not bound
 * GNI_RC_ERROR_NOMEM - Insufficient memory to allocate short message
 *                      internal structures
 * Description:
 * This function configures the short messaging protocol on the given Endpoint.
 **/ 
gni_return_t
GNI_SmsgInit(
    gni_ep_handle_t ep_hndl,
    gni_smsg_attr_t *local_smsg_attr,
    gni_smsg_attr_t *remote_smsg_attr
);

/**
 * GNI_SmsgSend - Send short message
 * 
 * Parameters:
 * IN
 * ep_hndl      Instance of an Endpoint.
 * header       Pointer to the header of a message.
 * header_length Length of the header bytes.
 * data         Pointer to the payload of the message.
 * data_length  Length of the payload byte.
 * msg_id       Identifier for application to track transaction.
 *              Only valid for short messaging using MBOX_PERSISTENT type,
 *              otherwise ignored.
 *
 * Returns:
 * GNI_RC_SUCCESS - The message was successfully sent.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid or
 *                        the Endpoint is not initialized for short messaging.
 * GNI_RC_NOT_DONE - No credits available to send the message
 * GNI_RC_ERROR_RESOURCE - The total size of the header plus data exceeds
 *                         the maximum short message size defined by GNI_SMSG_MAX_SIZE.
 *
 * Description:
 * This function sends a message to the remote peer, by copying it into the
 * pre-allocated remote buffer space using the FMA mechanism.  It returns
 * before the delivery is confirmed by the remote NIC.  With MBOX_PERSISTENT
 * type system attempts to re-transmit for certatransaction failures.  This
 * is a non-blocking call.  Completion events are delivered to local and remote
 * completion queues for each send.
 *
 * Note:
 * The SMSG interface uses the FMA mechanism with adaptive routing.  This
 * allows SMSG sends to arrive of order at the target node.  Due to this,
 * it is possible for completion events to be delivered to the remote
 * completion queue while GNI_SmsgGetNext reports that no new messages are
 * available.  To handle this case when using remote events to detect the
 * arrival of SMSG sends, be sure to clear all messages from an endpoint using
 * GNI_SmsgGetNext after receiving each remote completion event.
 *
 **/
gni_return_t
GNI_SmsgSend(
    gni_ep_handle_t ep_hndl,
    void *header,
    uint32_t header_length,
    void *data,
    uint32_t data_length,
    uint32_t msg_id
);

gni_return_t
GNI_SmsgSendHeader(
    gni_ep_handle_t ep_hndl,
    void *header,
    uint32_t header_length
);

/**     
 * GNI_SmsgSendWTag - Send short message with a tag
 *      
 * Parameters:
 *
 * ep_hndl      Instance of an Endpoint.
 * header       Pointer to the header of a message.  
 * header_length Length of the header bytes.
 * data         Pointer to the payload of the message.
 * data_length  Length of the payload byte.
 * msg_id       Identifier for application to track transaction.
 *              Only valid for short messaging using MBOX_PERSISTENT type
 * tag          Tag associated with the short message.
 *
 * Returns:
 * GNI_RC_SUCCESS - The message was successfully sent.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid or
 *                        the Endpoint is not initialized for short messaging.
 * GNI_RC_NOT_DONE - No credits available to send the message
 * GNI_RC_ERROR_RESOURCE - The total size of the header plus data exceeds
 *                         the maximum short message size defined by GNI_SMSG_MAX_SIZE.
 *      
 * Description:
 * This function sends a tagged message to the remote peer, by copying it into
 * the pre-allocated remote buffer space using the FMA mechanism.
 * It returns before the delivery is confirmed by the remote NIC.
 * With MBOX_PERSISTENT type system attempts to re-transmit
 * for certatransaction failures.
 * This is a non-blocking call.
 *
 **/

gni_return_t
GNI_SmsgSendWTag(
    gni_ep_handle_t ep_hndl,
    void *header,
    uint32_t header_length,
    void *data,
    uint32_t data_length,
    uint32_t msg_id,
    uint8_t tag
);

/**
 * GNI_SmsgGetNext - Get next available short message
 * 
 * Parameters:
 * IN
 * ep_hndl      Instance of an Endpoint.
 * 
 * OUT
 * header       Pointer to the header of the newly arrived message.
 * 
 * Returns:
 * GNI_RC_SUCCESS - The new message is successfully arrived.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid or the Endpoint
 *                        is not initialized for short messaging
 * GNI_RC_NOT_DONE - No new messages available.
 * GNI_RC_INVALID_STATE - The SMSG connection has entered an invalid state.
 *
 * Description:
 * This function returns a pointer to the header of the newly arrived message and 
 * makes this message current. An application may decide to copy the message
 * of the mailbox or process it immediately. This is a non-blocking call.
 * 
 **/
gni_return_t
GNI_SmsgGetNext(
    gni_ep_handle_t ep_hndl,
    void **header
);

/**
 * GNI_SmsgGetNextWTag  -   Get next available short message if input tag
 *                          matches that of the short message.
 *
 * Parameters:
 * IN
 * ep_hndl       Instance of an Endpoint.
 *
 * OUT
 * header   Pointer to the header of the newly arrived message.
 *          event value.
 * tag      On input, pointer to value of remote event to be matched.
 *          A wildcard value of GNI_SMSG_ANY_TAG can be used to match any
 *          tag value of the incoming message.
 *          The value is set to that of the matching remote event
 *          on output.
 *
 * Returns:
 * GNI_RC_SUCCESS - The new message is successfully arrived.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid or the Endpoint is
 *            not GNI_EP_STATE_CONNECTED state.
 * GNI_RC_NOT_DONE - No new messages available.
 * GNI_RC_NO_MATCH - Message available, but tag of message doesn't match
 *                   the value supplied the tag argument.
 * GNI_RC_INVALID_STATE - The SMSG connection has entered an invalid state.
 *
 * Description:
 * This function returns a pointer to the header of the newly arrived message and
 * makes this message current if the input tag matches the tag of the newly
 * arrived message. An application may decide to copy the message header out
 * of the mailbox or process the header immediately. This is a non-blocking call.
 *
 **/
gni_return_t
GNI_SmsgGetNextWTag(
    gni_ep_handle_t ep_hndl,
    void **header,
    uint8_t *tag
);


/**
 * GNI_SmsgRelease - Release current message
 * 
 * Parameters:
 * IN
 * ep_hndl      Instance of an Endpoint.
 * 
 * Returns:
 * GNI_RC_SUCCESS - The current message is successfully released.
 * GNI_RC_INVALID_PARAM - The Endpoint handle was invalid or the Endpoint
 *                        is not initialized for short messaging
 * GNI_RC_NOT_DONE - There is no current message.
 *                   The GetNext function must return GNI_RC_SUCCESS before
 *                   calling this function.
 *
 * Description:
 * This function releases the current message buffer. It must be called only 
 * after GetNext has returned GNI_RC_SUCCESS. This is a non-blocking call.
 * The message returned by the GetNext function must be copied or processed
 * prior to making this call.
 * 
 **/
gni_return_t
GNI_SmsgRelease(gni_ep_handle_t ep_hndl);

/**
 * GNI_SmsgsSetMaxRetrans - Configures SMSG max retransmit count.
 *
 * IN
 * gni_handle           The NIC handle to alter.
 * max_retrans          The new SMSG max retransmit count.
 *
 * Returns:
 * GNI_RC_SUCCESS - Operation completed successfully.
 * GNI_RC_INVALID_PARAM - Invalid NIC handle specified.
 *
 * Description:
 * This functions sets the maximum retransmit counts for SMSG transactions.
 * EPs associated with the NIC handle provided will give up retransmitting SMSG
 * transactions and return GNI_RC_TRANSACION_ERROR when the retransmit count has
 * been reached.
 **/
gni_return_t
GNI_SmsgSetMaxRetrans(
    gni_nic_handle_t gni_handle,
    uint16_t max_retrans
);

#ifdef __cplusplus
}
#endif

#endif /*_GNI_PUB_H_*/

