#ifndef SSTMACRO_SOFTWARE_LIBARIES_NICAPI_H
#define SSTMACRO_SOFTWARE_LIBARIES_NICAPI_H

#include <sstmac/software/api/gni/gni.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/common/errors.h>
#include <boost/unordered_map.hpp>

namespace sstmac {
    namespace sw {

class gni_invalid_param :
    public ssterror
{
    public:
        gni_invalid_param(const std::string& msg)
            : ssterror(msg)
        {
        }
};

#define printfstr_uint64_t "%lu"
#define printfstr(x) printfstr##_x


class NoStruct {};

template <class IntType, class T = NoStruct>
class api_handle_set {
    typedef enum { Null=0, Created=1, Initialized=2, Freed=3 } state_t;

    struct api_set_entry
    {
        T item;
        bool allocated;
    };

    const char* name_;

    std::vector<api_set_entry> entries_;

    std::list<IntType> free_ids_;

    IntType next_id;

    void bad_handle(IntType handle)
    {
        sst_throw(gni_invalid_param,
            "no handle %lu exists for %s",
            handle, name_);
    }

    api_set_entry& get(IntType handle){
        if (handle >= entries_.size())
            bad_handle(handle);

        api_set_entry& entry = entries_[handle];
        if (!entry.allocated)
            bad_handle(handle);

        return entry;
    }

    public:

    api_handle_set(const char* name)
        : next_id(0),
        name_(name)
    {
    }

    T& get_item(IntType handle){
        return get(handle).item;
    }

    /**
        Create the handle, but don't do initializiation work.
    */
    T& create(IntType* handle)
    {
        if (!free_ids_.empty()){
            *handle = free_ids_.front();
            free_ids_.pop_front();
            api_set_entry& entry = entries_[*handle];
            return entry.item;
        }
        else {
            *handle = entries_.size();
            entries_.push_back(api_set_entry());
            return entries_[*handle].item;
        }
    }

    void validate(IntType handle)
    {
        get(handle);
        //get performs validation
        api_set_entry& entry = get(handle);
    }

    T& free(IntType handle)
    {
        api_set_entry& entry = get(handle);
        entry.allocated = false;
        free_ids_.push_back(handle);
        return entry.item;
    }

};

struct endpoint
{
    nodeaddress::ptr addr;
    gni_smsg_attr_t* local_smsg_attr;
    gni_smsg_attr_t* remote_smsg_attr;
    bool bound;

    endpoint(){
      sst_throw(unimplementederror, "gni::endpoint::endpoint()");
    }
};

struct communication_domain
{
    bool attached;

    communication_domain(){
      sst_throw(unimplementederror, "gni::communication_domain::communication_domain()");
    }
};

struct NIC
{
};

class completion_queue
{
    public:
        uint32_t entry_count;
};

struct mem_registration
{
    uint64_t address;
    uint64_t length;
    gni_cq_handle_t cq;
    gni_nic_handle_t nic;

    bool contains(void* ptr);

    mem_registration(){
      sst_throw(unimplementederror, "gni::mem_registration::mem_registration()");
    }
};

class gniapi :
    public library
{

    private:
        api_handle_set<gni_cdm_handle_t,communication_domain> cdm_handles_;
        api_handle_set<gni_nic_handle_t,NIC> nic_handles_;
        api_handle_set<gni_ep_handle_t,endpoint> ep_handles_;
        api_handle_set<gni_cq_handle_t,completion_queue> cq_handles_;
        api_handle_set<gni_mem_handle_t,mem_registration> mem_handles_;

        int nproc_;

        int me_;

    public:
        typedef boost::intrusive_ptr<gniapi> ptr;

        uint64_t gni_cq_get_data(gni_cq_entry_t);
        uint64_t gni_cq_get_source(gni_cq_entry_t);
        uint64_t gni_cq_get_status(gni_cq_entry_t);
        uint64_t gni_cq_get_info(gni_cq_entry_t);
        uint64_t gni_cq_overrun(gni_cq_entry_t);
        uint64_t gni_cq_get_inst_id(gni_cq_entry_t);
        uint64_t gni_cq_get_tid(gni_cq_entry_t);
        uint64_t gni_cq_get_msg_id(gni_cq_entry_t);
        uint64_t gni_cq_get_type(gni_cq_entry_t);
        uint64_t gni_cq_get_block_id(gni_cq_entry_t);
        uint64_t gni_cq_get_unsuccessful_cnt(gni_cq_entry_t);
        uint64_t gni_cq_get_marker_id(gni_cq_entry_t);
        void gni_cq_set_inst_id(gni_cq_entry_t *,uint64_t);
        void gni_cq_set_tid(gni_cq_entry_t *,uint64_t);
        void gni_cq_set_msg_id(gni_cq_entry_t *,uint64_t);
        void gni_cq_set_type(gni_cq_entry_t *,uint64_t);
        void gni_cq_clr_status(gni_cq_entry_t *);

    
        void
        gni_cdm_create(
            uint32_t inst_id,
            uint8_t ptag,
            uint32_t cookie,
            uint32_t modes,
            gni_cdm_handle_t *cdm_hndl
        );

        void
        gni_cdm_destroy(gni_cdm_handle_t cdm_hndl);

        void
        gni_cdm_get_nic_address(
            uint32_t device_id,
            uint32_t *address,
            uint32_t *cpu_id
        );

        void
        gni_cdm_attach(
            gni_cdm_handle_t cdm_hndl,
            uint32_t device_id,
            uint32_t *local_addr,
            gni_nic_handle_t *nic_hndl
        );

        void
        gni_ep_create(
            gni_nic_handle_t nic_hndl,
            gni_cq_handle_t src_cq_hndl,
            gni_ep_handle_t *ep_hndl
        );

        void
        gni_ep_bind(
            gni_ep_handle_t ep_hndl,
            uint32_t remote_addr,
            uint32_t remote_id
        );

        void
        gni_ep_unbind(gni_ep_handle_t ep_hndl);

        void
        gni_ep_destroy(gni_ep_handle_t ep_hndl);

        void
        gni_mem_register(
            gni_nic_handle_t nic_hndl,
            uint64_t address,
            uint64_t length,
            gni_cq_handle_t dst_cq_hndl,
            uint32_t flags,
            uint32_t vmdh_index,
            gni_mem_handle_t *mem_hndl
        );

        void
        gni_mem_deregister(
            gni_nic_handle_t nic_hndl,
            gni_mem_handle_t *mem_hndl
        );

        void
        gni_cq_create(
            gni_nic_handle_t nic_hndl,
            uint32_t entry_count,
            uint32_t delay_count,
            gni_cq_mode_t mode,
            void (*handler)(gni_cq_entry_t*,void*),
            void *context,
            gni_cq_handle_t *cq_hndl
        );

        void
        gni_cq_destroy(gni_cq_handle_t cq_hndl);

        void
        gni_post_rdma(
            gni_ep_handle_t ep_hndl,
            gni_post_descriptor_t *post_descr
        );

        void
        gni_post_rdma_get(
            gni_ep_handle_t ep_hndl,
            gni_mem_handle_t remote_mem_hndl
        );

        void
        gni_post_rdma_get(
            gni_ep_handle_t ep_hndl,
            void* buffer,
            uint64_t length
        );

        void
        gni_post_rdma_put(
            gni_ep_handle_t ep_hndl,
            gni_mem_handle_t remote_mem_hndl
        );

        void
        gni_post_rdma_put(
            gni_ep_handle_t ep_hndl,
            void* buffer,
            uint64_t length
        );

        void
        gni_post_fma(
            gni_ep_handle_t ep_hndl,
            gni_post_descriptor_t *post_descr
        );

        void
        gni_post_fma_get(
            gni_ep_handle_t ep_hndl,
            gni_mem_handle_t remote_mem_hndl
        );

        void
        gni_post_fma_get(
            gni_ep_handle_t ep_hndl,
            void* buffer,
            uint64_t length
        );

        void
        gni_post_fma_put(
            gni_ep_handle_t ep_hndl,
            gni_mem_handle_t remote_mem_hndl
        );

        void
        gni_post_fma_put(
            gni_ep_handle_t ep_hndl,
            void* buffer,
            uint64_t length
        );

        void
        gni_get_completed(
            gni_cq_handle_t cq_hndl,
            gni_cq_entry_t event_data,
            gni_post_descriptor_t **post_descr
        );

        void
        gni_cq_get_event(
            gni_cq_handle_t cq_hndl,
            gni_cq_entry_t *event_data
        );

        void
        gni_cq_wait_event(
            gni_cq_handle_t cq_hndl,
            uint64_t timeout,
            gni_cq_entry_t *event_data
        );

        void
        gni_cq_test_event(gni_cq_handle_t cq_hndl);

        void
        gni_smsg_init(
            gni_ep_handle_t ep_hndl,
            gni_smsg_attr_t *local_smsg_attr,
            gni_smsg_attr_t *remote_smsg_attr
        );

        void
        gni_smsg_send(
            gni_ep_handle_t ep_hndl,
            void *header,
            uint32_t header_length,
            void *data,
            uint32_t data_length,
            uint32_t msg_id
        );

        void
        gni_smsg_send(
            gni_ep_handle_t ep_hndl,
            void *header,
            uint32_t header_length
        );

        void
        gni_smsg_send_wtag(
            gni_ep_handle_t ep_hndl,
            void *header,
            uint32_t header_length,
            void *data,
            uint32_t data_length,
            uint32_t msg_id,
            uint8_t tag
        );

        void
        gni_smsg_get_next(
            gni_ep_handle_t ep_hndl,
            void **header
        );

        void
        gni_smsg_get_next_wtag(
            gni_ep_handle_t ep_hndl,
            void **header,
            uint8_t *tag
        );

        void
        gni_smsg_release(gni_ep_handle_t ep_hndl);

        void
        gni_smsg_set_max_retrans(
            gni_nic_handle_t nic_handle,
            uint16_t max_retrans
        );
};

    }
}




#endif /*_gni_PUB_H_*/

