/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommdata.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>

namespace sstmac
{
  namespace sw
  {

    debug mpicomm::dbg_;

    //
    // NULL communicator.
    //
    mpicomm::ptr mpicomm::comm_null;

    mpicomm::mpicomm() :
      group_(), next_collective_tag_(10235), id_(-2), rank_(-1) //, appid_(-1)
    {
      topotype_ = TOPO_NONE;



      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> mpi | mpicomm", logger::yellow);
          logger::register_logger(&dbg_);
        }
    }

    mpicomm::mpicomm(
        const mpicommid &id, //const appid &aid,
        const mpiid &rank, const mpigroup::ptr peers,
        const environment::const_ptr &env, const appid &aid) :
      env_(env), group_(peers), next_collective_tag_(10235),
          aid_(aid), id_(id), rank_(rank)
    {

        if (peers->size() == 0){
            sst_throw(valueerror, "trying to build communicator of size 0");
        }

      topotype_ = TOPO_NONE;

      if (!comm_null)
        {
          comm_null = mpicomm::construct();
        }

      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> mpi | mpicomm", logger::yellow);
          logger::register_logger(&dbg_);
        }
    }

    void
    mpicomm::dup_keyvals(mpicomm::ptr m)
    {
      boost::unordered_map<int, keyval::ptr>::iterator it, end = m->keyvals_.end();
      for (it = m->keyvals_.begin(); it != end; it++)
        {
          keyval::ptr c = (it->second)->clone(keyval::get_new_key());
          keyvals_[c->key()] = c;
        }
    }

    std::string
    mpicomm::to_string() const
    {
      return "mpicomm(" + rank().to_string() + ")";
    }

    /// The rank of this peer in the communicator.
    const mpiid&
    mpicomm::rank() const
    {
      return rank_;
    }

    /// The size of the communicator.
    mpiid
    mpicomm::size() const
    {
      if (id_.id == -2)
        {
          throw nullerror("mpicomm: trying to call size() on a null mpicomm");
        }
      if (!group_)
        {
          throw nullerror("mpicomm: peers_ is null for some reason in size()");
        }
      return mpiid(group_->size());
    }

    /// The identifier for this communicator.
    /// To be used to tag messages to/from this communicator.
    const mpicommid&
    mpicomm::id() const
    {
      return id_;
    }

    void
    mpicomm::set_keyval(keyval::ptr k, void* val)
    {
      SSTMAC_DEBUG << "mpicomm(id=" << id_.id << ", rnk=" << rank_.id
            << "): set_keyval(" << k->key() << ")\n";
      keyvals_[k->key()] = k;
      k->set_val(val);
    }

    void
    mpicomm::get_keyval(keyval::ptr k, void* val, int* flag)
    {
      SSTMAC_DEBUG << "mpicomm(id=" << id_.id << ", rnk=" << rank_.id
            << "): get_keyval(" << k->key() << ")\n";

      if (keyvals_.find(k->key()) == keyvals_.end())
        {

          SSTMAC_DEBUG << "key not found, existing are: ";

          boost::unordered_map<int, keyval::ptr>::iterator it, end = keyvals_.end();
          for(it = keyvals_.begin(); it != end; it++){
              SSTMAC_DEBUG << it->first << " ";
          }

          SSTMAC_DEBUG << "\n";

          *flag = 1;
        }else{
            *flag = 0;
        }

      void** vcast = (void**)val;

      *vcast = k->val();

      //memcpy(val, (k->val()), sizeof(void*));

      //val = (k->val());
    }

    const appid&
    mpicomm::get_app() const
    {
      return aid_;
    }

    //
    // Get a unique tag for a collective operation.
    //
    mpitag
    mpicomm::next_collective_tag() const
    {
      next_collective_tag_.id++;
      return next_collective_tag_;
    }

    /// The task index of the caller.
    taskid
    mpicomm::my_task() const
    {
      if (id_.id < 0)
        throw illformederror("mpicomm::my_task:  NULL communicator.");
      if (rank_.id < 0 || rank_.id >= env_->get_app_count())
        throw illformederror(
            std::string("mpicomm::my_task:  Invalid rank ")
                + boost::lexical_cast<std::string>(rank_));

      if (!group_)
        {
          throw nullerror(
              "mpicomm: group_ is null for some reason in peer_task()");
        }
      return group_->at(rank_.id);
    }

    /// The task index of the given peer.
    taskid
    mpicomm::peer_task(const mpiid &rank) const
    {
      if (id_.id < 0)
        {
          throw nullerror(
              "mpicomm: trying to call peer_task() on a null mpicomm");
        }
      if (!group_)
        {
          throw nullerror(
              "mpicomm: group_ is null for some reason in peer_task()");
        }
      return group_->at(rank.id);
    }

    nodeaddress::ptr
    mpicomm::my_node() const
    {
      if (id_.id < 0)
        {
          throw nullerror("mpicomm: trying to call my_node() on a null mpicomm");
        }
      if (!env_)
        {
          throw nullerror("mpicomm: evn_ is null for some reason in my_node()");
        }

      taskid tid = my_task();
      //softwareid sid(appid_, tid);
      return env_->get_node_address(tid);
    }

    /// The list of nodes involved in this communicator.
    /// Indexing is done by mpiid::rank().id.
    nodeaddress::ptr
    mpicomm::node_at(const mpiid &rank) const
    {
      if (id_.id < 0)
        {
          throw nullerror("mpicomm: trying to call node_at() on a null mpicomm");
        }
      if (!env_)
        {
          throw nullerror("mpicomm: evn_ is null for some reason in node_at()");
        }

      taskid tid = peer_task(rank);
      //softwareid sid(appid_, tid);
      return env_->get_node_address(tid);
    }

    //
    // Default constructor.
    //
    //   mpicomm::mpicomm() :
    //     rank_(-1), data_()
    //   {}

    //
    // Hello.
    //
    //   mpicomm::mpicomm(const mpiid &rank,
    //                    boost::shared_ptr<const mpicommdata> data) :
    //     rank_(rank), data_(data)
    //   {}

    //
    // Goodbye.
    //
    //   mpicomm::~mpicomm() {
    //   }

    //
    // The rank of this process in the communicator.
    //
    //const mpiid& mpicomm::rank() const {
    //  return rank_;
    //}

    //
    // The size of the communicator.
    //
    //mpiid mpicomm::size() const {
    //  if(data_==0) throw nullerror("mpicomm::size: data pointer is null.");
    //  return mpiid(data_->peers().size());
    //}

    //
    // The communicator id.
    //
    //const mpicommid& mpicomm::id() const {
    //  if(data_==0) throw nullerror("mpicomm::id: data pointer is null.");
    //  return data_->id();
    //}

    //
    // Get this task id.
    //
    //taskid mpicomm::my_task() const {
    //  return this->peers().at(rank_.id);
    //}

    //
    // Get a specific peer.
    //
    //taskid mpicomm::peer_task(const mpiid &id) const {
    //  if(debugger::active()) debug() << "mpicomm::peer_task(" << id << ")\n";
    //  if((id.id < 0) || (id >= this->size()))
    //    throw valueerror("mpicomm::peer:  Index out of range.");
    //  return this->peers().at(id.id);
    //}

    //
    // My peers.
    //
    //const std::vector<taskid>& mpicomm::peers() const {
    //  if(data_==0) throw nullerror("mpicomm::peers: data pointer is null.");
    //  return data_->peers();
    //}

    //
    // Fairly self-explanatory.
    //
    std::ostream&
    operator<<(std::ostream &os, const mpicomm::ptr &comm)
    {
      os << "mpicomm(rank=" << comm->rank_ << ", id=" << comm->id_ << ")";
      return os;
    }

  }
} // end of namespace sstmac
