/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMM_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommid.h>
#include <sstmac/common/errors.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/software/process/tasklist.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_comm/keyval.h>
#include <boost/lexical_cast.hpp>
#include <sstmac/software/api/mpi/mpi_comm/mpigroup.h>
#include <iosfwd>
#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {

    // Forward declaration of important types.
    //class mpicommdata;

    /**
     * An MPI communicator handle.
     */
    class mpicomm : public ptr_type
    {
    public:
      enum topotypes
      {
        TOPO_NONE, TOPO_GRAPH, TOPO_CART
      };

      static const int proc_null = -1;

    protected:

      friend std::ostream&
      operator<<(std::ostream &os, const mpicomm::ptr &comm);

      friend class mpicommfactory;
      /// The data (possibly shared) associated with this communicator.
      //boost::shared_ptr<const mpicommdata> data_;

      // the definitely-shared environment (softwareid <--> nodeaddress)
      environment::const_ptr env_;

      static debug dbg_;

      /// The tasks participating in this communicator.  This is only used for an mpicomm::ptr which is NOT WORLD_COMM.
      //  tasklist::const_ptr peers_;
      mpigroup::ptr group_;

      mutable mpitag next_collective_tag_;

      boost::unordered_map<int, keyval::ptr> keyvals_;

      appid aid_;

      topotypes topotype_;

      std::string name_;

      mpicomm();

      /// Hello.
      mpicomm(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const mpigroup::ptr peers,
          const environment::const_ptr &env, const appid &aid);

    public:

      typedef boost::intrusive_ptr<mpicomm> ptr;

      static ptr
      construct(
          const mpicommid &id, //const appid &aid,
          const mpiid &rank, const mpigroup::ptr peers,
          const environment::const_ptr &env, const appid &aid)
      {
        return ptr(new mpicomm(id, rank, peers, env, aid));
      }

      static ptr
      construct()
      {
        return ptr(new mpicomm());
      }

      /// Goodbye.
      virtual
      ~mpicomm()
      {
      }

      void
      set_name(std::string name){
        name_ = name;
      }

      std::string
      get_name() const {
        return name_;
      }

      topotypes
      get_topo_type() const
      {
        return topotype_;
      }

      mpigroup::ptr
      get_group()
      {
        return group_;
      }

      void
      dup_keyvals(mpicomm::ptr m);

      /// This is the null communicator.
      static mpicomm::ptr comm_null;

      mpicommid id_;
      mpiid rank_;

      virtual std::string
      to_string() const;

      /// The rank of this peer in the communicator.
      const mpiid&
      rank() const;

      /// The size of the communicator.
      mpiid
      size() const;

      /// The identifier for this communicator.
      /// To be used to tag messages to/from this communicator.
      const mpicommid&
      id() const;

      void
      set_keyval(keyval::ptr k, void* val);

      void
      get_keyval(keyval::ptr k, void* val, int* flag);

      const appid&
      get_app() const;

      //
      // Get a unique tag for a collective operation.
      //
      mpitag
      next_collective_tag() const;

      /// The task index of the caller.
      taskid
      my_task() const;

      /// The task index of the given peer.
      taskid
      peer_task(const mpiid &rank) const;

      nodeaddress::ptr
      my_node() const;

      /// The list of nodes involved in this communicator.
      /// Indexing is done by mpiid::rank().id.
      nodeaddress::ptr
      node_at(const mpiid &rank) const;

      /// Equality comparison.
      inline bool
      operator==(const mpicomm::ptr &other) const
      {
        return ((rank_ == other->rank_) && (id_ == other->id_));
      }

      /// Inequality comparison.
      inline bool
      operator!=(const mpicomm::ptr &other) const
      {
        return !this->operator==(other);
      }

    };

    /// Allow communicators in sorted containers.
    inline bool
    operator<(const mpicomm::ptr &a, const mpicomm::ptr &b)
    {
      return (a->id() == b->id() ? a->rank() < b->rank() : a->id() < b->id());
    }

    /// Fairly self-explanatory.
    std::ostream&
    operator<<(std::ostream &os, const mpicomm::ptr &comm);

  }
} //end of namespace sstmac

#endif
