/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMMFACTORY_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_COMM_MPICOMMFACTORY_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommid.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>


namespace sstmac { namespace sw {

  // Important types.
  class mpiapi;
  /**
   * Construct mpi communicators.
   */
  class mpicommfactory : virtual public nocopy, public ptr_type {
    /// The private type used to manage communicators under construction.
    class pending;
    class pending_create;
    class pending_split;

    boost::intrusive_ptr<mpiapi> parent_;

    appid aid_;

    static debug dbg_;

    /// The scheduler used to block and unblock tasks.
    ////boost::shared_ptr<scheduler> scheduler_;

    /// We can restrict our run to use fewer than the nodes allocated.
    int mpirun_np_;

    /// The shared data for the world communicator.
    ////mpicommdata::ptr world_data_;

    /// The next available communicator index.
    mpicommid next_id_;

    /// Keyring(s) for pending comm_dup or comm_create requests.
    /// The signature for each pending request is determined by the
    /// input communicator (construction of communicators is global (blocking)
    /// across the communicator).
    typedef std::map<mpicommid, boost::intrusive_ptr<pending> > buildmap_t;
    buildmap_t under_construction_;

    /// The world communicators for all nodes.
   // std::map<mpiid, mpicomm> worldcomms_;

    /// The self communicators for all nodes.
   // std::map<mpiid, mpicomm> selfcomms_;

    mpicomm::ptr worldcomm_;
    mpicomm::ptr selfcomm_;
    mpitype splittype_;

  protected:
    /// Build comm_world using information retrieved from the environment.
    mpicommfactory(appid aid, boost::intrusive_ptr<mpiapi> parent);

    static std::map<std::pair<appid,mpicommid>, tasklist::ptr> tlists_;

  public:
    typedef boost::intrusive_ptr<mpicommfactory> ptr;


    virtual std::string to_string() const{
            return "mpicommfactory";
          }

    /// Build a new factory.
    static mpicommfactory::ptr construct(appid aid, boost::intrusive_ptr<mpiapi> parent);

    /// Goodbye.
    virtual ~mpicommfactory();

    /// Initialize the object.
    void init( ////const boost::shared_ptr<scheduler>& scheduler,
               environment::const_ptr& env, const mpiid &rank);

    void finalize();

  public:
    /// Get the world communicator for the given node.
    /// In this communicator, mpiid indices are the same as the taskid
    /// indices given by the environment.  Each node will only have one world.
    const mpicomm::ptr get_world();

    /// Get a 'self' communicator.  Each node will have a unique index
    /// for its self.  Each node will only have one self.
    const mpicomm::ptr get_self();


    /// Duplicate the given communicator.
    /// Blocks unti lall nodes have entered the call.
    /// Eventually we may opt to deal with the extended attributes (keyval)
    /// stuff in the MPI standard, even though I don't know anybody
    /// who actually uses it.
    mpicomm::ptr comm_dup(const mpicomm::ptr &caller);

    /// Make the given mpiid refer to a newly created communicator.
    /// Blocks the caller until all nodes have entered the call.
    /// This call must be performed by all nodes in the caller communicator
    /// a creating any new communicators.  Returns mpicomm::null_comm
    /// on all nodes that are not participants the new group.
    mpicomm::ptr comm_create(const mpicomm::ptr &caller, mpigroup::ptr group);

    /// MPI_Comm_split -- collective operation.
    mpicomm::ptr comm_split(const mpicomm::ptr &caller, int color, int key);

    mpicomm::ptr create_cart(const mpicomm::ptr &caller, int ndims, int *dims, int *periods,
        int reorder);
  };

} } // end of namespace sstmac.

#endif
