/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_implementation/rdmampi.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/common/sstmac_env.h>
namespace sstmac
{
  namespace sw
  {

  SSTRegister("rdma", mpi_implementation, rdmampi,
             "An MPI that uses RDMA pathways and has zero-copy semantics for both intra- and inter-node");


    rdmampi::rdmampi(int smp_single_copy_size, int max_eager_msg_size, int max_vshort_msg_size) :
      mpi_implementation(24), //envelope size
        smp_single_copy_size_(smp_single_copy_size),
        max_eager_msg_size_(max_eager_msg_size),
        max_vshort_msg_size_(max_vshort_msg_size)
    {
    }

    mpiprotocol::ptr
    rdmampi::get_protocol(long bytes, bool intranode, bool ssend, bool onesided) const
    {

      if (intranode)
      {
        if (onesided)
        {
          return mpiprotocol::rendezvous_mmap_protocol;
        }

        if (bytes > smp_single_copy_size_)
        {
          return mpiprotocol::rendezvous_mmap_protocol;
        }
        else
        {
          if (ssend)
          {
            return mpiprotocol::eager_ssend_protocol;
          }
          else
          {
            return mpiprotocol::eager0_mmap_protocol;
          }
        }
      }
      else
      {
        if (onesided)
        {
          return mpiprotocol::rendezvous_rdma_protocol;
        }

        if (bytes <= max_vshort_msg_size_)
        {
          return mpiprotocol::eager0_rdma_protocol;
        }
        else if (!ssend && bytes <= max_eager_msg_size_)
        {
          return mpiprotocol::eager1_rdma_singlecpy_protocol;
        }
        else
        {
          return mpiprotocol::rendezvous_rdma_protocol;
        }
      }
    }

    rdmampi::ptr
    rdmampi::construct(const sim_parameters::ptr& params)
    {
        /**
            sstkeyword {
                docstring=The crossover point at which SMP/on-node sends switch from
                an eager, double-copy protocol to a rendezvous, single-copy protocol.;
            }
        */
        int smp_single_copy_size = sstmac_env::params->get_optional_int_param(
            "smp_single_copy_size", 8192);
        /**
            sstkeyword {
                docstring=The crossover point at which internode sends switch from
                an eager RDMA protocol to a rendezvous, zero-copy RDMA protocol.;
            }
        */
        int max_eager_msg_size = sstmac_env::params->get_optional_int_param(
            "max_eager_msg_size", 8192);
        /**
            sstkeyword {
                docstring=The crossover point at which internode sends switch from
                using low-latency, short message mailboxes to an eager RDMA protocol;
            }
        */
        int max_vshort_msg_size = sstmac_env::params->get_optional_int_param(
            "max_vshort_msg_size", 1024);
        return new rdmampi(smp_single_copy_size, max_eager_msg_size, max_vshort_msg_size);
    }

    int64_t
    rdmampi::max_eager_msg_size() const
    {
      return max_eager_msg_size_;
    }

    int64_t
    rdmampi::max_vshort_msg_size() const
    {
      return max_vshort_msg_size_;
    }

  }
}
