/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue-recvrequest.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue-sendrequest.h>
#include <sstmac/software/api/mpi/rma/mpiqueue-getrequest.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue-proberequest.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpiserver.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/software/api/mpi/mpi_implementation/mpi_implementation.h>
#include <sstmac/common/sim_parameters.h>
#include <sstmac/software/api/mpi/mpi_protocol/mpiprotocol.h>
#include <sstmac/software/api/mpi/rma/mpi_rma_message.h>

#include <sstmac/common/factories/factory.h>

#include <sstmac/common/debug.h>
#include <sstmac/common/errors.h>
#include <stdint.h>

ImplementFactory(sstmac::sw::mpiqueue);

namespace sstmac
{
  namespace sw
  {

    // Pointer types.

    debug mpiqueue::dbg_;

    stat_spyplot::ptr mpiqueue::spy_;

    stat_spyplot::ptr mpiqueue::spy_bytes_;
    stat_spyplot_png::ptr mpiqueue::spy_bytes_png_;

    stat_spyplot::ptr mpiqueue::spy_congestion_;


    bool
    mpiqueue::sortbyseqnum::operator()(const mpimessage::ptr &a,
        const mpimessage::ptr &b) const
    {
      return (a->seqnum() < b->seqnum());
    }

    void
    mpiqueue::init(const mpiserverPtr &serv)
    {
        if (serv == 0)
          sst_throw(nullerror,"mpiqueue::init:  null server pointer");
        server_ = serv;
        mintrans_ = 0;
        os_ = serv->get_full_os();
    }

    void
    mpiqueue::set_eventmanager(const eventmanager::ptr& m)
    {
      if (!spy_)
      {
        spy_ = stat_spyplot::construct("<stats> spyplot | spyplot_csv", "mpi_spyplot.csv");
        m->register_stat(spy_);

        spy_bytes_ = stat_spyplot::construct("<stats> spyplot | spyplot_csv",
            "mpi_spyplot_bytes.csv");
        m->register_stat(spy_bytes_);

        spy_bytes_png_ = stat_spyplot_png::construct("<stats> spyplot | spyplot_png",
            "mpi_spyplot.png", true);
        m->register_stat(spy_bytes_png_);

        spy_congestion_
            = stat_spyplot::construct("<stats> congestion | mpi_congestion",
                "mpi_spyplot_congestion.csv");
        m->register_stat(spy_congestion_);
      }
      server_->set_eventmanager(m);
    }

    //
    // Hi there.
    //
    mpiqueue::mpiqueue() :
      mintrans_(0), next_id_(0), taskid_(0), appid_(0),
          mpi_impl_(0), user_lib_mem_(0), user_lib_time_(0),
          os_(0), server_(0)
    {
      mpiprotocol::init();

      if (!dbg_.is_real())
      {
        dbg_ = debug("<debug> mpi | mpiqueue", logger::magenta);
        logger::register_logger(&dbg_);
      }

    }

    void
    mpiqueue::init(const softwareid& myid)
    {
        taskid_ = myid.task_;
        appid_ = myid.app_;
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::init on node "
              << operatingsystem::get_current_node()->to_string() << "\n";

        // need to debug dangling messages.
        next_id_ = messageid(taskid_.id * int64_t(4000000000LL));

        std::string libname = sstprintf("mpiqueue-user-lib-mem-%d-%d",
            taskid_.id, appid_.id);
        user_lib_mem_ = lib_compute_memmove::construct(libname);
        server_->get_os()->register_lib(user_lib_mem_);

        user_lib_time_ = lib_compute_time::construct(
            sstprintf("mpiqueue-user-lib-time-%d-%d", taskid_.id, appid_.id));
        server_->get_os()->register_lib(user_lib_time_);
    }

    void
    mpiqueue::unregister(const softwareid& sid)
    {
      server_->get_os()->unregister_lib(user_lib_mem_);
      user_lib_mem_->unregister_all_libs();
      server_ = 0;
    }

    void
    mpiqueue::clear_statics()
    {
      spy_ = 0;
      spy_bytes_ = 0;
      spy_bytes_png_ = 0;
    }

    //
    // Goodbye.
    //
    mpiqueue::~mpiqueue() throw ()
    {
    }

    void
    mpiqueue::get(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &dest, const mpicomm::ptr &comm,
        const mpi_window::ptr &win, const mpi_rma_message::op_info& op,
        const eventhandler::ptr &completion)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::get("
            << "count=" << count << ", type=" << type.to_string() << ", dest="
            << dest.to_string() << ", comm=" << comm->to_string() << ")\n";

      const taskid& dst_tid = comm->peer_task(dest);
      const taskid& src_tid = comm->my_task();
      const mpiid& src_rank = comm->rank();

      nodeaddress::ptr src = comm->node_at(src_rank);
      nodeaddress::ptr dst = comm->node_at(dest);

      bool intranode = src->equals(dst);
      long bytes = count * int64_t(type.get_packed_size());
      mpiprotocol::ptr protocol = mpiprotocol::rendezvous_rma_protocol;

      bool handshake_only = protocol->handshake_only();
      int msg_count = -count; //negative size indicates handshake

      timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something

      //even if it carries the payload, we still tag this as a header since it's the first
      //message sent
      mpimessage::content_type_t ctype = mpimessage::header;
      mpimessage::MESSAGE_TYPES ty = mpimessage::RDMA_GET_REQ;
      mpi_rma_message::rmatype_t rmat = mpi_rma_message::get;

      mpi_rma_message::ptr mess = mpi_rma_message::construct(
          server_->get_lib_name(), sendtime, mpi_impl_->get_envelope_size(),
          mintrans_, msg_count, type, src_rank, dest, comm->id(),
          next_outbound_[dst_tid]++, next_id_++, mpimessage::user, src_tid,
          dst_tid, comm->get_app(), ctype, payload::const_ptr(), protocol,
          rmat, win->winid(), op,  ty);
      mess->fromaddr_ = src;
      mess->toaddr_ = dst;

      getrequest::ptr req = getrequest::construct(key, mpiqueue::ptr(this),
          completion);

      if (server_ == 0)
      {
        throw illformederror("mpiqueue::get:  null server pointer.");
      }

      /** push this on first. important! */
      mess->set_needs_ack(false);
      gets_waiting_for_data_[mess->id()] = req;

      spy_->add_one(taskid_.id, dst_tid.id);
      spy_bytes_->add(taskid_.id, dst_tid.id, bytes);
      spy_bytes_png_->add(taskid_.id, dst_tid.id, bytes);

      do_send(mess);

      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::get: Finished init get\n";
    }

    void
    mpiqueue::put(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &dest, const mpicomm::ptr &comm,
        const mpi_window::ptr &win, const mpi_rma_message::op_info& op,
        const eventhandler::ptr &completion, const payload::const_ptr& content)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::put("
            << "count=" << count << ", type=" << type.to_string() << ", dest="
            << dest.to_string() << ", comm=" << comm->to_string() << ")\n";

      const taskid& dst_tid = comm->peer_task(dest);
      const taskid& src_tid = comm->my_task();
      const mpiid& src_rank = comm->rank();

      nodeaddress::ptr src = comm->node_at(src_rank);
      nodeaddress::ptr dst = comm->node_at(dest);

      bool intranode = src->equals(dst);
      long bytes = count * int64_t(type.get_packed_size());
      mpiprotocol::ptr protocol = mpiprotocol::rendezvous_rma_protocol;

      bool handshake_only = protocol->handshake_only();
      int msg_count = count; //positive size indicates data

      timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something

      //even if it carries the payload, we still tag this as a header since it's the first
      //message sent
      mpimessage::content_type_t ctype = mpimessage::header;
      mpimessage::MESSAGE_TYPES ty = mpimessage::RDMA_PUT_REQ;
      mpi_rma_message::rmatype_t rmat = mpi_rma_message::put;

      mpi_rma_message::ptr mess = mpi_rma_message::construct(
          server_->get_lib_name(), sendtime, mpi_impl_->get_envelope_size(),
          mintrans_, msg_count, type, src_rank, dest, comm->id(),
          next_outbound_[dst_tid]++, next_id_++, mpimessage::user, src_tid,
          dst_tid, comm->get_app(), ctype, content, protocol,
          rmat, win->winid(), op, ty);
      mess->fromaddr_ = src;
      mess->toaddr_ = dst;

      sendrequest::ptr req = sendrequest::construct(mess, key, mpiqueue::ptr(this),
          completion);

      if (server_ == 0)
      {
        throw illformederror("mpiqueue::put:  null server pointer.");
      }

      /** push this on first. important! */
      mess->set_needs_ack(false);
      send_needs_completion_ack_[mess->id()] = req;

      spy_->add_one(taskid_.id, dst_tid.id);
      spy_bytes_->add(taskid_.id, dst_tid.id, bytes);
      spy_bytes_png_->add(taskid_.id, dst_tid.id, bytes);

      do_send(mess);

      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::get: Finished init put\n";
    }

    //
    // Send data.
    //
    void
    mpiqueue::send(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &dst_rank, const mpitag &tag, const mpicomm::ptr &comm,
        const sendinfo &sinfo, mpimessage::category message_type,
        const eventhandler::ptr &completion, const payload::const_ptr& content)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::send("
            << "count=" << count << ", type=" << type.to_string() << ", dest="
            << dst_rank.to_string() << ", tag=" << tag.to_string() << ", comm="
            << comm->to_string() << ", cat=" << mpimessage::str(message_type)
            << ")\n";

      const taskid& dst_tid = comm->peer_task(dst_rank);
      const taskid& src_tid = comm->my_task();
      const mpiid& src_rank = comm->rank();

      nodeaddress::ptr src = comm->node_at(src_rank);
      nodeaddress::ptr dst = comm->node_at(dst_rank);

      bool intranode = src->equals(dst);
      long bytes = count * int64_t(type.get_packed_size());
      mpiprotocol::ptr protocol = get_protocol(bytes, intranode, sinfo.ssend,
          sinfo.onesided);
      bool handshake_only = protocol->handshake_only();
      int msg_count = handshake_only ? -count : count; //negative size indicates handshake

      timestamp sendtime(1E-9); // in the future, can make this separate send to the ami or something

      //even if it carries the payload, we still tag this as a header since it's the first
      //message sent
      mpimessage::content_type_t ctype = mpimessage::header;
      mpimessage::MESSAGE_TYPES ty = mpimessage::MPI;

      mpimessage::ptr mess = mpimessage::construct(server_->get_lib_name(),
          sendtime, mpi_impl_->get_envelope_size(), mintrans_, msg_count, type,
          src_rank, dst_rank, tag, comm->id(), next_outbound_[dst_tid]++,
          next_id_++, sinfo.ssend, message_type, src_tid, dst_tid,
          comm->get_app(), ctype, content, protocol, ty);
      mess->fromaddr_ = src;
      mess->toaddr_ = dst;

      sendrequest::ptr req = sendrequest::construct(mess, key,
          mpiqueue::ptr(this), completion);

      if (server_ == 0)
      {
        throw illformederror("mpiqueue::send:  null server pointer.");
      }

      /** push this on first. important! */
      mess->set_needs_ack(false);
      if (mess->get_protocol()->send_needs_nic_ack())
      {
        mess->set_needs_ack(true);
        send_needs_nic_ack_.push_back(req);
      }
      else if (mess->get_protocol()->send_needs_eager_ack())
      {
        send_needs_eager_ack_.push_back(req);
      }
      else if (mess->get_protocol()->send_needs_rendezvous_ack())
      {
        send_needs_rendezvous_ack_[mess->id()] = req;
      }
      else if (mess->get_protocol()->send_needs_completion_ack())
      {
        send_needs_completion_ack_[mess->id()] = req;
      }

      spy_->add_one(taskid_.id, dst_tid.id);
      spy_bytes_->add(taskid_.id, dst_tid.id, bytes);
      spy_bytes_png_->add(taskid_.id, dst_tid.id, bytes);

      do_send(mess);

      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::send: Finished init send\n";
    }

    void
    mpiqueue::handle_put(const mpimessage::ptr &message)
    {

    }

    mpiprotocol::ptr
    mpiqueue::get_protocol(long bytes, bool intranode, bool rsend,
        bool onesided) const
    {
      return mpi_impl_->get_protocol(bytes, intranode, rsend, onesided);
    }

    void
    mpiqueue::start_recv(const recvrequest::ptr& req)
    {
      mpimessage::ptr mess = find_matching_recv(req);
      if (mess)
      {
        if (mess->count() > 0)
        {
            buffered_recv(mess, req);
        }
        else
        {
          req->handle(mess);
        }
      }
    }

    mpimessage::ptr
    mpiqueue::find_matching_recv(const recvrequestPtr &req)
    {
      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
      {
        mpimessage::ptr mess = *it;
        if (req->matches(mess))
        {
          SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                << "]::recv:  Matching message found.\n";
          // We're good to go.

          need_recv_.erase(it);
          return mess;

        }
      }
      // If we get here, we still need to wait for the message.
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::recv:  Matching message not found\n";
      pending_message_.push_back(req);
      return mpimessage::ptr();
    }

    //
    // Receive data.
    //
    void
    mpiqueue::recv(const mpirequest_t key, int count, const mpitype &type,
        const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
        mpimessage::category message_type, eventhandler::ptr &completion)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::recv("
            << "count=" << count << ", type=" << type.to_string()
            << ", source=" << source.to_string() << ", tag=" << tag.to_string()
            << ", comm=" << comm->to_string() << ", cat=" << mpimessage::str(
            message_type) << ")\n";

      recvrequest::ptr req = recvrequest::construct(key, mpiqueue::ptr(this),
          count, type, source, tag, comm, message_type, completion);

      do_recv(req);
    }

    void
    mpiqueue::finalize_recv(const mpimessage::ptr &msg)
    {
      spy_congestion_->add(msg->source().id, taskid_.id,
          msg->get_delay().msec());
    }

    //
    // Ask for a notification when a message with the given signature arrives.
    //
    void
    mpiqueue::probe(const mpirequest_t key, const mpicomm::ptr &comm,
        const mpiid &source, const mpitag &tag, eventhandler::ptr &completion)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::probe("
            << comm->to_string() << ", " << source.to_string() << ", "
            << tag.to_string() << ", " << ")\n";

      proberequest::ptr req = proberequest::construct(key, comm, source, tag,
          completion);
      // Figure out whether we already have a matching message.
      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
      {
        if ((*it)->cat() == mpimessage::user && req->matches(*it))
        {
          // We're good to go.
          req->complete(*it);
          return;
        }
      }
      // If we get here, we still need to wait for the message.
      probelist_.push_back(req);
    }

    //
    // Immediate-mode probe for a message with the given signature.
    //
    mpistatus::ptr
    mpiqueue::iprobe(const mpicomm::ptr &comm, const mpiid &source,
        const mpitag &tag)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string() << "]::iprobe("
            << comm->to_string() << ", " << source.to_string() << ", "
            << tag.to_string() << ")\n";
      proberequest::ptr req = proberequest::construct(mpirequest_t(), comm,
          source, tag, eventhandler::ptr());
      need_recv_t::iterator it, end = need_recv_.end();
      for (it = need_recv_.begin(); it != end; ++it)
      {
        if ((*it)->cat() == mpimessage::user && req->matches(*it))
        {
          // This is it.
          return (*it)->build_status();
        }
      }
      // If we got here, there was no match
      return mpistatus::ptr();
    }

    //
    // Used when a receive operation sends an ack.
    //
    void
    mpiqueue::send_rendezvous_ack(const mpimessage::ptr& message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::send_rendezvous_ack(" << message << ")\n";
      message->header_to_rendezvous_ack();
      server_->send(message);
    }

    void
    mpiqueue::send_completion_ack(const mpimessage::ptr& message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::send_completion_ack(" << message << ")\n";
      mpimessage::ptr copy = message->deep_clone();
      copy->payload_to_completion_ack();
      server_->send(copy);
    }

    //
    // CALLBACK used by mpiserver when this object has a message.
    //
    void
    mpiqueue::incoming_message(const mpimessage::ptr& message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::incoming_message(" << ", " << message->to_string() << ")\n";

      if (message->get_type() == sst_message::SEND_ACK)
      {
        this->handle_nic_ack(message);
        return;
      }

      /// These are the types of messages we may expect:
      /// (1)  Response to handshake request (we have the requestor).
      /// (2)  Data delivery following a handshake ack (we have the receiver)
      /// (3)  Acks from the NIC that a message has left
      /// (4)  New messages (either eager-send or a handshake request).
      /// Only category (4) is subject to order requirements.  The other
      /// categories get handled immediately

      switch (message->content_type())
      {
      case mpimessage::rendezvous_ack:
        this->incoming_rendezvous_ack(message);
        break;
      case mpimessage::completion_ack:
        this->incoming_completion_ack(message);
        break;
      case mpimessage::data:
        message->get_protocol()->handle_payload(mpiqueue::ptr(this), message);
        break;
      case mpimessage::header:
        this->incoming_new_message(message);
        break;
      default:
        sst_throw(valueerror, "received invalid messgae content type %s in incoming message",
            mpimessage::str(message->content_type()));
      }
    }

    //
    // Handle a response to an ack posted from this queue.
    //
    void
    mpiqueue::incoming_completion_ack(const mpimessage::ptr &message)
    {
      const messageid &id = message->id();
      SSTMAC_DEBUG << "  :  Incoming ack from task "
            << message->source_task().to_string() << " to task "
            << message->dest_task().to_string() << " with id "
            << id.to_string() << "\n" << " : Finding ack for "
            << message->id().to_string() << "\n";

      ack_needed_t::iterator it =
          send_needs_completion_ack_.find(message->id());
      if (it == send_needs_completion_ack_.end())
      {
        sst_throw(illformederror,
            "mpiqueue[%s]::incoming_message: completion ack with no match on %s",
            taskid_.to_string().c_str(), message->to_string().c_str());
      }

      // We have a winner.
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::incoming_message:  "
            << " rendezvous ack with matching request.\n";
      sendrequest::ptr req = it->second;
      send_needs_completion_ack_.erase(it);
      req->complete(message);
    }

    //
    // Handle a response to an ack posted from this queue.
    //
    void
    mpiqueue::incoming_rendezvous_ack(const mpimessage::ptr &message)
    {
      const messageid &id = message->id();
      SSTMAC_DEBUG << "  :  Incoming ack from task "
            << message->source_task().to_string() << " to task "
            << message->dest_task().to_string() << " with id "
            << id.to_string() << "\n" << " : Finding ack for "
            << message->id().to_string() << "\n";

      ack_needed_t::iterator it =
          send_needs_rendezvous_ack_.find(message->id());
      if (it == send_needs_rendezvous_ack_.end())
      {
        sst_throw(illformederror,
            "mpiqueue[%s]::incoming_message: rendezvous ack with no match on %s",
            taskid_.to_string().c_str(), message->to_string().c_str());
      }

      // We have a winner.
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::incoming_message:  "
            << "completion ack with matching request.\n";

      sendrequest::ptr req = it->second;
      send_needs_rendezvous_ack_.erase(it);
      message->rendezvous_ack_to_payload();

      /** the message either needs a nic ack or an mpi completion ack
       to finish everything off
       */
      if (message->get_protocol()->send_needs_completion_ack())
      {
        send_needs_completion_ack_[message->id()] = req;
        message->set_needs_ack(false);
      }
      else
      {
        send_needs_nic_ack_.push_back(req);
        message->set_needs_ack(true);
      }
      server_->send(message);
    }

    //
    // Handle a new incoming message.  Can be either an eager send (with data)
    // or a new handshake request
    //
    void
    mpiqueue::incoming_new_message(const mpimessage::ptr &message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]: incoming_new_message \n";
      taskid tid = message->source_task();

      if (message->seqnum() == next_inbound_[tid])
      {
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]: seqnum matched expected " << next_inbound_[tid]
              << " for task " << tid.id << "\n";
        // Handle this message.
        handle_new_message(message);
        ++next_inbound_[tid];
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]: advanced to seqnum " << next_inbound_[tid] << " for task "
              << tid.id << "\n";
        // Handle any messages that have been freed by the arrival of this one
        if (!held_[tid].empty())
        {
          SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                << "]: there are old messages for task " << tid.id << "\n";

          hold_list_t::iterator it = held_[tid].begin(), end = held_[tid].end();
          while (it != end)
          {
            mpimessage::ptr mess = *it;
            if (mess->seqnum() <= next_inbound_[tid])
            {
              //it = held_[tid].erase(it);
              it++;
              handle_new_message(mess);
              ++next_inbound_[tid];

            }
            else
            {
              break;
            }

          }
          // Now erase all the completed messages from the held queue.
          held_[tid].erase(held_[tid].begin(), it);
        }
      }
      else if (message->ignore_seqnum())
      {
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]: got rdma seqnum\n";
        handle_new_message(message);
      }
      else
      {
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]: putting message in held, seqnum " << message->seqnum()
              << " didn't match expected " << next_inbound_[tid]
              << " for task " << tid.id << "\n";
        held_[tid].insert(message);
      }

      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]: incoming_new_message done \n";
    }

    void
    mpiqueue::notify_probes(const mpimessage::ptr& message)
    {
      // And, finally, mpimessage::user messages are tested for outstanding probes
      // list<proberequest::ptr> proberemove;
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]:probelist size: " << probelist_.size() << "\n";
      probelist_t::iterator pit = probelist_.begin(), pend = probelist_.end();
      while (pit != pend)
      {
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]: testing probe \n";
        if ((*pit)->matches(message))
        {
          SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                << "]: probe match \n";
          // Match.
          proberequest::ptr preq = *pit;

          pit = probelist_.erase(pit);
          preq->complete(message);
        }
        else
        {
          pit++;
        }

        pend = probelist_.end();
      }
    }

    void
    mpiqueue::handle_new_message(const mpimessage::ptr &message)
    {
      message->get_protocol()->handle_header(mpiqueue::ptr(this), message);
    }

    //
    // Complete an inbound message.
    //
    mpiqueue::recvrequest::ptr
    mpiqueue::find_request(pending_message_t &pending,
        const mpimessage::ptr &message)
    {
      recvrequest::ptr req;
      pending_message_t::iterator it, end, tmp;
      end = pending.end();
      for (it = pending.begin(); it != end;)
      {
        req = *it;
        tmp = it++;
        if (req->is_cancelled())
        {
          pending.erase(tmp);
        }
        else if (req->matches(message))
        {
          SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                << "]::handle_new_message:  Receiver is ready.\n";
          pending.erase(tmp);
          return req;
        }
      }
      return recvrequest::ptr();
    }

    mpiqueue::recvrequest::ptr
    mpiqueue::find_pending_request(const mpimessage::ptr &message,
        bool set_need_recv)
    {
      recvrequest::ptr req = find_request(pending_message_, message);
      if (!req)
      {
        // We get here if no match was found.
        // Messages that don't have a respondent are added to the list
        SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
              << "]::handle_new_message:  receiver not ready for "
              << message->to_string() << "\n";
        if (set_need_recv)
          need_recv_.push_back(message);
      }
      return req;
    }

    mpiqueue::recvrequest::ptr
    mpiqueue::find_waiting_request(const mpimessage::ptr &message)
    {
      recvrequest::ptr req = find_request(waiting_message_, message);
      if (!req)
      {
        sst_throw(valueerror, "ERROR: matching request for waiting message not found");
      }
      return req;
    }

    void
    mpiqueue::complete_nic_ack(const mpimessage::ptr& message)
    {
      SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
            << "]::handle_nic_ack: " << message->to_string() << "\n";

      send_needs_ack_t::iterator it, end = send_needs_nic_ack_.end();
      for (it = send_needs_nic_ack_.begin(); it != end; ++it)
      {
        sendrequest::ptr sreq = *it;
        if (sreq->matches(message))
        {
          SSTMAC_DEBUG << "mpiqueue[" << taskid_.to_string()
                << "]::complete_nic_ack match found\n";

          // Match.

          send_needs_nic_ack_.erase(it);
          sreq->complete(message);
          return;
        }
      }
      sst_throw(valueerror, "no match found for nic ack on %s",
          message->to_string().c_str());
    }

    void
    mpiqueue::handle_nic_ack(const mpimessage::ptr &message)
    {
      message->get_protocol()->handle_nic_ack(mpiqueue::ptr(this), message);
    }

    const mpiserver::ptr&
    mpiqueue::server() const
    {
      return server_;
    }

    const lib_compute_memmove::ptr&
    mpiqueue::user_lib_mem() const
    {
      return user_lib_mem_;
    }

    const lib_compute_time::ptr&
    mpiqueue::user_lib_time() const
    {
      return user_lib_time_;
    }

    std::string
    mpiqueue::id_string() const
    {
      return taskid_.to_string();
    }


  }
} // end of namespace sstmac
