/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_QUEUE_MPIQUEUE_H_INCLUDED

#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/api/mpi/mpimessage.h>
#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpi_implementation/mpi_implementation.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/common/eventmanager.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/common/messages/messageid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/debug.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/common/types.h>
#include <sstmac/common/stats/stats_common.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/libraries/compute/lib_compute_time.h>
#include <sstmac/software/api/mpi/rma/mpi_window.h>
#include <sstmac/common/factories/factory.h>

namespace sstmac
{
  namespace sw
  {

    // Important types.
    class mpistatus;
    class nodeinterface;
    class mpiserver;
    class mpi_implementation;
    class mpiprotocol;
    typedef boost::intrusive_ptr<mpiprotocol> mpiprotocolPtr;
    typedef boost::intrusive_ptr<mpiserver> mpiserverPtr;
    typedef boost::intrusive_ptr<mpi_implementation> mpi_implementationPtr;

    struct appid;

    ///* A one-per-peer queue that handles inbound and outbound messages.
    ///*
    ///*


    class mpiqueue :
            virtual public nocopy,
            public factory_type
    {

      /** temporary fix until I figure out a better way to do this */
      friend class mpiserver;

      friend class mpiprotocol;
      friend class eager0_protocol;
      friend class eager0_socket;
      friend class eager0_rdma;
      friend class eager0_mmap;
      friend class eager1_rdma;
      friend class eager1_rdma_singlecpy;
      friend class eager1_rdma_doublecpy;
      friend class rendezvous_protocol;
      friend class rendezvous_mmap;
      friend class rendezvous_socket;
      friend class rendezvous_rdma;
      friend class creme_mpiprotocol;
      friend class eager_ssend;
      friend class rendezvous_rma;

    public:
      typedef boost::intrusive_ptr<mpiqueue> ptr;
      class recvrequest;
      class sendrequest;
      class proberequest;
      class getrequest;

      typedef boost::intrusive_ptr<recvrequest> recvrequestPtr;
      typedef boost::intrusive_ptr<sendrequest> sendrequestPtr;
      typedef boost::intrusive_ptr<proberequest> proberequestPtr;
      typedef boost::intrusive_ptr<getrequest> getrequestPtr;

      struct sendinfo
      {
        sendinfo() :
          rsend(false), ssend(false), onesided(false)
        {
        }
        bool rsend;
        bool ssend;
        bool onesided;
      };

    private:

      /// Nested types.

      /// Utility class to sort arriving messages in arrival order.
      struct sortbyseqnum
      {
        bool
        operator()(const mpimessage::ptr &a, const mpimessage::ptr &b) const;
      };

      /// Minimum transmission size.
      int64_t mintrans_;
      /// The sequence number for our next outbound transmission.
      boost::unordered_map<taskid, int> next_outbound_;

      /// The sequence number expected for our next inbound transmission.
      boost::unordered_map<taskid, int> next_inbound_;

      /// Hold messages that arrived out of order.
      typedef std::set<mpimessage::ptr, sortbyseqnum, std::allocator<
          mpimessage::ptr> > hold_list_t;

      boost::unordered_map<taskid, hold_list_t> held_;

      /// The (locally unique) id that will be given to the next message.
      messageid next_id_;

      /// Receive requests waiting for a matching message (or envelope).
      typedef std::list<recvrequestPtr> pending_message_t;
      pending_message_t pending_message_;

      pending_message_t waiting_message_;

      /// Inbound messages waiting for a matching receive request.
      typedef std::list<mpimessage::ptr> need_recv_t;
      need_recv_t need_recv_;
      need_recv_t need_rdma_get_;

      // Save all sends so we can match up the nic acks that come back
      typedef std::list<sendrequestPtr> send_needs_ack_t;
      send_needs_ack_t send_needs_nic_ack_;
      send_needs_ack_t send_needs_eager_ack_;

      /// Inbound messages that arrived out-of-order and are waiting for
      /// the message sequence to be complete.
      typedef boost::unordered_map<int, mpimessage::ptr> reorderlist_t;
      reorderlist_t reorderlist_;

      /// Requests waiting for an ack.
      typedef std::map<messageid, sendrequestPtr> ack_needed_t;
      ack_needed_t send_needs_rendezvous_ack_;
      ack_needed_t send_needs_completion_ack_;

      /// Requests that sent an ack and are waiting for the rest of their data.
      typedef std::map<messageid, recvrequestPtr> pending_req_map;
      pending_req_map recv_needs_payload_;

      /// Probe requests watching for a given envelope.
      typedef std::list<proberequestPtr> probelist_t;
      probelist_t probelist_;

      static stat_spyplot::ptr spy_;

      static stat_spyplot::ptr spy_bytes_;
      static stat_spyplot_png::ptr spy_bytes_png_;

      static stat_spyplot::ptr spy_congestion_;

    protected:
      static debug dbg_;

      taskid taskid_;

      appid appid_;

      mpi_implementationPtr mpi_impl_;

      boost::unordered_map<int, mpi_window::ptr> windows_;
      /// Requests waiting for a get rsp
      typedef std::map<messageid, getrequestPtr> get_requests_t;
      get_requests_t gets_waiting_for_data_;

      boost::unordered_map<long, std::queue<mpimessage::ptr> > headers_waiting_epoch_;

      /// A blocking memcpy library for buffered sends in which
      /// the memcpy happens in the application
      lib_compute_memmove::ptr user_lib_mem_;
      lib_compute_time::ptr user_lib_time_;

      operatingsystemPtr os_;

      mpiserverPtr server_;

      /// Hi there.
      mpiqueue();

    public:

      mpi_implementation::ptr
      get_implementation() const {
          return mpi_impl_;
      }

      void set_implementation(const mpi_implementation::ptr& impl){
          mpi_impl_ = impl;
      }

      virtual void init(const mpiserverPtr& serv);

      static void
      clear_statics();

      virtual std::string
      to_string() const
      {
        return "mpiqueue";
      }

      /// Goodbye.
      virtual
      ~mpiqueue() throw ();

      void
      unregister(const softwareid& sid);

      /// Send data.  All sends are triggered at time nodeinterface::now().
      /// Upon completion, the completion object have an updated
      /// status object relating to the send operation.
      void
      send(const mpirequest_t key, int count, const mpitype &type,
          const mpiid &dest, const mpitag &tag, const mpicomm::ptr &comm,
          const sendinfo &sinfo, mpimessage::category message_type,
          const eventhandler::ptr &completion,
          const payload::const_ptr& content = payload::null());

      void
      get(const mpirequest_t key, int count, const mpitype &type,
          const mpiid &dest, const mpicomm::ptr &comm,
          const mpi_window::ptr &win, const mpi_rma_message::op_info& op, const eventhandler::ptr &completion);

      void
      put(const mpirequest_t key, int count, const mpitype &type,
          const mpiid &dest, const mpicomm::ptr &comm,
          const mpi_window::ptr &win, const mpi_rma_message::op_info& op, const eventhandler::ptr &completion,
          const payload::const_ptr& content = payload::null());

      /// Receive data.  All receives are triggered at time nodeinterface::now().
      /// Upon completion, the receiver object will be updated to
      /// hold a copy of the inbound mpimessage with updated status information
      /// including any inbound content (null or not).
      void
      recv(const mpirequest_t key, int count, const mpitype &type,
          const mpiid &source, const mpitag &tag, const mpicomm::ptr &comm,
          mpimessage::category message_type, eventhandler::ptr &completion);

      /// Ask for a notification when a message with the given signature arrives.
      /// Probing is only enabled for message_type==mpimessage::user.
      /// Upon completion, the status object held by the completion object
      /// will be updated to reflect the envelope information for the incoming
      /// message, and the completion object will be notified
      /// (using eventhandler::complete).
      void
          probe(const mpirequest_t key, const mpicomm::ptr &comm,
              const mpiid &source, const mpitag &tag,
              eventhandler::ptr &completion);

      /// Immediate-mode probe for whether a message with the given signature
      /// has arrived.
      /// Probing is only enabled for message_type==mpimessage::user.
      /// \return null if no such message is pending.
      mpistatus::ptr
      iprobe(const mpicomm::ptr &comm, const mpiid &source, const mpitag &tag);

      /// CALLBACK used by mpiserver when this object has a message.
      virtual void
      incoming_message(const mpimessage::ptr& message) = 0;

      void
      set_eventmanager(const eventmanager::ptr &m);

      virtual void init(const softwareid& id);

      void
      win_create(mpi_window::ptr win)
      {
        windows_[win->winid()] = win;
      }

      void
      win_free(mpi_window::ptr win){
        windows_.erase(win->winid());
      }

      mpiprotocolPtr
      get_protocol(long bytes, bool intranode, bool rsend, bool onesided) const;

      int64_t
      get_envelope_size() const;

      const lib_compute_memmove::ptr&
      user_lib_mem() const;

      const lib_compute_time::ptr&
      user_lib_time() const;

      const mpiserverPtr&
      server() const;

      void
      finalize_recv(const mpimessage::ptr& msg);

      virtual timestamp
      progress_loop(const mpirequest_t& req) = 0;

      virtual void
      start_progress_loop(const std::vector<mpirequest_t>& req) = 0;

      virtual void
      start_progress_loop(const std::vector<mpirequest_t>& req,
          timestamp timeout) = 0;

      virtual void
      finish_progress_loop(const std::vector<mpirequest_t>& req) = 0;

      virtual eventhandler::ptr
      progress_done_handler(const OS_API::ptr& os, const mpirequest_t& req) = 0;

      void
      start_recv(const recvrequestPtr& req);

      virtual void
      buffered_send(const mpimessage::ptr& msg) = 0;

      virtual void
      buffered_recv(const mpimessage::ptr& msg, const recvrequestPtr& req) = 0;

      virtual void
      buffer_unexpected(const mpimessage::ptr& msg) = 0;

      virtual void
      post_rdma(const mpimessage::ptr& msg) = 0;

      virtual void
      post_header(const mpimessage::ptr& msg) = 0;

      void
      handle_put(const mpimessage::ptr &message);

    protected:

      virtual void
      do_send(const mpimessage::ptr& mess) = 0;

      virtual void
      do_recv(const mpiqueue::recvrequestPtr& req) = 0;

      /// The incoming_message method forwards to here when the message is
      /// a response to a handshake request.
      void
      incoming_rendezvous_ack(const mpimessage::ptr &message);

      void
      incoming_completion_ack(const mpimessage::ptr &message);

      /// The incoming_message method forwards to here when the message is
      /// a new message, whehter it's an eager send or a new handshake request.
      void
      incoming_new_message(const mpimessage::ptr &message);

      /// Called by the mpiserver when it figured out that the nic is sending
      /// back an ack that indicates that the message was sent
      void
      handle_nic_ack(const mpimessage::ptr &message);

      /// Called by the mpiserver when it figured out that the nic is sending
      /// back an ack that indicates that the message was sent
      void
      complete_nic_ack(const mpimessage::ptr &message);

      /// The incoming_new_message method calls here to complete an
      /// inbound message.  At this point, arrival ordering has been ensured.
      void
      handle_new_message(const mpimessage::ptr &message);

      void
      notify_probes(const mpimessage::ptr& message);

      recvrequestPtr
      find_request(pending_message_t& pending, const mpimessage::ptr& message);

      recvrequestPtr
      find_pending_request(const mpimessage::ptr &message,
          bool set_need_recv = true);

      recvrequestPtr
      find_waiting_request(const mpimessage::ptr &message);

      mpimessage::ptr
      find_matching_recv(const recvrequestPtr& req);

      void
      send_rendezvous_ack(const mpimessage::ptr& message);

      void
      send_completion_ack(const mpimessage::ptr& message);

      std::string
      id_string() const;

    };

    DeclareFactory1InitParam(mpiqueue,mpiserverPtr);

  }
} // end of namespace sstmac


#endif
