/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_strategies/mpi_collective/engines/mpiringbarrierengine.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpi_strategies/mpi_iterator/topologyiterator.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {

    //
    // Fire off the next set of send and receive operations.
    //
    void
    mpiringbarrierengine::sendrecv(bool firsttime)
    {
      SSTMAC_DEBUG << "mpiringbarrierengine[" << comm_->rank().id
            << "]:: sendrecv \n";

      long size = comm_->size().id;
      if (firsttime)
        {
          if (comm_->rank().id == 0)
            {

              this->start_send(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id + 1) % size), payload::null());

              this->start_recv(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id - 1 + size) % size));
            }
          else
            {

              this->start_recv(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id - 1 + size) % size));
            }
        }

    }

    //
    // Callback method to indicate that a send operation has completed.
    //
    void
    mpiringbarrierengine::send_complete(int count, const mpitype &type,
        const mpitag &tag, const mpiid &dest)
    {
      SSTMAC_DEBUG << "mpiringbarrierengine[" << comm_->rank().id
            << "]::send_complete(" << count << ", " << type.to_string() << ", "
            << tag.to_string() << ", " << dest.to_string() << ")\n";

    }

    //
    // Callback method to indicate that a receive operation has completed.
    //
    void
    mpiringbarrierengine::recv_complete(int count, const mpistatus_t& stat)
    {
      SSTMAC_DEBUG << "mpiringbarrierengine[" << comm_->rank().id
            << "]::recv_complete(" << count << ", " << ")\n";

      long size = comm_->size().id;

      if (forward_)
        {
          forward_ = false;

          this->start_recv(1, mpitype::mpi_byte, tag_,
              mpiid((comm_->rank().id + 1) % size));

          if (comm_->rank().id == 0)
            {
              this->start_send(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id - 1 + size) % size), payload::null());
            }
          else
            {
              this->start_send(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id + 1) % size), payload::null());
            }
        }
      else
        {
          if (comm_->rank().id != 0)
            {
              this->start_send(1, mpitype::mpi_byte, tag_,
                  mpiid((comm_->rank().id - 1 + size) % size), payload::null());
            }

          mpicollective::complete(payload::const_ptr());
        }
    }

    //
    // Copy complete.
    //
    void
    mpiringbarrierengine::copy_complete(int, const mpitype&, const mpitag&,
        const mpiid&)
    {
      throw illformederror("mpiringbarrierengine::copy_complete:  "
          "MPI barrier doesn't copy data.");
    }

    //
    // Hi.
    //
    mpiringbarrierengine::mpiringbarrierengine(
        const topologyiterator::ptr& iter, const mpirequest::ptr& the_key,
        const mpiqueue::ptr& queue, const mpitag &tag, const mpicomm::ptr &comm,
        eventhandler::ptr completion) :
        mpicollective(the_key, queue, tag, comm, completion), iter_(iter), forward_(true)
    {
      if (tag_ == mpitag::any_tag)
        throw valueerror("mpiringbarrierengine(...)"
            ":  refusing to send using mpitag::any_tag.");
    }

    //
    // Hi.
    //
    mpiringbarrierengine::ptr
    mpiringbarrierengine::construct(const topologyiterator::ptr& iter,
        const mpirequest::ptr& the_key, const mpiqueue::ptr& queue,
        const mpitag &tag, const mpicomm::ptr &comm, eventhandler::ptr completion)
    {
      mpiringbarrierengine::ptr retval(
          new mpiringbarrierengine(iter, the_key, queue, tag, comm,
              completion));

      return retval;
    }

    //
    // Goodbye.
    //
    mpiringbarrierengine::~mpiringbarrierengine() throw ()
    {
    }

    //
    // Start this kernel.
    //
    void
    mpiringbarrierengine::start()
    {
      SSTMAC_DEBUG << "mpiringbarrierengine[" << comm_->rank().id
            << "]::start \n";
      mpicollective::start();
      this->sendrecv(true);
    }

  }
} // end of namespace sstmac

