/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_MPICOLLECTIVE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPI_STRATEGIES_MPI_COLLECTIVE_MPICOLLECTIVE_H_INCLUDED

#include <sstmac/software/api/mpi/mpi_comm/mpicomm.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpi_queue/mpiqueue.h>
#include <sstmac/common/eventhandler.h>
#include <sstmac/software/api/mpi/mpi_types/mpiop.h>

#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {

    /// Important types.

    //class mpimagic;


    /**
     * Base type for mpikernel objects that do collective operations.
     */
    class mpicollective : public ptr_type
    {

      /// Nested type to handle the actual send operations.
      class sender;
      /// Nested type to handle the actual receive operations.
      class recver;
      /// Nested type to handle funneled (one-at-a-time) copy operations.
      class funneledcopy;
      /// Nested type to handle concurrent (start-at-once) copy operations.
      class immediatecopy;

      /// Copy object for funneled copies.
      boost::intrusive_ptr<funneledcopy> funnel_;

    protected:
      /// Our communicator.
      mpicomm::ptr comm_;

      mpitag tag_;

      eventhandler::ptr completion_;

      /// Our queue.
      mpiqueue::ptr queue_;

      mpirequest_t req_;

      /// Information on total bytes sent and received.
      mpistatus::ptr status_;

    protected:
      static debug dbg_;

    protected:
      /// Hi.
      mpicollective(const mpirequest::ptr& when_complete,
          const mpiqueue::ptr& queue, const mpitag& tag,
          const mpicomm::ptr &comm, const eventhandler::ptr& completion);

      /*
       * Callbacks for the derived type to request send or receive operations.
       */

      /// Called by the derived type to request a send.
      void
      start_send(int count, const mpitype &type, const mpitag &tag,
          const mpiid &dest,
          const payload::const_ptr& content = payload::const_ptr());

      /// Called by the derived type to request a receive.
      void
      start_recv(int count, const mpitype &type, const mpitag &tag,
          const mpiid &dest);

      /// Called by the derived type to request a funneled copy operation
      /// (into a shared buffer).  Calls copy_complete when complete.
      void
      start_funneled_copy(int count, const mpitype &type, const mpitag &tag,
          const mpiid &partner);

      /// Called by the derived type to request an immediate copy operation
      /// (not into a shared buffer).  Calls copy_complete when complete.
      void
      start_immediate_copy(int count, const mpitype &type, const mpitag &tag,
          const mpiid &partner);

      /*
       * Callbacks that the derived type needs to provide to complete
       * a send or receive operation.
       */

      /// Callback method to indicate that a send operation has completed.
      virtual void
      send_complete(int count, const mpitype &type, const mpitag &tag,
          const mpiid &dest) = 0;

      /// Callback method to indicate that a receive operation has completed.
      virtual void
      recv_complete(int count, const mpistatus_t& stat) =0;

      /// Callback method to indicate that a copy operation has completed.
      virtual void
      copy_complete(int count, const mpitype &type, const mpitag &tag,
          const mpiid &source) = 0;

      void
      start();

      const payload::const_ptr
      combine_content(const payload::const_ptr &p1,
          const payload::const_ptr &p2, mpiop &op, int rankhere,
          int rankincoming);

    public:

      virtual std::string
      to_string() const = 0;

      typedef boost::intrusive_ptr<mpicollective> ptr;

      /// Goodbye.
      virtual
      ~mpicollective() throw ();

      /// Complete.  Calls mpikernel::complete.
      void
      complete(const payload::const_ptr& content);

    };

  }
} // end of namespace sstmac

#endif
