/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <iostream>
#include <sstream>
#include <sstmac/common/errors.h>
#include <sstmac/software/api/mpi/mpitypes.h>

DeclareSerializable(sstmac::sw::pairdata)
  ;
DeclareSerializable(sstmac::sw::vecdata)
  ;
DeclareSerializable(sstmac::sw::inddata)
  ;

namespace sstmac
{
  namespace sw
  {
    static int go_debug = 0;
    static int nextid = 0;

    mpitype::mpitype() :
      id(-1)
    {
      idata_ = NULL;
      vdata_ = NULL;
      pdata_ = NULL;
      type_ = NONE;
      size = -1;
      label = "";
      native_id = -1;
      combiner_ = -1;
    }

    //
    // MPI datatypes.
    //
    mpitype::mpitype(const std::string &labelit, const int sizeit, int align,
        int comb) :
      label(labelit), size(sizeit), id(nextid++), type_(PRIM),
          alignment_(align)
    {
      vdata_ = NULL;
      idata_ = NULL;
      pdata_ = NULL;
      extent_ = size;
      numints_ = 0;
      numaddr_ = 0;
      numdtype_ = 1;
      combiner_ = comb;
    }

    mpitype::mpitype(const std::string &labelit, const mpitype &b1,
        const mpitype &b2, int s, int comb) :
      label(labelit), id(nextid++), type_(PAIR), extent_(s), alignment_(0)
    {
      // size = b1.size + b2.size;
      vdata_ = NULL;
      idata_ = NULL;
      size = b1.size + b2.size;
      numints_ = 0;
      numaddr_ = 0;
      numdtype_ = 1;
      combiner_ = comb;
      pdata_ = pairdata::construct();
      pdata_->base1_ = b1;
      pdata_->base2_ = b2;

    }

    mpitype::mpitype(const std::string &labelit, const mpitype &base,
        int count, int block, int str, bool in_elem, int comb) :
      label(labelit), id(nextid++), type_(VEC), alignment_(0)
    {
      vdata_ = vecdata::construct();
      idata_ = inddata::ptr();
      pdata_ = NULL;
      vdata_->base_ = base;
      vdata_->count_ = count;
      vdata_->blocklen_ = (block);
      if (in_elem)
        vdata_->stride_ = (str) * base.get_extent();
      else
        vdata_->stride_ = str;

      //size = count * abs(vdata_->stride_);
      extent_ = (count - 1) * abs(vdata_->stride_) + base.extent_;
      size = count * block * base.size;
      native_id = id;

      numints_ = 3;
      numaddr_ = 0;
      numdtype_ = 1;
      combiner_ = comb;
    }

    mpitype::mpitype(const std::string &labelit,
        boost::intrusive_ptr<inddata> dat, int sz, int ext, int numint,
        int numdtype, int comb) :
      label(labelit), id(nextid++), type_(IND), size(sz), extent_(ext),
          alignment_(0), numints_(numint), numdtype_(numdtype)
    {
      idata_ = dat;
      vdata_ = vecdata::ptr();
      pdata_ = NULL;
      native_id = id;

      numaddr_ = 0;
      combiner_ = comb;
    }

    mpitype::mpitype(const mpitype &other)
    {
      deep_copy(other);
    }

    mpitype&
    mpitype::operator=(const mpitype& other)
    {
      deep_copy(other);
      return *this;
    }

    void
    mpitype::deep_copy(const mpitype& other)
    {
      label = other.label;
      size = other.size;
      type_ = other.type_;
      id = other.id;
      extent_ = other.extent_;
      alignment_ = other.alignment_;
      combiner_ = other.combiner_;
      numints_ = other.numints_;
      numaddr_ = other.numaddr_;
      numdtype_ = other.numdtype_;
      native_id = other.native_id;

      pdata_ = other.pdata_;
      vdata_ = other.vdata_;
      idata_ = other.idata_;

      //std::cout << "---mpitype deep copy for " << id << "! -- \n";


    }

    mpitype::~mpitype()
    {
    }

    int
    mpitype::get_lb() const
    {
      if (type_ == PRIM || type_ == PAIR)
      {
        return 0;
      }
      else if (type_ == VEC)
      {
        return vdata_->base_.get_lb();
      }
      else if (type_ == IND)
      {
        return idata_->lb_;
      }
      else
      {
        sst_throw(valueerror, "mpitype::get_lb - unknown type category");
      }
    }

    int
    mpitype::get_true_lb() const
    {
      if (type_ == PRIM || type_ == PAIR || type_ == VEC)
      {
        return 0;
      }
      else if (type_ == IND)
      {
        return idata_->mindisp_;
      }
      else
      {
        sst_throw(valueerror, "mpitype::get_lb - unknown type category");
      }
    }

    int
    mpitype::get_ub() const
    {
      if (type_ == PRIM || type_ == PAIR || type_ == VEC)
      {
        return extent_;
      }
      else if (type_ == IND)
      {
        return idata_->ub_;
      }
      else
      {
        sst_throw(valueerror, "mpitype::get_lb - unknown type category");
      }
    }

    int
    mpitype::get_extent() const
    {
      if (type_ == PRIM || type_ == PAIR || type_ == VEC)
      {
        return extent_;
      }
      else if (type_ == IND)
      {
        return idata_->ub_ - idata_->lb_;
      }
      else
      {
        sst_throw(valueerror, "mpitype::get_lb - unknown type category");
      }
    }

    int
    mpitype::get_true_extent() const
    {
      if (type_ == PRIM || type_ == PAIR || type_ == VEC)
      {
        return extent_;
      }
      else if (type_ == IND)
      {
        return idata_->maxbyte_ - idata_->mindisp_;
      }
      else
      {
        sst_throw(valueerror, "mpitype::get_lb - unknown type category");
      }
    }

    int
    mpitype::bytes_to_elements(size_t bytes) const
    {
      if (size == 0)
        return 0;

      if (type_ == PRIM)
      {
        return bytes / size;
      }
      else if (type_ == PAIR)
      {
        int ret = 0;
        size_t total = 0;
        bool one = true;
        while (total <= bytes)
        {
          if (one)
            total += pdata_->base1_.size;
          else
            total += pdata_->base2_.size;

          ret++;
          one = !one;
        }
        return ret - 1;
      }
      else if (type_ == VEC)
      {
        if (vdata_->blocklen_ == 0)
          return 0;

        int ret = 0;
        size_t total = 0;
        while (total < bytes)
        {
          int before = total;
          for (int i = 0; i < vdata_->blocklen_ && total < bytes; i++)
          {
            ret += vdata_->base_.bytes_to_elements(
                vdata_->base_.get_packed_size());
            total += vdata_->base_.get_packed_size();
            //ret++;

          }
          if (total == before)
            return 0; //stupid check for an empty struct
        }

        return ret;

      }
      else if (type_ == IND)
      {

        int ret = 0;
        size_t total = 0;
        while (total < bytes)
        {
          int before = total;
          for (int i = 0; i < idata_->blocks.size() && total < bytes; i++)
          {
            for (int j = 0; j < idata_->blocks[i].num && total < bytes; j++)
            {
              ret += idata_->blocks[i].base.bytes_to_elements(
                  idata_->blocks[i].base.get_packed_size());
              total += idata_->blocks[i].base.get_packed_size();
              //ret++;
            }
          }
          if (total == before)
            return 0; //stupid check for an empty struct
        }

        return ret;
      }
      sst_throw(valueerror,
        "mpitype::bytes_to_elements: invalid type %d",
        type_);
      return 0;
    }

    void
    mpitype::get_envelope(int* numint, int* numaddr, int* numdtype, int* comb)
    {
      *comb = combiner_;
      *numint = numints_;
      *numaddr = numaddr_;
      *numdtype = numdtype_;
    }

    void
    mpitype::get_contents(int* ints, int* addr, long* dtypes)
    {
      if (type_ == PRIM || type_ == PAIR)
      {
        dtypes[0] = native_id;
      }
      else if (type_ == VEC)
      {
        ints[0] = vdata_->count_;
        ints[1] = vdata_->blocklen_;
        if (combiner_ == MPI_COMBINER_HVECTOR)
          ints[2] = vdata_->stride_;
        else
          ints[2] = vdata_->stride_ / vdata_->base_.get_extent();
        dtypes[0] = vdata_->base_.native_id;
      }
      else if (type_ == IND)
      {
        ints[0] = idata_->blocks.size();

        for (int i = 0; i < ints[0]; i++)
        {
          ints[i + 1] = idata_->blocks[i].num;
          if (combiner_ == MPI_COMBINER_HINDEXED || combiner_
              == MPIX_COMBINER_HINDEXED_BLOCK || combiner_
              == MPI_COMBINER_STRUCT)
            ints[i + 1 + ints[0]] = idata_->blocks[i].disp;
          else
            ints[i + 1 + ints[0]] = idata_->blocks[i].disp
                / idata_->blocks[i].base.get_extent();
          if (numdtype_ > i)
            dtypes[i] = idata_->blocks[i].base.native_id;
        }

      }
    }

    void
    pairdata::serialize_order(sst_serializer* ser)
    {
      base1_.serialize_order(ser);
      base2_.serialize_order(ser);
    }

    void
    vecdata::serialize_order(sst_serializer* ser)
    {
      base_.serialize_order(ser);
      ser->do_primitive(count_);
      ser->do_primitive(blocklen_);
      ser->do_primitive(stride_);
    }

    void
    inddata::serialize_order(sst_serializer* ser)
    {
      ser->do_primitive(mindisp_);
      ser->do_primitive(maxbyte_);
      ser->do_primitive(ub_);
      ser->do_primitive(lb_);

        size_t s = blocks.size();
      ser->do_primitive(s);

      for (int i = 0; i < s; i++)
      {
        blocks[i].base.serialize_order(ser);
        ser->do_primitive(blocks[i].disp);
        ser->do_primitive(blocks[i].num);
      }



    }

    void
    mpitype::serialize_order(sst_serializer* ser)
    {
      ser->do_primitive(id);
      ser->do_primitive(label);
      ser->do_primitive(native_id);
      ser->do_enum(type_);
      ser->do_primitive(alignment_);
      ser->do_primitive(size);
      ser->do_primitive(extent_);

      if (type_ == PAIR)
      {
        ser->do_obj(pdata_);
      }
      else if (type_ == VEC)
      {
        ser->do_obj(vdata_);
      }
      else if (type_ == IND)
      {
        ser->do_obj(idata_);
      }

    }

    mpitype
    mpitype::clone() const
    {
      if (type_ == PRIM)
      {
        mpitype ret = mpitype(label, size, alignment_, combiner_);
        return ret;
      }
      else if (type_ == PAIR)
      {
        mpitype ret = mpitype(label, pdata_->base1_, pdata_->base2_, extent_,
            combiner_);
        return ret;
      }
      else if (type_ == VEC)
      {
        mpitype ret = mpitype(label, vdata_->base_, vdata_->count_,
            vdata_->blocklen_, vdata_->stride_, false, combiner_);
        return ret;
      }
      else if (type_ == IND)
      {
        mpitype ret = mpitype(label, idata_, size, extent_, numints_,
            numdtype_, combiner_);
        ret.deep_copy(*this);

        /*  boost::unordered_map<int, ind_block>::iterator it, end =
         idata_->blocks.end();
         for (it = idata_->blocks.begin(); it != end; it++)
         {
         ret.idata_->blocks[it->first] = it->second;
         }
         ret.size = size;
         ret.extent_ = extent_;*/
        return ret;
      }
      sst_throw(valueerror,
        "mpitype::bytes_to_elements: invalid type %d",
        type_);
      return mpitype();
    }

    int
    mpitype::get_packed_size() const
    {
      /* if (type_ == PRIM)
       {
       return size;
       }
       else if (type_ == PAIR)
       {
       return pdata_->base1_.size + pdata_->base2_.size;
       }
       else if (type_ == VEC)
       {

       int ret = vdata_->base_.get_packed_size() * vdata_->blocklen_
       * vdata_->count_;
       return ret;
       }
       else if (type_ == IND)
       {

       int ret = 0;
       boost::unordered_map<int, ind_block>::iterator it, end =
       idata_->blocks.end();
       for (it = idata_->blocks.begin(); it != end; it++)
       {
       ret += it->second.num * it->second.base.get_packed_size();
       }
       return ret;
       }
       else
       {
       sst_throw(ssterror, "mpitype::get_packed_size() - not a valid type: %s", to_string().c_str());
       }*/
      return size;
    }

    void
    mpitype::pack(const void *inbuf, int incount, void *outbuf, int outsize,
        int *position) const
    {
      char* incast = (char*) inbuf;
      char* outcast = (char*) outbuf;
      int outdex = *position;

      int psize = get_packed_size();
      if (incount * psize > outsize - *position)
      {
        sst_throw(
            valueerror,
            "mpitype::pack - data to be packed (%d bytes) is larger than remainging buffer size (%d)",
            incount * psize, outsize - *position);
      }

      if (type_ == PRIM)
      {
        if (go_debug)
        {
          std::cout << "<-- packing primitive type: incount=" << incount
              << ", outsize=" << outsize << ", pos=" << *position << "\n";
        }
        memcpy(outcast + outdex, inbuf, incount * size);
        *position = *position + (incount * size);
      }
      else if (type_ == PAIR)
      {
        int first = pdata_->base1_.size;
        int firststep = first;
        if (pdata_->base1_.alignment_ > first)
        {
          firststep = pdata_->base1_.alignment_;
        }
        for (int i = 0; i < incount; i++)
        {

          int step = i * extent_;
          memcpy(outcast + outdex + (i * size), incast + step, first);
          memcpy(outcast + outdex + (i * size) + first,
              incast + step + firststep, pdata_->base2_.size);

        }

        *position = *position + (incount * size);
      }
      else if (type_ == VEC)
      {
        if (go_debug)
        {
          std::cout << "<-- packing vector type: incount=" << incount
              << ", outsize=" << outsize << ", pos=" << *position << "\n";
        }
        for (int i = 0; i < incount; i++)
        {

          for (int j = 0; j < vdata_->count_; j++)
          {
            int index = i * extent_ + vdata_->stride_ * j;
            if (go_debug)
            {
              std::cout << "vector: calling pack on " << vdata_->base_.label
                  << " with index " << index << "\n";
              std::cout << "sending " << (long) (incast + index) << " to "
                  << (long) (outcast + outdex) << "\n";
            }
            vdata_->base_.pack(incast + index, vdata_->blocklen_, outcast,
                outsize, &outdex);
          }
        }

        *position = outdex;
      }
      else if (type_ == IND)
      {
        if (go_debug)
        {
          std::cout << "<-- packing indexed type: incount=" << incount
              << ", outsize=" << outsize << ", pos=" << *position << "\n";
        }
        for (int i = 0; i < incount; i++)
        {
          for (int j = 0; j < idata_->blocks.size(); j++)
          {
            int index = i * extent_ + idata_->blocks[j].disp;
            if (go_debug)
            {
              std::cout << "indexed: calling pack on "
                  << idata_->blocks[j].base.label << " with index " << index
                  << "\n";
            }
            idata_->blocks[j].base.pack(incast + index, idata_->blocks[j].num,
                outcast, outsize, &outdex);
          }
        }

        *position = outdex;
      }

    }

    void
    mpitype::unpack(const void *inbuf, int insize, void *outbuf, int outcount,
        int *position) const
    {
      char* incast = (char*) inbuf;
      char* outcast = (char*) outbuf;
      int index = *position;

      if (type_ == PRIM)
      {
        if (go_debug)
        {
          std::cout << "--> unpacking primitive type: outcount=" << outcount
              << ", insize=" << insize << ", pos=" << *position << "\n";
        }
        int psize = get_packed_size();
        int realbytescount = outcount * psize;

        if (realbytescount > insize - *position)
        {
          // sst_throw(
          //     valueerror,
          //     "mpiindexed::unpack - data to be unpacked (%d bytes) is larger than remaining buffer size (%d)",
          //     outcount * psize, insize - *position);
          realbytescount = (insize - *position);
        }

        memcpy(outbuf, incast + index, realbytescount);
        *position = *position + (realbytescount);
      }
      else if (type_ == PAIR)
      {
        if (go_debug)
        {
          std::cout << "--> unpacking pair type: outcount=" << outcount
              << ", insize=" << insize << ", pos=" << *position << "\n";
        }

        int unpackedbytes = 0;
        int first = pdata_->base1_.size;
        int second = pdata_->base2_.size;
        int firststep = first;
        if (pdata_->base1_.alignment_ > first)
        {
          firststep = pdata_->base1_.alignment_;
        }
        for (int i = 0; unpackedbytes < insize - *position && i < outcount; i++)
        {

          memcpy(outcast + i * extent_, incast + index + i * size, first);
          memcpy(outcast + i * extent_ + firststep,
              incast + index + i * size + first, second);
          unpackedbytes += first + second;
        }

        *position = *position + (unpackedbytes);
      }
      else if (type_ == VEC)
      {
        int unpackedbytes = 0;
        if (go_debug)
        {
          std::cout << "--> unpacking vector type: outcount=" << outcount
              << ", insize=" << insize << ", pos=" << *position << "\n";
        }
        for (int i = 0; i < outcount && unpackedbytes < insize - *position; i++)
        {
          for (int j = 0; j < vdata_->count_; j++)
          {
            int outdex = i * extent_ + vdata_->stride_ * j;
            if (go_debug)
            {
              std::cout << "vector: calling pack on " << vdata_->base_.label
                  << " with outdex " << outdex << "\n";
            }
            int before = index;
            vdata_->base_.unpack(incast, insize, outcast + outdex,
                vdata_->blocklen_, &index);
            int unpacked = index - before;
            unpackedbytes += unpacked;
          }
        }
        *position = index;
      }
      else if (type_ == IND)
      {
        int unpackedbytes = 0;
        if (go_debug)
        {
          std::cout << "--> unpacking indexed type: outcount=" << outcount
              << ", insize=" << insize << ", pos=" << *position << "\n";
          std::cout << "this size: " << size << "\n";
        }
        for (int i = 0; i < outcount && unpackedbytes < insize - *position; i++)
        {
          for (int j = 0; j < idata_->blocks.size(); j++)
          {
            int outdex = i * extent_ + idata_->blocks[j].disp;
            if (go_debug)
            {
              std::cout << "indexed: calling unpack on "
                  << idata_->blocks[j].base.label << " with outdex " << outdex
                  << " with disp " << idata_->blocks[j].disp << "\n";
            }
            int before = index;
            idata_->blocks[j].base.unpack(incast, insize, outcast + outdex,
                idata_->blocks[j].num, &index);
            int unpacked = index - before;
            unpackedbytes += unpacked;
          }
        }

        *position = index;
      }

    }

    //
    // Test for (in)equality.
    //
    bool
    operator==(const mpitype &a, const mpitype &b)
    {
      return ((a.get_packed_size() == b.get_packed_size()) && (a.label
          == b.label) && (a.get_type() == b.get_type()));
    }
    bool
    operator!=(const mpitype &a, const mpitype &b)
    {
      return !(a == b);
    }

    std::string
    mpitype::to_string() const
    {
      std::stringstream ss(std::stringstream::in | std::stringstream::out);
      if (type_ == PRIM)
      {
        ss << "mpitype(primitive, label=\"" << label << "\", size=" << size
            << ")";
      }
      if (type_ == PAIR)
      {
        ss << "mpitype(pair, label=\"" << label << "\", size=" << size << ")";
      }
      else if (type_ == VEC)
      {
        ss << "mpitype(vector, label=\"" << label << "\", size=" << size
            << ", base=" << vdata_->base_.to_string() << ", count="
            << vdata_->count_ << ", blocklen=" << vdata_->blocklen_
            << ", stride=" << vdata_->stride_ << ")";
      }
      else if (type_ == IND)
      {
        ss << "mpitype(indexed/struct, label=\"" << label << "\", size="
            << size << ", base=" << idata_->blocks[0].base.to_string()
            << ", blocks=" << idata_->blocks.size() << ")";

      }
      return ss.str();
    }

    //
    // Fairly self-explanatory.
    //
    std::ostream&
    operator<<(std::ostream &os, const mpitype &type)
    {
      os << type.to_string();
      return os;
    }

    //
    // The data types defined by the MPI standard.
    //
    const mpitype mpitype::mpi_null("MPI_NULL", 0, 0, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_char("MPI_CHAR", 1, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_signed_char("MPI_SIGNED_CHAR", 1, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_wchar("MPI_WCHAR", 2, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_unsigned_long_long("MPI_UNSIGNED_LONG_LONG", 8,
        8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_lb("MPI_LB", 0, 0, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_ub("MPI_UB", 0, 0, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_unsigned_char("MPI_UNSIGNED_CHAR", 1, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_byte("MPI_BYTE", 1, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_short("MPI_SHORT", 2, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_unsigned_short("MPI_UNSIGNED_SHORT", 2, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_int("MPI_INT", 4, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_unsigned("MPI_UNSIGNED", 4, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long("MPI_LONG", 8, 8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_unsigned_long("MPI_UNSIGNED_LONG", 8, 8,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_float("MPI_FLOAT", 4, sizeof(float),
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_double("MPI_DOUBLE", 8, 8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long_double("MPI_LONG_DOUBLE", 16, 16,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long_long_int("MPI_LONG_LONG_INT", 8, 8,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long_long("MPI_LONG_LONG", 8, 8,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_packed("MPI_PACKED", 1, 0, MPI_COMBINER_NAMED);

    struct fi
    {
      float a;
      int b;
    };

    struct di
    {
      double a;
      int b;
    };
    struct ii
    {
      int a;
      int b;
    };
    struct si
    {
      short a;
      int b;
    };
    struct li
    {
      long a;
      int b;
    };
    struct ldi
    {
      long double a;
      int b;
    };

    //pair types
    const mpitype mpitype::mpi_float_int("MPI_FLOAT_INT", mpitype::mpi_float,
        mpitype::mpi_int, sizeof(fi), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_double_int("MPI_DOUBLE_INT",
        mpitype::mpi_double, mpitype::mpi_int, sizeof(di), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long_int("MPI_LONG_INT", mpitype::mpi_long,
        mpitype::mpi_int, sizeof(li), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_short_int("MPI_SHORT_INT", mpitype::mpi_short,
        mpitype::mpi_int, sizeof(si), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_2int("MPI_2INT", mpitype::mpi_int,
        mpitype::mpi_int, sizeof(ii), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_long_double_int("MPI_LONG_DOUBLE_INT",
        mpitype::mpi_long_double, mpitype::mpi_int, sizeof(ldi),
        MPI_COMBINER_NAMED);

    //fortran nonsense
    const mpitype mpitype::mpi_complex("MPI_COMPLEX", 8, 8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_double_complex("MPI_DOUBLE_COMPLEX", 16, 16,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_logical("MPI_LOGICAL", 1, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_real("MPI_REAL", 4, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_double_precision("MPI_DOUBLE_PRECISION", 8, 8,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_integer("MPI_INTEGER", 4, 4, MPI_COMBINER_NAMED);

    const mpitype mpitype::mpi_integer1("MPI_INTEGER1", 1, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_integer2("MPI_INTEGER2", 2, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_integer4("MPI_INTEGER4", 4, 4,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_integer8("MPI_INTEGER8", 8, 8,
        MPI_COMBINER_NAMED);

    const mpitype mpitype::mpi_real4("MPI_REAL4", 4, 4, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_real8("MPI_REAL8", 8, 8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_real16("MPI_REAL16", 16, 16, MPI_COMBINER_NAMED);

    const mpitype mpitype::mpi_complex8("MPI_COMPLEX8", 8, 8,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_complex16("MPI_COMPLEX16", 16, 16,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_complex32("MPI_COMPLEX32", 32, 32,
        MPI_COMBINER_NAMED);

    //fortran pairs
    const mpitype mpitype::mpi_2integer("MPI_2INTEGER", mpitype::mpi_int,
        mpitype::mpi_int, sizeof(ii), MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_2complex("MPI_2COMPLEX", mpitype::mpi_complex,
        mpitype::mpi_complex, 16, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_2double_complex("MPI_2DOUBLE_COMPLEX",
        mpitype::mpi_double_complex, mpitype::mpi_double_complex, 16,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_2real("MPI_2REAL", mpitype::mpi_real,
        mpitype::mpi_real, 8, MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_2double_precision("MPI_2DOUBLE_PRECISION",
        mpitype::mpi_double_precision, mpitype::mpi_double_precision, 16,
        MPI_COMBINER_NAMED);
    const mpitype mpitype::mpi_character("MPI_CHARACTER", 1, 4,
        MPI_COMBINER_NAMED);

  }
} // end of namespace sstmac
