/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIMESSAGE_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPIMESSAGE_H_INCLUDED

#include <sstmac/hardware/network/network_message.h>
#include <sstmac/common/messages/messageid.h>
#include <sstmac/common/messages/payload.h>
#include <sstmac/software/api/mpi/mpi_comm/mpicommid.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpi_types/mpiid.h>
#include <sstmac/software/api/mpi/mpi_types/mpitag.h>
#include <sstmac/software/api/mpi/mpi_types/mpitype.h>
#include <sstmac/software/process/taskid.h>
#include <sstmac/software/process/appid.h>
#include <sstmac/common/nocopy.h>
#include <sstmac/common/ptr_type.h>
#include <sstmac/hardware/network/network_message.h>

namespace sstmac
{
  namespace sw
  {

    class operatingsystem;
    class mpiprotocol;
    typedef boost::intrusive_ptr<mpiprotocol> mpiprotocolPtr;

    // Important types.
    /**
     * A specialization of networkdata that contains envelope information
     * relevant to MPI messaging.
     */
    class mpimessage : public library_message,
        public timed_message,
        public hw::network_message,
        public serializable_type<mpimessage>
    {
    ImplementSerializable(mpimessage)
    public:
      typedef boost::intrusive_ptr<mpimessage> ptr;

      /// The category label used to differentiate user data (which can be
      /// matched using wildcards or [later] probed using probe or iprobe)
      /// from internal communications (e.g. collective operations).
      /// The send_ready and recv_ready categories are intended for setting
      /// up rendezvous operations.  The recv_ack category is indented for
      /// completing synchronous send operations.
      typedef enum
      {
        user = 0, collective, onesided
      } category;

      typedef enum
      {
        header = 0, data, rendezvous_ack, completion_ack, fake
      } content_type_t;

    protected:
      int64_t envelope_;
      int64_t mintrans_;
      int count_;
      mpitype type_;
      mpiid source_;
      mpiid dest_;
      mpitag tag_;
      mpicommid commid_;
      int seqnum_;
      messageid msgid_;
      category cat_;
      content_type_t content_type_;
      bool ssend_;
      // category cat_;
      taskid source_task_;
      taskid dest_task_;
      appid aid_;
      payload::const_ptr content_;
      int protocol_;
      bool ignore_seqnum_;
      long bytes_;
      timestamp delay_;

      void
      recompute_bytes();

    protected:
      /// Hello.
      mpimessage(const std::string& libn, const timestamp& time,
          int64_t envelope, int64_t mintrans, int count, const mpitype &type,
          const mpiid &source, const mpiid &dest, const mpitag &tag,
          const mpicommid &commid, int seqnum, const messageid &msgid,
          bool ssend, category cat, const taskid &source_task,
          const taskid &dest_task, const appid &aid,
          content_type_t content_type, const payload::const_ptr& content,
          const mpiprotocolPtr& protocol, MESSAGE_TYPES ty);

      mpimessage(content_type_t content_type);

          mpimessage(const payload::const_ptr& content,
              content_type_t content_type);

      mpimessage();

    public:
      /// Construct a new message.
      static mpimessage::ptr
      construct(const std::string& libn, const timestamp& time,
          int64_t envelope, int64_t mintrans, int count, const mpitype &type,
          const mpiid &source, const mpiid &dest, const mpitag &tag,
          const mpicommid &commid, int seqnum, const messageid &msgid,
          bool ssend, category cat, const taskid &source_task,
          const taskid &dest_task, const appid &aid,
          content_type_t content_type, const payload::const_ptr& content,
          const mpiprotocolPtr& protocol, MESSAGE_TYPES ty = MPI);

      static const char*
      str(content_type_t content_type);

      static const char*
      str(category cat);

      static mpimessage::ptr
      construct(content_type_t content_type)
      {
        mpimessage::ptr ret(new mpimessage(content_type));
        return ret;
      }

      //used for deserialization
      static mpimessage::ptr
      construct()
      {
        return ptr(new mpimessage());
      }

      static mpimessage::ptr
      construct(const payload::const_ptr& content, content_type_t content_type)
      {
        mpimessage::ptr ret(new mpimessage(content, content_type));
        return ret;
      }

      /// Goodbye.
      virtual
      ~mpimessage() throw ();

      virtual sst_message::ptr
      clone() const;

      virtual sst_message::ptr
      clone(MESSAGE_TYPES ty) const;

      virtual mpimessage::ptr
      deep_clone() const;

      void
      add_delay(const timestamp& delay)
      {
        delay_ += delay;
      }

      timestamp
      get_delay() const
      {
        return delay_;
      }

      /**
       * Serialize this message during parallel simulation.
       * @param ser The serializer to use
       */
      virtual void
      serialize_order(sst_serializer* ser);

      /// Parent method to get the complemetary part of this message.
      ////virtual boost::shared_ptr<networkdata> complement() const;

      virtual std::string
      to_string() const;

      long
      payload_bytes() const;

      virtual long
      get_byte_length() const
      {
        return bytes_;
      }

      /// The source task.  Required by the parent type.
      virtual const taskid&
      source_task() const
      {
        return source_task_;
      }

      /// The destination task.  Required by the parent type.
      virtual const taskid&
      dest_task() const
      {
        return dest_task_;
      }

      virtual long
      source_thread(const boost::intrusive_ptr<operatingsystem>& os) const;

      virtual long
      dest_thread(const boost::intrusive_ptr<operatingsystem>& os) const;

      mpiprotocolPtr
      get_protocol() const;

      void
      set_protocol(const mpiprotocolPtr& protocol);

      void
      header_to_rendezvous_ack();

      void
      header_to_completion_ack();

      void
      payload_to_completion_ack();

      void
      rendezvous_ack_to_payload();

      void
      reverse();

      /// The number of elements sent.
      int
      count() const
      {
        return count_;
      }

      /// Access the type label associted with this mesage.
      const mpitype&
      type() const
      {
        return type_;
      }

      /// Access the index of the sender.
      const mpiid&
      source() const
      {
        return source_;
      }

      /// Access the index of the target.
      const mpiid&
      dest() const
      {
        return dest_;
      }

      const appid&
      app() const
      {
        return aid_;
      }

      /// Access the tag associated with this message.
      const mpitag&
      tag() const
      {
        return tag_;
      }

      /// Access the id of the communicator that owns this message.
      const mpicommid&
      commid() const
      {
        return commid_;
      }

      /// Access the sequence number for this message.
      int
      seqnum() const
      {
        return seqnum_;
      }

      /// A (hopfully unique) message id.
      const messageid&
      id() const
      {
        return msgid_;
      }

      bool
      intranode() const
      {
        return fromaddr_->equals(toaddr_);
      }

      /// Return true if the receiver needs to acknowledge receipt.
      bool
      ssend() const
      {
        return ssend_;
      }

      /// Access the category label for this message.
      category
      cat() const
      {
        return cat_;
      }

      void
      cat(category c)
      {
        cat_ = c;
      }

      content_type_t
      content_type()
      {
        return content_type_;
      }

      void
      content_type(content_type_t ty)
      {
        content_type_ = ty;
      }

      /// Access the payload associated with this message.
      payload::const_ptr
      content() const
      {
        return content_;
      }

      void
      set_content(payload::const_ptr p)
      {
        content_ = p;
      }

      /// Create a new status object for this message.
      virtual mpistatus_t
      build_status() const;

      void
      mutate(mutation_t mut);

      void
      set_ignore_seqnum(bool flag);

      bool
      ignore_seqnum() const;

    };

  }
} // end of namespace sstmac.

#endif
