/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2011 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_MPITRACER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_MPITRACER_H_INCLUDED


#include <sstmac/software/api/mpi/mpirequest.h>
#include <sstmac/software/api/mpi/mpistatus.h>
#include <sstmac/software/api/mpi/mpi_types/mpiop.h>

#include <sstmac/common/logger.h>


namespace sstmac { namespace sw {

  /**
   * Base class to provide tracing for mpi calls.
   */
  class mpitracer : public logger {
  protected:
    mpitracer(std::string logname) : logger(logname, NULL) {}

  public:

    virtual std::string to_string() const{
            return "mpitracer";
          }

    typedef boost::intrusive_ptr<mpitracer> ptr;

    //typedef mpirequest mpirequest;
   // typedef mpirequest::ptr mpirequest_t;
   // typedef const mpistatus::ptr const_mpistatus_t;

    virtual ~mpitracer() {}

    /// Clone this tracer.
    virtual mpitracer::ptr clone() const = 0;

    /// Finish tracing.  Marks the tracer as inactive.
    virtual void finish() = 0;

    /// Get the rank of a node.
    virtual void comm_rank(const timestamp &start, const timestamp &finish,
                           const mpicomm::ptr &comm, int rank) = 0;

    /// The size of a communicator.
    virtual void comm_size(const timestamp &start, const timestamp &finish,
                           const mpicomm::ptr &comm, int size) = 0;

    /// Get the number of elements received assuming a specific type.
    virtual void get_count(const timestamp &start, const timestamp &finish,
                           const_mpistatus_t status,
                           const mpitype &type, int count) = 0;

    /// Split a communicator.
    virtual void comm_split(const timestamp &start, const timestamp &finish,
                            const mpicomm::ptr &incomm, int color, int key,
                            const mpicomm::ptr &outcomm) = 0;

    /// Duplicate a type.
    virtual void type_dup(const timestamp &start, const timestamp &finish,
                          const mpitype &intype, const mpitype &outtype) = 0;

    /// Commit a type.
    virtual void type_commit(const timestamp &start, const timestamp &finish,
                             const mpitype& type) = 0;

    /// Create a new type.
    virtual void type_contiguous(const timestamp &start, const timestamp &finish,
                                 int count, const mpitype &oldtype,
                                 const mpitype &newtype) = 0;

    /// Free a type.
    virtual void type_free(const timestamp &start, const timestamp &finish,
                           const mpitype &type) = 0;

    /// Get type size.
    virtual void type_size(const timestamp &start, const timestamp &finish,
                           const mpitype &type, int size) = 0;

    /// Test whether MPI has been initialized.
    virtual void
    initialized(const timestamp &start, const timestamp &finish, bool ini) = 0;

    /// Test whether MPI has been finalized.
    virtual void
    finalized(const timestamp &start, const timestamp &finish, bool fini) = 0;

    /// Initialize MPI.  This call will also create the actual trace file.
    virtual void init(const timestamp &start, const timestamp &finish,
                      const mpicomm::ptr &world, const mpicomm::ptr &self) = 0;

    /// Finalize MPI.
    virtual void finalize(const timestamp &start, const timestamp &finish) = 0;

    /// Duplicate a communicator.
    virtual void comm_dup(const timestamp &start, const timestamp &finish,
                          const mpicomm::ptr &input, const mpicomm::ptr &output) = 0;

    /// Create a communicator containing a subset of an existing comm.
    virtual void
    comm_create(const timestamp &start, const timestamp &finish,
                const mpicomm::ptr &input, const std::vector<mpiid> &members,
                const mpicomm::ptr &output) = 0;

    /// Destroy a communicator.
    virtual void comm_free(const timestamp &start, const timestamp &finish,
                           const mpicomm::ptr &input) = 0;

    /// Blocking send.
    virtual void send(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type, const mpiid &target,
                      const mpitag &tag, const mpicomm::ptr &comm) = 0;

    /// Blocking send using a user-defined buffer.
    virtual void bsend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm::ptr &comm) = 0;

    /// Blocking syncronous send.
    virtual void ssend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm::ptr &comm) = 0;

    /// Ready send.
    virtual void rsend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm::ptr &comm) = 0;

    /// Nonblocking send.
    virtual void isend(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &target,
                       const mpitag &tag, const mpicomm::ptr &comm,
                       const mpirequest_t &req) = 0;

    /// Nonblocking user-buffered send.
    virtual void ibsend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm::ptr &comm,
                        const mpirequest_t &req) = 0;

    /// Nonblocking synchronous send.
    virtual void issend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm::ptr &comm,
                        const mpirequest_t &req) = 0;

    /// Nonblocking ready send.
    virtual void irsend(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiid &target,
                        const mpitag &tag, const mpicomm::ptr &comm,
                        const mpirequest_t &req) = 0;

    /// Blocking receive.
    virtual void recv(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type, const mpiid &source,
                      const mpitag &tag, const mpicomm::ptr &comm,
                      const const_mpistatus_t &stat) = 0;

    /// Asynchronous receive.
    virtual void irecv(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type, const mpiid &source,
                       const mpitag &tag, const mpicomm::ptr &comm,
                       const mpirequest_t &req) = 0;

    /// Send/receieve exchange.
    virtual void
    sendrecv(const timestamp &start, const timestamp &finish,
             int sendcount, const mpitype &sendtype,
             const mpiid &target, const mpitag &sendtag,
             int recvcount, const mpitype &recvtype,
             const mpiid &source, const mpitag &recvtag,
             const mpicomm::ptr &comm, const const_mpistatus_t &stat) = 0;

    /// Send and receive using a single buffer.
    virtual void
    sendrecv_replace(const timestamp &start, const timestamp &finish,
                     int count, const mpitype &type,
                     const mpiid &dest, const mpitag &sendtag,
                     const mpiid &source, const mpitag &recvtag, 
                     const mpicomm::ptr &comm, const const_mpistatus_t &stat) = 0;

    /// Initiate a send.
    virtual void send_init(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiid &target,
                           const mpitag &tag, const mpicomm::ptr &comm,
                           const mpirequest_t &req) = 0;

    /// Initiate a buffered send.
    virtual void bsend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm::ptr &comm,
                            const mpirequest_t &req) = 0;

    /// Initiate a synchronous send.
    virtual void ssend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm::ptr &comm,
                            const mpirequest_t &req) = 0;

    /// Initiate a ready send.
    virtual void rsend_init(const timestamp &start, const timestamp &finish,
                            int count, const mpitype &type, const mpiid &target,
                            const mpitag &tag, const mpicomm::ptr &comm,
                            const mpirequest_t &req) = 0;
    
    /// Initiate a receive.
    virtual void recv_init(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type, const mpiid &target,
                           const mpitag &tag, const mpicomm::ptr &comm,
                           const mpirequest_t &req) = 0;

    /// Start a previously initiated request.
    virtual void start(const timestamp &start, const timestamp &finish,
                       const mpirequest_t &req) = 0;

    /// Start a collection of previously initiated requests.
    virtual void startall(const timestamp &start, const timestamp &finish,
                          const std::vector< mpirequest_t > &requests) = 0;

    /// Cancel a request that had been previously initiated.
    virtual void cancel(const timestamp &start, const timestamp &finish,
                        const mpirequest_t &req) = 0;

    /// Test whether a request has been cancelled.
    virtual void test_cancelled(const timestamp &start, const timestamp &finish,
                                const const_mpistatus_t &req,
                                bool cancelled) = 0;

    /// Wait for a request to finish.
    virtual void wait(const timestamp &start, const timestamp &finish,
                      const mpirequest_t &req, const const_mpistatus_t &stat)=0;

    /// Wait for all the given requests to finish.
    virtual void waitall(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req,
                         const std::vector< const_mpistatus_t > &statuses) = 0;

    /// Wait for at least one of the given requests to complete.
    virtual void waitany(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req,
                         int index, const const_mpistatus_t &stat) = 0;

    /// Wait for some of the given requests to complete.
    virtual void waitsome(const timestamp &start, const timestamp &finish,
                          const std::vector<mpirequest_t > &req,
                          const std::vector<int> &indices,
                          const std::vector< const_mpistatus_t > &statuses) = 0;

    /// Test whether a request has finished.
    virtual void test(const timestamp &start, const timestamp &finish,
                      const mpirequest_t &req, bool flag, 
                      const const_mpistatus_t &stat) = 0;

    /// Test whether all the given requests have finished.
    virtual void testall(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req, bool flag,
                         const std::vector< const_mpistatus_t > &statuses) = 0;
    
    /// Test whether at least one of the given requests has completed.
    virtual void testany(const timestamp &start, const timestamp &finish,
                         const std::vector< mpirequest_t > &req, int index,
                         bool flag, const const_mpistatus_t &stat) = 0;

    /// Test whether some of the given requests have completed.
    virtual void testsome(const timestamp &start, const timestamp &finish,
                          const std::vector<mpirequest_t > &req,
                          const std::vector<int> &indices,
                          const std::vector< const_mpistatus_t > &statuses) = 0;

    /// Blocking probe for a message that matches the given signature.
    virtual void probe(const timestamp &start, const timestamp &finish,
                       const mpiid &source, const mpitag &tag,
                       const mpicomm::ptr &comm, const const_mpistatus_t &stat) = 0;

    /// Non-blocking probe for messages.
    virtual void iprobe(const timestamp &start, const timestamp &finish,
                        const mpiid &source, const mpitag &tag,
                        const mpicomm::ptr &comm, bool &flag,
                        const const_mpistatus_t &stat) = 0;

    /// Attach a user-defined buffer for *bsend operations.
    virtual void buffer_attach(const timestamp &start, const timestamp &finish,
                               int bytes) = 0;

    /// Detach the user-defined buffer.
    virtual void buffer_detach(const timestamp &start, const timestamp &finish,
                               int bytes) = 0;

    /// Set a barrier.
    virtual void barrier(const timestamp &start, const timestamp &finish,
                         const mpicomm::ptr &comm) = 0;

    /// Broadcast data with no payload.
    virtual void bcast(const timestamp &start, const timestamp &finish,
                       int count, const mpitype &type,
                       const mpiid &root, const mpicomm::ptr &comm) = 0;

    /// Scatter data uniformly.
    virtual void scatter(const timestamp &start, const timestamp &finish,
                         int sendcount, const mpitype &sendtype,
                         int recvcount, const mpitype &recvtype,
                         const mpiid &root, const mpicomm::ptr &comm) = 0;

    /// Scatter data non-uniformly.
    virtual void scatterv(const timestamp &start, const timestamp &finish,
                          const std::vector<int> &sendcounts,
                          const mpitype &sendtype,
                          int recvcount, const mpitype &recvtype,
                          const mpiid &root, const mpicomm::ptr &comm) = 0;
    
    /// Gather uniformly distributed data.
    virtual void gather(const timestamp &start, const timestamp &finish,
                        int sendcount, const mpitype &sendtype,
                        int recvcount, const mpitype &recvtype,
                        const mpiid &root, const mpicomm::ptr &comm) = 0;

    /// Gather non-uniformly distributed data.
    virtual void gatherv(const timestamp &start, const timestamp &finish,
                         int sendcount, const mpitype &sendtype,
                         const std::vector<int> &recvcnts,
                         const mpitype &recvtype,
                         const mpiid &root, const mpicomm::ptr &comm) = 0;

    /// Gather uniformly distributed data and make it available to all nodes.
    virtual void allgather(const timestamp &start, const timestamp &finish,
                           int sendcount, const mpitype &sendtype,
                           int recvcount, const mpitype &recvtype,
                           const mpicomm::ptr &comm) = 0;

    /// Gather non-uniformly distributed data onto all nodes.
    virtual void allgatherv(const timestamp &start, const timestamp &finish,
                            int sendcount, const mpitype &sendtype,
                            const std::vector<int> &recvcnts,
                            const mpitype &recvtype, const mpicomm::ptr &comm) = 0;

    /// Send uniformly distributed data from all to all processes.
    virtual void alltoall(const timestamp &start, const timestamp &finish,
                          int sendcount, const mpitype &sendtype,
                          int recvcount, const mpitype &recvtype,
                          const mpicomm::ptr &comm) = 0;

    /// Send non-uniformly distributed data from all to all processors.
    virtual void
    alltoallv(const timestamp &start, const timestamp &finish,
              const std::vector<int> &sendcnts, const mpitype &sendtype,
              const std::vector<int> &recvcnts, const mpitype &recvtype,
              const mpicomm::ptr &comm) = 0;

    /// Reduce data from all nodes onto the root node.
    virtual void reduce(const timestamp &start, const timestamp &finish,
                        int count, const mpitype &type, const mpiop &op,
                        const mpiid &root, const mpicomm::ptr &comm) = 0;

    /// Reduce data from all nodes to all nodes.
    virtual void allreduce(const timestamp &start, const timestamp &finish,
                           int count, const mpitype &type,
                           const mpiop &op, const mpicomm::ptr &comm) = 0;

    /// Reduce data and scatter results.
    virtual void reduce_scatter(const timestamp &start, const timestamp &finish,
                                const std::vector<int> &recvcnts,
                                const mpitype &type, const mpiop &op,
                                const mpicomm::ptr &comm) = 0;

    /// Compute partial reductions on data.
    virtual void scan(const timestamp &start, const timestamp &finish,
                      int count, const mpitype &type,
                      const mpiop &op, const mpicomm::ptr &comm) = 0;
  };

} } // end of namespace sstmac

#endif
