/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_MPI_RMA_MPI_WINDOW_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_MPI_RMA_MPI_WINDOW_H_INCLUDED

#include <iosfwd>
#include <sstmac/common/types.h>

#include <sstmac/software/api/mpi/mpi_types/mpiinfo.h>

#include <sstmac/common/messages/wrapperpayload.h>
#include <sstmac/software/api/mpi/rma/mpi_rma_message.h>

#include <stdio.h>
#include <string.h>

namespace sstmac
{
  namespace sw
  {
    class mpi_window : public ptr_type
    {

    protected:

      void* base_;
      size_t size_;
      int disp_;
      mpiinfo::ptr inf_;
      mpicomm::ptr comm_;
      long winid_;

      long epoch_;

      mpi_window(void* base, size_t size, int disp, mpiinfo::ptr inf,
          mpicomm::ptr comm, long id) :
        base_(base), size_(size), disp_(disp), inf_(inf), comm_(comm),
            winid_(id)
      {
        epoch_ = 0;
      }

    public:

      typedef boost::intrusive_ptr<mpi_window> ptr;

      static ptr
      construct(void* base, size_t size, int disp, mpiinfo::ptr inf,
          mpicomm::ptr comm, long id)
      {
        return ptr(new mpi_window(base, size, disp, inf, comm, id));
      }

      virtual
      ~mpi_window()
      {
      }

      virtual std::string
      to_string() const
      {
        return "mpi_window";
      }

      bool
      contains_addr(void* addr)
      {
        size_t a1 = (size_t) addr;
        size_t a2 = (size_t) base_;
        size_t upper = (size_t) ((char*) base_ + size_);
        return a1 >= a2 && a1 < upper;
      }

      mpicomm::ptr
      comm()
      {
        return comm_;
      }

      long
      winid() const
      {
        return winid_;
      }

      long
      get_current_epoch() const
      {
        return epoch_;
      }

      void
      next_epoch()
      {
        epoch_++;
      }

      payload::const_ptr
      get_data(const mpi_rma_message::op_info &inf) const
      {
        char* bcast = (char*) base_;
        bcast += (disp_ * inf.disp_);

        return wrapperpayload::construct(bcast, inf.count_);

      }

      void
      put_data(const mpi_rma_message::op_info &inf, payload::const_ptr load)
      {
        char* bcast = (char*) base_;
        bcast += (disp_ * inf.disp_);

        wrapperpayload::const_ptr cast = boost::dynamic_pointer_cast<const wrapperpayload>(load);

        if(!cast){
          sst_throw(ssterror, "mpi_window::put_data - didn't cast to wrapperpayload");
        }

        char* d = (char*)cast->data();
        memcpy(bcast, d, inf.count_);


      }
    };
  }
}
#endif
