/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "shmem/api.h"
#include "shmem/server.h"

#include "shmem/memory/symmetric.h"
#include "shmem/messages/put_transfer.h"
#include "shmem/messages/inc_transfer.h"
#include "shmem/messages/get_transfer.h"

#include <sstmac/common/messages/valuepayload.h>

// ImplementFactory(sstmac::sw::api);
ImplementAPI(sstmac::sw::shmem, api, "shmem")


namespace sstmac
{
  namespace sw
  {

    namespace shmem
    {
      SSTRegister("shmem", sstmac::sw::api, api, "Create bindings for SHMEM runtime");

      debug api::dbg_;
      boost::unordered_map<int, shared_heap_info::ptr> api::heap_info_;

      api::api() :
        parent_api()
      {
        if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> shmem | api", logger::yellow);
          logger::register_logger(&dbg_);
        }

        global_key_ = key::construct();
        waiting_on_quiet_ = false;
        get_arrived_ = payload::ptr();
      }

      void
      api::init(const softwareid& sid)
      {
        parent_api::init(sid);
        id_ = sid;
        libname_ = "shmemapi" + id_.to_string();
        //parent_api::init(sid);
      }

      api::~api()
      {
      }

      void
      api::init_os(const OS_API::ptr& os)
      {
        parent_api::init_os(os);

        // lib_mem_ = lib_compute_memmove::construct(libname_ + "-lib-mem");
        // lib_mem_->init_callback();
        // register_lib(lib_mem_);
      }

      // ----- helpers -----
      symmetric::ptr
      api::get_symmetric_by_addr(long addr)
      {
        return heap_->get_symmetric_by_addr(addr);
      }

      void
      api::put_done(long tid)
      {
        if (pending_puts_.find(tid) == pending_puts_.end())
        {
          sst_throw(ssterror, "shemm::api::put_done - couldn't find put %l", tid);
        }

        pending_puts_.erase(tid);
        if (waiting_on_quiet_ && pending_puts_.size() == 0)
        {
          os_->unblock(global_key_);
        }

      }

      void
      api::unblock()
      {
        os_->unblock(global_key_);
      }

      void
      api::send_put_complete(transfer::const_ptr t) const
      {
        payload::const_ptr load = valuepayload<long>::construct(t->get_tid());
        server_->send_tran_response(t, load, message::PUT_COMPLETE);
      }

      void
      api::get_arrived(payload::const_ptr pay)
      {
        get_arrived_ = pay;
        // if (!get_arrived_)
        // {
        //   sst_throw(ssterror, "shmem::api::get_arrived with null payload ");
        // }
        os_->unblock(global_key_);
      }

      void
      api::send_get_response(transfer::const_ptr t, payload::const_ptr pay)
      {
        server_->send_tran_response(t, pay, message::GET_RESPONSE);
      }

      void
      api::start_trans(transfer::ptr tran, bool block, bool unblock_on_mem)
      {

        server_->start_transfer(tran,
            (unblock_on_mem) ? global_key_ : key::ptr());

        if (block)
          os_->block(global_key_);

      }

      symmetric::ptr
      api::get_fake_sym(size_t s)
      {
        if (!current_fake_)
        {
          symmetric::ptr ret = heap_->shmalloc(s, false);
          current_fake_ = ret;
          return ret;
        }
        else
        {
          return current_fake_;
        }
      }

      //-------

      void
      api::start_pes(int npes)
      {
        SSTMACBacktrace("shmem_start_pes");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::start_pes() - start \n";
        std::string server_libname = sstprintf("shmem::server-%d", id_.app_.id);
        library::ptr shmemlib = os_->get_lib(server_libname);

        if (shmemlib == library::ptr()) // only do one mpiserver per app per node
        {
          server::ptr serv = server::construct(id_.app_.id);
          shmemlib = serv;
          os_->register_ser(serv);
          serv->start();
        }

        server_ = boost::dynamic_pointer_cast<server>(shmemlib);
        server_->add_pe(ptr(this));
        server_->set_eventmanager(os_->get_eventmanager());

        heap_ = heap::construct();
        heap_info_[id_.task_.id] = shared_heap_info::construct(heap_);

        //check for globals
        init_globals<long> ();
        init_globals<int> ();
        init_globals<double> ();
        init_globals<char> ();
        init_globals<bool> ();
        init_globals<long long> ();
        init_globals<long double> ();
        init_globals<short> ();
        init_globals<float> ();

        init_globals<long*> ();
        init_globals<int*> ();
        init_globals<double*> ();
        init_globals<char*> ();
        init_globals<bool*> ();
        init_globals<long long*> ();
        init_globals<long double*> ();
        init_globals<short*> ();
        init_globals<float*> ();

        SSTMAC_DEBUG << "shmem::api::start_pes() - done \n";
        end_api_call();
      }

      int
      api::_num_pes()
      {
        SSTMAC_DEBUG << "shmem::api::num_pes() \n";
        environment::const_ptr env = os_->get_env(id_.app_);
        return env->get_app_count();
      }

      int
      api::_my_pe()
      {
        SSTMAC_DEBUG << "shmem::api::my_pe() \n";
        return id_.task_.id;
      }

      void*
      api::shmalloc(size_t size)
      {
        SSTMACBacktrace("shmem_shmalloc");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmalloc \n";
        symmetric::ptr mem = heap_->shmalloc(size);

        shmem_barrier(0, 0, _num_pes());

        void* ret = (void*) mem->get_raw();
        end_api_call();
        return ret;
      }

      void
      api::shfree(void* ptr)
      {
        SSTMACBacktrace("shmem_shfree");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shfree \n";

        heap_->shfree(ptr);
        end_api_call();
      }

      int
      api::shmem_addr_accessible(void* addr, int pe)
      {
        SSTMACBacktrace("shmem_addr_accessible");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_addr_accessible \n";
        int ret;
        if (pe == _my_pe())
        {
          long a = heap_->find_addr(addr);

          ret = (a >= 0) ? 1 : 0;
        }
        else
        {
          long shmaddr = heap_->find_addr(addr);
          if (shmaddr >= 0)
            ret = heap_info_[pe]->addr_accessible(shmaddr);
          else
            ret = 0;
        }
        end_api_call();
        return ret;
      }

      bool
      api::pe_in_set(int pe, int PE_start, int logPE_stride, int PE_size){

          int diff = pe - PE_start;
          double std = (double)pow(2.0, (double)logPE_stride);
          double divd = diff / std;
          int divi = diff / (int)pow(2, logPE_stride);
          return divi == divd && divi < PE_size;

      }
      /**
       * This function was written with false-sharing,
       * it currently does not actually model the correct
       * communication necessary for this call.
       */
      void
      api::shmem_barrier(int PE_start, int logPE_stride, int PE_size)
      {
        SSTMACBacktrace("shmem_barrier");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_barrier \n";
        static int count = 0;
        static boost::unordered_map<key::ptr, server::ptr> servmap;

        if (pe_in_set(_my_pe(), PE_start, logPE_stride, PE_size))
        {

          shmem_quiet();

          count++;

          SSTMAC_DEBUG << "api: pe " << id_.task_.id
                << " entering shmem_barrier(), count is " << count << "\n";

          if (count == PE_size)
          {
            count = 0;

            SSTMAC_DEBUG << "api: shmem_barrier(), unblocking everyone \n";

            boost::unordered_map<key::ptr, server::ptr>::iterator it, end =
                servmap.end();
            for (it = servmap.begin(); it != end; it++)
            {
              server_->unblock_other_server(it->second, it->first);
            }

            servmap.clear();

          }
          else
          {
            SSTMAC_DEBUG << "api: shmem_barrier(), blocking \n";

            servmap[global_key_] = server_;
            os_->block(global_key_);

            SSTMAC_DEBUG << "api: shmem_barrier(), pe " << id_.task_.id
                  << " unblocked \n";
          }
        }
        end_api_call();
      }

      void
      api::shmem_quiet()
      {
        SSTMACBacktrace("shmem_quiet");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_quiet \n";

        if (pending_puts_.size() > 0)
        {
          waiting_on_quiet_ = true;
          os_->block(global_key_);
          waiting_on_quiet_ = false;
        }

        end_api_call();
      }

      void
      api::shmem_int_inc(int *target, int pe)
      {
        SSTMACBacktrace("shmem_int_inc");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_int_inc \n";
        long a;
        int o;
        heap_->get_addr_off<int> (target, &a, &o);
        inc_transfer<int>::ptr tran = inc_transfer<int>::construct(a, o,
            id_.task_.id, pe);
        server_->start_transfer(tran, global_key_);
        end_api_call();
      }
      void
      api::shmem_long_inc(long *target, int pe)
      {
        SSTMACBacktrace("shmem_long_inc");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_long_inc \n";
        long a;
        int o;
        heap_->get_addr_off<long> (target, &a, &o);
        inc_transfer<long>::ptr tran = inc_transfer<long>::construct(a, o,
            id_.task_.id, pe);
        server_->start_transfer(tran, global_key_);
        end_api_call();
      }
      void
      api::shmem_longlong_inc(long long *target, int pe)
      {
        SSTMACBacktrace("shmem_longlong_inc");
        start_api_call();
        SSTMAC_DEBUG << "shmem::api::shmem_longlong_inc \n";
        long a;
        int o;
        heap_->get_addr_off<long long> (target, &a, &o);
        inc_transfer<long long>::ptr tran = inc_transfer<long long>::construct(
            a, o, id_.task_.id, pe);
        server_->start_transfer(tran, global_key_);
        end_api_call();
      }
    }
  }
}
