/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SHMEM_api_H_
#define SSTMAC_SHMEM_api_H_

#include <sstmac/common/ptr_type.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/process/softwareid.h>
#include <sstmac/software/libraries/library.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/common/messages/vectorpayload.h>

#include "shmem/memory/heap.h"
#include "shmem/memory/shared_heap_info.h"
#include "shmem/memory/global.h"
#include "shmem/messages/transfer.h"
#include "shmem/transfer_api.h"

#include "shmem/messages/get_transfer.h"
#include "shmem/messages/put_transfer.h"
#include "shmem/messages/op_transfer.h"

#include "shmem/messages/fake_data.h"

#include <sstmac/software/api/api.h>

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {
      class server;

      class api : public ::sstmac::sw::api, public virtual transfer_api
      {
      protected:
        typedef ::sstmac::sw::api parent_api;

        static debug dbg_;
        softwareid id_;
        boost::intrusive_ptr<server> server_;
        heap::ptr heap_;

        //GH: this map gets shared by everyone, modeling the information that would be
        //  passed to everyone. I made it a separate object so that we don't
        //  accidentally false-share some stuff
        static boost::unordered_map<int, shared_heap_info::ptr> heap_info_;

        // lib_compute_memmove::ptr lib_mem_;

        key::ptr global_key_;

        boost::unordered_map<long, transfer::ptr> pending_puts_;
        bool waiting_on_quiet_;
        payload::const_ptr get_arrived_;

        symmetric::ptr current_fake_;

        api();

      public:
        typedef boost::intrusive_ptr<api> ptr;

        static ptr
        construct(const sim_parameters::ptr& params)
        {
          return new api;
        }

        void
        init(const softwareid &sid);

        virtual
        ~api();

        virtual std::string
        to_string() const
        {
          return "api";
        }

        virtual void
        init_os(const OS_API::ptr& os);

        // ----- helpers -----
        virtual symmetric::ptr
        get_symmetric_by_addr(long addr);

        virtual void
        put_done(long tid);

        virtual void
        get_arrived(payload::const_ptr pay);

        virtual void
        send_put_complete(transfer::const_ptr t) const;

        virtual void
        send_get_response(transfer::const_ptr t, payload::const_ptr pay);

        virtual void
        unblock();

        virtual symmetric::ptr
        get_fake_sym(size_t s);

        template<typename T>
          void
          copy_target(const T* from, T* to, size_t n, int stride = 1)
          {
            for (int i = 0; i < n; i++)
            {
              to[i * stride] = from[i];
            }
          }

        void
        start_trans(transfer::ptr tran, bool block = true,
            bool unblock_on_mem = true);

        template<typename T>
          void
          init_globals()
          {
            if (global<T>::constructed_)
            {
              for (int i = 0; i < global<T>::constructed_->size(); i++)
              {
                global<T>* gptr = (*global<T>::constructed_)[i];

                heap_->add_global(gptr);

              }
            }

            // global<T>::constructed_->clear();
          }

        bool
        pe_in_set(int pe, int PE_start, int logPE_stride, int PE_size);

        //--------------
        //---- these are the real API functions ----
        //--------------

        void
        start_pes(int npes);

        int
        _num_pes();

        int
        _my_pe();

        void*
        shmalloc(size_t size);

        void
        shfree(void* ptr);

        int
        shmem_addr_accessible(void* addr, int pe);

        void
        shmem_barrier(int PE_start, int logPE_stride, int PE_size);

        void
        shmem_quiet();

        template<typename T>
          void
          shmem_p(T *addr, T value, int pe)
          {
            SSTMACBacktrace("shmem_p");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_p \n";
            long a;
            int o;
            heap_->get_addr_off(addr, &a, &o);
            boost::intrusive_ptr<put_transfer<T> > tran =
                put_transfer<T>::construct(a, o, value, id_.task_.id, pe, 1, 1);
            pending_puts_[tran->get_tid()] = tran;
            start_trans(tran, true, true);
            end_api_call();
          }

        template<typename T>
          void
          shmem_put(T *addr, const T *value, size_t nelems, int pe,
              ptrdiff_t tstride = 1, ptrdiff_t sstride = 1)
          {
            SSTMACBacktrace("shmem_put");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_put \n";
            long a;
            int o;
            heap_->get_addr_off<T> (addr, &a, &o);
            boost::intrusive_ptr<put_transfer<T> > tran =
                put_transfer<T>::construct(a, o, value, nelems, id_.task_.id,
                    pe, tstride, sstride);
            pending_puts_[tran->get_tid()] = tran;
            start_trans(tran, true, true);
            end_api_call();
          }

        template<typename T>
          T
          shmem_g(T *addr, int pe)
          {
            SSTMACBacktrace("shmem_g");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_g \n";
            long a;
            int o;
            heap_->get_addr_off<T> (addr, &a, &o);
            boost::intrusive_ptr<get_transfer<T> > tran =
                get_transfer<T>::construct(a, o, id_.task_.id, pe, 1);
            start_trans(tran, true, false);
            T ret;
            if (addr != NULL)
            {
              boost::intrusive_ptr<const vector1payload<T, T*> > cast =
                  boost::dynamic_pointer_cast<const vector1payload<T, T*> >(
                      get_arrived_);
              ret = cast->data()[0];
              get_arrived_ = payload::ptr();
            }
            end_api_call();
            return ret;
          }

        template<typename T>
          T
          shmem_get(T *target, const T *source, size_t nelems, int pe,
              ptrdiff_t tstride = 1, ptrdiff_t sstride = 1)
          {
            SSTMACBacktrace("shmem_get");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_get \n";
            long a;
            int o;
            heap_->get_addr_off<T> (source, &a, &o);
            boost::intrusive_ptr<get_transfer<T> > tran =
                get_transfer<T>::construct(a, o, nelems, id_.task_.id, pe,
                    sstride);
            start_trans(tran, true, false);
            if (source != NULL && target != NULL)
            {
              boost::intrusive_ptr<const vector1payload<T, T*> > cast =
                  boost::dynamic_pointer_cast<const vector1payload<T, T*> >(
                      get_arrived_);
              copy_target<T> (cast->data(), target, nelems, tstride);
              get_arrived_ = payload::ptr();
            }
            end_api_call();
            return *target;
          }

        /**
         * This is not really implemented that well, just a bunch of gets and a barrier
         */
        template<typename T>
          void
          broadcast(T* target, const T* source, size_t nelems, int PE_root,
              int PE_start, int logPE_stride, int PE_size)
          {
            SSTMACBacktrace("shmem_get");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::broadcast \n";

            if (_my_pe() != PE_root)
            {
              shmem_get(target, source, nelems, PE_root);
            }

            shmem_barrier(PE_start, logPE_stride, PE_size);

            end_api_call();
          }

        void
        shmem_int_inc(int *target, int pe);
        void
        shmem_long_inc(long *target, int pe);
        void
        shmem_longlong_inc(long long *target, int pe);

        template<typename T>
          void
          do_op(T *target, T *source, int nreduce, int PE_start,
              int logPE_stride, int PE_size, T *pWrk, long *pSync,
              symmetric::OPS op)
          {

            long a1, a2;
            int o1, o2;

            if (target == NULL)
            {
              if (!current_fake_)
              {
                symmetric::ptr fake = heap_->shmalloc(nreduce, false);
                current_fake_ = fake;
              }
              a1 = current_fake_->get_shmem_addr();
              o1 = -1;
            }
            else
            {
              heap_->get_addr_off<T> (target, &a1, &o1);
            }

            payload::const_ptr pay;

            if (source != NULL)
            {
              heap_->get_addr_off<T> (source, &a2, &o2);
              T* raw = (T*) get_symmetric_by_addr(a2)->get_raw();
              raw += o2;

              pay = vector1payload<T, T*>::construct(raw, nreduce);

            }
            else
            {
              pay = fake_data::construct(sizeof(T) * nreduce);
            }

            int penum = PE_start;
            for (int i = 0; i < PE_size; i++)
            {
              boost::intrusive_ptr<op_transfer<T> > tran =
                  op_transfer<T>::construct(a1, o1, pay, nreduce, penum,
                      id_.task_.id, op, PE_size);
              start_trans(tran, false, false);
              penum = (penum + (int) pow(2, logPE_stride)) % _num_pes();
            }

            os_->block(global_key_);
            if (target == NULL)
            {
              heap_->free_by_addr(current_fake_->get_shmem_addr());
              current_fake_ = symmetric::ptr();
            }
          }

        template<typename T>
          void
          shmem_sum_to_all(T *target, T *source, int nreduce, int PE_start,
              int logPE_stride, int PE_size, T *pWrk, long *pSync)
          {
            SSTMACBacktrace("shmem_sum_to_all");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_sum_to_all \n";
            do_op(target, source, nreduce, PE_start, logPE_stride, PE_size,
                pWrk, pSync, symmetric::SUM);
            end_api_call();
          }

        template<typename T>
          void
          shmem_min_to_all(T *target, T *source, int nreduce, int PE_start,
              int logPE_stride, int PE_size, T *pWrk, long *pSync)
          {
            SSTMACBacktrace("shmem_min_to_all");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_sum_to_all \n";
            do_op(target, source, nreduce, PE_start, logPE_stride, PE_size,
                pWrk, pSync, symmetric::MIN);
            end_api_call();
          }

        template<typename T>
          void
          shmem_max_to_all(T *target, T *source, int nreduce, int PE_start,
              int logPE_stride, int PE_size, T *pWrk, long *pSync)
          {
            SSTMACBacktrace("shmem_max_to_all");
            start_api_call();
            SSTMAC_DEBUG << "shmem::api::shmem_sum_to_all \n";
            do_op(target, source, nreduce, PE_start, logPE_stride, PE_size,
                pWrk, pSync, symmetric::MAX);
            end_api_call();
          }

      };

    }
  }
}
#endif /* api_H_ */
