/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SHMEM_symmetric_H_
#define SSTMAC_SHMEM_symmetric_H_

#include <sstmac/common/ptr_type.h>

#include <sstmac/common/debug.h>

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {
      class symmetric : public ptr_type
      {
        void* raw_;
        size_t size_; //bytes
        long addr_;

        bool real_;

        static debug dbg_;

        int op_counter_;

        symmetric(size_t size, long addr, bool real);

        symmetric(size_t size, long addr, void* raw);
      public:
        typedef boost::intrusive_ptr<symmetric> ptr;

        enum OPS
        {
          MIN, MAX, SUM
        };

        static ptr
        construct(size_t size, long addr, bool real = true)
        {
          return ptr(new symmetric(size, addr, real));
        }

        static ptr
        construct(size_t size, long addr, void* raw)
        {
          return ptr(new symmetric(size, addr, raw));
        }

        virtual
        ~symmetric()
        {
          if (real_)
            free(raw_);
        }

        virtual std::string
        to_string() const
        {
          return "symmetric";
        }

        size_t
        get_size() const
        {
          return size_;
        }

        void*
        get_raw() const
        {
          return raw_;
        }

        long
        get_shmem_addr() const
        {
          return addr_;
        }

        bool
        is_real() const
        {
          return real_;
        }

        void
        reset_op_counter()
        {
          op_counter_ = 0;
        }

        int
        get_op_counter() const
        {
          return op_counter_;
        }

        void
        perform_fake_op(size_t num)
        {
          op_counter_++;
        }

        template<typename T>
          void
          perform_op(const T* vals, int offset, size_t num, OPS op)
          {
            SSTMAC_DEBUG << "operating on addr " << addr_ << " with op count "
                  << op_counter_ << "\n";

            if (real_)
            {
              T* cast = (T*) raw_;
              cast += offset;

              switch (op)
              {

              case SUM:
              {
                for (size_t i = 0; i < num; i++)
                {

                  if (op_counter_ == 0)
                  {
                    cast[i] = 0;
                  }
                  cast[i] += vals[i];
                }
                break;
              }

              case MIN:
              {

                for (size_t i = 0; i < num; i++)
                {

                  if (op_counter_ == 0)
                  {
                    cast[i] = vals[i];
                  }
                  else if (vals[i] < cast[i])
                  {
                    cast[i] = vals[i];
                  }
                }
                break;
              }

              case MAX:
              {

                for (size_t i = 0; i < num; i++)
                {

                  if (op_counter_ == 0)
                  {
                    cast[i] = vals[i];
                  }
                  else if (vals[i] > cast[i])
                  {
                    cast[i] = vals[i];
                  }
                }
                break;
              }

              default:
                sst_throw(ssterror, "symmetric::perform_op() - unknown op %d", op);
              }
            }

            op_counter_++;
          }

      };

    }
  }
}
#endif /* symmetric_H_ */
