/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include "shmem/server.h"
#include "shmem/api.h"
#include <sstmac/common/sstevent.h>
#include <sstmac/common/eventcallback.h>
#include <sstmac/software/process/operatingsystem.h>

#include <sstmac/common/messages/valuepayload.h>

#include "shmem/messages/message.h"
#include "shmem/messages/put_transfer.h"

namespace sstmac
{
  namespace sw
  {
    namespace shmem
    {
      debug server::dbg_;

      server::server(int aid) :
        aid_(aid)
      {

        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> server | shmem", logger::blue);
            logger::register_logger(&dbg_);
          }
        libname_ = sstprintf("shmem::server-%d", aid);

        q_server_busy_ = false;
      }

      server::~server()
      {
      }

      service::ptr
      server::clone()
      {
        return construct(aid_);
      }

      void
      server::start()
      {
        service::start();

        //sender_ = msg_callback(mpiserver::ptr(this), &mpiserver::start_send);
        lib_mem_ = lib_compute_memmove::construct(libname_ + "-lib-mem");
        lib_mem_->init_as_callback_lib();
        register_lib(lib_mem_);
      }

      void
      server::incoming_mem_done(transfer::const_ptr t)
      {
        SSTMAC_DEBUG << "shmem::server(" << os_->get_my_addr()->unique_id()
              << ")::incoming_mem_done - executing \n";

        t->execute(apis_[t->get_remote_pe()]);

        q_server_busy_ = false;
        check_q();

      }

      void
      server::send_tran_response(transfer::const_ptr t, payload::const_ptr pay,
          message::SHMEM_MSG_TYPE type)
      {
        //tell the API that it's done, for fences and quiets
        appid a(aid_);
        environment::const_ptr env = os_->get_env(a);
        nodeaddress::ptr to = env->get_node_address(t->get_local_pe());
        nodeaddress::ptr from = os_->get_my_addr();

        sst_message::MESSAGE_TYPES ty;

        switch(type){
        case message::GET_RESPONSE:
          ty = sst_message::RDMA_GET_RSP;
          break;
        case message::PUT_COMPLETE:
          ty = sst_message::RDMA_PUT_RSP;
          break;
        default:
          sst_throw(ssterror, "shmem::server::send_tran_response - unexpected msg type");

        }

        message::ptr msg = message::construct(type, libname_, to, from,
            taskid(t->get_remote_pe()), taskid(t->get_local_pe()), ty, pay);

        if (to != from)
          {
            os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
          }
        else
          {
           // sst_throw(ssterror, "shmem::server::incoming_mem_done - need to implement intra-node transfer");
            handle(msg);
          }
      }

      void
      server::check_q()
      {
        if (!transfer_q_.empty() && !q_server_busy_)
          {
            SSTMAC_DEBUG << "shmem::server(" << os_->get_my_addr()->unique_id()
                  << ")::check_q start new transfer \n";

            transfer::const_ptr t = transfer_q_.front();
            transfer_q_.pop();

            if (apis_.find(t->get_remote_pe()) == apis_.end())
              {
                sst_throw(ssterror, "shmem::server::check_q() - cannot find pe %d at node %s", t->get_remote_pe(), os_->get_my_addr()->to_string().c_str());
              }

            if (t->size_bytes() > 0)
              {
                taskid tas(t->get_remote_pe());
                int destth = fullos_->get_task_threadid(tas);

                eventcallback::ptr cb = new_callback(ptr(this),
                    &server::incoming_mem_done, t);

                lib_mem_->copy(t->size_bytes(), thread::nic_thread, destth, cb);

              }
            else
              {
                event::ptr ev = new_event(ptr(this),
                    &server::incoming_mem_done, t);
                schedule(now(), ev);
              }

            q_server_busy_ = true;
          }
      }

      void
      server::handle(const sst_message::ptr &msg)
      {
        SSTMAC_DEBUG << "shmem::server(" << os_->get_my_addr()->unique_id()
              << ")::handle...";
        message::ptr shm = boost::dynamic_pointer_cast<message>(msg);

        if (!shm)
          {
            sst_throw(ssterror, "shmem::server::handle - couldn't cast to a shmem::message");
          }

        switch (shm->get_shmem_type())
          {
        case message::TRAN:
          {
            SSTMAC_DEBUG << "EXECUTE_TRANSACTION\n";
            transfer::const_ptr t =
                boost::dynamic_pointer_cast<const transfer>(shm->get_data());
            transfer_q_.push(t);
            check_q();
            break;
          }

        case message::PUT_COMPLETE:
          {
            SSTMAC_DEBUG << "PUT_COMPLETE\n";
            valuepayload<long>::const_ptr tid = boost::dynamic_pointer_cast<
                const valuepayload<long> >(shm->get_data());
            taskid dst = shm->dest_task();
            apis_[dst.id]->put_done(tid->data());
            break;
          }

        case message::GET_RESPONSE:
          {
            SSTMAC_DEBUG << "GET_RESPONSE\n";
            payload::const_ptr load =
                boost::dynamic_pointer_cast<const payload>(shm->get_data());
            taskid dst = shm->dest_task();
            apis_[dst.id]->get_arrived(load);
            break;
          }
          }
      }

      void
      server::start_transfer(transfer::ptr tran, key::ptr k)
      {
        SSTMAC_DEBUG << "shmem::server(" << os_->get_my_addr()->unique_id()
              << ")::tran_transfer \n";

        if (tran->size_bytes() > 0)
          {
            eventcallback::ptr cb = new_callback(ptr(this),
                &server::tran_copy_complete, tran, k);
            lib_mem_->copy(tran->size_bytes(), os_->get_current_threadid(),
                thread::nic_thread, cb);
          }
        else
          {
            event::ptr ev = new_event(ptr(this), &server::tran_copy_complete,
                tran, k);
            schedule(now(), ev);
          }
      }

      void
      server::tran_copy_complete(transfer::const_ptr tran, key::ptr k)
      {
        SSTMAC_DEBUG << "shmem::server(" << os_->get_my_addr()->unique_id()
              << ")::tran_copy_complete \n";
        appid a(aid_);
        environment::const_ptr env = os_->get_env(a);
        nodeaddress::ptr to = env->get_node_address(tran->get_remote_pe());
        nodeaddress::ptr from = os_->get_my_addr();

        sst_message::MESSAGE_TYPES ty;

       /* switch(tran->get_trans_type()){
        case transfer::GET:
          ty = sst_message::RDMA_GET_REQ;
          break;
        case transfer::ATOMIC:
        case transfer::PUT:
          ty = sst_message::RDMA_PUT_REQ;
          break;
        default:
          sst_throw(ssterror, "shmem::server::send_tran_response - unexpected msg type");

        }*/
        ty = sst_message::NETWORK;

        message::ptr msg = message::construct(message::TRAN, libname_, to,
            from, taskid(tran->get_local_pe()), taskid(tran->get_remote_pe()), ty,
            tran);

        if (to != from)
          {

            os_->execute_kernel(AMI::AMI_COMM_SEND, msg);
          }
        else
          {
            // sst_throw(ssterror, "shmem::server::put_copy_complete - need to implement intra-node put");
            handle(msg);
          }

        if (k)
          os_->unblock(k);
      }

      void
      server::add_pe(boost::intrusive_ptr<api> api)
      {
        apis_[api->_my_pe()] = api;
      }

      void
      server::unblock_other_server(ptr serv, key::ptr k)
      {
        event::ptr ev = new_event(serv, &server::unblock_api, k);
        schedule(now(), ev);
      }

      void
      server::unblock_api(key::ptr k)
      {
        os_->unblock(k);
      }
    }
  }
}
