/*  api.cc                                                                   */

#include "sstmac_upc/api.h"
#include "sstmac_upc/server.h"
#include "sstmac_upc/memory/shared_ptr.h"

#include <sstmac/common/messages/valuepayload.h>

ImplementAPI(sstmac::sw::upc, api, "upc")
;

namespace sstmac
{
    namespace sw
    {
        namespace upc
        {
            SSTRegister("upc", sstmac::sw::api, api, "Create bindings for UPC runtime");
            
            debug api::dbg_;
            
            long api::notify_count_ = 0;
            long api::wait_index_ = 0;
            int  api::notify_key_1_ = 0;
            int  api::notify_key_2_ = 0;
            boost::unordered_map<key::ptr, server::ptr> api::notify_servmap_;
            
            api::api() :
                parent_api()
            {
                if (!dbg_.is_real())
                {
                    dbg_ = debug("<debug> upc | api", logger::yellow);
                    logger::register_logger(&dbg_);
                }
                
                global_key_ = key::construct();
                waiting_on_fence_ = false;
                inside_notify_ = false;
                loc_wait_index_ = 0;
                get_arrived_ = payload::ptr();
                setup_ = false;
            }
            
            void
            api::setup()
            {
                if (!setup_)
                {
                    SSTMAC_DEBUG << "upc::api::setup() - start \n";
                    std::string server_libname = sstprintf("upc::server-%d", id_.app_.id);
                    library::ptr upclib = os_->get_lib(server_libname);

                    if (upclib == library::ptr()) // only do one mpiserver per app per node
                    {
                        server::ptr serv = server::construct(id_.app_.id);
                        upclib = serv;
                        os_->register_ser(serv);
                        serv->start();
                    }

                    server_ = boost::dynamic_pointer_cast<server>(upclib);
                    server_->add_thread(ptr(this));
                    server_->set_eventmanager(os_->get_eventmanager());

                    heap_ = shared_heap::construct();

                    //check for shared objects
                    init_shared();
                    
                    setup_ = true;
                    
                    SSTMAC_DEBUG << "upc::api::setup() - done \n";
                }
            }
            
            void
            api::init(const softwareid& sid)
            {
                parent_api::init(sid);
                id_ = sid;
                libname_ = "upcapi" + id_.to_string();
                //parent_api::init(sid);
            }
            
            api::~api()
            {
            }
            
            void
            api::init_os(const OS_API::ptr& os)
            {
                parent_api::init_os(os);
                
                // lib_mem_ = lib_compute_memmove::construct(libname_ + "-lib-mem");
                // lib_mem_->init_callback();
                // register_lib(lib_mem_);
            }
            
            void
            api::put_done(long tid)
            {
                if (pending_puts_.find(tid) == pending_puts_.end())
                {
                    sst_throw(ssterror, "upc::api::put_done - couldn't find put %l", tid);
                }
                
                pending_puts_.erase(tid);
                if (waiting_on_fence_ && pending_puts_.size() == 0)
                {
                    os_->unblock(global_key_);
                }
                
            }
            
            void
            api::unblock()
            {
                os_->unblock(global_key_);
            }
            
            void
            api::send_put_complete(transfer::const_ptr t) const
            {
                payload::const_ptr load = valuepayload<long>::construct(t->get_tid());
                server_->send_tran_response(t, load, message::PUT_COMPLETE);
            }
            
            void
            api::get_arrived(payload::const_ptr pay)
            {
                get_arrived_ = pay;
                // if (!get_arrived_)
                // {
                //      sst_throw(ssterror, "shmem::api::get_arrived with null payload ");
                // }
                os_->unblock(global_key_);
            }
            
            void
            api::send_get_response(transfer::const_ptr t, payload::const_ptr pay)
            {
                server_->send_tran_response(t, pay, message::GET_RESPONSE);
            }
            
            void
            api::start_trans(transfer::ptr tran, bool block, bool unblock_on_mem)
            {
                
                server_->start_transfer(tran,
                    (unblock_on_mem) ? global_key_ : key::ptr());
                
                if (block)
                    os_->block(global_key_);
                
            }
            
            bool
            api::check_shared_heap(shared::ptr sobj)
            {
                return heap_->contains_shared(sobj);
            }
            
            // UPC API calls.
            
            size_t
            api::threads()
            {
                SSTMAC_DEBUG << "upc::api::threads() \n";
                environment::const_ptr env = os_->get_env(id_.app_);
                return env->get_app_count();
            }
            
            size_t
            api::mythread()
            {
                SSTMAC_DEBUG << "upc::api::mythread() \n";
                return id_.task_.id;
            }
            
            size_t
            api::maxblocksize()
            {
                SSTMAC_DEBUG << "upc::api::maxblocksize() \n";
                return UPC_MAX_BLOCKSIZE;
            }
            
            void
            api::notify()
            {
                notify(DEFAULT_KEY);
            }
            
            void
            api::notify(int nkey)
            {
                SSTMACBacktrace("notify");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::notify(" << nkey << ") \n";
                
                if (inside_notify_)
                {
                    sst_throw(ssterror, "upc::api::notify called inside another notify-wait pair.");
                }
                else 
                {
                    fence();
                    
                    notify_count_++;
                    loc_notify_key_ = nkey;
                    inside_notify_ = true;
                    
                    SSTMAC_DEBUG << "api: thread " << id_.task_.id
                                 << " entering notify(), count is " << notify_count_ << "\n";
                    
                    if (notify_count_ == 1)
                    {
                        notify_key_1_ = nkey;
                    }
                    else if (notify_count_ == threads())
                    {
                        notify_count_ = 0;
                        wait_index_++;
                        
                        notify_key_2_ = notify_key_1_;
                        
                        boost::unordered_map<key::ptr, server::ptr>::iterator it, end =
                            notify_servmap_.end();
                        for (it = notify_servmap_.begin(); it != end; it++)
                        {
                            server_->unblock_other_server(it->second, it->first);
                        }
                        notify_servmap_.clear();
                    }
                }
                
                end_api_call();
            }
            
            void
            api::wait()
            {
                wait(DEFAULT_KEY);
            }
            
            void
            api::wait(int wkey)
            {
                SSTMACBacktrace("wait");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::wait(" << wkey << ")\n";
                
                if (wkey != DEFAULT_KEY && loc_notify_key_ != DEFAULT_KEY && wkey != loc_notify_key_)
                {
                    sst_throw(ssterror, "upc::api::wait key(%d) did not match the notify key (%d).", wkey, loc_notify_key_);
                }
                else if (!inside_notify_)
                {
                    sst_throw(ssterror, "upc::api::wait called without prior call to upc::api::notify.");
                }
                else
                {
                    if (wait_index_ == loc_wait_index_)
                    {
                        SSTMAC_DEBUG << "api: wait(), blocking \n";

                        notify_servmap_[global_key_] = server_;
                        os_->block(global_key_);

                        SSTMAC_DEBUG << "api: wait(), thread " << id_.task_.id
                                     << " unblocked \n";
                        
                    }
                    
                    if (wkey != DEFAULT_KEY && notify_key_2_ != DEFAULT_KEY && wkey != notify_key_2_)
                    {
                        sst_throw(ssterror, "upc::api::wait key(%d) did not match the notify key (%d).", wkey, notify_key_2_);
                    }
                    
                    inside_notify_ = false;
                    loc_wait_index_++;
                }
                
                end_api_call();
            }
            
            void
            api::barrier()
            {
                barrier(DEFAULT_KEY);
            }
            
            void
            api::barrier(int bkey)
            {
                SSTMACBacktrace("barrier");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::barrier(" << bkey << ")\n";
                static int keys = DEFAULT_KEY;
                static boost::unordered_map<key::ptr, server::ptr> servmap;
                static long count = 0;
                
                if (inside_notify_)
                {
                    sst_throw(ssterror, "upc::api::barrier cannot be called inside a notify-wait block.");
                }
                else
                {
                    fence();
                    
                    if (count == 0)
                        keys = bkey;
                    else if (bkey != DEFAULT_KEY && keys != DEFAULT_KEY && bkey != keys)
                        sst_throw(ssterror, "upc::api::barrier called with mismatched keys: %d and %d.", bkey, keys);
                    
                    inside_notify_ = true;
                    count++;
                    
                    SSTMAC_DEBUG << "api: thread " << id_.task_.id
                                 << " entering barrier(), count is " << count << "\n";
                    
                    if (count == threads())
                    {
                        count = 0;
                        
                        SSTMAC_DEBUG << "api: barrier(), unblocking everyone \n";
                        
                        boost::unordered_map<key::ptr, server::ptr>::iterator it, end =
                            servmap.end();
                        for (it = servmap.begin(); it != end; it++)
                        {
                            server_->unblock_other_server(it->second, it->first);
                        }
                        
                        servmap.clear();
                    }
                    else
                    {
                        SSTMAC_DEBUG << "api: barrier(), blocking \n";

                        servmap[global_key_] = server_;
                        os_->block(global_key_);

                        SSTMAC_DEBUG << "api: barrier(), thread " << id_.task_.id
                                     << " unblocked \n";
                    }
                    
                    inside_notify_ = false;
                }
                
                end_api_call();
            }
            
            void
            api::fence()
            {
                SSTMACBacktrace("fence");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::fence() \n";
                
                if (pending_puts_.size() > 0)
                {
                    waiting_on_fence_ = true;
                    os_->block(global_key_);
                    waiting_on_fence_ = false;
                }

                end_api_call();
            }
            
            size_t
            api::localsizeof(std::string type_name)
            {
                // Will need to parse the string for a shared type and return the localsize of that type.
                sst_throw(unimplementederror, "upc::api::localsizeof from a type name is unimplemented.");
            }
            
            size_t
            api::blocksizeof(std::string type_name)
            {
                // Will need to parse the string for a shared type and return the blocksize of that type.
                sst_throw(unimplementederror, "upc::api::blocksizeof from a type name is unimplemented.");
            }
            
            size_t
            api::elemsizeof(std::string type_name)
            {
                // Will need to parse the string for a shared type and return the elemsize of that type.
                sst_throw(unimplementederror, "upc::api::elemsizeof from a type name is unimplemented.");
            }
            
            void
            api::global_exit(int status)
            {
                SSTMACBacktrace("global_exit");
                SSTMAC_DEBUG << "upc::api::global_exit(" << status << ")\n";
                
                // Maybe right... may not need the barrier (perhaps just a fence?)
                barrier(status);
                
                end_api_call();
                exit(status);
            }
            
            size_t
            api::affinitysize(size_t totalsize, size_t nbytes, size_t threadid)
            {
                SSTMACBacktrace("affinitysize");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::affinitysize(" << totalsize << ", " << nbytes << ", " << threadid << ")\n";
                
                size_t ret = 0;
                size_t numthreads = threads();
                
                // Start with the size of the blocks on all threads.
                ret = totalsize / numthreads;
                // Check for the last set of blocks.
                ret += (totalsize % numthreads >= nbytes*mythread()) ? nbytes : 0;
                
                end_api_call();
                return ret;
            }
            
            // Shared locking functions.
            
            lock *
            api::global_lock_alloc()
            {
                SSTMACBacktrace("global_lock_alloc");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::global_lock_alloc()\n";
                
                lock *lck = new lock();
                heap_->add_lock(lck);
                
                end_api_call();
                return lck;
            }
            
            lock *
            api::all_lock_alloc()
            {
                SSTMACBacktrace("all_lock_alloc");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::all_lock_alloc()\n";
                static lock *lck;
                static long count = 0;
                lock *loc;
                
                if (inside_notify_)
                {
                    sst_throw(ssterror, "upc::api::all_lock_alloc cannot be called inside a notify-wait block.");
                }
                else
                {
                    count++;
                    
                    if (count == 1)
                    {
                        lck = new lock();
                    }
                    else if (count == threads())
                    {
                        count = 0;
                    }
                    
                    heap_->add_lock(lck);
                    loc = lck;
                }
                
                // Make sure that we don't get into another collective and mess with our static variables...
                barrier(-161);
                
                end_api_call();
                return loc;
            }
            
            void
            api::lock_free(lock *lck)
            {
                SSTMACBacktrace("lock_free");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::lock_free(" << lck->to_string() << ")\n";
                
                if (lck == 0)
                    return;
                
                heap_->delete_lock(lck);
                delete lck;
                
                end_api_call();
            }
            
            void
            api::lock_wait(lock *lck)
            {
                SSTMACBacktrace("lock_wait");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::lock_wait(" << lck->to_string() << ")\n";
                bool have_lock = lck->take_lock(global_key_);
                
                while (!have_lock)
                {
                    lck->add_server(global_key_, server_);
                    os_->block(global_key_);
                    have_lock = lck->take_lock(global_key_);
                }
                
                end_api_call();
            }
            
            int
            api::lock_attempt(lock *lck)
            {
                SSTMACBacktrace("lock_attempt");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::lock_attempt(" << lck->to_string() << ")\n";
                bool have_lock = lck->take_lock(global_key_);
                int ret = (have_lock) ? 1 : 0;
                end_api_call();
                return ret;
            }
            
            void
            api::unlock(lock *lck)
            {
                SSTMACBacktrace("unlock");
                start_api_call();
                SSTMAC_DEBUG << "upc::api::unlock(" << lck->to_string() << ")\n";
                
                lck->unlock(global_key_);
                
                boost::unordered_map<key::ptr, server::ptr> servermap = lck->get_servermap();
                boost::unordered_map<key::ptr, server::ptr>::iterator it = servermap.begin();
                
                server_->unblock_other_server(it->second, it->first);
                
                end_api_call();
            }
            
        }
    }
}
