/*  shared_ptr.h                                                         */

#ifndef SSTMAC_UPC_shared_ptr_H_
#define SSTMAC_UPC_shared_ptr_H_

#include <sstmac/software/process/os_api.h>
#include <sstmac/common/debug.h>

#include "sstmac_upc/memory/shared.h"
#include "sstmac_upc/api.h"

#include "sstmac_upc/messages/transfer.h"
#include "sstmac_upc/messages/put_transfer.h"
#include "sstmac_upc/messages/get_transfer.h"

namespace sstmac
{
  namespace sw
  {
    namespace upc
    {
      template<typename T, size_t B>
        class shared_reference;

      template<typename T, size_t B>
        class shared_ptr
        {
        private:
          shared::ptr ref_; // Reference to a shared object.
          size_t baseaddr_; // Offset of this block (in bytes) from the begining of the shared object's memory on this thread.
          size_t threadid_; // This thread.
          size_t phase_; // Offset (in units of 1/sizeof(T)) from the begining of this block
          size_t blocksize_;

          size_t numthreads_;
          size_t mythread_;
          api::ptr api_;

          static debug dbg_;

        public:

          shared_ptr() :
            ref_(NULL), baseaddr_(0), threadid_(0), phase_(0), blocksize_(B)
          {
            //set_api();
            //setup_debug();
          }

          shared_ptr(shared::ptr ref) :
            ref_(ref), baseaddr_(0), phase_(0), blocksize_(B)
          {
            if (ref)
              threadid_ = (ref->get_starting_thread());
            else
              threadid_ = 0;
            set_api();
            setup_debug();
          }

          //                     shared_ptr(shared_ptr& sptr) :
          //                         ref_(sptr.ref_), baseaddr_(sptr.baseaddr_), threadid_(sptr.threadid_), phase_(sptr.phase_), blocksize_(B)
          //                     {
          //                         set_api();
          //                         setup_debug();
          //                     }

          template<typename U, size_t C>
            shared_ptr(shared_ptr<U, C>& sptr) :
              ref_(sptr.get_ref()), baseaddr_(sptr.get_baseaddr()),
                  threadid_(sptr.get_threadid()), phase_(0), blocksize_(B)
            {
              if (sizeof(T) == sizeof(U) && B == C)
                phase_ = sptr.get_phase();
              set_api();
              setup_debug();
            }

          template<size_t C>
            shared_ptr(shared_ptr<void, C>& sptr) :
              ref_(sptr.get_ref()), baseaddr_(sptr.get_baseaddr()),
                  threadid_(sptr.get_threadid()), phase_(sptr.get_phase()),
                  blocksize_(B)
            {
              set_api();
              setup_debug();
            }

          //                     // Creation with NULL value.
          //                     shared_ptr(int null)
          //                     {
          //                         if (null != 0)
          //                         {
          //                             sst_throw(ssterror, "upc::shared_ptr cannot be assigned from a non-NULL integer.");
          //                         }
          //                         baseaddr_ = 0;
          //                         threadid_ = 0;
          //                         phase_ = 0;
          //                         blocksize_ = B;
          //                         set_api();
          //                         setup_debug();
          //                     }

          virtual
          ~shared_ptr()
          {
          }

          // Helpers.

          void
          set_api()
          {
            thread::ptr t = OS_API::get_current_thread();
            api::ptr a = t->get_api<api> ();

            numthreads_ = a->threads();
            mythread_ = a->mythread();
            api_ = a;
          }

          void
          setup_debug()
          {
            if (!dbg_.is_real())
            {
              dbg_ = debug("<debug> shared_ptr | upc", logger::yellow);
              logger::register_logger(&dbg_);
            }
          }

          void
          shared_ptr_add(long increment)
          {
            SSTMAC_DEBUG << "upc::shared_ptr::shared_ptr_add(" << increment
                  << ")\n";

            size_t new_baseaddr, new_threadid, new_phase;
            long numthreads = numthreads_; //api_->threads();
            long block_inc, thread_inc, phase_inc;

            if (blocksize_ == 0)
            {
              // Nothing fancy - Just increment the base address like a normal pointer.
              new_baseaddr = baseaddr_ + (increment * sizeof(T));
              new_phase = 0;
              new_threadid = threadid_;
            }
            else if (blocksize_ == 1)
            {
              long modoff = (increment >= 0) ? 0 : 1 - numthreads;
              thread_inc = threadid_ + increment + modoff;
              block_inc = thread_inc / numthreads;
              new_threadid = thread_inc % numthreads - modoff;
              new_baseaddr = baseaddr_ + block_inc * sizeof(T);
              new_phase = 0;
            }
            else
            {
              if (increment >= 0)
              {
                thread_inc = (phase_ + increment) / blocksize_;
                new_phase = (phase_ + increment) % blocksize_;
                block_inc = (threadid_ + thread_inc) / numthreads;
                new_threadid = (threadid_ + thread_inc) % numthreads;
                new_baseaddr = baseaddr_ + block_inc * (blocksize_ * sizeof(T));
              }
              else
              {
                long modoff = 1 - blocksize_;
                phase_inc = phase_ + increment + modoff;
                thread_inc = phase_inc / blocksize_;
                new_phase = phase_inc % blocksize_ - modoff;
                modoff = 1 - numthreads;
                block_inc = (threadid_ + thread_inc + modoff) / numthreads;
                new_threadid = (threadid_ + thread_inc + modoff) % numthreads
                    - modoff;
                new_baseaddr = baseaddr_ + block_inc * (blocksize_ * sizeof(T));
              }
            }

            phase_ = new_phase;
            threadid_ = new_threadid;
            baseaddr_ = new_baseaddr;
          }

          T&
          get_val()
          {
            T* tptr = (T*) ref_->get_offset_raw(baseaddr_, threadid_, phase_,
                sizeof(T), blocksize_);
            return *tptr;
          }

          T&
          get_val(long index)
          {
            shared_ptr<T, B> other = *this + index;
            return other.get_val();
          }

          T
          get_val_message()
          {
            SSTMAC_DEBUG << "upc::shared_ptr::get_val_message()\n";
            T value;
            if (threadid_ != mythread_)
            {
              SSTMAC_DEBUG
                    << "shared_ptr::get_val_message(), getting data from thread "
                    << threadid_ << "\n";
              api_->upc_get<T, B> (&value, *this, 1);
            }
            else
            {
              SSTMAC_DEBUG
                    << "shared_ptr::get_val_message(), getting local data\n";
              value = get_val();
            }
            return value;
          }

          T
          put_val_message(T value)
          {
            SSTMAC_DEBUG << "upc::shared_ptr::put_val_message(" << value
                  << ")\n";
            T temp_val = value;
            if (threadid_ != mythread_)
            {
              SSTMAC_DEBUG
                    << "shared_ptr::put_val_message(), putting data on thread "
                    << threadid_ << "\n";
              api_->upc_put<T, B> (*this, &temp_val, 1);
            }
            else
            {
              SSTMAC_DEBUG
                    << "shared_ptr::put_val_message(), putting in local memory\n";
              get_val() = value;
            }
            return value;
          }

          shared::ptr
          get_ref() const
          {
            return ref_;
          }

          size_t
          get_baseaddr() const
          {
            return baseaddr_;
          }

          size_t
          get_threadid() const
          {
            return threadid_;
          }

          size_t
          get_phase() const
          {
            return phase_;
          }

          size_t
          get_blocksize() const
          {
            return blocksize_;
          }

          size_t
          get_mythread() const
          {
            return mythread_;
          }

          size_t
          get_numthreads() const
          {
            return numthreads_;
          }

          // UPC api calls:

          size_t
          localsizeof() const
          {
            return (((ref_->get_size() / (blocksize_ * sizeof(T)))
                + numthreads_ - 1) / numthreads_) * (blocksize_ * sizeof(T));
          }

          size_t
          blocksizeof() const
          {
            return get_blocksize();
          }

          size_t
          elemsizeof() const
          {
            return sizeof(T); // maybe correct?
          }

          size_t
          threadof() const
          {
            return get_threadid();
          }

          size_t
          phaseof() const
          {
            return get_phase();
          }

          void
          resetphase()
          {
            phase_ = 0;
          }

          size_t
          addrfield() const
          {
            return (size_t) ref_->get_offset_raw(baseaddr_, threadid_, phase_,
                sizeof(T), blocksize_);
          }

          // Operators.

          virtual std::string
          to_string() const
          {
            return "shared_ptr";
          }

          // Pointer arithmetic:

          //                     shared_ptr<T, B>&
          //                     operator = (const shared_ptr& sptr)
          //                     {
          //                         ref_ = sptr.ref_;
          //                         threadid_ = sptr.threadid_;
          //                         phase_ = sptr.phase_;
          //                         baseaddr_ = sptr.baseaddr_;
          //                         return *this;
          //                     }

          template<typename U, size_t C>
            shared_ptr<T, B>
            operator =(const shared_ptr<U, C>& sptr)
            {
              SSTMAC_DEBUG << "upc::shared_ptr:: = (" << sptr.get_baseaddr()
                    << ", " << sptr.get_threadid() << ", " << sptr.get_phase()
                    << ", " << sptr.get_blocksize() << ")\n";
              ref_ = sptr.get_ref();
              threadid_ = sptr.get_threadid();
              phase_ = 0;
              if (sizeof(T) == sizeof(U) && B == C)
                phase_ = sptr.get_phase();
              baseaddr_ = sptr.get_baseaddr();
              return *this;
            }

          template<size_t C>
            shared_ptr<T, B>
            operator =(const shared_ptr<void, C>& sptr)
            {
              SSTMAC_DEBUG << "upc::shared_ptr:: = (" << sptr.get_baseaddr()
                    << ", " << sptr.get_threadid() << ", " << sptr.get_phase()
                    << ", " << sptr.get_blocksize() << ")\n";
              ref_ = sptr.get_ref();
              threadid_ = sptr.get_threadid();
              phase_ = sptr.get_phase();
              baseaddr_ = sptr.get_baseaddr();
              return *this;
            }

          shared_ptr<T, B>
          operator =(shared::ptr ref)
          {
            ref_ = ref;
            if(ref)
              threadid_ = ref->get_starting_thread();
            else{
              threadid_ = 0;
              ref_ = NULL;
              baseaddr_ = 0;
            }
            phase_ = 0;
            baseaddr_ = 0;
            return *this;
          }

          // Allows assignment from NULL / integer 0
          shared_ptr<T, B>
          operator =(int null)
          {
            if (null != 0)
            {
              sst_throw(ssterror, "upc::shared_ptr cannot be assigned from a non-NULL integer.");
            }
            baseaddr_ = 0;
            threadid_ = 0;
            phase_ = 0;
            return *this;
          }

          template<typename U>
            shared_ptr<T, B>
            operator +=(const U& inc)
            {
              long increment = static_cast<long> (inc);
              shared_ptr_add(increment);
              return *this;
            }

          template<typename U>
            shared_ptr<T, B>
            operator -=(const U& inc)
            {
              long increment = -(static_cast<long> (inc));
              shared_ptr_add(increment);
              return *this;
            }

          template<typename U>
            shared_ptr<T, B>
            operator +(const U& inc) const
            {
              long increment = static_cast<long> (inc);
              shared_ptr<T, B> other;
              other = *this;
              other.shared_ptr_add(increment);
              return other;
            }

          template<typename U>
            shared_ptr<T, B>
            operator -(const U& inc) const
            {
              long increment = -(static_cast<long> (inc));
              shared_ptr<T, B> other;
              other = *this;
              other.shared_ptr_add(increment);
              return other;
            }

          template<typename U, size_t C>
            ptrdiff_t
            operator -(const shared_ptr<U, C>& other) const
            {
              sst_throw(unimplementederror, "upc::shared_ptr subtraction is not implemented yet.");
            }

          shared_ptr<T, B>
          operator ++()
          {
            shared_ptr_add(1);
            return *this;
          }

          shared_ptr<T, B>
          operator --()
          {
            shared_ptr_add(-1);
            return *this;
          }

          shared_ptr<T, B>
          operator ++(int dummy)
          {
            shared_ptr<T, B> *other = new shared_ptr<T, B> (*this);
            shared_ptr_add(1);
            return *other;
          }

          shared_ptr<T, B>
          operator --(int dummy)
          {
            shared_ptr<T, B> *other = new shared_ptr<T, B> (*this);
            shared_ptr_add(-1);
            return *other;
          }

          template<typename U, size_t C>
            bool
            operator ==(const shared_ptr<U, C>& sptr) const
            {
              return (addrfield() == sptr.addrfield() && threadof()
                  == sptr.threadof());
            }

          template<typename U, size_t C>
            bool
            operator !=(const shared_ptr<U, C>& sptr) const
            {
              return !(addrfield() == sptr.addrfield() && threadof()
                  == sptr.threadof());
            }

          bool
          operator ==(const int null) const
          {
            if (null != 0)
            {
              sst_throw(ssterror, "upc::shared_ptr cannot be compared to non-NULL integer.");
            }

            return (ref_ == NULL) && (baseaddr_ == 0) && (threadid_ == 0)
                && (phase_ == 0);
          }

          bool
          operator !=(const int null) const
          {
            if (null != 0)
            {
              sst_throw(ssterror, "upc::shared_ptr cannot be compared to non-NULL integer.");
            }

            return (ref_ != NULL) || (baseaddr_ != 0) || (threadid_ != 0)
                || (phase_ != 0);
          }

          // Type casting.

          operator T*()
          {
            SSTMAC_DEBUG << "upc::shared_ptr:: T*()\n";
            if (!ref_)
            {
              return NULL;
            }
            if (mythread_ != threadid_)
            {
              sst_throw(ssterror, "cannot cast upc::shared_ptr to local on thread %d, points to thread %d", mythread_, threadid_);
            }
            return (T*) ref_->get_offset_raw(baseaddr_, threadid_, phase_,
                sizeof(T), blocksize_);
          }

          //                     operator void* () const
          //                     {
          //                         if (mythread_ != threadid_)
          //                         {
          //                             sst_throw(ssterror, "cannot cast upc::shared_ptr to local on thread %d, points to thread %d", mythread_, threadid_);
          //                         }
          //                         return (void*) ref_->get_offset_raw(baseaddr_, threadid_, phase_, sizeof(T), blocksize_);
          //                     }

          template<typename U, size_t C>
            operator shared_ptr<U, C>() const
            {
              shared_ptr<U, C> other(*this);
              return other;
            }

          template<size_t C>
            operator shared_ptr<void, C>() const
            {
              shared_ptr<void, C> other(*this);
              return other;
            }

          shared_reference<T, B>
          create_ref(size_t index = 0)
          {
            shared_ptr<T, B> other = *this + index;
            shared_reference<T, B> ret(other);
            return ret;
          }

          shared_reference<T, B>
          operator *()
          {
            SSTMAC_DEBUG << "upc::shared_ptr::dereference\n";
            return create_ref();
          }

          T*
          operator ->()
          {
            // Register some message passing.
            // Not real sure how this one will work...
            T* tptr = (T*) ref_->get_offset_raw(baseaddr_, threadid_, phase_,
                sizeof(T), blocksize_);
            return tptr;
          }

          shared_reference<T, B>
          operator [](size_t idx)
          {
            SSTMAC_DEBUG << "upc::shared_ptr:: [" << idx << "]\n";
            return create_ref(idx);
          }

          // Build transfer objects from the current pointer.

          transfer::ptr
          build_put(T* val, long from, size_t size)
          {
            long to = threadid_;
            SSTMAC_DEBUG << "upc::shared_ptr::build_put(" << from << ", " << to
                  << ", " << size << ")\n";

            return put_transfer<T>::construct(ref_, val, size, baseaddr_,
                phase_, from, to, ref_->is_real());
          }

          transfer::ptr
          build_get(long from, size_t size)
          {
            long to = threadid_;
            SSTMAC_DEBUG << "upc::shared_ptr::build_get(" << from << ", " << to
                  << ", " << size << ")\n";

            return get_transfer<T>::construct(ref_, size, baseaddr_, phase_,
                from, to, ref_->is_real());
          }
        };

      // Dereferencing and array access.

      template<typename T, size_t B>
        class shared_reference
        {
        private:
          shared_ptr<T, B>* pointer_;

        public:

          shared_reference(shared_ptr<T, B>& ptr)
          {
            pointer_ = new shared_ptr<T, B> (ptr);
          }

          ~shared_reference()
          {
            delete pointer_;
          }

          T
          operator =(T val)
          {
            pointer_->put_val_message(val);
            return val;
          }

          operator T() const
          {
            return pointer_->get_val_message();
          }

          T*
          operator &()
          {
            return pointer_->operator T*();
          }

          template<typename U>
          bool
          operator ==(const U val) const
          {
            return pointer_->get_val() == val;
          }

          template<typename U>
          bool
          operator !=(const U val) const
          {
            return pointer_->get_val() != val;
          }

         /* template<typename U, size_t C>
          bool
          operator ==(const shared_ptr<U, C>& sptr) const
          {
            return pointer_->get_val() == sptr.get_val();
          }

          template<typename U, size_t C>
          bool
          operator !=(const shared_ptr<U, C>& sptr) const
          {
            return pointer_->get_val() != sptr.get_val();
          }*/

          template<typename U, size_t C>
            bool
            operator ==(const shared_reference<U, C>& sptr) const
            {
              return pointer_->get_val() == (sptr.pointer_->get_val());
            }

          template<typename U, size_t C>
            bool
            operator !=(const shared_reference<U, C>& sptr) const
            {
              return pointer_->get_val() != (sptr.pointer_->get_val());
            }
        };

      template<typename T, size_t B>
        debug shared_ptr<T, B>::dbg_;
    }
  }
}

#endif
