/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <sstmac/software/libraries/actor/simplemailserver-mbox.h>
#include <sstmac/software/libraries/actor/simplemboxapp.h>
#include <sstmac/software/process/key.h>
#include <boost/lexical_cast.hpp>
#include <algorithm>
#include <iostream>

namespace sstmac { namespace sw
{

  
  
  using boost::lexical_cast;

  simplembox::tagmap_t simplemailserver::mbox::tag_map_;

  //
  // Constructor.
  //
  simplemailserver::mbox::mbox(softwareid sid) :
         sid_(sid), active_(false), initialized_(false)
  {
    SSTMAC_DEBUG << "begin simplemailserver::mbox constructor \n";
    libname_ = "simplemailserver::mbox" + sid.to_string();
    SSTMAC_DEBUG << "simplemailserver::mbox constructed \n";
  }



  //
  // Allow the parent to send us a message to handle.
  //
  void
  simplemailserver::mbox::message_arrived(const actorid &sender,
      const actormsg::ptr& payload)
  {
    SSTMAC_DEBUG << "mbox[" << id_.to_string() << "]: handling actor msg from "
        << sender.to_string() << "\n";
    unhandled_.push_back(std::make_pair(sender, payload));
    // And finally test whether there is a pending receive for this payload.
    for (blocked_t::iterator it = blocked_.begin(); it != blocked_.end(); ++it)
      {
        if (it->first.first.matches(sender) && it->first.second->matches(payload))
          {
            key::ptr thekey = it->second;
            blocked_.erase(it);
            os_->unblock(thekey);
            break;
          }
      }
  }

  /* Methods from here on down are part of the simplembox interface */

  //
  // What's my pid.
  //
  actorid
  simplemailserver::mbox::mypid() const
  {
    return id_;
  }



  //
  // What nodes are known to the current system?
  //
  //const std::vector<nodeaddress::ptr >& simplemailserver::mbox::nodes() const {
  //    return parent_->nodes_;
  //}

  //
  // All peer identifiers (magically, we know about everybody).
  //
  const simplembox::world_t
  simplemailserver::mbox::world() const
  {
    // return parent_->magic_->world();
    return world_;
  }

  //
  // Register the given tag mapping in the registry.
  //
  void
  simplemailserver::mbox::set_tag(int key, actorid value)
  {
    tagmap_t &tm = tag_map_;
    if (tm.find(key) != tm.end())
      throw valueerror(std::string("simplemailserver::mbox::set_tag:  "
          "Attempt to redefine tag ") + lexical_cast<std::string>(key));
    tm[key] = value;
  }

  //
  // Test whether the given key is defined.
  //
  bool
  simplemailserver::mbox::has_tag(int key) const
  {
    const tagmap_t &tm = tag_map();
    return tm.find(key) != tm.end();
  }

  //
  // Access the value of the given key.
  //
  actorid
  simplemailserver::mbox::get_tag(int key) const
  {
    const tagmap_t &tm = tag_map();
    tagmap_t::const_iterator it = tm.find(key);
    if (it == tm.end())
      throw valueerror(std::string("simplemailserver::mbox::get_tag:  "
          "No match for the tag ") + lexical_cast<std::string>(key));
    return it->second;
  }

  //
  // Access the tag mapping in a const context.
  //
  const simplemailserver::mbox::tagmap_t&
  simplemailserver::mbox::tag_map() const
  {
    // return parent_->magic_->tagmap();
    return tag_map_;
  }



  //
  // Initialize.  This method is collective and "magic".
  //
  timestamp
  simplemailserver::mbox::init()
  {
    /* if(active_)
 throw illformederror("simplemailserver::mbox::init:  Already active\n");
 parent_->magic_->barrier(1, id_, parent_->magic_->world());
 return parent_->node_->now();*/


    library::ptr server = os_->get_lib(simplemailserver::lib_name());

    if (server == library::ptr()) // only do one mpiserver per node
      {
        server = simplemailserver::construct();
        os_->register_lib(server);
        simplemailserver::ptr sms = boost::dynamic_pointer_cast<simplemailserver>(server);
        sms->start();
      }
    parent_ = boost::dynamic_pointer_cast<simplemailserver>(server);

    id_ = actorid(os_->get_my_addr(), parent_->get_next_child(sid_.app_), sid_.app_);

    parent_->register_mbox(mbox::ptr(this));

    world_ = actorworld::construct(os_->get_env(id_.app()));

    initialized_ = true;

    return os_->now();
  }

  //
  // Finalize.  This is a magic operation.
  //
  timestamp
  simplemailserver::mbox::finalize()
  {
    /* if(active_)
 throw illformederror("simplemailserver::mbox::finalize:  Not active\n");
 parent_->magic_->barrier(2, id_, parent_->magic_->world());
 return parent_->node_->now();*/
    return os_->now();
  }

  //
  // Spawn the given task on the given node and return its pid.
  //
  actorid
  simplemailserver::mbox::spawn(nodeaddress::ptr &target,
      const boost::intrusive_ptr<simplemboxapp>& entity, int64_t communication_size)
  {
    SSTMAC_DEBUG << "simplemailserver::mbox::spawn \n";
    if (!entity)
      throw nullerror("simplemailserver::mbox::spawn:  NULL app pointer.");
    //simplembox::ptr the_mbox = parent_->new_mbox();
    //if (!the_mbox)
    //  throw illformederror("simplemailserver::mbox::spawn: Got NULL "
    //          "mbox from parent.");
    //shared_ptr<simplemboxapp> the_app = boost::dynamic_pointer_cast<
    //          simplemboxapp>(entity->clone(library::id_));
    //if (!the_app)
    //  throw illformederror(std::string("simplemailserver::mbox::spawn: "
    //          "'clone()' method returns NULL for ") + typeid(*entity).name());
    os_->start_thread(entity);
    return entity->mbox()->mypid();
  }

  //
  // Send a message to the given pid.
  //
  void
  simplemailserver::mbox::send(const actorid &target, actormsg::ptr msg,
      int64_t size)
  {
    if (!msg)
      throw nullerror("simplemailserver::mbox::send: "
          "NULL messages are not allowed.");
    parent_->send(id_, target, msg, size);
  }

  //
  // Test whether there is any message that matches the given pattern.
  bool
  simplemailserver::mbox::test(const actorid &source,
      const actorpattern::ptr& pat) const
  {
    unhandled_t::const_iterator it, end = unhandled_.end();
    for (it = unhandled_.begin(); it != end; ++it)
      {
        if (source.matches(it->first) && pat->matches(it->second))
          return true;
      }
    return false;
  }

  //
  // Block until we get a message matching the given pattern.
  //
  std::pair<actorid, actormsg::ptr >
  simplemailserver::mbox::recv(const actorid &source,
      const actorpattern::ptr& pat)
  {
    key::ptr thekey;
    std::pair<actorid, actormsg::ptr > retval;
    unhandled_t::iterator it;
    while (1)
      {
        for (it = unhandled_.begin(); it != unhandled_.end(); ++it)
          {
            if (source.matches(it->first) && pat->matches(it->second))
              {
                retval.first = it->first;
                retval.second = it->second;
                unhandled_.erase(it);
                break;
              }
          }
        if (!retval.second)
          {
            // We get here because there has been no match yet.
            if (!thekey)
              {
                thekey = key::construct();
                actorp_t matcher = std::make_pair(source, pat);
                blocked_.push_back(std::make_pair(matcher, thekey));
              }
            os_->block(thekey);
          }
        else
          {
            break;
          }
      }
    // and we get here when we have a match.
    // The match routine has already erased our 'blocked_' entry.
    return retval;
  }

} } // end of namespace sstmac
