/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMAILSERVER_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_ACTOR_SIMPLEMAILSERVER_H_INCLUDED

#include <sstmac/software/libraries/actor/actorid.h>
#include <sstmac/common/nodeaddress.h>
#include <sstmac/software/services/service.h>
#include <sstmac/software/libraries/actor/actormsg.h>
#include <sstmac/software/libraries/actor/simplembox.h>
#include <stdint.h>

namespace sstmac { namespace sw
{


  /**
   * A very simple dataflow process network server (one per node).
   */

  class simplemailserver : public service
  {
    /// The actual implementation of a simplembox -- we make this a private
    /// nested type to make it easier to safely manage internal state.
  public:
    class mbox;
    class data;

  protected:
    static debug dbg_;

  private:
    /// The identifiers of nodes we were given.  Stored in a vector to allow
    /// the user to query for nodes.
    std::vector<nodeaddress::ptr> nodes_;
    /// Nodes are also stored in a set to ensure that we only take valid ids.
    //type<nodeaddress::ptr >::set valid_nodes_;

    /// The node id that this server resides on.
    //  nodeaddress::ptr my_node_id_;

    /// The rank id that we will give to our next spawned process.
    std::map<int, int32_t> next_child_;

    /// The client processes we have (weak pointers, because clients are
    /// welcome to go out of scope any time they want to).
    typedef std::map<actorid, boost::intrusive_ptr<mbox> >  mboxes_t;
    mboxes_t mboxes_;


    /// Private method to send data.
    void
    send(const actorid &source, const actorid &target,
        actormsg::ptr& msg, int64_t size);

    /// Private constructor.
    simplemailserver() {

      if(!dbg_.is_real()){
           dbg_ = debug("<debug> actor | simplemailserver");
           logger::register_logger(&dbg_);
       }

      libname_ = lib_name();
    }

  public:
    typedef boost::intrusive_ptr<simplemailserver> ptr;

    /// Get us a new, heap-allocated object.
    static ptr
    construct(){
      return ptr(new simplemailserver());
    }

    /// Clean up.
    virtual
    ~simplemailserver() { }

    /// Create a new mbox interface that can talk to this object.
    simplembox::ptr
    new_mbox(softwareid sid);

    void
    register_mbox(boost::intrusive_ptr<mbox> m);

    virtual service::ptr
    clone() {
      return construct();
    }

    /* Interface required by the 'server' base type. */

    /// Complete a receive.
    /// Forwarded by the node when an incoming message arrives.
    /// It is the responsibility of the server to figure out how to
    /// understand any envelope information in the networkdata.
    virtual void
    handle(const sst_message::ptr &msg);

    int32_t
    get_next_child(appid aid)
    {
      if(next_child_.find(aid.id) == next_child_.end()){
          next_child_[aid.id] = 0;
      }
      return next_child_[aid.id]++;
    }

    static std::string
    lib_name(){
      return "simplemailserver";
    }

    virtual std::string
    to_string() const {
      return lib_name();
    }


  };

} } // end of namespace sstmac

#endif
