/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_CHECKPOINT_LIB_CHECKPOINT_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_CHECKPOINT_LIB_CHECKPOINT_H_INCLUDED

#include <sstream>

#include <sstmac/software/libraries/library.h>
#include <sstmac/software/process/softwareid.h>
#include <sstmac/common/sstmac-config.h>
#include <sstmac/software/libraries/filesystem/nppfs.h>

namespace sstmac
{
  namespace sw
  {

    class lib_checkpoint : public library
    {
      // ------- constructor / boost stuff -------------//

    protected:
      lib_checkpoint(softwareid id);
      lib_checkpoint(std::string id);

      class checkvar : public ptr_type
      {

      public:
        typedef boost::intrusive_ptr<checkvar> ptr;

        long filepos_;

        virtual void
        save() = 0;

        virtual size_t
        size() = 0;

        virtual std::string
        print() = 0;

        virtual std::string
        to_string() const
        {
          return "lib_checkpoint::checkvar";
        }
      };

      template<typename Type>
        class checktypedvar : public checkvar
        {
        protected:
          checktypedvar()
          {
          }
        public:
          typedef boost::intrusive_ptr<checktypedvar<Type> > ptr;
          static ptr
          construct()
          {
            return ptr(new checktypedvar);
          }
          virtual
          ~checktypedvar()
          {
          }
          Type* ptr_;
          Type saved_;
          virtual void
          save()
          {
            saved_ = (*ptr_);
          }
          virtual std::string
          print()
          {
            std::stringstream ss;
            ss << "val: " << saved_;
            return ss.str();
          }
          virtual size_t
          size()
          {
            return sizeof(Type);
          }
        };

      class checkfakevar : public checkvar
      {
      protected:
        checkfakevar()
        {
        }
      public:
        typedef boost::intrusive_ptr<checkfakevar> ptr;
        static ptr
        construct()
        {
          return ptr(new checkfakevar);
        }
        virtual
        ~checkfakevar()
        {
        }
        size_t size_;
        virtual void
        save()
        {
        }
        virtual size_t
        size()
        {
          return size_;
        }
        virtual std::string
        print()
        {
          std::stringstream ss;
          ss << "size: " << size_;
          return ss.str();
        }

      };
      typedef boost::unordered_map<std::string, checkvar::ptr> maptype;
      maptype vars_;
      long check_id_;

      nppfs::ptr fs_;
      openfile::ptr file_;
      long current_filepos_;

      softwareid id_;
    public:
      typedef boost::intrusive_ptr<lib_checkpoint> ptr;

      virtual
      ~lib_checkpoint();

      static ptr
      construct(softwareid id);

      static ptr
      construct(std::string str);
      // --------------------------------------//

    public:

      long
      get_check_id() const
      {
        return check_id_;
      }

      bool
      restarted()
      {
        return restarted_;
      }

      virtual void
      register_child_libs()
      {

        fs_ = nppfs::construct("nppfs-" + libname_);
        register_lib(fs_);
      }

      template<typename Type>
        void
        register_var(std::string name, Type* val)
        {

          SSTMAC_DEBUG << "lib_checkpoint[" << os_->get_my_addr()->to_string()
                << "]: register_var(" << name << ", " << val << ")\n";

          if (!file_)
            file_ = fs_->create("temp-checkpoint" + id_.to_string());

          if (check_id_ == 0)
            {
              boost::intrusive_ptr<checktypedvar<Type> > var = checktypedvar<
                  Type>::construct();
              var->ptr_ = val;
              var->saved_ = *val;
              var->filepos_ = current_filepos_;
              current_filepos_ += sizeof(Type);
              file_->seek(sizeof(Type));
              vars_[name] = (var);
            }
          else
            { //load saved variables
              SSTMAC_DEBUG << "lib_checkpoint["
                    << os_->get_my_addr()->to_string()
                    << "]: recovering from checkpoint " << check_id_ << "\n";

              if (vars_.find(name) == vars_.end())
                {
                  throw ssterror(
                      "lib_checkpoint[" + os_->get_my_addr()->to_string()
                          + "]: can't find var " + name);
                }

              boost::intrusive_ptr<checktypedvar<Type> >
                  v = boost::dynamic_pointer_cast<checktypedvar<Type> >(
                      vars_[name]);
              //get it from disk
              SSTMAC_DEBUG << "reading file of size " << sizeof(Type)
                    << " at pos " << v->filepos_ << "\n";
              file_->clear_offset();
              fs_->seek(file_, v->filepos_);
              fs_->read(file_, sizeof(Type));

              //now assign value
              SSTMAC_DEBUG << "lib_checkpoint["
                    << os_->get_my_addr()->to_string()
                    << "]: setting value to " << v->saved_ << "\n";
              *val = v->saved_;
              v->ptr_ = val; //update to new ptr
            }
        }

      template<typename Type>
        void
        register_fake(std::string name, size_t size)
        {

          SSTMAC_DEBUG << "lib_checkpoint[" << os_->get_my_addr()->to_string()
                << "]: register_fake(" << name << ", " << size << ")\n";

          if (!file_)
            file_ = fs_->create("temp-checkpoint" + id_.to_string());

          if (check_id_ == 0)
            {
              checkfakevar::ptr var = checkfakevar::construct();
              var->size_ = size * sizeof(Type);
              var->filepos_ = current_filepos_;
              current_filepos_ += size * sizeof(Type);
              vars_[name] = (var);
              file_->seek(size * sizeof(Type));
            }
          else
            { //load saved variables
              SSTMAC_DEBUG << "lib_checkpoint["
                    << os_->get_my_addr()->to_string()
                    << "]: recovering from checkpoint " << check_id_ << "\n";

              if (vars_.find(name) == vars_.end())
                {
                  throw ssterror(
                      "lib_checkpoint[" + os_->get_my_addr()->to_string()
                          + "]: can't find var " + name);
                }

              boost::intrusive_ptr<checkfakevar> v =
                  boost::dynamic_pointer_cast<checkfakevar>(vars_[name]);

              //get it from disk
              SSTMAC_DEBUG << "reading file of size " << v->size()
                    << " at pos " << v->filepos_ << "\n";
              file_->clear_offset();
              fs_->seek(file_, v->filepos_);
              fs_->read(file_, v->size());
            }
        }

      void
      checkpoint()
      {
        if (!file_)
          file_ = fs_->create("temp-checkpoint" + id_.to_string());

        maptype::iterator iter, end = vars_.end();
        file_->clear_offset();
        for (iter = vars_.begin(); iter != end; iter++)
          {
            fs_->write(file_, iter->second->size());
            fs_->seek(file_, iter->second->size());
            (iter)->second->save();

            SSTMAC_DEBUG << "lib_checkpoint: saved variable " << iter->first
                  << ": " << iter->second->print() << "\n";
          }
      }

      /**
       * Only call this function when all tasks have passed the checkpoint
       */
      void
      finalize_checkpoint()
      {
        check_id_++;
      }

      virtual bool
      supported() const
      {
        return os_->kernel_supported(AMI::AMI_COMM_SEND);
      }

      void
      restart()
      {
        restarted_ = true;
      }
    protected:

      static debug dbg_;

      bool restarted_;
    };

  }
} //end of namespace sstmac

#endif
