/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <math.h>

#include <sstmac/software/libraries/compute/lib_compute_loops.h>

#include <sstmac/software/process/key.h>
#include <sstmac/common/sstmac_env.h>
#include <stdint.h>

namespace sstmac
{
  namespace sw
  {

    int lib_compute_loops::mem_op_size_ = 16;

    double lib_compute_loops::mem_line_ratio_ = -1;
    double lib_compute_loops::flop_line_ratio_ = -1;
    bool lib_compute_loops::do_loops_ = true;

    debug lib_compute_loops::loopsdbg_;

      lib_compute_loops::lib_compute_loops(softwareid id) :
        lib_compute_inst(id),
        lib_compute_memmove(id)
      {
        libname_ = "computelibloops" + id.to_string();
        key_cat_ = libcompute::key_category;

        check_init();

        if (!loopsdbg_.is_real())
        {
          loopsdbg_ = sstmac::debug("<debug> lib_compute_loops | loops",
              sstmac::logger::red);
          sstmac::logger::register_logger(&loopsdbg_);
        }
      }

      lib_compute_loops::lib_compute_loops(const std::string& id) :
        lib_compute_inst(id),
        lib_compute_memmove(id)
      {
        libname_ = id;
        key_cat_ = libcompute::key_category;

        if (!loopsdbg_.is_real())
        {
          loopsdbg_ = sstmac::debug("<debug> lib_compute_loops | loops",
              sstmac::logger::red);
          sstmac::logger::register_logger(&loopsdbg_);
        }
      }

    void
    lib_compute_loops::check_init()
    {
      if (mem_line_ratio_ < 0 && sstmac_env::params)
      {
        mem_line_ratio_ = sstmac_env::params->get_optional_double_param(
            "lib_compute_loops_mem_ratio", 0.8);
        flop_line_ratio_ = sstmac_env::params->get_optional_double_param(
            "lib_compute_loops_flop_ratio", 0.8);
        do_loops_ = sstmac_env::params->get_optional_bool_param(
            "lib_compute_loops_enable", true);
      }
    }

    void
    lib_compute_loops::compute_loop_work(long long loop, double numlines,
        long tid)
    {
      SSTMACBacktrace("Compute Loops");
      if (loop < 0)
      {
        loop = 0;
      }

      if (numlines < 0)
      {
        numlines = 0;
      }


      if (loopsdbg_.is_active())
      {
        loopsdbg_ << "compute_loops (" << get_lib_name() << "): loop: " << loop
            << ", numlines: " << numlines << ", tid: " << tid << "\n";
      }

      doing_memory_ = true;
      long long bytes = std::max((long long) 1,
          (long long) (loop * mem_line_ratio_ * numlines * mem_op_size_));
      lib_compute_memmove::read(bytes, tid);
      doing_memory_ = false;

      eventdata::ptr inst = eventdata::construct(get_lib_name());
      long long flops = std::max((long long) 1,
          (long long) (loop * flop_line_ratio_ * numlines));
      inst->set_event_value(eventdata::INST_FLOP, flops);
      lib_compute_inst::compute_inst(inst);
    }



    void
    lib_compute_loops::compute_fft()
    {
      check_init();
      if (do_loops_)
      {
        eventdata::ptr inst = eventdata::construct(get_lib_name());
        inst->set_event_value(eventdata::INST_FLOP, 500);
        lib_compute_inst::compute_inst(inst);
      }
    }

  }
} //end of namespace sstmac
