/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/graphviz.h>
#include <sstmac/software/libraries/compute/lib_compute_memmove.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/libraries/compute/lib_compute_inst.h>
#include <sstmac/software/libraries/compute/mem_access.h>
#include <sstmac/software/libraries/libwaiters.h>
#include <sstmac/common/eventcallback.h>

namespace sstmac
{
  namespace sw
  {

    debug lib_compute_memmove::dbg_;

    key::category lib_compute_memmove::key_category("Memory");

    lib_compute_memmove::lib_compute_memmove(softwareid id)
        : libcompute()
    {
      libname_ = get_name(id.to_string());
      key_cat_ = key_category;
      init();
    }

    lib_compute_memmove::lib_compute_memmove(const std::string& libname)
        : libcompute()
    {
      libname_ = get_name(libname);
      key_cat_ = key_category;
      init();
    }

    void
    lib_compute_memmove::init()
    {
        if (!dbg_.is_real())
        {
            dbg_ = debug("<debug> lib_compute_memmove | lib_memory");
            logger::register_logger(&dbg_);
        }

        compute_cb_ = msg_callback(ptr(this), &lib_compute_memmove::compute);

        finalize_cb_ = msg_callback(ptr(this), &lib_compute_memmove::finish);

        rdma_cb_ = msg_callback(ptr(this), &lib_compute_memmove::finish_rdma);
    }

    void
    lib_compute_memmove::finish_rdma(const sst_message::ptr &cmsg)
    {
        SSTMAC_DEBUG
              << "lib_compute_memmove: finish rdma mem access kernel for msg "
              << (long) cmsg.get() << ":"
              << " func=" << AMI::tostr(cmsg->get_function())
              << " bytes=" << cmsg->get_byte_length()
              << " thread=" << fullos_->get_current_threadid()
              << "\n";
        library::incoming_message(cmsg);
    }

    void
    lib_compute_memmove::finish(const sst_message::ptr& cmsg)
    {
        SSTMAC_DEBUG
              << "lib_compute_memmove: finish local mem access kernel for msg "
              << (long) cmsg.get() << ":"
              << " func=" << AMI::tostr(cmsg->get_function())
              << " bytes=" << cmsg->get_byte_length()
              << " thread=" << fullos_->get_current_threadid()
              << "\n";
        os_->execute_kernel(AMI::AMI_COMP_RELEASE, cmsg);
        library::incoming_message(cmsg);
    }

    void
    lib_compute_memmove::compute(const sst_message::ptr& cmsg)
    {
        long bytes = cmsg->get_byte_length();
        SSTMAC_DEBUG
              << "lib_compute_memmove: compute local mem access kernel for msg "
              << (long) cmsg.get() << ":"
              << " func=" << AMI::tostr(cmsg->get_function())
              << " bytes=" << cmsg->get_byte_length()
              << " thread=" << fullos_->get_current_threadid()
              << "\n";

        long memmove_ninstr = bytes / sizeof(int64_t);
        long loop_ninstr = 2 * bytes / loop_unroll_ / sizeof(int64_t);
        long ninstr = (memmove_ninstr + loop_ninstr) / pipeline_efficiency_;
        cmsg->set_event_value(eventdata::INST_MEMOP_LOCAL, ninstr);
        cmsg->add_callback(finalize_cb_);
        //first compute the length of time necessary to execute the instructions
        //this does not actually increment any time values or schedule any events
        //it just computes a certain amount of time and stores the value in cmsg
        os_->execute_kernel(AMI::AMI_COMP_INSTR_TIME, cmsg);
        //now actually go ahead and execute the memory access
        os_->execute_kernel(cmsg->get_function(), cmsg);
    }

    void
    lib_compute_memmove::reserve_cpu(const sst_message::ptr& cmsg)
    {
        SSTMAC_DEBUG
              << "lib_compute_memmove: reserve cpu for mem access for msg "
              << (long) cmsg.get() << ":"
              << " func=" << AMI::tostr(cmsg->get_function())
              << " bytes="
              << cmsg->get_byte_length()
              << " thread=" << fullos_->get_current_threadid()
              << "\n";
        cmsg->add_callback(compute_cb_);
        os_->execute_kernel(AMI::AMI_COMP_SCHEDULE, cmsg);
    }

    void
    lib_compute_memmove::do_access(
            long bytes,
            AMI::AMI_COMP_FUNC func,
            long srcthread,
            long dstthread,
            const eventcallback::ptr& cb
    )
    {
        if (!supported())
            throw valueerror("lib_compute_memmove requires the presence of a memory model");

        mem_access::ptr cmsg = mem_access::construct(bytes, get_lib_name(),
                                                     func,
                                                     srcthread, dstthread);

        SSTMAC_DEBUG
              << "lib_compute_memmove: start memory access "
              << AMI::tostr(func)
              << " on node " << fullos_->get_my_addr()
              << " for msg "<< (long) cmsg.get() << ":"
              << " bytes=" << cmsg->get_byte_length()
              << " thread=" << fullos_->get_current_threadid()
              << "\n";

        if (cb){
            event_step_callback::ptr page_event_cb = boost::dynamic_pointer_cast<event_step_callback>(cb);
            if (page_event_cb){
                cmsg->set_page_event_callback(page_event_cb);
            }
            cmsg->add_callback(cb);
        }
        reserve_cpu(cmsg);
        wait_this(cmsg);
    }

    void
    lib_compute_memmove::unregister_all_libs()
    {
        //clear all the callbacks to avoid circular references
        compute_cb_ = 0;
        nic_cb_ = 0;
        finalize_cb_ = 0;
        rdma_cb_ = 0;
        library::unregister_all_libs();
    }

    void
    lib_compute_memmove::read(long bytes, const eventcallback::ptr& cb)
    {
        read(bytes, os_->get_current_threadid(), cb);
    }

    void
    lib_compute_memmove::read(long bytes, long srcthread, const eventcallback::ptr& cb)
    {
        SSTMACBacktrace("memread");
        do_access(bytes, AMI::AMI_COMP_MEMREAD, srcthread, srcthread, cb);
    }

    void
    lib_compute_memmove::read(long bytes, long srcthread, long dstthread, const eventcallback::ptr& cb)
    {
        SSTMACBacktrace("memread");
        do_access(bytes, AMI::AMI_COMP_MEMREAD, srcthread, dstthread, cb);
    }

    void
    lib_compute_memmove::write(long bytes, const eventcallback::ptr& cb)
    {
        write(bytes, os_->get_current_threadid(), cb);
    }

    void
    lib_compute_memmove::write(long bytes, long dstthread, const eventcallback::ptr& cb)
    {
        SSTMACBacktrace("memwrite");
        do_access(bytes, AMI::AMI_COMP_MEMWRITE, dstthread, dstthread, cb);
    }

    void
    lib_compute_memmove::write(long bytes, long srcthread, long destthread, const eventcallback::ptr &cb)
    {
        SSTMACBacktrace("memwrite");
        do_access(bytes, AMI::AMI_COMP_MEMWRITE, srcthread, destthread, cb);
    }

    void
    lib_compute_memmove::copy(long bytes, const eventcallback::ptr& cb)
    {
        long thr = os_->get_current_threadid();
        copy(bytes, thr, thr, cb);
    }

    void
    lib_compute_memmove::copy(long bytes, long srcthread, long dstthread, const eventcallback::ptr& cb)
    {
        SSTMACBacktrace("memcpy");
        do_access(bytes, AMI::AMI_COMP_MEMWRITE, srcthread, dstthread, cb);
    }

    void
    lib_compute_memmove::consume_params(const sim_parameters::ptr &params)
    {
        pipeline_efficiency_ = params->get_optional_double_param("lib_compute_pipeline", 1.0);
        loop_unroll_ = params->get_optional_double_param("lib_compute_unroll_loops", 1.0);
        prefetch_ = params->get_optional_bool_param("lib_compute_prefetch_loops", false);
    }

    void
    lib_compute_memmove::read(long bytes, const taskid &srctask, const eventcallback::ptr &cb)
    {
        long srcthread = fullos_->get_task_threadid(srctask);
        do_access(bytes, AMI::AMI_COMP_MEMREAD, srcthread, thread::nic_thread, cb);
    }

    void
    lib_compute_memmove::write(long bytes, const taskid &dsttask, const eventcallback::ptr &cb)
    {
        long dstthread = fullos_->get_task_threadid(dsttask);
        do_access(bytes, AMI::AMI_COMP_MEMWRITE, thread::nic_thread, dstthread, cb);
    }

    bool
    lib_compute_memmove::supported() const
    {
        return os_->kernel_supported(AMI::AMI_COMP_MEMREAD);
    }

  }
} //end of namespace sstmac
