/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#ifndef SSTMAC_SOFTWARE_LIBRARIES_FILESYSTEM_FILE_METADATA_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_FILESYSTEM_FILE_METADATA_H_INCLUDED

#include <sstream>

#include <sstmac/hardware/network/network_message.h>

namespace sstmac
{
  namespace sw
  {

    class file_metadata :
        public library_message,
        public hw::network_message
    {
    public:
      enum FILE_OPS
      {
        CREATE,
        CREATE_RESP,
        OPEN,
        OPEN_RESP,
        CLOSE,
        READ,
        READ_RESP,
        WRITE,
        WRITE_RESP
      };

    protected:
      file_metadata(const std::string& libn, const std::string& fname,
          FILE_OPS op) :
            sst_message(
                (op == CREATE || op == OPEN || op == CLOSE) ? sst_message::FILEMETA
                    : sst_message::FILEDATA), library_message(libn), op_(op),
            fname_(fname), bytes_(0), write_resp_(false)
      {

      }

    public:

      typedef boost::intrusive_ptr<file_metadata> ptr;

      static ptr
      construct(const std::string& libn, const std::string& fname, FILE_OPS op)
      {
        return ptr(new file_metadata(libn, fname, op));
      }

      virtual sst_message::ptr
      clone() const
      {
        ptr ret = construct(lib_name_, fname_, op_);
        ret->data_ = data_;
        ret->bytes_ = bytes_;
        ret->access_size_ = access_size_;
        ret->from_lib_ = from_lib_;
        ret->write_resp_ = write_resp_;
        ret->src_task_ = src_task_;
        ret->dest_task_ = dest_task_;
        return ret;

      }

      virtual sst_message::ptr
      clone(MESSAGE_TYPES t) const
      {
        sst_message::ptr ret = clone();
        ret->set_type(t);
        return ret;
      }

      virtual
      ~file_metadata()
      {
      }

      virtual long
      get_byte_length() const
      {
        return bytes_;
      }

      virtual const sw::taskid&
      source_task() const
      {
        return src_task_;
      }

      virtual const sw::taskid&
      dest_task() const
      {
        return dest_task_;
      }

      FILE_OPS op_;
      std::string fname_;
      ptr_type::ptr data_;
      long bytes_;
      long access_size_;
      std::string from_lib_;
      bool write_resp_;
      taskid src_task_;
      taskid dest_task_;

      virtual std::string
      to_string() const
      {
        std::stringstream ss;
        ss << "file_metadata( op = " << op_ << ", fname = " << fname_
            << ", bytes = " << bytes_ << ", access_size " << access_size_
            << ", from_lib = " << from_lib_ << ")";
        return ss.str();
      }
    };

  }
}/* namespace sstmac */
#endif /* FILE_METADATA_H_ */
