/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_LIBRARIES_FILESYSTEM_NPPFS_H_INCLUDED
#define SSTMAC_SOFTWARE_LIBRARIES_FILESYSTEM_NPPFS_H_INCLUDED

#include <sstmac/software/services/service.h>
#include <sstmac/software/libraries/filesystem/openfile_parallel.h>
#include <sstmac/software/libraries/filesystem/file_metadata.h>

namespace sstmac
{
  namespace sw
  {

    /**
     * nppfs stands for No Particular Parallel File System,
     * an approximate implementation of a file system which
     * can have files spread out among different nodes.
     */
    class nppfs : public library
    {


    protected:
      nppfs(softwareid id)
      {

        libname_ = "nppfs" + id.to_string();

        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> nppfs | nppfs_client");
            logger::register_logger(&dbg_);
          }
      }

      nppfs(std::string name)
      {

        libname_ = name;

        if (!dbg_.is_real())
          {
            dbg_ = debug("<debug> nppfs | nppfs_client");
            logger::register_logger(&dbg_);
          }
      }

      static debug dbg_;

      class node_access : public ptr_type
      {
      protected:
        node_access(nodeaddress::ptr addr, long offset, long sector, long bytes) :
          addr_(addr), offset_(offset), sector_(sector), bytes_(bytes)
        {

        }

      public:
        typedef boost::intrusive_ptr<node_access> ptr;
        static ptr
        construct(nodeaddress::ptr addr, long offset, long sector, long bytes)
        {
          return ptr(new node_access(addr, offset, sector, bytes));
        }
        virtual std::string
        to_string() const
        {
          return "node_access";
        }

        nodeaddress::ptr addr_;
        long offset_;
        long sector_;
        long bytes_;
      };

      std::vector<node_access::ptr>
      get_access_nodes(openfile_parallel::ptr file, long bytes);

    public:
      typedef boost::intrusive_ptr<nppfs> ptr;

      static ptr
      construct(softwareid id)
      {
        return ptr(new nppfs(id));
      }

      static ptr
      construct(std::string id)
      {
        return ptr(new nppfs(id));
      }

      virtual
      ~nppfs()
      {
      }

      openfile::ptr
      create(std::string name);

      openfile::ptr
      open(std::string name);

      void
      read(openfile::ptr file, long bytes);
      void
      write(openfile::ptr file, long bytes, bool wait_complete = false);

      void
      seek(openfile::ptr file, long bytes);

      void
      close(openfile::ptr file);

      virtual bool
      supported() const;

      openfile::ptr
      process_openfile_meta(sst_message::ptr msg);

    };

  }
} /* namespace sstmac */
#endif /* NPPFS_H_ */
