#include "libtopomap.hpp"

/* greedily maps a logical topology graph (ltg) to a physical topology
 * graph (ptg) with nprocs[i] processors per vertex i. Results in a
 * mapping from ltg to ptg in mapping. The greedy algorithm begins at
 * the vertex start and maps routing-aware (more expensive than without routing) */
int TPM_Map_greedy_routing(int r, TPM_Graph *ptg_ref, TPM_Graph_ewgts *ptgc_ref, TPM_Routes *routes_ref, TPM_Graph_vwgts *nprocs_ref, TPM_Graph *ltg_ref, TPM_Graph_ewgts *weights_ref, TPM_Mapping *mapping_ref, int start) {
  // turn refs into normal objects
  TPM_Graph& ptg = *ptg_ref; // physical topology graph
  TPM_Graph_ewgts& ptgc = *ptgc_ref; // ptg edge capacities!
  TPM_Graph& ltg = *ltg_ref; // logical topology graph
  TPM_Graph_ewgts& ltgw = *weights_ref; // ltg edge weights!
  TPM_Graph_vwgts& nprocs = *nprocs_ref; // number of PEs per node in ptg
  TPM_Mapping& mapping = *mapping_ref; // returned mapping
  TPM_Routes& routes = *routes_ref; // routes in h

  // number of slots in ptg (nprocs) must be identical to number of
  // nodes in ltg! This is just a correctness check!
#ifdef CHECKS
  if(std::accumulate(nprocs.begin(), nprocs.end(), 0) != ltg.size()) {
    if(!r) printf("number of slots does not match (nprocs: %i, ltg.size(): %lu!\n", std::accumulate(nprocs.begin(), nprocs.end(), 0), ltg.size());
    MPI_Abort(MPI_COMM_WORLD, 5);
  }
#endif

//printf("ltgw: %i, ltg:%i \n", ltgw.size(), ltg.size());

  TPM_Graph_dewgts wgt(ptg.size());
  double init = 1;
  if(routes.type == SSSP) {
    // init edge weights in ptg to max(nprocs)*n^2 to force shortest paths in SSSP!
    int maxw=1;
    for(int i=0; i<ltgw.size(); ++i) if(ltgw[i].size() > 0) maxw=std::max(maxw, *std::max_element(ltgw[i].begin(), ltgw[i].end()) );
    init = (*std::max_element(nprocs.begin(), nprocs.end())) * maxw * ptg.size()*ptg.size();
    assert(init > 0); // check for overflow
  }
  for(int i=0; i<ptg.size(); ++i) wgt[i].resize(ptg[i].size(), init);

  TPM_Graph_dists ptdists(ptg.size());
  TPM_Graph_preds ptpreds(ptg.size());

  std::vector<bool> mapped(ltg.size(), false);
  std::vector<bool> discovered(ltg.size(), false);

  /********************************************************
   * The actual mapping code
   ********************************************************/

  int current=start; // the current vertex in ptg to be mapped to
  int num_map=0; // the number of mapped vertices in ltg
  while(num_map < ltg.size()) {
    /* find vertex in ltg with most edges (most edge weights) that is
     * not mapped yet and map it to next free vertex in ptg (starting
     * with myself ) */
    // TODO: linear for now, should be done with sort ...
    int maxEdgeSum=-1;
    int maxvert=-1;
    for(int i=0; i<ltg.size(); ++i) {
      if(mapped[i]) continue;
      int thisEdgeSum=0;
      for(int j=0; j<ltg[i].size(); ++j) {
        assert(i<ltgw.size() && j<ltgw[i].size());
        thisEdgeSum+=ltgw[i][j]; // edge weight of logical edge
      }
      //printf("%i -- %i\n", i, ltg[i].size());
      if(thisEdgeSum > maxEdgeSum) {
        maxEdgeSum=thisEdgeSum;
        maxvert=i;
      }
    }

    DBG4(if(!r) printf("most 'expensive' vertex in logical topology graph: %i (cost: %i)\n", maxvert, maxEdgeSum));

    // if current vertex in ptg does not have free slots, choose different
    // (pretty much random) one -- should this pick a particular one?
    while(nprocs[current] <= 0) { current = (current+1)%nprocs.size(); }

    DBG4(if(!r) printf("mapped %i to %i\n", maxvert, current));
    discovered[maxvert] = true;
    mapped[maxvert] = true;
    num_map++;
    mapping[maxvert] = current; // map rank maxvert to current
    nprocs[current]--; // one of the procs of current is occupied now

    /* add all edges of this vertex that lead to non-discovered vertices
     * to prio queue */
    std::priority_queue<prioQedge_t, std::vector<prioQedge_t>, prioQedge_compare_func > Q;
    for(int j=0; j<ltg[maxvert].size(); ++j) {
      if(!discovered[ltg[maxvert][j]]) {
        prioQedge_t edge;
        edge.src = maxvert; edge.dst = ltg[maxvert][j]; edge.srcmapped = current; edge.weight=ltgw[maxvert][j];
        Q.push(edge);
        discovered[ltg[maxvert][j]] = 1;
      }
    }

    /* while Q not empty find most expensive edge in queue */
    while(!Q.empty()) {
      // take the most expensive edge out of the queue
      prioQedge_t Qel = Q.top();
      Q.pop();
      int target = Qel.dst;

      DBG4(if(!r) printf("most expensive edge in topology graph leads to: %i (cost: %f)\n", target, Qel.weight));

      /* try mapping to other close vertices and compute the cost using routing
       * the vertex to check mapping is selected with a BFS and the number of
       * vertices to check can be adjusted by maxCheckVert
       */
      int maxCheckVert=100; // maximum route checks
      // start a BFS in ptg at current
      std::queue<int> bfsQ;
//      bfsQ.push(current);
      bfsQ.push(Qel.srcmapped); // htor change: use the vertex where the source of the heaviest edge is mapped to as source in ptg!

      double minMaxC = std::numeric_limits<double>::max();
      int minMaxMappingDest = -1;
      std::vector<bool> visited(ptg.size(), false); // BFS visited map

      while(!bfsQ.empty() && maxCheckVert > 0) {
        int bfsCurrent = bfsQ.front();
        bfsQ.pop();
        //if(!r) printf("[%i] in BFS. mapped: %i/%i, size: %i, maxCheckVert: %i\n", r, num_map, ltg.size(), bfsQ.size(), maxCheckVert);

        // add all neighbors to bfsQ
        for(int j=0; j<ptg[bfsCurrent].size(); ++j) {
          if(!visited[ptg[bfsCurrent][j]]) {
              bfsQ.push(ptg[bfsCurrent][j]);
              visited[ptg[bfsCurrent][j]] = true;
          }
        }

        // check route to bfsCurrent vertex (if it has available PEs)!
        if(nprocs[bfsCurrent] > 0) {
          maxCheckVert--; // we checked one more

          // get the route from the routing module
          std::vector<int> path;
          int src = current;
          int tgt = bfsCurrent;
          TPM_RouteGet(&ptg, &wgt, &routes, src, tgt, &path);

          int dilation = 0;
          double maxc = 0;
          int idx = 0;

          int v=src;
          while(v != tgt) {
            dilation++;
            int w=path[idx]; // next hop, we're looking at edge v -> w right now
            idx++;

            // find position of edge in weights map (same as on ptg)
            assert(ptg.size() > v);
            int pos; for(pos=0; ptg[v][pos] != w; pos++) if(pos >= ptg[v].size()) { printf("[%i] edge %i->%i not found\n", r, v, w); break; }
    //if(!r) printf(".. step: %i, src: %i, dst: %i, pos: %i, cost: %.2f\n", dilation, v, w, pos, wgt[v][pos]);
            assert(wgt.size() > v);
            assert(wgt[v].size() > pos);

            //wgt[v][pos] += (double)Qel.first/ptgc[v][pos]; // add weight of logical edge divided by capacity of physical edge
            // don't add the weight to wgt but pretend one did!
            maxc=std::max(wgt[v][pos] + (double)Qel.weight/ptgc[v][pos] - init, maxc);

            //if(print) printf("increasing edge weight on edge %i -> %i to %i\n", v, w, wgt[v][pos]-init);
            v = w;
          }

          // if we found a better target -- swap!
          if(maxc < minMaxC) {
            minMaxC = maxc;
            minMaxMappingDest = bfsCurrent;
          }
        }
      } // end BFS

      // map most expensive process to the nearest physical neighbor!
      DBG4(if(!r) printf("mapped %i to %i\n", target, smallestdistvert));
      nprocs[minMaxMappingDest]--;
      mapping[target] = minMaxMappingDest;
      mapped[target] = true;
      num_map++;

      // update occupied edges in route in graph
      std::vector<int> path;
      int src = current;
      int tgt = minMaxMappingDest;
      TPM_RouteGet(&ptg, &wgt, &routes, src, tgt, &path);

      int idx=0;
      int v=src;
      while(v != tgt) {
        int w=path[idx]; // next hop, we're looking at edge v -> w right now
        idx++;

        assert(ptg.size() > v);
        int pos; for(pos=0; ptg[v][pos] != w; pos++) if(pos >= ptg[v].size()) { printf("[%i] edge %i->%i not found\n", r, v, w); break; }
//if(!r) printf(".. step: %i, src: %i, dst: %i, pos: %i, cost: %.2f\n", dilation, v, w, pos, wgt[v][pos]);
        assert(wgt.size() > v);
        assert(wgt[v].size() > pos);
        // now really add the weight of the path selected
        wgt[v][pos] += (double)Qel.weight/ptgc[v][pos];
        v = w;
      }

      /* add all edges of just mapped vertex (target) that lead to
       * not-discovered vertices to prio queue */
      for(int j=0; j<ltg[target].size(); ++j) {
        if(!discovered[ltg[target][j]]) {
          prioQedge_t edge;
          edge.src = target; edge.dst = ltg[target][j]; edge.srcmapped = minMaxMappingDest; edge.weight=ltgw[target][j];
          Q.push(edge);
          discovered[ltg[target][j]] = 1;
        }
      }
    } // while !Q.empty()
  }
  return 0;
}
