#include "libtopomap.hpp"

#ifdef HAVE_SCOTCH
#include <scotch.h>
#endif /* HAVE_SCOTCH */

extern "C" {
#ifdef HAVE_SCOTCH
  int                         SCOTCH_archInit     (SCOTCH_Arch * const);
  void                        SCOTCH_archExit     (SCOTCH_Arch * const);
  int                         SCOTCH_archBuild    (SCOTCH_Arch * const, const SCOTCH_Graph * const, const SCOTCH_Num, const SCOTCH_Num * const, const SCOTCH_Strat * const);
  int                         SCOTCH_graphInit    (SCOTCH_Graph * const);
  void                        SCOTCH_graphExit    (SCOTCH_Graph * const);
  int                         SCOTCH_graphMap     (SCOTCH_Graph * const, const SCOTCH_Arch * const, SCOTCH_Strat * const, SCOTCH_Num * const);
  int                         SCOTCH_stratInit    (SCOTCH_Strat * const);
  void                        SCOTCH_stratExit    (SCOTCH_Strat * const);
#endif
}

int TPM_Map_scotch(int r, TPM_Graph *ptg_ref, TPM_Graph_ewgts *ptgc_ref, TPM_Graph_vwgts *nprocs_ref, TPM_Graph *ltg_ref, TPM_Graph_ewgts *weights_ref, TPM_Mapping *mapping_ref, int me, MPI_Comm topocomm) {
  // turn refs into normal objects
 TPM_Graph& ptg = *ptg_ref; // the physical graph topology as adjacency list
  TPM_Graph& ptgc = *ptgc_ref; // the edge capacities in the physical graph topology
  TPM_Graph& ltg = *ltg_ref; // the logical graph topology as adjacency list
  TPM_Graph& ltgw = *weights_ref; // the edge weights in the logical graph topology
  TPM_Graph_vwgts& nprocs = *nprocs_ref; // the number of CPUs per node in the physical topology (enables heterogeneous systems)
  TPM_Mapping& mapping = *mapping_ref; // the mapping (output)

#ifdef HAVE_SCOTCH
  SCOTCH_Strat        logsdat;                    // Graph mapping strategy (default of its kind)
  SCOTCH_Strat        physdat;                    // Architecture building strategy (default of its kind)
  SCOTCH_Arch         phyadat;                    // Physical graph architecture
  SCOTCH_Graph        phygdat;                    // Physical graph data
  SCOTCH_Graph        loggdat;                    // Logical graph

  SCOTCH_archInit  (&phyadat);                    // Initialize architecture structure

  int edgenbr=0;
  for(int i=0; i<ptg.size(); ++i) edgenbr += ptg[i].size();
  std::vector<int> edgetab(edgenbr);
  std::vector<int> edlotab(edgenbr);
  std::vector<int> verttab(ptg.size()+1);
  int iter=0;
  verttab[0] = 0;
  for(int i=0; i<ptg.size(); ++i) {
    std::copy(ptg[i].begin(), ptg[i].end(), &edgetab[iter]);
    iter += ptg[i].size();
    verttab[i+1]=iter;
  }

  {
    assert(ptg.size() == ptgc.size());
    iter=0;
    for(int i=0; i<ptgc.size(); ++i) { assert(ptg[i].size() == ptgc[i].size()); std::copy(ptgc[i].begin(), ptgc[i].end(), &edlotab[iter]); iter += ptgc[i].size(); }

    SCOTCH_graphInit  (&phygdat);

    /*
    printf("verttab (%i): ", ptg.size());
    for(int i=0; i<verttab.size(); ++i) { printf("%i ", verttab[i]); } printf("\n");
    printf("edgetab (%i): ", edgenbr);
    for(int i=0; i<edgetab.size(); ++i) { printf("%i ", edgetab[i]); } printf("\n");
    printf("edlotab: ");
    for(int i=0; i<edlotab.size(); ++i) { printf("%i ", edlotab[i]); } printf("\n");*/

    SCOTCH_graphBuild (&phygdat,
      0, // const SCOTCH_Num            baseval,              /* Base value                          */
      ptg.size(), //const SCOTCH_Num            vertnbr,              /* Number of vertices                  */
      &verttab[0], // const SCOTCH_Num * const    verttab,              /* Vertex array [vertnbr or vertnbr+3] */
      NULL, // const SCOTCH_Num * const    vendtab,              /* Vertex end array [vertnbr]          */
      &nprocs[0], // const SCOTCH_Num * const    velotab,              /* Vertex load array                   */
      NULL, // const SCOTCH_Num * const    vlbltab,              /* Vertex label array                  */
      edgenbr, // const SCOTCH_Num            edgenbr,              /* Number of edges (arcs)              */
      &edgetab[0], // const SCOTCH_Num * const    edgetab,              /* Edge array [edgenbr]                */
      &edlotab[0]); // const SCOTCH_Num * const    edlotab);              /* Edge load array                     */
    //assert (SCOTCH_graphCheck (&phygdat) == 0);
  }
// provide list of to-be-mapped to vertices and size ...
  assert(nprocs.size() == ptg.size());
  std::vector<int> mapping_nodes;
  for(int i=0; i<ptg.size(); ++i) if(nprocs[i] > 0) { mapping_nodes.push_back(i); }

  //printf("mapping_nodes: "); for(int i=0; i<mapping_nodes.size(); ++i) { printf("%i ", mapping_nodes[i]); } printf("\n");

  SCOTCH_stratInit (&physdat);
  SCOTCH_archBuild (&phyadat, &phygdat, mapping_nodes.size(), &mapping_nodes[0], &physdat); // Compute target architecture from physical graph
  SCOTCH_stratExit (&physdat);                    // Free architecture recursive bipartitioning strategy
  SCOTCH_graphExit (&phygdat);                    // Free architecture graph data (vectors can be safely re-used)

  edgenbr=0;
  for(int i=0; i<ltg.size(); ++i) edgenbr += ltg[i].size();
  edgetab.resize(edgenbr);
  edlotab.resize(edgenbr);
  verttab.resize(ltg.size()+1);

  {
    int iter=0;
    verttab[0] = 0;
    for(int i=0; i<ltg.size(); ++i) { std::copy(ltg[i].begin(), ltg[i].end(), &edgetab[iter]); iter += ltg[i].size(); verttab[i+1]=iter; }
    iter=0;
    assert(ltg.size() == ltgw.size());
    for(int i=0; i<ltgw.size(); ++i) { std::copy(ltgw[i].begin(), ltgw[i].end(), &edlotab[iter]); iter += ltgw[i].size(); }

    SCOTCH_graphInit (&loggdat);                    // Initialize logical graph structure
    SCOTCH_graphBuild (&loggdat,
      0, // const SCOTCH_Num            baseval,              /* Base value                          */
      ltg.size(), // const SCOTCH_Num            vertnbr,              /* Number of vertices                  */
      &verttab[0], // const SCOTCH_Num * const    verttab,              /* Vertex array [vertnbr or vertnbr+1] */
      NULL, // const SCOTCH_Num * const    vendtab,              /* Vertex end array [vertnbr]          */
      NULL, // const SCOTCH_Num * const    velotab,              /* Vertex load array                   */
      NULL, // const SCOTCH_Num * const    vlbltab,              /* Vertex label array                  */
      edgenbr, // const SCOTCH_Num            edgenbr,              /* Number of edges (arcs)              */
      &edgetab[0], // const SCOTCH_Num * const    edgetab,              /* Edge array [edgenbr]                */
      &edlotab[0]); // const SCOTCH_Num * const    edlotab)              /* Edge load array                     */
    //assert (SCOTCH_graphCheck (&loggdat) == 0);
  }
  SCOTCH_stratInit (&logsdat);
  SCOTCH_stratGraphMapBuild (&logsdat, SCOTCH_STRATBALANCE, mapping_nodes.size (), 0.05);
  SCOTCH_graphMap (&loggdat, &phyadat, &logsdat, &mapping[0]); // Compute mapping of logical graph to physical graph (mapping of SCOTCH_Num kind)
  SCOTCH_stratExit (&logsdat);
  SCOTCH_graphExit (&loggdat);                    // Free logical graph data
  SCOTCH_archExit  (&phyadat);                    // Free logical graph to physical graph static mapping strategy
#else
  // gather the mapping from all processes - \Theta(P) space
  MPI_Allgather(&me, 1, MPI_INT, &mapping[0], 1, MPI_INT, topocomm);
  if(!r) printf("******* SCOTCH mapping requested but not compiled in, returning original mapping! ********\n");
#endif

  return 0;
}


