#include "libtopomap.hpp"
#include <sstmac/software/libraries/parmetis/parmetis.h>

typedef struct {
    double a;
    int b;
} doubleint;

// this is an exported variable that can be set in order to use the fake
// names file ;)
char *TPM_Fake_names_file = NULL;

void TPM_Get_cong(int r, TPM_Graph *h_ref, TPM_Graph_ewgts *ptgc_ref, TPM_Routes *routes_ref, TPM_Graph *g_ref, TPM_Graph_ewgts *weights_ref, TPM_Mapping *map_ref, double *max_cong, int *max_dil, bool print, double *avg_cong, double *avg_dil, const char *outfile) {
  TPM_Graph_ewgts& ltgw= *weights_ref; // ltg edge weights!
  TPM_Graph_ewgts& ptgc= *ptgc_ref; // ptg capacity weights!
  TPM_Mapping& mapping = *map_ref; // mapping from vertex in g to vertex in h
  TPM_Routes& routes = *routes_ref; // routes in h
  TPM_Graph& g= *g_ref; // embedded graph
  TPM_Graph& h= *h_ref; // host graph

  TPM_Graph_dewgts wgt(h.size());
  // init edge weights in h to n*n to force shortest paths
  int maxw=0; 
  assert(g.size() == ltgw.size());
  for(int i=0; i<g.size(); ++i) {
    if(ltgw[i].size() != 0) {
      //printf("[%i] ... %i %i %x %i\n", r, i, ltgw.size(), &ltgw[i][0], ltgw[i].size());
      maxw=std::max(maxw, *std::max_element(ltgw[i].begin(), ltgw[i].end()) );
    }
  }
  assert(maxw>0); // check for overflow
  const double init=maxw*h.size()*h.size();
  assert(init>0); // check for overflow
  for(int i=0; i<h.size(); ++i) {
    wgt[i].resize(h[i].size());
    std::fill(wgt[i].begin(), wgt[i].end(), init);
  }

  // for each vertex, annotate the edges on shortest path
  // this would simulate a routing which ideally balances packets along
  // shortest paths !
  // Oblivious static routing would cause more congestion!
  *max_cong=*max_dil=0;
  TPM_Graph_dists dist(h.size());
  TPM_Graph_preds pred(h.size());
  int g_edges=0, dil_sum=0; // to compute average dilation
  // iterate over all vertices in logical topology (p)
  for(int i=0; i<g.size(); i++) {
    // iterate over all neighbors of each vertex in topology
    for(int j=0; j<g[i].size(); j++) {
      // find vertices in physical topology
      int src = mapping[i];
      int tgt = mapping[g[i][j]];

      if(print) printf("checking virtual topo edge %i -> %i (mapped %i -> %i)\n", i, g[i][j], src, tgt);
      // new: get the route from the routing module
      std::vector<int> path;
      TPM_RouteGet(&h, &wgt, &routes, src, tgt, &path);

      // walk path from source to destination in physical graph
      int idx=0;
      //assert(path[0] == tgt);
      int v=src;
      int dilation = 0; // two cables at least per hop but one is already in ltg but if we map on multicore, the average dilation can become negative :-/ -- thus init to 0
      double maxc=0;
//if(!r) printf("--- src: %i, tgt: %i, pathlen: %i\n", src, tgt, path.size());
      while(v != tgt) {
        dilation++;
        int w=path[idx]; // next hop, we're looking at edge v -> w right now
        idx++;
//if(!r) printf("--- edge %i->%i\n", v, w);
        // find position of edge v -> w in edge weights map (same
        // as in h -- thus, find position in h :)
        assert(h.size() > v);
        int pos; for(pos=0; h[v][pos] != w; pos++) if(pos >= h[v].size()) { printf("[%i] edge %i->%i not found\n", r, v, w); break; }
//if(!r) printf(".. step: %i, src: %i, dst: %i, pos: %i, cost: %.2f\n", dilation, v, w, pos, wgt[v][pos]);
        assert(wgt.size() > v);
        assert(wgt[v].size() > pos);

        wgt[v][pos] += (double)ltgw[i][j]/ptgc[v][pos]; // add weight of logical edge divided by capacity of physical edge
        maxc=std::max(wgt[v][pos] - init, maxc);

        if(print) printf("increasing edge weight on edge %i -> %i to %f\n", v, w, wgt[v][pos]-init);
        v = w;
      }
      assert(idx == path.size()); // just safety
      *max_cong = std::max(maxc,*max_cong);
      *max_dil = std::max(dilation,*max_dil);
      
      if(avg_dil != NULL) {
        g_edges++;
        dil_sum+=dilation;
      }

      DBG3(printf("%i -> %i: %.2f, %i\n", src, tgt, *max_cong, dilation));
    }
  }
  // at this point, we have fully annotated the graph and might want
  // to loop over it again to get the congestion sum, however, for the
  // congestion max, we don't need to do this!
  if(avg_cong!=NULL) {
    int h_used_edges=0;
    double h_cong_sum=0;
    for(int i=0; i<h.size(); ++i) {
      for(int j=0; j<h[i].size(); ++j) {
        if(wgt[i][j] > init) {
          h_used_edges++;
          h_cong_sum+=(wgt[i][j]-init);
        }
      }
    }
    *avg_cong = h_cong_sum/h_used_edges;
  }
  if(avg_dil != NULL) *avg_dil = (double)dil_sum/g_edges;
  
  /* this section only writes the graph if necessary */
  if(r == 0 && outfile != NULL) {
    FILE *f = fopen(outfile, "w");
    fprintf(f, "digraph net {\n");
    
    // print mapping 
    for(int i=0; i<mapping.size(); ++i) fprintf(f, "%i [ rank=\"%i\" ]\n", mapping[i], i);

    for(int i=0; i<h.size(); ++i) {
      for(int j=0; j<h[i].size(); ++j) {
        // only print edges that carry some weight (they don't have to
        // be mapped if they're routed through ...)
        if(wgt[i][j] > init) fprintf(f, "%i -> %i [ weight=\"%f\" ]\n", i, h[i][j], wgt[i][j]-init);
      }
    }
    fprintf(f, "}\n");
    fclose(f);
    INFO(printf("[%s written]\n", outfile));
  }
  /* end of graph output section */
}

static int TPM_Get_myid(int r, TPM_Nodenames *names) {
  // determine my hostname
  const int len=1024;
  char name[len];
  if(TPM_Fake_names_file != NULL) {
    TPM_Fake_hostname(TPM_Fake_names_file, r, name, len);
  } else {
    // determine my hostname
    gethostname(name, len);
  }
  // scan for and remove first "."
  for(int i=0; i<len; ++i) if(name[i] == '.') { name[i] = '\0'; break; }
  std::string hostname = name;

  // find my vertex-id in graph -- linear search (but is only done once,
  // could turn name-table into a map!
  int me=-1;
  for(int i=0; i<names->size(); ++i) if((*names)[i] == hostname) { me=i; break; }
  if(me == -1) {
    printf("[%i] didn'n find my own hostname (%s) in nametable! Aborting job.\n", r, hostname.c_str());
    MPI_Abort(MPI_COMM_WORLD, 2);
  }
  DBG2(printf("[%i] found myself (%s) as vertex %i!\n", r, hostname.c_str(), me));

  return me;
}

static inline int TPM_Graph_size(TPM_Graph *g_ref) {
  TPM_Graph& g = *g_ref;
  int ret=0;
  for(int i=0; i<g.size(); ++i) ret += g[i].size();
  return ret;
}

int TPM_Topomap(MPI_Comm topocomm, const char *topofile, int max_mem, int *newrank) {
  int r,p;
  MPI_Comm_size(topocomm, &p);
  MPI_Comm_rank(topocomm, &r);

  TPM_Nodenames names;
  TPM_Graph ptg; // physical topology graph
  TPM_Graph_ewgts  ptgc; // physical topology graph capacity
  TPM_Routes routes;

#ifdef BGP
  int me;
  TPM_Get_bgp_topo(r, &ptg, &me);

  //  fill ptgc with 1
  ptgc.resize(ptg.size());
  for(int i=0; i<ptgc.size(); ++i) {
    ptgc[i].resize(ptg[i].size());
    std::fill(ptgc[i].begin(), ptgc[i].end(), 1);
  }
#else
  TPM_Read_topo(topofile, &names, &ptg, &ptgc);
  int me = TPM_Get_myid(r, &names);
  //printf("[%i] myid: %i\n", r, me);

  // read routes if necessary
  TPM_RoutesInit( !r ? 1 : 0, &routes);
#endif

  int n=ptg.size();

  const int NONE = 0; // no mapper
  const int RCM = 1; // RCM mapper
  const int GREEDY = 2; // greedy mapper
  const int RECURSIVE = 3; // recursive mapper
  const int SCOTCH = 4; // recursive mapper
  const int GREEDY_ROUTE = 5; // recursive mapper
  const int RECURSIVE_ROUTE = 6; // recursive mapper

  int mapper = NONE; // which mapping algorithm to use, can be different on different processes!

  int use_anneal = 0; // use simulated annealing to refine result
  if(n>500) use_anneal = 0; // too slow in those cases :-(
  int use_parmetis = 1; // variable that indicates if parmetis can be used or not (will be disabled if not all nodes have the same number of processes OR ltg is not symmetric!)
  int fix_parmetis = 1; // attempt to fix parmetis partition if it is not perfectly balanced. The fix just balances it randomly and guarantees bad outcome :-/
  int fix_graph= 1; // this is a DIRTY hack if the graph is not symmetric (ParMETIS needs a symmetric graph), make it symmetric by adding edges!

  double t_parm=0, t_map=0;

  /* check if environment variables determine our selection! -- mainly
   * for testing purposes */
  char *env = getenv("TPM_STRATEGY");
  if(env != NULL) if(!strcmp(env,"none")) mapper=NONE;
                  else if(!strcmp(env,"greedy")) mapper=GREEDY;
                  else if(!strcmp(env,"greedy_route")) mapper=GREEDY_ROUTE;
                  else if(!strcmp(env,"recursive")) mapper=RECURSIVE;
                  else if(!strcmp(env,"rcm")) mapper=RCM;
                  else if(!strcmp(env,"scotch")) mapper=SCOTCH;
                  else if(!strcmp(env,"recursive_route")) mapper=RECURSIVE_ROUTE;
  env = getenv("TPM_ANNEAL");
  if(env != NULL ) if(!strcmp(env,"yes")) use_anneal=1; else use_anneal=0;
  env = getenv("TPM_PARMETIS");
  if(env != NULL) if(!strcmp(env,"yes")) use_parmetis=1; else use_parmetis=0;
  env = getenv("TPM_FIX_GRAPH");
  if(env != NULL) if(!strcmp(env,"yes")) fix_graph=1; else fix_graph=0;
  env = getenv("TPM_FIX_PARMETIS");
  if(env != NULL) if(!strcmp(env,"yes")) fix_parmetis=1; else fix_parmetis=0;

  // RCM mapper needs ParMETIS!
  if(mapper == RCM) fix_parmetis = fix_graph = 1;

  const char *fake_file="none"; if(TPM_Fake_names_file != NULL) fake_file=TPM_Fake_names_file;

  INFO(if(!r) printf("INFO p: %i, fake_file: \"%s\", topo_file: \"%s\", mapper: %i (0=none, 1=rcm, 2=greedy, 3=rec, 4=scotch, 5=greedy_route, 6=rec_route), use_anneal: %i, use_parmetis: %i, fix_parmetis: %i, fix_graph: %i\n", p, fake_file, topofile, mapper, use_anneal, use_parmetis, fix_parmetis, fix_graph));

  // determine the number of processes that run on each host
  // all processes need to know the result -- this would be a
  // Gather_reduce :-) -- for now an Alltoall though :-/
  TPM_Graph_vwgts nprocs(ptg.size());
  std::fill(nprocs.begin(), nprocs.end(), 0);
  // alltoall buffers --- TWO!!! I WANT MPI-2.2!!!
  {
    std::vector<int> sbuf(p), rbuf(p);
    std::fill(sbuf.begin(), sbuf.end(), me);
    MPI_Alltoall(&sbuf[0], 1, MPI_INT, &rbuf[0], 1, MPI_INT, topocomm);
    // fill nprocs structure for each vertex
    for(int i=0; i<p; ++i) nprocs[rbuf[i]]++;
  }
  
  DBG2(if(!r) for(int i=0; i<n; ++i) if(nprocs[i]) printf("%i ranks on node %i (%s)\n", nprocs[i], i, names[i].c_str()));
  
  /* the hardware graph is initialized at this point -- the weight of
   * each vertex is the number of processes that run on this vertex */
  MEM(if(!r) printf("ptg: cardinality %lu, size: %i (mem: %.2f kiB)\n", ptg.size(), TPM_Graph_size(&ptg), (float)TPM_Graph_size(&ptg)*(float)sizeof(int)/1024));

  // check if all nodes have the same number of processes! This is
  // required for the metis partitioning!
  int nproc=0;
  for(int i=0; i<ptg.size(); ++i) if(nprocs[i] != 0) {
    if(nproc==0) {
      nproc = nprocs[i];
    }
    if(nproc != nprocs[i]) {
      nproc = -1;
      break;
    }
  }
  
  if(use_parmetis) if(nproc < 2) { 
    INFO(if(!r) printf("no symmetric multicore allocation found (nproc=%i) disabling ParMETIS!\n", nproc));
    use_parmetis = 0;
  } DBG2(else if(!r) printf("all nodes have %i processes\n", nproc));

  // now collect the graph that we want to map to the topology ... we
  // assume that each process *can* store the whole graph! This needs to
  // be checked before collecting it (we just assume it here). 
  
  // now, each process queries its neighbors (all edges that it has) and
  // then all those edges are allgatherv'd

  int indegree, outdegree, weighted;
  std::vector<int> in, out, inw, outw;

  int topo;
  MPI_Topo_test(topocomm, &topo);

  switch(topo) {
    case MPI_CART: /* Cartesian */
      {
        int ndims, i, rpeer, speer;
        MPI_Cartdim_get(topocomm, &ndims);

        for(i = 0; i<ndims; i++) {
          MPI_Cart_shift(topocomm, i, 1, &rpeer, &speer);
          if(rpeer != MPI_PROC_NULL) { in.push_back(rpeer); out.push_back(rpeer); }
          if(speer != MPI_PROC_NULL) { in.push_back(speer); out.push_back(speer); }
        }
        indegree = in.size();
        outdegree = out.size();
        weighted = 0;
      }
      break;
    case MPI_GRAPH: /* graph */
      {
        int rank, nneighbors;
        MPI_Comm_rank(topocomm, &rank);
        MPI_Graph_neighbors_count(topocomm, rank, &nneighbors);
        in.resize(nneighbors);
        out.resize(nneighbors);
        MPI_Graph_neighbors(topocomm, rank, nneighbors, &in[0]);
        std::copy(in.begin(), in.end(), out.begin());

        indegree = in.size();
        outdegree = out.size();
        assert(indegree == outdegree);
        weighted = 0;
      }
      break;
#ifndef FAKE_MPI22
    case MPI_DIST_GRAPH: /* dist graph */
      {
        int res = MPI_Dist_graph_neighbors_count(topocomm, &indegree, &outdegree, &weighted);
        in.resize(indegree);
        out.resize(outdegree);
        if(weighted) { inw.resize(indegree); outw.resize(outdegree); }
        MPI_Dist_graph_neighbors(topocomm, indegree, &in[0], &inw[0], outdegree, &out[0], &outw[0]);
      }
      break;
#endif
    case MPI_UNDEFINED:
#ifdef FAKE_MPI22
      // get information about local neighbors (we need only out-vertices
      // ... we could save memory here, but I don't want to trigger bugs yet ;)
      MPIX_Dist_graph_neighbors_count(topocomm, &indegree, &outdegree, &weighted);

      in.resize(indegree);
      out.resize(outdegree);
      if(weighted) { inw.resize(indegree); outw.resize(outdegree); }
      MPIX_Dist_graph_neighbors(topocomm, indegree, &in[0], &inw[0], outdegree, &out[0], &outw[0]);
      assert(outdegree == out.size());
#endif
      break;
    default:
      return -20;
      break;
  }

  // we might get double edges here -- transform them into a weight map
  // (TODO: we loose the actual edge weights in this process but they
  // can easily be recovered if needed!)

  std::vector<int> wgts;

  // unweighted graphs might indicate weight with multiple edges
  if(!weighted) {
    if(outdegree > 0) {
      std::vector<int> tmp(outdegree); // temp vector

      std::sort(out.begin(), out.end());
      std::copy(out.begin(), out.end(), tmp.begin());
      out.erase(unique(out.begin(), out.end()), out.end());
      outdegree = out.size();
        
      // get weights by counting elements in tmp
      wgts.resize(outdegree);

      int elem=tmp[0], cnt=0, pos=0;
      for(int i=0; i<tmp.size()+1; ++i) {
        if(i < tmp.size() && tmp[i] == elem) cnt++;
        else { 
          if(i<tmp.size()) elem = tmp[i]; // this if guard seems silly -- might want to restructure this whole loop!
          assert(pos < out.size());
          //DBG2(printf("[%i] weight[%i]=%i\n", r, pos, cnt));
          wgts[pos++] = cnt;
          cnt=1; 
  } } } } else {
    // weigthed graphs ignore double edges TODO: should be merged too at
    // some point ...
    wgts.resize(outdegree);
    std::copy(outw.begin(), outw.end(), wgts.begin());
  }

  //printf("[%i] in: %i, out %i\n", r, indegree, outdegree);

  TPM_Graph ltg(p);
  TPM_Graph_ewgts ltgw(ltg.size());
  // now collect information for allgatherv -- we only send out-vertices!
  {
    std::vector<int> rcounts(p);
    MPI_Allgather(&outdegree, 1, MPI_INT, &rcounts[0], 1, MPI_INT, topocomm);
    int bufsize = std::accumulate(rcounts.begin(), rcounts.end(), 0);
    
    std::vector<int> rbuf(bufsize);
    std::vector<int> displs(p);
    displs[0] = 0; for(int i=1; i<p; ++i) displs[i] = displs[i-1]+rcounts[i-1]; 
    MPI_Allgatherv(&out[0], outdegree, MPI_INT, &rbuf[0], &rcounts[0], &displs[0], MPI_INT, topocomm);

    // weights
    std::vector<int> wgtsrbuf(bufsize);
    MPI_Allgatherv(&wgts[0], outdegree, MPI_INT, &wgtsrbuf[0], &rcounts[0], &displs[0], MPI_INT, topocomm);

    // save full graph in ltg
    for(int i=0; i<p; ++i) {
      DBG2(if(r==0) printf("vertex %i adjacency list: ", i));
      ltg[i].resize(rcounts[i]);
      ltgw[i].resize(rcounts[i]);
      for(int j=displs[i]; j<displs[i]+rcounts[i]; j++) {
        ltg[i][j-displs[i]] = rbuf[j];
        ltgw[i][j-displs[i]] = wgtsrbuf[j];
        DBG2(if(r==0) printf(" %i (%i) ", rbuf[j], wgtsrbuf[j]));
      }
      DBG2(if(r==0) printf("\n"));
    }
  }

  /* at this point each process has access to the physical topology
   * graph ptg and the logical topology graph ltg */
  MEM(if(!r) printf("ltg: cardinality %lu, size: %i (mem: %.2f kiB)\n", ltg.size(), TPM_Graph_size(&ltg), (float)TPM_Graph_size(&ltg)*(float)sizeof(int)/1024));

  /* get the edge cut with the default mapping */
  TPM_Mapping origmapping(p);
  double origmax_cong;
  int origmax_dil;
  double origavgs[2];
  // gather the mapping from all processes - \Theta(P) space
  MPI_Allgather(&me, 1, MPI_INT, &origmapping[0], 1, MPI_INT, topocomm);
  TPM_Get_cong(r, &ptg, &ptgc, &routes, &ltg, &ltgw, &origmapping, &origmax_cong, &origmax_dil, false, &origavgs[0], &origavgs[1], "origcong.dot");
  //if(!r) printf("orig max. congestion: %i, max. dilation %i\n", origmax_cong, origmax_dil);
  /* end getting edge cut for default mapping */


  // the ltg graph *must* be symmetric for METIS!
  // check ltg for symmetry
  if(use_parmetis) for(int i=0; i<ltg.size(); ++i) {
    for(int j=0; j<ltg[i].size(); ++j) {
      int src = i, tgt = ltg[i][j], found=0;
      //if(!r) printf("%i -> %i (%i/%i)\n", src, tgt, j, ltg[i].size());
      for(int k=0; k<ltg[tgt].size(); ++k) if(ltg[tgt][k] == src) found=1;
      if(!found) if(fix_graph) {
        // add edge that does not exist!
        ltg[tgt].push_back(src);
        ltgw[tgt].push_back(ltgw[src][tgt]);
        INFO(if(!r) printf("added edge %i->%i to make graph symmetric\n", tgt, src));
      } else {
        INFO(if(!r) printf("edge (%i,%i) exists but edge (%i,%i) not! Graph not symmetric - disabling ParMETIS\n", src,tgt,tgt,src));
        use_parmetis = 0;
        break;
      }
    } 
    if(!use_parmetis) break;
  }

  TPM_Mapping mapping(ltg.size()); // topology mapping
  std::vector<int> fullpart(p); // full ParMETIS partition vector
  
  if(use_parmetis) {
    assert(ltg.size() == p);
    std::vector<idx_t>  vtxdist(p+1), // vertex distribution
                          xadj(2), // CSR index 
                          adjncy(ltg[r].size()); // CSR list

    vtxdist[0] = 0; for(int i=1; i<(p+1); i++) vtxdist[i]=vtxdist[i-1]+1;
    xadj[0]=0; xadj[1]=ltg[r].size();
    
    std::copy(ltg[r].begin(), ltg[r].end(), adjncy.begin());

    std::vector<idx_t> vwgt(1,1); // vertex weights 
    std::vector<idx_t> adjwgt(ltg[r].size()); // edge weights
    std::copy(ltgw[r].begin(), ltgw[r].end(), adjwgt.begin());

    //for(int i=0; i<adjncy.size(); i++) printf("[%i] -> %i (%i)\n", r, adjncy[i], adjwgt[i]);
    int wgtflag=3; // 0 No weights, 1 Weights on the edges only, 2 Weights on the vertices only, 3 Weights on both the vertices and edges. 
    int numflag=0; // numbering scheme - 0 C-style numbering that starts from 0
    int ncon=1; // number of weights that each vertex has.
    int nparts = p/nproc; // number of partitions
    std::vector<float> tpwgts(ncon*nparts,1.0/nparts); // specify the fraction of vertex weight 
    std::vector<float> ubvec(ncon,1.05); // specify the imbalance tolerance for each vertex weight
    int options[3] = {0,0,0};
    int edgecut;
    std::vector<idx_t> part(1);

    if(nparts != p) {
      MPI_Barrier(topocomm); // for timing only -- I know, it's dumb
      t_parm = -MPI_Wtime();
      ParMETIS_V3_PartKway(&vtxdist[0], &xadj[0], &adjncy[0], &vwgt[0], &adjwgt[0], &wgtflag, &numflag, 
                        &ncon, &nparts, &tpwgts[0], &ubvec[0], options, &edgecut, &part[0], &topocomm);
      MPI_Barrier(topocomm); // for timing only -- I know, it's dumb
      t_parm += MPI_Wtime();
    } else {
      part[0] = r;
      edgecut = -1; 
    }
    INFO(if(!r) printf("[ParMETIS: edge-cut: %i (splitted graph of cardinality %lu in %i partitions, nproc=%i)]\n", edgecut, ltg.size(), nparts, nproc));

    /* build node-graph from the parmetis partition */
    // first collect the parmetis partition everywhere
    assert(sizeof(idx_t) == sizeof(int));
    fullpart.resize(p);
    MPI_Allgather(&part[0], 1, MPI_INT, &fullpart[0], 1, MPI_INT, topocomm);
    DBG2(if(!r) printf("[%i] parmetis partition: ", r); if(!r) for(int i=0; i<p; ++i) printf("%i ", fullpart[i]); if(!r) printf("\n"));

    /* parmetis does not guarantee to generate perfectly balanced
     * partitions. They are most likely well balanced but our
     * application *requires* perfect balancing. Thus, we might need to
     * correct the partitioning here */
    std::vector<int> cnts(p,0);
    // traverse fullpart and check if each host exists exactly nprocs times!
    for(int i=0; i<p; ++i) cnts[fullpart[i]]++;
    int balanced = 1;
    for(int i=0; i<nparts; ++i) if(cnts[i] != nproc) balanced = 0;
    // fix parmetis mapping to be balanced
    if(!balanced) if(fix_parmetis) {
      int nvert_fixed=0; // number of vertices that have been fixed
      while(!balanced) {
        nvert_fixed++;
        // find overloaded node
        int overloaded = 0;
        for(int i=0; i<nparts; ++i) if(cnts[i] > nproc) { overloaded=i; break; }
        // find underloaded node 
        int underloaded = 0;
        for(int i=0; i<nparts; ++i) if(cnts[i] < nproc) { underloaded=i; break; }
        // exchange overloaded and underloaded -- TODO: we should pick
        // the least connected vertex from the overloaded node here!
        int overloaded_index=0;
        for(int i=0; i<p; ++i) if(fullpart[i] == overloaded) { overloaded_index=i; break; }
        fullpart[overloaded_index] = underloaded;
        cnts[overloaded]--;
        cnts[underloaded]++;
        
        // re-check balance
        balanced = 1; for(int i=0; i<nparts; ++i) if(cnts[i] != nproc) balanced = 0;
      }
      INFO(if(!r && nvert_fixed>0) printf("fixed (moved) %i ParMETIS vertices!\n", nvert_fixed));
    } else { 
      INFO(if(!r) printf("ParMETIS partition not perfectly balanced and fix_parmetis=false -- disabling ParMETIS\n"));
      use_parmetis = 0;
    }
  }

  // this block is just to debug parmetis partitioning
  if(use_parmetis) {
    double xcong; int xdil;
    std::vector<int> xmap(ltg.size());
    TPM_Get_cong(r, &ptg, &ptgc, &routes, &ltg, &ltgw, &fullpart, &xcong, &xdil, false, NULL, NULL, "parmetis.dot");
  }

  if(use_parmetis) {
    // build node topology graph
    TPM_Graph ntg(p/nproc);
    TPM_Graph_ewgts ntgw(ntg.size());
    for(int i=0; i<ltg.size(); ++i) {
      for(int j=0; j<ltg[i].size(); ++j) {
        int src=i, tgt=ltg[i][j];
        int nsrc=fullpart[src];
        int ntgt=fullpart[tgt];
        // add edge to node topology graph if necessary
        if(nsrc != ntgt) {
          // check if edge exists -- TODO: change lin. search to map lookup!
          int found=0, pos=0;
          for(int k=0; k<ntg[nsrc].size(); ++k, ++pos) if(ntg[nsrc][pos] == ntgt) { found = 1; break; }
          if(found) {
            ntgw[nsrc][pos] += ltgw[i][j];
          } else {
            ntg[nsrc].push_back(ntgt);
            ntgw[nsrc].push_back(ltgw[i][j]);
          }
        }
      }
    }
    MEM(if(!r) printf("ntg: cardinality %lu, size: %i (mem: %.2f kiB)\n", ntg.size(), TPM_Graph_size(&ntg), (float)TPM_Graph_size(&ntg)*(float)sizeof(int)/1024));
    
    //if(!r) for(int i=0; i<ntg.size(); ++i) { printf("ntg vertex %i adjacency list: ", i); for(int j=0; j<ntg[i].size(); j++) printf(" %i ", ntg[i][j]); printf("\n"); }

    for(int i=0; i<nprocs.size(); ++i) if(nprocs[i] > 0) nprocs[i] = 1;
    
    TPM_Mapping ntgmap(ntg.size()); // mapping from ntg to ptg

    t_map = -MPI_Wtime();
    if(mapper == RCM) TPM_Map_rcm(r, &ptg, &nprocs, &ntg, &ntgmap, me /* start */);
    else if(mapper == RECURSIVE) { // set up maps for recursive mapping
      TPM_Mapping lmap(ntg.size());
      for(int i=0; i<ntg.size(); ++i) lmap[i] = i;

      // copy ptg and ptgc graphs
      TPM_Graph tmpptg(ptg.size()), tmpptgc(ptgc.size());
      std::copy(ptg.begin(), ptg.end(), tmpptg.begin());
      for(int i=0; i<ptg.size(); i++) {
        tmpptg[i].resize(ptg[i].size());
        std::copy(ptg[i].begin(), ptg[i].end(), tmpptg[i].begin());
      }
      std::copy(ptgc.begin(), ptgc.end(), tmpptgc.begin());
      for(int i=0; i<ptgc.size(); i++) {
        tmpptgc[i].resize(ptgc[i].size());
        std::copy(ptgc[i].begin(), ptgc[i].end(), tmpptgc[i].begin());
      }
      TPM_Mapping pmap;
      for(int i=0; i<ptg.size(); ++i) pmap.push_back(i);
      
      if(mapper == RECURSIVE) {
        if(!r) printf("calling recursive mapper with ltg: %lu and ptg: %i\n", lmap.size(), std::accumulate(nprocs.begin(), nprocs.end(), 0));
        TPM_Map_recursive(r, &tmpptg, &tmpptgc, &nprocs, &ntg, &ntgw, &ntgmap, lmap, &pmap);
      } else {
        if(!r) printf("calling recursive route mapper with ltg: %lu and ptg: %i\n", lmap.size(), std::accumulate(nprocs.begin(), nprocs.end(), 0));
        TPM_Map_recursive_routing(r, &tmpptg, &tmpptgc, &nprocs, &ntg, &ntgw, &ntgmap, lmap, &pmap, &routes);
      }
    } else if(mapper == SCOTCH) {
      TPM_Map_scotch(r, &ptg, &ptgc, &nprocs, &ntg, &ntgw, &ntgmap, me, topocomm);
    } else if(mapper == GREEDY_ROUTE) {
        TPM_Map_greedy_routing(r, &ptg, &ptgc, &routes, &nprocs, &ntg, &ntgw, &ntgmap, me /* start */);
    } else TPM_Map_greedy(r, &ptg, &ptgc, &nprocs, &ntg, &ntgw, &ntgmap, me /* start */);
    t_map += MPI_Wtime();

    // inflate mapping back to logical topology size
    for(int i=0; i<ltg.size(); ++i) {
      int pos_in_ntg=fullpart[i];
      assert(pos_in_ntg<ntgmap.size());
      mapping[i] = ntgmap[pos_in_ntg];
    }
  } else { // if(use_parmetis)
    // do greedy mapping on full graph
    t_map = -MPI_Wtime();
    if(mapper == RCM) TPM_Map_rcm(r, &ptg, &nprocs, &ltg, &mapping, me /* start */);
    else if(mapper == RECURSIVE || mapper == RECURSIVE_ROUTE) { // set up maps for recursive mapping
      TPM_Mapping lmap(ltg.size());
      for(int i=0; i<ltg.size(); ++i) lmap[i] = i;
      
      // copy ptg and ptgc graphs
      TPM_Graph tmpptg(ptg.size());
      TPM_Graph_ewgts tmpptgc(ptgc.size());
      std::copy(ptg.begin(), ptg.end(), tmpptg.begin());
      for(int i=0; i<ptg.size(); i++) {
        tmpptg[i].resize(ptg[i].size());
        std::copy(ptg[i].begin(), ptg[i].end(), tmpptg[i].begin());
      }
      std::copy(ptgc.begin(), ptgc.end(), tmpptgc.begin());
      for(int i=0; i<ptgc.size(); i++) {
        tmpptgc[i].resize(ptgc[i].size());
        std::copy(ptgc[i].begin(), ptgc[i].end(), tmpptgc[i].begin());
      }
      TPM_Mapping pmap;
      for(int i=0; i<ptg.size(); ++i) pmap.push_back(i);
      

      if(mapper == RECURSIVE) {
        if(!r) printf("calling recursive mapper with ltg: %lu and ptg: %i\n", lmap.size(), std::accumulate(nprocs.begin(), nprocs.end(), 0));
        TPM_Map_recursive(r, &tmpptg, &tmpptgc, &nprocs, &ltg, &ltgw, &mapping, lmap, &pmap);
      } else {
        if(!r) printf("calling recursive route mapper with ltg: %lu and ptg: %i\n", lmap.size(), std::accumulate(nprocs.begin(), nprocs.end(), 0));
        TPM_Map_recursive_routing(r, &tmpptg, &tmpptgc, &nprocs, &ltg, &ltgw, &mapping, lmap, &pmap, &routes);
      }
    } else if(mapper == SCOTCH) {
      TPM_Map_scotch(r, &ptg, &ptgc, &nprocs, &ltg, &ltgw, &mapping, me, topocomm);
    } else if(mapper == GREEDY_ROUTE) {
        TPM_Map_greedy_routing(r, &ptg, &ptgc, &routes, &nprocs, &ltg, &ltgw, &mapping, me /* start */);
    } else TPM_Map_greedy(r, &ptg, &ptgc, &nprocs, &ltg, &ltgw, &mapping, me /* start */);
    t_map += MPI_Wtime();
  } // else(use_parmetis)

  double max_cong;
  int max_dil;
  double avgs_gr[3];
  TPM_Get_cong(r, &ptg, &ptgc, &routes, &ltg, &ltgw, &mapping, &max_cong, &max_dil, false, &avgs_gr[1], &avgs_gr[2]);
  avgs_gr[0] = max_dil;
  //if(!r) printf("max. congestion: %i, max. dilation %i\n", max_cong, max_dil);

  // only to print it later ... 
  doubleint ti_greedy = { max_cong, r };
  MPI_Allreduce(MPI_IN_PLACE, &ti_greedy, 1, MPI_DOUBLE_INT, MPI_MINLOC, topocomm);
  MPI_Bcast(avgs_gr, 3, MPI_DOUBLE, ti_greedy.b, topocomm);

  double new_cong;
  int new_dil, anneal_iters=0;
  double t_anneal = -MPI_Wtime();
  if(use_anneal) {
    // bcast best mapping such that annealing step starts with best! --
    // this makes the results WORSE (against all expectations) -- at least
    // for our test-cases ... thus disable for now
    /* each process computed a different mapping here -- see which one
     * found the best one! */
    /*MPI_Bcast(&mapping[0], p, MPI_INT, ti_greedy.b, topocomm);*/
    
    /* do optimization */
    TPM_Map_annealing(r, &ptg, &ptgc, &routes, &nprocs, &ltg, &ltgw, &mapping, &new_cong, &new_dil, &anneal_iters);
    //printf("[%i] annealing %i -> %i (dil: %i -> %i)\n", r, max_cong, new_cong, max_dil, new_dil);
  } else {
    INFO(if(!r) printf("annealing disabled\n"));
    new_cong = max_cong;
    new_dil = max_dil;
  }
  t_anneal += MPI_Wtime();

  if(!r) printf("ParMETIS time: %.6f s; Mapping time strategy %f: %i s; Annealing time: %.6f iters: %i\n", t_parm, t_map, mapper, t_anneal, anneal_iters);

  double avgs[2];
  TPM_Get_cong(r, &ptg, &ptgc, &routes, &ltg, &ltgw, &mapping, &new_cong, &new_dil, false, &avgs[0], &avgs[1], "cong.dot");
  doubleint ti = { new_cong, r };
  MPI_Allreduce(MPI_IN_PLACE, &ti, 1, MPI_DOUBLE_INT, MPI_MINLOC, topocomm);
  MPI_Bcast(avgs, 2, MPI_DOUBLE, ti.b, topocomm);

  int new_mapping_found=1;
  if(ti.a >= origmax_cong) {
    max_cong=origmax_cong;
    max_dil=origmax_dil;
    INFO(if(!r) printf("no better mapping found with strategy %i (found cong: %.2f, %i) -- using orig (cong: %.2f, %i (%.2f, %.2f))\n", mapper, new_cong, new_dil, max_cong, max_dil, origavgs[0], origavgs[1]));
    *newrank = r;
    new_mapping_found=0;
    mapping = origmapping;
  } else {
    double data[3]={new_dil, avgs[0], avgs[1]};
    if(ti.b != 0) {
      /* get dilation to rank 0 -- only for printing */
      MPI_Request req;
      if(!r) MPI_Irecv(&data, 3, MPI_DOUBLE, ti.b, 0, topocomm, &req);
      if(r==ti.b) MPI_Send(&data, 3, MPI_DOUBLE, 0, 0, topocomm);
      if(!r) MPI_Wait(&req, MPI_STATUS_IGNORE);
      /* end get dilation */
    }

    INFO(if(!r) printf("process %i found minimum congestion %.4f, %.0f (%.2f, %.2f) (orig: %.4f, %i (%.2f, %.2f); strategy %i: %.4f, %.0f (%.2f, %.2f))\n", ti.b, ti.a, data[0], data[1], data[2], origmax_cong, origmax_dil, origavgs[0], origavgs[1], mapper, ti_greedy.a, avgs_gr[0], avgs_gr[1], avgs_gr[2]));
    // broadcast best mapping
    MPI_Bcast(&mapping[0], p, MPI_INT, ti.b, topocomm);
  }
  
   
  // I am rank r and run on node me, mapping[r] indicates the new node
  // where I want to go (newme). 
  int newme = mapping[r]; // my target node
#ifdef HAVE_NODE_MAPPING 
  // if on-node mapping is provided, call 
  std::vector<int> neighbors; // ranks in ltg that are on the same node as I am
  TPM_Graph ntg; // on-node topology graph -- ntg[i] contains vertex neighbor[i]
  TPM_Graph_ewgts ntgw; // weights of on-node graph
  std::map<int,int> neighmap; // neighbor map for faster lookup
  
  // print mapping
  //if(!r) for(int i=0; i<p; ++i) printf("%i ", mapping[i]); if(!r) printf("\n");
  
  for(int i=0; i<p; ++i) if(mapping[i] == newme) { neighmap[i]=neighbors.size(); neighbors.push_back(i); }
  ntg.resize(neighbors.size());
  ntgw.resize(neighbors.size());

  // extract subgraph of ltg that connects neighbors
  for(int i=0; i<neighbors.size(); ++i) {
    int u = neighbors[i]; // source vertex
    for(int j=0; j<ltg[u].size(); ++j) {
      int v = ltg[u][j]; // target vertex
      std::map<int,int>::iterator n = neighmap.find(v);
      if (n != neighmap.end()) {
        ntg[i].push_back((*n).second);
        ntgw[i].push_back(ltgw[u][j]);
      }
    }
  }

  *newrank = TPM_Node_mapping(&ntg, &ntgw, &neighbors, topocomm, r);
#else 
  if(new_mapping_found) {
    // no on-node mapping is provided, essentially random order
    int before=0; // the number of ranks smaller than r (myself) that are mapped to my target node (newme)!
    for(int i=0; i<r; ++i) if(origmapping[i] == origmapping[r]) before++;

    INFO(if(!r) printf("origmapping "); if(!r) for(int i=0; i<p; ++i) printf("%i ", origmapping[i]); if(!r) printf("\n"));
    INFO(if(!r) printf("mapping "); if(!r) for(int i=0; i<p; ++i) printf("%i ", mapping[i]); if(!r) printf("\n"));
    
    // Find my new rank.
    // This process is bound to a physical node (origmapping[r]), so we
    // must search the new mapping for the rank(s) re-assigned to this
    // physical node. If more than one rank, chose the one that has
    // the same position as my order in the original mapping (enforced by
    // before).
    int pos = 0;
    //printf("%i searching a rank allocated to mynid=%i (before: %i)\n", r, origmapping[r], before);
    for (pos = 0; pos < p; ++pos) {
      if (origmapping[r] == mapping[pos]) before--;  // we found a target (other ranks might be before us)
      if (before < 0) break;  // we found *our* target!
    }
    *newrank = pos;
  }
#endif
 

  return 0;
} // TPM_Topomap

int TPM_Map_comm(MPI_Comm topocomm, const char *topofile, MPI_Comm *outcomm) {
  int newrank;
  TPM_Topomap(topocomm, topofile, 0, &newrank);
  MPI_Comm_split(topocomm, 0, newrank, outcomm);

  return 0;
}

int TPM_Translate_rank(MPI_Comm sourcecomm, MPI_Comm targetcomm, int rank) {
  MPI_Group sourcegroup, targetgroup;
  MPI_Comm_group(sourcecomm, &sourcegroup);
  MPI_Comm_group(targetcomm, &targetgroup);
  int retrank;
  MPI_Group_translate_ranks(sourcegroup, 1, &rank, targetgroup, &retrank);
  return retrank;
}

int TPM_Write_graph_comm(MPI_Comm grphcomm, const char *filename) {
  int inneighbors, outneighbors, weighted;
  MPIX_Dist_graph_neighbors_count(grphcomm, &inneighbors, &outneighbors, &weighted);
  int r,p;
  MPI_Comm_size(grphcomm, &p);
  MPI_Comm_rank(grphcomm, &r);

  std::vector<int> sources(inneighbors), sourcesw(inneighbors), dests(outneighbors), destsw(outneighbors);
  MPIX_Dist_graph_neighbors(grphcomm, inneighbors, &sources[0], &sourcesw[0], outneighbors, &dests[0], &destsw[0]);

  //for(int i=0; i<inneighbors; ++i) printf("[%i] (%i) %i\n", r, sources[i], sourcesw[i]);

  for(int i=0; i<inneighbors; ++i) assert(sources[i] < p && sources[i] >= 0);

  std::vector<int> rcnts(p), rdispls(p);
  MPI_Gather(&outneighbors, 1, MPI_INT, &rcnts[0], 1, MPI_INT, 0, grphcomm);
  for(int i=0; i<p; ++i) if(i==0) rdispls[i] = 0; else rdispls[i] = rdispls[i-1] + rcnts[i-1];

  int num_edges = std::accumulate(rcnts.begin(), rcnts.end(), 0);
  std::vector<int> tmpedges(num_edges), tmpedgesw(num_edges,1);
  MPI_Gatherv(&dests[0], outneighbors, MPI_INT, &tmpedges[0], &rcnts[0], &rdispls[0], MPI_INT, 0, grphcomm);
  if(weighted) MPI_Gatherv(&destsw[0], outneighbors, MPI_INT, &tmpedgesw[0], &rcnts[0], &rdispls[0], MPI_INT, 0, grphcomm);
  
  if(r == 0) {
    std::map<std::pair<int,int>,int> alledges;
    int cnt=0;
    for(int i=0; i<p; ++i) 
      for(int j=0; j<rcnts[i]; ++j) {
        std::pair<int,int> e=std::make_pair(tmpedges[cnt], i);
        if(alledges.find(e) == alledges.end()) alledges[e] = tmpedgesw[cnt];
        else alledges[e] += tmpedgesw[cnt];
        cnt++;
      }

    FILE *f = fopen(filename, "w");
    fprintf(f, "digraph net {\n");
    for(std::map<std::pair<int,int>,int>::iterator it=alledges.begin(); it!=alledges.end(); ++it) {
      std::pair<int,int> e=it->first;
      fprintf(f, "%i -> %i [ weight=\"%i\" ]\n", e.first, e.second, it->second);
    }
    fprintf(f, "}\n");
    fclose(f);
    INFO(printf("[%s written]\n", filename));
  }
  return 0;
}

int TPM_Write_phystopo(MPI_Comm grphcomm, const char *filename, const char *topofile) {
  int r,p;
  MPI_Comm_size(grphcomm, &p);
  MPI_Comm_rank(grphcomm, &r);

  TPM_Nodenames names;
  TPM_Graph  ptg; // physical topology graph 
  TPM_Graph_ewgts  ptgw; // physical topology graph weights
  TPM_Read_topo(topofile, &names, &ptg, &ptgw);

  int n=ptg.size();

  int me = TPM_Get_myid(r, &names);

  // determine the number of processes that run on each host
  // all processes need to know the result -- this would be a
  // Gather_reduce :-) -- for now an Alltoall though :-/
  TPM_Graph_vwgts nprocs(ptg.size());
  std::fill(nprocs.begin(), nprocs.end(), 0);
  // alltoall buffers --- TWO!!! I WANT MPI-2.2!!!
  std::vector<int> sbuf(p), rbuf(p);
  std::fill(sbuf.begin(), sbuf.end(), me);
  MPI_Alltoall(&sbuf[0], 1, MPI_INT, &rbuf[0], 1, MPI_INT, grphcomm);
  // fill nprocs structure for each vertex
  for(int i=0; i<p; ++i) nprocs[rbuf[i]]++;
  
  
  DBG2(if(!r) for(int i=0; i<n; ++i) if(nprocs[i]) printf("%i ranks on node %i (%s)\n", nprocs[i], i, names[i].c_str()));
  
  /* the hardware graph is initialized at this point -- the weight of
   * each vertex is the number of processes that run on this vertex */
  
  if(r == 0) {
    FILE *f = fopen(filename, "w");
    fprintf(f, "digraph net {\n");
    // print rank allocations!
    for(int i=0; i<ptg.size(); ++i) if(nprocs[i] > 0) {
      fprintf(f, "%i [ ranks=\"", i);
      int first=1;
      for(int j=0; j<p; ++j) if(rbuf[j] == i) {
        if(first) { first=0; } else { fprintf(f, ","); }
        fprintf(f, "%i", j);
      }
      fprintf(f, "\" ]\n");
    }
    
    for(int i=0; i<ptg.size(); ++i) 
      for(int j=0; j<ptg[i].size(); ++j)
        fprintf(f, "%i -> %i \n", i, ptg[i][j]);
    fprintf(f, "}\n");
    fclose(f);
    INFO(printf("[%s written]\n", filename));
  }

  return 0;
}





void TPM_Benchmark_graphtopo(MPI_Comm topocomm, int newrank, int dsize, double *before, double *after) {
  int r,p;
  MPI_Comm_size(topocomm, &p);
  MPI_Comm_rank(topocomm, &r);

  assert(newrank >=0 && newrank <p);
  int indegree, outdegree, weighted;
  MPIX_Dist_graph_neighbors_count(topocomm, &indegree, &outdegree, &weighted);
  //assert(!weighted); // the code below works only for unweighted graphs (but accumulates multiple edges into weights)

  std::vector<int> in(indegree), out(outdegree), inw(indegree), outw(outdegree);
  MPIX_Dist_graph_neighbors(topocomm, indegree, &in[0], &inw[0], outdegree, &out[0], &outw[0]);

  //-- stolen from TPM_Topomap() !!!
  // we might get double edges here -- transform them into a weight map
  // (TODO: we loose the actual edge weights in this process but they
  // can easily be recovered if needed!)
  //std::vector<int> outw;
  if(!weighted) {
    if(outdegree > 0) {
      std::vector<int> tmp(out.size()); // temp vector

      std::sort(out.begin(), out.end());
      std::copy(out.begin(), out.end(), tmp.begin());
      out.erase(unique(out.begin(), out.end()), out.end());
      outdegree = out.size();
        
      // get weights by counting elements in tmp
      outw.resize(outdegree);

      int elem=tmp[0], cnt=0, pos=0;
      for(int i=0; i<tmp.size()+1; ++i) {
        if(tmp[i] == elem && i<tmp.size()) cnt++;
        else { 
          elem = tmp[i]; 
          assert(pos < out.size());
          //DBG2(printf("[%i] weight[%i]=%i\n", r, pos, cnt));
          outw[pos++] = cnt;
          cnt=1; 
  } } } } else {
    // weigthed graphs ignore double edges TODO: should be merged too at
    // some point ...
    assert(outw.size() == outdegree);
  }

  //std::vector<int> inw;
  if(!weighted) {
    if(indegree > 0) {
      std::vector<int> tmp(in.size()); // temp vector

      std::sort(in.begin(), in.end());
      std::copy(in.begin(), in.end(), tmp.begin());
      in.erase(unique(in.begin(), in.end()), in.end());
      indegree = in.size();
        
      // get weights by counting elements in tmp
      inw.resize(indegree);

      int elem=tmp[0], cnt=0, pos=0;
      for(int i=0; i<tmp.size()+1; ++i) {
        if(tmp[i] == elem && i<tmp.size()) cnt++;
        else { 
          elem = tmp[i]; 
          assert(pos < in.size());
          //DBG2(printf("[%i] weight[%i]=%i\n", r, pos, cnt));
          inw[pos++] = cnt;
          cnt=1; 
  } } } } else {
    // weigthed graphs ignore double edges TODO: should be merged too at
    // some point ...
    assert(inw.size() == indegree);
  }

  /* do benchmark - send along edges */

  static const int mult=dsize;
  static const int trials=11;
  std::vector<std::vector<char> > sbufs(outdegree), rbufs(indegree);
  for(int i=0; i<outdegree; ++i) sbufs[i].resize(mult*outw[i]);
  for(int i=0; i<indegree; ++i) rbufs[i].resize(mult*inw[i]);
  std::vector<MPI_Request> sreqs(outdegree), rreqs(indegree);

  DBG5(if(!r) printf("starting benchmark!\n"));

  MPI_Barrier(topocomm);

  MPI_Aint w;
  double t[2];
  for(int i=0; i < trials; ++i) {
    if (i == 1) {
      t[0] = -MPI_Wtime();
    }
    for(int j=0; j<in.size(); ++j) {
      MPI_Irecv(&rbufs[j][0], inw[j]*mult, MPI_BYTE, in[j], 0, topocomm, &rreqs[j]);
    }
    for(int j=0; j<out.size(); ++j) {
      MPI_Isend(&sbufs[j][0], outw[j]*mult, MPI_BYTE, out[j], 0, topocomm, &sreqs[j]);
    }
    // silly AMPI doesn't support MPI_STATUSES_IGNORE
    std::vector<MPI_Status> rstats(rreqs.size());
    std::vector<MPI_Status> sstats(sreqs.size());
    MPI_Waitall(rreqs.size(), &rreqs[0], &rstats[0]);
    MPI_Waitall(sreqs.size(), &sreqs[0], &sstats[0]);
  }
  t[0] += MPI_Wtime();

  /* create reordered comm */
  MPI_Comm reorderedcomm;
  MPI_Comm_split(topocomm, 0, newrank, &reorderedcomm);

  int newr;
  MPI_Comm_rank(reorderedcomm, &newr);
  assert(newr == newrank);

  DBG5(if(!r) printf("finished stage one (%f s), rearranging edge connectivity!\n", t[0]));

  /* now we need to send our neighbor list to the host that has our rank
   * now -- first figure out the permutation in a non-scalable way :) */
  std::vector<int> permutation(p);
  MPI_Allgather(&newrank, 1, MPI_INT, &permutation[0], 1, MPI_INT, topocomm);
  int rpeer=permutation[r]; // new rank r in reorderedcomm!
  int speer=0; while(permutation[speer++] != r); speer--;
  //printf("[%i] >%i <%i\n", r, speer, rpeer);

  MPI_Request reqs[4];
  MPI_Isend(&in[0], in.size(), MPI_INT, speer, 1, topocomm, &reqs[0]);
  assert(in.size() == inw.size());
  MPI_Isend(&inw[0], in.size(), MPI_INT, speer, 1, topocomm, &reqs[1]);
  //printf("%i sending in %i %i to %i\n", r, in[0], in[1], speer);
  MPI_Isend(&out[0], out.size(), MPI_INT, speer, 2, topocomm, &reqs[2]);
  assert(out.size() == outw.size());
  MPI_Isend(&outw[0], out.size(), MPI_INT, speer, 2, topocomm, &reqs[3]);
  //printf("%i sending out %i %i to %i\n", r, out[0], out[1], speer);

  MPI_Status stat;
  /* tag == 1 -> in edges */
  MPI_Probe(rpeer, 1, topocomm, &stat);
  int count;
  MPI_Get_count(&stat, MPI_INT, &count);
  std::vector<int> rin(count), rinw(count);
  MPI_Recv(&rin[0], count, MPI_INT, rpeer, 1, topocomm, MPI_STATUS_IGNORE);
  MPI_Recv(&rinw[0], count, MPI_INT, rpeer, 1, topocomm, MPI_STATUS_IGNORE);
  //printf("%i recvd in %i %i from %i\n", r, rin[0], rin[1], rpeer);

  /* tag == 2 -> out edges */
  MPI_Probe(rpeer, 2, topocomm, &stat);
  MPI_Get_count(&stat, MPI_INT, &count);
  std::vector<int> rout(count), routw(count);
  MPI_Recv(&rout[0], count, MPI_INT, rpeer, 2, topocomm, MPI_STATUS_IGNORE);
  MPI_Recv(&routw[0], count, MPI_INT, rpeer, 2, topocomm, MPI_STATUS_IGNORE);
  //printf("%i recvd out %i %i from %i\n", r, rout[0], rout[1], rpeer);
 
  // silly AMPI doesn't support MPI_STATUSES_IGNORE
  std::vector<MPI_Status> stats(4);
  MPI_Waitall(4, reqs, &stats[0]);

  DBG5(MPI_Barrier(topocomm); if(!r) printf("finished rearranging edge connectivity, starting phase two!\n"));

  /* reset all data structures for new permutation */ 
  sreqs.resize(rout.size());
  rreqs.resize(rin.size());

  sbufs.resize(rout.size());
  for(int i=0; i<sbufs.size(); ++i) sbufs[i].resize(mult*routw[i]);
  rbufs.resize(rin.size());
  for(int i=0; i<rbufs.size(); ++i) rbufs[i].resize(mult*rinw[i]);

  MPI_Barrier(topocomm);

  for(int i=0; i < trials; ++i) {
    if(i == 1) {
      t[1] = -MPI_Wtime();
    }
    for(int j=0; j<rin.size(); ++j) {
      MPI_Irecv(&rbufs[j][0], rinw[j]*mult, MPI_BYTE, rin[j], 0, reorderedcomm, &rreqs[j]);
    }
    for(int j=0; j<rout.size(); ++j) {
      MPI_Isend(&sbufs[j][0], routw[j]*mult, MPI_BYTE, rout[j], 0, reorderedcomm, &sreqs[j]);
    }
    // silly AMPI doesn't support MPI_STATUSES_IGNORE
    std::vector<MPI_Status> rstats(rreqs.size());
    std::vector<MPI_Status> sstats(sreqs.size());
    MPI_Waitall(rreqs.size(), &rreqs[0], &rstats[0]);
    MPI_Waitall(sreqs.size(), &sreqs[0], &sstats[0]);
  }
  t[1] += MPI_Wtime();

  double rt[2];
  MPI_Allreduce(t, rt, 2, MPI_DOUBLE, MPI_MAX, topocomm);

  *before=rt[0];
  *after=rt[1];
}

/* this function is supposed to perform on-node mapping, it is called
 * after the global communication graph has been mapped to the network
 * topology (this mapping assumed infinite in-node bandwidth). This
 * second stage gets the on-node sub-graphs as input and maps the
 * processes to cores on NUMA nodes, the inputs are:
 *   - ltg: the logical (application) graph topology as adjacency list,
 *     this is the on-node subgraph of the global graph
 *   - ltgw: the weights of the on-node graph topology (by whatever
 *     metric they're provided to MPI)
 *   - idx2rank: an array to translate vertices in ltg to ranks in the
 *     MPI communicator comm
 *   - comm: the MPI topology communicator for the mapping, can be used
 *     for MPI communication
 *   - r: the rank of the local process in comm
 * 
 * This function returns the *NEW* rank of the current process in the
 * topology mapping.
 */
int TPM_Node_mapping(TPM_Graph* ltg_ref, TPM_Graph_ewgts *ltgw_ref, TPM_Mapping* idx2rank_ref, MPI_Comm comm, int r) {
  TPM_Graph& ltg = *ltg_ref;
  TPM_Graph_ewgts& ltgw = *ltgw_ref;
  TPM_Mapping& idx2rank = *idx2rank_ref;

  printf("[%i] TPM_Node_mapping %lu neighbors\n", r, ltg.size()); fflush(stdout);
  MPI_Barrier(MPI_COMM_WORLD); // this is horrible
  // print the on-node graph adjacency list 
  if(0 == r) {
    for(int i=0; i<ltg.size(); ++i) {
      printf("%i: (rank %i): ", i, idx2rank[i]);
      for(int j=0; j<ltg[i].size(); ++j) {
        printf("%i ", ltg[i][j]);
      }
      printf("\n");
    }
  }

  // identity mapping for now -- should change :-)
  return r;
}
