/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#include <sstmac/software/libraries/topomap/topomap_api.h>
#include <sstmac/hardware/network/topology/topology.h>
#include <sstmac/hardware/network/interconnect/switchinterconnect.h>
#include <sstmac/software/libraries/topomap/libtopomap.hpp>
#include <sstmac/software/api/mpi/sstmac_mpi.h>
#include <sstmac/common/fileio.h>
#include <sstmac/common/types.h>
#include <deque>
ImplementAPI(sstmac::sw, topomap_api, "topomap")

namespace sstmac
{
  namespace sw
  {
    SSTRegister("topomap", api, topomap_api,
        "Create bindings for topomap optimizer");

    debug topomap_api::dbg_;

    using namespace sstmac::hw;

    topomap_api::topomap_api(std::string f) :
      app_fname_(f)
    {
      if (!dbg_.is_real())
      {
        dbg_ = sstmac::debug("<debug> topomap", sstmac::debug::magenta);
        sstmac::logger::register_logger(&dbg_);
      }
    }

    void
    topomap_api::make_system_graph(const char *sys_graph_file)
    {
      interconnect::ptr intercon = get_interconnect();
      switchinterconnect::ptr swin = boost::dynamic_pointer_cast<
          switchinterconnect>(intercon);
      topology::maptype sw = swin->get_switches();
      topology::ptr top = get_topology();

      std::ofstream out;
      out.open(sys_graph_file, std::ios::out);

      int n2rfactor = 10; //this number is the weight of node-switch connection, relative to redundant

      if (out.is_open())
      {

        long numnodes = top->num_nodes();
        long numsw = sw.size();
        long edges = top->radix() * numsw / 2; //divide by two to make it unidirectional

        out << "num: " << (numnodes + numsw) << "w" << std::endl;

        //first print out all their names (dumb)
        for (long a = 0; a < numnodes; a++)
        {
          out << a << " " << "node" << a << "\n";
        }

        for (long a = 0; a < numsw; a++)
        {
          out << (numnodes + a) << " " << "switch" << a << "\n";
        }

        //then print out their connections
        for (long a = 0; a < sw.size(); a++)
        {
          nodeaddress::ptr addr = nodeaddress::convert_from_uniqueid(a);
          networkswitch::ptr net = sw[addr];
          discrete_network_switch::ptr dnet = boost::dynamic_pointer_cast<
              discrete_network_switch>(net);
          if (!dnet)
          {
            sst_throw(ssterror, "topomap_api::make_system_graph - switch didn't cast to discrete_network_switch");
          }
          std::vector<int> redund = dnet->get_connected_redund();

          std::vector<long> nodes = top->get_nodes_connected_to_switch(addr);

          out << (numnodes + a) << " ";

          std::vector<nodeaddress::ptr> con = net->get_connected_switches();
          for (int i = 0; i < con.size(); i++)
          {
            out << (numnodes + con[i]->unique_id()) << "(" << redund[i] << ") ";
          }

          for (int i = 0; i < nodes.size(); i++)
          {
            out << nodes[i] << "(" << n2rfactor << ") ";
          }

          out << std::endl;

          for (int i = 0; i < nodes.size(); i++)
          {
            out << nodes[i] << " " << (numnodes + a) << "(" << n2rfactor << ") \n";
          }
        }
        out.close();
      }
    }

    void
    topomap_api::read_app_graph(std::string app_graph_file,
        std::vector<int> &sources, std::vector<int> &degrees,
        std::vector<int> &destinations, std::vector<int> &weights, int &nranks)
    {
      if (app_graph_file == "")
      {
        sst_throw(valueerror, "topomap_api::read_app_graph - topomap_app_graph_file wasn't set. must be set to use libtopomap");
      }
      std::ifstream in;
      SSTFileIO::open_file(in, app_graph_file);

      if (in.is_open())
      {
        std::deque<std::string> tok;

        std::string space = " ";

        std::string line;

        //first, take out the num of ranks
        std::getline(in, line);
        line = trim_str(line);

        sscanf(line.c_str(), "num: %dw", &nranks);

        SSTMAC_DEBUG << "topomap_app: found " << nranks
              << " in app graph file \n";
        //skip over next n lines
        for (int i = 0; i < nranks; i++)
        {
          std::getline(in, line);
        }

        for (int i = 0; i < nranks && in.good(); i++)
        {
          SSTMAC_DEBUG3 << "iteration " << i << "\n";
          std::deque<std::string> toks;
          std::getline(in, line);
          line = trim_str(line);
          SSTMAC_DEBUG3 << "line: " << line << "\n";

          pst::BasicStringTokenizer::tokenize(line, toks, space);
          std::deque<std::string>::iterator it, end = toks.end();

          int degree = 0;
          for (it = toks.begin(); it != end; it++)
          {
            if (degree == 0)
            {
              int src = atoi((*it).c_str());
              sources.push_back(src);
            }
            else
            {
              int dst;
              int weight;
              sscanf((*it).c_str(), "%d(%d)", &dst, &weight);
              SSTMAC_DEBUG2 << "dst=" << dst << ", weight=" << weight << "\n";
              destinations.push_back(dst);
              weights.push_back(weight);
            }
            ++degree;
          }

          SSTMAC_DEBUG2 << "degree: " << degree << "\n";

          degrees.push_back(degree - 1);
        }
        in.close();
      }
      else
      {
        sst_throw(ssterror, "topomap_app::TPM_Init: can't open file %s",
            app_graph_file.c_str());
      }
    }

    void
    topomap_api::make_allocation_file(const char* fakefname)
    {
      thread::ptr t = OS_API::get_current_thread();
      environment::const_ptr env = t->get_os()->get_env(t->aid());

      std::ofstream myfile;
      myfile.open(fakefname, std::ios::out);
      if (!myfile.is_open())
      {
        sst_throw(ioerror,
            "topomap_app: can't open allocation file for writing.", fakefname);
      }

      size_t s = env->get_app_count();
      for (int i = 0; i < s; i++)
      {
        taskid tid(i);
        nodeaddress::ptr n = env->get_node_address(tid);
        myfile << i << " node" << n->unique_id() << "\n";
      }

      myfile.close();
    }

    void
    topomap_api::tpm_init(bool usetopo)
    {
      MPI_Init(NULL, NULL);

      int num_ranks;
      int my_rank;

      MPI_Comm_size(MPI_COMM_WORLD, &num_ranks);
      MPI_Comm_rank(MPI_COMM_WORLD, &my_rank);

      if (usetopo)
      {
        const char *sys_graph_file = "TPM_sys_graph_file.graph";

        //create allocation file
        if (TPM_Fake_names_file == NULL)
        {

          char* fakefname = "temp_fakefnamefile.txt";

          make_allocation_file(fakefname);

          SSTMAC_DEBUG << "Setting TPM_Fake_names_file to: " << fakefname
                << "\n";

          TPM_Fake_names_file = fakefname;
        }

        //printf("num_ranks=%d, my_rank=%d, sys_graph_file=%s, app_graph_file=%s\n", num_ranks, my_rank, sys_graph_file, app_graph_file);
        //  MPI_Barrier(MPI_COMM_WORLD);

        // Step 1: Build an MPI Graph Communicator for the input application communication graph

        std::vector<int> sources, degrees, destinations, weights;

        int nranks = 0;
        if (my_rank == 0)
        {

          read_app_graph(app_fname_, sources, degrees, destinations, weights,
              nranks);

          make_system_graph(sys_graph_file);

        }

        MPI_Barrier(MPI_COMM_WORLD);

#if 0
        int k = 0;
        for (int i = 0; i < nranks; i++)
        {
          printf("%d ", sources[i]);
          for (int j = 0; j < degrees[i]; j++)
          {
            printf("%d(%d) ", destinations[k], weights[k]);
            ++k;
          }
          printf("\n");
        }

        //    MPI_Barrier(MPI_COMM_WORLD);
        //  MPI_Finalize();
        //  return EXIT_SUCCESS;
#endif

        // Create an MPI communicator representing the input app communication graph
        int status;
        MPI_Comm distgr;
        status = MPIX_Dist_graph_create(MPI_COMM_WORLD, nranks, &sources[0],
            &degrees[0], &destinations[0],
            (&weights[0]) ? &weights[0] : (int *) 1, MPI_INFO_NULL, 1, &distgr);
        if (status != MPI_SUCCESS)
        {
          printf("MPIX_Dist_graph_create failed.\n");
          exit(-1);
        }

        // Call libtopomap several times, each using a different strategy

        //   setenv("TPM_STRATEGY", "greedy", 1);
        setenv("TPM_STRATEGY", "recursive", 1);
        //   setenv("TPM_STRATEGY", "rcm", 1);


        int new_rank;
        int ret = TPM_Topomap(distgr, sys_graph_file, 0, &new_rank);

        if (ret != 0)
        {
          printf("topomap failed -- aborting\n");
          MPI_Abort(MPI_COMM_WORLD, 1);
          return;
        }

        SSTMAC_DEBUG << "old rank: " << my_rank << " --> new rank: "
              << new_rank << "\n";

        rank_ = new_rank;

        MPI_Comm_split(distgr, 0, new_rank, &comm_);

      }
      else //in case we want to disable TPM, and have things still work
      {
        comm_ = MPI_COMM_WORLD;
        rank_ = my_rank;
      }

    }

  }

}
