/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/process/app.h>
#include <sstmac/software/process/api.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/api/mpi/mpiapi.h>
#include <sstmac/software/services/launch/launchinfo.h>

ImplementFactory(sstmac::sw::app);

namespace sstmac
{
  namespace sw
  {

    debug app::dbg_;
    std::vector<char**> app::argv_for_app_;
    std::vector<int> app::argc_for_app_;

    void
    app_factory::print_apps()
    {
        std::map<std::string, SSTDesc_base*>::const_iterator
                it = object_map_->begin(), end = object_map_->end();
        std::cout << "Valid SST/macro apps are:\n";
        for ( ; it != end; ++it){
            std::cout << it->first << "\n";
        }
        std::cout << std::endl;
    }

    void
    app_factory::clear_apps()
    {
        std::map<std::string, SSTDesc_base*>::const_iterator
                it = object_map_->begin(), end = object_map_->end();
        for ( ; it != end; ++it){
            it->second->clear();
        }
    }

    app::app(const sim_parameters::ptr &params)
        : params_(params)
    {
    }

    app::app(softwareid id) :
      thread(id), id_(id)
    {
      if (!dbg_.is_real())
        {
          dbg_ = debug("<debug> thread | app", logger::green);
          logger::register_logger(&dbg_);
        }
    }

    app::~app()
    {
    }

    void
    app::set_launch_info(const launchinfo::ptr &linf)
    {
      linfo_ = linf;
    }

    lib_compute_loops::ptr
    app::get_compute_loops_lib()
    {
      if(!computeloops_){
          computeloops_ = lib_compute_loops::construct(id_);
          register_lib(computeloops_);
      }

      return computeloops_;
    }

    void
    app::kill()
    {
      linfo_ = 0;
      thread::kill();

      int appnum = id_.app_.id;
      char** argv = (int) argv_for_app_.size() > appnum ? argv_for_app_[appnum]
          : 0;

      if (argv)
        {
          char* argv_buffer = argv[0];
          delete[] argv;
          delete[] argv_buffer;
          argv_for_app_[appnum] = 0;
        }

    }

    launchinfo::ptr
    app::get_launch_info()
    {
      return linfo_;
    }

    void
    app::compute(timestamp time)
    {

      if (!computetime_)
      {
          computetime_ = lib_compute_time::construct(id_);
          computetime_->init_as_multithreaded_lib();
          register_lib(computetime_);
      }

      computetime_->compute(time);
    }

    void
    app::compute_inst(const eventdata::ptr &cmsg)
    {
      if (!computeinst_)
        {
          computeinst_ = lib_compute_inst::construct(id_);
          register_lib(computeinst_);
        }
      computeinst_->compute_inst(cmsg);
    }

    void
    app::compute_block_read(long bytes)
    {
      if (!computememmove_)
        init_mem_lib();

      computememmove_->read(bytes);
    }

    void
    app::compute_block_write(long bytes)
    {
      if (!computememmove_)
        init_mem_lib();

      computememmove_->write(bytes);
    }

    void
    app::init_argv()
    {
      int appnum = id_.app_.id;
      /** haven't yet computed argv for this app */
      if ((int) argv_for_app_.size() <= appnum || argv_for_app_[appnum] == 0)
        {
          for (int idx = argv_for_app_.size(); idx <= appnum; ++idx)
            {
              argv_for_app_.push_back(0);
              argc_for_app_.push_back(0);
            }
          char param_str[64];
          sprintf(param_str, "launch_app%d", appnum);
          std::string appname = params_->get_param(param_str);

          sprintf(param_str, "launch_app%d_argv", appnum);
          std::vector<std::string> argv_param_vec;
          argv_param_vec.push_back(appname);
          if (params_->has_param(param_str))
            {
              params_->get_vector_param(param_str, argv_param_vec);
            }
          int argc = argv_param_vec.size();
          char* argv_buffer = new char[256 * argc];
          char* argv_buffer_ptr = argv_buffer;
          char** argv = new char*[argc];
          for (int i = 0; i < argc; ++i)
            {
              const std::string& src_str = argv_param_vec[i];
              ::strcpy(argv_buffer_ptr, src_str.c_str());
              argv[i] = argv_buffer_ptr;
              //increment pointer for next strcpy
              argv_buffer_ptr += src_str.size() + 1; //+1 for null terminator
            }
          argv_for_app_[appnum] = argv;
          argc_for_app_[appnum] = argc;
        }
    }

    int
    app::get_argc()
    {
      init_argv();
      return argc_for_app_[id_.app_.id];
    }

    char**
    app::get_argv()
    {
      init_argv();
      return argv_for_app_[id_.app_.id];
    }

    void
    app::init_mem_lib()
    {
      computememmove_ = lib_compute_memmove::construct(id_);
      computememmove_->init_as_multithreaded_lib();
      register_lib(computememmove_);
    }

    void
    app::compute_block_memcpy(long bytes)
    {
      if (!computememmove_)
        init_mem_lib();

      computememmove_->copy(bytes);
    }

    void
    app::sleep(timestamp time)
    {
      if (!sleeplib_)
        {
          sleeplib_ = lib_sleep::construct(id_);
          register_lib(sleeplib_);
        }
      sleeplib_->sleep(time);
    }

    ptr_type::ptr
    app::build_api(int aid, const std::string &name)
    {
        // an underlying thread may have built this
        ptr_type::ptr& my_api = apis_[aid];
        if (!my_api){
            api::ptr new_api = api_factory::get_value(name, sstmac_env::params, id_);
            register_lib(new_api);
            my_api = new_api;
        }
        return my_api;
    }

    void
    app::init_os(const OS_API::ptr &os)
    {
        if (params_->has_param("startup_apis")){
            std::vector<std::string> apis;
            /**
                sstkeyword {
                    docstring=In some advanced cases, certains APIs (MPI, GNI, SHMEM) might need to be
                    launched manually.  In most cases, though, SST/macro automatically initializes
                    the API upon first usaage.;
                }
            */
            params_->get_vector_param("startup_apis", apis);
            for (int i=0; i < apis.size(); ++i){
                const std::string& libname = apis[i];
                api::ptr new_api = api_factory::get_value(libname, params_, id_);
                register_lib(new_api);
                int api_id = APIBase::api_id(libname);
                apis_[api_id] = new_api;
            }
        }
        else if (params_->has_param("app_startup_libs")){
            std::vector<std::string> libs;
            /**
                sstkeyword {
                    docstring=In some advanced cases, certains libraries might need to be
                    loaded manually.  In most cases, though, SST/macro automatically initializes
                    the library upon first usaage.;
                }
            */
            params_->get_vector_param("app_startup_libs", libs);
            for (int i=0; i < libs.size(); ++i){
                const std::string& libname = libs[i];
                library::ptr lib = library::construct_lib(libname, id_);
                register_lib(lib);
            }
        }
    }

    std::string
    app::get_compute_name()
    {

      if (!computetime_)
        {
          computetime_ = lib_compute_time::construct(id_);
          register_lib(computetime_);
        }
      return computetime_->get_lib_name();
    }

    std::string
    app::get_compute_inst_name()
    {
      if (!computeinst_)
        {
          computeinst_ = lib_compute_inst::construct(id_);
          register_lib(computeinst_);
        }
      return computeinst_->get_lib_name();
    }

    std::string
    app::get_compute_memmove_name()
    {
      if (!computememmove_)
        {
          computememmove_ = lib_compute_memmove::construct(id_);
          register_lib(computememmove_);
        }
      return computememmove_->get_lib_name();
    }

    std::string
    app::get_sleep_name()
    {
      if (!sleeplib_)
        {
          sleeplib_ = lib_sleep::construct(id_);
          register_lib(sleeplib_);
        }
      return sleeplib_->get_lib_name();
    }

    softwareid
    app::get_sid() const
    {
      return id_;
    }

    int
    app::get_appnum() const
    {
      return id_.app_.id;
    }

    sim_parameters::ptr
    app::parameters()
    {
      return params_;
    }

    void
    app::init_libs(sim_parameters::ptr params)
    {
      params_ = params;
    }

    void
    app::run()
    {
      logger::check_user_params();
      skeleton_main();
    }

    void compute_time(double tsec){
        thread::ptr t = OS_API::get_current_thread();
        app::ptr a = boost::dynamic_pointer_cast<app>(t);
        if (!a){
            sst_throw(illformederror, "cannot cast current thread to app in compute_time function");
        }
        a->compute(timestamp(tsec));
    }

  }
}
