#include <sstmac/software/process/ftq.h>
#include <sstmac/common/debug.h>
namespace sstmac {
    namespace sw {


    FTQCalendar::FTQCalendar()
        : stat_logger("<stats> ftq | intensity", "ftq.csv"),
          num_ticks_epoch_(0),
          show_vis_(false),
          show_vis_thread_chart_(false),
          show_vis_thread_chart_type_("bar"),
          show_vis_thread_chart_separate_(true),
          show_vis_summary_chart_(false)
    {
    }

    void
    FTQCalendar::init(
            long nticks_per_epoch,
            bool show_vis,
            bool show_vis_thread_chart,
            const std::string& show_vis_thread_chart_type,
            bool show_vis_thread_chart_separate,
            bool show_vis_summary_chart)
    {
        num_ticks_epoch_ = nticks_per_epoch;
        show_vis_ = show_vis;
        show_vis_thread_chart_ = show_vis_thread_chart;
        show_vis_thread_chart_type_ = show_vis_thread_chart_type;
        show_vis_thread_chart_separate_ = show_vis_thread_chart_separate;
        show_vis_summary_chart_ = show_vis_summary_chart;
    }

    FTQCalendar::ptr
    FTQCalendar::construct()
    {
        return new FTQCalendar();
    }

    void
    FTQCalendar::register_app(int aid, const std::string& appname)
    {
        calendars_[aid] = new AppFTQCalendar(aid, appname, num_ticks_epoch_, show_vis_,
        		show_vis_thread_chart_, show_vis_thread_chart_type_, show_vis_thread_chart_separate_, show_vis_summary_chart_);
    }

    void
    FTQCalendar::simulation_finished(timestamp t_end)
    {
        if (!active_)
            return;

        boost::unordered_map<int, AppFTQCalendar::ptr>::iterator it = calendars_.begin(), end = calendars_.end();
        for ( ; it != end; ++it){
            it->second->dump();
        }
    }

    void
    FTQCalendar::collect(int event_typeid, int aid, int tid, long ticks_begin, long num_ticks)
    {
        calendars_[aid]->collect(event_typeid, tid, ticks_begin, num_ticks);
    }

    AppFTQCalendar::ptr
    FTQCalendar::get_calendar(int aid) const
    {
        boost::unordered_map<int, AppFTQCalendar::ptr>::const_iterator it = calendars_.find(aid);
        if (it == calendars_.end()){
            sst_throw(valueerror, "no FTQ calendar found for app %d", aid);
        }
        return it->second;
    }

    AppFTQCalendar::AppFTQCalendar(int aid,
           const std::string& appname,
            long nticks_epoch,
            bool show_vis, bool show_vis_thread_chart,
            const std::string& show_vis_thread_chart_type,
            bool show_vis_thread_chart_separate,
            bool show_vis_summary_chart)
        : epochs_(0),
          aid_(aid),
          max_tid_(0),
          max_epoch_(0),
          num_ticks_epoch_(nticks_epoch),
          appname_(appname), show_vis_(show_vis),
          show_vis_thread_chart_(show_vis_thread_chart),
          show_vis_thread_chart_type_(show_vis_thread_chart_type),
          show_vis_thread_chart_separate_(show_vis_thread_chart_separate),
          show_vis_summary_chart_(show_vis_summary_chart),
          buffers_(0)
    {
        int num_categories = key::num_categories();
        totals_ = new long long[num_categories];
        for (int i=0; i < num_categories; ++i){
            totals_[i] = 0;
        }

#ifdef HAVE_VTK
        if (show_vis_ && view == 0 && chart == 0) {
        	view = vtkSmartPointer<vtkContextView>::New();
        	chart = vtkSmartPointer<vtkChartXY>::New();
        	view->GetRenderWindow()->SetSize(500, 500);
        	view->GetScene()->AddItem(chart);

        	chart->SetTitle("CPU and MPI Time for App: " + appname_);
        	chart->GetAxis(vtkAxis::LEFT)->SetTitle("Thread ID");
            chart->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch(us)");

        	chart->SetShowLegend(true);
        	chart->GetLegend()->SetInline(false);
        	chart->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
        	chart->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);

        	view->GetRenderWindow()->SetMultiSamples(0);
        	if (!show_vis_summary_chart_)
        		view->GetInteractor()->Initialize();
        }
#endif
    }

    void
    AppFTQCalendar::allocate_epochs(long max_epoch)
    {
        if (max_epoch < max_epoch_allocated_)
            return;

        long epoch_start = max_epoch_allocated_;
        max_epoch_allocated_ += allocation_num_epochs;
        epochs_.resize(max_epoch_allocated_);

        int num_categories = key::num_categories();
        long long* buffer = new long long[num_categories * allocation_num_epochs];
        long long* buffer_ptr = buffer;
        for (long epoch=epoch_start; epoch <= max_epoch_allocated_; buffer_ptr += num_categories, ++epoch){
            epochs_[epoch].init(num_categories, buffer_ptr);
        }
        buffers_.push_back(buffer);
    }

    void
    AppFTQCalendar::collect(int event_typeid, int tid, long ticks_begin, long num_ticks)
    {
        #if 0
        TaskFTQCalendar::ptr& cal = calendars_[tid];
        if (!cal){
            cal = new TaskFTQCalendar();
            if (tid > max_tid_){
                max_tid_ = tid;
            }
        }

        FTQEpoch*& t_epoch = thread_epochs_[tid];
        if (!t_epoch){
        	t_epoch = new FTQEpoch[num_epochs_];
            if (tid > max_tid_){
                max_tid_ = tid;
            }
            t_epoch->init(key::num_categories(), new long long[key::num_categories()]);
        }
        cal->collect(event_typeid, ticks_begin, num_ticks);
        #endif
        /** aggregate for all time intervals and all threads for given event type */
        totals_[event_typeid] += num_ticks;

        long ticks_end = ticks_begin + num_ticks;
        long max_epoch = ticks_end / num_ticks_epoch_;
        allocate_epochs(max_epoch);

        long first_epoch = ticks_begin / num_ticks_epoch_;
        long nticks_first_epoch = num_ticks_epoch_ - ticks_begin % num_ticks_epoch_;
        // this might not go until the end of the epoch
        nticks_first_epoch = std::min(nticks_first_epoch, num_ticks);
        /** aggregate for all threads for given time interval and event type */
        epochs_[first_epoch].collect(event_typeid, nticks_first_epoch);
        num_ticks -= nticks_first_epoch;
        //t_epoch[epoch].collect(event_typeid, nticks_next_epoch);
        for (long epoch=first_epoch+1; epoch <= max_epoch; ++epoch){
            long nticks_this_epoch = std::min(num_ticks, num_ticks_epoch_);
            epochs_[epoch].collect(event_typeid, nticks_this_epoch);
            //t_epoch[epoch].collect(event_typeid, nticks_next_epoch);
            num_ticks -= nticks_this_epoch;
        }

        if (max_epoch > max_epoch_){
            if (show_vis_ && !show_vis_summary_chart_){
                display_charts(tid, max_epoch_);
            }
            max_epoch_ = max_epoch;
        }

    }

    void
    AppFTQCalendar::display_charts(int tid, long max_epoch){

#ifdef HAVE_VTK
    	// Data structure for main chart with all threads showing
    	std::map<int, vtkTable*> table_cpu_map;
    	std::map<int, vtkLongArray*> X_cpu_map;
    	std::map<int, vtkLongArray*> Y_cpu_map;
    	std::map<int, vtkPlot *> line_cpu_map;

    	// Data structure for per thread chart on separate window
    	std::map<int, vtkTable*> table_thread_cpu_map;
    	std::map<int, vtkLongArray*> X_thread_cpu_map;
    	std::map<int, vtkLongArray*> Y_thread_cpu_map;
    	std::map<int, vtkPlot *> thread_cpu_plot_map;
    	long cpu_max[max_tid_];

    	// Data structure for main chart with all threads showing
    	std::map<int, vtkTable*> table_mpi_map;
    	std::map<int, vtkLongArray*> X_mpi_map;
    	std::map<int, vtkLongArray*> Y_mpi_map;
    	std::map<int, vtkPlot *> line_mpi_map;
    	long mpi_max[max_tid_];

    	long epoch_max[max_tid_];
    	long maximum_epoch = 0;

    	// Data structure for per thread chart on separate window
    	std::map<int, vtkTable*> table_thread_mpi_map;
    	std::map<int, vtkLongArray*> X_thread_mpi_map;
    	std::map<int, vtkLongArray*> Y_thread_mpi_map;
    	std::map<int, vtkPlot *> thread_mpi_plot_map;

    	// Data structure for per thread chart upto 9 threads per window
     	std::map <int, vtkTable*> table_both_map;
    	std::map <int, vtkFloatArray*> X_both_map;
    	std::map <int, vtkFloatArray*> Y_both_cpu_map;
    	std::map <int, vtkFloatArray*> Y_both_mpi_map;
    	std::map <int, vtkPlotStacked *> line_both_map;

    	chart->ClearPlots();

    	for (int i = 0; i <= max_tid_; ++i) {
    		FTQEpoch* t_epoch = thread_epochs_[i];
    		if (t_epoch) {
    			table_cpu_map.insert(std::make_pair(i, vtkTable::New()));
    			X_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
    			Y_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
    			// CPU will be shown points with CROSS characteristics (X mark)
    			line_cpu_map.insert(
    					std::make_pair(i, chart->AddPlot(vtkChart::POINTS)));
    			vtkPlotPoints::SafeDownCast(line_cpu_map[i])->SetMarkerStyle(vtkPlotPoints::CROSS);
    			X_cpu_map[i]->SetName("X");
    			table_cpu_map[i]->AddColumn(X_cpu_map[i]);
    			Y_cpu_map[i]->SetName("Compute Time");
    			table_cpu_map[i]->AddColumn(Y_cpu_map[i]);

    			table_mpi_map.insert(std::make_pair(i, vtkTable::New()));
    			X_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
    			Y_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
    			// MPI will be shown solid points (dot)
    			line_mpi_map.insert(
    					std::make_pair(i, chart->AddPlot(vtkChart::POINTS)));
    			X_mpi_map[i]->SetName("X");
    			table_mpi_map[i]->AddColumn(X_mpi_map[i]);
    			Y_mpi_map[i]->SetName("MPI Time");
    			table_mpi_map[i]->AddColumn(Y_mpi_map[i]);

    			cpu_max[i] = 0;
    			mpi_max[i] = 0;
    			epoch_max[i] = 0;

    			if (show_vis_thread_chart_) {
    				if (show_vis_thread_chart_separate_) {
						if (view_thread.find(i) == view_thread.end()) { // If it is first time being called, init data structures
							view_thread.insert(std::make_pair(i, vtkContextView::New()));
							chart_thread.insert(std::make_pair(i, vtkChartXY::New()));
							view_thread[i]->GetRenderWindow()->SetSize(500, 500);
							view_thread[i]->GetScene()->AddItem(chart_thread[i]);

							chart_thread[i]->SetTitle(
									"CPU and MPI Time for Thread: " + boost::lexical_cast<std::string>(i));
							chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetTitle("Number of Ticks");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch (us)");

							chart_thread[i]->SetShowLegend(true);
							// Show legend outside the chart box
							chart_thread[i]->GetLegend()->SetInline(false);
							chart_thread[i]->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
							chart_thread[i]->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);

							view_thread[i]->GetRenderWindow()->SetMultiSamples(0);
							view_thread[i]->GetInteractor()->Initialize();
						}
						else
							chart_thread[i]->ClearPlots();
    				}
    				else { // show_vis_thread_chart_separate_ == false. Show upto 9 threads in one chart
    					if (view_thread.find(i/9) == view_thread.end()) { // If it is first time being called, init data structures
							view_thread.insert(std::make_pair(i/9, vtkContextView::New()));
							view_thread[i/9]->GetRenderWindow()->SetSize(900, 900);

							view_thread[i/9]->GetRenderWindow()->SetMultiSamples(0);
							view_thread[i/9]->GetInteractor()->Initialize();
    					}
						if (chart_thread.find(i) == chart_thread.end()) {
							chart_thread.insert(std::make_pair(i, vtkChartXY::New()));
							view_thread[i/9]->GetScene()->AddItem(chart_thread[i]);

							chart_thread[i]->SetTitle(
									"Thread: " + boost::lexical_cast<std::string>(i));
							chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetTitle("Number of Ticks");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetTitle("Epoch (us)");
							chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetBehavior(vtkAxis::FIXED);
							chart_thread[i]->SetAutoSize(false);
							chart_thread[i]->SetSize(vtkRectf((i%3)*300, ((8-(i%9))/3)*300, 300.0, 300.0));

							chart_thread[i]->SetShowLegend(true);
							chart_thread[i]->GetLegend()->SetInline(true);
							chart_thread[i]->GetLegend()->SetHorizontalAlignment(vtkChartLegend::RIGHT);
		 					chart_thread[i]->GetLegend()->SetVerticalAlignment(vtkChartLegend::TOP);
						}
    				}

    				if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
						chart_thread[i]->ClearPlots();
						if (show_vis_thread_chart_type_ == "area") {
							table_both_map.insert(std::make_pair(i, vtkTable::New()));
							X_both_map.insert(std::make_pair(i, vtkFloatArray::New()));
							Y_both_cpu_map.insert(std::make_pair(i, vtkFloatArray::New()));
							Y_both_mpi_map.insert(std::make_pair(i, vtkFloatArray::New()));
							line_both_map.insert(std::make_pair(i, vtkPlotStacked::SafeDownCast(chart_thread[i]->AddPlot(vtkChart::STACKED))));

							X_both_map[i]->SetName("X");
							table_both_map[i]->AddColumn(X_both_map[i]);
							Y_both_cpu_map[i]->SetName("CPU");
							table_both_map[i]->AddColumn(Y_both_cpu_map[i]);
							Y_both_mpi_map[i]->SetName("MPI");
							table_both_map[i]->AddColumn(Y_both_mpi_map[i]);
						} else { // if chart type is bar or line
							table_thread_cpu_map.insert(std::make_pair(i, vtkTable::New()));
							X_thread_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
							Y_thread_cpu_map.insert(std::make_pair(i, vtkLongArray::New()));
							if (show_vis_thread_chart_type_ == "bar") {
								thread_cpu_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::BAR)));
							} else if (show_vis_thread_chart_type_ == "line") {
								thread_cpu_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::LINE)));
							}
							X_thread_cpu_map[i]->SetName("X");
							table_thread_cpu_map[i]->AddColumn(X_thread_cpu_map[i]);
							Y_thread_cpu_map[i]->SetName("Compute Time");
							table_thread_cpu_map[i]->AddColumn(Y_thread_cpu_map[i]);

							table_thread_mpi_map.insert(std::make_pair(i, vtkTable::New()));
							X_thread_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
							Y_thread_mpi_map.insert(std::make_pair(i, vtkLongArray::New()));
							if (show_vis_thread_chart_type_ == "bar") {
								thread_mpi_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::BAR)));
							} else if (show_vis_thread_chart_type_ == "line") {
								thread_mpi_plot_map.insert(
										std::make_pair(i, chart_thread[i]->AddPlot(vtkChart::LINE)));
							}
							X_thread_mpi_map[i]->SetName("X");
							table_thread_mpi_map[i]->AddColumn(X_thread_mpi_map[i]);
							Y_thread_mpi_map[i]->SetName("MPI Time");
							table_thread_mpi_map[i]->AddColumn(Y_thread_mpi_map[i]);
						}
    				}
    			}
    		}
    	}

    	// Fill in the data that will be plotted in the chart
    	for (int i = 0; i < max_epoch; i++) {
    		for (int j = 0; j <= max_tid_; ++j) {
    			FTQEpoch* t_epoch = thread_epochs_[j];
    			if (t_epoch) {
    				long nticks_cpu = t_epoch[i].get_event_time(key::compute);
    				long nticks_mpi = t_epoch[i].get_event_time(key::mpi);
    				if (nticks_cpu > nticks_mpi && nticks_cpu > 0) {
    					table_cpu_map[j]->InsertNextBlankRow();
					    table_cpu_map[j]->SetValue(table_cpu_map[j]->GetNumberOfRows()-1, 0, i);
    					table_cpu_map[j]->SetValue(table_cpu_map[j]->GetNumberOfRows()-1, 1, j);
    					if (i > maximum_epoch)
    						maximum_epoch = i;
    				}
    				if (nticks_mpi > nticks_cpu && nticks_mpi > 0) {
    					table_mpi_map[j]->InsertNextBlankRow();
					    table_mpi_map[j]->SetValue(table_mpi_map[j]->GetNumberOfRows()-1, 0, i);
    					table_mpi_map[j]->SetValue(table_mpi_map[j]->GetNumberOfRows()-1, 1, j);
    					if (i > maximum_epoch)
    						maximum_epoch = i;
    				}

    				if (show_vis_thread_chart_) {
    					if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
    						if (show_vis_thread_chart_type_ == "area") {
    				        	table_both_map[j]->InsertNextBlankRow();
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 0, table_both_map[j]->GetNumberOfRows()-1);
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 1, nticks_cpu);
    				        	table_both_map[j]->SetValue(table_both_map[j]->GetNumberOfRows()-1, 2, nticks_mpi);
								if (cpu_max[j] < nticks_cpu)
									cpu_max[j] = nticks_cpu;
								if (mpi_max[j] < nticks_mpi)
									mpi_max[j] = nticks_mpi;
								if (epoch_max[j] < i)
									epoch_max[j] = i;

    						} else { // chart type is bar or line
								if (nticks_cpu > 0) {
									table_thread_cpu_map[j]->InsertNextBlankRow();
									table_thread_cpu_map[j]->SetValue(table_thread_cpu_map[j]->GetNumberOfRows()-1, 0, i);
									table_thread_cpu_map[j]->SetValue(table_thread_cpu_map[j]->GetNumberOfRows()-1, 1, nticks_cpu);
									if (cpu_max[j] < nticks_cpu)
										cpu_max[j] = nticks_cpu;
									if (epoch_max[j] < i)
										epoch_max[j] = i;
								}
								if (nticks_mpi > 0) {
									table_thread_mpi_map[j]->InsertNextBlankRow();
									table_thread_mpi_map[j]->SetValue(table_thread_mpi_map[j]->GetNumberOfRows()-1, 0, i);
									table_thread_mpi_map[j]->SetValue(table_thread_mpi_map[j]->GetNumberOfRows()-1, 1, nticks_mpi);
									if (mpi_max[j] < nticks_mpi)
										mpi_max[j] = nticks_mpi;
									if (epoch_max[j] < i)
										epoch_max[j] = i;
								}
    						}
    					}
					}
    			}
    		}
    	}

    	// Display the chart(s)
    	bool labeled = false;
    	for (int i = 0; i <= max_tid_; ++i) {
    		FTQEpoch* t_epoch = thread_epochs_[i];
    		if (t_epoch) {
    			  line_cpu_map[i]->SetInputData(table_cpu_map[i], 0, 1);
    			  line_cpu_map[i]->SetColor(255, 0, 0, 255);
    			  line_cpu_map[i]->SetWidth(1.0);
    			  // Show the legend for the first thread since, same thing will be repeated per thread
    			  if (!labeled)
    				  line_cpu_map[i]->SetLabel("CPU");
    			  else
    				  line_cpu_map[i]->SetLabel("");

    			  line_mpi_map[i]->SetInputData(table_mpi_map[i], 0, 1);
    			  line_mpi_map[i]->SetColor(0, 255, 0, 255);
    			  line_mpi_map[i]->SetWidth(1.0);
    			  // Show the legend for the first thread since, same thing will be repeated per thread
    			  if (!labeled) {
    				  line_mpi_map[i]->SetLabel("MPI");
    				  labeled = true;
    			  } else
    				  line_mpi_map[i]->SetLabel("");

    			  if (show_vis_thread_chart_) {
    				  if (show_vis_thread_chart_separate_ || (!show_vis_thread_chart_separate_)) {
    					  if (show_vis_thread_chart_type_ == "area") {
    					  	  line_both_map[i]->SetUseIndexForXSeries(true);
    					  	  line_both_map[i]->SetInputData(table_both_map[i]);
    					  	  line_both_map[i]->SetInputArray(1, "CPU");
    					  	  line_both_map[i]->SetInputArray(2, "MPI");

    					  	  // Due to bug in area plot with more than one Legend, show the last legend only
    					  	  line_both_map[i]->SetLabel("MPI");

    					  	  VTK_CREATE(vtkColorSeries,colorSeries);
    					  	  //colorSeries->SetColorScheme(vtkColorSeries::COOL);
    					  	  colorSeries->SetColorScheme(vtkColorSeries::WILD_FLOWER);
    					  	  line_both_map[i]->SetColorSeries(colorSeries);
    					  } else {
							  if (table_thread_cpu_map[i]->GetNumberOfRows() > 1) {
								  if (show_vis_thread_chart_separate_ || ((!show_vis_thread_chart_separate_) && (i <= 8))) {
									  thread_cpu_plot_map[i]->SetInputData(table_thread_cpu_map[i], 0, 1);
									  thread_cpu_plot_map[i]->SetColor(255, 0, 0, 255);
									  thread_cpu_plot_map[i]->SetWidth(0.5);
									  thread_cpu_plot_map[i]->SetLabel(boost::lexical_cast<std::string>(i) + " CPU");
								  }
							  }

							  if (table_thread_mpi_map[i]->GetNumberOfRows() > 1) {
								  if (show_vis_thread_chart_separate_ || ((!show_vis_thread_chart_separate_) && (i <= 8))) {
									  thread_mpi_plot_map[i]->SetInputData(table_thread_mpi_map[i], 0, 1);
									  thread_mpi_plot_map[i]->SetColor(0, 255, 0, 255);
									  thread_mpi_plot_map[i]->SetWidth(0.5);
									  thread_mpi_plot_map[i]->SetLabel(boost::lexical_cast<std::string>(i) + " MPI");
								  }
							  }

    					  }
						  chart_thread[i]->GetAxis(vtkAxis::LEFT)->SetRange(0, std::max(cpu_max[i], mpi_max[i])+1);
						  chart_thread[i]->GetAxis(vtkAxis::BOTTOM)->SetRange(0, epoch_max[i]+1);

						  if (show_vis_thread_chart_separate_) { // Show one thread per window
							  view_thread[i]->GetRenderWindow()->SetMultiSamples(0);
							  chart_thread[i]->Update();
							  view_thread[i]->GetInteractor()->Initialize();

							  view_thread[i]->GetInteractor()->Render();
							  view_thread[i]->GetInteractor()->Enable();
						  }
						  else { // Show only charts applicable for 9 threads per chart case
							  view_thread[i/9]->GetRenderWindow()->SetMultiSamples(0);
							  chart_thread[i]->Update();
							  view_thread[i/9]->GetInteractor()->Initialize();

							  view_thread[i/9]->GetInteractor()->Render();
							  view_thread[i/9]->GetInteractor()->Enable();

						  }
					  }
    			  }
    		}
    	}

      chart->GetAxis(vtkAxis::LEFT)->SetRange(0, max_tid_+2);
	  chart->GetAxis(vtkAxis::BOTTOM)->SetRange(0, maximum_epoch+1);
   	  view->GetRenderWindow()->SetMultiSamples(0);
   	  chart->Update();
   	  view->GetInteractor()->Initialize();

   	  view->GetInteractor()->Render();
   	  view->GetInteractor()->Enable();
#endif
    }

    static const char* gnuplot_histogram_header =
        "set terminal postscript enhanced\n"
        "set border 3 front linetype -1 linewidth 1.000\n"
        "set boxwidth 0.75 absolute\n"
        "set style fill solid 1.00 border lt -1\n"
        "set grid nopolar\n"
        "set grid noxtics nomxtics ytics nomytics noztics nomztics nox2tics nomx2tics noy2tics nomy2tics nocbtics nomcbtics\n"
        "set grid layerdefault   linetype 0 linewidth 1.000,  linetype 0 linewidth 1.000\n"
        "set key outside right top vertical Left reverse noenhanced autotitles columnhead nobox\n"
        "set key invert samplen 4 spacing 1.5 width 0 height 0 font \"Arial,28\" \n"
        "set size 0.95,0.95\n"
        "set style histogram rowstacked title  offset character 0, 0, 0\n"
        "set style data histograms\n"
        "set style fill solid noborder\n"
        "set xtics border in scale 0,0 nomirror rotate by -45  offset character 0, 0, 0 autojustify\n"
        "set xtics norangelimit font \",8\"\n"
        "set xtics ()\n"
        "set xtics font \"Arial,24\" \n"
        "set noytics\n"
        "set title \"Application Activity Over Time\" font \"Arial,28\" \n"
        "set ylabel \"% of total\" font \"Arial,28\" \n"
        "set xlabel \"Time(ms)\" font \"Arial,28\" offset 0,-1\n"
        "set yrange [0.00000 : 100.000] noreverse nowriteback\n"
        "set boxwidth 1\n";

    void
    AppFTQCalendar::dumpi_gnuplot_histogram(int num_categories)
    {
        int last_data_column = num_categories + 1;
        int totals_column = num_categories + 2;
        std::string fname = sstprintf("plot_app%d.p", aid_);
        std::ofstream out(fname.c_str());
        out << gnuplot_histogram_header;
        out << sstprintf("set xrange[0.0:%d.0]\n", max_epoch_);
        out << "plot 'ftq_app1.dat' using (100.*$2/$" << totals_column << ") title column(2) lc 2 lw -1 lt -1";
        if (num_categories == 1)
            return;

        out << ", \\";
        out << "\n";

        out << sstprintf("   for [i=3:%d] '' using (100.*column(i)/column(%d)) title column(i) lc i lw -1 lt -1",
                         last_data_column, totals_column);
        out.close();
    }

    void
    AppFTQCalendar::dump()
    {
        int num_categories = key::num_categories();
        std::string fname = sstprintf("ftq_app%d.dat", aid_);
        std::ofstream out(fname.c_str());
        //print the first line header
        out << sstprintf("%12s", "Epoch(us)");

        int nonzero_categories[num_categories];
        int num_nonzero_cats = 0;
        for (int i=0; i < num_categories; ++i){
            if (totals_[i] > 0){
                out << sstprintf(" %12s", key::name(i).c_str());
                nonzero_categories[num_nonzero_cats] = i;
                ++num_nonzero_cats;
            }
        }
        out << sstprintf(" %12s", "Total");

        out << "\n";
        timestamp one_ms(1e-3);
        int64_t ticks_ms = one_ms.ticks_int64();
        for (int ep=0; ep < max_epoch_; ++ep){
            //figure out how many us
            double num_ms = double(ep * num_ticks_epoch_) / (double) ticks_ms;
            out << sstprintf("%12.4f", num_ms);
            long total_ticks = 0;
            for (int i=0; i < num_nonzero_cats; ++i){
                int cat_id = nonzero_categories[i];
                long num_ticks = epochs_[ep].get_event_time(cat_id);
                out << sstprintf(" %12ld", num_ticks);
                total_ticks += num_ticks;
            }
            out << sstprintf(" %12ld", total_ticks);
            out << "\n";
        }
        out.close();

        dumpi_gnuplot_histogram(num_nonzero_cats);

#if 0
        fname = sstprintf("nodes_app%d.out", aid_);
        std::ofstream node_out(fname.c_str());
        for (int i=0; i <= max_tid_; ++i){
            TaskFTQCalendar::ptr cal = calendars_[i];
            if (cal){
                cal->dump(node_out);
            }
            else
                node_out << "\n";
        }
        node_out.close();
#endif


         timestamp stamp_sec = timestamp(1);
         int64_t ticks_s = stamp_sec.ticks_int64();
         *debug::debug_stream_ << sstprintf("Average time stats for application %s: \n", appname_.c_str());
         /** print the stat totals */
         int ntasks = max_tid_ + 1;
         for (int i=0; i < num_categories; ++i){
             double av_per_app = (double) totals_[i] / ntasks;
             double t_sec = av_per_app / (double) ticks_s;
             *debug::debug_stream_ << sstprintf("%12s: %16.8f s\n", key::name(i).c_str(), t_sec);
         }
#ifdef HAVE_VTK
         if (show_vis_) {
        	 display_charts(max_tid_, max_epoch_);
        	 std::cout << "--- Please close the chart window to exit \n";
			 view->GetInteractor()->Render();
			 view->GetInteractor()->Enable();
			 view->GetInteractor()->Start();
         }
#endif
     }

    TaskFTQCalendar::TaskFTQCalendar()
        : head_(0), tail_(0)
    {
    }

    void
    TaskFTQCalendar::dump(std::ofstream &os)
    {
        //this isn'y really good for much
        #if 0
        int num_categories = key::num_categories();
        char ev_chars[num_categories];
        ev_chars[key::mpi] = 'M';
        ev_chars[key::compute] = 'C';
        ev_chars[key::sleep] = 'S';
        ev_chars[key::general] = 'G';
        ev_chars[key::buffer] = 'B';

        long num_ticks_per_char = 1000 * 1000 * 100;
        long num_chars = max_tick_ / num_ticks_per_char;

        char* newline = new char[num_chars+1];
        newline[num_chars] = '\0';
        ::memset(newline, ' ' , num_chars);
        EventNode* node = head_;
        while (node){
            long idx_begin = node->ticks_begin / num_ticks_per_char;
            long idx_end = node->ticks_end / num_ticks_per_char;
            char next_char = ev_chars[node->ev_type];
            for (long idx=idx_begin; idx < idx_end; ++idx){
                newline[idx] = next_char;
            }
            node = node->next;
        }
        os << newline;
        os << "\n";
        #endif
    }

    void
    TaskFTQCalendar::collect(int event_typeid, long ticks_begin, long ticks)
    {
        if (tail_){
            if (tail_->event_typeid == event_typeid){
                //combine the two events
                tail_->ticks_end = ticks_begin + ticks;
            }
            else {
                EventNode* node = new EventNode;
                node->ticks_begin = ticks_begin;
                node->ticks_end = ticks_begin + ticks;
                node->next = 0;
                node->event_typeid = event_typeid;
                tail_->next = node;
                tail_ = node;
            }
        }
        else {
            EventNode* node = new EventNode;
            node->ticks_begin = ticks_begin;
            node->ticks_end = ticks_begin + ticks;
            node->next = 0;
            node->event_typeid = event_typeid;
            tail_ = head_ = node;
        }
        max_tick_ = ticks_begin + ticks;
    }

    FTQEpoch::FTQEpoch()
        : totals_(0)
    {
    }

    void
    FTQEpoch::init(int num_events, long long *buffer)
    {
        totals_ = buffer;
        for (int i=0; i < num_events; ++i){
            totals_[i] = 0;
        }
    }


    }
}
