/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_GLOBAL_PTR_ARRAY_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_GLOBAL_PTR_ARRAY_H_INCLUDED

#include <sstmac/software/process/global.h>

namespace sstmac
{
  namespace sw
  {

    template<typename T, int N>
      class sstmac_global_builtin_arr<T*, N> : public sstmac_global
      {

      protected:
        typedef typename boost::unordered_map<long, T*> val_map;
        typedef T* Tptr;
        val_map vals_[N];
        T* init_;

        bool
        check_init(process_context::ptr ptxt, int n = 0)
        {
          if (n >= N)
          {
            sst_throw(
                valueerror,
                "sstmac_global*::check_init: trying to access index %d outside of array size %d",
                n, N);
          }
          if (ptxt)
          {
            if (vals_[n].find(ptxt->get_key()) == vals_[n].end())
              vals_[n][ptxt->get_key()] = init_;
            return true;
          }
          else
            return false;
        }

      public:
        typedef typename val_map::const_iterator const_iterator;

        explicit
        sstmac_global_builtin_arr() :
          init_(NULL)
        {
        }

        explicit
        sstmac_global_builtin_arr(T* init) :
          init_(init)
        {
        }

        virtual
        ~sstmac_global_builtin_arr()
        {
        }

        virtual T*&
        get_val(int n = 0) const
        {
          if (n >= N)
          {
            sst_throw(
                valueerror,
                "sstmac_global*::get_val: trying to access index %d outside of array size %d",
                n, N);
          }
          process_context::ptr ptxt = get_current_context();
          val_map& vals = const_cast<val_map&> (vals_[n]);
          if (ptxt)
          {
            typename val_map::iterator it = vals.find(ptxt->get_key());
            if (it == vals.end())
            {
              T*& ret = vals[ptxt->get_key()] = init_;
              return ret;
            }
            else
            {
              return it->second;
            }
          }
          return const_cast<T*&> (init_);
        }

        virtual std::string
        to_string() const
        {
          std::stringstream ss;
          ss << get_val();
          return ss.str();
        }

        template<typename U>
          T*&
          operator=(const sstmac_global_builtin<U>& b)
          {
            T*& myval = get_val();
            myval = (T*) b;
            return myval;
          }

        template<typename U>
          T*&
          operator=(U* b)
          {
            T*& myval = get_val();
            myval = (T*) b;
            return myval;
          }

        T*
        operator++()
        {
          Tptr& p = get_val();
          return ++p;
        }

        T*
        operator--()
        {
          Tptr& p = get_val();
          return --p;
        }

        T*
        operator++(int dummy)
        {
          Tptr& p = get_val();
          return p++;
        }

        T*
        operator--(int dummy)
        {
          Tptr& p = get_val();
          return p--;
        }

        template<typename U>
          T*
          operator+(const U& b)
          {
            Tptr p = get_val();
            return p + b;
          }

        template<typename U>
          T*
          operator-(const U& b)
          {
            Tptr p = get_val();
            return p - b;
          }

        operator char*()
        {
          return (char*) get_val();
        }

        operator int*()
        {
          return (int*) get_val();
        }

        operator long*()
        {
          return (long*) get_val();
        }

        operator void*()
        {
          return (void*) get_val();
        }

        operator const char*() const
        {
          return (const char*) get_val();
        }

        operator const int*() const
        {
          return (const int*) get_val();
        }

        operator const long*() const
        {
          return (const long*) get_val();
        }

        operator const void*() const
        {
          return (const void*) get_val();
        }

        T*&
        operator*()
        {
          return *(get_val());
        }

        template<typename U>
          bool
          operator ==(const U& b) const
          {
            T* myval = get_val();
            T* otherval = (T*) b;
            return myval == otherval;
          }

        T*
        operator->()
        {
          return get_val();
        }

        T*&
        operator[](int idx)
        {
          return get_val(idx);
        }

        T*&
        operator[](long idx)
        {
          return get_val(idx);
        }

      };

  }
}
#endif
