#include <sstream>

#include <sstmac/software/process/graphviz.h>
#include <sstmac/common/basicstringtokenizer.h>
#include <sstmac/common/config.h>
#include <sstmac/software/process/operatingsystem.h>

#if HAVE_EXECINFO_H
#include <execinfo.h>
#endif

#if HAVE_DLFCN_H
#include <dlfcn.h>
#endif

namespace sstmac {
    namespace sw {

#define BACKTRACE_NFXN 50
typedef void* graphviz_trace[BACKTRACE_NFXN];
static std::list<void**> trace_pool_;


char debug[2048];
char cmd[512];

GraphVizIncrementStack::GraphVizIncrementStack(const char *fxn)
{
    thread::ptr thr = OS_API::get_current_thread();
    if (thr){
        thr->append_backtrace(const_cast<char*>(fxn));
    }
}

GraphVizIncrementStack::~GraphVizIncrementStack()
{
    thread::ptr thr = OS_API::get_current_thread();
    if (thr){
        thr->pop_backtrace();
    }
}

void
GraphViz::Trace::add_call(void* fxn, int ncalls, long count)
{
    graphviz_call& call = calls_[fxn];
    call.first += ncalls;
    call.second += count;
}

void
GraphViz::Trace::add_self(long count)
{
    self_ += count;
}

std::string
GraphViz::Trace::summary() const
{
    std::stringstream sstr;
    sstr << "fn=" << (const char*) fxn_ << "\n";
    sstr << 0 << " " << self_ << "\n";
    std::map<void*, graphviz_call>::const_iterator it = calls_.begin(), end = calls_.end();
    for ( ; it != end; ++it)
    {
        const char* fxn = (const char*) it->first;
        const graphviz_call& call = it->second;
        sstr << "cfn=" << fxn << "\n";
        sstr << "calls=" << call.first << " " << 0 << "\n";
        sstr << 0 << " " << call.second << "\n";
    }
    return sstr.str();
}

void*
GraphViz::Trace::fxn() const
{
    return fxn_;
}


GraphViz::Trace::ptr
GraphViz::Trace::construct(const GraphViz::ptr& parent, void* fxn)
{
    return new Trace(parent, fxn);
}

GraphViz::Trace::Trace(const GraphViz::ptr& parent, void* fxn)
    : self_(0), fxn_(fxn), parent_(parent)
{

}

GraphViz::Trace::ptr
GraphViz::get_trace(void* fxn)
{
    return traces_[fxn];
}

GraphViz::ptr
GraphViz::construct()
{
    return new GraphViz;
}

void
GraphViz::simulation_finished(timestamp t)
{
    check_open();
    myfile << "events: Instructions\n\n";

    std::map<void*,Trace::ptr>::const_iterator it = traces_.begin(), end = traces_.end();
    for (; it != end; ++it)
    {
        myfile << it->second->summary();
        myfile << "\n";
    }
    myfile.close();
}

GraphViz::GraphViz()
    : stat_logger("<stats> callgraph | graphviz", "callgrind.out")
{
}

void
GraphViz::add_call(
    int ncalls,
    long count,
    void* fxn,
    void* callfxn
)
{
    Trace::ptr tr = traces_[fxn];
    if (!tr)
    {
        tr = Trace::construct(GraphViz::ptr(this), fxn);
        traces_[fxn] = tr;
    }
    tr->add_call(callfxn, ncalls, count);
}

void**
GraphViz::allocate_trace()
{
    void** stack = 0;
    if (trace_pool_.empty()){
        stack = new graphviz_trace;
    }
    else{
        stack = trace_pool_.front();
        trace_pool_.pop_front();
    }
    return stack;
}

void
GraphViz::count_trace(long count, const thread::ptr& thr)
{
#if !HAVE_GRAPHVIZ
    return; //nothing to do here
#endif

    if (!is_active())
        return;


    /** see how much of the backtrace is new.  we don't
        want to double count the number of calls */

    void** stack = thr->get_backtrace();

    //this stack is actually backwards - oldest call is index 0
    //most recent call is at the end of the list

    int last_collect_nfxn = thr->get_last_backtrace_nfxn();
    int nfxn_total = thr->get_backtrace_nfxn();
    int stack_end = nfxn_total - 1;
    int recollect_stop = std::min(stack_end,last_collect_nfxn);
    for (int i=0; i < recollect_stop; ++i){
        void* fxn = stack[i];
        void* callfxn = stack[i+1];
        add_call(0, count, fxn, callfxn);
    }

    for (int i=recollect_stop; i < stack_end; ++i)
    {
        void* fxn = stack[i];
        void* callfxn = stack[i+1];
        add_call(1, count, fxn, callfxn);
    }

    void* fxn = stack[stack_end];
    Trace::ptr tr = traces_[fxn];
    if (!tr)
    {
        tr = Trace::construct(GraphViz::ptr(this), fxn);
        traces_[fxn] = tr;
    }
    tr->add_self(count);

    void** dummy_param = 0;
    thr->collect_backtrace(nfxn_total);
}

    }
}


