/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#ifndef SSTMAC_SOFTWARE_PROCESS_KEY_H_INCLUDED
#define SSTMAC_SOFTWARE_PROCESS_KEY_H_INCLUDED

#include <sstmac/software/threading/threading_interface.h>
#include <sstmac/software/process/thread.h>
#include <sstmac/common/ptr_type.h>
#include <cstring>

namespace sstmac
{
  namespace sw
  {

    class library;
    /**
     * A base type and default (empty) implementation of a handle
     * to block and unblock processes.
     */
    class key :
        public ptr_type
    {

    /**
        Global identifier for all keys of a given type.
        This should be a STATIC object that gets passed
        into every key when selecting type.
    */
    public:


    class category
    {
        private:
            friend class library;

            int id_;

            //only callable by library
            category();

        public:
            category(const std::string& name);


            std::string
            name() const {
                return key::name(id_);
            }

            int id() const {
                return id_;
            }
    };

    private:
        static boost::unordered_map<std::string, int>* category_name_to_id_;
        static boost::unordered_map<int, std::string>* category_id_to_name_;

    public:
        static category general;

        static int
        allocate_category_id(const std::string& name);

        static int
        event_typeid(const std::string& event_name);

        static int
        num_categories(){
            return category_name_to_id_->size();
        }

      // thread_data_t defined in process/thread.h
      typedef std::set<thread_data_t> blocking_t;
    private:
      friend class operatingsystem;

      blocking_t blocking_;

      int keyname_id_;

      key();

      key(const category& name);

    public:
      virtual std::string
      to_string() const
      {
        return "key";
      }

      typedef boost::intrusive_ptr<key> ptr;

      static key::ptr
      construct();

      static key::ptr
      construct(const category& name);

      static std::string
      name(int keyname_id){
           return (*category_id_to_name_)[keyname_id];
      }

      std::string
      get_name() const {
        return name(keyname_id_);
      }

      int event_typeid() const {
        return keyname_id_;
      }

      virtual
      ~key();

      void*
      operator new(size_t size);

      void
      operator delete(void* ptr);

      void
      add_blocker(thread_data_t t);

      blocking_t
      get_blocking()
      {
        return blocking_;
      }

      void
      remove_blocker(thread_data_t t);

      bool
      still_blocked()
      {
        return blocking_.size() > 0;
      }

      void
      clear()
      {
        blocking_.clear();
      }

      static void
      clear_statics();

    };

  }
} // end of namespace sstmac

#endif
