/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <algorithm>
#include <exception>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/common/errors.h>
#include <iterator>

ImplementFactory(sstmac::sw::allocationstrategy)
;

namespace sstmac
{
  namespace sw
  {
    SSTRegister("firstavailable | first", allocationstrategy, firstavailableallocation,
                "Allocate the first set of nodes from the list of available nodes. In most cases, allocating from the available node list will give you a regular, contiguous allocation");
    SSTRegister("random", allocationstrategy, randomallocation,
                "Allocate a random set of nodes from the list of available nodes. This will give a non-contiguous allocation.");

    //////////////////////////////////////////////////////////////////////
    // allocationstrategy

    debug allocationstrategy::dbg_;

    sstmac::hw::topology::ptr allocationstrategy::topology_;

    bool allocationstrategy::node_map_built_ = false;
    allocationstrategy::node_set allocationstrategy::available_;
    allocationstrategy::node_set allocationstrategy::allocated_;

    void
    allocationstrategy::clear_statics()
    {
      topology_ = 0;
    }

    allocationstrategy::allocationstrategy()
    {
      if (!dbg_.is_real())

      {
        dbg_ = debug("<debug> allocation | allocationstrategy");
        logger::register_logger(&dbg_);
      }

    }

    allocationstrategy::~allocationstrategy() throw ()
    {
    }

    void
    allocationstrategy::check_request(int nnode_requested, node_set& allocation)
    {
      if (nnode_requested <= 0 || nnode_requested > (int) available_.size())
      {
        std::cerr << "allocationstrategy:  Request for " << nnode_requested
            << " nodes when " << available_.size() << " are available.\n";
        throw valueerror("allocationstrategy: Not enough nodes available.");
      }
    }

    void
    allocationstrategy::release(node_set &alloc)
    {
      node_set::iterator it, end = alloc.end();
      for (it = alloc.begin(); it != end; it++)
      {
        allocated_.erase(*it);
        available_.insert(*it);
      }
    }

    void
    allocationstrategy::remove_from_available(nodeaddress::ptr addr)
    {
      available_.erase(addr);
    }

    //////////////////////////////////////////////////////////////////////
    // firstavailableallocation

    firstavailableallocation::firstavailableallocation()
    {
    }

    firstavailableallocation::~firstavailableallocation() throw ()
    {
    }

    boost::intrusive_ptr<firstavailableallocation>
    firstavailableallocation::construct()
    {
      return boost::intrusive_ptr<firstavailableallocation>(
          new firstavailableallocation());
    }

    void
    firstavailableallocation::allocate(int nnode_requested,
        node_set &allocation)
    {
      //node_set::const_iterator
      //        it = available_.begin(), end = available_.end();
      //for ( ; it != end; ++it){
      //    std::cout << "next node " << (*it)->to_string() << std::endl;
      //}

      if (nnode_requested > (int) available_.size())
      {
        sst_throw(valueerror, "insufficient number of nodes to complete allocation.  available: %d  requested %d:",
            available_.size(), nnode_requested);
      }
      else
      {

        // allocation.resize(nnode_requested);
        for (int i = 0; i < nnode_requested; i++)
        {
          SSTMAC_DEBUG << "firstavailable_: i=" << i << "\n";
          node_set::const_iterator nextnode = available_.begin();
          SSTMAC_DEBUG << "firstavailable_: nextnode= "
                << (*nextnode)->to_string() << "\n";
          allocation.insert(*nextnode);
          allocated_.insert(*nextnode);
          available_.erase(nextnode);
        }
      }
    }

    //////////////////////////////////////////////////////////////////////
    // randomallocation

    randomallocation::randomallocation(unsigned seed)
    {
      rng_ = RNG::SimpleCombo::construct(seed);
    }

    randomallocation::randomallocation(const RNG::UniformInteger::ptr &rng)
    {
      rng_ = rng;
    }

    randomallocation::~randomallocation() throw ()
    {
    }

    boost::intrusive_ptr<randomallocation>
    randomallocation::construct(unsigned seed)
    {
      return boost::intrusive_ptr<randomallocation>(new randomallocation(seed));
    }

    boost::intrusive_ptr<randomallocation>
    randomallocation::construct(const RNG::UniformInteger::ptr &rng)
    {
      return boost::intrusive_ptr<randomallocation>(new randomallocation(rng));
    }

    void
    randomallocation::allocate(int nnode_requested, node_set &allocation)
    {
      if (nnode_requested > (int) available_.size())
      {
        sst_throw(valueerror, "insufficient number of nodes to complete allocation.  available: %d  requested %d:",
            available_.size(), nnode_requested);
      }
      else
      {
        std::vector<nodeaddress::ptr> availvec(available_.size());
        std::copy(available_.begin(), available_.end(), availvec.begin());
        RNG::UniformInteger_functor rngf(rng_);
        std::random_shuffle(availvec.begin(), availvec.end(), rngf);
        for (int i = 0; i < nnode_requested; i++)
        {
          nodeaddress::ptr node = availvec[i];
          allocation.insert(node);
          allocated_.insert(node);
          available_.erase(node);
        }
      }
      // we only want to randomly pick nodes, not randomize the nodes
      // picked, so sort them
      //std::sort(allocation.begin(), allocation.end());
    }

  }
} //end of namespace sstmac
