#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/allocation/cartallocation.h>
#include <sstmac/common/factories/factory.h>

using sstmac::hw::hdtorus;

namespace sstmac
{
  namespace sw
  {
    SSTRegister("cart | cartesian", allocationstrategy, cartallocation,
                "Allocate a regular, cartesian volume of nodes.  This is meant mostly for torus topologies, but is also meaningful for dragonfly and hypercube.");

    cartallocation::cartallocation()
    {

    }

    cartallocation::ptr
    cartallocation::construct(sim_parameters::ptr params)
    {

      return new cartallocation();
    }

    void
    cartallocation::init(const hw::topology::ptr &top)
    {
      allocationstrategy::init(top);
      regtop_ = safe_cast(top, sstmac::hw::regular_topology);
    }

    void
    cartallocation::init_factory_params(const sim_parameters::ptr &params)
    {
      allocationstrategy::init_factory_params(params);

      /**
        sstkeyword {
            gui=4 4 4;
            docstring=The span of the allocation in each dimension.ENDL
            An allocation [4,4,4] would allocated 64 nodes in a cube of torus.ENDL
            This number of sizes should match the dimensionality of the torus.;
        }
      */
      params->get_vector_param("cart_launch_sizes", sizes_);
      /**
        sstkeyword {
            gui=0 0 0;
            docstring=The offset of the allocation in each dimension.ENDL
            Nonzero values shift the allocation along a given dimension.ENDL
            This number of offsets should match the dimensionality of the torus.;
        }
      */
      params->get_vector_param("cart_launch_offsets", offsets_);
      if (sizes_.size() != offsets_.size())
      {
        sst_throw(valueerror, "cartesian allocator: offsets and sizes have different dimensions");
      }
    }

    void
    cartallocation::set_topology(const sstmac::hw::topology::ptr& top)
    {
      allocationstrategy::set_topology(top);
      regtop_ = safe_cast(top, sstmac::hw::regular_topology);
    }

    void
    cartallocation::allocate_dim(int dim, std::vector<int>& vec,
        node_set& allocation)
    {
      if (dim == sizes_.size())
      {
        long nid = regtop_->get_node_id(vec);
        allocation.insert(nodeaddress::convert_from_uniqueid(nid));
        return;
      }

      int dim_size = sizes_[dim];
      int dim_offset = offsets_[dim];
      for (int i = 0; i < dim_size; ++i)
      {
        vec[dim] = dim_offset + i;
        allocate_dim(dim + 1, vec, allocation);
      }
    }

    void
    cartallocation::allocate(int nnode, node_set &allocation)
    {
      // sst_throw(unimplementederror, "cartesion allocator is implemented but not yet tested");
      if (nnode > (int) available_.size())
      {
        sst_throw(sstmac::valueerror,
            "cartallocation: requested %d nodes, only %ld are available",
            nnode, available_.size());
      }
      std::vector<int> vec(sizes_.size(), 0);
      allocate_dim(0, vec, allocation);
    }

  } // end namespace sw
} // end namespace sstmac
