/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */
#include <sstmac/software/services/launch/allocation/hostname_allocation.h>
#include <sstmac/software/services/launch/allocation/allocationstrategy.h>
#include <sstmac/common/fileio.h>
#include <sstmac/common/config.h>
#include <sstmac/hardware/network/interconnect/interconnect.h>
#include <sstmac/common/driver_util.h>
#include <sstmac/hardware/network/topology/topology.h>

namespace sstmac
{
  namespace sw
  {
  SSTRegister("hostname | dumpi",
              allocationstrategy,
              hostname_allocation,
              "Given a file containing one hostname per line, maps task id i to line i+1 in the file [zero-based task ids, one-based file lines]");

    hostname_allocation::nodemap_t hostname_allocation::hostnamemap_;

    std::map<long, std::string> hostname_allocation::nodenum_to_host_map_;

    hostname_allocation::hostname_allocation(const std::string& mapname) : mapfile_(mapname)
    {

    }

    hostname_allocation::ptr
    hostname_allocation::construct(const std::string& mapname)
    {
      return ptr(new hostname_allocation(mapname));
    }

    hostname_allocation::ptr
    hostname_allocation::construct(sim_parameters::ptr params)
    {
      std::string mapname = params->get_param("launch_dumpi_mapname");
      return construct(mapname);
    }

    void
    hostname_allocation::init(const hw::topology::ptr &top)
    {
        allocationstrategy::init(top);
        regtop_ = safe_cast(top, sstmac::hw::regular_topology);
    }

    void
    hostname_allocation::set_topology(const hw::topology::ptr &top)
    {
        allocationstrategy::set_topology(top);
        regtop_ = safe_cast(top, sstmac::hw::regular_topology);
    }

    void
    hostname_allocation::read_map_file(const std::string &here,
        const std::string &mapfile,
        std::map<std::string, std::vector<int> >& hostmap)
    {
      SSTMAC_DEBUG << "dumpiallocation: reading map file \n";
      hostmap.clear();
      std::ifstream in;
      SSTFileIO::open_file(in, mapfile);

      if (!in.is_open())
        {
          throw valueerror(
              here
                  + "could not find node map file in current folder or configuration include path");
        }
      SSTMAC_DEBUG << "dumpiallocation: found map file \n";
      int nnode = -1;
      in >> nnode;
      if (nnode <= 0)
        throw valueerror(here + "bad nnode in node map file");
      int ncoor = -1;
      in >> ncoor;
      if (ncoor <= 0)
        throw valueerror(here + "bad ncoor in node map file");

      SSTMAC_DEBUG << "dumpiallocation: looks good to go \n";
      for (int i = 0; i < nnode; i++)
        {

          std::string hostname;
          in >> hostname;

          if (hostname.size() == 0)
            throw valueerror(here + "bad hostname in map file");
          std::vector<int> coor(ncoor);
          for (int j = 0; j < ncoor; j++)
            {
              coor[j] = -1;
              in >> coor[j];
              if (coor[j] < 0)
                throw valueerror(here + "bad coordinate in map file");
            }

          if (dbg_.is_active())
            {
              SSTMAC_DEBUG << "inserting " << hostname << " into map at coor";
              for (int j = 0; j < ncoor; j++)
                dbg_ << " " << coor[j];
              dbg_ << "\n";
            }
          hostmap[hostname] = coor;
        }
    }

    void
    hostname_allocation::allocate(int nnode_requested,
        node_set &allocation)
    {
      static std::string here("hostname_allocation::allocate: ");

      std::map<std::string, std::vector<int> > hostmap;
      read_map_file(here, mapfile_, hostmap);

      if (!topology_)
      {
          sst_throw(valueerror, "hostname_allocation: null topology");
      }

      std::map<std::string, std::vector<int> >::iterator it, end = hostmap.end();

      for (it = hostmap.begin(); it != end; it++)
        {
          std::vector<int> coords = it->second;

          // find node index for this vertex
          int64_t nodeindex = regtop_->get_node_id(coords);
          nodeaddress::ptr nid = nodeaddress::convert_from_uniqueid(nodeindex);
          hostnamemap_[it->first] = nid;
          nodenum_to_host_map_[nid->unique_id()] = it->first;
          allocation.insert(nid);
        }

    }

  }
}
