/*
 *  This file is part of SST/macroscale:
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top
 *  SST/macroscale directory.
 */

#include <sstmac/software/services/launch/indexing/hostname_indexer.h>

#include <fstream>
#include <sstream>
#include <sstmac/common/errors.h>
#include <sstmac/software/services/launch/allocation/hostname_allocation.h>
#include <sstmac/common/fileio.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/launch/launcher.h>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("hostname", indexstrategy, hostname_indexer,
                "assigns tasks to nodes based on hostname map of topology and hostname list in file");

    hostname_indexer::hostname_indexer(const std::string& listfile)
        : listfile_(listfile)
    {
    }

    hostname_indexer::ptr
    hostname_indexer::construct(sim_parameters::ptr params)
    {
      return ptr(new hostname_indexer(params->get_param("launch_hostname_list")));
    }

    void
    hostname_indexer::allocate(
        const appid& aid,
        const node_set &nodes,
        int ppn,
        std::vector<nodeaddress::ptr> &result,
        int nproc)
    {

      node_set nodeidset;
      int nrank = nproc;
      result.resize(nrank);

      std::ifstream nodelist;
      SSTFileIO::open_file(nodelist, listfile_);
      if (!nodelist.is_open())
          SSTFileIO::not_found(listfile_);

      std::stringstream sstr;
      environment::ptr env = launcher::get_env(aid);
      for (int i = 0; i < nrank; i++)
        {
          std::string hostname;
          nodelist >> hostname;

          sstr << hostname << "\n";

          hostname_allocation::nodemap_t::const_iterator
              aptr_it = hostname_allocation::hostnamemap_.find(hostname),
              end = hostname_allocation::hostnamemap_.end();

          if (aptr_it == end)
          {
                std::stringstream sstr;
                sstr << hostname << " from file " << listfile_ << " does not exist in node map.";

                hostname_allocation::nodemap_t::const_iterator
                        it = hostname_allocation::hostnamemap_.begin();

                if (it == end)
                {
                    sstr << " No hostnames are registered with hostname_allocation."
                         << " This is perhaps not surprising then."
                         << " Maybe check that launch_allocation is set to hostname.";
                }
                else
                {
                    sstr << std::endl << "Valid hostnames are: ";
                    for ( ; it != end; ++it)
                    {
                        sstr << std::endl << it->first;
                    }
                    sstr << std::endl << std::endl
                         << "Are you sure the hostname file and node map"
                            " are from the same machine?";
                }
                throw sstmac::inputerror(sstr.str());
          }

          nodeaddress::ptr aptr = aptr_it->second;

          SSTMAC_DEBUG << " rank " << i << " is on " << hostname << ", at " << aptr->to_string() << "\n";

          result[i] = aptr;
          env->add_hostname(i, hostname, aptr);

        }
      env->setenv("SSTMAC_NODELIST", sstr.str());

    }
  }
}
