/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <math.h>

#include <sstmac/software/services/launch/instantlaunch.h>
#include <sstmac/software/ami/environment.h>
//#include <sstmac/simulator/nodeinterface.h>
//#include <sstmac/backends/native/components/node/node.h>

//#include <sstmac/backends/native/launch/instantlaunch-completion.h>

#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>

#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/launch/messages/arrive_message.h>

#include <sstmac/software/services/launch/machine_descriptor.h>

#include <sstmac/software/libraries/checkpoint/restartable_app.h>

#include <boost/unordered_map.hpp>

namespace sstmac
{
  namespace sw
  {

    SSTRegister("instant | instantlaunch", launcher, instantlaunch,
            "Launch the application instantly at t=0 on all nodes.  This is not physically realistic, but the simplest to work with.");

#define RESTART_TIME 10

    void
    instantlaunch::clear_statics()
    {
        app_to_task_map_.clear();
        envmap_.clear();
    }

    // Pointer types.

    //
    // Protected constructor.
    //
    instantlaunch::instantlaunch(const sim_parameters::ptr& params) :
        launcher(params),
      block_(key::construct())
    {

    }

    service::ptr
    instantlaunch::clone()
    {
      instantlaunch::ptr ret = new instantlaunch(params_);
      ret->launcher::init(tolaunch_, alloc_, indexing_);
      return ret;
    }

    //
    // Public constructor
    //
    instantlaunch::ptr
    instantlaunch::construct(const sim_parameters::ptr& params)
    {
        instantlaunch::ptr retval = new instantlaunch(params);
        return retval;
    }

    void
    instantlaunch::handle(const sst_message::ptr &msg)
    {

      if (msg->get_type() == sst_message::SEND_ACK)
        {
          return;
        }

      launch_message::ptr lmsg = boost::dynamic_pointer_cast<launch_message>(msg);

      if (!lmsg)
        {
          std::cerr << "msg: " << msg->to_string() << "\n";
          std::cerr << "from: " << msg->fromaddr()->to_string() << "\n";
          std::cerr << "to: " << msg->toaddr()->to_string() << "\n";
          throw valueerror(
              "instantlaunch[" + os_->get_my_addr()->to_string()
                  + "]: wanted a launch_message, but got something else");
        }

      launchinfo::ptr linfo = lmsg->get_info();

      /*if (lmsg->is_restart_)
       {
       fullos_->restarting(true);
       restart_app(linfo);
       launch_message::ptr sleep = launch_message::construct(linfo);
       sleep->is_restart_node_ = true;
       sleep->set_time(timestamp(RESTART_TIME / 2));

       os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);
       return;
       }

       if (lmsg->is_restart_node_)
       {
       fullos_->restarting(false);
       return;
       }

       if (lmsg->is_restart_wait_)
       {

       fullos_->cleanup();

       //restart with checkpoint state
       std::list::iterator<app::ptr>:: it, end = lmsg->newapps_.end();
       for (it = lmsg->newapps_.begin(); it != end; it++)
       {
       SSTMAC_DEBUG << "restarting app " << (*it)->get_sid().to_string()
       << " on node " << os_->get_my_addr()->to_string() << "\n";
       fullos_->add_application(*it, linfo->params_);
       }

       states_[linfo->aid_] = LAUNCHED;
       fullos_->start_apps();

       return;
       }*/

      if (lmsg->get_launch_type() == launch_message::ARRIVE)
        {
          current_ = lmsg;

          if (envmap_.find(linfo->aid_) == envmap_.end())
            {
              environment::ptr env = environment::construct();
              envmap_[linfo->aid_] = env;
              SSTMAC_DEBUG
                    << "instantlaunch: initializing allocation for starting at "
                    << linfo->start_ << "\n";

              //this app has not been started anywhere yet, need to allocate, index, and build environment


              node_to_task_map_t& taskmap = app_to_task_map_[linfo->aid_.id];
              task_to_node_map_t& nodemap = app_to_node_map_[linfo->aid_.id];
              taskmap.clear();
              nodemap.clear();

              SSTMAC_DEBUG << "instantlaunch: requesting allocation for "
                    << linfo->num_tasks_ << " tasks \n";

              long nnode;
              if (linfo->num_tasks_ < 0)
                {
                  nnode = -1;
                }
              else
                {
                  nnode = linfo->extra_nodes_ + ceil(
                      double(linfo->num_tasks_)
                          / (double) linfo->ntask_per_node_);
                }
              allocationstrategy::node_set nodes;
              alloc_->allocate(nnode, nodes);

              SSTMAC_DEBUG << "instantlaunch: allocator returned "
                    << nodes.size() << " nodes \n";

              std::vector<nodeaddress::ptr> indexed;
              indexing_->allocate(linfo->aid_, nodes, linfo->ntask_per_node_, indexed, linfo->num_tasks_);

              std::vector<nodeaddress::ptr>::iterator it, end = indexed.end();
              std::map<int,int> app_to_tasknum;
              int ntasks = 0;
              for (it = indexed.begin(); it != end && ntasks < linfo->num_tasks_; it++)
                {

                  int appnum = linfo->aid_.id;
                  int tasknum = app_to_tasknum[appnum]++;
                  sstmac::sw::taskid tid(tasknum);

                  SSTMAC_DEBUG << "taskid: " << tid << "\n";

                  SSTMAC_DEBUG << "adding task " << tid.to_string()
                        << " to node " << (*it)->to_string()
                        << " for app " << appnum
                        << "\n";

                  env->add_item(tid, *it);

                  taskmap.insert(std::pair<nodeaddress::ptr, taskid>(*it, tid));
                  nodemap[tid.id] = *it;

                  ++ntasks;
                }

              if (ntasks < (int) nodes.size())
                {
                  SSTMAC_DEBUG << "instantlaunch: there were "
                               << (nodes.size() - ntasks)
                               << " extra nodes allocated \n";
                  allocationstrategy::node_set::iterator iter = nodes.begin();
                  for (long j = 0; j < ntasks; j++)
                    {
                      iter++;
                    }

                  for (; iter != nodes.end(); iter++)
                    {
                      env->add_extra(*iter);
                    }
                }

            }

          fullos_->set_environment(envmap_[linfo->aid_], linfo->aid_);

          nodeaddress::ptr myaddr = os_->get_my_addr();
          SSTMAC_DEBUG << "instantlaunch(" << myaddr->to_string()
                << "): considering launching a task...";

          node_to_task_map_t& taskmap = app_to_task_map_[linfo->aid_.id];
          if (taskmap.find(myaddr) != taskmap.end())
            {

              std::multimap<nodeaddress::ptr, taskid>::iterator first =
                  taskmap.lower_bound(myaddr), last = taskmap.upper_bound(
                  myaddr);
              int intranode_rank = 0;
              do
                {
                  softwareid sid(linfo->aid_, first->second);

                  app::ptr theapp = linfo->apptype_->clone(sid);
                  int core_affinity = linfo->core_affinity(intranode_rank);
                  theapp->set_affinity(core_affinity);

                  SSTMAC_DEBUG << "launch " << first->second
                        << " for app " << sid.app_.id
                        << " as intranode rank " << intranode_rank
                        << " with core affinity " << core_affinity
                        << "\n";

                  theapp->init_libs(linfo->apptype_->parameters());
                  restartable_app::ptr rest = boost::dynamic_pointer_cast<
                      restartable_app>(theapp);
                  if (rest)
                    {
                      rest->init_restartable(fullos_, sid, linfo->params_);
                    }
                  fullos_->add_application(theapp, linfo);
                  first++;
                  intranode_rank++;
                }
              while (first != last);
              fullos_->start_apps();
              states_[linfo->aid_] = LAUNCHED;
            }
          else
            {
              SSTMAC_DEBUG << "nah, don't feel like it \n";
              states_[linfo->aid_] = NOT_RUNNING_HERE;
            }
        }else if(lmsg->get_launch_type() == launch_message::COMPLETE){

        }
    }

    void
    instantlaunch::start()
    {
      service::start();

      SSTMAC_DEBUG << "instantlaunch: starting...\n";

      if (!os_)
        {
          throw valueerror("instantlaunch: OS hasn't been registered yet");
        }
      if (!os_->kernel_supported(AMI::AMI_COMP_SLEEP))
        {
          throw unimplementederror(
              "instantlaunch: needs the AMI::AMI_COMP_SLEEP kernel to be able to launch");
        }

      std::list<launchinfo::ptr>::iterator it, end = tolaunch_.end();
      for (it = tolaunch_.begin(); it != end; it++)
        {
          arrive_message::ptr launchmsg = arrive_message::construct(*it);
          SSTMAC_DEBUG << "app " << (*it)->apptype_->to_string()
                << " is going to launch " << (*it)->num_tasks_ << " at "
                << (*it)->start_ << "\n";

          os_->execute_kernel(AMI::AMI_COMP_SLEEP, launchmsg);
        }

      SSTMAC_DEBUG << "instantlaunch: launches scheduled\n";
      return;
    }

    /**
     *  This function is called when the local node has failed.
     */
    void
    instantlaunch::apps_failed()
    {
      /*  SSTMAC_DEBUG << "instantlaunch: apps_failed() on "
       << os_->get_my_addr()->to_string()
       << " - restarting all apps running here \n";
       std::map::iterator<appid, APP_STATE>:: iter, end = states_.end();

       fullos_->restarting(true);

       for (iter = states_.begin(); iter != end; iter++)
       {
       if (iter->second == LAUNCHED)
       { //this has been launched

       std::list<app::ptr> osapps = fullos_->get_app_ptrs(iter->first);

       SSTMAC_DEBUG << "instantlaunch: apps_failed() - considering app "
       << iter->first.id << ", it has " << osapps.size()
       << " instances running here \n";

       if (osapps.size() > 0)
       { //check to see if any are still running here

       //get the info pointer, for ease
       launchinfo::ptr linfo;

       std::list::iterator<launchinfo::ptr>:: it, end =
       tolaunch_.end();
       for (it = tolaunch_.begin(); it != end; it++)
       {
       if ((*it)->aid_ == iter->first)
       {
       linfo = *it;
       break;
       }
       }

       //first, tell all the other nodes (have to get a unique set first)
       typedef boost::unordered_map<nodeaddress::ptr, int> maptype;
       maptype nodes;
       for (size_t i = 0; i < linfo->num_tasks_; i++)
       {
       sstmac::sw::taskid tid(i);
       nodes[nodemap_[tid.id]] = 1;
       }

       //don't send it to myself
       nodes.erase(os_->get_my_addr());

       maptype::iterator nodeiter, nodeend = nodes.end();
       for (nodeiter = nodes.begin(); nodeiter != nodeend; nodeiter++)
       {

       launch_message::ptr restartmsg =
       launch_message::construct(linfo);
       restartmsg->set_toaddr(nodeiter->first);
       restartmsg->is_restart_ = true;
       os_->execute_kernel(AMI::AMI_COMM, AMI::AMI_COMM_SEND,
       restartmsg);
       }

       //now restart my own
       restart_app(linfo);

       }
       }
       }
       launchinfo::ptr linfo = *tolaunch_.begin();
       launch_message::ptr sleep = launch_message::construct(linfo);
       sleep->is_restart_node_ = true;
       sleep->set_time(timestamp(RESTART_TIME / 2));

       os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);
       */
    }

    void
    instantlaunch::restart_app(launchinfo::ptr linfo)
    {
      /* SSTMAC_DEBUG << "instantlaunch[" << os_->get_my_addr()->to_string()
       << "]::restart_app() - restarting app " << linfo->aid_.id << "\n";
       //save the checkpoint states, if possible

       std::list<app::ptr> theapps = fullos_->get_app_ptrs(linfo->aid_);
       std::list<app::ptr> newapps;
       std::list::iterator<app::ptr>:: it, end = theapps.end();
       for (it = theapps.begin(); it != end; it++)
       {
       restartable_app::ptr rest = boost::dynamic_pointer_cast<
       restartable_app>(*it);
       if (rest)
       {
       SSTMAC_DEBUG << "instantlaunch::restart_app() - app "
       << (*it)->get_sid().to_string() << " is checkpointable \n";
       lib_checkpoint::ptr check = rest->get_check_lib();
       check->unregister_all_libs();
       fullos_->unregister_lib(check);
       check->restart();
       app::ptr app2 = linfo->apptype_->clone((*it)->get_sid());
       restartable_app::ptr rest2 = boost::dynamic_pointer_cast<
       restartable_app>(app2);
       rest2->init_restartable(check);
       newapps.push_back(app2);

       }
       else
       {
       SSTMAC_DEBUG << "instantlaunch::restart_app() - app "
       << (*it)->get_sid().to_string()
       << " not restartable with checkpoint \n";
       newapps.push_back(linfo->apptype_->clone((*it)->get_sid()));
       }

       fullos_->kill_app(*it);
       }

       launch_message::ptr sleep = launch_message::construct(linfo);
       sleep->is_restart_wait_ = true;
       sleep->newapps_ = newapps;
       sleep->set_time(timestamp(RESTART_TIME));

       os_->execute_kernel(AMI::AMI_COMPUTE, AMI::AMI_COMP_SLEEP, sleep);
       */
    }

  }
} // end of namespace sstmac
