/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

#include <math.h>
#include <sstmac/software/services/launch/reallaunch.h>
#include <sstmac/software/ami/environment.h>
#include <sstmac/common/errors.h>
#include <sstmac/common/debug.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/services/launch/machine_descriptor.h>
#include <sstmac/software/libraries/checkpoint/restartable_app.h>
#include <sstmac/common/driver_util.h>
#include <boost/unordered_map.hpp>

namespace sstmac
{
  namespace sw
  {
    SSTRegister("real | reallaunch", launcher, reallaunch);
#define RESTART_TIME 10

    // Pointer types.

    //
    // Protected constructor.
    //
    reallaunch::reallaunch(const sim_parameters::ptr& params) :
      launcher(params),
      block_(key::construct())
    {
      clock_ = sleep_message::construct("launcher", timestamp(1e-6));
    }

    service::ptr
    reallaunch::clone()
    {
      reallaunch::ptr ret = construct(params_);
      ret->init(tolaunch_, alloc_, indexing_);
      return ret;
    }

    //
    // Public constructor
    //
    reallaunch::ptr
    reallaunch::construct(sim_parameters::ptr params)
    {
      reallaunch::ptr retval(new reallaunch(params));
      return retval;
    }

    void
    reallaunch::msg_arrive(arrive_message::ptr msg)
    {
      launchinfo::ptr linfo = msg->get_info();

      states_[linfo->aid_] = WAITING;

      waiting_.push(linfo);

      try_launch();
    }

    void
    reallaunch::msg_start(start_message::ptr smsg)
    {
      launchinfo::ptr linfo = smsg->get_info();
      if (intranode_rank_.find(linfo->aid_) == intranode_rank_.end())
        {
          intranode_rank_[linfo->aid_] = 0;
        }

      softwareid sid(linfo->aid_, smsg->task());

      app::ptr theapp = linfo->apptype_->clone(sid);
      int core_affinity = linfo->core_affinity(intranode_rank_[linfo->aid_]);
      intranode_rank_[linfo->aid_]++;
      theapp->set_affinity(core_affinity);

      SSTMAC_DEBUG << "launch app " << linfo->aid_.id << ", task "
            << smsg->task().id << " as intranode rank "
            << intranode_rank_[linfo->aid_] << " with core affinity "
            << core_affinity << "\n";

      theapp->init_libs(linfo->apptype_->parameters());
      fullos_->add_application(theapp, linfo);
      fullos_->set_environment(smsg->env(), linfo->aid_);
      fullos_->start_apps();

      linfos_[linfo->aid_] = linfo;

      states_[linfo->aid_] = LAUNCHED;

      vis::vis_obj::VIS_COLOR col;
      if ((linfo->aid_.id - 1) % 6 == 0)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is RED \n";
          col = vis::vis_obj::RED;
        }
      else if ((linfo->aid_.id - 1) % 6 == 1)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is BLUE \n";
          col = vis::vis_obj::BLUE;
        }
      else if ((linfo->aid_.id - 1) % 6 == 2)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is GREEN \n";
          col = vis::vis_obj::GREEN;
        }
      else if ((linfo->aid_.id - 1) % 6 == 3)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is YELLOW \n";
          col = vis::vis_obj::YELLOW;
        }
      else if ((linfo->aid_.id - 1) % 6 == 4)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is ORANGE \n";
          col = vis::vis_obj::ORANGE;
        }
      else if ((linfo->aid_.id - 1) % 6 == 5)
        {
          SSTMAC_DEBUG << "app " << linfo->aid_.id << " is PURPLE \n";
          col = vis::vis_obj::PURPLE;
        }

      vis_set_color(col, os_->now());
    }

    void
    reallaunch::msg_complete(complete_message::ptr cmsg)
    {
      launchinfo::ptr linfo = cmsg->get_info();

      if (cmsg->is_local()) //it came from me, I should forward this to the master
        {

          states_[linfo->aid_] = FINISHED;
          linfos_.erase(linfo->aid_);
          if (myid_ == master_->unique_id())
            {
              fullos_->cleanup();
            }
          cmsg->set_local();
          cmsg->set_toaddr(master_);
          os_->execute_kernel(AMI::AMI_COMM_SEND, cmsg);
        }
      else
        {

          if (cmsg->success())
            {

              if (complete_.find(linfo->aid_) == complete_.end())
                {
                  complete_[linfo->aid_] = 0;
                }

              complete_[linfo->aid_]++;
            }
          else
            {
              SSTMAC_DEBUG << "launcher: app " << linfo->aid_.id
                    << " has failed, killing the rest \n";

              if (states_[linfo->aid_] == FAILED)
                {
                  SSTMAC_DEBUG << "it's already been killed \n";
                }
              else
                {
                  environment::const_ptr env = envmap_[linfo->aid_];
                  for (uint i = 0; i < env->get_app_count(); i++)
                    {
                      taskid tid(i);
                      nodeaddress::ptr at = env->get_node_address(tid);
                      kill_message::ptr k = kill_message::construct(linfo, at);

                      os_->execute_kernel(AMI::AMI_COMM_SEND, k);
                    }
                }

              states_[linfo->aid_] = FAILED;
              complete_[linfo->aid_] = linfo->num_tasks_;
            }

          if (complete_[linfo->aid_] == (long) envmap_[linfo->aid_]->get_app_count())
            {
              states_[linfo->aid_] = FINISHED;

              //release allocation
              allocationstrategy::node_set nodes;
              for (long i = 0; i < complete_[linfo->aid_]; i++)
                {
                  taskid tid(i);
                  nodes.insert(envmap_[linfo->aid_]->get_node_address(tid));
                }

              alloc_->release(nodes);

              try_launch();
            }
        }

      vis_set_color(vis::vis_obj::CLEAR, os_->now());
    }

    void
    reallaunch::msg_restart(restart_message::ptr msg)
    {

    }

    void
    reallaunch::msg_kill(kill_message::ptr msg)
    {
      launchinfo::ptr linfo = msg->get_info();
      SSTMAC_DEBUG << "launcher[" << myid_ << "]: killing app "
            << linfo->aid_.id << "\n";
      allocationstrategy::node_set nodes;
      if (myid_ == master_->unique_id()) //make sure to send to everyone else
        {

          for (long i = 0; i < envmap_[linfo->aid_]->get_app_count(); i++)
            {
              taskid tid(i);
              nodeaddress::ptr addr = envmap_[linfo->aid_]->get_node_address(
                  tid);
              nodes.insert(envmap_[linfo->aid_]->get_node_address(tid));
              if (!addr->equals(msg->fromaddr()) && !addr->equals(master_))
                {
                  SSTMAC_DEBUG << "sending kill signal to node "
                        << addr->to_string() << "\n";
                  kill_message::ptr kill = kill_message::construct(linfo, addr);
                  os_->execute_kernel(AMI::AMI_COMM_SEND, kill);
                }
            }

          states_[linfo->aid_] = FAILED;
          complete_[linfo->aid_] = linfo->num_tasks_;

        }
      else
        {
          os_->execute_kernel(AMI::AMI_CATASTROPHIC_INTERRUPT);
        }

      std::list<app::ptr> apps = fullos_->get_app_ptrs(linfo->aid_);
      std::list<app::ptr>::iterator it, end = apps.end();
      for (it = apps.begin(); it != end; it++)
        {
          fullos_->kill_app(*it);
        }

      if (myid_ == master_->unique_id()) //make sure to send to everyone else
        {
          //release allocation
          alloc_->release(nodes);

          os_->execute_kernel(AMI::AMI_COMP_SLEEP, clock_);
        }
    }

    void
    reallaunch::handle(const sst_message::ptr &msg)
    {
      if (msg && msg == visupdate_)
        {
          vis_->vis_update(os_->now());
          if (waiting_.size() > 0 || complete_.size() != tolaunch_.size())
            {
              os_->execute_kernel(AMI::AMI_COMP_SLEEP, visupdate_);
            }
          return;
        }

      if (msg == clock_)
        {
          try_launch();
        }
      else if (msg->get_type() == sst_message::SEND_ACK)
        {
          return;
        }
      else if (msg->get_type() == sst_message::LAUNCH)
        {
          launch_message::ptr lmsg =
              boost::dynamic_pointer_cast<launch_message>(msg);

          if (!lmsg)
            {
              std::cerr << "msg: " << msg->to_string() << "\n";
              std::cerr << "from: " << msg->fromaddr()->to_string() << "\n";
              std::cerr << "to: " << msg->toaddr()->to_string() << "\n";
              throw valueerror(
                  "reallaunch[" + os_->get_my_addr()->to_string()
                      + "]: wanted a launch_message, but got something else");
            }

          if (lmsg->get_launch_type() == launch_message::ARRIVE)
            {
              arrive_message::ptr amsg = boost::dynamic_pointer_cast<
                  arrive_message>(lmsg);

              msg_arrive(amsg);

            }
          else if (lmsg->get_launch_type() == launch_message::START)
            {
              start_message::ptr smsg = boost::dynamic_pointer_cast<
                  start_message>(lmsg);

              msg_start(smsg);

            }
          else if (lmsg->get_launch_type() == launch_message::COMPLETE)
            {
              complete_message::ptr cmsg = boost::dynamic_pointer_cast<
                  complete_message>(lmsg);

              msg_complete(cmsg);

            }
          else if (lmsg->get_launch_type() == launch_message::KILL)
            {
              kill_message::ptr kmsg =
                  boost::dynamic_pointer_cast<kill_message>(lmsg);

              msg_kill(kmsg);

            }

          else
            {
              throw ssterror("reallaunch: unknown message type in handle");
            }

        }

    }

    void
    reallaunch::try_launch()
    {
      if (waiting_.size() > 0)
        {
          launchinfo::ptr linfo = waiting_.front();

          /*if (envmap_.find(linfo->aid_) == envmap_.end())
           {
           SSTMAC_DEBUG
           << "reallaunch: initializing allocation for starting at "
           << linfo->start_ << "\n";

           //this app has not been started anywhere yet, need to allocate, index, and build environment
           environment::ptr env = environment::construct();

           intranode_rank_[linfo->aid_] = 0;

           SSTMAC_DEBUG << "reallaunch: requesting allocation for "
           << linfo->num_tasks_ << " tasks \n";

           long nnode;
           if (linfo->num_tasks_ < 0)
           {
           nnode = -1;
           }
           else
           {
           nnode = linfo->extra_nodes_ + ceil(
           double(linfo->num_tasks_)
           / (double) linfo->ntask_per_node_);
           }
           allocationstrategy::node_set nodes;
           alloc_->allocate(nnode, nodes);

           SSTMAC_DEBUG << "reallaunch: allocator returned " << nodes.size()
           << " nodes \n";

           if (nodes.size() > 0)
           {

           std::vector<nodeaddress::ptr> indexed;
           indexing_->allocate(linfo->aid_, nodes, linfo->ntask_per_node_, indexed,
           linfo->num_tasks_);

           long i = 0;

           std::vector::iterator<nodeaddress::ptr>:: it, end =
           indexed.end();
           for (it = indexed.begin(); it != end && i < linfo->num_tasks_; it++)
           {

           sstmac::sw::taskid id(i);

           SSTMAC_DEBUG << "taskid: " << id << "\n";

           SSTMAC_DEBUG << "adding task " << id.to_string()
           << " to node " << (*it)->to_string() << "\n";

           env->add_item(id, *it);

           if ((*it)->equals(os_->get_my_addr()))
           {
           softwareid sid(linfo->aid_, id);

           app::ptr theapp = linfo->apptype_->clone(sid);
           int core_affinity = linfo->core_affinity(
           intranode_rank_[linfo->aid_]);
           intranode_rank_[linfo->aid_]++;
           theapp->set_affinity(core_affinity);

           SSTMAC_DEBUG << "launch " << id.id
           << " as intranode rank "
           << intranode_rank_[linfo->aid_]
           << " with core affinity " << core_affinity
           << "\n";

           theapp->init_libs(linfo->apptype_->parameters());
           fullos_->add_application(theapp, linfo);
           }
           else
           {

           start_message::ptr start = start_message::construct(
           linfo, *it, id, env);
           os_->execute_kernel(AMI::AMI_COMM_SEND, start);

           }

           i++;
           }

           if (i < nodes.size())
           {
           SSTMAC_DEBUG << "reallaunch: there were "
           << (nodes.size() - i)
           << " extra nodes allocated \n";
           std::set::iterator<nodeaddress::ptr, nodeaddress::nodeaddresscomp>::
           iter = nodes.begin();
           for (long j = 0; j < i; j++)
           {
           iter++;
           }

           for (; iter != nodes.end(); iter++)
           {
           env->add_extra(*iter);
           }
           }

           envmap_[linfo->aid_] = env;
           fullos_->set_environment(envmap_[linfo->aid_], linfo->aid_);

           fullos_->start_apps();
           states_[linfo->aid_] = LAUNCHED;

           waiting_.pop();

           }
           else
           {
           SSTMAC_DEBUG << "app failed to launch right now. \n";
           }

           }*/

          if (envmap_.find(linfo->aid_) == envmap_.end())
            {
              environment::ptr env = environment::construct();
              envmap_[linfo->aid_] = env;
              SSTMAC_DEBUG
                    << "reallaunch: initializing allocation for starting at "
                    << linfo->start_ << "\n";

              //this app has not been started anywhere yet, need to allocate, index, and build environment

              intranode_rank_[linfo->aid_] = 0;

              node_to_task_map_t& taskmap = app_to_task_map_[linfo->aid_.id];
              task_to_node_map_t& nodemap = app_to_node_map_[linfo->aid_.id];
              taskmap.clear();
              nodemap.clear();

              SSTMAC_DEBUG << "reallaunch: requesting allocation for "
                    << linfo->num_tasks_ << " tasks \n";

              long nnode;
              if (linfo->num_tasks_ < 0)
                {
                  nnode = -1;
                }
              else
                {
                  nnode = linfo->extra_nodes_ + ceil(
                      double(linfo->num_tasks_)
                          / (double) linfo->ntask_per_node_);
                }
              allocationstrategy::node_set nodes;
              alloc_->allocate(nnode, nodes);

              SSTMAC_DEBUG << "reallaunch: allocator returned " << nodes.size()
                    << " nodes \n";

              std::vector<nodeaddress::ptr> indexed;
              indexing_->allocate(linfo->aid_, nodes, linfo->ntask_per_node_,
                  indexed, linfo->num_tasks_);

              std::vector<nodeaddress::ptr>::iterator it, end = indexed.end();
              std::map<int, int> app_to_tasknum;
              int ntasks = 0;
              for (it = indexed.begin(); it != end && ntasks
                  < linfo->num_tasks_; it++)
                {

                  int appnum = linfo->aid_.id;
                  int tasknum = app_to_tasknum[appnum]++;
                  sstmac::sw::taskid tid(tasknum);

                  SSTMAC_DEBUG << "taskid: " << tid << "\n";

                  SSTMAC_DEBUG << "adding task " << tid.to_string()
                        << " to node " << (*it)->to_string() << " for app "
                        << appnum << "\n";

                  env->add_item(tid, *it);

                  taskmap.insert(std::pair<nodeaddress::ptr, taskid>(*it, tid));
                  nodemap[tid.id] = *it;

                  if ((*it)->equals(os_->get_my_addr()))
                    {
                      softwareid sid(linfo->aid_, tid);

                      app::ptr theapp = linfo->apptype_->clone(sid);
                      int core_affinity = linfo->core_affinity(
                          intranode_rank_[linfo->aid_]);
                      intranode_rank_[linfo->aid_]++;
                      theapp->set_affinity(core_affinity);

                      SSTMAC_DEBUG << "launch " << tid.id
                            << " as intranode rank "
                            << intranode_rank_[linfo->aid_]
                            << " with core affinity " << core_affinity << "\n";

                      theapp->init_libs(linfo->apptype_->parameters());
                      fullos_->add_application(theapp, linfo);
                    }
                  else
                    {

                      start_message::ptr start = start_message::construct(
                          linfo, *it, tid, env);
                      os_->execute_kernel(AMI::AMI_COMM_SEND, start);

                    }

                }

              if (ntasks < (int) nodes.size())
                {
                  SSTMAC_DEBUG << "reallaunch: there were " << (nodes.size()
                        - ntasks) << " extra nodes allocated \n";
                  allocationstrategy::node_set::iterator iter = nodes.begin();
                  for (long j = 0; j < ntasks; j++)
                    {
                      iter++;
                    }

                  for (; iter != nodes.end(); iter++)
                    {
                      env->add_extra(*iter);
                    }
                }

            }

          fullos_->set_environment(envmap_[linfo->aid_], linfo->aid_);

          waiting_.pop();
        }

    }

    void
    reallaunch::start()
    {
      service::start();

      SSTMAC_DEBUG << "reallaunch: starting...\n";

      if (!os_)
        {
          throw valueerror("reallaunch: OS hasn't been registered yet");
        }
      if (!os_->kernel_supported(AMI::AMI_COMP_SLEEP))
        {
          throw unimplementederror(
              "reallaunch: needs the AMI::AMI_COMP_SLEEP kernel to be able to launch");
        }

      myid_ = os_->get_my_addr()->unique_id();

      master_ = nodeaddress::convert_from_uniqueid(
          sstmac_env::params->get_long_param("launcher_nodeid"));

      vismap_[os_->get_my_addr()] = ptr(this);

      if ((int)vismap_.size() == alloc_->get_topology()->num_switches())
        {
          SSTMAC_DEBUG << "starting vis, if applicable \n";

          vis::vis_engine::ptr veng
            = vis::vis_engine_factory::get_extra_param("vis_engine", sstmac_env::params);
          if (veng)
            {
              vis::vis_topology::ptr vistop = boost::dynamic_pointer_cast<
                  vis::vis_topology>(alloc_->get_topology());
              if (vistop)
                {
                  vis_ = vis::vis_display::construct("launcher", veng, vismap_,
                      vistop);
                  vis_->vis_start(
                      sstmac_env::params->has_param("vis_file_interval"));

                }
              else
                {
                  SSTMAC_DEBUG << "topology does not support vis \n";
                }

            }
          else
            {
              vismap_.clear();
            }
        }

      if (myid_ == master_->unique_id())
        {
          alloc_->remove_from_available(master_);

          std::list<launchinfo::ptr>::iterator it, end = tolaunch_.end();
          for (it = tolaunch_.begin(); it != end; it++)
            {
              arrive_message::ptr launchmsg = arrive_message::construct(*it);
              SSTMAC_DEBUG << "app " << (*it)->apptype_->to_string()
                    << " is going to launch " << (*it)->num_tasks_ << " at "
                    << (*it)->start_ << "\n";

              states_[(*it)->aid_] = ARRIVING;

              os_->execute_kernel(AMI::AMI_COMP_SLEEP, launchmsg);
            }

          if (sstmac_env::params->has_param("vis_file_interval"))
            {
              timestamp t = sstmac_env::params->get_time_param(
                  "vis_file_interval");
              visupdate_ = sleep_message::construct("launcher", t);
              os_->execute_kernel(AMI::AMI_COMP_SLEEP, visupdate_);

            }
        }

      SSTMAC_DEBUG << "reallaunch: launches scheduled\n";
      return;
    }

    /**
     *  This function is called when the local node has failed.
     */
    void
    reallaunch::apps_failed()
    {
      SSTMAC_DEBUG << "reallaunch[" << myid_
            << "]: apps have failed here, must kill all humans. I mean peer tasks. \n";

      std::map<appid, APP_STATE>::iterator it, end = states_.end();

      for (it = states_.begin(); it != end; it++)
        {
          if (it->second == LAUNCHED)
            {
              SSTMAC_DEBUG << "app " << it->first.id
                    << " is still running, killing \n";

              std::list<app::ptr> appshere = fullos_->get_app_ptrs(it->first);

              if (appshere.size() > 0) //tell master to kill every one else
                {
                  kill_message::ptr kill = kill_message::construct(
                      linfos_[(*appshere.begin())->get_sid().app_], master_);

                  os_->execute_kernel(AMI::AMI_COMM_SEND, kill);

                }

              states_[it->first] = FAILED;
            }
        }

      fullos_->apps_kill();
    }

  }
} // end of namespace sstmac
