#include <sstmac/common/test/test.h>
#include <sstmac/software/process/app.h>
#include <sstmac/software/process/operatingsystem.h>
#include <sstmac/software/api/simpmsg/simpmsg.h>
#include <sstmac/software/process/thread.h>


using namespace sstmac;
using namespace sstmac::sw;
using namespace sstmac::hw;

namespace simpmsg {

sstmac_register_app(test_simpmsg);

void sleep(int seconds)
{
    sstmac::sw::thread::ptr thr = sstmac::sw::operatingsystem::get_current_thread();
    sstmac::sw::app::ptr app = sstmac::safe_cast(thr, sstmac::sw::app);
    app->sleep(timestamp(seconds));
}

class test_message :
    public simple_message
{
    public:
        typedef enum { compute, request, terminate, data } action_t;

    private:
        int value_;

        action_t action_;

        static const long ctrl_msg_num_bytes = 64;

    public:
        typedef boost::intrusive_ptr<test_message> ptr;

        #define enumcase(x) case x: return #x;
        static const char*
        tostr(action_t action)
        {
            switch(action)
            {
                enumcase(compute);
                enumcase(request);
                enumcase(terminate);
                enumcase(data);
            }
        }

        test_message(int value, action_t action)
            : value_(value), action_(action),
            simple_message(ctrl_msg_num_bytes)

        {
        }

        sst_message::ptr
        clone(MESSAGE_TYPES ty) const
        {
            test_message::ptr cln = new test_message(value_, action_);
            simple_message::clone_into(ty, cln);
            simple_message::clone_into(cln);
            return cln;
        }

        test_message(action_t action)
            : action_(action), value_(0),
            simple_message(ctrl_msg_num_bytes)
        {
        }

        test_message(long num_bytes)
            : value_(0), action_(data),
            simple_message(num_bytes)
        {
        }

        action_t
        action() const {
            return action_;
        }

        int
        value() const {
            return value_;
        }
};

class messenger_thread :
    public simpmsg_thread
{
    private:
        simpmsg_queue::ptr work_queue_;

    public:
        messenger_thread(const simpmsg_queue::ptr& queue)
            : work_queue_(queue)
        {
        }

        virtual void run();
};

void
messenger_thread::run()
{
    int me = simp_rank();
    while(1)
    {
        simple_message::ptr msg = simp_poll();
        test_message::ptr tmsg = safe_cast(msg, test_message);
        //std::cout << sstprintf("receiver got message of type %s on %d\n",
        //                       test_message::tostr(tmsg->action()), me);
        work_queue_->put_message(msg);
        if (tmsg->action() == test_message::terminate)
            break;
    }
}

static int max_num_tasks = 5;
static int max_num_sends = 5;

bool
terminate(int num_tasks, int num_sends)
{
    if (num_tasks >= max_num_tasks &&
        num_sends >= max_num_sends)
        return true;
    return false;
}

void
run_work_loop(int me, int nproc, const simpmsg_queue::ptr& work_queue)
{
    // send out some data requests and some tasks
    for (int t=1; t <= max_num_tasks; ++t){
        int dst = (me + t) % nproc;
        simp_send(dst, new test_message(1e4, test_message::compute));
        simp_send(dst, new test_message(1e6, test_message::request));
    }

    int num_tasks = 0;
    int num_sends = 0;
    while (1)
    {
        simple_message::ptr msg = work_queue->poll_until_message();
        test_message::ptr tmsg = safe_cast(msg, test_message);
        //std::cout << sstprintf("got work message of type %s on %d\n",
        //                       test_message::tostr(tmsg->action()), me);
        switch (tmsg->action())
        {
            case test_message::compute:
                //compute for x microseconds
                compute(tmsg->value()*1e-6);
                ++num_tasks;
                if (terminate(num_tasks, num_sends))
                    return;
                break;
            case test_message::request:
                //send back x bytes of data
                simp_rdma_put(tmsg->src(), new test_message(tmsg->value()));
                break;
            case test_message::rdma_put_ack:
                ++num_sends;
                if (terminate(num_tasks, num_sends))
                    return;
                break;
            case test_message::terminate:
                return;

        }
    }
}

int
test_simpmsg_main(int argc, char** argv)
{
    simp_init();
    int me = simp_rank();
    int nproc = simp_nproc();
    std::cout << sstprintf("Rank %d starting\n", me);

    //spin off a messenger thread
    simpmsg_queue::ptr work_queue = simpmsg_queue::construct();
    messenger_thread::ptr thr = new messenger_thread(work_queue);
    thr->start();

    run_work_loop(me, nproc, work_queue);

    thr->join();

    //if I am 0, send out terminate messages to everybody
    if (me == 0)
    {
        for (int i=1; i < nproc; ++i){
            simp_send(i, new test_message(test_message::terminate));
        }
    }

    std::cout << sstprintf("Rank %d terminating\n", me);
    simp_finalize();
    return 0;
}

}
