/*
 *  This file is part of SST/macroscale: 
 *               The macroscale architecture simulator from the SST suite.
 *  Copyright (c) 2009-2010 Sandia Corporation.
 *  This software is distributed under the BSD License.
 *  Under the terms of Contract DE-AC04-94AL85000 with Sandia Corporation,
 *  the U.S. Government retains certain rights in this software.
 *  For more information, see the LICENSE file in the top 
 *  SST/macroscale directory.
 */

//#include <sstmac/sstmac.h>
#include <unistd.h> // getopt
#include <sstmac_testutil.h>
#include <gtest/gtest.h>
#include <stdlib.h>
#include <stdio.h>
#include <sstream>

#include <sstmac/sstmacro.h>
#include <sstmac/backends/native/eventmap.h>
#include <sstmac/backends/native/nodeid.h>


#if 0
#include <sstmac/keywords/force_link.h>
#include <sstmac/software/force_link.h>
#include <sstmac/hardware/force_link.h>


//#include <sstmac/hardware/network/topology/composite.h>
//#include <sstmac/hardware/network/topology/line.h>
//#include <sstmac/hardware/network/topology/ring.h>
//#include <sstmac/hardware/network/topology/fattree.h>
//#include <sstmac/hardware/network/topology/hdtorus.h>


#include <sstmac/hardware/network/congestion/flower.h>
#include <sstmac/hardware/network/congestion/ordered_flower.h>
#include <sstmac/hardware/network/congestion/throttle_flower.h>
#include <sstmac/hardware/network/congestion/sharedcircuit.h>
#include <sstmac/hardware/network/congestion/circuit.h>
#include <sstmac/hardware/network/congestion/fullnetwork.h>
#include <sstmac/hardware/network/congestion/circuit.h>
#include <sstmac/hardware/network/congestion/sharedbuffer.h>
#endif

using namespace sstmac::hw;
using namespace sstmac;
using namespace sstmac::native;
using namespace sstmac::sw;

std::string
get_token(std::string &tokens, int *narg = 0)
{
  std::string result;

  if (tokens.size() != 0)
    {
      size_t pos = tokens.find_first_of("_(),");

      if (pos == std::string::npos)
        {
          result = tokens;
          tokens.clear();
        }
      else
        {
          result = tokens.substr(0, pos);
          tokens = tokens.substr(pos + 1, std::string::npos);
        }
    }

  if (result.size() > 0 && narg)
    *narg = *narg + 1;

  return result;
}

struct test_results : public ptr_type
{
  timestamp end_;
  int numsent_;
  int numrecv_;
  int nummsgs_;

  test_results() :
    end_(0), numsent_(0), numrecv_(0), nummsgs_(0)
  {
  }

  virtual std::string
  to_string() const
  {
    return "testnetworks::test_results";
  }
};

class message_sent_handler : public eventhandler
{
  eventmanager::ptr em_;
  test_results* res_;

  message_sent_handler(eventmanager::ptr e, test_results* res) :
    em_(e), res_(res)
  {
  }

public:
  virtual void
  handle(const sst_message::ptr &msg)
  {
    // std::cout << em_->now() << ": sent from " << msg->fromaddr_->to_string() << " to " << msg->toaddr_->to_string() << std::endl;
    res_->numsent_++;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::message_sent_handler";
  }

  static boost::intrusive_ptr<message_sent_handler>
  construct(eventmanager::ptr e, test_results* res)
  {
    boost::intrusive_ptr<message_sent_handler> ret(
        new message_sent_handler(e, res));
    return ret;
  }

};

class message_recv_handler : public eventhandler
{
  eventmanager::ptr em_;
  test_results* res_;

  message_recv_handler(eventmanager::ptr e, test_results* res) :
    em_(e), res_(res)
  {
  }

public:
  virtual void
  handle(const sst_message::ptr &msg)
  {

    // std::cout << em_->now() << ": recv from " << msg->fromaddr_->to_string() << " to " << msg->toaddr_->to_string() << std::endl;
    res_->numrecv_++;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::message_recv_handler";
  }

  static boost::intrusive_ptr<message_recv_handler>
  construct(eventmanager::ptr e, test_results* res)
  {
    boost::intrusive_ptr<message_recv_handler> ret(
        new message_recv_handler(e, res));
    return ret;
  }

};

class test_message : 
    public network_message
{

  static int ID_CNT;
  int msg_id_;
  test_message() :
    sst_message(sst_message::SST)
  {
    msg_id_ = ID_CNT++;
  }

public:

  static boost::intrusive_ptr<test_message>
  construct()
  {
    boost::intrusive_ptr<test_message> ret(new test_message());
    return ret;
  }

  int bytes_;

  virtual long
  get_byte_length() const
  {
    return bytes_;
  }

  virtual bool
  get_needs_ack() const
  {
    return true;
  }

  virtual std::string
  to_string() const
  {
    return "test_networks::test_message";
  }

  virtual sst_message::ptr
  clone() const
  {
    boost::intrusive_ptr<test_message> ret = construct();
    ret->bytes_ = bytes_;
    ret->fromaddr_ = fromaddr_;
    ret->toaddr_ = toaddr_;
    return ret;
  }

  virtual sst_message::ptr
  clone(MESSAGE_TYPES t) const
  {
    boost::intrusive_ptr<test_message> ret = construct();
    ret->bytes_ = bytes_;
    ret->msgtype_ = msgtype_;
    ret->fromaddr_ = fromaddr_;
    ret->toaddr_ = toaddr_;
    return ret;
  }
};

int test_message::ID_CNT = 0;

void
test_network(interconnect::ptr net, std::string testfile, test_results* res)
{

  eventmanager::ptr em = eventmap::construct();

  net->set_eventmanager(em);

  boost::intrusive_ptr<message_sent_handler> sent_handler =
      message_sent_handler::construct(em, res);
  boost::intrusive_ptr<message_recv_handler> recv_handler =
      message_recv_handler::construct(em, res);

  std::string line;
  std::ifstream myfile;
  myfile.open(testfile.c_str());
  if (!myfile.is_open())
    {
      std::cout
      << "testnetworks::test_network -- ERROR - cannot open test file: "
      << testfile << std::endl;
      exit(1);
    }

  while (!myfile.eof())
    {
      std::getline(myfile, line);
      std::istringstream iss(line);
      std::string temp;
      iss >> temp;
      timestamp when = timestamp(atof(temp.c_str()));
      iss >> temp;
      nodeid::ptr from = nodeid::construct(atoi(temp.c_str()));
      iss >> temp;
      nodeid::ptr to = nodeid::construct(atoi(temp.c_str()));
      iss >> temp;
      int size = atoi(temp.c_str());

      boost::intrusive_ptr<test_message> the_msg = test_message::construct();
      net->set_endpoint(to, recv_handler);
      net->set_endpoint(from, sent_handler);
      the_msg->fromaddr_ = from;
      the_msg->toaddr_ = to;
      the_msg->bytes_ = size;
      res->nummsgs_++;

      net->send(when, the_msg);

    }

  em->run();
  res->end_ = em->now();
  em->clear(timestamp(0));

}

void
print_help(const char *name)
{
  std::cout << name << " [-h] [-f file] [-p file_prefix] [-d debug_string]"
      << std::endl;
}

std::vector<std::string> test_file_names;
std::string prefix;

int
main(int argc, char **argv)
{
  int opt;
  int nodes = 10;

  ::testing::InitGoogleTest(&argc, argv);

  std::vector<int> top_args;

  std::vector<std::string> files;
  bool helpflag = false;
  while (!helpflag && (opt = getopt(argc, argv, "hf:p:d:")) != -1)
    {
      switch (opt)
      {
      case 'h':
        helpflag = true;
        break;
      case 'f':
        test_file_names.push_back(optarg);
        break;

      case 'p':
        prefix = optarg;
        prefix += "/";
        break;

      case 'd':
        logger::set_user_param(optarg);
        break;

      default:
        std::cerr << "Unknown argument -" << opt << "\n";
        helpflag = true;
      }
    }
  if (helpflag)
    {
      print_help(argv[0]);
    }

  if (test_file_names.size() == 0)
    {
      test_file_names.push_back(prefix + "testnetworks_tests/all2all.txt");
      test_file_names.push_back(prefix + "testnetworks_tests/hotspot.txt");
      test_file_names.push_back(prefix + "testnetworks_tests/neighbor.txt");
      test_file_names.push_back(prefix + "testnetworks_tests/tornado.txt");
      test_file_names.push_back(prefix + "testnetworks_tests/random.txt");
    }

  return RUN_ALL_TESTS();

}

void
print_err(timestamp exp, timestamp act)
{
  int esecs = exp.sec();
  int actsecs = act.sec();
  double edec = (exp.sec() - esecs);
  double adec = (act.sec() - actsecs);

  std::stringstream ess;
  std::stringstream ass;
  for (int i = 0; i < 7; i++)
    {
      if (edec * 10 >= 1)
        {
          int temp = edec * pow(10.0, 7 - i);
          ess << temp;
          break;
        }
      else
        {
          ess << "0";
        }
      edec *= 10;
    }

  for (int i = 0; i < 7; i++)
    {
      if (adec * 10 >= 1)
        {
          int temp = adec * pow(10.0, 7 - i);
          ass << temp;
          break;
        }
      else
        {
          ass << "0";
        }
      adec *= 10;
    }
  std::cout << "ERROR: expected: " << esecs << "." << ess.str() << ", got: "
      << actsecs << "." << ass.str() << std::endl;
}

TEST(testnetworks, fullnetwork)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000011000000));
  expected.push_back(timestamp(1.000101000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000101000000));

  for (int i = 0; i < test_file_names.size(); i++)
    {
      empty_parameters::ptr par = empty_parameters::construct();
      par->add_param("network_latency", "1us");
      par->add_param("network_bandwidth", "1GBps");
      par->add_param("topology_geometry", "16");
      par->add_param("topology_name", "xbar");

      interconnect::ptr net =
          interconnect_factory::get_value("fullnetwork", par);

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "fullnetwork, t=" << res->end_  << ", expecting " << expected[i] << std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-6);
      if (!good)
        {
          print_err(expected[i], res->end_);
        }
      EXPECT_TRUE(good)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numsent_)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_)
      ;
      delete res;
    }

}

TEST(testnetworks, circuit)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.0000110000000));
  expected.push_back(timestamp(1.00080790000));
  expected.push_back(timestamp(1.000002000000));
  expected.push_back(timestamp(1.000003900000));
  expected.push_back(timestamp(1.000101000000));

  for (int i = 0; i < test_file_names.size(); i++)
    {

      empty_parameters::ptr par = empty_parameters::construct();
      par->add_param("topology_geometry", "4,4");
      par->add_param("topology_name", "torus");
      par->add_param("network_bandwidth", "1GBps");
      par->add_param("network_latency", "1us");

      interconnect::ptr net = interconnect_factory::get_value("circuit", par);

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "circuit, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-6);
      if (!good)
        {
          print_err(expected[i], res->end_);
        }
      EXPECT_TRUE(good)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numsent_)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_)
      ;
      delete res;
    }
}

TEST(testnetworks, sharedcircuit_ordered)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.000007600000));
  expected.push_back(timestamp(1.000534300000));
  expected.push_back(timestamp(1.000001600000));
  expected.push_back(timestamp(1.000002300000));
  expected.push_back(timestamp(1.0000676000000));

  for (int i = 0; i < test_file_names.size(); i++)
    {
      empty_parameters::ptr par = empty_parameters::construct();
      par->add_param("topology_geometry", "4,4");
      par->add_param("topology_name", "torus");
      par->add_param("network_bandwidth_link", "1GBps");
      par->add_param("network_bandwidth_node", "1.5GBps");
      par->add_param("network_latency", "1us");
      par->add_param("sharedcircuit_flower", "ordered");

      interconnect::ptr net = interconnect_factory::get_value("sharedcircuit", par);

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "sharedcircuit_ordered, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_, expected[i], 1e-6);
      if (!good)
        {
          print_err(expected[i], res->end_);
        }
      EXPECT_TRUE(good)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numsent_)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_)
      ;
      delete res;
    }
}

TEST(testnetworks, sharedcircuit_throttle)
{
  std::vector<timestamp> expected;
  expected.push_back(timestamp(254.0000076000000));
  expected.push_back(timestamp(1.007858700000));
  expected.push_back(timestamp(1.000001600000));
  expected.push_back(timestamp(1.000002300000));
  expected.push_back(timestamp(1.000067600000));

  for (int i = 0; i < test_file_names.size(); i++)
    {
      empty_parameters::ptr par =
          empty_parameters::construct();
      par->add_param("topology_geometry", "4,4");
      par->add_param("topology_name", "torus");
      par->add_param("network_bandwidth_link", "1GBps");
      par->add_param("network_bandwidth_node", "1.5GBps");
      par->add_param("network_latency", "1us");
      par->add_param("sharedcircuit_flower", "throttle");

      interconnect::ptr net = interconnect_factory::get_value("sharedcircuit", par);

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      //std::cout << "sharedcircuit_throttle, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good =
          timestamp_near(res->end_, expected[i], 1e-6);
      if (!good)
        {
          print_err(expected[i], res->end_);
        }
      EXPECT_TRUE(good)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numsent_)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_)
      ;
      delete res;
    }
}

TEST(testnetworks, sharedbuffer)
{
  std::vector<timestamp> expected;

  expected.push_back(timestamp(254.00001230001));
  expected.push_back(timestamp(1.000102100001));
  expected.push_back(timestamp(1.000002610001));
  expected.push_back(timestamp(1.000002610001));
  expected.push_back(timestamp(1.000102100001));

  for (int i = 0; i < test_file_names.size(); i++)
    {
      empty_parameters::ptr par =
          empty_parameters::construct();
      par->add_param("topology_geometry", "4,4");
      par->add_param("topology_name", "torus");
      par->add_param("network_bandwidth_link", "1GBps");
      par->add_param("network_bandwidth_node", "2GBps");
      par->add_param("network_latency_node", "1us");
      par->add_param("network_latency_link", "10ns");
      par->add_param("network_latency_router", "100ns");
      par->add_param("network_buffer", "81920");
      par->add_param("network_vcs", "2");

      interconnect::ptr net = interconnect_factory::get_value("sharedbuffer", par);

      test_results *res = new test_results();
      test_network(net, test_file_names[i], res);

      // std::cout << "sharedbuffer, t=" << res->end_ << ", expecting " << expected[i]<< std::endl;

      bool good = timestamp_near(res->end_,
          expected[i], 1e-6);
      if (!good)
        {
          print_err(expected[i], res->end_);
        }
      EXPECT_TRUE(good)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numsent_)
      ;
      EXPECT_TRUE(res->nummsgs_ == res->numrecv_)
      ;

      delete res;
    }
}

