//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "common/utilities/assert.h"
#include "common/utilities/dumpihelper.h"
#include "common/stats/entry/entrydataimpl.h"

namespace sstmac { namespace tools { namespace analyzer {

////////////////////////////////////////////////////////////////////////////////
// General implementation
template< typename T >
EntryGenerate<T>::EntryGenerate(int nodeCount)
    : EntryGenerateBase(nodeCount)
{} 

template< typename T >
void EntryGenerate<T>::Start()
{} 

template< typename T >
void EntryGenerate<T>::Stop()
{} 

template< typename T >
void EntryGenerate<T>::OnCallRecord(const CallRecordBase& record)
{}

////////////////////////////////////////////////////////////////////////////////
// Overall_Characteristics
template<>
void EntryGenerate<statsentry::Overall>::OnCallRecord(const CallRecordBase& record)
{
    m_bytes.Accumulate(record.GetBytesSend());

    TimeRange range = record.GetWallTime();
    m_waitTime.Accumulate(range.second - range.first);
    m_computeTime.Accumulate(range.first - range.second);

    if(record.IsInit())
    {
        m_computeTime.Accumulate(-range.first);
    }
    else if(record.IsFinalize())
    {
        m_computeTime.Accumulate(range.second);
    }
}

////////////////////////////////////////////////////////////////////////////////
// ByNode_Characteristics
template<>
void EntryGenerate<statsentry::ByNode>::Start()
{
    m_bytes.resize(m_nodeCount);
    m_waitTime.resize(m_nodeCount);
    m_computeTime.resize(m_nodeCount);
    m_inDegree.resize(m_nodeCount);
    m_outDegree.resize(m_nodeCount);
    m_inNodes.resize(m_nodeCount);
    m_outNodes.resize(m_nodeCount);
}

template<>
void EntryGenerate<statsentry::ByNode>::OnCallRecord(const CallRecordBase& record)
{
    int nodeRank = record.GetNodeRank();

    m_bytes[nodeRank].Accumulate(record.GetBytesSend());

    TimeRange range = record.GetWallTime();
    m_waitTime[nodeRank].Accumulate(range.second - range.first);
    m_computeTime[nodeRank].Accumulate(range.first - range.second);

    //Merge in/out nodes
    vector<int> inNodes;
    vector<int> outNodes;
    record.GetDest(outNodes);
    record.GetSource(inNodes);
    //m_inNodes[nodeRank].insert(inNodes.begin(), inNodes.end());  
    //m_outNodes[nodeRank].insert(outNodes.begin(), outNodes.end());  

    if(record.IsInit())
    {
        m_computeTime[nodeRank].Accumulate(-range.first);
    }
    else if(record.IsFinalize())
    {
        m_computeTime[nodeRank].Accumulate(range.second);
    }
}

template<>
void EntryGenerate<statsentry::ByNode>::Stop()
{
    for (int i = 0; i < m_nodeCount; i++) 
    {
        m_inDegree[i] = m_inNodes[i].size();
        m_outDegree[i] = m_outNodes[i].size();

        m_avgBytes.Accumulate(m_bytes[i]);
        m_avgWaitTime.Accumulate(m_waitTime[i]);
        m_avgComputeTime.Accumulate(m_computeTime[i]);
        m_avgInDegree.Accumulate(m_inDegree[i]);
        m_avgOutDegree.Accumulate(m_outDegree[i]);
        m_sdBytes.Accumulate(m_bytes[i]);
        m_sdWaitTime.Accumulate(m_waitTime[i]);
        m_sdComputeTime.Accumulate(m_computeTime[i]);
        m_sdInDegree.Accumulate(m_inDegree[i]);
        m_sdOutDegree.Accumulate(m_outDegree[i]);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Mpi_Call_Stats
template<>
void EntryGenerate<statsentry::MpiCall>::Start()
{
    m_count.resize(Dumpi::GetCount());
    m_avgBlockTime.resize(Dumpi::GetCount());
    m_sdBlockTime.resize(Dumpi::GetCount());
    m_time.resize(Dumpi::GetCount());

    for (unsigned int i = 0; i < m_time.size(); i++) 
    {
        m_time[i].resize(m_nodeCount);
    }
}

template<>
void EntryGenerate<statsentry::MpiCall>::OnCallRecord(const CallRecordBase& record)
{
    int func = record.GetFunction();
    int nodeRank = record.GetNodeRank();
    ++m_count[func];

    double blockTime = record.GetWallTime().second - record.GetWallTime().first;
    m_avgBlockTime[func].Accumulate(blockTime);
    m_sdBlockTime[func].Accumulate(blockTime);
    m_time[func][nodeRank].Accumulate(blockTime);
}

////////////////////////////////////////////////////////////////////////////////
// Collective_Variation
template<>
void EntryGenerate<statsentry::Collective>::Start()
{
    m_currRegion = -1;
    m_time.reserve(500);
    m_collective.reserve(500);
}

template<>
void EntryGenerate<statsentry::Collective>::OnCallRecord(const CallRecordBase& record)
{
    if (!record.IsCollective())
    {
        return;
    }

    int nodeRank = record.GetNodeRank();
    int func = record.GetFunction();

    if(!record.IsInit())
    {
        if(nodeRank == 0)
        {
            //Insert new collective pair only for root node
            m_collective.back().m_endFunc = func;
        }

        //Complet last region if this is not MPI_Init
        m_time[m_currRegion][nodeRank].Accumulate(record.GetWallTime().first);

        //Accumulate average and sd
        m_avgTime[m_currRegion].Accumulate(m_time[m_currRegion][nodeRank]);
        m_sdTime[m_currRegion].Accumulate(m_time[m_currRegion][nodeRank]);
    }
    else
    {
        m_currRegion = -1;
    }

    if(!record.IsFinalize())
    {
        //Insert a new region is this is not MPI_Finalize
        if(nodeRank == 0)
        {
            m_collective.push_back(statsentry::Collective::CollRegion(m_collective.size() + 1, func, Dumpi::GetCount()));
            m_time.push_back(vector<Sum>());
            m_avgTime.push_back(Average());
            m_sdTime.push_back(Deviation());
            m_time.back().resize(m_nodeCount);
        }
        
        //Start new region timing
        m_time[++m_currRegion][nodeRank].Accumulate(-record.GetWallTime().first);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Transmission_SpyPlot
template<>
void EntryGenerate<statsentry::TransmissionTime>::Start()
{
    m_time.resize(m_nodeCount);

    for (int i = 0; i < m_nodeCount; i++)
    {
        m_time[i].resize(m_nodeCount);
    }
}

template<>
void EntryGenerate<statsentry::TransmissionBytes>::Start()
{
    m_traffic.resize(m_nodeCount);

    for (int i = 0; i < m_nodeCount; i++) 
    {
        m_traffic[i].resize(m_nodeCount);
    }
}

template<>
void EntryGenerate<statsentry::TransmissionTime>::OnCallRecord(const CallRecordBase& record)
{
    int srcRank = record.GetNodeRank();
    std::vector<int> destRank;
    record.GetDest(destRank);
    TimeRange range = record.GetWallTime();

    for (unsigned int i = 0; i < destRank.size(); i++)
    {
        m_time[srcRank][destRank[i]].Accumulate(range.second - range.first);

    }
}

template<>
void EntryGenerate<statsentry::TransmissionBytes>::OnCallRecord(const CallRecordBase& record)
{
    int srcRank = record.GetNodeRank();
    std::vector<int> destRank;
    record.GetDest(destRank);
    TimeRange range = record.GetWallTime();

    for (unsigned int i = 0; i < destRank.size(); i++) 
    {
        m_traffic[srcRank][destRank[i]].Accumulate(record.GetBytesSend(destRank[i]));
    }
}

} } } //namespace
