//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "common/comparison/criterion/criteriondataimpl.h"
#include "common/utilities/dumpihelper.h"
#include "common/factory/factory.h"
#include "common/log/logger.h"
#include "common/utilities/serializeadapter.h"
#include "common/numeric/numericdefine.h"

FACTORY_REGISTER_CLASS(sstmac::tools, CriterionDataBase, sstmac::tools::criterion, Overall);
FACTORY_REGISTER_CLASS(sstmac::tools, CriterionDataBase, sstmac::tools::criterion, ByNode);
FACTORY_REGISTER_CLASS(sstmac::tools, CriterionDataBase, sstmac::tools::criterion, MpiCall);
FACTORY_REGISTER_CLASS(sstmac::tools, CriterionDataBase, sstmac::tools::criterion, Collective);
FACTORY_REGISTER_CLASS(sstmac::tools, CriterionDataBase, sstmac::tools::criterion, TransmissionBytes);

namespace sstmac { namespace tools { namespace criterion {

void Overall::Serialize(ostream& os) const
{
    os << "Transmit Bytes, Wait Time, Compute Time\n"
        << m_bytesDiff << ", "
        << m_waitTimeDiff << ", "
        << m_computeTimeDiff;
}

void Overall::Deserialize(istream& is)
{
    is >> Ignore('\n') >> m_bytesDiff
        >> Ignore(' ') >> m_waitTimeDiff
        >> Ignore(' ') >> m_computeTimeDiff;
}

void ByNode::Serialize(ostream& os) const
{
    if(IsEqual(m_logicSimilarity, 1.0f))
    {
        int nodeCount = m_bytesDiff.size();

        os << nodeCount << " Nodes, Transmit Bytes, Wait Time, Compute Time";

        for (int i = 0; i < nodeCount; i++) 
        {
            os << "\nRank_" << i 
                << ", " << m_bytesDiff[i]
                << ", " << m_waitTimeDiff[i]
                << ", " << m_computeTimeDiff[i];
        }
    }
    else
    {
        os << "Number of nodes is different: "
            << m_nodeCount.first << " : "
            << m_nodeCount.second;
    }
}

void ByNode::Deserialize(istream& is)
{
    if(IsEqual(m_logicSimilarity, 1.0f))
    {
        int nodeCount = 0;
        is >> nodeCount >> Ignore('\n');

        m_bytesDiff.resize(nodeCount);
        m_waitTimeDiff.resize(nodeCount);
        m_computeTimeDiff.resize(nodeCount);

        for (int i = 0; i < nodeCount; i++) 
        {
            ASSERT(is.good(), "Expect to have %d lines of %s\n", nodeCount, GetClassId());

            is >> Ignore(' ') >> m_bytesDiff[i] 
                >> Ignore(' ') >> m_waitTimeDiff[i]
                >> Ignore(' ') >> m_computeTimeDiff[i];
        }
    }
    else
    {
        is >> Ignore(':') >> m_nodeCount.first 
            >> Ignore(':') >> m_nodeCount.second;
    }
}

void MpiCall::Serialize(ostream& os) const
{
    int funcCount = m_func.size(); 
    os <<  funcCount << " MPI Functions, Count, Avg. Block Time, SD. Block Time";

    for (int i = 0; i < funcCount; ++i) 
    {
        os << '\n' 
            << DumpiFunction(m_func[i])
            << ", " << m_callCountDiff[i] 
            << ", " << m_avgBlockTimeDiff[i]
            << ", " << m_sdBlockTimeDiff[i];
    }
}

void MpiCall::Deserialize(istream& is)
{
    int funcCount = 0;
    is >> funcCount;

    m_func.resize(funcCount);
    m_callCountDiff.resize(funcCount);
    m_avgBlockTimeDiff.resize(funcCount);
    m_sdBlockTimeDiff.resize(funcCount);

    for (int i = 0; i < funcCount; i++) 
    {
        ASSERT(is.good(), "Expect to have %d lines of %s\n", funcCount, GetClassId());

        is >> Ignore('\n') >> DumpiFunction(m_func[i])
            >> Ignore(' ') >> m_callCountDiff[i]
            >> Ignore(' ') >> m_avgBlockTimeDiff[i]
            >> Ignore(' ') >> m_sdBlockTimeDiff[i];
    }
}

void Collective::Serialize(ostream& os) const
{
    if(IsEqual(m_logicSimilarity, 1.0f))
    {
        int collCount = m_collective.size();
        os << collCount << " Collective Regions, Avg. Time, SD. Time";

        for (int i = 0; i < collCount; i++) 
        {
            os << '\n' 
                << m_collective[i] << ", " 
                << m_avgTimeDiff[i] << ", " 
                << m_sdTimeDiff[i];
        }
    }
    else
    {
        os << "Collective Region Number is different: "
            << m_collCount.first << " : " << m_collCount.second;
    }
}

void Collective::Deserialize(istream& is)
{
    if(IsEqual(m_logicSimilarity, 1.0f))
    {
        int collCount = 0;
        is >> collCount >> Ignore('\n');

        m_collective.resize(collCount);
        m_avgTimeDiff.resize(collCount);
        m_sdTimeDiff.resize(collCount);

        for (int i = 0; i < collCount; i++) 
        {
            ASSERT(is.good(), "Expect to have %d lines of %s\n", collCount, GetClassId());

            is >> m_collective[i]
                >> Ignore(' ') >> m_avgTimeDiff[i] 
                >> Ignore(' ') >> m_sdTimeDiff[i];
        }
    }
    else
    {
        is >> Ignore(':') >> m_collCount.first
            >> Ignore(':') >> m_collCount.second;
    }
}

void TransmissionBytes::Serialize(ostream& os) const
{
    if(IsEqual(m_logicSimilarity, 1.0f))
    {
        int nodeCount = m_spyPlotDiff.size();

        os << nodeCount << " Src\\Dest";

        for (int i = 0; i < nodeCount; i++) 
        {
            os << ", Rank_" << i;
        }

        for (int src = 0; src < nodeCount; src++) 
        {
            os << "\nRank_" << src;

            for (int dest = 0; dest < nodeCount; dest++) 
            {
                os << ", " << m_spyPlotDiff[src][dest]; 
            }
        }
    }
    else
    {
        os <<  "Node count " << m_nodeCount.first << " : " << m_nodeCount.second;

        os << '\n' << m_unexpectedBytes.size() << " Unepected transmission";

        for (unsigned int i = 0; i < m_unexpectedBytes.size(); i++) 
        {
            os << "\nsrc[" << m_unexpectedBytes[i].first << "] : dest[" 
                << m_unexpectedBytes[i].second << "]";
        }
    }
}

void TransmissionBytes::Deserialize(istream& is)
{
    if (IsEqual(m_logicSimilarity, 1.0f))
    {
        int nodeCount = 0;
        is >> nodeCount >> Ignore('\n');

        m_spyPlotDiff.resize(nodeCount);

        for (int src = 0; src < nodeCount; src++) 
        {
            m_spyPlotDiff[src].resize(nodeCount);

            for (int dest = 0; dest < nodeCount; dest++) 
            {
                is >> Ignore(' ') >> m_spyPlotDiff[src][dest];            
            }
        }
    }
    else
    {
        is >> Ignore(' ') >> m_nodeCount.first >> Ignore(' ') >> m_nodeCount.second;

        int count;
        is >> Ignore('\n') >> count;

        m_unexpectedBytes.resize(count);

        for (int i = 0; i < count; i++) 
        {
            ASSERT(is.good(), "Expect to have %d lines of %s\n", count, GetClassId());

            is >> Ignore('[') >> m_unexpectedBytes[i].first 
                >> Ignore('[') >> m_unexpectedBytes[i].second;
        }
    }
}

} } } //namespace
