//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#ifndef SSTMAC_TOOLS_COMMON_REGISTRY_REGISTRY_INCLUDED_
#define SSTMAC_TOOLS_COMMON_REGISTRY_REGISTRY_INCLUDED_

#include <map>
#include "common/registry/policy/sequencekey.h"
#include "common/registry/policy/identicalkey.h"
#include "common/log/logger.h"

namespace sstmac { namespace tools {

////////////////////////////////////////////////////////////////////////////////
// A registry keeps one and only copy of the StoredType specified by KeyType
template<typename StubType, typename StoredType, template<typename> class KeyAllocator = IdenticalKey>
class Registry : public KeyAllocator<StubType> 
{
public:
    typedef typename KeyAllocator<StubType>::KeyType KeyType;
    typedef std::map<KeyType, StoredType> StorageType;
    typedef typename StorageType::value_type ValueType;
    typedef typename StorageType::const_iterator ConstIterator;
    typedef typename StorageType::iterator Iterator;

public:
    virtual ~Registry (){}

    bool Register(const StubType& stub, const StoredType& val){
        const KeyType& key = KeyAllocator<StubType>::AllocateKey(stub);
        if (m_storage.find(key) != m_storage.end())
        {
            OUTPUT_WARNING("Duplicate register.\n");
            return false;
        }
        m_storage[key] = val;
        return true;
    }

    bool Unregister(const StubType& stub){
        typename StorageType::iterator iter = m_storage.find(GetKey(stub));
        if (iter == m_storage.end())
        {
            OUTPUT_WARNING("Already unregistered..\n");
            return false;
        }
        FreeKey(stub);
        m_storage.erase(iter);
        return true;
    }

    bool IsRegistered(const StubType& stub) const {
        return m_storage.find(GetKey(stub)) != m_storage.end();
    }

    StoredType* Get(const StubType& stub){
        typename StorageType::iterator iter = m_storage.find(KeyAllocator<StubType>::GetKey(stub));
        if (iter == m_storage.end())
        {
            return NULL;
        }
        return &(iter->second);
    }

    const StoredType* Get(const StubType& stub) const {
        typename StorageType::const_iterator iter = m_storage.find(KeyAllocator<StubType>::GetKey(stub));
        if (iter == m_storage.end())
        {
            return NULL;
        }
        return &(iter->second);
    }

    int Size() const {
        return m_storage.size();
    }

    ConstIterator Begin() const {
        return m_storage.begin();
    }

    ConstIterator End() const {
        return m_storage.end();
    }

    Iterator Begin() {
        return m_storage.begin();
    }

    Iterator End() {
        return m_storage.end();
    }

private:
    StorageType m_storage;
};

} } //namespace

#endif //SSTMAC_TOOLS_COMMON_REGISTRY_REGISTRY_INCLUDED_
