//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include <sstream>
#include "common/report/reportcomparison.h"
#include "common/log/logger.h"
#include "common/factory/factory.h"
#include "common/comparison/comparison.h"
#include "common/comparison/criterion/criteriondatabase.h"
#include "common/utilities/serializeadapter.h"

namespace sstmac { namespace tools {

const char* ReportComparison::ROOT_TAG = "DUMPI_Comparison_Comparison";
const char* ReportComparison::META_TAG[2] = {"ReferenceMeta", "SimulationMeta"};
const char* ReportComparison::COMPARISON_TAG = "Comparison";
const char* ReportComparison::COMPARISON_ATTR_COUNT = "count";
const char* ReportComparison::CRITERION_TAG = "Criterion";
const char* ReportComparison::CRITERION_ATTR_NAME = "name";
const char* ReportComparison::CRITERION_ATTR_LOGICSIM = "similarity_logic";
const char* ReportComparison::CRITERION_ATTR_NUMERICSIM = "similarity_numeric";

ReportComparison::ReportComparison(Comparison& comparison)
    : m_comparison(comparison)
{}

bool ReportComparison::ReadImpl(std::istream& is)
{
    is >> XmlHeader();

    StartTag rootStart;
    is >> rootStart; 

    if(rootStart.m_tag.compare(ROOT_TAG))
    {
        return false;
    }

    //Meta data
    for (int i = 0; i < 2; i++) 
    {
        StartTag metaStart;
        is >> metaStart;

        if(metaStart.m_tag.compare(META_TAG[i]))
        {
            return false;
        }

        m_comparison.GetMetaData(i) = metaStart.m_attr;

        EndTag metaEnd;
        is >> metaEnd;
    }
      
    StartTag comparisonStart;
    is >> comparisonStart;
    std::string* comparisonCountStr = comparisonStart.m_attr.Get(COMPARISON_ATTR_COUNT);

    if(!comparisonCountStr || comparisonStart.m_tag.compare(COMPARISON_TAG))
    {
        return false;
    }

    stringstream ss(*comparisonCountStr);
    int comparisonCount;
    ss >> comparisonCount;

    for (int i = 0; i < comparisonCount; i++) 
    {
        StartTag entryStart;
        is >> entryStart;

        std::string* entryName = entryStart.m_attr.Get(CRITERION_ATTR_NAME);
        std::string* logicSimilarity= entryStart.m_attr.Get(CRITERION_ATTR_LOGICSIM);
        std::string* numericSimilarity = entryStart.m_attr.Get(CRITERION_ATTR_NUMERICSIM);

        if(!entryName || !logicSimilarity || !numericSimilarity)
        {
            OUTPUT_WARNING("Failed to read entry\n");
            return false;
        }

        CriterionDataBase::Ptr newData(Factory<CriterionDataBase>::Singleton().Create(*entryName));

        if(!newData)
        {
            OUTPUT_WARNING("Failed to read entry\n");
            return false;
        }

        stringstream logicSimStream (*logicSimilarity);
        stringstream numericSimStream (*numericSimilarity);

        logicSimStream >> newData->m_logicSimilarity;
        numericSimStream >> newData->m_numericSimilarity;

        is >> StartData() >> (*newData) >> EndData(); 
        m_comparison.Register(newData->GetClassId(), newData);

        //std::cout << *newData << std::endl;

        EndTag entryEnd;
        is >> entryEnd;
    }

    EndTag comparisonEnd; 
    is >> comparisonEnd;

    EndTag rootEnd;
    is >> rootEnd;

    return true;
}

bool ReportComparison::WriteImpl(std::ostream& os) const
{
    os << XmlHeader();

    os << StartTag(ROOT_TAG);

    for (int i = 0; i < 2; i++) 
    {
        os << StartTag(META_TAG[i], m_comparison.GetMetaData(i));
        os << EndTag(META_TAG[i]);
    }

    StartTag::AttributeHolder attr;
    stringstream ss;
    ss << m_comparison.Size();
    attr.Register(COMPARISON_ATTR_COUNT, ss.str());
    os << StartTag(COMPARISON_TAG, attr);

    for (Comparison::Iterator iter = m_comparison.Begin(); iter != m_comparison.End(); ++iter) 
    {
        StartTag::AttributeHolder attr;
        std::stringstream logicSimStream;
        logicSimStream << iter->second->m_logicSimilarity;
        std::stringstream numericSimStream;
        numericSimStream << iter->second->m_numericSimilarity;

        attr.Register(CRITERION_ATTR_NAME, iter->second->GetClassId());
        attr.Register(CRITERION_ATTR_LOGICSIM, logicSimStream.str());
        attr.Register(CRITERION_ATTR_NUMERICSIM, numericSimStream.str());

        os << StartTag(CRITERION_TAG, attr);
        os << StartData() << (*iter->second) << EndData();
        os << EndTag(CRITERION_TAG);
    }

    os << EndTag(COMPARISON_TAG);
    os << EndTag(ROOT_TAG);

    return true;
}

} } //namespace
