//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include "common/stats/entry/entrydataimpl.h"
#include "common/factory/factory.h"
#include "common/log/logger.h"
#include "common/utilities/serializeadapter.h"
#include "common/utilities/dumpihelper.h"

FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, Overall);
FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, ByNode);
FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, MpiCall);
FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, Collective);
FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, TransmissionTime);
FACTORY_REGISTER_CLASS(sstmac::tools, EntryDataBase, sstmac::tools::statsentry, TransmissionBytes);

namespace sstmac { namespace tools { namespace statsentry {

///////////////////////////////////////////////////////////////////////////////
//
void Overall::Serialize(ostream& os) const
{
    os << "Transmit Bytes, Wait Time, Compute Time\n";
    os << m_bytes << ", "
        << m_waitTime << ", "
        << m_computeTime;
}

void Overall::Deserialize(istream& is)
{
    is >> Ignore('\n') >> m_bytes 
        >> Ignore(' ') >> m_waitTime 
        >> Ignore(' ') >> m_computeTime;
}

////////////////////////////////////////////////////////////////////////////////
//
void ByNode::Serialize(ostream& os) const
{
    int nodeCount = m_bytes.size();

    os << nodeCount << " Nodes, Transmit Bytes, Wait Time, Compute Time, In Degree, Out Degree";

    for (int i = 0; i < nodeCount; i++) 
    {
        os << "\nRank_" << i 
            << ", " << m_bytes[i]
            << ", " << m_waitTime[i]
            << ", " << m_computeTime[i]
            << ", " << m_inDegree[i]
            << ", " << m_outDegree[i];
    }

    os << "\nAverage" 
        << ", " << m_avgBytes
        << ", " << m_avgWaitTime
        << ", " << m_avgComputeTime
        << ", " << m_avgInDegree
        << ", " << m_avgOutDegree;

    os << "\nDeviation" 
        << ", " << m_sdBytes
        << ", " << m_sdWaitTime
        << ", " << m_sdComputeTime
        << ", " << m_sdInDegree
        << ", " << m_sdOutDegree;
}

void ByNode::Deserialize(istream& is)
{
    int nodeCount = 0;
    is >> nodeCount >> Ignore('\n');
    
    m_bytes.resize(nodeCount);
    m_waitTime.resize(nodeCount);
    m_computeTime.resize(nodeCount);
    m_inDegree.resize(nodeCount);
    m_outDegree.resize(nodeCount);

    for (int i = 0; i < nodeCount; i++) 
    {
        ASSERT(is.good(), "Expect to have %d lines of %s\n", nodeCount, GetClassId());
        
        is >> Ignore(' ') >> m_bytes[i] 
            >> Ignore(' ') >> m_waitTime[i]
            >> Ignore(' ') >> m_computeTime[i]
            >> Ignore(' ') >> m_inDegree[i]
            >> Ignore(' ') >> m_outDegree[i];
    }

    is >> Ignore(' ') >> m_avgBytes
        >> Ignore(' ') >> m_avgWaitTime
        >> Ignore(' ') >> m_avgComputeTime
        >> Ignore(' ') >> m_avgInDegree
        >> Ignore(' ') >> m_avgOutDegree;

    is >> Ignore(' ') >> m_sdBytes
        >> Ignore(' ') >> m_sdWaitTime
        >> Ignore(' ') >> m_sdComputeTime
        >> Ignore(' ') >> m_sdInDegree
        >> Ignore(' ') >> m_sdOutDegree;
}

////////////////////////////////////////////////////////////////////////////////
//
void MpiCall::Serialize(ostream& os) const
{
    int funcCount = 0;
    int nodeCount = m_time[0].size();

    for (int i = 0; i < Dumpi::GetCount(); ++i) 
    {
        if(m_count[i] > 0)
        {
            funcCount++; 
        }
    }

    os <<  funcCount << " Functions :" << nodeCount << " Nodes, Count, Avg. Block Time, SD. Block Time";

    for (int i = 0; i < Dumpi::GetCount(); ++i) 
    {
        if(m_count[i] > 0)
        {
            os << '\n' 
                << DumpiFunction(i)
                << ", " << m_count[i] 
                << ", " << m_avgBlockTime[i]
                << ", " << m_sdBlockTime[i];

            for (int nodeRank = 0; nodeRank < nodeCount; nodeRank++) 
            {
                os << ", " << m_time[i][nodeRank]; 
            }
        }
    }    
}

void MpiCall::Deserialize(istream& is)
{
    int totalCount = Dumpi::GetCount();
    int funcCount = 0;
    int nodeCount = 0;

    is >> funcCount >> Ignore(':') >> nodeCount;

    m_count.resize(totalCount);
    m_avgBlockTime.resize(totalCount);
    m_sdBlockTime.resize(totalCount);
    m_time.resize(totalCount);

    for (unsigned int i = 0; i < m_time.size(); i++) 
    {
        m_time[i].resize(nodeCount);
    }

    for (int i = 0; i < funcCount; i++) 
    {
        ASSERT(is.good(), "Expect to have %d lines of %s\n", funcCount, GetClassId());

        string funcName;
        is >>Ignore('\n') >> funcName;
        int funcId = Dumpi::GetEnum(funcName);

        is >> Ignore(' ') >> m_count[funcId]
            >> Ignore(' ') >> m_avgBlockTime[funcId]
            >> Ignore(' ') >> m_sdBlockTime[funcId];

        for (int nodeRank = 0; nodeRank < nodeCount; nodeRank++) 
        {
            is >> Ignore(' ') >> m_time[funcId][nodeRank];            
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//
Collective::CollRegion::CollRegion()
    : m_seqId(0)
    , m_beginFunc(0)
    , m_endFunc(0)
{}

Collective::CollRegion::CollRegion(int seqId, int beginFunc, int endFunc)
    : m_seqId(seqId)
    , m_beginFunc(beginFunc)
    , m_endFunc(endFunc)
{}

void Collective::CollRegion::Serialize(ostream& os) const
{
    os << m_seqId << ". " 
        << DumpiFunction(m_beginFunc) << " - " 
        << DumpiFunction(m_endFunc) << ' '; 
}

void Collective::CollRegion::Deserialize(istream& is)
{
    is >> m_seqId >> Ignore(' ') 
        >> DumpiFunction(m_beginFunc) >> Ignore(3) 
        >> DumpiFunction(m_endFunc) >> Ignore(1);
}

bool Collective::CollRegion::operator==(const CollRegion& rhs) const
{
    return m_seqId == rhs.m_seqId 
        && m_beginFunc == rhs.m_beginFunc 
        && m_endFunc == rhs.m_endFunc;
}

void Collective::Serialize(ostream& os) const
{
    int collCount = m_collective.size();
    int nodeCount = m_time[0].size();

    os << collCount << " Phases :" << nodeCount << " Nodes" << ", Avg. Time, SD. Time";

    for (int i = 0; i < nodeCount; i++) 
    {
        os << ", Rank_" << i;
    }

    for (int i = 0; i < collCount; i++) 
    {
        os << '\n' << m_collective[i]
           << ", " << m_avgTime[i]
           << ", " << m_sdTime[i];

        for (int nodeRank = 0; nodeRank < nodeCount; nodeRank++) 
        {
            os << ", " << m_time[i][nodeRank]; 
        }
    }
}

void Collective::Deserialize(istream& is)
{
    int collCount = 0;
    int nodeCount = 0;
    is >> collCount >> Ignore(':') >> nodeCount;

    m_collective.resize(collCount);

    for (int i = 0; i < collCount; i++) 
    {
        ASSERT(is.good(), "Expect to have %d lines of %s\n", collCount, GetClassId());

        is >> Ignore('\n') >> m_collective[i]
           >> Ignore(' ') >> m_avgTime[i] 
           >> Ignore(' ') >> m_sdTime[i];

        m_time.resize(nodeCount);

        for (int nodeRank = 0; nodeRank < nodeCount; nodeRank++) 
        {
            is >> Ignore(' ') >> m_time[i][nodeRank];            
        }
    }
}

////////////////////////////////////////////////////////////////////////////////
//
void TransmissionTime::Serialize(ostream& os) const
{
    int nodeCount = m_time.size();

    os << nodeCount << " Src\\Dest (Time)";

    for (int i = 0; i < nodeCount; i++) 
    {
        os << ", Rank_" << i;
    }

    for (int src = 0; src < nodeCount; src++) 
    {
        os << "\nRank_" << src;

        for (int dest = 0; dest < nodeCount; dest++) 
        {
            os << ", " << m_time[src][dest]; 
        }

        //os << flush;
    }


}

void TransmissionTime::Deserialize(istream& is)
{
    int nodeCount = 0;
    is >> nodeCount >> Ignore('\n');

    m_time.resize(nodeCount);

    for (int src = 0; src < nodeCount; src++) 
    {
        m_time[src].resize(nodeCount);
        
        for (int dest = 0; dest < nodeCount; dest++) 
        {
            is >> Ignore(' ') >> m_time[src][dest];            
        }
    }


}


void TransmissionBytes::Serialize(ostream& os) const
{
    int nodeCount = m_traffic.size();

    os << nodeCount << " Src\\Dest (Bytes)";

    for (int i = 0; i < nodeCount; i++)
    {
        os << ", Rank_" << i;
    }

    for (int src = 0; src < nodeCount; src++)
    {
        os << "\nRank_" << src;

        for (int dest = 0; dest < nodeCount; dest++)
        {
            os << ", " << m_traffic[src][dest];
        }
    }
}

void TransmissionBytes::Deserialize(istream& is)
{
    int nodeCount = 0;
    is >> nodeCount >> Ignore('\n');

    m_traffic.resize(nodeCount);

    for (int src = 0; src < nodeCount; src++)
    {
        m_traffic[src].resize(nodeCount);

        for (int dest = 0; dest < nodeCount; dest++)
        {
            is >> Ignore(' ') >> m_traffic[src][dest];
        }
    }
}

} } } //namespace
