//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include <climits>
#include <vector>
#include "common/utilities/serializeadapter.h"
#include "common/utilities/dumpihelper.h"

namespace sstmac { namespace tools { 

////////////////////////////////////////////////////////////////////////////////
// Ignore
Ignore::Ignore(int skipChar)
    : m_skipChar(skipChar)
    , m_delim(0)
{}

Ignore::Ignore(char delim)
    : m_skipChar(0)
    , m_delim(delim)
{}

void Ignore::Serialize(ostream& os) const
{
    for (int i = 0; i < m_skipChar; i++) 
    {
        os.put(' ');
    }
}

void Ignore::Deserialize(istream& is) const
{
    if (m_delim == 0)
    {
        is.ignore(m_skipChar);
    }
    else
    {
        is.ignore(INT_MAX, m_delim);
    }
}

////////////////////////////////////////////////////////////////////////////////
// Percentage
Percentage::Percentage()
    : m_val(0.0)
{}

Percentage::Percentage(double val)
    : m_val(val)
{}

Percentage::operator double() const
{
    return m_val;
}

void Percentage::Serialize(ostream& os) const
{
    os << m_val * 100.0 << '%';
}

void Percentage::Deserialize(istream& is)
{
    is >> m_val >> Ignore(1);
    m_val /= 100.0;
}

////////////////////////////////////////////////////////////////////////////////
// DumpiFunction
DumpiFunction::DumpiFunction(int& func)
    : m_funcVal(func)
    , m_funcRef(func)
{}

DumpiFunction::DumpiFunction(const int& func)
    : m_funcVal(func)
    , m_funcRef(m_funcVal)
{}

void DumpiFunction::Serialize(ostream& os) const
{
    os << Dumpi::GetName(m_funcRef);     
}

void DumpiFunction::Deserialize(istream& is) const
{
    string funcName;
    is >> funcName;

    m_funcRef = Dumpi::GetEnum(funcName);
}

////////////////////////////////////////////////////////////////////////////////
// StartTag
StartTag::StartTag()
{}

StartTag::StartTag(const string& tag)
    : m_tag(tag)
{}

StartTag::StartTag(const string& tag, const AttributeHolder& attr)
    : m_tag(tag)
    , m_attr(attr)
{}

void StartTag::Serialize(ostream& os) const
{
    os << "\n<" << m_tag;

    for (AttributeHolder::ConstIterator iter = m_attr.Begin(); iter != m_attr.End(); ++iter) 
    {
        os << ' ' << iter->first << '=' << '"' << iter->second << '"';
    }
 
    os << '>';
}

void StartTag::Deserialize(istream& is)
{
    string line;
    getline(is, line);

    //Remove prefixing and trailing bracket
    line.erase(line.end() - 1);
    line.erase(line.begin());

    //split the string by space
    vector<string> substring;
    string::size_type delimPos = line.find_first_of(' ');

    while(delimPos != string::npos)
    {
        substring.push_back(line.substr(0, delimPos));
        line = line.substr(delimPos + 1);
        delimPos = line.find_first_of(' ');
    }

    substring.push_back(line);

    //tag is the first substring
    m_tag = substring.front();

    //handle attributes
    for (unsigned int i = 1; i < substring.size(); i++) 
    {
        delimPos = substring[i].find_first_of('=');
        ASSERT(delimPos != string::npos, "Invalid attribute format");

        string key = substring[i].substr(0, delimPos);
        string val = substring[i].substr(delimPos + 1);

        val.erase(val.end() - 1);
        val.erase(val.begin());

        m_attr.Register(key, val);
    }
}

////////////////////////////////////////////////////////////////////////////////
// EndTag
EndTag::EndTag()
{}

EndTag::EndTag(const string& tag)
    : m_tag(tag)
{}

void EndTag::Serialize(ostream& os) const
{
    os << "\n</" << m_tag << '>'; 
}

void EndTag::Deserialize(istream& is)
{
    getline(is, m_tag);

    m_tag.erase(0, 2);
    m_tag.erase(m_tag.end() - 1);
}

////////////////////////////////////////////////////////////////////////////////
// CDATA
void StartData::Serialize(ostream& os) const
{
    os << "\n<![CDATA[";
}

void StartData::Deserialize(istream& is) const
{
    is.ignore(9);
}

void EndData::Serialize(ostream& os) const
{
    os << "\n]]>";
}

void EndData::Deserialize(istream& is) const
{
    is.ignore(5);
}

////////////////////////////////////////////////////////////////////////////////
// XmlHeader
void XmlHeader::Serialize(ostream& os) const
{
    os << "<?xml version=\"1.0\" ?>";
}

void XmlHeader::Deserialize(istream& is) const
{
    is.ignore(INT_MAX, '\n');
}

} } //namespace
