//------------------------------------------------------------------------------
// 
//     SST/Macro tools set
//
//------------------------------------------------------------------------------

#include <wx/sizer.h>
#include <wx/filedlg.h>
#include <wx/filename.h>
#include <wx/menu.h>
#include <wx/log.h>
#include "visualizer/ui/plotframe.h"

namespace sstmac { namespace tools { namespace visualizer {

////////////////////////////////////////////////////////////////////////////////
// 
PlotWindow::PlotWindow(wxWindow* parent, PlotBase::Ptr plot)
    : wxPLplotwindow(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize, wxBORDER_SUNKEN,
#if wxUSE_GRAPHICS_CONTEXT
        wxPLPLOT_BACKEND_GC | wxPLPLOT_DRAW_TEXT )
#else
        wxPLPLOT_BACKEND_AGG | wxPLPLOT_DRAW_TEXT )
#endif
    , m_plot(plot)
{
    SetMinSize(ClientToWindowSize(wxSize(800, 600)));

    //Set background to white
    m_stream->scol0( 0, 255, 255, 255 );
    m_stream->scol0( 15, 0, 0, 0 );
}

wxString PlotWindow::GetPlotName() const
{
    return wxString(m_plot->GetClassId(), wxConvUTF8); 
}

void PlotWindow::PlotToDC()
{
    Plot(m_stream);
}

bool PlotWindow::PlotToFile(const wxString& dev, const wxString& filePath)
{
    plstream pls;

    pls.sfnam(filePath.mb_str());
    pls.sdev(dev.mb_str());

    pls.scol0(0, 255, 255, 255);
    pls.scol0(15, 0, 0, 0);

    pls.init();

    Plot(&pls);

    return true;
}

void PlotWindow::Plot(plstream* pls)
{
    m_plot->Plot(pls);
}

////////////////////////////////////////////////////////////////////////////////
// 
PlotScrollWindow::PlotScrollWindow(wxWindow* parent)
    : wxScrolledWindow(parent, wxID_ANY, wxDefaultPosition, wxDefaultSize) 
{
    PlotWindow* temp = new PlotWindow(this, PlotBase::Ptr()); 
    wxSize minSize = ClientToWindowSize(temp->GetMinSize());
    temp->Destroy();

    SetMinSize(minSize);
    wxBoxSizer* scrollSizer = new wxBoxSizer(wxVERTICAL);
    SetScrollRate(10, 10);
    SetSizer(scrollSizer);
}

////////////////////////////////////////////////////////////////////////////////
// 
BEGIN_EVENT_TABLE(PlotFrame, wxDocChildFrame)
    EVT_MENU(PLOT_EXPORT, PlotFrame::OnMenuExport)
END_EVENT_TABLE()

PlotFrame::PlotFrame(wxDocument* document, wxView* view, wxDocParentFrame* parent, const wxString& title)
    : wxDocChildFrame(document, view, parent, wxID_ANY, title, wxDefaultPosition, wxDefaultSize)
{
    CreateMenuBar();

    m_scrollWindow = new PlotScrollWindow(this);

    wxBoxSizer* topSizer = new wxBoxSizer(wxVERTICAL);
    topSizer->Add(m_scrollWindow, 1, wxEXPAND, 0);
    SetSizerAndFit(topSizer);
}

void PlotFrame::AddPlotWindow(PlotBase::Ptr plot)
{
    PlotWindow* plotWindow = new PlotWindow(m_scrollWindow, plot);

    m_scrollWindow->GetSizer()->Add(plotWindow, 1, wxTOP, 1);
    m_scrollWindow->FitInside(); // Need to call this to resize the virtual area in scroll window

    plotWindow->PlotToDC();

    m_plotWindows.push_back(plotWindow);
}

void PlotFrame::CreateMenuBar()
{
    wxMenuBar* menuBar = new wxMenuBar();
    SetMenuBar(menuBar);

    wxMenu* fileMenu = new wxMenu();
    menuBar->Append(fileMenu, wxT("&File"));
    fileMenu->Append(wxID_OPEN, wxT("&Open..."));
    fileMenu->Append(wxID_CLOSE, wxT("&Close"));
    fileMenu->Append(PLOT_EXPORT, wxT("&Export..."));
    fileMenu->AppendSeparator();
    fileMenu->Append(wxID_EXIT, wxT("&Exit"));

    wxMenu* aboutMenu = new wxMenu();
    menuBar->Append(aboutMenu, wxT("&Help"));
    aboutMenu->Append(wxID_ABOUT, wxT("&About..."));
}

void PlotFrame::OnMenuExport(wxCommandEvent& WXUNUSED(cmd))
{
    static const wxString WILDCARD = _("PNG files (*.png)|*.png|PostScript files (*.ps)|*.ps|Portable Document Format (*.pdf)|*.pdf|Scalable Vector Graphics (*.svg)|*.svg|");
    static const wxString FILE_SUFFIX[] = {_(".png"), _(".ps"), _(".pdf"), _(".svg")};
    static const wxString FILE_EXT[] = {_("png"), _("ps"), _("pdf"), _("svg")};
    static const wxString PLOT_DEV[] = {wxT("pngcairo"), wxT("pscairo"), wxT("pdfcairo"), wxT("svgcairo")};

    wxFileName originalName = GetDocument()->GetFilename();

    wxFileDialog *saveDialog= new wxFileDialog(this, wxT("Export plot"), originalName.GetPath(), originalName.GetName(), WILDCARD, wxFD_SAVE | wxFD_OVERWRITE_PROMPT);

    if ( saveDialog->ShowModal() == wxID_OK )
    {
        int filterIndex = saveDialog->GetFilterIndex();
        const wxString& inputName = saveDialog->GetFilename();
        wxString baseName;

        if(!inputName.EndsWith(FILE_SUFFIX[filterIndex].c_str(), &baseName))
        {
            baseName = inputName;
        }

        for (unsigned int i = 0; i < m_plotWindows.size(); i++) 
        {
            wxString finalName = baseName ;
            finalName.Append(wxT("-"));
            finalName.Append(m_plotWindows[i]->GetPlotName());

            wxFileName fullPath(saveDialog->GetDirectory(), finalName, FILE_EXT[filterIndex]);

            bool saveOk = m_plotWindows[i]->PlotToFile(PLOT_DEV[filterIndex], fullPath.GetFullPath());

            wxLogMessage(wxT("Plot exported to %s %s"), fullPath.GetFullPath().c_str(), saveOk ? wxT("done") : wxT("failed"));
        }
    }

    saveDialog->Close();     
}

} } } //namespace
